Package: posterior
Title: Tools for Working with Posterior Distributions
Version: 1.7.0
Date: 2026-04-01
Authors@R: c(person("Paul-Christian", "Bürkner", email = "paul.buerkner@gmail.com", role = c("aut", "cre")),
             person("Jonah", "Gabry", email = "jgabry@gmail.com", role = c("aut")),
             person("Matthew", "Kay", email = "mjskay@northwestern.edu", role = c("aut")),
             person("Aki", "Vehtari", email = "Aki.Vehtari@aalto.fi", role = c("aut")),
             person("Måns", "Magnusson", role = c("ctb")),
             person("Rok", "Češnovar", role = c("ctb")),
             person("Ben", "Lambert", role = c("ctb")),
             person("Ozan", "Adıgüzel", role = c("ctb")), 
             person("Jacob", "Socolar", role = c("ctb")),
             person("Noa", "Kallioinen", role = c("ctb")),
             person("Teemu", "Säilynoja", role = c("ctb")))
Description: Provides useful tools for both users and developers of packages 
  for fitting Bayesian models or working with output from Bayesian models.
  The primary goals of the package are to: 
  (a) Efficiently convert between many different useful formats of
  draws (samples) from posterior or prior distributions.
  (b) Provide consistent methods for operations commonly performed on draws, 
  for example, subsetting, binding, or mutating draws.
  (c) Provide various summaries of draws in convenient formats.
  (d) Provide lightweight implementations of state of the art posterior 
  inference diagnostics. References: Vehtari et al. (2021) 
  <doi:10.1214/20-BA1221>.
Depends: R (>= 3.2.0)
Imports: methods, abind, checkmate, rlang (>= 1.0.6), stats, tibble (>=
        3.1.0), vctrs (>= 0.5.0), tensorA, pillar, distributional,
        parallel, matrixStats
Suggests: testthat (>= 2.1.0), caret (>= 6.0-84), gbm (>= 2.1.8),
        randomForest (>= 4.6.14), e1071 (>= 1.7-3), dplyr, tidyr,
        knitr, ggplot2, ggdist, rmarkdown
Config/Needs/website: stan-dev/pkgdown-config
License: BSD_3_clause + file LICENSE
Encoding: UTF-8
URL: https://mc-stan.org/posterior/, https://discourse.mc-stan.org/
BugReports: https://github.com/stan-dev/posterior/issues
RoxygenNote: 7.3.3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-01 11:52:27 UTC; wa.bocktif1
Author: Paul-Christian Bürkner [aut, cre],
  Jonah Gabry [aut],
  Matthew Kay [aut],
  Aki Vehtari [aut],
  Måns Magnusson [ctb],
  Rok Češnovar [ctb],
  Ben Lambert [ctb],
  Ozan Adıgüzel [ctb],
  Jacob Socolar [ctb],
  Noa Kallioinen [ctb],
  Teemu Säilynoja [ctb]
Maintainer: Paul-Christian Bürkner <paul.buerkner@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-01 13:00:02 UTC
Built: R 4.5.2; ; 2026-04-01 17:18:39 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("[",draws_array)
S3method("[",draws_df)
S3method("[",draws_list)
S3method("[",draws_matrix)
S3method("[",draws_rvars)
S3method("[",rvar)
S3method("[<-",rvar)
S3method("[[",rvar)
S3method("[[<-",rvar)
S3method("dim<-",rvar)
S3method("dimnames<-",rvar)
S3method("levels<-",rvar)
S3method("names<-",rvar)
S3method("variables<-",draws_array)
S3method("variables<-",draws_df)
S3method("variables<-",draws_list)
S3method("variables<-",draws_matrix)
S3method("variables<-",draws_rvars)
S3method(.subset_draws,draws_array)
S3method(.subset_draws,draws_df)
S3method(.subset_draws,draws_list)
S3method(.subset_draws,draws_matrix)
S3method(.subset_draws,draws_rvars)
S3method(Math,rvar)
S3method(Math,rvar_factor)
S3method(Ops,rvar)
S3method(Ops,rvar_factor)
S3method(Ops,rvar_ordered)
S3method(Pr,default)
S3method(Pr,logical)
S3method(Pr,rvar)
S3method(Summary,rvar)
S3method(all,rvar)
S3method(all.equal,rvar)
S3method(any,rvar)
S3method(anyDuplicated,rvar)
S3method(anyDuplicated,rvar_factor)
S3method(anyNA,rvar)
S3method(aperm,rvar)
S3method(as.data.frame,rvar)
S3method(as.list,rvar)
S3method(as.vector,rvar)
S3method(as_draws,default)
S3method(as_draws,draws)
S3method(as_draws,rvar)
S3method(as_draws_array,default)
S3method(as_draws_array,draws_array)
S3method(as_draws_array,draws_df)
S3method(as_draws_array,draws_list)
S3method(as_draws_array,draws_matrix)
S3method(as_draws_array,draws_rvars)
S3method(as_draws_array,mcmc)
S3method(as_draws_array,mcmc.list)
S3method(as_draws_df,data.frame)
S3method(as_draws_df,default)
S3method(as_draws_df,draws_array)
S3method(as_draws_df,draws_df)
S3method(as_draws_df,draws_list)
S3method(as_draws_df,draws_matrix)
S3method(as_draws_df,draws_rvars)
S3method(as_draws_df,mcmc)
S3method(as_draws_df,mcmc.list)
S3method(as_draws_list,default)
S3method(as_draws_list,draws_array)
S3method(as_draws_list,draws_df)
S3method(as_draws_list,draws_list)
S3method(as_draws_list,draws_matrix)
S3method(as_draws_list,draws_rvars)
S3method(as_draws_list,mcmc)
S3method(as_draws_list,mcmc.list)
S3method(as_draws_matrix,default)
S3method(as_draws_matrix,draws_array)
S3method(as_draws_matrix,draws_df)
S3method(as_draws_matrix,draws_list)
S3method(as_draws_matrix,draws_matrix)
S3method(as_draws_matrix,draws_rvars)
S3method(as_draws_matrix,mcmc)
S3method(as_draws_matrix,mcmc.list)
S3method(as_draws_rvars,default)
S3method(as_draws_rvars,draws_array)
S3method(as_draws_rvars,draws_df)
S3method(as_draws_rvars,draws_list)
S3method(as_draws_rvars,draws_matrix)
S3method(as_draws_rvars,draws_rvars)
S3method(as_draws_rvars,list)
S3method(as_draws_rvars,mcmc)
S3method(as_draws_rvars,mcmc.list)
S3method(as_draws_rvars,rvar)
S3method(as_tibble,rvar)
S3method(bind_draws,"NULL")
S3method(bind_draws,draws_array)
S3method(bind_draws,draws_df)
S3method(bind_draws,draws_list)
S3method(bind_draws,draws_matrix)
S3method(bind_draws,draws_rvars)
S3method(bind_draws,list)
S3method(broadcast_and_bind_rvars,rvar)
S3method(broadcast_and_bind_rvars,rvar_factor)
S3method(c,rvar)
S3method(cbind,rvar)
S3method(cdf,rvar)
S3method(cdf,rvar_factor)
S3method(cdf,rvar_ordered)
S3method(chain_ids,"NULL")
S3method(chain_ids,draws_array)
S3method(chain_ids,draws_df)
S3method(chain_ids,draws_list)
S3method(chain_ids,draws_matrix)
S3method(chain_ids,draws_rvars)
S3method(chol,rvar)
S3method(density,rvar)
S3method(density,rvar_factor)
S3method(dim,rvar)
S3method(dimnames,rvar)
S3method(dissent,default)
S3method(dissent,rvar)
S3method(draw_ids,"NULL")
S3method(draw_ids,draws_array)
S3method(draw_ids,draws_df)
S3method(draw_ids,draws_list)
S3method(draw_ids,draws_matrix)
S3method(draw_ids,draws_rvars)
S3method(draw_ids,rvar)
S3method(duplicated,rvar)
S3method(duplicated,rvar_factor)
S3method(entropy,default)
S3method(entropy,rvar)
S3method(ess_basic,default)
S3method(ess_basic,rvar)
S3method(ess_bulk,default)
S3method(ess_bulk,rvar)
S3method(ess_mean,default)
S3method(ess_mean,rvar)
S3method(ess_quantile,default)
S3method(ess_quantile,rvar)
S3method(ess_sd,default)
S3method(ess_sd,rvar)
S3method(ess_tail,default)
S3method(ess_tail,rvar)
S3method(extract_list_of_variable_arrays,default)
S3method(extract_list_of_variable_arrays,draws)
S3method(extract_variable,default)
S3method(extract_variable,draws)
S3method(extract_variable,draws_df)
S3method(extract_variable,draws_list)
S3method(extract_variable,draws_rvars)
S3method(extract_variable_array,default)
S3method(extract_variable_array,draws)
S3method(extract_variable_matrix,default)
S3method(extract_variable_matrix,draws)
S3method(extract_variable_matrix,draws_df)
S3method(extract_variable_matrix,draws_list)
S3method(extract_variable_matrix,draws_rvars)
S3method(format,rvar)
S3method(format_glimpse,rvar)
S3method(get_rvar_class,default)
S3method(get_rvar_class,factor)
S3method(get_rvar_class,ordered)
S3method(is.array,rvar)
S3method(is.finite,rvar)
S3method(is.infinite,rvar)
S3method(is.matrix,rvar)
S3method(is.na,rvar)
S3method(is.nan,rvar)
S3method(iteration_ids,"NULL")
S3method(iteration_ids,draws_array)
S3method(iteration_ids,draws_df)
S3method(iteration_ids,draws_list)
S3method(iteration_ids,draws_matrix)
S3method(iteration_ids,draws_rvars)
S3method(iteration_ids,rvar)
S3method(length,rvar)
S3method(levels,rvar)
S3method(mad,default)
S3method(mad,rvar)
S3method(mad,rvar_ordered)
S3method(match,default)
S3method(match,rvar)
S3method(max,rvar)
S3method(mcse_mean,default)
S3method(mcse_mean,rvar)
S3method(mcse_quantile,default)
S3method(mcse_quantile,rvar)
S3method(mcse_sd,default)
S3method(mcse_sd,rvar)
S3method(mean,rvar)
S3method(median,rvar)
S3method(merge_chains,draws_array)
S3method(merge_chains,draws_df)
S3method(merge_chains,draws_list)
S3method(merge_chains,draws_matrix)
S3method(merge_chains,draws_rvars)
S3method(merge_chains,rvar)
S3method(min,rvar)
S3method(modal_category,default)
S3method(modal_category,rvar)
S3method(mutate_variables,draws_array)
S3method(mutate_variables,draws_df)
S3method(mutate_variables,draws_list)
S3method(mutate_variables,draws_matrix)
S3method(mutate_variables,draws_rvars)
S3method(names,rvar)
S3method(nchains,"NULL")
S3method(nchains,draws_array)
S3method(nchains,draws_df)
S3method(nchains,draws_list)
S3method(nchains,draws_matrix)
S3method(nchains,draws_rvars)
S3method(nchains,rvar)
S3method(ndraws,"NULL")
S3method(ndraws,draws_array)
S3method(ndraws,draws_df)
S3method(ndraws,draws_list)
S3method(ndraws,draws_matrix)
S3method(ndraws,draws_rvars)
S3method(ndraws,rvar)
S3method(niterations,"NULL")
S3method(niterations,draws_array)
S3method(niterations,draws_df)
S3method(niterations,draws_list)
S3method(niterations,draws_matrix)
S3method(niterations,draws_rvars)
S3method(niterations,rvar)
S3method(nvariables,"NULL")
S3method(nvariables,draws)
S3method(order_draws,draws_array)
S3method(order_draws,draws_df)
S3method(order_draws,draws_list)
S3method(order_draws,draws_matrix)
S3method(order_draws,draws_rvars)
S3method(order_draws,rvar)
S3method(pareto_convergence_rate,default)
S3method(pareto_convergence_rate,rvar)
S3method(pareto_diags,default)
S3method(pareto_diags,rvar)
S3method(pareto_khat,default)
S3method(pareto_khat,rvar)
S3method(pareto_khat_threshold,default)
S3method(pareto_khat_threshold,rvar)
S3method(pareto_min_ss,default)
S3method(pareto_min_ss,rvar)
S3method(pareto_pit,default)
S3method(pareto_pit,draws_matrix)
S3method(pareto_pit,rvar)
S3method(pareto_smooth,default)
S3method(pareto_smooth,rvar)
S3method(pillar_shaft,rvar)
S3method(pit,default)
S3method(pit,draws_matrix)
S3method(pit,rvar)
S3method(print,draws_array)
S3method(print,draws_df)
S3method(print,draws_list)
S3method(print,draws_matrix)
S3method(print,draws_rvars)
S3method(print,draws_summary)
S3method(print,rvar)
S3method(prod,rvar)
S3method(quantile,rvar)
S3method(quantile,rvar_factor)
S3method(quantile,rvar_ordered)
S3method(quantile2,default)
S3method(quantile2,rvar)
S3method(range,rvar)
S3method(rbind,rvar)
S3method(remove_variables,draws_array)
S3method(remove_variables,draws_df)
S3method(remove_variables,draws_list)
S3method(remove_variables,draws_matrix)
S3method(remove_variables,draws_rvars)
S3method(remove_variables,list)
S3method(rename_variables,draws)
S3method(rep,rvar)
S3method(rep.int,rvar,rep_int_rvar)
S3method(rep_len,rvar)
S3method(repair_draws,draws_array)
S3method(repair_draws,draws_df)
S3method(repair_draws,draws_list)
S3method(repair_draws,draws_matrix)
S3method(repair_draws,draws_rvars)
S3method(repair_draws,rvar)
S3method(resample_draws,draws)
S3method(resample_draws,rvar)
S3method(reserved_variables,default)
S3method(reserved_variables,draws_array)
S3method(reserved_variables,draws_df)
S3method(reserved_variables,draws_list)
S3method(reserved_variables,draws_matrix)
S3method(reserved_variables,draws_rvars)
S3method(rhat,default)
S3method(rhat,rvar)
S3method(rhat_basic,default)
S3method(rhat_basic,rvar)
S3method(rhat_nested,default)
S3method(rhat_nested,rvar)
S3method(sd,default)
S3method(sd,rvar)
S3method(split_chains,draws)
S3method(str,rvar)
S3method(subset,draws)
S3method(subset_draws,draws_array)
S3method(subset_draws,draws_df)
S3method(subset_draws,draws_list)
S3method(subset_draws,draws_matrix)
S3method(subset_draws,draws_rvars)
S3method(subset_draws,rvar)
S3method(sum,rvar)
S3method(summarise_draws,default)
S3method(summarise_draws,draws)
S3method(summarise_draws,rvar)
S3method(summary,draws)
S3method(summary,rvar)
S3method(t,rvar)
S3method(thin_draws,draws)
S3method(thin_draws,rvar)
S3method(unique,rvar)
S3method(unique,rvar_factor)
S3method(var,default)
S3method(var,rvar)
S3method(variables,"NULL")
S3method(variables,draws_array)
S3method(variables,draws_df)
S3method(variables,draws_list)
S3method(variables,draws_matrix)
S3method(variables,draws_rvars)
S3method(variance,draws_array)
S3method(variance,draws_matrix)
S3method(variance,rvar)
S3method(vec_cast,character.rvar)
S3method(vec_cast,character.rvar_factor)
S3method(vec_cast,character.rvar_ordered)
S3method(vec_cast,distribution.rvar)
S3method(vec_cast,rvar.character)
S3method(vec_cast,rvar.distribution)
S3method(vec_cast,rvar.double)
S3method(vec_cast,rvar.factor)
S3method(vec_cast,rvar.integer)
S3method(vec_cast,rvar.logical)
S3method(vec_cast,rvar.ordered)
S3method(vec_cast,rvar.rvar)
S3method(vec_cast,rvar.rvar_factor)
S3method(vec_cast,rvar.rvar_ordered)
S3method(vec_cast,rvar_factor.character)
S3method(vec_cast,rvar_factor.double)
S3method(vec_cast,rvar_factor.factor)
S3method(vec_cast,rvar_factor.integer)
S3method(vec_cast,rvar_factor.logical)
S3method(vec_cast,rvar_factor.ordered)
S3method(vec_cast,rvar_factor.rvar)
S3method(vec_cast,rvar_factor.rvar_factor)
S3method(vec_cast,rvar_factor.rvar_ordered)
S3method(vec_cast,rvar_ordered.character)
S3method(vec_cast,rvar_ordered.double)
S3method(vec_cast,rvar_ordered.factor)
S3method(vec_cast,rvar_ordered.integer)
S3method(vec_cast,rvar_ordered.logical)
S3method(vec_cast,rvar_ordered.ordered)
S3method(vec_cast,rvar_ordered.rvar)
S3method(vec_cast,rvar_ordered.rvar_factor)
S3method(vec_cast,rvar_ordered.rvar_ordered)
S3method(vec_proxy,rvar)
S3method(vec_proxy_compare,rvar)
S3method(vec_proxy_equal,rvar)
S3method(vec_proxy_order,rvar)
S3method(vec_ptype,rvar)
S3method(vec_ptype,rvar_factor)
S3method(vec_ptype,rvar_ordered)
S3method(vec_ptype2,character.rvar_factor)
S3method(vec_ptype2,character.rvar_ordered)
S3method(vec_ptype2,distribution.rvar)
S3method(vec_ptype2,double.rvar)
S3method(vec_ptype2,factor.rvar_factor)
S3method(vec_ptype2,factor.rvar_ordered)
S3method(vec_ptype2,integer.rvar)
S3method(vec_ptype2,logical.rvar)
S3method(vec_ptype2,ordered.rvar_factor)
S3method(vec_ptype2,ordered.rvar_ordered)
S3method(vec_ptype2,rvar.distribution)
S3method(vec_ptype2,rvar.double)
S3method(vec_ptype2,rvar.integer)
S3method(vec_ptype2,rvar.logical)
S3method(vec_ptype2,rvar.rvar)
S3method(vec_ptype2,rvar_factor.character)
S3method(vec_ptype2,rvar_factor.factor)
S3method(vec_ptype2,rvar_factor.ordered)
S3method(vec_ptype2,rvar_factor.rvar_factor)
S3method(vec_ptype2,rvar_ordered.character)
S3method(vec_ptype2,rvar_ordered.factor)
S3method(vec_ptype2,rvar_ordered.ordered)
S3method(vec_ptype2,rvar_ordered.rvar_ordered)
S3method(vec_ptype_abbr,rvar)
S3method(vec_ptype_abbr,rvar_factor)
S3method(vec_ptype_abbr,rvar_ordered)
S3method(vec_ptype_full,rvar)
S3method(vec_restore,rvar)
S3method(vec_restore,rvar_factor)
S3method(vec_restore,rvar_ordered)
S3method(weight_draws,draws_array)
S3method(weight_draws,draws_df)
S3method(weight_draws,draws_list)
S3method(weight_draws,draws_matrix)
S3method(weight_draws,draws_rvars)
S3method(weights,draws)
export("%**%")
export("%in%")
export("draws_of<-")
export("variables<-")
export(E)
export(Pr)
export(as_draws)
export(as_draws_array)
export(as_draws_df)
export(as_draws_list)
export(as_draws_matrix)
export(as_draws_rvars)
export(as_rvar)
export(as_rvar_factor)
export(as_rvar_integer)
export(as_rvar_logical)
export(as_rvar_numeric)
export(as_rvar_ordered)
export(autocorrelation)
export(autocovariance)
export(bind_draws)
export(cdf)
export(chain_ids)
export(default_convergence_measures)
export(default_mcse_measures)
export(default_summary_measures)
export(diag)
export(dissent)
export(draw_ids)
export(draws_array)
export(draws_df)
export(draws_list)
export(draws_matrix)
export(draws_of)
export(draws_rvars)
export(drop)
export(entropy)
export(ess_basic)
export(ess_bulk)
export(ess_mean)
export(ess_median)
export(ess_quantile)
export(ess_sd)
export(ess_tail)
export(example_draws)
export(extract_list_of_variable_arrays)
export(extract_variable)
export(extract_variable_array)
export(extract_variable_matrix)
export(for_each_draw)
export(gpdfit)
export(is_constant)
export(is_draws)
export(is_draws_array)
export(is_draws_df)
export(is_draws_list)
export(is_draws_matrix)
export(is_draws_rvars)
export(is_rvar)
export(is_rvar_factor)
export(is_rvar_ordered)
export(iteration_ids)
export(mad)
export(match)
export(mcse_mean)
export(mcse_median)
export(mcse_quantile)
export(mcse_sd)
export(merge_chains)
export(modal_category)
export(mutate_variables)
export(nchains)
export(ndraws)
export(niterations)
export(nvariables)
export(order_draws)
export(pareto_convergence_rate)
export(pareto_diags)
export(pareto_khat)
export(pareto_khat_threshold)
export(pareto_min_ss)
export(pareto_pit)
export(pareto_smooth)
export(pgeneralized_pareto)
export(pit)
export(ps_convergence_rate)
export(ps_khat_threshold)
export(ps_min_ss)
export(ps_tail)
export(ps_tail_length)
export(qgeneralized_pareto)
export(quantile2)
export(r_scale)
export(rdo)
export(rename_variables)
export(repair_draws)
export(resample_draws)
export(reserved_variables)
export(rfun)
export(rhat)
export(rhat_basic)
export(rhat_nested)
export(rstar)
export(rvar)
export(rvar_all)
export(rvar_any)
export(rvar_apply)
export(rvar_factor)
export(rvar_ifelse)
export(rvar_is_finite)
export(rvar_is_infinite)
export(rvar_is_na)
export(rvar_is_nan)
export(rvar_mad)
export(rvar_max)
export(rvar_mean)
export(rvar_median)
export(rvar_min)
export(rvar_ordered)
export(rvar_prod)
export(rvar_quantile)
export(rvar_range)
export(rvar_rng)
export(rvar_sd)
export(rvar_sum)
export(rvar_var)
export(sd)
export(set_variables)
export(split_chains)
export(subset_draws)
export(summarise_draws)
export(summarize_draws)
export(thin_draws)
export(u_scale)
export(uniformity_test)
export(var)
export(variables)
export(variance)
export(weight_draws)
export(z_scale)
exportMethods(diag)
exportMethods(drop)
import(checkmate)
import(stats)
importFrom(abind,abind)
importFrom(distributional,cdf)
importFrom(distributional,variance)
importFrom(methods,callNextMethod)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(methods,setOldClass)
importFrom(pillar,format_glimpse)
importFrom(pillar,new_pillar_shaft_simple)
importFrom(pillar,pillar_shaft)
importFrom(rlang,as_label)
importFrom(rlang,as_name)
importFrom(rlang,caller_env)
importFrom(rlang,dots_list)
importFrom(rlang,enquo)
importFrom(rlang,enquos)
importFrom(rlang,eval_tidy)
importFrom(rlang,expr)
importFrom(rlang,inject)
importFrom(rlang,is_missing)
importFrom(rlang,missing_arg)
importFrom(rlang,quo_get_env)
importFrom(rlang,quo_get_expr)
importFrom(rlang,quos)
importFrom(tensorA,as.tensor)
importFrom(tensorA,chol.tensor)
importFrom(tensorA,mul.tensor)
importFrom(tibble,as_tibble)
importFrom(utils,lsf.str)
importFrom(utils,str)
importFrom(vctrs,new_vctr)
importFrom(vctrs,vec_cast)
importFrom(vctrs,vec_cast_common)
importFrom(vctrs,vec_proxy)
importFrom(vctrs,vec_proxy_compare)
importFrom(vctrs,vec_proxy_equal)
importFrom(vctrs,vec_proxy_order)
importFrom(vctrs,vec_ptype)
importFrom(vctrs,vec_ptype2)
importFrom(vctrs,vec_ptype_abbr)
importFrom(vctrs,vec_ptype_full)
importFrom(vctrs,vec_restore)
importFrom(vctrs,vec_slice)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# posterior <img src="man/figures/logo.svg" align="right" width="120" />

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/posterior)](https://CRAN.R-project.org/package=posterior)
[![R-CMD-check](https://github.com/stan-dev/posterior/workflows/R-CMD-check/badge.svg)](https://github.com/stan-dev/posterior/actions?workflow=R-CMD-check)
[![Coverage
Status](https://codecov.io/gh/stan-dev/posterior/branch/master/graph/badge.svg)](https://app.codecov.io/gh/stan-dev/posterior)
<!-- badges: end -->

The **posterior** R package is intended to provide useful tools for both
users and developers of packages for fitting Bayesian models or working
with output from Bayesian models. The primary goals of the package are
to:

- Efficiently convert between many different useful formats of draws
  (samples) from posterior or prior distributions.
- Provide consistent methods for operations commonly performed on draws,
  for example, subsetting, binding, or mutating draws.
- Provide various summaries of draws in convenient formats.
- Provide lightweight implementations of state of the art posterior
  inference diagnostics.

If you are new to **posterior** we recommend starting with these
vignettes:

- [*The posterior R
  package*](https://mc-stan.org/posterior/articles/posterior.html): an
  introduction to the package and its main functionality
- [*rvar: The Random Variable
  Datatype*](https://mc-stan.org/posterior/articles/rvar.html): an
  overview of the new random variable datatype

### Installation

You can install the latest official release version via

``` r
install.packages("posterior")
```

or build the developmental version directly from GitHub via

``` r
# install.packages("remotes")
remotes::install_github("stan-dev/posterior")
```

### Examples

Here we offer a few examples of using the package. For a more detailed
overview see the vignette [*The posterior R
package*](https://mc-stan.org/posterior/articles/posterior.html).

``` r
library("posterior")
#> This is posterior version 1.6.1
#> 
#> Attaching package: 'posterior'
#> The following objects are masked from 'package:stats':
#> 
#>     mad, sd, var
#> The following objects are masked from 'package:base':
#> 
#>     %in%, match
```

To demonstrate how to work with the **posterior** package, we will use
example posterior draws obtained from the eight schools hierarchical
meta-analysis model described in Gelman et al. (2013). Essentially, we
have an estimate per school (`theta[1]` through `theta[8]`) as well as
an overall mean (`mu`) and standard deviation across schools (`tau`).

#### Draws formats

``` r
eight_schools_array <- example_draws("eight_schools")
print(eight_schools_array, max_variables = 3)
#> # A draws_array: 100 iterations, 4 chains, and 10 variables
#> , , variable = mu
#> 
#>          chain
#> iteration   1    2     3   4
#>         1 2.0  3.0  1.79 6.5
#>         2 1.5  8.2  5.99 9.1
#>         3 5.8 -1.2  2.56 0.2
#>         4 6.8 10.9  2.79 3.7
#>         5 1.8  9.8 -0.03 5.5
#> 
#> , , variable = tau
#> 
#>          chain
#> iteration   1    2    3   4
#>         1 2.8 2.80  8.7 3.8
#>         2 7.0 2.76  2.9 6.8
#>         3 9.7 0.57  8.4 5.3
#>         4 4.8 2.45  4.4 1.6
#>         5 2.8 2.80 11.0 3.0
#> 
#> , , variable = theta[1]
#> 
#>          chain
#> iteration     1     2    3     4
#>         1  3.96  6.26 13.3  5.78
#>         2  0.12  9.32  6.3  2.09
#>         3 21.25 -0.97 10.6 15.72
#>         4 14.70 12.45  5.4  2.69
#>         5  5.96  9.75  8.2 -0.91
#> 
#> # ... with 95 more iterations, and 7 more variables
```

The draws for this example come as a `draws_array` object, that is, an
array with dimensions iterations x chains x variables. We can easily
transform it to another format, for instance, a data frame with
additional meta information.

``` r
eight_schools_df <- as_draws_df(eight_schools_array)
print(eight_schools_df)
#> # A draws_df: 100 iterations, 4 chains, and 10 variables
#>      mu tau theta[1] theta[2] theta[3] theta[4] theta[5] theta[6]
#> 1  2.01 2.8     3.96    0.271    -0.74      2.1    0.923      1.7
#> 2  1.46 7.0     0.12   -0.069     0.95      7.3   -0.062     11.3
#> 3  5.81 9.7    21.25   14.931     1.83      1.4    0.531      7.2
#> 4  6.85 4.8    14.70    8.586     2.67      4.4    4.758      8.1
#> 5  1.81 2.8     5.96    1.156     3.11      2.0    0.769      4.7
#> 6  3.84 4.1     5.76    9.909    -1.00      5.3    5.889     -1.7
#> 7  5.47 4.0     4.03    4.151    10.15      6.6    3.741     -2.2
#> 8  1.20 1.5    -0.28    1.846     0.47      4.3    1.467      3.3
#> 9  0.15 3.9     1.81    0.661     0.86      4.5   -1.025      1.1
#> 10 7.17 1.8     6.08    8.102     7.68      5.6    7.106      8.5
#> # ... with 390 more draws, and 2 more variables
#> # ... hidden reserved variables {'.chain', '.iteration', '.draw'}
```

Different formats are preferable in different situations and hence
posterior supports multiple formats and easy conversion between them.
For more details on the available formats see `help("draws")`. All of
the formats are essentially base R object classes and can be used as
such. For example, a `draws_matrix` object is just a `matrix` with a
little more consistency and additional methods.

#### Summarizing draws

Computing summaries of posterior or prior draws and convergence
diagnostics for posterior draws is one of the most common tasks when
working with Bayesian models fit using Markov Chain Monte Carlo (MCMC)
methods. The **posterior** package provides a flexible interface for
this purpose via `summarise_draws()`:

``` r
# summarise_draws or summarize_draws
summarise_draws(eight_schools_df)
#> # A tibble: 10 × 10
#>    variable  mean median    sd   mad      q5   q95  rhat ess_bulk ess_tail
#>    <chr>    <dbl>  <dbl> <dbl> <dbl>   <dbl> <dbl> <dbl>    <dbl>    <dbl>
#>  1 mu        4.18   4.16  3.40  3.57  -0.854  9.39  1.02     558.     322.
#>  2 tau       4.16   3.07  3.58  2.89   0.309 11.0   1.01     246.     202.
#>  3 theta[1]  6.75   5.97  6.30  4.87  -1.23  18.9   1.01     400.     254.
#>  4 theta[2]  5.25   5.13  4.63  4.25  -1.97  12.5   1.02     564.     372.
#>  5 theta[3]  3.04   3.99  6.80  4.94 -10.3   11.9   1.01     312.     205.
#>  6 theta[4]  4.86   4.99  4.92  4.51  -3.57  12.2   1.02     695.     252.
#>  7 theta[5]  3.22   3.72  5.08  4.38  -5.93  10.8   1.01     523.     306.
#>  8 theta[6]  3.99   4.14  5.16  4.81  -4.32  11.5   1.02     548.     205.
#>  9 theta[7]  6.50   5.90  5.26  4.54  -1.19  15.4   1.00     434.     308.
#> 10 theta[8]  4.57   4.64  5.25  4.89  -3.79  12.2   1.02     355.     146.
```

Basically, we get a data frame with one row per variable and one column
per summary statistic or convergence diagnostic. The summaries `rhat`,
`ess_bulk`, and `ess_tail` are described in Vehtari et al. (2020). We
can choose which summaries to compute by passing additional arguments,
either functions or names of functions. For instance, if we only wanted
the mean and its corresponding Monte Carlo Standard Error (MCSE) we
would use:

``` r
summarise_draws(eight_schools_df, "mean", "mcse_mean")
#> # A tibble: 10 × 3
#>    variable  mean mcse_mean
#>    <chr>    <dbl>     <dbl>
#>  1 mu        4.18     0.150
#>  2 tau       4.16     0.213
#>  3 theta[1]  6.75     0.319
#>  4 theta[2]  5.25     0.202
#>  5 theta[3]  3.04     0.447
#>  6 theta[4]  4.86     0.189
#>  7 theta[5]  3.22     0.232
#>  8 theta[6]  3.99     0.222
#>  9 theta[7]  6.50     0.250
#> 10 theta[8]  4.57     0.273
```

For a function to work with `summarise_draws`, it needs to take a vector
or matrix of numeric values and returns a single numeric value or a
named vector of numeric values.

#### Subsetting draws

Another common task when working with posterior (or prior) draws, is
subsetting according to various aspects of the draws (iterations,
chains, or variables). **posterior** provides a convenient interface for
this purpose via the `subset_draws()` method. For example, here is the
code to extract the first five iterations of the first two chains of the
variable `mu`:

``` r
subset_draws(eight_schools_df, variable = "mu", chain = 1:2, iteration = 1:5)
#> # A draws_df: 5 iterations, 2 chains, and 1 variables
#>      mu
#> 1   2.0
#> 2   1.5
#> 3   5.8
#> 4   6.8
#> 5   1.8
#> 6   3.0
#> 7   8.2
#> 8  -1.2
#> 9  10.9
#> 10  9.8
#> # ... hidden reserved variables {'.chain', '.iteration', '.draw'}
```

The same call to `subset_draws()` can be used regardless of whether the
object is a `draws_df`, `draws_array`, `draws_list`, etc.

#### Mutating and renaming draws

The magic of having obtained draws from the joint posterior (or prior)
distribution of a set of variables is that these draws can also be used
to obtain draws from any other variable that is a function of the
original variables. That is, if are interested in the posterior
distribution of, say, `phi = (mu + tau)^2` all we have to do is to
perform the transformation for each of the individual draws to obtain
draws from the posterior distribution of the transformed variable. This
procedure is automated in the `mutate_variables` method:

``` r
x <- mutate_variables(eight_schools_df, phi = (mu + tau)^2)
x <- subset_draws(x, c("mu", "tau", "phi"))
print(x)
#> # A draws_df: 100 iterations, 4 chains, and 3 variables
#>      mu tau   phi
#> 1  2.01 2.8  22.8
#> 2  1.46 7.0  71.2
#> 3  5.81 9.7 240.0
#> 4  6.85 4.8 135.4
#> 5  1.81 2.8  21.7
#> 6  3.84 4.1  62.8
#> 7  5.47 4.0  88.8
#> 8  1.20 1.5   7.1
#> 9  0.15 3.9  16.6
#> 10 7.17 1.8  79.9
#> # ... with 390 more draws
#> # ... hidden reserved variables {'.chain', '.iteration', '.draw'}
```

When we do the math ourselves, we see that indeed for each draw, `phi`
is equal to `(mu + tau)^2` (up to rounding two 2 digits for the purpose
of printing).

We may also easily rename variables, or even entire vectors of variables
via `rename_variables`, for example:

``` r
x <- rename_variables(eight_schools_df, mean = mu, alpha = theta)
variables(x)
#>  [1] "mean"     "tau"      "alpha[1]" "alpha[2]" "alpha[3]" "alpha[4]" "alpha[5]"
#>  [8] "alpha[6]" "alpha[7]" "alpha[8]"
```

As with all **posterior** methods, `mutate_variables` and
`rename_variables` can be used with all draws formats.

#### Binding draws together

Suppose we have multiple draws objects that we want to bind together:

``` r
x1 <- draws_matrix(alpha = rnorm(5), beta = 1)
x2 <- draws_matrix(alpha = rnorm(5), beta = 2)
x3 <- draws_matrix(theta = rexp(5))
```

Then, we can use the `bind_draws` method to bind them along different
dimensions. For example, we can bind `x1` and `x3` together along the
`'variable'` dimension:

``` r
x4 <- bind_draws(x1, x3, along = "variable")
print(x4)
#> # A draws_matrix: 5 iterations, 1 chains, and 3 variables
#>     variable
#> draw alpha beta theta
#>    1 -0.14    1 2.080
#>    2  0.13    1 1.077
#>    3  0.71    1 0.037
#>    4 -0.99    1 1.447
#>    5 -1.39    1 1.031
```

Or, we can bind `x1` and `x2` together along the `'draw'` dimension:

``` r
x5 <- bind_draws(x1, x2, along = "draw")
print(x5)
#> # A draws_matrix: 10 iterations, 1 chains, and 2 variables
#>     variable
#> draw  alpha beta
#>   1  -0.140    1
#>   2   0.132    1
#>   3   0.705    1
#>   4  -0.990    1
#>   5  -1.394    1
#>   6  -2.889    2
#>   7  -0.072    2
#>   8  -0.051    2
#>   9  -1.983    2
#>   10 -0.204    2
```

As with all **posterior** methods, `bind_draws` can be used with all
draws formats.

#### Converting from regular R objects to draws formats

The `eight_schools` example already comes in a format natively supported
by posterior but we could of course also import the draws from other
sources, for example, from common base R objects:

``` r
x <- matrix(rnorm(50), nrow = 10, ncol = 5)
colnames(x) <- paste0("V", 1:5)
x <- as_draws_matrix(x)
print(x)
#> # A draws_matrix: 10 iterations, 1 chains, and 5 variables
#>     variable
#> draw     V1     V2     V3      V4     V5
#>   1   1.477 -1.804 -0.979  0.4196  0.046
#>   2   0.081 -0.695 -1.275 -1.2993  0.407
#>   3   1.384 -0.963  0.465  0.9283 -0.101
#>   4  -0.852  1.574  1.118 -0.2389  0.067
#>   5  -0.515  0.017 -1.651  0.0054 -1.756
#>   6   0.308 -1.376  1.099 -0.6325 -0.437
#>   7  -0.818  0.703 -0.027 -0.9165  0.706
#>   8  -0.415 -0.232 -1.818 -1.6000 -0.551
#>   9   2.599  0.958  0.522  0.5579  1.159
#>   10 -1.070  1.259 -0.475  1.2223 -0.963

summarise_draws(x, "mean", "sd", "median", "mad")
#> # A tibble: 5 × 5
#>   variable    mean    sd  median   mad
#>   <chr>      <dbl> <dbl>   <dbl> <dbl>
#> 1 V1        0.218  1.22  -0.167  0.990
#> 2 V2       -0.0558 1.16  -0.108  1.42 
#> 3 V3       -0.302  1.10  -0.251  1.33 
#> 4 V4       -0.155  0.951 -0.117  1.09 
#> 5 V5       -0.142  0.839 -0.0273 0.710
```

Instead of `as_draws_matrix()` we also could have just used
`as_draws()`, which attempts to find the closest available format to the
input object. In this case this would result in a `draws_matrix` object
either way.

The above matrix example contained only one chain. Multi-chain draws
could be stored in base R 3-D array object, which can also be converted
to a draws object:

``` r
x <- array(data=rnorm(200), dim=c(10, 2, 5))
x <- as_draws_matrix(x)
variables(x) <-  paste0("V", 1:5)
print(x)
#> # A draws_matrix: 10 iterations, 2 chains, and 5 variables
#>     variable
#> draw     V1    V2    V3      V4    V5
#>   1   0.288 -1.61 -0.36 -0.5534  0.58
#>   2   0.267  0.05 -1.41 -2.2698  1.90
#>   3   1.051  1.78  1.00 -1.1469  0.40
#>   4   0.679  1.57  0.38 -0.3360  0.98
#>   5  -1.952 -1.18 -0.29  0.0928 -1.09
#>   6  -0.380 -1.48 -0.88  0.3971  1.19
#>   7   0.398 -1.86 -0.77 -0.1377 -0.21
#>   8   0.023  1.97 -1.66 -0.5317 -0.21
#>   9   0.380  0.15 -1.91  0.2331  0.49
#>   10 -0.653  0.53  0.24 -0.0038  0.26
#> # ... with 10 more draws
```

#### Converting from mcmc objects to draws formats

The **coda** and **rjags** packages use `mcmc` and `mcmc.list` objects
which can also be converted to draws objects:

``` r
data(line, package = "coda")
line <- as_draws_df(line)
print(line)
#> # A draws_df: 200 iterations, 2 chains, and 3 variables
#>    alpha  beta sigma
#> 1    7.2 -1.57 11.23
#> 2    3.0  1.50  4.89
#> 3    3.7  0.63  1.40
#> 4    3.3  1.18  0.66
#> 5    3.7  0.49  1.36
#> 6    3.6  0.21  1.04
#> 7    2.7  0.88  1.29
#> 8    3.0  1.09  0.46
#> 9    3.5  1.07  0.63
#> 10   2.1  1.48  0.91
#> # ... with 390 more draws
#> # ... hidden reserved variables {'.chain', '.iteration', '.draw'}
```

### Contributing to posterior

We welcome contributions! The **posterior** package is under active
development. If you find bugs or have ideas for new features (for us or
yourself to implement) please [open an
issue](https://github.com/stan-dev/posterior/issues) on GitHub. See
[CONTRIBUTING.md](https://github.com/stan-dev/posterior/blob/master/.github/CONTRIBUTING.md)
for more details.

### Citing posterior

Developing and maintaining open source software is an important yet
often underappreciated contribution to scientific progress. Thus,
whenever you are using open source software (or software in general),
please make sure to cite it appropriately so that developers get credit
for their work.

When using **posterior**, please cite it as follows:

- Bürkner P. C., Gabry J., Kay M., & Vehtari A. (2020). “posterior:
  Tools for Working with Posterior Distributions.” R package version
  XXX, \<URL: <https://mc-stan.org/posterior/>\>.

When using the MCMC convergence diagnostics `rhat`, `ess_bulk`,
`ess_tail`, `ess_median`, `ess_quantile`, `mcse_median`, or
`mcse_quantile` please also cite

- Vehtari A., Gelman A., Simpson D., Carpenter B., & Bürkner P. C.
  (2021). Rank-normalization, folding, and localization: An improved
  Rhat for assessing convergence of MCMC (with discussion). *Bayesian
  Analysis*. 16(2), 667–718. doi.org/10.1214/20-BA1221

When using the MCMC convergence diagnostic `rhat_nested` please also
cite

- Margossian, C. C., Hoffman, M. D., Sountsov, P., Riou-Durand, L.,
  Vehtari, A., and Gelman, A. (2024). Nested $\widehat{R}$: Assessing
  the convergence of Markov chain Monte Carlo when running many short
  chains. *Bayesian Analysis*, <doi:10.1214/24-BA1453>.

When using the MCMC convergence diagnostic `rstar` please also cite

- Lambert, B. and Vehtari, A. (2022). $R^*$: A robust MCMC convergence
  diagnostic with uncertainty using decision tree classifiers. *Bayesian
  Analysis*, 17(2):353-379. <doi:10.1214/20-BA1252>

When using the Pareto-k diagnostics `pareto_khat`, `pareto_min_ss`,
`pareto_convergence_rate`, `khat_threshold` or `pareto_diags`, or Pareto
smoothing `pareto_smooth` please also cite

- Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
  Pareto smoothed importance sampling. *Journal of Machine Learning
  Research*, 25(72):1-58.

The same information can be obtained by running `citation("posterior")`.

### References

Gelman A., Carlin J. B., Stern H. S., David B. Dunson D. B., Aki Vehtari
A., & Rubin D. B. (2013). *Bayesian Data Analysis, Third Edition*.
Chapman and Hall/CRC.

Lambert, B. and Vehtari, A. (2022). $R^*$: A robust MCMC convergence
diagnostic with uncertainty using decision tree classifiers. *Bayesian
Analysis*, 17(2):353-379. <doi:10.1214/20-BA1252>

Margossian, C. C., Hoffman, M. D., Sountsov, P., Riou-Durand, L.,
Vehtari, A., and Gelman, A. (2024). Nested $\widehat{R}$: Assessing the
convergence of Markov chain Monte Carlo when running many short chains.
*Bayesian Analysis*, <doi:10.1214/24-BA1453>.

Vehtari A., Gelman A., Simpson D., Carpenter B., & Bürkner P. C. (2021).
Rank-normalization, folding, and localization: An improved Rhat for
assessing convergence of MCMC (with discussion). *Bayesian Analysis*.
16(2), 667–718. doi.org/10.1214/20-BA1221

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., and Gabry, J. (2024).
Pareto smoothed importance sampling. *Journal of Machine Learning
Research*, 25(72):1-58.

### Licensing

The **posterior** package is licensed under the following licenses:

- Code: BSD 3-clause (<https://opensource.org/license/bsd-3-clause>)
- Documentation: CC-BY 4.0
  (<https://creativecommons.org/licenses/by/4.0/>)

## ----setup--------------------------------------------------------------------
library(posterior)
library(dplyr)
options(pillar.neg = FALSE, pillar.subtle=FALSE, pillar.sigfig=2)

## ----simulate-data-1----------------------------------------------------------
N <- 1000
phi <- 0.3
set.seed(6534)
dr <- array(data=replicate(4,as.numeric(arima.sim(n = N,
                                                list(ar = c(phi)),
                                                sd = sqrt((1-phi^2))))),
         dim=c(N,4,1)) %>%
  as_draws_df() %>%
  set_variables('xn')

## ----simulate-data-2----------------------------------------------------------
drt <- dr %>%
  mutate_variables(xt3=qt(pnorm(xn), df=3),
                   xt2_5=qt(pnorm(xn), df=2.5),
                   xt2=qt(pnorm(xn), df=2),
                   xt1_5=qt(pnorm(xn), df=1.5),
                   xt1=qt(pnorm(xn), df=1))

## ----summarise_draws----------------------------------------------------------
drt %>%
  summarise_draws()

## ----summarise_draws-mcse-----------------------------------------------------
drt %>%
  summarise_draws(mean, sd, mcse_mean, ess_bulk, ess_basic)

## ----summarise_draws-pareto_khat----------------------------------------------
drt %>%
  summarise_draws(mean, sd, mcse_mean, ess_basic, pareto_khat)

## ----pareto_smooth------------------------------------------------------------
drts <- drt %>% 
  mutate_variables(xt3_s=pareto_smooth(xt3),
                   xt2_5_s=pareto_smooth(xt2_5),
                   xt2_s=pareto_smooth(xt2),
                   xt1_5_s=pareto_smooth(xt1_5),
                   xt1_s=pareto_smooth(xt1)) %>%
  subset_draws(variable="_s", regex=TRUE)

## ----summarise_draws-pareto_khat-2--------------------------------------------
drts %>%
  summarise_draws(mean, mcse_mean, ess_basic, pareto_khat)

## ----summarise_draws-min_ss---------------------------------------------------
drt %>%
  summarise_draws(mean, mcse_mean, ess_basic,
                  pareto_khat, min_ss=pareto_min_ss)

## ----summarise_draws-conv_rate------------------------------------------------
drt %>%
  summarise_draws(mean, mcse_mean, ess_basic,
                  pareto_khat, min_ss=pareto_min_ss,
                  conv_rate=pareto_convergence_rate)

## ----summarise_draws-khat_threshold-------------------------------------------
drt %>%
  subset_draws(iteration=1:100) %>%
  summarise_draws(mean, mcse_mean, ess_basic,
                  pareto_khat, min_ss=pareto_min_ss,
                  khat_thres=pareto_khat_threshold,
                  conv_rate=pareto_convergence_rate)

## ----summarise_draws-pareto_diags---------------------------------------------
drt %>%
  mutate_variables(xt2_5_sq=xt2_5^2) %>%
  subset_draws(variable="xt2_5_sq") %>%
  summarise_draws(mean, mcse_mean,
                  pareto_diags)

# concatenation and binding ----------------------------------------------------

#' @export
c.rvar <- function(...) {
  common_type <- vctrs::vec_ptype_common(...)
  if (!is_rvar(common_type)) {
    # if the common type of the arguments is not an rvar, fall back to the
    # vctrs implementation of c()
    return(vctrs::vec_c(...))
  }

  args <- list(...)
  # drop NULLs (for compatibility with R < 4: in R >= 4, NULLs will have already been dropped)
  args[vapply(args, is.null, logical(1))] <- NULL

  out <- make_1d(args[[1]], names(args)[[1]])

  # process remaining args in succession, binding them to the output
  for (i in seq_along(args)[-1]) {
    arg_name <- names(args)[[i]]
    arg <- make_1d(vec_cast(args[[i]], common_type), arg_name)
    out <- broadcast_and_bind_rvars(out, arg)
  }

  out
}

#' @export
rbind.rvar <- function(...) {
  # not sure why deparse.level is not passed here correctly...
  deparse.level <- rlang::caller_env()$deparse.level %||% 1
  bind_rvars(list(...), as.list(substitute(list(...))[-1]), deparse.level)
}

#' @export
cbind.rvar <- function(...) {
  # not sure why deparse.level is not passed here correctly...
  deparse.level <- rlang::caller_env()$deparse.level %||% 1
  bind_rvars(list(...), as.list(substitute(list(...))[-1]), deparse.level, axis = 2)
}

#' bind a list of objects together, as in cbind or rbind (depending on `axis`),
#' converting to rvars as needed
#' @noRd
bind_rvars <- function(args, arg_exprs, deparse.level = 1, axis = 1) {
  if (any(sapply(args, is.data.frame))) {
    # if there is a data frame in args, we should just make the first arg
    # into a data frame and then use the data frame bind implementation

    # data frames always deparse at level 2 (expressions get named)
    args <- deparse_names(args, arg_exprs, deparse.level = 2)
    args[[1]] <- as.data.frame(make_at_least_2d(args[[1]], axis, names(args)[[1]]))
    bind <- if (axis == 1) rbind else cbind
    return(do.call(bind, args))
  }

  args <- deparse_names(args, arg_exprs, deparse.level)

  out <- make_at_least_2d(as_rvar(args[[1]]), axis, names(args)[[1]])

  # process remaining args in succession, binding them to the output
  for (i in seq_along(args)[-1]) {
    arg_name <- names(args)[[i]]
    arg <- make_at_least_2d(as_rvar(args[[i]]), axis, arg_name)
    out <- broadcast_and_bind_rvars(out, arg, axis)
  }

  out
}


# helpers: concatenation and binding --------------------------------------

#' broadcast two rvars to compatible dimensions and bind along the `axis` dimension
#' @noRd
broadcast_and_bind_rvars <- function(x, y, axis = 1) {
  UseMethod("broadcast_and_bind_rvars")
}

#' @export
broadcast_and_bind_rvars.rvar <- function(x, y, axis = 1) {
  # TODO: should really replace this with something that takes a list of rvars,
  # conforms their chains, broadcasts them all to common dimensions and uses a
  # single abind to bind them.
  if (!length(x)) return(y)
  if (!length(y)) return(x)

  draws_axis <- axis + 1 # because first dim is draws

  # conform nchains
  # (don't need to do draws here since that's part of the broadcast below)
  c(x, y) %<-% conform_rvar_nchains(list(x, y))

  # broadcast each array to the desired dimensions
  # (except along the axis we are binding along)
  draws_x <- draws_of(x)
  draws_y <- draws_of(y)
  new_dim <- dim2_common(dim(draws_x), dim(draws_y))

  new_dim[draws_axis] <- dim(draws_x)[draws_axis]
  draws_x <- broadcast_array(draws_x, new_dim)

  new_dim[draws_axis] <- dim(draws_y)[draws_axis]
  draws_y <- broadcast_array(draws_y, new_dim)

  # factors may not bind properly with abind, so convert them to characters first
  if (is.factor(draws_x)) draws_x <- while_preserving_dims(as.character, draws_x)
  if (is.factor(draws_y)) draws_y <- while_preserving_dims(as.character, draws_y)

  # bind along desired axis
  result <- new_rvar(
    abind(draws_x, draws_y, along = draws_axis, use.dnns = TRUE),
    .nchains = nchains(x)
  )
}

#' @export
broadcast_and_bind_rvars.rvar_factor <- function(x, y, axis = 1) {
  result <- NextMethod()
  combine_rvar_factor_levels(
    result, list(levels(x), levels(y)), is_rvar_ordered(x) && is_rvar_ordered(y)
  )
}


#' Deparse argument names roughly following the rules of the deparse.level
#' argument to cbind / rbind
#' @importFrom rlang as_name as_label
#' @noRd
deparse_names <- function(args, arg_exprs, deparse.level) {
  # give arguments names if needed
  if (deparse.level > 0) {
    if (is.null(names(args))) {
      names(args) <- rep("", length(args))
    }
    for (i in seq_along(arg_exprs)) {
      arg_name <- names(args)[[i]]
      arg_expr <- arg_exprs[[i]]
      if (!isTRUE(nzchar(arg_name))) {
        if (deparse.level == 1 && is.name(arg_expr)) {
          names(args)[[i]] <- as_name(arg_expr)
        } else if (deparse.level > 1) {
          names(args)[[i]] <- as_label(arg_expr)
        }
      }
    }
    if (all(names(args) == "")) {
      names(args) = NULL
    }
  }
  args
}

#' restructure a variable for concatenation with `c()` by ensuring that `x` has
#' only 1 dimension and merging the high-level `name` with names in `x`
#' @noRd
make_1d <- function(x, name) {
  if (length(dim(x)) > 1) {
    dim(x) <- length(x)
  }
  if (isTRUE(nzchar(name))) {
    # name provided, merge with names in the vector
    if (length(x) > 1) {
      if (!length(names(x))) names(x) <- rep("", length(x))
      empty_names <- !nzchar(names(x))
      names(x)[empty_names] <- seq_along(x)[empty_names]
      names(x)[!empty_names] <- paste0(".", names(x)[!empty_names])
      names(x) <- paste0(name, names(x))
    } else {
      names(x) <- name
    }
  }
  x
}

#' restructure a variable for binding along `axis` with rbind (axis = 1) or
#' cbind (axis = 2). Ensures that `x` has at least two dimensions (filling
#' setting extra dimension `axis` to 1 if needed) and if `x` is vector-like
#' (is 1-dimensional), sets the name of the newly-added dimension to `axis_name`
#' @noRd
make_at_least_2d <- function(x, axis, axis_name) {
  if (length(dim(x)) <= 1) {
    # input is a vector, turn it into a matrix
    if (axis == 1) { # rbind
      dim(x) = c(1, length(x))
    } else {         # cbind
      dim(x) = c(length(x), 1)
    }
    if (isTRUE(nchar(axis_name) > 0)) {
      dimnames(x)[[axis]] <- axis_name
    }
  }
  x
}

#' Resample `draws` objects
#'
#' Resample [`draws`] objects according to provided weights, for example weights
#' obtained through importance sampling.
#'
#' @template args-methods-x
#' @param weights (numeric vector) A vector of positive weights of length
#'   `ndraws(x)`. The weights will be internally normalized. If `weights` is not
#'   specified, an attempt will be made to extract any weights already stored in
#'   the draws object (via [weight_draws()]). If no weights are stored in the
#'   draws object, equal weight is supplied to each draw. How exactly the
#'   weights influence the resampling depends on the `method` argument.
#' @param method (string) The resampling method to use:
#'  * `"simple"`: simple random resampling with replacement
#'  * `"simple_no_replace"`: simple random resampling without replacement
#'  * `"stratified"`: stratified resampling with replacement
#'  * `"deterministic"`: deterministic resampling with replacement
#'
#'  Currently, `"stratified"` is the default as it has comparably low variance
#'  and bias with respect to ideal resampling. The latter would sample perfectly
#'  proportional to the weights, but this is not possible in practice due to the
#'  finite number of draws available. For more details about resampling methods,
#'  see Kitagawa (1996).
#' @param ndraws (positive integer) The number of draws to be returned. By
#'   default `ndraws` is set internally to the total number of draws in `x` if
#'   sensible.
#' @template args-methods-dots
#' @template return-draws
#'
#' @details Upon usage of `resample_draws()`, chains will automatically be merged
#'   due to subsetting of individual draws (see [`subset_draws`] for details).
#'   Also, weights stored in the `draws` object will be removed in the process,
#'   as resampling invalidates existing weights.
#'
#' @references
#' Kitagawa, G., Monte Carlo Filter and Smoother for Non-Gaussian Nonlinear '
#' State Space Models, *Journal of Computational and Graphical Statistics*,
#' 5(1):1-25, 1996.
#'
#' @seealso [resample_draws()]
#'
#' @examples
#' x <- as_draws_df(example_draws())
#'
#' # random weights for justr for demonstration
#' w <- runif(ndraws(x), 0, 10)
#'
#' # use default stratified sampling
#' x_rs <- resample_draws(x, weights = w)
#' summarise_draws(x_rs, default_summary_measures())
#'
#' # use simple random sampling
#' x_rs <- resample_draws(x, weights = w, method = "simple")
#' summarise_draws(x_rs, default_summary_measures())
#'
#' @export
resample_draws <- function(x, ...) {
  UseMethod("resample_draws")
}

#' @rdname resample_draws
#' @export
resample_draws.draws <- function(x, weights = NULL, method = "stratified",
                                 ndraws = NULL, ...) {
  ndraws_total <- ndraws(x)
  assert_numeric(weights, len = ndraws_total, lower = 0, null.ok = TRUE)
  assert_choice(method, supported_resample_methods())
  assert_number(ndraws, null.ok = TRUE, lower = 0)
  if (is.null(weights)) {
    weights <- weights(x, normalize = TRUE)
    if (is.null(weights)) {
      weights <- rep.int(1/ndraws_total, ndraws_total)
    }
    # resampling invalidates stored weights
    x <- remove_variables(x, ".log_weight")
  } else {
    weights <- weights / sum(weights)
  }
  if (is.null(ndraws)) {
    if (grepl("_no_replace$", method)) {
      stop_no_call("Argument 'ndraws' is required when sampling without replacement.")
    }
    ndraws <- length(weights)
  }
  method_fun <- paste0(".resample_", method)
  method_fun <- get(method_fun, asNamespace("posterior"))
  draw_ids <- method_fun(weights = weights, ndraws = ndraws, ...)
  subset_draws(x, draw = draw_ids, unique = FALSE)
}

#' @rdname resample_draws
#' @export
resample_draws.rvar <- function(x, ...) {
  resample_draws(draws_rvars(x = x), ...)$x
}

# simple random resampling with replacement
# @return index vector of length 'ndraws'
.resample_simple <- function(weights, ndraws, ...) {
  out <- seq_along(weights)
  sample(out, ndraws, replace = TRUE, prob = weights)
}


# internal ----------------------------------------------------------------

# simple random resampling without replacement
# @return index vector of length 'ndraws'
.resample_simple_no_replace <- function(weights, ndraws, ...) {
  if (ndraws > length(weights)) {
    stop_no_call("Argument 'ndraws' must be smaller than the total ",
          "number of draws in method 'simple_no_replace'.")
  }
  out <- seq_along(weights)
  sample(out, ndraws, replace = FALSE, prob = weights)
}

# Stratified resampling
#   Kitagawa, G., Monte Carlo Filter and Smoother for Non-Gaussian
#   Nonlinear State Space Models, Journal of Computational and
#   Graphical Statistics, 5(1):1-25, 1996.
# @return index vector of length 'ndraws'
.resample_stratified <- function(weights, ndraws, ...) {
  # expected number of repetitions for each original draw
  w <- ndraws * weights
  out <- integer(ndraws)
  c <- 0
  j <- 0
  for (i in seq_along(w)) {
    c <- c + w[i]
    if (c >= 1) {
      a <- floor(c)
      c <- c - a
      out[j + seq_len(a)] <- i
      j <- j + a
    }
    if (j < ndraws && c >= runif(1)) {
      c <- c - 1
      j <- j + 1
      out[j] <- i
    }
  }
  out
}

# Deterministic resampling
#   Kitagawa, G., Monte Carlo Filter and Smoother for Non-Gaussian
#   Nonlinear State Space Models, Journal of Computational and
#   Graphical Statistics, 5(1):1-25, 1996.
# @return index vector of length 'ndraws'
.resample_deterministic <- function(weights, ndraws, ...) {
  # expected number of repetitions for each original draw
  w <- ndraws * weights
  fw <- floor(w)
  out <- integer(ndraws)
  k <- 0
  c <- 0.5
  for (i in seq_along(w)) {
    if (w[i] >= 1) {
      a <- fw[i]
      w[i] <- w[i] - a
      out[k + seq_len(a)] <- i
      k <-  k + a;
    }
    c <- c + w[i]
    if (c >= 1) {
      k <- k + 1
      out[k] <- i
      c <- c - 1
    }
  }
  out
}

# names of supported resampling methods
supported_resample_methods <- function() {
  c("simple", "simple_no_replace", "stratified", "deterministic")
}

#' Factor random variables of arbitrary dimension
#'
#' Random variables backed by [factor]-like arrays of arbitrary dimension.
#'
#' @name rvar_factor
#'
#' @inheritParams rvar
#' @inheritDotParams base::factor
#'
#' @details
#'
#' A subtype of [rvar()] that represents a (possibly multidimensional) sample of
#' a [factor] or an [ordered] factor. It is otherwise very similar to the basic [rvar()]:
#' it is backed by a multidimensional array with draws as the first dimension.
#' The primary difference is that the backing array has class `"factor"` (for [rvar_factor()])
#' or `c("ordered", "factor")` (for [rvar_ordered()]). If you
#' pass a [factor] or [ordered] factor to [rvar()] it will automatically return
#' an object with the classes `"rvar_factor"` or `c("rvar_ordered", "rvar_factor")`.
#'
#' See [rvar()] for more details on the internals of the random variable datatype.
#'
#' @seealso [as_rvar_factor()] to convert objects to `rvar_factor`s. See [rdo()], [rfun()], and
#' [rvar_rng()] for higher-level interfaces for creating `rvar`s.
#'
#' @return An object of class `"rvar_factor"` representing a `factor`-like random variable.
#'
#' @examples
#' set.seed(1234)
#'
#' # To create a "scalar" `rvar_factor`, pass a one-dimensional array or a vector
#' # whose length (here `4000`) is the desired number of draws:
#' x <- rvar(sample(c("a","a","a","b","c"), 4000, replace = TRUE))
#' x
#'
#' # Create random vectors by adding an additional dimension:
#' x_array <- array(c(
#'     sample(c("a","a","a","b","c"), 4000, replace = TRUE),
#'     sample(c("a","a","b","c","c"), 4000, replace = TRUE),
#'     sample(c("b","b","b","b","c"), 4000, replace = TRUE),
#'     sample(c("d","d","b","b","c"), 4000, replace = TRUE)
#'   ), dim = c(4000, 4))
#' rvar_factor(x_array)
#'
#' # You can also create ordered factors
#' rvar_ordered(x_array)
#'
#' # arguments of factor() and ordered() are passed down by the constructor
#' # e.g. we can reorder levels of an ordered factor:
#' rvar_ordered(x_array, levels = c("d","c","b","a"))
#'
#' # Unlike base factors, rvar factors can be matrices or arrays:
#' rvar_factor(x_array, dim = c(2, 2))
#'
#' # If the input to rvar_factor() is an array with a `"levels"` attribute, it
#' # will use those as the levels of the factor
#' y_array <- t(array(rbinom(3000, 1, c(0.1, 0.5, 0.9)) + 1, dim = c(3, 1000)))
#' rvar(y_array)
#' # with levels
#' attr(y_array, "levels") = c("a", "b")
#' rvar_factor(y_array)
#'
#' @export
rvar_factor <- function(
  x = factor(), dim = NULL, dimnames = NULL, nchains = NULL, with_chains = FALSE, ...
) {

  # to ensure we pick up levels already attached to x (if there are any), we
  # need to convert x to a factor here if it has levels
  if (!is.factor(x) && !is.null(attr(x, "levels"))) {
    x <- while_preserving_dims(factor, x, labels = attr(x, "levels"))
  }

  out <- rvar(
    x, dim = dim, dimnames = dimnames, nchains = nchains, with_chains = with_chains
  )
  .rvar_to_rvar_factor(out, ...)
}

#' @rdname rvar_factor
#' @export
rvar_ordered <- function(
  x = ordered(NULL), dim = NULL, dimnames = NULL, nchains = NULL, with_chains = FALSE, ...
) {

  rvar_factor(
    x, dim = dim, dimnames = dimnames, nchains = nchains, with_chains = with_chains, ordered = TRUE, ...
  )
}

# factor-like rvar methods ------------------------------------------------

#' @export
levels.rvar <- function(x) {
  levels(draws_of(x))
}

#' @export
`levels<-.rvar` <- function(x, value) {
  levels(draws_of(x)) <- value
  x
}


# type predicates and casting ---------------------------------------------------------

#' Is `x` a factor random variable?
#'
#' Test if `x` is an [`rvar_factor`] or [`rvar_ordered`].
#'
#' @inheritParams is_rvar
#'
#' @seealso [as_rvar_factor()] and [as_rvar_ordered()] to convert objects to
#' `rvar_factor`s and `rvar_ordered`s.
#'
#' @return `TRUE` if `x` is an [`rvar_factor`] or [`rvar_ordered`], `FALSE` otherwise.
#'
#' @export
is_rvar_factor <- function(x) {
  inherits(x, "rvar_factor")
}

#' @rdname is_rvar_factor
#' @export
is_rvar_ordered <- function(x) {
  inherits(x, "rvar_ordered")
}

#' Coerce to a factor random variable
#'
#' Convert `x` to an [`rvar_factor`] or [`rvar_ordered`] object.
#'
#' @inheritParams as_rvar
#' @inheritDotParams base::factor
#'
#' @details For objects that are already [`rvar`]s, returns them (with modified dimensions
#' if `dim` is not `NULL`), possibly adding levels using the unique values of the draws of
#' the `rvar` (if the object is not already factor-like).
#'
#' For numeric, logical, factor, or character vectors or arrays, returns an [`rvar_factor`]
#' or [`rvar_ordered`] with a single draw and the same dimensions as `x`. This is in contrast
#' to the [rvar_factor()] and [rvar_ordered()] constructors, which treats the first dimension
#' of `x` as the draws dimension. As a result, `as_rvar_factor()` and `as_rvar_ordered()`
#' are useful for creating constants.
#'
#' @seealso [rvar()], [rvar_factor()], and [rvar_ordered()] to construct [`rvar`]s directly.
#' See [rdo()], [rfun()], and [rvar_rng()] for higher-level interfaces for creating `rvar`s.
#'
#' @return An object of class `"rvar_factor"` or `"rvar_ordered"` representing a random variable.
#'
#' @examples
#'
#' # You can use as_rvar_factor() to create "constant" rvars (having only one draw):
#' x <- as_rvar_factor("a")
#' x
#'
#' # Such constants can be of arbitrary shape:
#' as_rvar_factor(letters[1:4])
#' as_rvar_ordered(matrix(letters[1:10], nrow = 5))
#' as_rvar_factor(array(letters[1:12], dim = c(2, 3, 2)))
#'
#' @export
as_rvar_factor <- function(x, dim = NULL, dimnames = NULL, nchains = NULL, ...) {
  out <- .as_rvar(
    x, dim = dim, dimnames = dimnames, nchains = nchains, ptype = new_rvar(factor())
  )
  .rvar_to_rvar_factor(out, ...)
}

#' @rdname as_rvar_factor
#' @export
as_rvar_ordered <- function(x, dim = NULL, dimnames = NULL, nchains = NULL, ...) {
  out <- .as_rvar(
    x, dim = dim, dimnames = dimnames, nchains = nchains, ptype = new_rvar(ordered(NULL))
  )
  .rvar_to_rvar_factor(out, ordered = TRUE, ...)
}


# misc standard methods ---------------------------------------------------

#' @export
unique.rvar_factor <- function(x, incomparables = FALSE, MARGIN = 1, ...) {
  draws_margin <- check_rvar_margin(x, MARGIN)

  .draws <- draws_of(x)
  .levels <- levels(.draws)
  .class <- oldClass(.draws)

  .draws <- unique(unclass(.draws), incomparables = incomparables, MARGIN = draws_margin, ...)

  levels(.draws) <- .levels
  oldClass(.draws) <- .class
  draws_of(x) <- .draws

  x
}

#' @export
duplicated.rvar_factor <- function(x, incomparables = FALSE, MARGIN = 1, ...) {
  draws_margin <- check_rvar_margin(x, MARGIN)
  duplicated(unclass(draws_of(x)), incomparables = incomparables, MARGIN = draws_margin, ...)
}

#' @export
anyDuplicated.rvar_factor <- function(x, incomparables = FALSE, MARGIN = 1, ...) {
  draws_margin <- check_rvar_margin(x, MARGIN)
  anyDuplicated(unclass(draws_of(x)), incomparables = incomparables, MARGIN = draws_margin, ...)
}


# helpers for factor levels -----------------------------------------------

#' Given a list of levels and an output rvar (which is derived from rvars with those levels
#' bound together in some way, e.g. with c() or vec_c() or whatever),
#' combine the factor levels together appropriately and update out as needed
#' @param x an rvar_factor
#' @param list_of_levels a list of character vectors, each a set of levels
#' @param ordered should the output be an ordered factor if possible?
#' @noRd
combine_rvar_factor_levels <- function(x, list_of_levels, ordered = FALSE) {
  .draws <- draws_of(x)

  unique_levels <- unique(list_of_levels)
  # zero-length levels lists don't count (since can only come from factors with only missing values)
  unique_levels <- unique_levels[lengths(unique_levels) > 0]
  if (length(unique_levels) <= 1) {
    # levels are the same in all variables, so preserve level order when binding
    .levels <- unique_levels[1][[1]]
    # We only keep the "ordered" class when the levels were all the same (this
    # mimics base-R, which demotes to unordered factor when combining ordered
    # factors with different levels)
    .draws <- while_preserving_dims(factor, .draws, .levels, ordered = ordered)
  } else {
    # levels are not the same in all variables, so preserve any old levels by
    # merging them together, but do not apply the "ordered" class
    .levels <- unique(do.call(c, list_of_levels))
    .draws <- while_preserving_dims(factor, .draws, .levels)
  }
  if (!is.factor(.draws)) {
    .draws <- while_preserving_dims(factor, .draws)
  }
  draws_of(x) <- .draws

  x
}

# as_rvar -----------------------------------------------------------------

test_that("as_rvar works", {
  expect_equal(draws_of(as_rvar(1L)), matrix(1L, dimnames = list("1", NULL)))
  expect_equal(draws_of(as_rvar(c(TRUE, FALSE))), matrix(c(TRUE, FALSE), nrow = 1, dimnames = list("1", NULL)))
  expect_equal(draws_of(as_rvar(1:3L)), matrix(1:3L, nrow = 1, dimnames = list("1", NULL)))

  expect_equal(nchains(as_rvar(1, nchains = 2)), 2)

  expect_equal(draws_of(as_rvar(1:6, dim = c(2,3))), array(1:6, dim = c(1,2,3), dimnames = list("1", NULL, NULL)))
  expect_equal(
    draws_of(as_rvar(1:6, dim = c(2,3), dimnames = list(letters[1:2], letters[1:3]))),
    array(1:6, dim = c(1,2,3), dimnames = list("1", letters[1:2], letters[1:3]))
  )

  expect_equal(
    draws_of(as_rvar(factor(letters[1:3], levels = letters[3:1]))),
    structure(matrix(3:1, nrow = 1, dimnames = list("1", NULL)), levels = letters[3:1], class = "factor")
  )
  expect_equal(
    draws_of(as_rvar(ordered(letters[1:3], levels = letters[3:1]))),
    structure(matrix(3:1, nrow = 1, dimnames = list("1", NULL)), levels = letters[3:1], class = c("ordered", "factor"))
  )
})

test_that("as_rvar preserves dimension names", {
  m <- diag(1:3)
  dimnames(m) <- list(a = paste0("a", 1:3), b = paste0("b", 1:3))
  m_rvar <- as_rvar(m)
  expect_equal(dimnames(m_rvar), dimnames(m))

  x <- 1:3
  names(x) <- c("a","b","c")
  x_rvar <- as_rvar(x)
  expect_equal(names(x_rvar), names(x))
})


# as_rvar_numeric ---------------------------------------------------------

test_that("as_rvar_numeric works", {
  x_array = array(
    1:24, dim = c(2,4,3),
    dimnames = list(NULL, A = paste0("a", 1:4), B = paste0("b", 1:3))
  )
  x <- rvar(x_array)
  x_array_letters = array(
    letters[1:24], dim = c(2,4,3),
    dimnames = list(NULL, A = paste0("a", 1:4), B = paste0("b", 1:3))
  )
  x_fct <- rvar_factor(x_array)
  x_ord <- rvar_ordered(x_array)

  expect_equal(as_rvar_numeric(x_fct), x)
  expect_equal(as_rvar_numeric(x_ord), x)
})


# as_rvar_factor -----------------------------------------------------------------

test_that("as_rvar_factor works", {
  expect_equal(
    draws_of(as_rvar_factor(array(1:4, dim = c(2,2)))),
    structure(1:4L, levels = c("1", "2", "3", "4"), dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = "factor")
  )
  expect_equal(
    draws_of(as_rvar_factor(array(c(TRUE, TRUE, FALSE, FALSE), dim = c(2,2)))),
    structure(c(2, 2, 1, 1), levels = c("FALSE", "TRUE"), dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = "factor")
  )
  expect_equal(
    draws_of(as_rvar_factor(array(letters[1:4], dim = c(2,2)))),
    structure(1:4L, levels = letters[1:4], dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = "factor")
  )
  expect_equal(
    draws_of(as_rvar_factor(`dim<-`(factor(letters[1:4], levels = letters[4:1]), c(2,2)))),
    structure(4:1, dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), levels = letters[4:1], class = "factor")
  )
  expect_equal(
    draws_of(as_rvar_factor(`dim<-`(ordered(letters[1:4], levels = letters[4:1]), c(2,2)))),
    structure(4:1, dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), levels = letters[4:1], class = c("ordered", "factor"))
  )

  expect_equal(nchains(as_rvar_factor(1, nchains = 2)), 2)

  expect_equal(
    as_rvar_factor(rvar(array(1:12, dim = c(2,2,3)))),
    rvar_factor(array(as.character(1:12), dim = c(2,2,3)), levels = as.character(1:12))
  )
})

test_that("as_rvar_ordered works", {
  expect_equal(
    draws_of(as_rvar_ordered(array(1:4, dim = c(2,2)))),
    structure(1:4L, levels = c("1", "2", "3", "4"), dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = c("ordered", "factor"))
  )
  expect_equal(
    draws_of(as_rvar_ordered(as_rvar(array(1:4, dim = c(2,2))))),
    structure(1:4L, levels = c("1", "2", "3", "4"), dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = c("ordered", "factor"))
  )
  expect_equal(
    draws_of(as_rvar_ordered(as_rvar_factor(array(1:4, dim = c(2,2))))),
    structure(1:4L, levels = c("1", "2", "3", "4"), dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = c("ordered", "factor"))
  )
  expect_equal(
    draws_of(as_rvar_ordered(array(c(TRUE, TRUE, FALSE, FALSE), dim = c(2,2)))),
    structure(c(2, 2, 1, 1), levels = c("FALSE", "TRUE"), dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = c("ordered", "factor"))
  )
  expect_equal(
    draws_of(as_rvar_ordered(array(letters[1:4], dim = c(2,2)))),
    structure(1:4L, levels = letters[1:4], dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = c("ordered", "factor"))
  )
  expect_equal(
    draws_of(as_rvar_ordered(`dim<-`(factor(letters[1:4], levels = letters[4:1]), c(2,2)))),
    structure(4:1, dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), levels = letters[4:1], class = c("ordered", "factor"))
  )
  expect_equal(
    draws_of(as_rvar_ordered(`dim<-`(ordered(letters[1:4], levels = letters[4:1]), c(2,2)))),
    structure(4:1, dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), levels = letters[4:1], class = c("ordered", "factor"))
  )

  expect_equal(nchains(as_rvar_ordered(1, nchains = 2)), 2)

  expect_equal(
    as_rvar_ordered(rvar(array(1:12, dim = c(2,2,3)))),
    rvar_ordered(array(as.character(1:12), dim = c(2,2,3)), levels = as.character(1:12))
  )
})

test_that("as_rvar(<character>) produces an rvar_factor ", {
  expect_equal(
    draws_of(as_rvar(array(letters[1:4], dim = c(2,2)))),
    structure(1:4L, levels = letters[1:4], dim = c(1, 2, 2), dimnames = list("1", NULL, NULL), class = "factor")
  )
})

# casting to/from rvar/distribution ---------------------------------------

test_that("casting to/from rvar/distribution objects works", {
  x_dist <- distributional::dist_sample(list(a = c(1,1), b = 3:4))
  null_dist <- vctrs::vec_ptype(x_dist)
  x_rvar <- rvar(matrix(c(1,1,3:4), ncol = 2, dimnames = list(NULL, c("a","b"))))

  # casting to rvar
  expect_equal(vctrs::vec_cast(x_dist, rvar()), x_rvar)
  expect_equal(as_rvar(x_dist), x_rvar)

  # casting to rvar with a broadcast
  x_dist_bc <- distributional::dist_sample(list(a = 1, b = 3:4))
  expect_equal(vctrs::vec_cast(x_dist_bc, rvar()), x_rvar)

  # can't cast non-sample distributions to rvar
  expect_error(vctrs::vec_cast(distributional::dist_normal(), rvar()))

  # can't cast samples of incompatible sizes to rvar
  expect_error(vctrs::vec_cast(distributional::dist_sample(list(1:3, 1:2)), rvar()))

  # casting to distribution
  expect_equal(vctrs::vec_cast(x_rvar, null_dist), x_dist)

  # can't cast multivariate rvars to distributions
  x_mv <- rvar(array(1:8, dim = c(2,2,2)))
  expect_error(vctrs::vec_cast(x_mv, null_dist))
})


# type predicates ---------------------------------------------------------

test_that("is.matrix/array on rvar works", {
  x_mat <- rvar(array(1:24, dim = c(2,2,6)))
  x_arr <- rvar(array(1:24, dim = c(2,2,3,2)))

  expect_true(is.matrix(x_mat))
  expect_true(is.array(x_mat))
  expect_false(is.matrix(x_arr))
  expect_true(is.array(x_arr))
})

test_that("vec_is_list(<rvar>) is FALSE", {
  expect_equal(vctrs::vec_is_list(rvar()), FALSE)
})


# type conversion -----------------------------------------------------------

test_that("as.list works", {
  x_array <- array(
    1:24, dim = c(2,4,3),
    dimnames = list(NULL, A = paste0("a", 1:4), B = paste0("b", 1:3))
  )
  x <- new_rvar(x_array)

  expect_equal(as.list(x),
    list(
      a1 = new_rvar(x_array[,1,]),
      a2 = new_rvar(x_array[,2,]),
      a3 = new_rvar(x_array[,3,]),
      a4 = new_rvar(x_array[,4,])
    )
  )

  x_array <- array(
    1:12, dim = c(6, 2),
    dimnames = list(NULL, A = paste0("a", 1:2))
  )
  x <- new_rvar(x_array)

  expect_equal(as.list(x),
    list(
      a1 = new_rvar(x_array[,1]),
      a2 = new_rvar(x_array[,2])
    )
  )
})

test_that("as.vector works", {
  x = rvar(array(1:12, dim = c(2, 2, 3)))
  dimnames(x) <- list(c("a","b"), c("c","d","e"))

  expect_equal(as.vector(x), rvar(array(1:12, dim = c(2, 6))))
})

test_that("as.data.frame and as_tibble work on rvars", {
  x1 = rvar(array(1:9, dim = c(3,3)),
    dimnames = list(A = paste0("a", 1:3))
  )
  x2 = rvar(array(1:12, dim = c(2,2,3)),
    dimnames = list(A = paste0("a", 1:2), B = paste0("b", 1:3))
  )
  x3 = rvar(array(1:24, dim = c(2,2,2,4)),
    dimnames = list(A = paste0("a", 1:2), B = paste0("b", 1:2), C = paste0("c", 1:4))
  )

  # constructing reference data frames with rvars in them without having that
  # code call as.data.frame() (defeating the purpose of the test) requires
  # bypassing the data.frame() constructor being called on an rvar, as it would
  # call as.data.frame.rvar(). Hence the twisty code below.

  # nulls
  df0 <- data.frame()
  df0[["rvar()"]] <- rvar()
  row.names(df0) <- numeric()
  expect_equal(as.data.frame(rvar()), df0)

  tibble0 <- as_tibble(df0)
  names(tibble0) <- "value"
  expect_equal(as_tibble(rvar()), tibble0)

  # 1-dim arrays
  df1 <- as.data.frame(mean(x1))
  names(df1) <- "x1"
  df1$x1 <- x1
  dimnames(df1$x1)["A"] <- list(NULL)
  expect_equal(as.data.frame(x1), df1)

  tibble1 <- as_tibble(df1)
  names(tibble1) <- "value"
  expect_equal(as_tibble(x1), tibble1)

  # 2-dim arrays
  df2 <- as.data.frame(mean(x2))
  for (i in 1:3) {
    col <- x2[,i,drop = TRUE]
    dimnames(col) <- list(NULL)
    df2[[i]] <- col
  }
  expect_equal(as.data.frame(x2), df2)
  expect_equal(dimnames(as.data.frame(unname(x2))), dimnames(as.data.frame(mean(unname(x2)))))

  tibble2 <- as_tibble(df2)
  expect_equal(as_tibble(x2), tibble2)

  # 3-dim arrays
  df3 <- as.data.frame(mean(x3))
  for (c_i in 1:4) for (b_i in 1:2) {
    col <- x3[,b_i,c_i,drop = TRUE]
    dimnames(col) <- list(NULL)
    df3[[b_i + (c_i - 1) * 2]] <- col
  }
  expect_equal(as.data.frame(x3), df3)
  expect_equal(dimnames(as.data.frame(unname(x3))), dimnames(as.data.frame(mean(unname(x3)))))

  tibble3 <- as_tibble(df3)
  expect_equal(as_tibble(x3), tibble3)
})

test_that("as.character works", {
  x <- rvar(c(1,1))
  expect_equal(as.character(x), format(x))
  x <- rvar_factor(letters[1:2])
  expect_equal(as.character(x), format(x))
  x <- rvar_ordered(letters[1:2])
  expect_equal(as.character(x), format(x))
})

test_that("proxy restore works", {
  x1 = rvar(array(1:9, dim = c(3,3)),
    dimnames = list(A = paste0("a", 1:3))
  )
  x2 = rvar(array(1:12, dim = c(2,2,3)),
    dimnames = list(A = paste0("a", 1:2), B = paste0("b", 1:3))
  )
  x3 = rvar(array(1:24, dim = c(2,2,2,4)),
    dimnames = list(A = paste0("a", 1:2), B = paste0("b", 1:2), C = paste0("c", 1:4))
  )

  expect_equal(vec_restore(vec_proxy(unname(x1)), rvar()), unname(x1))
  expect_equal(vec_restore(vec_proxy(unname(x2)), rvar()), unname(x2))
  expect_equal(vec_restore(vec_proxy(unname(x3)), rvar()), unname(x3))
  expect_equal(vec_restore(vec_proxy(x1), rvar()), x1)
  expect_equal(vec_restore(vec_proxy(x2), rvar()), x2)
  expect_equal(vec_restore(vec_proxy(x3), rvar()), x3)

  expect_equal(
    vec_restore(c(
      vec_proxy(vctrs::vec_init(rvar(), 2)),
      vec_proxy(rvar(1:10, nchains = 2))
    ), rvar()),
    rvar(array(c(rep(NA, 10), rep(NA, 10), 1:10), dim = c(10, 3)), nchains = 2)
  )

  expect_equal(
    vec_restore(c(
      vec_proxy(vctrs::vec_init(rvar_factor(), 2)),
      vec_proxy(rvar_factor(letters[1:10], nchains = 2))
    ), rvar_factor()),
    rvar_factor(array(letters[c(rep(NA, 10), rep(NA, 10), 1:10)], dim = c(10, 3)), nchains = 2)
  )

  expect_equal(
    vec_restore(c(
      vec_proxy(vctrs::vec_init(rvar_ordered(), 2)),
      vec_proxy(rvar_ordered(letters[1:10], nchains = 2))
    ), rvar_ordered()),
    rvar_ordered(array(letters[c(rep(NA, 10), rep(NA, 10), 1:10)], dim = c(10, 3)), nchains = 2)
  )
})

# c.rvar ------------------------------------------------------------------

test_that("c works on rvar", {
  x <- rvar(array(1:9, dim = c(3,3)))
  y <- rvar(array(2:10, dim = c(3,3), dimnames = list(NULL, c("a","b","c"))))
  x_y <- rvar(array(c(1:9, 2:10), dim = c(3,6), dimnames = list(NULL, c("","","","a","b","c"))))

  expect_equal(c(x), x)
  expect_equal(c(x, NULL), x)
  expect_equal(c(x, y), x_y)
  expect_equal(c(x, NULL, y), x_y)
  expect_equal(c(x, x), rvar(array(c(1:9, 1:9), dim = c(3,6))))

  expect_equal(c(z = x),
    rvar(array(1:9, dim = c(3,3), dimnames = list(NULL, c("z1","z2","z3"))))
  )

  expect_equal(c(z = c(x, y)),
    rvar(array(c(1:9, 2:10), dim = c(3,6), dimnames = list(NULL, c("z1","z2","z3","z.a","z.b","z.c"))))
  )

  expect_equal(c(x, 5), rvar(array(c(1:9, 5, 5, 5), dim = c(3,4))))
  expect_equal(vctrs::vec_c(5, x), rvar(array(c(5, 5, 5, 1:9), dim = c(3,4))))
  expect_equal(c(x, 5L), rvar(array(c(1:9, 5, 5, 5), dim = c(3,4))))
  expect_equal(vctrs::vec_c(5L, x), rvar(array(c(5, 5, 5, 1:9), dim = c(3,4))))

  expect_equal(c(x == 1, TRUE), rvar(array(c(1:9 == 1, TRUE, TRUE, TRUE), dim = c(3,4))))
  expect_equal(vctrs::vec_c(TRUE, x == TRUE), rvar(array(c(TRUE, TRUE, TRUE, 1:9 == 1), dim = c(3,4))))

  x_col <- x
  dim(x_col) <- c(3,1)
  expect_equal(c(x_col), x)
  expect_equal(c(x_col, y), x_y)
})

test_that("c works on rvar_factor", {
  x <- rvar(array(letters[1:9], dim = c(3,3)))
  y <- rvar(array(letters[2:10], dim = c(3,3), dimnames = list(NULL, c("a","b","c"))))
  x_y <- rvar(array(letters[c(1:9, 2:10)], dim = c(3,6), dimnames = list(NULL, c("","","","a","b","c"))))

  expect_equal(c(x), x)
  expect_equal(c(x, NULL), x)
  expect_equal(c(x, y), x_y)
  expect_equal(c(x, NULL, y), x_y)
  expect_equal(c(x, x), rvar(array(letters[c(1:9, 1:9)], dim = c(3,6))))

  expect_equal(c(x, "a"),
    rvar(array(c(letters[1:9], "a", "a", "a"), dim = c(3,4)))
  )
  expect_equal(vctrs::vec_c("a", x),
    rvar(array(c("a", "a", "a", letters[1:9]), dim = c(3,4)))
  )
  expect_equal(c(x, factor("a")),
    rvar(array(c(letters[1:9], "a", "a", "a"), dim = c(3,4)))
  )
  expect_equal(vctrs::vec_c(factor("a"), x),
    rvar(array(c("a", "a", "a", letters[1:9]), dim = c(3,4)))
  )
  expect_equal(c(x, "xx"),
    rvar(array(c(letters[1:9], "xx", "xx", "xx"), dim = c(3,4)))
  )
  expect_equal(c(x, factor("xx")),
    rvar(array(c(letters[1:9], "xx", "xx", "xx"), dim = c(3,4)))
  )
  expect_equal(c(x, ordered("xx")),
    rvar(array(c(letters[1:9], "xx", "xx", "xx"), dim = c(3,4)))
  )
  expect_equal(vctrs::vec_c(ordered("xx"), x),
    rvar_factor(array(c("xx", "xx", "xx", letters[1:9]), dim = c(3,4)), levels = c("xx", letters[1:9]))
  )

  expect_error(c(x, 5), "Can't combine")
  expect_error(c(x, 5L), "Can't combine")
  expect_error(c(rvar(), factor("a")), "Can't combine")
  expect_error(vctrs::vec_c(factor("a"), rvar()), "Can't combine")

  x_col <- x
  dim(x_col) <- c(3,1)
  expect_equal(c(x_col), x)
  expect_equal(c(x_col, y), x_y)
})

test_that("c works on rvar_ordered", {
  x <- rvar_ordered(array(letters[1:9], dim = c(3,3)), levels = letters)
  y <- rvar_ordered(array(letters[2:10], dim = c(3,3), dimnames = list(NULL, c("a","b","c"))), levels = letters)
  x_y <- rvar_ordered(
    array(letters[c(1:9, 2:10)], dim = c(3,6), dimnames = list(NULL, c("","","","a","b","c"))), levels = letters
  )

  expect_equal(c(x), x)
  expect_equal(c(x, NULL), x)
  expect_equal(c(x, y), x_y)
  expect_equal(c(x, NULL, y), x_y)
  expect_equal(c(x, x), rvar_ordered(array(letters[c(1:9, 1:9)], dim = c(3,6)), levels = letters))

  # demotion to factor when levels aren't all in common
  x_partial <- rvar_ordered(array(letters[1:9], dim = c(3,3)))
  y_partial <- rvar_ordered(array(letters[2:10], dim = c(3,3), dimnames = list(NULL, c("a","b","c"))))
  x_y_fct <- rvar_factor(array(letters[c(1:9, 2:10)], dim = c(3,6), dimnames = list(NULL, c("","","","a","b","c"))))
  expect_equal(c(x, y), x_y)

  expect_equal(
    c(x, "a"),
    rvar_ordered(array(c(letters[1:9], "a", "a", "a"), dim = c(3,4)), levels = letters)
  )
  expect_equal(vctrs::vec_c("a", x),
    rvar_ordered(array(c("a", "a", "a", letters[1:9]), dim = c(3,4)), levels = letters)
  )
  expect_equal(c(x, factor("a")),
    rvar_factor(array(c(letters[1:9], "a", "a", "a"), dim = c(3,4)), levels = letters)
  )
  expect_equal(vctrs::vec_c(factor("a"), x),
    rvar_factor(array(c("a", "a", "a", letters[1:9]), dim = c(3,4)), levels = letters)
  )
  expect_equal(c(x, "xx"),
    rvar_factor(array(c(letters[1:9], "xx", "xx", "xx"), dim = c(3,4)), levels = c(letters, "xx"))
  )
  expect_equal(c(x, factor("xx")),
    rvar_factor(array(c(letters[1:9], "xx", "xx", "xx"), dim = c(3,4)), levels = c(letters, "xx"))
  )
  expect_equal(c(x, ordered("xx")),
    rvar_factor(array(c(letters[1:9], "xx", "xx", "xx"), dim = c(3,4)), levels = c(letters, "xx"))
  )
  expect_equal(vctrs::vec_c(ordered("xx"), x),
    rvar_factor(array(c("xx", "xx", "xx", letters[1:9]), dim = c(3,4)), levels = c("xx", letters))
  )
  expect_equal(vctrs::vec_c(ordered("a"), x),
    rvar_factor(array(c("a", "a", "a", letters[1:9]), dim = c(3,4)), levels = letters)
  )
  expect_equal(vctrs::vec_c(ordered("a", levels = letters), x),
    rvar_ordered(array(c("a", "a", "a", letters[1:9]), dim = c(3,4)), levels = letters)
  )

  expect_error(c(x, 5), "Can't combine")
  expect_error(c(x, 5L), "Can't combine")
  expect_error(c(rvar(), ordered("a")), "Can't combine")
  expect_error(vctrs::vec_c(ordered("a"), rvar()), "Can't combine")

  x_col <- x
  dim(x_col) <- c(3,1)
  expect_equal(c(x_col), x)
  expect_equal(c(x_col, y), x_y)
})


# cbind.rvar --------------------------------------------------------------

test_that("cbind works on rvar", {
  x = rvar(array(1:9, dim = c(3,3)))
  y = rvar(array(2:10, dim = c(3,3)))

  expect_equal(cbind(rvar(array(1:9, dim = c(3,3)))), rvar(array(1:9, dim = c(3,3,1))))
  expect_equal(cbind(x), rvar(array(1:9, dim = c(3,3,1), dimnames = list(NULL, NULL, "x"))))

  expect_equal(cbind(x, y, deparse.level = 0), rvar(array(c(1:9, 2:10), dim = c(3,3,2))))
  expect_equal(cbind(a = x, y, deparse.level = 0),
    rvar(array(c(1:9, 2:10), dim = c(3,3,2), dimnames = list(NULL, NULL, c("a", ""))))
  )
  expect_equal(cbind(a = x, y),
    rvar(array(c(1:9, 2:10), dim = c(3,3,2), dimnames = list(NULL, NULL, c("a", "y"))))
  )
  expect_equal(cbind(x, b = y, deparse.level = 0),
    rvar(array(c(1:9, 2:10), dim = c(3,3,2), dimnames = list(NULL, NULL, c("", "b"))))
  )
  expect_equal(cbind(x, y + 1, deparse.level = 2),
    rvar(array(c(1:9, 2:10 + 1), dim = c(3,3,2), dimnames = list(NULL, NULL, c("x", "y + 1"))))
  )
  expect_equal(cbind(x, y, y + 1, deparse.level = 2),
    rvar(array(c(1:9, 2:10, 2:10 + 1), dim = c(3,3,3), dimnames = list(NULL, NULL, c("x", "y", "y + 1"))))
  )
  expect_equal(cbind(x, y, y + 1, deparse.level = 2), cbind(cbind(x, y), y + 1, deparse.level = 2))

  x_col <- x
  dim(x_col) <- c(3,1)
  expect_equal(cbind(x_col, y, deparse.level = 0), rvar(array(c(1:9, 2:10), dim = c(3,3,2))))
  expect_equal(cbind(a = x_col, y),
    rvar(array(c(1:9, 2:10), dim = c(3,3,2), dimnames = list(NULL, NULL, c("", "y"))))
  )
  dimnames(x_col)[[2]] = "b"
  expect_equal(cbind(a = x_col, y),
    rvar(array(c(1:9, 2:10), dim = c(3,3,2), dimnames = list(NULL, NULL, c("b", "y"))))
  )

  expect_equal(cbind(NULL, x), cbind(x))
  expect_equal(cbind(x, NULL, y), cbind(x, y))
})

test_that("cbind works on rvar_factor", {
  x = rvar(array(letters[1:9], dim = c(3,3)))
  y = rvar(array(letters[2:10], dim = c(3,3)))

  expect_equal(cbind(rvar(array(letters[1:9], dim = c(3,3)))), rvar_factor(array(letters[1:9], dim = c(3,3,1))))
  expect_equal(cbind(x), rvar_factor(array(letters[1:9], dim = c(3,3,1), dimnames = list(NULL, NULL, "x"))))

  expect_equal(cbind(x, y, deparse.level = 0), rvar_factor(array(letters[c(1:9, 2:10)], dim = c(3,3,2))))
  expect_equal(cbind(a = x, y, deparse.level = 0),
    rvar_factor(array(letters[c(1:9, 2:10)], dim = c(3,3,2), dimnames = list(NULL, NULL, c("a", ""))))
  )
  expect_equal(cbind(a = x, y),
    rvar_factor(array(letters[c(1:9, 2:10)], dim = c(3,3,2), dimnames = list(NULL, NULL, c("a", "y"))))
  )
  expect_equal(cbind(x, b = y, deparse.level = 0),
    rvar_factor(array(letters[c(1:9, 2:10)], dim = c(3,3,2), dimnames = list(NULL, NULL, c("", "b"))))
  )

  x_col <- x
  dim(x_col) <- c(3,1)
  expect_equal(cbind(x_col, y, deparse.level = 0), rvar_factor(array(letters[c(1:9, 2:10)], dim = c(3,3,2))))
  expect_equal(cbind(a = x_col, y),
    rvar(array(letters[c(1:9, 2:10)], dim = c(3,3,2), dimnames = list(NULL, NULL, c("", "y"))))
  )
  dimnames(x_col)[[2]] = "b"
  expect_equal(cbind(a = x_col, y),
    rvar_factor(array(letters[c(1:9, 2:10)], dim = c(3,3,2), dimnames = list(NULL, NULL, c("b", "y"))))
  )

  expect_equal(cbind(NULL, x), cbind(x))
  expect_equal(cbind(x, NULL, y), cbind(x, y))
})

test_that("cbind works on rvar with data frames", {
  # these do not work on R < 4 for some reason related to how data frames
  # handle binding (so, not much we can do about it?)
  skip_if_not(R.version$major >= 4)

  x = rvar(array(1:9, dim = c(3,3)))
  y = rvar(array(2:10, dim = c(3,3)))

  expect_equal(cbind(data.frame(x), y + 1), data.frame(x = x, `y + 1` = y + 1, check.names = FALSE))
  expect_equal(cbind(x + 1, data.frame(y)), data.frame(`x + 1` = x + 1, y = y, check.names = FALSE))
})


# rbind.rvar --------------------------------------------------------------

test_that("rbind works on rvar", {
  x <- rvar(array(1:9, dim = c(3,3)))
  y <- rvar(array(2:10, dim = c(3,3)))
  x_y_array <- abind(
    array(1:9, dim = c(3,1,3)),
    array(2:10, dim = c(3,1,3)),
    along = 2
  )
  x_yp1_array <- abind(
    array(1:9, dim = c(3,1,3)),
    array(2:10 + 1, dim = c(3,1,3)),
    along = 2
  )

  expect_equal(rbind(rvar(array(1:9, dim = c(3,3)))), rvar(array(1:9, dim = c(3,1,3))))
  expect_equal(rbind(x), rvar(array(1:9, dim = c(3,1,3), dimnames = list(NULL, "x", NULL))))

  expect_equal(rbind(x, y, deparse.level = 0), rvar(x_y_array))
  expect_equal(rbind(a = x, y, deparse.level = 0), rvar(x_y_array, dimnames = list(c("a",""), NULL)))
  expect_equal(rbind(a = x, y), rvar(x_y_array, dimnames = list(c("a","y"), NULL)))
  expect_equal(rbind(x, b = y, deparse.level = 0), rvar(x_y_array, dimnames = list(c("","b"), NULL)))
  expect_equal(rbind(x, y + 1, deparse.level = 2), rvar(x_yp1_array, dimnames = list(c("x","y + 1"), NULL)))
  expect_equal(rbind(x, y, y + 1, deparse.level = 2), rbind(rbind(x, y), y + 1, deparse.level = 2))

  x_row <- x
  dim(x_row) <- c(1,3)
  expect_equal(rbind(x_row, y, deparse.level = 0), rvar(x_y_array))
  expect_equal(rbind(a = x_row, y), rvar(x_y_array, dimnames = list(c("","y"), NULL)))
  dimnames(x_row)[[1]] = "b"
  expect_equal(rbind(a = x_row, y), rvar(x_y_array, dimnames = list(c("b","y"), NULL)))

  expect_equal(rbind(NULL, x), rbind(x))
  expect_equal(rbind(x, NULL, y), rbind(x, y))

  expect_equal(rbind(data.frame(x), data.frame(x = y)), data.frame(x = c(x, y)))
})

test_that("rbind works on rvar_factor", {
  x <- rvar_factor(array(letters[1:9], dim = c(3,3)))
  y <- rvar_factor(array(letters[2:10], dim = c(3,3)))
  x_y_array <- abind(
    array(letters[1:9], dim = c(3,1,3)),
    array(letters[2:10], dim = c(3,1,3)),
    along = 2
  )

  expect_equal(rbind(rvar(array(letters[1:9], dim = c(3,3)))), rvar_factor(array(letters[1:9], dim = c(3,1,3))))
  expect_equal(rbind(x), rvar(array(letters[1:9], dim = c(3,1,3), dimnames = list(NULL, "x", NULL))))

  expect_equal(rbind(x, y, deparse.level = 0), rvar_factor(x_y_array))
  expect_equal(rbind(a = x, y, deparse.level = 0), rvar_factor(x_y_array, dimnames = list(c("a",""), NULL)))
  expect_equal(rbind(a = x, y), rvar_factor(x_y_array, dimnames = list(c("a","y"), NULL)))
  expect_equal(rbind(x, b = y, deparse.level = 0), rvar_factor(x_y_array, dimnames = list(c("","b"), NULL)))

  x_row <- x
  dim(x_row) <- c(1,3)
  expect_equal(rbind(x_row, y, deparse.level = 0), rvar_factor(x_y_array))
  expect_equal(rbind(a = x_row, y), rvar_factor(x_y_array, dimnames = list(c("","y"), NULL)))
  dimnames(x_row)[[1]] = "b"
  expect_equal(rbind(a = x_row, y), rvar_factor(x_y_array, dimnames = list(c("b","y"), NULL)))

  expect_equal(rbind(NULL, x), rbind(x))
  expect_equal(rbind(x, NULL, y), rbind(x, y))

  # rbind does not work with rvar_factor because it treats anything with non-null
  # levels() as a factor and skips the rvar code entirely, causing an error. So
  # we'll test with vctrs::vec_rbind() instead
  expect_equal(vctrs::vec_rbind(data.frame(x), data.frame(x = y)), data.frame(x = c(x, y)))
})

# print -------------------------------------------------------------------

test_that("print(<rvar>) works", {
  x <- rvar(array(1:12, dim = c(2,2,3)))
  x_with_chains <- rvar(array(1:12, dim = c(2,2,3)), nchains = 2)

  out <- capture.output(print(rvar(), color = FALSE))
  expect_match(
    out,
    regexp = "rvar<1>\\[0\\] mean . sd:",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "rvar\\(\\)",
    all = FALSE
  )

  out <- capture.output(print(x, color = FALSE))
  expect_match(
    out,
    regexp = "rvar<2>\\[2,3\\] mean . sd:",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "     \\[,1\\]         \\[,2\\]         \\[,3\\]        ",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "\\[1,\\]  1.5 . 0.71   5.5 . 0.71   9.5 . 0.71 ",
    all = FALSE
  )
  out <- capture.output(print(x_with_chains, color = FALSE))
  expect_match(
    out,
    regexp = "rvar<1,2>\\[2,3\\] mean . sd:",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "     \\[,1\\]         \\[,2\\]         \\[,3\\]        ",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "\\[1,\\]  1.5 . 0.71   5.5 . 0.71   9.5 . 0.71 ",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "\\[2,\\]  3.5 . 0.71   7.5 . 0.71  11.5 . 0.71",
    all = FALSE
  )
})

test_that("print(<rvar_factor>) works", {
  x <- rvar_factor(array(letters[1:12], dim = c(2,2,3)), nchains = 2)

  out <- capture.output(print(x, color = FALSE))
  expect_match(
    out,
    regexp = "rvar_factor<1,2>\\[2,3\\] mode <entropy>:",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "\\[1,\\] a <0\\.28>  e <0\\.28>  i <0\\.28> ",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "\\[2,\\] c <0\\.28>  g <0\\.28>  k <0\\.28> ",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "12 levels: a b c d e f g h i j k l",
    all = FALSE
  )
})

test_that("print(<rvar_ordered>) works", {
  x <- rvar_ordered(array(letters[1:12], dim = c(2,2,3)), nchains = 2)

  out <- capture.output(print(x, color = FALSE))
  expect_match(
    out,
    regexp = "rvar_ordered<1,2>\\[2,3\\] mode <dissent>:",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "\\[1,\\] a <0.067>  e <0.067>  i <0.067> ",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "\\[2,\\] c <0.067>  g <0.067>  k <0.067> ",
    all = FALSE
  )
  expect_match(
    out,
    regexp = "12 levels: a < b < c < d < e < f < g < h < i < j < k < l",
    all = FALSE
  )
})


# str ---------------------------------------------------------------------

test_that("str(<rvar>) works", {
  x <- rvar(array(1:24, dim = c(2,3,4)))
  x_with_chains <- rvar(array(1:24, dim = c(2,3,4)), nchains = 2)

  expect_output(str(rvar()),
    " rvar<1>\\[0\\] "
  )
  out <- capture.output(str(rvar(1:3)))
  expect_match(
    out,
    regexp = " rvar<3>\\[1\\]  2 . 1",
    all = FALSE
  )
  out <- capture.output(str(x, vec.len = 5))
  expect_match(
    out,
    regexp = " rvar<2>\\[3,4\\]  1.5 . 0.71  3.5 . 0.71  5.5 . 0.71  7.5 . 0.71  9.5 . 0.71 ...",
    all = FALSE
  )

  x_with_attrs <- x
  dimnames(x_with_attrs)[1] <- list(c("a","b","c"))
  attr(draws_of(x_with_attrs), "foo") <- list(1,2)
  attr(x_with_attrs, "bar") <- list(1,2)

  out <- capture.output(str(x_with_attrs, vec.len = 5))
  expect_match(
    out,
    regexp = " rvar<2>\\[3,4\\]  1.5 . 0.71  3.5 . 0.71  5.5 . 0.71  7.5 . 0.71  9.5 . 0.71 ...",
    all = FALSE
  )
  expect_match(
    out,
    regexp = " - dimnames\\(\\*\\)=List of 2",
    all = FALSE
  )
  expect_match(
    out,
    regexp = '  \\.\\.\\$ : chr \\[1:3\\] "a" "b" "c"',
    all = FALSE
  )
})

test_that("str(<rvar_factor>) works", {
  x <- rvar_factor(array(letters[1:24], dim = c(2,3,4)), nchains = 2)
  x[[1]] <- "a"
  attr(x, "foo") <- "bar"

  out <- capture.output(str(x, vec.len = 5))
  expect_match(
    out,
    regexp = " rvar_factor<1,2>\\[3,4\\]  a <0\\.00>  c <0\\.22>  e <0\\.22>  g <0\\.22>  i <0\\.22> \\.\\.\\.",
    all = FALSE
  )
  expect_match(
    out,
    regexp = " - 24 levels: a b c d \\.\\.\\. x",
    all = FALSE
  )
  expect_match(
    out,
    regexp = ' - attr\\(\\*, "foo"\\)= chr "bar"',
    all = FALSE
  )
})

test_that("str(<rvar_ordered>) works", {
  x <- rvar_ordered(array(letters[1:24], dim = c(2,3,4)), nchains = 2)
  x[[1]] <- "a"
  attr(x, "foo") <- "bar"

  out <- capture.output(str(x, vec.len = 5))
  expect_match(
    out,
    regexp = " rvar_ordered<1,2>\\[3,4\\]  a <0\\.000>  c <0\\.032>  e <0\\.032>  g <0\\.032>  i <0\\.032> \\.\\.\\.",
    all = FALSE
  )
  expect_match(
    out,
    regexp = " - 24 levels: a < b < c < d < \\.\\.\\. < x",
    all = FALSE
  )
  expect_match(
    out,
    regexp = ' - attr\\(\\*, "foo"\\)= chr "bar"',
    all = FALSE
  )
})


# other -------------------------------------------------------------------

test_that("glimpse on rvar works", {
  skip_on_cran()
  x_vec <- rvar(array(1:24, dim = c(6,4)))
  x_matrix <- rvar(array(1:24, dim = c(2,3,4)))

  expect_equal(format_glimpse(rvar()), character(0))
  expect_equal(
    format_glimpse(x_vec),
    c("3.5 ± 1.9", "9.5 ± 1.9", "15.5 ± 1.9", "21.5 ± 1.9"),
    check.attributes = FALSE
  )
  expect_equal(format_glimpse(x_matrix), "<rvar[3 x 4]>")
})

test_that("format summary arg works", {
  x = rvar(c(1,9,10))

  expect_equal(as.vector(format(x, summary = "mean_sd")), "6.7 \u00b1 4.9")
  expect_equal(as.vector(format(x, summary = "median_mad")), "9 \u00b1 1.5")
  expect_error(format(x, summary = "foo"))
})

test_that("vec_ptype_abbr and vec_ptype_full work", {
  x <- rvar(array(1:24, dim = c(2,3,4)))
  x_with_chains <- rvar(array(1:24, dim = c(2,3,4)), nchains = 2)

  expect_equal(vec_ptype_abbr(x, suffix_shape = FALSE), "rvar")
  expect_equal(vec_ptype_full(rvar(1:3)), "rvar<3>")
  expect_equal(vec_ptype_full(x), "rvar<2>[,4]")
  expect_equal(vec_ptype_full(x_with_chains), "rvar<1,2>[,4]")

  x_fct <- rvar_factor(array(letters[1:24], dim = c(2,3,4)))
  expect_equal(vec_ptype_abbr(x_fct, suffix_shape = FALSE), "rvar_fct")

  x_ord <- rvar_ordered(array(letters[1:24], dim = c(2,3,4)))
  expect_equal(vec_ptype_abbr(x_ord, suffix_shape = FALSE), "rvar_ord")
})

test_that("formatting NA rvars works", {
  expect_equal(c(format(rvar(c(NA, 1)))), "NA \u00b1 NA")
  expect_equal(c(format(rvar_factor(c(NA, 1)))), "NA <NA>")
  expect_equal(c(format(rvar_ordered(c(NA, 1)))), "NA <NA>")
})

# numeric summaries -------------------------------------------------------

test_that("numeric summaries work", {
  x_array <- array(1:24, dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x <- new_rvar(x_array)
  x_letters <- array(letters[1:24], dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x_ord <- rvar_ordered(x_letters, levels = letters)
  x_fct <- rvar_factor(x_letters, levels = letters)

  expect_equal(median(x), apply(x_array, c(2,3), median))
  expect_equal(sum(x), apply(x_array, c(2,3), sum))
  expect_equal(prod(x), apply(x_array, c(2,3), prod))
  expect_equal(min(x), apply(x_array, c(2,3), min))
  expect_equal(max(x), apply(x_array, c(2,3), max))

  ordered_out <- function(x) structure(
    x, dim = c(2,3), dimnames = list(a = c("a1", "a2"), b = c("b1", "b2", "b3")),
    levels = letters, class = c("ordered", "factor")
  )
  expect_equal(median(x_ord), ordered_out(c(2, 6, 10, 14, 18, 22)))
  expect_equal(min(x_ord), ordered_out(c(1, 5, 9, 13, 17, 21)))
  expect_equal(max(x_ord), ordered_out(c(4, 8, 12, 16, 20, 24)))
  expect_error(sum(x_ord))
  expect_error(prod(x_ord))

  expect_error(median(x_fct))
  expect_error(min(x_fct))
  expect_error(max(x_fct))
  expect_error(sum(x_ord))
  expect_error(prod(x_ord))
})


# mean --------------------------------------------------------------------

test_that("means work", {
  x_array <- array(1:24, dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x <- new_rvar(x_array)

  expect_equal(Pr(x < 2), apply(x_array < 2, c(2,3), mean))
  expect_error(Pr(x))
  expect_equal(E(x), apply(x_array, c(2,3), mean))
  expect_equal(mean(x), apply(x_array, c(2,3), mean))
  # E() and Pr() should also work on base arrays
  expect_equal(Pr(x_array < 2), mean(x_array < 2))
  expect_error(Pr(x_array))
  expect_equal(E(x_array), mean(x_array))

  # test vector rvars as well since these should be summarized down to vectors
  # (not one-dimensional arrays)
  y_array <- array(1:24, dim = c(4,6), dimnames = list(NULL, paste0("a", 1:6)))
  y <- new_rvar(y_array)
  expect_equal(mean(y), apply(y_array, 2, mean))

  expect_error(mean(rvar_factor("a")))
  expect_error(mean(rvar_ordered("a")))
})


# spread ------------------------------------------------------------------

test_that("spread functions work", {
  x_array <- array(1:24, dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x <- new_rvar(x_array)
  x_letters <- array(letters[1:24], dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x_ord <- rvar_ordered(x_letters, levels = letters)
  x_fct <- rvar_factor(x_letters, levels = letters)

  expect_equal(sd(x), apply(x_array, c(2,3), sd))
  expect_equal(variance(x), apply(x_array, c(2,3), var))
  expect_equal(var(x), apply(x_array, c(2,3), var))
  expect_equal(mad(x), apply(x_array, c(2,3), mad))

  expect_error(sd(x_ord))
  expect_error(variance(x_ord))
  expect_error(var(x_ord))
  expect_equal(mad(x_ord), apply(x_array, c(2,3), mad))

  expect_error(sd(x_fct))
  expect_error(variance(x_fct))
  expect_error(var(x_fct))
  expect_error(mad(x_fct))

  y_array <- array(1:24, dim = c(4,6), dimnames = list(NULL, paste0("a", 1:6)))
  y <- new_rvar(y_array)
  expect_equal(sd(y), apply(y_array, 2, sd))
  expect_equal(variance(y), apply(y_array, 2, var))
  expect_equal(var(y), apply(y_array, 2, var))
  expect_equal(mad(y), apply(y_array, 2, mad))
})


# range -------------------------------------------------------------------

test_that("range works", {
  x_array <- array(1:24, dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x <- new_rvar(x_array)
  x_letters <- array(letters[1:24], dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x_ord <- rvar_ordered(x_letters, levels = letters)
  x_fct <- rvar_factor(x_letters, levels = letters)

  expect_equal(range(x), apply(x_array, c(2,3), range))
  expect_equal(range(x_ord),
    structure(c(1, 4, 5, 8, 9, 12, 13, 16, 17, 20, 21, 24), levels = letters,
      class = c("ordered", "factor"), dim = c(2, 2, 3),
      dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3"))
    )
  )
  expect_error(range(x_fct))

  y_array <- array(1:24, dim = c(4,6), dimnames = list(NULL, paste0("a", 1:6)))
  y <- new_rvar(y_array)
  expect_equal(range(y), apply(y_array, 2, range))

  # range over a scalar should return a vector
  z_array <- array(1:6, dim = c(6,1), dimnames = list(NULL, "a"))
  z <- new_rvar(z_array)
  expect_equal(range(z), range(z_array))
})


# logical summaries -------------------------------------------------------

test_that("logical summaries work", {
  x_array <- array(1:24, dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x <- new_rvar(x_array)

  expect_equal(all(x > 10), apply(x_array > 10, c(2,3), all))
  expect_equal(any(x > 10), apply(x_array > 10, c(2,3), any))

  y_array <- array(1:24, dim = c(4,6), dimnames = list(NULL, paste0("a", 1:6)))
  y <- new_rvar(y_array)
  expect_equal(all(y > 10), apply(y_array > 10, 2, all))
  expect_equal(any(y > 10), apply(y_array > 10, 2, any))

  expect_error(all(rvar("a")))
  expect_error(any(rvar("a")))
})


# special value predicates ------------------------------------------------

test_that("special value predicates work", {
  x_array <- array(c(1,NA,3:4, 5:6,Inf,8, 9,-Inf,11:12, NaN,14:24),
    dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3"))
  )
  x <- new_rvar(x_array)
  x_letters <- array(c("a",NA,letters[3:12], NaN, letters[14:24]), dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x_ord <- rvar_ordered(x_letters, levels = letters)
  x_fct <- rvar_factor(x_letters, levels = letters)

  .dimnames = list(a = c("a1", "a2"), b = c("b1", "b2", "b3"))
  expect_equal(is.finite(x), array(c(rep(FALSE, 4), rep(TRUE, 2)), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.infinite(x), array(c(FALSE, TRUE, TRUE, FALSE, FALSE, FALSE), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.nan(x), array(c(FALSE, FALSE, FALSE, TRUE, FALSE, FALSE), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.na(x), array(c(TRUE, FALSE, FALSE, TRUE, FALSE, FALSE), dim = c(2,3), dimnames = .dimnames))

  .dimnames = list(a = c("a1", "a2"), b = c("b1", "b2", "b3"))
  expect_equal(is.finite(x_ord), array(c(FALSE, TRUE, TRUE, FALSE, rep(TRUE, 2)), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.finite(x_fct), array(c(FALSE, TRUE, TRUE, FALSE, rep(TRUE, 2)), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.infinite(x_ord), array(rep(FALSE, 6), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.infinite(x_fct), array(rep(FALSE, 6), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.nan(x_ord), array(c(FALSE, FALSE, FALSE, FALSE, FALSE, FALSE), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.nan(x_fct), array(c(FALSE, FALSE, FALSE, FALSE, FALSE, FALSE), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.na(x_ord), array(c(TRUE, FALSE, FALSE, TRUE, FALSE, FALSE), dim = c(2,3), dimnames = .dimnames))
  expect_equal(is.na(x_fct), array(c(TRUE, FALSE, FALSE, TRUE, FALSE, FALSE), dim = c(2,3), dimnames = .dimnames))

  y_array <- array(x_array, dim = c(4,6), dimnames = list(NULL, paste0("a", 1:6)))
  y <- new_rvar(y_array)
  expect_equal(is.finite(y), matrixStats::colAlls(apply(y_array, 2, is.finite), useNames = TRUE))
  expect_equal(is.infinite(y), matrixStats::colAnys(apply(y_array, 2, is.infinite), useNames = TRUE))
  expect_equal(is.nan(y), matrixStats::colAnys(apply(y_array, 2, is.nan), useNames = TRUE))
  expect_equal(is.na(y), matrixStats::colAnys(apply(y_array, 2, is.na), useNames = TRUE))
})


# anyNA -------------------------------------------------------------------

test_that("anyNA works", {
  x_array <- array(1:24, dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x <- new_rvar(x_array)
  x_letters <- array(letters[1:24], dim = c(4,2,3), dimnames = list(NULL, a = c("a1", "a2"), b = c("b1", "b2", "b3")))
  x_ord <- rvar_ordered(x_letters, levels = letters)
  x_fct <- rvar_factor(x_letters, levels = letters)

  expect_equal(anyNA(x), FALSE)
  expect_equal(anyNA(x_fct), FALSE)
  expect_equal(anyNA(x_ord), FALSE)
  x[2,1] <- NA
  expect_equal(anyNA(x), TRUE)
  x_fct[2,1] <- NA
  expect_equal(anyNA(x_fct), TRUE)
  x_ord[2,1] <- NA
  expect_equal(anyNA(x_ord), TRUE)
})

test_that("distributional functions work on a scalar rvar", {
  x_values <- c(2,4,3,5)
  x <- rvar(x_values)

  x_density <- density(x_values, cut = 0)
  expect_equal(density(x, at = x_density$x), x_density$y)

  x_cdf <- ecdf(x_values)(x_values)
  expect_equal(cdf(x, x_values), x_cdf)

  expect_equal(quantile(x, 1:4/4), quantile(x_values, 1:4/4, names = FALSE))
})

test_that("distributional functions work on an rvar array", {
  x <- rvar(array(1:12, dim = c(3,2,2)))

  d11 <- c(density(1:3, n = 3, from = 1, to = 3)$y, rep(0, 9))
  d21 <- c(rep(0, 3), density(4:6, n = 3, from = 4, to = 6)$y, rep(0, 6))
  d12 <- c(rep(0, 6), density(7:9, n = 3, from = 7, to = 9)$y, rep(0, 3))
  d22 <- c(rep(0, 9), density(10:12, n = 3, from = 10, to = 12)$y)
  x_density <- array(c(d11, d21, d12, d22), dim = c(12, 2, 2))
  expect_equal(density(x, at = 1:12), x_density)

  cdf11 <- ecdf(1:3)(1:12)
  cdf21 <- ecdf(4:6)(1:12)
  cdf12 <- ecdf(7:9)(1:12)
  cdf22 <- ecdf(10:12)(1:12)
  x_cdf <- array(c(cdf11, cdf21, cdf12, cdf22), dim = c(12, 2, 2))
  expect_equal(cdf(x, 1:12), x_cdf)

  p <- ppoints(9, a = 1)
  q11 <- quantile(1:3, p)
  q21 <- quantile(4:6, p)
  q12 <- quantile(7:9, p)
  q22 <- quantile(10:12, p)
  x_quantiles <- array(c(q11, q21, q12, q22), dim = c(9, 2, 2), dimnames = list(NULL))
  expect_equal(quantile(x, p), x_quantiles)
})

test_that("distributional functions work on an rvar_factor", {
  x_values <- c(2,2,2,4,4,4,4,3,5,3)
  x_letters <- letters[x_values]
  x <- rvar_factor(x_letters, levels = letters[1:5])

  expect_equal(density(x, letters[1:6]), c(0, .3, .2, .4, .1, NA))

  expect_equal(cdf(x, letters[1:5]), c(NA, NA, NA, NA, NA))

  expect_equal(quantile(x, 1:4/4), c(NA, NA, NA, NA))
})

test_that("distributional functions work on an rvar_ordered", {
  x_values <- c(2,2,2,4,4,4,4,3,5,3)
  x_letters <- letters[x_values]
  x <- rvar_ordered(x_letters, levels = letters[1:6])

  expect_equal(density(x, letters[1:7]), c(0, .3, .2, .4, .1, 0, NA))

  expect_equal(cdf(x, letters[1:7]), c(0, .3, .5, .9, 1, 1, NA))

  expect_equal(quantile(x, c(.3, .5, .9, 1)), letters[2:5])
})

#' Subset `draws` objects
#'
#' Subset [`draws`] objects by variables, iterations, chains, and draws indices.
#'
#' @template args-methods-x
#' @param variable (character vector) The variables to select. All elements of
#'   non-scalar variables can be selected at once.
#' @param iteration (integer vector) The iteration indices to select.
#' @param chain (integer vector) The chain indices to select.
#' @param draw (integer vector) The draw indices to be select. Subsetting draw
#'   indices will lead to an automatic merging of chains via [`merge_chains`].
#' @param regex (logical) Should `variable` should be treated as a
#'   (vector of) regular expressions? Any variable in `x` matching at least one
#'   of the regular expressions will be selected. Defaults to `FALSE`.
#' @param unique (logical) Should duplicated selection of chains, iterations, or
#'   draws be allowed? If `TRUE` (the default) only unique chains, iterations,
#'   and draws are selected regardless of how often they appear in the
#'   respective selecting arguments.
#'
#' @template args-methods-dots
#' @template return-draws
#'
#' @details
#' To ensure that multiple consecutive subsetting operations work correctly,
#' `subset()` *[repairs][repair_draws]* the `draws` object before and after
#' subsetting.
#'
#' @examples
#' x <- example_draws()
#' subset_draws(x, variable = c("mu", "tau"))
#' subset_draws(x, chain = 2)
#' subset_draws(x, iteration = 5:10, chain = 3:4)
#'
#' # extract the first chain twice
#' subset_draws(x, chain = c(1, 1), unique = FALSE)
#'
#' # extract all elements of 'theta'
#' subset_draws(x, variable = "theta")
#'
#' @export
subset_draws <- function(x, ...) {
  UseMethod("subset_draws")
}

#' @rdname subset_draws
#' @export
subset_draws.draws_matrix <- function(x, variable = NULL, iteration = NULL,
                                      chain = NULL, draw = NULL, regex = FALSE,
                                      unique = TRUE, ...) {
  if (all_null(variable, iteration, chain, draw)) {
    return(x)
  }
  x <- repair_draws(x)
  variable <- check_existing_variables(variable, x, regex = regex)
  iteration <- check_iteration_ids(iteration, x, unique = unique)
  chain <- check_chain_ids(chain, x, unique = unique)
  draw <- check_draw_ids(draw, x, unique = unique)
  x <- prepare_subsetting(x, iteration, chain, draw)
  x <- .subset_draws(x, iteration, chain, draw, variable, reserved = TRUE)
  if (!is.null(chain) || !is.null(iteration)) {
    x <- repair_draws(x, order = FALSE)
  }
  x
}

#' @rdname subset_draws
#' @export
subset_draws.draws_array <- function(x, variable = NULL, iteration = NULL,
                                     chain = NULL, draw = NULL, regex = FALSE,
                                     unique = TRUE, ...) {
  if (all_null(variable, iteration, chain, draw)) {
    return(x)
  }
  x <- repair_draws(x)
  variable <- check_existing_variables(variable, x, regex = regex)
  iteration <- check_iteration_ids(iteration, x, unique = unique)
  chain <- check_chain_ids(chain, x, unique = unique)
  draw <- check_draw_ids(draw, x, unique = unique)
  x <- prepare_subsetting(x, iteration, chain, draw)
  if (!is.null(draw)) {
    iteration <- draw
  }
  x <- .subset_draws(x, iteration, chain, variable, reserved = TRUE)
  if (!is.null(chain) || !is.null(iteration)) {
    x <- repair_draws(x, order = FALSE)
  }
  x
}

#' @rdname subset_draws
#' @export
subset_draws.draws_df <- function(x, variable = NULL, iteration = NULL,
                                  chain = NULL, draw = NULL, regex = FALSE,
                                  unique = TRUE, ...) {
  if (all_null(variable, iteration, chain, draw)) {
    return(x)
  }
  x <- repair_draws(x)
  unique <- as_one_logical(unique)
  variable <- check_existing_variables(variable, x, regex = regex)
  iteration <- check_iteration_ids(iteration, x, unique = unique)
  chain <- check_chain_ids(chain, x, unique = unique)
  draw <- check_draw_ids(draw, x, unique = unique)
  x <- prepare_subsetting(x, iteration, chain, draw)
  x <- .subset_draws(
    x, iteration, chain, draw, variable,
    unique = unique, reserved = TRUE
  )
  x
}

#' @rdname subset_draws
#' @export
subset_draws.draws_list <- function(x, variable = NULL, iteration = NULL,
                                    chain = NULL, draw = NULL, regex = FALSE,
                                    unique = TRUE, ...) {
  if (all_null(variable, iteration, chain, draw)) {
    return(x)
  }
  x <- repair_draws(x)
  variable <- check_existing_variables(variable, x, regex = regex)
  iteration <- check_iteration_ids(iteration, x, unique = unique)
  chain <- check_chain_ids(chain, x, unique = unique)
  draw <- check_draw_ids(draw, x, unique = unique)
  x <- prepare_subsetting(x, iteration, chain, draw)
  if (!is.null(draw)) {
    iteration <- draw
  }
  x <- .subset_draws(x, iteration, chain, variable, reserved = TRUE)
  if (!is.null(chain) || !is.null(iteration)) {
    x <- repair_draws(x, order = FALSE)
  }
  x
}

#' @rdname subset_draws
#' @export
subset_draws.draws_rvars <- function(x, variable = NULL, iteration = NULL,
                                     chain = NULL, draw = NULL, regex = FALSE,
                                     unique = TRUE, ...) {
  if (all_null(variable, iteration, chain, draw)) {
    return(x)
  }
  x <- repair_draws(x)
  variable <- check_existing_variables(variable, x, regex = regex)
  iteration <- check_iteration_ids(iteration, x, unique = unique)
  chain <- check_chain_ids(chain, x, unique = unique)
  draw <- check_draw_ids(draw, x, unique = unique)
  x <- prepare_subsetting(x, iteration, chain, draw)
  if (!is.null(draw)) {
    iteration <- draw
  }
  x <- .subset_draws(x, iteration, chain, variable, reserved = TRUE)
  if (!is.null(chain) || !is.null(iteration)) {
    x <- repair_draws(x, order = FALSE)
  }
  x
}

#' @rdname subset_draws
#' @export
subset_draws.rvar <- function(x, variable = NULL, ...) {
  if (!is.null(variable)) {
    stop_no_call("Cannot subset an rvar by variable.")
  }
  subset_draws(draws_rvars(x = x), ...)$x
}

#' @rdname subset_draws
#' @export
subset.draws <- function(x, ...) {
  subset_draws(x, ...)
}

#' subset specified non-NULL dimensions
#' @param x an object to be subsetted
#' @param ... arguments containing indices for subsetting a dimension
#'   NULL is treated as not subsetting that dimension
#' @noRd
subset_dims <- function(x, ...) {
  dots <- list(...)
  if (!length(dots)) {
    return(x)
  }
  dim_x <- max(length(dim(x)), 1L)
  if (length(dots) > dim_x) {
    stop_no_call("'x' has only ", dim_x, " dimensions.")
  }
  if (length(dots) < dim_x) {
    dots <- c(dots, repl(NULL, dim_x - length(dots)))
  }
  names(dots) <- paste0("i", seq_along(dots))
  args <- rep("", length(dots))
  for (i in seq_along(dots)) {
    if (!is.null(dots[[i]])) {
      args[i] <- names(dots)[i]
    }
  }
  args <- paste0(args, collapse = ", ")
  call <- paste0("x[", args, "]")
  dots$x <- x
  eval2(call, dots)
}

# internal method for subset_draws
.subset_draws <- function(x, ...) {
  UseMethod(".subset_draws")
}

#' @export
.subset_draws.draws_matrix <- function(x, iteration = NULL, chain = NULL,
                                       draw = NULL, variable = NULL,
                                       reserved = FALSE, ...) {
  out <- subset_dims(x, NULL, variable)
  if (!is.null(draw)) {
    out <- out[draw, ]
  } else {
    if (!is.null(chain)) {
      chain <- unique(chain)
      nchains <- length(chain)
      chain_ids <- rep(chain_ids(out), each = niterations(out))
      slice_index <- chain_ids %in% chain
      out <- out[slice_index, ]
      attr(out, "nchains") <- nchains
    }
    if (!is.null(iteration)) {
      niterations <- length(iteration)
      slice_index <- iteration +
        (rep(chain_ids(out), each = niterations) - 1) * niterations(out)
      nchains <- nchains(out)
      out <- out[slice_index, ]
      attr(out, "nchains") <- nchains
    }
  }
  if (reserved && !is.null(variable)) {
    new_vars <- variables(out, reserved = TRUE)
    reserved_vars <- setdiff(reserved_variables(x), new_vars)
    if (length(reserved_vars)) {
      x_reserved <- subset_dims(x, draw, reserved_vars)
      out <- cbind(out, x_reserved)
      class(out) <- class(x)
    }
  }
  out
}

#' @export
.subset_draws.draws_array <- function(x, iteration = NULL, chain = NULL,
                                      variable = NULL, reserved = FALSE, ...) {
  out <- subset_dims(x, iteration, chain, variable)
  if (reserved && !is.null(variable)) {
    new_vars <- variables(out, reserved = TRUE)
    reserved_vars <- setdiff(reserved_variables(x), new_vars)
    if (length(reserved_vars)) {
      x_reserved <- subset_dims(x, iteration, chain, reserved_vars)
      out <- abind(out, x_reserved, along = 3L)
      class(out) <- class(x)
    }
  }
  out
}

#' @export
.subset_draws.draws_df <- function(x, iteration = NULL, chain = NULL,
                                   draw = NULL, variable = NULL,
                                   unique = TRUE, reserved = FALSE, ...) {
  if (!is.null(variable)) {
    x <- x[, variable, reserved = TRUE]
  }
  if (!is.null(draw)) {
    # each x$.draw is unique so using 'match' is valid here
    x <- x[match(draw, x$.draw), ]
    x$.draw <- repair_iteration_ids(x$.draw)
    x$.iteration <- x$.draw
  } else if (!is.null(chain) || !is.null(iteration)) {
    if (unique) {
      if (!is.null(chain)) {
        x <- x[x$.chain %in% chain, ]
      }
      if (!is.null(iteration)) {
        x <- x[x$.iteration %in% iteration, ]
      }
      x <- repair_draws(x, order = FALSE)
    } else {
      if (!is.null(chain)) {
        x <- x[unlist(lapply(chain, function(c) which(x$.chain == c))), ]
        x$.chain <- rep(seq_along(chain), each = nrow(x) / length(chain))
      }
      if (!is.null(iteration)) {
        x <- x[unlist(lapply(iteration, function(i) which(x$.iteration == i))), ]
        x$.iteration <- rep(seq_along(iteration), each = nrow(x) / length(iteration))
      }
      x <- repair_draws(x)
    }
  }
  if (!reserved && !is.null(variable)) {
    # remove reserved variables which were not selected
    reserved_vars <- setdiff(reserved_variables(x), variable)
    x <- remove_variables(x, reserved_vars)
  }
  x
}

#' @export
.subset_draws.draws_list <- function(x, iteration = NULL, chain = NULL,
                                     variable = NULL, reserved = FALSE, ...) {
  if (!is.null(chain)) {
    x <- x[chain]
  }
  if (!is.null(variable)) {
    if (reserved) {
      z <- x
    }
    for (i in seq_along(x)) {
      x[[i]] <- x[[i]][variable]
    }
    if (reserved) {
      new_vars <- variables(x, reserved = TRUE)
      reserved_vars <- setdiff(reserved_variables(z), new_vars)
      if (length(reserved_vars)) {
        for (i in seq_along(x)) {
          x[[i]] <- c(x[[i]], z[[i]][reserved_vars])
        }
      }
      remove(z)
    }
  }
  if (!is.null(iteration)) {
    for (i in seq_along(x)) {
      x[[i]] <- lapply(x[[i]], `[`, iteration)
    }
  }
  x
}

#' @importFrom vctrs vec_slice
#' @export
.subset_draws.draws_rvars <- function(x, iteration = NULL, chain = NULL,
                                      variable = NULL, reserved = FALSE, ...) {
  if (!is.null(variable)) {
    if (reserved) {
      z <- x
    }
    x <- x[variable]
    if (reserved) {
      new_vars <- variables(x, reserved = TRUE)
      reserved_vars <- setdiff(reserved_variables(z), new_vars)
      x <- c(x, z[reserved_vars])
      # c() currently removes the 'draws' classes
      class(x) <- class_draws_rvars()
      remove(z)
    }
  }
  if (!is.null(chain)) {
    chain <- unique(chain)
    nchains <- length(chain)
    chain_ids <- rep(chain_ids(x), each = niterations(x))
    slice_index <- chain_ids %in% chain
    for (i in seq_along(x)) {
      draws_of(x[[i]]) <- vec_slice(draws_of(x[[i]]), slice_index)
      nchains_rvar(x[[i]]) <- nchains
    }
  }
  if (!is.null(iteration)) {
    niterations <- length(iteration)
    slice_index <- iteration +
      (rep(chain_ids(x), each = niterations) - 1) * niterations(x)
    for (i in seq_along(x)) {
      draws_of(x[[i]]) <- vec_slice(draws_of(x[[i]]), slice_index)
    }
  }
  x
}

# prepare object to be subsetted via 'subset_draws'
prepare_subsetting <- function(x, iteration = NULL, chain = NULL,
                               draw = NULL) {
  if (!is.null(draw)) {
    if (!is.null(iteration)) {
      stop_no_call("Cannot subset 'iteration' and 'draw' at the same time.")
    }
    if (!is.null(chain)) {
      stop_no_call("Cannot subset 'chain' and 'draw' at the same time.")
    }
    if (nchains(x) > 1L) {
      message("Merging chains in order to subset via 'draw'.")
      x <- merge_chains(x)
    }
  }
  x
}

#' The `draws_array` format
#'
#' @name draws_array
#' @family formats
#'
#' @templateVar draws_format draws_array
#' @templateVar base_class "array"
#' @template draws_format-skeleton
#' @template args-format-nchains
#'
#' @details Objects of class `"draws_array"` are 3-D arrays with dimensions
#'   `"iteration"`, `"chain"`, and `"variable"`. See **Examples**.
#'
NULL

#' @rdname draws_array
#' @export
as_draws_array <- function(x, ...) {
  UseMethod("as_draws_array")
}

#' @rdname draws_array
#' @export
as_draws_array.default <- function(x, ...) {
  x <- as_draws(x)
  as_draws_array(x, ...)
}

#' @rdname draws_array
#' @export
as_draws_array.draws_array <- function(x, ...) {
  x
}

#' @rdname draws_array
#' @export
as_draws_array.draws_matrix <- function(x, ...) {
  old_dim <- dim(x)
  old_dimnames <- dimnames(x)
  iteration_ids <- iteration_ids(x)
  chain_ids <- chain_ids(x)
  dim(x) <- c(niterations(x), nchains(x), old_dim[2])
  dimnames(x) <- list(
    iteration = iteration_ids,
    chain = chain_ids,
    variable = old_dimnames[[2]]
  )
  class(x) <- class_draws_array()
  attr(x, "nchains") <- NULL
  x
}

#' @rdname draws_array
#' @export
as_draws_array.draws_df <- function(x, ...) {
  if (ndraws(x) == 0) {
    return(empty_draws_array(variables(x)))
  }
  iterations <- iteration_ids(x)
  chains <- chain_ids(x)
  x <- check_variables_are_numeric(x, to = "draws_array")
  out <- vector("list", length(chains))
  for (i in seq_along(out)) {
    if (length(chains) == 1) {
      out[[i]] <- x
    } else {
      out[[i]] <- x[x$.chain == i, ]
    }
    out[[i]] <- remove_reserved_df_variables(out[[i]])
    out[[i]] <- as.matrix(out[[i]])
  }
  out <- as_array_matrix_list(out)
  dimnames(out) <- list(
    iteration = iterations,
    chain = chains,
    variable = dimnames(out)[[3]]
  )
  class(out) <- class_draws_array()
  out
}

#' @rdname draws_array
#' @export
as_draws_array.draws_list <- function(x, ...) {
  x <- as_draws_df(x)
  as_draws_array(x, ...)
}

#' @rdname draws_array
#' @export
as_draws_array.draws_rvars <- function(x, ...) {
  if (ndraws(x) == 0) {
    return(empty_draws_array(variables(x)))
  }

  x <- check_variables_are_numeric(
    x, to = "draws_array", is_non_numeric = is_rvar_factor, convert = FALSE
  )

  # cbind discards class information when applied to vectors, which converts
  # the underlying factors to numeric
  draws <- do.call(cbind, lapply(seq_along(x), function(i) {
    # flatten each rvar so it only has two dimensions: draws and variables
    # this also collapses indices into variable names in the format "var[i,j,k,...]"
    x_i <- flatten_array(x[[i]], names(x)[[i]])
    draws_of(x_i)
  }))

  # add chain info back into the draws array
  # ([draws, variables] -> [iterations, chains, variables])
  .dimnames <- dimnames(draws)
  dim(draws) <- c(niterations(x), nchains(x), dim(draws)[-1])
  dimnames(draws) <- c(list(NULL, NULL), .dimnames[-1])

  as_draws_array(draws, ...)
}

#' @rdname draws_array
#' @export
as_draws_array.mcmc <- function(x, ...) {
  as_draws_array(as_draws_matrix(x), ...)
}

#' @rdname draws_array
#' @export
as_draws_array.mcmc.list <- function(x, ...) {
  class(x) <- "list"
  .as_draws_array(as_array_matrix_list(x))
}

# try to convert any R object into a 'draws_array' object
.as_draws_array <- function(x) {
  x <- as.array(x)
  new_dimnames <- list(iteration = NULL, chain = NULL, variable = NULL)
  if (!is.null(dimnames(x)[[3]])) {
    new_dimnames[[3]] <- dimnames(x)[[3]]
  } else {
    new_dimnames[[3]] <- default_variables(dim(x)[3])
  }
  check_new_variables(new_dimnames[[3]])
  new_dimnames[[1]] <- as.character(seq_rows(x))
  new_dimnames[[2]] <- as.character(seq_cols(x))
  dimnames(x) <- new_dimnames
  class(x) <- class_draws_array()
  x
}

#' @rdname draws_array
#' @export
draws_array <- function(..., .nchains = 1) {
  out <- validate_draws_per_variable(...)
  .nchains <- as_one_integer(.nchains)
  if (.nchains < 1) {
    stop_no_call("Number of chains must be positive.")
  }
  ndraws <- length(out[[1]])
  if (ndraws %% .nchains != 0) {
    stop_no_call("Number of chains does not divide the number of draws.")
  }
  niterations <- ndraws / .nchains
  variables <- names(out)
  out <- unlist(out)
  out <- array(out, dim = c(niterations, .nchains, length(variables)))
  dimnames(out) <- list(NULL, NULL, variables)
  as_draws_array(out)
}

class_draws_array <- function() {
  c("draws_array", "draws", "array")
}

#' @rdname draws_array
#' @export
is_draws_array <- function(x) {
  inherits(x, "draws_array")
}

# is an object looking like a 'draws_array' object?
is_draws_array_like <- function(x) {
  is.array(x) && length(dim(x)) == 3L
}

#' Extract parts of a `draws_array` object
#'
#' Extract parts of a `draws_array` object. They are strictly defined as arrays
#' of 3 dimensions (iteration x chain x variable) so dropping any of the
#' dimensions breaks the expected structure of the object. Accordingly, no
#' dropping of dimensions is done by default even if the extracted slices are of
#' length 1. If `drop` is manually set to `TRUE` and any of the dimensions is
#' actually dropped, this will lead to dropping the `"draws_array"` class as
#' well.
#'
#' @param x,i,j,...,drop Same as in the default extraction method but with
#'   `drop` being set to `FALSE` by default.
#'
#' @return An object of class `"draws_array"` unless any of the dimensions
#' was dropped during the extraction.
#'
#' @keywords internal
#' @export
`[.draws_array` <- function(x, i, j, ..., drop = FALSE) {
  # TODO: allow for argument 'reserved' as in '[.draws_df'
  #   right now this fails because NextMethod() cannot ignore arguments
  out <- NextMethod("[", drop = drop)
  if (length(dim(out)) == length(dim(x))) {
    class(out) <- class(x)
  }
  out
}

#' @export
variance.draws_array <- function(x, ...) {
  var(as.vector(x))
}

# convert a list of matrices to an array
as_array_matrix_list <- function(x) {
  stopifnot(is.list(x))
  if (length(x) == 1) {
    tmp <- dimnames(x[[1]])
    x <- x[[1]]
    dim(x) <- c(dim(x), 1)
    dimnames(x) <- tmp
  } else {
    x <- abind::abind(x, along = 3L)
  }
  x <- aperm(x, c(1, 3, 2))
}

# create an empty draws_array object
empty_draws_array <- function(variables = character(0), nchains = 0,
                              niterations = 0) {
  assert_character(variables, null.ok = TRUE)
  assert_number(nchains, lower = 0)
  assert_number(niterations, lower = 0)
  out <- array(
    numeric(0),
    dim = c(niterations, nchains, length(variables)),
    dimnames = list(
      iteration = seq_len(niterations),
      chain = seq_len(nchains),
      variable = variables
    )
  )
  class(out) <- class_draws_array()
  out
}

test_that("extract_variable_matrix works the same for different formats", {
  draws_array <- as_draws_array(example_draws())
  mu_array <- extract_variable_matrix(draws_array, "mu")

  draws_df <- as_draws_df(example_draws())
  mu_df <- extract_variable_matrix(draws_df, "mu")
  expect_equal(mu_df, mu_array)

  draws_list <- as_draws_list(example_draws())
  mu_list <- extract_variable_matrix(draws_list, "mu")
  expect_equal(mu_list, mu_array)

  draws_matrix <- as_draws_matrix(example_draws())
  mu_matrix <- extract_variable_matrix(draws_matrix, "mu")
  expect_equal(as.vector(mu_matrix), as.vector(mu_array))

  draws_rvars <- as_draws_rvars(example_draws())
  mu_matrix <- extract_variable_matrix(draws_rvars, "mu")
  expect_equal(as.vector(mu_matrix), as.vector(mu_array))
})

test_that("extract_variable_matrix works for draws_rvars on an indexed variable", {
  draws_array <- as_draws_array(example_draws())
  theta1_array <- extract_variable_matrix(draws_array, "theta[1]")

  draws_rvars <- as_draws_rvars(example_draws())
  theta1_matrix <- extract_variable_matrix(draws_rvars, "theta[1]")
  expect_equal(as.vector(theta1_matrix), as.vector(theta1_array))

  expect_error(extract_variable_matrix(draws_rvars, "theta"), "Cannot extract non-scalar value")
})

test_that("extract_variable_matrix default method works", {
  # it should convert matrix to draws object
  x <- matrix(1:20, nrow = 10, ncol = 2)
  colnames(x) <- c("A", "B")
  expect_equivalent(extract_variable_matrix(x, "A"), x[, 1, drop=FALSE])
  expect_equivalent(extract_variable_matrix(x, "B"), x[, 2, drop=FALSE])
})

test_that("weight_draws works on draws_matrix", {
  x <- as_draws_matrix(example_draws())
  weights <- rexp(ndraws(x))

  x1 <- weight_draws(x, weights)
  weights1 <- weights(x1, normalize = FALSE)
  expect_equal(weights1, weights)

  x2 <- weight_draws(x, log(weights), log = TRUE)
  weights2 <- weights(x2)
  expect_equal(weights2, weights / sum(weights))
})

test_that("weight_draws works on draws_array", {
  x <- as_draws_array(example_draws())
  weights <- rexp(ndraws(x))

  x1 <- weight_draws(x, weights)
  weights1 <- weights(x1)
  expect_equal(weights1, weights / sum(weights))

  x2 <- weight_draws(x, log(weights), log = TRUE)
  weights2 <- weights(x2, normalize = FALSE)
  expect_equal(weights2, weights)
})

test_that("weight_draws works on draws_df", {
  x <- as_draws_df(example_draws())
  weights <- rexp(ndraws(x))

  x1 <- weight_draws(x, weights)
  weights1 <- weights(x1, normalize = FALSE)
  expect_equal(weights1, weights)

  x2 <- weight_draws(x, log(weights), log = TRUE)
  weights2 <- weights(x2)
  expect_equal(weights2, weights / sum(weights))
})

test_that("weight_draws works on draws_list", {
  x <- as_draws_list(example_draws())
  weights <- rexp(ndraws(x))

  x1 <- weight_draws(x, weights)
  weights1 <- weights(x1)
  expect_equal(weights1, weights / sum(weights))

  x2 <- weight_draws(x, log(weights), log = TRUE)
  weights2 <- weights(x2, normalize = FALSE)
  expect_equal(weights2, weights)
})

test_that("weight_draws works on draws_rvars", {
  x <- as_draws_rvars(example_draws())
  weights <- rexp(ndraws(x))

  x1 <- weight_draws(x, weights)
  weights1 <- weights(x1)
  expect_equal(weights1, weights / sum(weights))

  x2 <- weight_draws(x, log(weights), log = TRUE)
  weights2 <- weights(x2, normalize = FALSE)
  expect_equal(weights2, weights)
})


# conversion preserves weights --------------------------------------------

test_that("conversion between formats preserves weights", {
  draws <- list(
    matrix = weight_draws(draws_matrix(x = 1:10), 1:10),
    array = weight_draws(draws_array(x = 1:10), 1:10),
    df = weight_draws(draws_df(x = 1:10), 1:10),
    list = weight_draws(draws_list(x = 1:10), 1:10),
    rvars = weight_draws(draws_rvars(x = 1:10), 1:10)
  )

  # chain/iteration/draw columns are placed at the end by conversion functions,
  # so our reference format will keep that order
  reserved = names(draws$df) %in% reserved_df_variables()
  draws$df = draws$df[, c(names(draws$df)[!reserved], names(draws$df)[reserved])]

  for (type in names(draws)) {
    expect_equal(as_draws_matrix(draws[[!!type]]), draws$matrix)
    expect_equal(as_draws_array(draws[[!!type]]), draws$array)
    expect_equal(as_draws_df(draws[[!!type]]), draws$df)
    expect_equal(as_draws_list(draws[[!!type]]), draws$list)
    expect_equal(as_draws_rvars(draws[[!!type]]), draws$rvars)
  }
})

test_that("mutate_variables works correctly for draws_df objects", {
  x <- as_draws_df(example_draws())

  x <- mutate_variables(x, tau2 = tau^2)
  expect_equal(x$tau2, x$tau^2)

  x <- mutate_variables(x, scale = 1.96 * tau, lower = mu - scale)
  expect_equal(x$lower, x$mu - 1.96 * x$tau)

  x <- mutate_variables(x, mean = mean(mu))
  expect_equal(x$mean, rep(mean(x$mu), NROW(x)))

  expect_error(
    mutate_variables(x, mu = as.character(mu)),
    "does not evaluate to a numeric vector"
  )
  expect_error(
    mutate_variables(x, mu = mu[1:2]),
    "does not evaluate to a vector of length 1 or 400"
  )
})

test_that("mutate_variables works correctly for draws_list objects", {
  x <- as_draws_list(example_draws())

  x <- mutate_variables(x, tau2 = tau^2)
  expect_equal(x[[1]]$tau2, x[[1]]$tau^2)
  expect_equal(x[[2]]$tau2, x[[2]]$tau^2)

  x <- mutate_variables(x, scale = 1.96 * tau, lower = mu - scale)
  expect_equal(x[[2]]$lower, x[[2]]$mu - 1.96 * x[[2]]$tau)

  x <- mutate_variables(x, mean = mean(mu))
  expect_equal(x[[1]]$mean, rep(mean(x[[1]]$mu), niterations(x)))

  expect_error(
    mutate_variables(x, mu = as.character(mu)),
    "does not evaluate to a numeric vector"
  )
  expect_error(
    mutate_variables(x, mu = mu[1:2]),
    "does not evaluate to a vector of length 1 or 100"
  )
})

test_that("mutate_variables works correctly for draws_matrix objects", {
  x <- as_draws_matrix(example_draws())
  x <- mutate_variables(x, tau2 = tau^2)
  expect_equal(x[, "tau2"], x[, "tau"]^2, check.attributes = FALSE)
})

test_that("mutate_variables works correctly for draws_array objects", {
  x <- as_draws_array(example_draws())
  x <- mutate_variables(x, tau2 = tau^2)
  expect_equal(x[, , "tau2"], x[, , "tau"]^2, check.attributes = FALSE)
})

test_that("mutate_variables works correctly for draws_rvars objects", {
  x <- as_draws_rvars(example_draws())
  x <- mutate_variables(x, tau_sq = tau^2, theta_2_sq = theta[[2]]^2)
  expect_equal(x$tau_sq, x$tau^2, check.attributes = FALSE)
  expect_equal(x$theta_2_sq, x$theta[[2]]^2, check.attributes = FALSE)
})

test_that("mutate_variables drops variables on assigning NULL", {
  x <- example_draws()
  x <- mutate_variables(x, mu = NULL, tau = NULL)
  expect_equal(variables(x), paste0("theta[", 1:8, "]"))
})

test_that("resample_draws returns expected format", {
  x <- example_draws()
  w <- runif(ndraws(x), 0, 10)

  x <- as_draws_matrix(x)
  x_rs <- resample_draws(x, weights = w, method = "stratified")
  expect_true(is_draws_matrix(x_rs))
  expect_equal(ndraws(x_rs), ndraws(x))

  x <- as_draws_array(x)
  x_rs <- resample_draws(x, weights = w, method = "deterministic", ndraws = 200)
  expect_true(is_draws_array(x_rs))
  expect_equal(ndraws(x_rs), 200)

  x <- as_draws_df(x)
  x_rs <- resample_draws(x, weights = w, method = "simple")
  expect_true(is_draws_df(x_rs))
  expect_equal(ndraws(x_rs), ndraws(x))

  x <- as_draws_list(x)
  x_rs <- resample_draws(x, w, method = "simple_no_replace", ndraws = 100)
  expect_true(is_draws_list(x_rs))
  expect_equal(ndraws(x_rs), 100)
  expect_error(
    resample_draws(x, w, method = "simple_no_replace"),
    "Argument 'ndraws' is required"
  )

  x <- as_draws_rvars(x)
  x_rs <- resample_draws(x, weights = w, method = "simple")
  expect_true(is_draws_rvars(x_rs))
  expect_equal(ndraws(x_rs), ndraws(x))
})

test_that("Resampling algorithms return the correct result in expectation", {
  set.seed(1234)
  x <- as_draws_df(cbind(mu = 1:10000))
  w <- 1:10000 / 777
  expected_mean <- sum(x$mu * (w / sum(w)))

  x_rs <- resample_draws(x, w, method = "stratified")
  mean_rs <- mean(x_rs$mu)
  expect_true(mean_rs > 6660 && mean_rs < 6670)

  x_rs <- resample_draws(x, w, method = "deterministic")
  mean_rs <- mean(x_rs$mu)
  expect_true(mean_rs > 6660 && mean_rs < 6670)

  x_rs <- resample_draws(x, w, method = "simple")
  mean_rs <- mean(x_rs$mu)
  expect_true(mean_rs > 6650 && mean_rs < 6690)

  # method 'simple_no_replace' will be biased for weights with large variance
})

test_that("resample_draws uses stored weights when available", {
  x <- example_draws()

  x_rs <- resample_draws(x)
  expect_true(is_draws_array(x_rs))
  expect_equal(ndraws(x_rs), ndraws(x))

  w <- runif(ndraws(x), 0, 10)
  x <- weight_draws(x, w)
  x_rs <- resample_draws(x)
  expect_true(is_draws_array(x_rs))
  expect_equal(ndraws(x_rs), ndraws(x))
  # .log_weight variable has been dropped
  expect_true(!".log_weight" %in% variables(x_rs, reserved = TRUE))
})

test_that("resample_draws works on rvars", {
  set.seed(1234)
  x <- rvar(1:10000)
  w <- 1:10000 / 777

  x_rs <- resample_draws(x, w, method = "stratified")
  mean_rs <- mean(x_rs)
  expect_true(mean_rs > 6660 && mean_rs < 6670)
  expect_true(is_rvar(x_rs))

  # without weights
  x_rs <- resample_draws(x, method = "stratified")
  mean_rs <- mean(x_rs)
  expect_true(mean_rs > 4990 && mean_rs < 5010)
})

test_that("print.draws_matrix runs without errors", {
  x <- as_draws_matrix(example_draws())
  expect_output(print(x),
    "A draws_matrix: 100 iterations, 4 chains, and 10 variables"
  )

  x <- weight_draws(x, rep(1, ndraws(x)))
  expect_output(
    print(x),
    "hidden reserved variables ..\\.log_weight.."
  )
})

test_that("print.draws_array runs without errors", {
  x <- as_draws_array(example_draws())
  expect_output(print(x),
    "A draws_array: 100 iterations, 4 chains, and 10 variables"
  )

  x <- weight_draws(x, rep(1, ndraws(x)))
  expect_output(
    print(x),
    "hidden reserved variables ..\\.log_weight.."
  )
})

test_that("print.draws_df runs without errors", {
  x <- as_draws_df(example_draws())
  expect_output(print(x),
    "A draws_df: 100 iterations, 4 chains, and 10 variables"
  )

  x <- weight_draws(x, rep(1, ndraws(x)))
  expect_output(print(x), "'\\.log_weight'")

  x <- subset(x, variable = c("mu", "tau"))
  x <- mutate_variables(x, tau2 = tau^2)
  expect_output(print(x), "tau2")
})

test_that("print.draws_list runs without errors", {
  x <- as_draws_list(example_draws())
  expect_output(print(x),
    "A draws_list: 100 iterations, 4 chains, and 10 variables"
  )

  x <- weight_draws(x, rep(1, ndraws(x)))
  expect_output(
    print(x),
    "hidden reserved variables ..\\.log_weight.."
  )
})

test_that("print.draws_rvars runs without errors", {
  skip_on_cran()
  skip_on_os("windows")
  x <- as_draws_rvars(example_draws())
  out <- capture.output(print(x))
  expect_match(
    out,
    regexp = "A draws_rvars: 100 iterations, 4 chains, and 3 variables",
    all = FALSE
  )

  x <- weight_draws(x, rep(1, ndraws(x)))
  expect_output(
    print(x),
    "hidden reserved variables ..\\.log_weight.."
  )
})

test_that("print.draws_array handles reserved variables correctly", {
  x <- as_draws_array(example_draws())
  variables(x)[1] <- ".log_weight" # reserved name
  expect_output(print(x, max_variables = 1), "variable = tau")
  expect_output(
    print(x),
    "hidden reserved variables ..\\.log_weight.."
  )
})

test_that("print.draws_matrix handles reserved variables correctly", {
  x <- as_draws_matrix(example_draws())
  variables(x)[1] <- ".log_weight" # reserved name
  expect_output(print(x, max_variables = 1), "tau")
  expect_output(
    print(x),
    "hidden reserved variables ..\\.log_weight.."
  )
})

test_that("print.draws_df handles reserved variables correctly", {
  x <- as_draws_df(example_draws())
  variables(x)[1] <- ".log_weight" # reserved name
  expect_output(print(x, max_variables = 1), "tau")
  expect_output(
    print(x),
    "hidden reserved variables {'.log_weight', '.chain', '.iteration', '.draw'}",
    fixed = TRUE
  )
})

test_that("print.draws_list handles reserved variables correctly", {
  x <- as_draws_list(example_draws())
  variables(x)[1] <- ".log_weight" # reserved name
  expect_output(print(x, max_variables = 1), "tau")
  expect_output(
    print(x),
    "hidden reserved variables ..\\.log_weight.."
  )
})

test_that("print.draws_rvars handles reserved variables correctly", {
  skip_on_cran()
  skip_on_os("windows")
  x <- as_draws_rvars(example_draws())
  variables(x)[1] <- ".log_weight" # reserved name
  expect_output(print(x, max_variables = 1), "tau")
  expect_output(
    print(x),
    "hidden reserved variables ..\\.log_weight.."
  )
})

test_that("print.draws_df correctly handles data frames with unrepaired draws", {
  x <- as_draws_df(list(x = 1:10, y = 2:11))
  x_slice <- x[c(1,3,5),]
  expect_output(
    print(x_slice),
"x +y
1 +1 +2
2 +3 +4
3 +5 +6"
  )
})

test_that("remove_variables works correctly for draws_matrix objects", {
  x <- as_draws_matrix(example_draws())
  expect_equal(posterior:::remove_variables(x, NULL), x)
  x <- posterior:::remove_variables(x, c("mu", "tau"))
  expect_equal(variables(x), paste0("theta[", 1:8, "]"))
})

test_that("remove_variables works correctly for draws_array objects", {
  x <- as_draws_array(example_draws())
  expect_equal(posterior:::remove_variables(x, NULL), x)
  x <- posterior:::remove_variables(x, c("mu", "tau"))
  expect_equal(variables(x), paste0("theta[", 1:8, "]"))
})

test_that("remove_variables works correctly for draws_df objects", {
  x <- as_draws_df(example_draws())
  expect_equal(posterior:::remove_variables(x, NULL), x)
  x <- posterior:::remove_variables(x, c("mu", "tau"))
  expect_equal(variables(x), paste0("theta[", 1:8, "]"))
})

test_that("remove_variables works correctly for draws_list objects", {
  x <- as_draws_list(example_draws())
  expect_equal(posterior:::remove_variables(x, NULL), x)
  x <- posterior:::remove_variables(x, c("mu", "tau"))
  expect_equal(variables(x), paste0("theta[", 1:8, "]"))
})

test_that("remove_variables works correctly for draws_rvars objects", {
  x <- as_draws_rvars(example_draws())
  expect_equal(posterior:::remove_variables(x, NULL), x)
  x <- posterior:::remove_variables(x, c("mu", "tau"))
  expect_equal(variables(x), "theta")
})

test_that("transformations to and from draws_matrix objects work", {
  draws_matrix <- as_draws_matrix(example_draws())

  draws_array <- as_draws_array(draws_matrix)
  draws_matrix2 <- as_draws_matrix(draws_array)
  expect_equal(draws_matrix, draws_matrix2)

  draws_df <- as_draws_df(draws_matrix)
  draws_matrix2 <- as_draws_matrix(draws_df)
  expect_equal(draws_matrix, draws_matrix2)

  draws_list <- as_draws_list(draws_matrix)
  draws_matrix2 <- as_draws_matrix(draws_list)
  expect_equal(draws_matrix, draws_matrix2)

  draws_rvars <- as_draws_rvars(draws_matrix)
  draws_matrix2 <- as_draws_matrix(draws_rvars)
  expect_equal(draws_matrix, draws_matrix2)
})

test_that("transformations to and from draws_array objects work", {
  draws_array <- as_draws_array(example_draws())

  draws_matrix <- as_draws_matrix(draws_array)
  draws_array2 <- as_draws_array(draws_matrix)
  expect_equal(draws_array, draws_array2)

  draws_df <- as_draws_df(draws_array)
  draws_array2 <- as_draws_array(draws_df)
  expect_equal(draws_array, draws_array2)

  draws_list <- as_draws_list(draws_array)
  draws_array2 <- as_draws_array(draws_list)
  expect_equal(draws_array, draws_array2)

  draws_rvars <- as_draws_rvars(draws_array)
  draws_array2 <- as_draws_array(draws_rvars)
  expect_equal(draws_array, draws_array2)
})

test_that("transformations to and from draws_df objects work", {
  draws_df <- as_draws_df(example_draws())

  draws_matrix <- as_draws_matrix(draws_df)
  draws_df2 <- as_draws_df(draws_matrix)
  expect_equal(draws_df, draws_df2)

  draws_array <- as_draws_array(draws_df)
  draws_df2 <- as_draws_df(draws_array)
  expect_equal(draws_df, draws_df2)

  draws_list <- as_draws_list(draws_df)
  draws_df2 <- as_draws_df(draws_list)
  expect_equal(draws_df, draws_df2)

  draws_rvars <- as_draws_rvars(draws_df)
  draws_df2 <- as_draws_df(draws_rvars)
  expect_equal(draws_df, draws_df2)

  # test that a single draw does not lead to a drop of dimensions
  draws_rvars2 <- subset_draws(draws_rvars, draw = 1)
  draws_df2 <- as_draws_df(draws_rvars2)
  expect_equal(subset_draws(draws_df, draw = 1), draws_df2)
})

test_that("transformations to and from draws_list objects work", {
  draws_list <- as_draws_list(example_draws())

  draws_matrix <- as_draws_matrix(draws_list)
  draws_list2 <- as_draws_list(draws_matrix)
  expect_equal(draws_list, draws_list2)

  draws_array <- as_draws_array(draws_list)
  draws_list2 <- as_draws_list(draws_array)
  expect_equal(draws_list, draws_list2)

  draws_df <- as_draws_df(draws_list)
  draws_list2 <- as_draws_list(draws_df)
  expect_equal(draws_list, draws_list2)

  draws_rvars <- as_draws_rvars(draws_list)
  draws_list2 <- as_draws_list(draws_rvars)
  expect_equal(draws_list, draws_list2)
})

test_that("transformations to and from draws_rvars objects work", {
  draws_rvars <- as_draws_rvars(example_draws())

  draws_matrix <- as_draws_matrix(draws_rvars)
  draws_rvars2 <- as_draws_rvars(draws_matrix)
  expect_equal(draws_rvars, draws_rvars2)

  draws_array <- as_draws_array(draws_rvars)
  draws_rvars2 <- as_draws_rvars(draws_array)
  expect_equal(draws_rvars, draws_rvars2)

  draws_df <- as_draws_df(draws_rvars)
  draws_rvars2 <- as_draws_rvars(draws_df)
  expect_equal(draws_rvars, draws_rvars2)

  draws_list <- as_draws_list(draws_rvars)
  draws_rvars2 <- as_draws_rvars(draws_list)
  expect_equal(draws_rvars, draws_rvars2)
})

test_that("matrices can be transformed to draws_matrix objects", {
  x <- round(rnorm(200), 2)
  x <- array(x, dim = c(40, 5))
  dimnames(x) <- list(NULL, paste0("theta", 1:5))

  y <- as_draws(x)
  expect_is(y, "draws_matrix")
  expect_equal(variables(y), colnames(x))
  expect_equal(niterations(y), 40)
  expect_equal(nchains(y), 1)
})

test_that("arrays can be transformed to draws_array objects", {
  x <- round(rnorm(200), 2)
  x <- array(x, dim = c(10, 4, 5))
  dimnames(x) <- list(NULL, NULL, paste0("theta", 1:5))

  y <- as_draws(x)
  expect_is(y, "draws_array")
  expect_equal(variables(y), dimnames(x)[[3]])
  expect_equal(niterations(y), 10)
  expect_equal(nchains(y), 4)
})

test_that("data.frames can be transformed to draws_df objects", {
  x <- data.frame(
    v1 = rnorm(100),
    v2 = rnorm(100)
  )
  y <- as_draws(x)
  expect_is(y, "draws_df")
  expect_equal(variables(y), names(x))
  expect_equal(niterations(y), 100)
  expect_equal(nchains(y), 1)

  # are .iteration and .chain automatically used?
  x2 <- x
  x2$.iteration <- 1:100
  x2$.chain <- rep(1:4, each = 25)
  y2 <- as_draws(x2)
  expect_is(y2, "draws_df")
  expect_equal(variables(y2), names(x))
  expect_equal(niterations(y2), 25)
  expect_equal(nchains(y2), 4)
})

test_that("lists can be transformed to draws_list objects", {
  x <- list(
    list(
      v1 = rnorm(50),
      v2 = rnorm(50)
    ),
    list(
      v1 = rnorm(50),
      v2 = rnorm(50)
    )
  )
  y <- as_draws(x)
  expect_is(y, "draws_list")
  expect_equal(variables(y), names(x[[1]]))
  expect_equal(ndraws(y), 100)
  expect_equal(nchains(y), 2)
})

test_that("numeric vectors can be transformed to draws_matrix objects", {
  draws_matrix <- draws_matrix(a = 1:10, b = 11:20, c = 1)
  draws_matrix2 <- as_draws_matrix(cbind(1:10, 11:20, 1))
  expect_equivalent(draws_matrix, draws_matrix2)
})

test_that("numeric vectors can be transformed to draws_array objects", {
  draws_array <- draws_array(a = 1:10, b = 11:20, c = 1, .nchains = 2)
  draws_array2 <- array(c(1:10, 11:20, rep(1, 10)), c(5, 2, 3))
  dimnames(draws_array2)[[3]] <- c("a", "b", "c")
  draws_array2 <- as_draws_array(draws_array2)
  expect_equal(draws_array, draws_array2)
})

test_that("numeric vectors can be transformed to draws_df objects", {
  draws_df <- draws_df(a = 1:10, b = 11:20, c = 1, .nchains = 2)
  draws_array <- array(c(1:10, 11:20, rep(1, 10)), c(5, 2, 3))
  dimnames(draws_array)[[3]] <- c("a", "b", "c")
  draws_df2 <- as_draws_df(draws_array)
  expect_equal(draws_df, draws_df2)
})

test_that("numeric vectors can be transformed to draws_list objects", {
  draws_list <- draws_list(a = 1:10, b = 11:20, c = 1, .nchains = 2)
  draws_array <- array(c(1:10, 11:20, rep(1, 10)), c(5, 2, 3))
  dimnames(draws_array)[[3]] <- c("a", "b", "c")
  draws_list2 <- as_draws_list(draws_array)
  expect_equal(draws_list, draws_list2)
})

test_that("lists of data.frames are correctly transformed to draws_list objects", {
  x <- list(data.frame(y=1:10, x = 0), data.frame(y = 11:20, x = 1))
  y <- as_draws_list(x)
  expect_equal(y[[1]]$y, 1:10)
  expect_true(is.list(y[[1]]))
  expect_true(!is.data.frame(y[[1]]))
})

test_that("numeric vectors can be transformed to draws_rvars objects", {
  draws_rvars <- draws_rvars(a = 1:10, b = 11:20, c = 1, .nchains = 2)
  draws_array <- array(c(1:10, 11:20, rep(1, 10)), c(5, 2, 3))
  dimnames(draws_array)[[3]] <- c("a", "b", "c")
  draws_rvars2 <- as_draws_rvars(draws_array)
  expect_equal(draws_rvars, draws_rvars2)
})

test_that("rvars can be transformed to draws objects", {
  rv_array <- array(c(1:10, 11:20, rep(1, 10)), c(5, 2, 3))
  rv <- rvar(rv_array)
  draws_rvars <- draws_rvars(x = rv)
  expect_equal(as_draws(rv), draws_rvars)
  expect_equal(as_draws_rvars(rv), draws_rvars)
  expect_equal(as_draws_matrix(rv), as_draws_matrix(draws_rvars))
  expect_equal(as_draws_array(rv), as_draws_array(draws_rvars))
  expect_equal(as_draws_df(rv), as_draws_df(draws_rvars))
  expect_equal(as_draws_list(rv), as_draws_list(draws_rvars))
})

test_that("mcmc and mcmc.list objects can be transformed to draws objects", {
  # don't want to add coda as dependency so construct equivalent of
  # mcmc and mcmc.list objects
  x1 <- matrix(1:20, 10, 2)
  x2 <- matrix(21:40, 10, 2)
  dimnames(x1) <- dimnames(x2) <- list(1:10, c("A", "B"))
  attr(x1, "mcpar") <- attr(x2, "mcpar") <- c(1, 10, 1)
  class(x1) <- class(x2) <- "mcmc"
  xlist <- structure(list(x1 = x1, x2 = x2), class = "mcmc.list")

  mcmc_draws <- list(
    as_draws(x1),
    as_draws_matrix(x1),
    as_draws_array(x1),
    as_draws_df(x1),
    as_draws_list(x1),
    as_draws_rvars(x1)
  )
  mcmc_list_draws <- list(
    as_draws(xlist),
    as_draws_matrix(xlist),
    as_draws_array(xlist),
    as_draws_df(xlist),
    as_draws_list(xlist),
    as_draws_rvars(xlist)
  )
  for (j in seq_along(mcmc_draws)) {
    xj <- mcmc_draws[[j]]
    expect_equal(ndraws(xj), 10)
    expect_equal(nvariables(xj), 2)
    expect_equal(nchains(xj), 1)
    expect_equal(variables(xj), c("A", "B"))

    xj <- mcmc_list_draws[[j]]
    expect_equal(ndraws(xj), 20)
    expect_equal(nchains(xj), 2)
    expect_equal(nvariables(xj), 2)
    expect_equal(variables(xj), c("A", "B"))
  }
})

test_that("empty draws objects can be converted", {
  empty_draws <- list(
    empty_draws_matrix(),
    empty_draws_array(),
    empty_draws_list(),
    empty_draws_df(),
    empty_draws_rvars()
  )
  for (j in seq_along(empty_draws)) {
    # basically just check they don't error and preserve 0 draws
    empty_j <- empty_draws[[j]]
    expect_equal(ndraws(as_draws_matrix(empty_j)), 0)
    expect_equal(ndraws(as_draws_array(empty_j)), 0)
    expect_equal(ndraws(as_draws_df(empty_j)), 0)
    expect_equal(ndraws(as_draws_list(empty_j)), 0)
    expect_equal(ndraws(as_draws_rvars(empty_j)), 0)
  }
})

test_that("as_draws throws appropriate error if no close format", {
  expect_error(
    as_draws("A"),
    "Don't know how to transform an object of class 'character'"
  )
  expect_error(
    as_draws(TRUE),
    "Don't know how to transform an object of class 'logical'"
  )
})

test_that("draws_* constructors throw correct errors", {
  expect_error(draws_array(a = 1, .nchains = 0), "Number of chains must be positive")
  expect_error(draws_df(a = 1, .nchains = 0), "Number of chains must be positive")
  expect_error(draws_list(a = 1, .nchains = 0), "Number of chains must be positive")

  expect_error(draws_array(a = 1, .nchains = 2), "Number of chains does not divide the number of draws")
  expect_error(draws_df(a = 1, .nchains = 2), "Number of chains does not divide the number of draws")
  expect_error(draws_list(a = 1, .nchains = 2), "Number of chains does not divide the number of draws")
})

test_that("as_draws_rvars correctly reshapes missing, out-of-order, and string array indices", {
  x_array <- as_draws_array(example_draws())
  variables(x_array) <- paste0("var[", rep(1:2, each = 5), ",", rep(1:5, 2), "]")
  x_rvars <- as_draws_rvars(x_array)

  x_array2 <- remove_variables(x_array, "var[2,3]")
  x_rvars2 <- x_rvars
  x_rvars2$var[2,3] <- NA
  expect_equal(as_draws_rvars(x_array2), x_rvars2)

  x_array2 <- subset_draws(x_array, variable = c("var[1,4]", "var[2,3]"))
  x_rvars2 <- x_rvars
  x_rvars2$var <- x_rvars2$var[1:2,1:4]
  x_rvars2$var[1,1:3] <- NA
  x_rvars2$var[2,c(1:2,4)] <- NA
  expect_equal(as_draws_rvars(x_array2), x_rvars2)

  x_rvars2 <- x_rvars
  rownames(x_rvars2$var) <- letters[1:2]
  colnames(x_rvars2$var) <- rev(letters[1:5])
  expect_equal(as_draws_rvars(as_draws_array(x_rvars2)), x_rvars2)
})

test_that("as_draws_rvars handles integer indices < 1", {
  draws_df <- draws_df(`x[-1]` = 1:2, `x[-3]` = 3:4, `x[-2]` = 5:6)

  x <- rvar(array(c(3:4, 5:6, 1:2), dim = c(2, 3), dimnames = list(NULL, -3:-1)))
  expect_equal(as_draws_rvars(draws_df), draws_rvars(x = x))
})

test_that("as_draws_rvars and draws_rvars handle unnamed variables", {
  # as_draws_rvars should provide a default name
  expect_equal(as_draws_rvars(list(rvar(1:2))), draws_rvars(`...1` = rvar(1:2)))
  # draws_rvars should throw an error
  expect_error(draws_rvars(rvar(1:2)), "All variables must be named")
})

test_that("as_draws_rvars can accept lists of lists as input", {
  # for https://github.com/stan-dev/posterior/issues/192
  draws_list <- as_draws_list(example_draws())
  list_of_lists <- unclass(draws_list)

  expect_equal(as_draws_rvars(example_draws()), as_draws_rvars(list_of_lists))
})

test_that("as_draws_rvars preserves dims of length-1 arrays", {
  draws_array <- draws_array(`x[1,1,1]` = 1:5)
  expect_equal(
    draws_of(as_draws_rvars(draws_array)$x),
    array(1:5, dim = c(5, 1, 1, 1), dimnames = list(1:5, NULL, NULL, NULL))
  )
})

test_that("draws_df does not munge variable names", {
  draws_df <- draws_df(`x[1]` = 1:2, `x[2]` = 3:4)
  expect_equal(variables(draws_df), c("x[1]", "x[2]"))
})

test_that("draws_df can roundtrip through data.frame", {
  draws_df <- draws_df(`x[1]` = 1:2, `x[2]` = 3:4)
  expect_equal(as_draws_df(as.data.frame(draws_df)), draws_df)
  expect_equal(as_draws_df(tibble::as_tibble(draws_df)), draws_df)
  draws_dataframe <- as.data.frame(draws_df)
  draws_dataframe$.draw <- NULL
  expect_equal(as_draws_df(draws_dataframe), draws_df)
})

test_that("draws_df drops the draws class when metadata is removed", {
  skip_if_not_installed("dplyr")

  draws_df <- as_draws_df(example_draws())

  expect_equal(dplyr::count(draws_df, .chain), tibble::tibble(.chain = 1:4, n = 100L))
})

test_that("as_draws_rvars can parse nested and malformed indices", {
  draws_df <- draws_df(`x[y[1],1]` = 1:2, `x[y[2,1]` = 3:4, `x[y[2,2]` = 4:5, `y[` = 5:6, `y]` = 6:7)

  draws_rvars <- draws_rvars(
    x = rvar(array(
      c(1:4, NA_real_, NA_real_, 4:5),
      dim = c(2, 2, 2),
      dimnames = list(NULL, c("y[1]", "y[2"), NULL)
    )),
    `y[` = rvar(5:6),
    `y]` = rvar(6:7)
  )
  expect_equal(as_draws_rvars(draws_df), draws_rvars)
})

test_that("draws_rvars broadcasts 0-length rvars to the size of other rvars", {
  draws <- draws_rvars(x = rvar(), y = rvar(1:10))

  expect_equal(draws_of(draws$x), array(numeric(), dim = c(10, 0), dimnames = list(1:10)))
})

test_that("0-length rvars can be cast to draws formats", {
  rvar0 <- rvar(array(numeric(), dim = c(10, 0)), nchains = 2)

  draws_rvars <- as_draws_rvars(rvar0)
  expect_equal(ndraws(draws_rvars), 10)
  expect_equal(niterations(draws_rvars), 5)
  expect_equal(nchains(draws_rvars), 2)
  expect_equal(nvariables(draws_rvars), 1)

  draws_matrix <- as_draws_matrix(rvar0)
  expect_equal(ndraws(draws_matrix), 10)
  expect_equal(niterations(draws_matrix), 5)
  expect_equal(nchains(draws_matrix), 2)
  # draws_matrix can't represent 0-length variables, they are dropped
  expect_equal(nvariables(draws_matrix), 0)

  draws_array <- as_draws_array(rvar0)
  expect_equal(ndraws(draws_array), 10)
  expect_equal(niterations(draws_array), 5)
  expect_equal(nchains(draws_array), 2)
  # draws_array can't represent 0-length variables, they are dropped
  expect_equal(nvariables(draws_array), 0)

  draws_df <- as_draws_df(rvar0)
  expect_equal(ndraws(draws_df), 10)
  expect_equal(niterations(draws_df), 5)
  expect_equal(nchains(draws_df), 2)
  # draws_df can't represent 0-length variables, they are dropped
  expect_equal(nvariables(draws_df), 0)

  draws_list <- as_draws_list(rvar0)
  expect_equal(nchains(draws_list), 2)
  # draws_df can't represent 0-length variables, or the iterations / draws
  # containing them, they are dropped
  expect_equal(ndraws(draws_list), 0)
  expect_equal(niterations(draws_list), 0)
  expect_equal(nvariables(draws_list), 0)

  # a zero-length draws_rvars with no names should also be equivalent
  draws_rvars_0 <- draws_rvars(x = rvar0)
  names(draws_rvars_0) <- NULL
  expect_equal(as_draws_matrix(draws_rvars_0), as_draws_matrix(draws_rvars))
  expect_equal(as_draws_array(draws_rvars_0), as_draws_array(draws_rvars))
  expect_equal(as_draws_df(draws_rvars_0), as_draws_df(draws_rvars))
  expect_equal(as_draws_list(draws_rvars_0), as_draws_list(draws_rvars))

  # now a draws_rvars with one 0-length and one non-0-length variable
  y <- rvar(1:10, nchains = 2)
  draws_rvars_xy <- draws_rvars(x = rvar0, y = y)
  draws_rvars_y <- draws_rvars(y = y)
  expect_equal(as_draws_matrix(draws_rvars_xy), as_draws_matrix(draws_rvars_y))
  expect_equal(as_draws_array(draws_rvars_xy), as_draws_array(draws_rvars_y))
  expect_equal(as_draws_df(draws_rvars_xy), as_draws_df(draws_rvars_y))
  expect_equal(as_draws_list(draws_rvars_xy), as_draws_list(draws_rvars_y))

})

test_that("1-length rvars can be cast to other formats", {
  # scalars should stay as scalars...
  x <- rvar(1:10)
  expect_equal(as_draws_array(x), draws_array(x = 1:10))
  expect_equal(as_draws_df(x), draws_df(x = 1:10))
  expect_equal(as_draws_list(x), draws_list(x = 1:10))
  expect_equal(as_draws_matrix(x), draws_matrix(x = 1:10))

  # ... but extra dims should not be dropped
  x <- rvar(array(1:10, dim = c(10, 1, 1)))
  expect_equal(as_draws_array(x), draws_array(`x[1,1]` = 1:10))
  expect_equal(as_draws_df(x), draws_df(`x[1,1]` = 1:10))
  expect_equal(as_draws_list(x), draws_list(`x[1,1]` = 1:10))
  expect_equal(as_draws_matrix(x), draws_matrix(`x[1,1]` = 1:10))
})


# discrete variables ------------------------------------------------------

test_that("conversion between formats supporting discrete variables work", {
  draws_rvars <- draws_rvars(
    y = rvar_factor(array(letters[1:24], dim = c(2,2,3,2)), with_chains = TRUE),
    z = rvar_ordered(array(letters[1:12], dim = c(2,2,3)), with_chains = TRUE),
    x = rvar(array(1:12, dim = c(2,2,3)), with_chains = TRUE)
  )

  draws <- list(
    df = as_draws_df(draws_rvars),
    list = as_draws_list(draws_rvars),
    rvars = draws_rvars
  )

  expect_equal(levels(draws$df[["z[1]"]]), levels(draws_rvars$z))
  expect_equal(levels(draws$list[[1]][["z[1]"]]), levels(draws_rvars$z))

  for (type in names(draws)) {
    expect_equal(as_draws_df(draws[[!!type]]), draws$df)
    expect_equal(as_draws_list(draws[[!!type]]), draws$list)
    expect_equal(as_draws_rvars(draws[[!!type]]), draws$rvars)
  }
})

test_that("lossy conversion to formats that don't support discrete variables works", {
  draws_rvars <- draws_rvars(
    y = rvar_factor(array(letters[1:24], dim = c(2,2,3,2)), with_chains = TRUE),
    z = rvar_ordered(array(letters[1:12], dim = c(2,2,3)), with_chains = TRUE),
    x = rvar(array(1:12, dim = c(2,2,3)), with_chains = TRUE)
  )

  draws <- list(
    df = as_draws_df(draws_rvars),
    list = as_draws_list(draws_rvars),
    rvars = draws_rvars
  )

  lossy_draws_rvars <- draws_rvars(
    y = rvar(array(1:24, dim = c(2,2,3,2)), with_chains = TRUE),
    z = rvar(array(1:12, dim = c(2,2,3)), with_chains = TRUE),
    x = rvar(array(1:12, dim = c(2,2,3)), with_chains = TRUE)
  )

  lossy_draws <- list(
    array = as_draws_array(lossy_draws_rvars),
    matrix = as_draws_matrix(lossy_draws_rvars)
  )

  for (type in names(draws)) {
    expect_warning(
      expect_equal(as_draws_array(draws[[!!type]]), lossy_draws$array),
      "draws_array does not support non-numeric variables"
    )
    expect_warning(
      expect_equal(as_draws_matrix(draws[[!!type]]), lossy_draws$matrix),
      "draws_matrix does not support non-numeric variables"
    )
  }
})

#' @export
length.rvar <- function(x) {
  prod(dim(x))
}

#' @export
dim.rvar <- function(x) {
  dim(draws_of(x))[-1]
}

#' @export
`dim<-.rvar` <- function(x, value) {
  if (length(value) == 0) {
    # vectors have NULL dim; for us that means
    # dim of c(ndraws(x), length(x))
    value = length(x)
  }
  # must keep old rowname around and restore them, since changing dim will drop them
  old_rownames <- rownames(draws_of(x))
  dim(draws_of(x)) <- c(ndraws(x), value)
  rownames(draws_of(x)) <- old_rownames
  x
}

#' @export
dimnames.rvar <- function(x) {
  .dimnames <- dimnames(draws_of(x))
  .dimnames[-1]
}

#' @export
`dimnames<-.rvar` <- function(x, value) {
  dimnames(draws_of(x)) <- c(list(rownames(draws_of(x))), value)
  x
}

#' @export
names.rvar <- function(x) {
  .dimnames <- dimnames(draws_of(x))
  .dimnames[[2]]
}

#' @export
`names<-.rvar` <- function(x, value) {
  dimnames(draws_of(x))[2] <- list(value)
  x
}

#' @importFrom methods setGeneric
#' @export
setGeneric("drop")

#' Drop redundant dimensions
#'
#' Delete the dimensions of an [`rvar`] which are of size one. See [`base::drop()`]
#'
#' @param x (rvar) an [`rvar`].
#'
#' @return
#' An [`rvar`] with the same length as `x`, but where any entry equal to `1`
#' in `dim(x)` has been removed. The exception is if `dim(x) == 1`, in which
#' case `dim(drop(x)) == 1` as well (this is because [`rvar`]s, unlike [`numeric`]s,
#' never have `NULL` dimensions).
#'
#' @examples
#' # Sigma is a 3x3 covariance matrix
#' Sigma <- as_draws_rvars(example_draws("multi_normal"))$Sigma
#' Sigma
#'
#' Sigma[1, ]
#'
#' drop(Sigma[1, ])
#'
#' # equivalently ...
#' Sigma[1, drop = TRUE]
#'
#' @importFrom methods setMethod
#' @export
setMethod("drop", signature(x = "rvar"), function(x) {
  .dim <- dim(x)

  if (length(.dim) > 1) {
    # with exactly 1 dimension left we don't want to drop anything
    # (otherwise names get lost), so only do this with > 1 dimension
    keep_dim <- .dim != 1
    .dimnames <- dimnames(x)
    dim(x) <- .dim[keep_dim]
    # for comparison / testing, ensure if no dimnames have names that we
    # actually have those names be NULL (rather than just empty strings)
    new_dimnames <- .dimnames[keep_dim]
    if (all(names(new_dimnames) == "")) names(new_dimnames) <- NULL
    dimnames(x) <- new_dimnames
  }

  x
})

#' The `draws_df` format
#'
#' @name draws_df
#' @family formats
#'
#' @templateVar draws_format draws_df
#' @templateVar base_class class(tibble::tibble())
#' @template draws_format-skeleton
#' @template args-format-nchains
#'
#' @details Objects of class `"draws_df"` are [tibble][tibble::tibble] data
#'   frames. They have one column per variable as well as additional metadata
#'   columns `".iteration"`, `".chain"`, and `".draw"`. The difference between
#'   the `".iteration"` and `".draw"` columns is that the former is relative to
#'   the MCMC chain while the latter ignores the chain information and has all
#'   unique values. See **Examples**.
#'
#'   If a `data.frame`-like object is supplied to `as_draws_df` that contains
#'   columns named `".iteration"` or `".chain"`, they will be treated as
#'   iteration and chain indices, respectively. See **Examples**.
#'
#' @examples
#'
#' # the difference between iteration and draw is clearer when contrasting
#' # the head and tail of the data frame
#' print(head(x1), reserved = TRUE, max_variables = 2)
#' print(tail(x1), reserved = TRUE, max_variables = 2)
#'
#' # manually supply chain information
#' xnew <- data.frame(mu = rnorm(10), .chain = rep(1:2, each = 5))
#' xnew <- as_draws_df(xnew)
#' print(xnew)
#'
NULL


#' @rdname draws_df
#' @export
as_draws_df <- function(x, ...) {
  UseMethod("as_draws_df")
}

#' @rdname draws_df
#' @export
as_draws_df.default <- function(x, ...) {
  x <- as_draws(x)
  as_draws_df(x, ...)
}

#' @rdname draws_df
#' @export
as_draws_df.data.frame <- function(x, ...) {
  .as_draws_df(x)
}

#' @rdname draws_df
#' @export
as_draws_df.draws_df <- function(x, ...) {
  x
}

#' @rdname draws_df
#' @export
as_draws_df.draws_matrix <- function(x, ...) {
  if (ndraws(x) == 0L) {
    return(empty_draws_df(variables(x)))
  }
  iteration_ids <- iteration_ids(x)
  chain_ids <- chain_ids(x)
  attr(x, "nchains") <- NULL
  x <- tibble::as_tibble(unclass(x))
  x[[".chain"]] <- rep(chain_ids, each = max(iteration_ids))
  x[[".iteration"]] <- rep(iteration_ids, max(chain_ids))
  x[[".draw"]] <- seq_len(nrow(x))
  class(x) <- class_draws_df()
  x
}

#' @rdname draws_df
#' @export
as_draws_df.draws_array <- function(x, ...) {
  if (ndraws(x) == 0L) {
    return(empty_draws_df(variables(x)))
  }
  x <- as_draws_matrix(x)
  as_draws_df(x)
}

#' @rdname draws_df
#' @export
as_draws_df.draws_list <- function(x, ...) {
  if (ndraws(x) == 0L) {
    return(empty_draws_df(variables(x)))
  }
  iteration_ids <- iteration_ids(x)
  chain_ids <- chain_ids(x)
  vars <- names(x[[1L]])
  x <- do.call(vctrs::vec_rbind, lapply(x, vctrs::new_data_frame))
  x <- tibble::as_tibble(x)
  x[[".chain"]] <- rep(chain_ids, each = max(iteration_ids))
  x[[".iteration"]] <- rep(iteration_ids, max(chain_ids))
  x[[".draw"]] <- seq_len(nrow(x))
  class(x) <- class_draws_df()
  x
}

#' @rdname draws_df
#' @export
as_draws_df.draws_rvars <- function(x, ...) {
  if (ndraws(x) == 0L) {
    return(empty_draws_df(variables(x)))
  }
  out <- do.call(cbind, lapply(seq_along(x), function(i) {
    # flatten each rvar so it only has two dimensions: draws and variables
    # this also collapses indices into variable names in the format "var[i,j,k,...]"
    flatten_rvar_draws_to_df(x[[i]], names(x)[[i]])
  }))
  iteration_ids <- iteration_ids(x)
  chain_ids <- chain_ids(x)
  out[[".chain"]] <- rep(chain_ids, each = max(iteration_ids))
  out[[".iteration"]] <- rep(iteration_ids, max(chain_ids))
  out[[".draw"]] <- draw_ids(x)
  class(out) <- class_draws_df()
  out
}

#' @rdname draws_df
#' @export
as_draws_df.mcmc <- function(x, ...) {
  as_draws_df(as_draws_matrix(x), ...)
}

#' @rdname draws_df
#' @export
as_draws_df.mcmc.list <- function(x, ...) {
  as_draws_df(as_draws_array(x), ...)
}

#' Convert any \R object into a `draws_df` object
#' @param x An \R object.
#' @noRd
.as_draws_df <- function(x) {
  x <- tibble::as_tibble(x, .name_repair = "unique")

  # prepare iteration and chain indices
  has_iteration_column <- ".iteration" %in% names(x)
  if (has_iteration_column) {
    iteration_ids <- x[[".iteration"]]
    x[[".iteration"]] <- NULL
  } else {
    iteration_ids <- seq_len(nrow(x))
  }
  has_chain_column <- ".chain" %in% names(x)
  if (has_chain_column) {
    chain_ids <- x[[".chain"]]
    x[[".chain"]] <- NULL
  } else {
    chain_ids <- rep(1L, nrow(x))
  }
  # drop draw indices since they are regenerated below
  x[[".draw"]] <- NULL

  # add reserved variables to the data
  check_new_variables(names(x))
  x[[".chain"]] <- chain_ids
  x[[".iteration"]] <- iteration_ids
  if (has_iteration_column || has_chain_column) {
    x[[".chain"]] <- repair_chain_ids(x[[".chain"]])
    x[[".iteration"]] <- repair_iteration_ids(x[[".iteration"]], x[[".chain"]])
  }
  x[[".draw"]] <- compute_draw_ids(x[[".chain"]], x[[".iteration"]])
  class(x) <- class_draws_df()
  x
}

#' @rdname draws_df
#' @export
draws_df <- function(..., .nchains = 1) {
  out <- validate_draws_per_variable(...)
  .nchains <- as_one_integer(.nchains)
  if (.nchains < 1) {
    stop_no_call("Number of chains must be positive.")
  }
  ndraws <- length(out[[1]])
  if (ndraws %% .nchains != 0) {
    stop_no_call("Number of chains does not divide the number of draws.")
  }
  niterations <- ndraws %/% .nchains
  out <- as.data.frame(out, optional = TRUE)
  out[[".iteration"]] <- rep(seq_len(niterations), .nchains)
  out[[".chain"]] <- rep(seq_len(.nchains), each = niterations)
  as_draws_df(out)
}

class_draws_df <- function() {
  # inherits for tibbles
  c("draws_df", "draws", "tbl_df", "tbl", "data.frame")
}

#' @rdname draws_df
#' @export
is_draws_df <- function(x) {
  inherits(x, "draws_df")
}

# is an object looking like a 'draws_df' object?
is_draws_df_like <- function(x) {
  is.data.frame(x)
}

#' @export
`[.draws_df` <- function(x, i, j, drop = FALSE, ..., reserved = FALSE) {
  reserved <- as_one_logical(reserved)
  # draws_df is a tibble so drop = FALSE by default anyway
  out <- NextMethod("[")
  if (reserved) {
    reserved_vars <- all_reserved_variables(x)
    reserved_vars <- setdiff(reserved_vars, names(out))
    out[, reserved_vars] <- NextMethod("[", j = reserved_vars, drop = FALSE)
  } else {
    out <- drop_draws_class_if_metadata_removed(out, warn = TRUE)
  }
  out
}

# This generic is not exported here as {dplyr} is only in Suggests, so
# we must export it in .onLoad() for compatibility with R < 3.6.
dplyr_reconstruct.draws_df <- function(data, template) {
  data <- NextMethod("dplyr_reconstruct")
  data <- drop_draws_class_if_metadata_removed(data, warn = FALSE)
  data
}

# drop "draws_df" and "draws" classes if metadata columns were removed
# from the data frame
drop_draws_class_if_metadata_removed <- function(x, warn = TRUE) {
  if (!all(reserved_df_variables() %in% names(x))) {
    if (warn) warning_no_call("Dropping 'draws_df' class as required metadata was removed.")
    class(x) <- setdiff(class(x), c("draws_df", "draws"))
  }
  x
}

# create an empty draws_df object
empty_draws_df <- function(variables = character(0)) {
  assert_character(variables, null.ok = TRUE)
  x <- tibble::tibble()
  x[variables] <- numeric(0)
  x[c(".chain", ".iteration", ".draw")] <- integer(0)
  class(x) <- class_draws_df()
  x
}

test_that("rstar returns reasonable values", {
  skip_if_not_installed("caret")
  skip_if_not_installed("randomForest")
  x <- example_draws()
  val <- rstar(x)
  expect_true(val > 0.8 & val < 10)
})

test_that("rstar works with 1d example", {
  skip_if_not_installed("caret")
  skip_if_not_installed("randomForest")
  x <- example_draws()
  x <- as_draws_df(x)
  # remove all bar one variable
  x <- x[, c(variables(x)[1], ".chain", ".iteration", ".draw")]
  val <- rstar(x)
  expect_true(val > 0.5 & val < 10)
})

test_that("rstar works with draws_df example", {
  skip_if_not_installed("caret")
  skip_if_not_installed("randomForest")
  x <- example_draws()
  x <- as_draws_df(x)
  val <- rstar(x)
  expect_true(val > 0.5 & val < 10)
})

test_that("rstar with uncertainty returns vectors of correct length", {
  skip_if_not_installed("caret")
  x <- example_draws()
  val <- rstar(x, method = "gbm", uncertainty = T, verbose = F)
  expect_equal(length(val), 1000)
  val <- rstar(x, method = "knn", uncertainty = T, nsimulations = 10)
  expect_equal(length(val), 10)
})

test_that("incorrect nsimulations values throws error", {
  skip_if_not_installed("caret")
  x <- example_draws()
  expect_error(rstar(x, method = "knn", nsimulations = 0),
               "'nsimulations' must be greater than or equal to 1.")
})

test_that("rstar with uncertainty returns reasonable values", {
  skip_if_not_installed("caret")
  x <- example_draws()
  val <- rstar(x, method = "gbm", uncertainty = T, verbose = F)
  expect_true(max(val) > 0.3 & min(val) < 10)
})

test_that("rstar accepts different classifiers", {
  skip_if_not_installed("caret")
  x <- example_draws()
  val <- rstar(x, method = "gbm", verbose=F)
  expect_true(is.numeric(val))
  val <- rstar(x, method = "knn")
  expect_true(is.numeric(val))
})

test_that("rstar accepts different hyperparameters", {
  skip_if_not_installed("caret")
  x <- example_draws()

  # use fast hyperparameters
  caret_grid <- data.frame(interaction.depth=c(3),
                       n.trees = 1,
                       shrinkage=c(0.1),
                       n.minobsinnode=10)
  start <- Sys.time()
  val <- rstar(x, method = "gbm", verbose=F,
               hyperparameters = caret_grid)
  end <- Sys.time()
  dif1 <- end - start
  # use slower hyperparameters
  caret_grid <- data.frame(interaction.depth=c(3),
                       n.trees = 1000,
                       shrinkage=c(0.1),
                       n.minobsinnode=10)
  start <- Sys.time()
  val <- rstar(x, method = "gbm", verbose=F,
               hyperparameters = caret_grid)
  end <- Sys.time()
  dif2 <- end - start
  expect_true(dif1 < dif2)
})

test_that("rstar accepts different training proportion", {
  skip_if_not_installed("caret")
  x <- example_draws()
  val1 <- rstar(x, method = "knn")
  val2 <- rstar(x, method = "knn", training_proportion = 0.1)
  expect_true(val1 > val2)
})

test_that("rstar throws error when passed invalid training_proportion", {
  skip_if_not_installed("caret")
  x <- example_draws()
  expect_error(rstar(x, method = "knn", training_proportion = 0),
               "'training_proportion' must be greater than 0 and less than 1")
  expect_error(rstar(x, method = "knn", training_proportion = 1),
               "'training_proportion' must be greater than 0 and less than 1")
})

test_that("split-chain R* returns generally higher values", {
  skip_if_not_installed("caret")
  skip_on_cran()  # reduces test time
  x <- example_draws()
  n <- 10
  vals_split <- vector(length = n)
  vals_unsplit <- vector(length = n)
  for(i in 1:n) {
    vals_split[i] <- rstar(x, method = "knn")
    vals_unsplit[i] <- rstar(x, method = "knn", split = FALSE)
  }
  expect_true(median(vals_split) > median(vals_unsplit))
})

#' Mutate variables in `draws` objects
#'
#' Mutate variables in a [`draws`] object.
#'
#' @param .x (draws) A [`draws`] object.
#' @param ... Name-value pairs of expressions, each with either length 1 or the
#'   same length as in the entire input (i.e., number of iterations or draws).
#'   The name of each argument will be the name of a new variable, and the value
#'   will be its corresponding value. Use a `NULL` value in `mutate_variables`
#'   to drop a variable. New variables overwrite existing variables of the same
#'   name.
#'
#' @return
#' Returns a [`draws`] object of the same format as `.x`, with variables mutated
#' according to the expressions provided in `...`.
#'
#' @details
#' In order to mutate variables in [`draws_matrix`] and [`draws_array`] objects,
#' they are transformed to [`draws_df`] objects first and then transformed back
#' after mutation. As those transformations are quite expensive for larger
#' number of draws, we recommend using `mutate_variables` on [`draws_df`] and
#' [`draws_list`] objects if speed is an issue.
#'
#' In [`draws_rvars`] objects, the output of each expression in `...` is
#' coerced to an [`rvar`] object if it is not already one using `as_rvar()`.
#'
#' @seealso [`variables`], [`rename_variables`]
#'
#' @examples
#' x <- as_draws_df(example_draws())
#' x <- subset(x, variable = c("mu", "tau"))
#'
#' mutate_variables(x, tau2 = tau^2)
#' mutate_variables(x, scale = 1.96 * tau, lower = mu - scale)
#'
#' @importFrom rlang enquos caller_env eval_tidy as_label
#' @export
mutate_variables <- function(.x, ...) {
  UseMethod("mutate_variables")
}

#' @rdname mutate_variables
#' @export
mutate_variables.draws_matrix <- function(.x, ...) {
  as_draws_matrix(mutate_variables(as_draws_df(.x), ...))
}

#' @rdname mutate_variables
#' @export
mutate_variables.draws_array <- function(.x, ...) {
  as_draws_array(mutate_variables(as_draws_df(.x), ...))
}

#' @rdname mutate_variables
#' @export
mutate_variables.draws_df <- function(.x, ...) {
  dots <- enquos(..., .named = TRUE)
  names(dots) <- check_reserved_variables(names(dots))
  env <- caller_env()
  for (var in names(dots)) {
    .x[[var]] <- .mutate_variable(dots[[var]], .x, env)
  }
  .x
}

#' @rdname mutate_variables
#' @export
mutate_variables.draws_list <- function(.x, ...) {
  dots <- enquos(..., .named = TRUE)
  names(dots) <- check_reserved_variables(names(dots))
  env <- caller_env()
  for (chain in seq_along(.x)) {
    for (var in names(dots)) {
      .x[[chain]][[var]] <- .mutate_variable(dots[[var]], .x[[chain]], env)
    }
  }
  .x
}

#' @rdname mutate_variables
#' @export
mutate_variables.draws_rvars <- function(.x, ...) {
  dots <- enquos(..., .named = TRUE)
  names(dots) <- check_reserved_variables(names(dots))
  env <- caller_env()
  for (var in names(dots)) {
    .x[[var]] <- as_rvar(eval_tidy(dots[[var]], .x, env))
  }
  conform_rvar_ndraws_nchains(.x)
}

# evaluate an expression passed to 'mutate_variables' and check its validity
.mutate_variable <- function(expr, data, env = caller_env()) {
  out <- eval_tidy(expr, data, env)
  if (is.null(out)) {
    return(NULL)
  }
  if (!is.numeric(out)) {
    stop_no_call("{", as_label(expr), "} does not evaluate to a numeric vector.")
  }
  n <- length(data[[1]])
  if (length(out) == 1L) {
    out <- rep(out, n)
  }
  if (length(out) != n) {
    stop_no_call("{", as_label(expr), "} does not evaluate ",
          "to a vector of length 1 or ", n, ".")
  }
  out
}

#' Nested Rhat convergence diagnostic
#'
#' Compute the nested Rhat convergence diagnostic for a single
#' variable as proposed in Margossian et al. (2024).
#'
#' @family diagnostics
#' @template args-conv
#' @param superchain_ids (numeric) Vector of length nchains specifying
#'   which superchain each chain belongs to. There should be equal
#'   numbers of chains in each superchain. All chains within the same
#'   superchain are assumed to have been initialized at the same
#'   point.
#' @template args-methods-dots
#'
#' @details Nested Rhat is a convergence diagnostic useful when
#'   running many short chains. It is calculated on superchains, which
#'   are groups of chains that have been initialized at the same
#'   point.
#'
#' Note that there is a slight difference in the calculation of Rhat
#'   and nested Rhat, as nested Rhat is lower bounded by 1. This means
#'   that nested Rhat with one chain per superchain will not be
#'   exactly equal to basic Rhat (see Footnote 3 in Margossian et
#'   al. (2024)).
#'
#' @template return-conv
#' @template ref-margossian-nestedrhat-2024
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' rhat_nested(mu, superchain_ids = c(1, 1, 2, 2))
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' rhat_nested(d$Sigma, superchain_ids = c(1, 1, 2, 2))
#'
#' @export
rhat_nested <- function(x, ...) UseMethod("rhat_nested")

#' @rdname rhat_nested
#' @export
rhat_nested.default <- function(x, superchain_ids, ...) {
  .rhat_nested(x, superchain_ids = superchain_ids)
}

#' @rdname rhat_nested
#' @export
rhat_nested.rvar <- function(x, superchain_ids, ...) {
  summarise_rvar_by_element_with_chains(
    x, rhat_nested, superchain_ids = superchain_ids, ...
  )
}

.rhat_nested <- function(x, superchain_ids, ...) {
  if (should_return_NA(x)) {
    return(NA_real_)
  }

  x <- as.matrix(x)
  niterations <- NROW(x)
  nchains <- NCOL(x)

  # check that all chains are assigned a superchain
  if (length(superchain_ids) != nchains) {
    warning_no_call("Length of superchain_ids not equal to number of chains, ",
                    "returning NA.")
    return(NA_real_)
  }

  # check that superchains are equal length
  superchain_id_table <- table(superchain_ids)
  nchains_per_superchain <- max(superchain_id_table)

  if (nchains_per_superchain != min(superchain_id_table)) {
    warning_no_call("Number of chains per superchain is not the same for ",
                    "each superchain, returning NA.")
    return(NA_real_)
  }

  superchains <- unique(superchain_ids)

  # mean and variance of chains calculated as in Rhat
  chain_mean <- matrixStats::colMeans2(x)
  chain_var <- matrixStats::colVars(x, center = chain_mean)

  # mean of superchains calculated by only including specified chains
  # (equation 4 in Margossian et al. 2024)
  superchain_mean <- sapply(
    superchains, function(k) mean(x[, which(superchain_ids == k)])
  )

  # between-chain variance estimate (Bhat_k in equation 7 in Margossian et al. 2024)
  if (nchains_per_superchain == 1) {
    var_between_chain <- 0
  } else {
    var_between_chain <- sapply(
      superchains,
      function(k) var(chain_mean[which(superchain_ids == k)])
    )
  }

  # within-chain variance estimate (What_k in equation 7 in Margossian et al. 2024)
  if (niterations == 1) {
    var_within_chain <- 0
  } else {
    var_within_chain <- sapply(
      superchains,
      function(k) mean(chain_var[which(superchain_ids == k)])
    )
  }

  # between-superchain variance (Bhat_nu in equation 6 in Margossian et al. 2024)
  var_between_superchain <- var(superchain_mean)

  # within-superchain variance (What_nu in equation 7 in Margossian et al. 2024)
  var_within_superchain <- mean(var_within_chain + var_between_chain)

  # nested Rhat (Rhat_nu in equation 8 in Margossian et al. 2024)
  sqrt(1 + var_between_superchain / var_within_superchain)
}

test_that("rhat_nested returns reasonable values", {
  tau <- extract_variable_matrix(example_draws(), "tau")

  nested_rhat <- rhat_nested(tau, superchain_ids = c(1, 2, 1, 2))
  expect_true(nested_rhat > 1 & nested_rhat < 1.05)

  nested_rhat <- rhat_nested(tau, superchain_ids = c(1, 2, 1, 2))
  expect_true(nested_rhat > 1 & nested_rhat < 1.05)
})


test_that("rhat_nested handles special cases correctly", {
  set.seed(1234)
  x <- c(rnorm(10), NA)
  expect_true(is.na(rhat_nested(x, superchain_ids = c(1))))

  x <- c(rnorm(10), Inf)
  expect_true(is.na(rhat_nested(x, superchain_ids = c(1, 2, 1, 2))))

  tau <- extract_variable_matrix(example_draws(), "tau")
  expect_warning(
    rhat_nested(tau, superchain_ids = c(1, 1, 1, 3)),
    "Number of chains per superchain is not the same for each superchain, returning NA."
  )

  tau <- extract_variable_matrix(example_draws(), "tau")
  expect_warning(
    rhat_nested(tau, superchain_ids = c(1, 2)),
    "Length of superchain_ids not equal to number of chains, returning NA."
  )
})

# initialize a named list
# @param names names of the elements
# @param values optional values of the elements
named_list <- function(names, values = NULL) {
  if (!is.null(values)) {
    if (length(values) <= 1L) {
      values <- replicate(length(names), values)
    }
    values <- as.list(values)
    stopifnot(length(values) == length(names))
  } else {
    values <- vector("list", length(names))
  }
  setNames(values, names)
}

# unlist lapply output
ulapply <- function(X, FUN, ..., recursive = TRUE, use.names = TRUE) {
  unlist(lapply(X, FUN, ...), recursive, use.names)
}

seq_rows <- function(x) {
  seq_len(NROW(x))
}

seq_cols <- function(x) {
  seq_len(NCOL(x))
}

# selectively drop one-level dimensions of an array and/or reset object classes
drop_dims_or_classes <- function(x, dims = NULL, reset_class = FALSE) {
  assert_array(x)
  assert_integerish(dims, null.ok = TRUE)
  reset_class <- as_one_logical(reset_class)
  old_dims <- dim(x)
  # proceed to drop dimensions if the input array has any non-NULL dimensions
  if (length(old_dims)) {
    # base::drop if all one-level dimensions are to be dropped non-selectively
    if (is.null(dims) || setequal(dims, which(old_dims == 1L))) {
      x <- drop(x)
      # custom drop if certain one-level dimensions are to be dropped selectively
    } else {
      dim(x) <- old_dims[-dims]
      old_dimnames <- dimnames(x)
      # if all names of new dimnames are empty strings (""), set them to NULL
      new_dimnames <- old_dimnames[-dims]
      if (all(names(new_dimnames) == "")) {
        names(new_dimnames) <- NULL
      }
      dimnames(x) <- new_dimnames
    }
  }
  # optionally, set class to NULL and let R decide appropriate classes
  if (reset_class) {
    class(x) <- NULL
  }
  x
}

'%||%' <- function(x, y) {
  if (is.null(x)) x <- y
  x
}

# cat with without separating elements
cat0 <- function(..., file = "", fill = FALSE, labels = NULL, append = FALSE) {
  cat(..., sep = "", file = file, fill = fill, labels = labels, append = append)
}

# coerce 'x' to a single logical value
as_one_logical <- function(x, allow_na = FALSE) {
  s <- substitute(x)
  x <- as.logical(x)
  if (length(x) != 1L || anyNA(x) && !allow_na) {
    s <- deparse_pretty(s)
    stop_no_call("Cannot coerce '", s, "' to a single logical value.")
  }
  x
}

# coerce 'x' to a single integer value
as_one_integer <- function(x, allow_na = FALSE) {
  s <- substitute(x)
  x <- SW(as.integer(x))
  if (length(x) != 1L || anyNA(x) && !allow_na) {
    s <- deparse_pretty(s)
    stop_no_call("Cannot coerce '", s, "' to a single integer value.")
  }
  x
}

# coerce 'x' to a single numeric value
as_one_numeric <- function(x, allow_na = FALSE) {
  s <- substitute(x)
  x <- SW(as.numeric(x))
  if (length(x) != 1L || anyNA(x) && !allow_na) {
    s <- deparse_pretty(s)
    stop_no_call("Cannot coerce '", s, "' to a single numeric value.")
  }
  x
}

# coerce 'x' to a single character string
as_one_character <- function(x, allow_na = FALSE) {
  s <- substitute(x)
  x <- as.character(x)
  if (length(x) != 1L || anyNA(x) && !allow_na) {
    s <- deparse_pretty(s)
    stop_no_call("Cannot coerce '", s, "' to a single character value.")
  }
  x
}

# check if all inputs are NULL
all_null <- function(...) {
  all(ulapply(list(...), is.null))
}

# check if two objects are equal
is_equal <- function(x, y, ...) {
  isTRUE(all.equal(x, y, ...))
}

# move elements to the start of a named object
move_to_start <- function(x, start) {
  assert_named(x)
  start <- intersect(names(x), start)
  if (!length(start)) {
    return(x)
  }
  out <- x[c(start, setdiff(names(x), start))]
  class(out) <- class(x)
  out
}

# prettily deparse an expression
# @return a single character string
deparse_pretty <- function(x, max_chars = NULL, max_wsp = 1L) {
  out <- collapse(deparse(x))
  out <- rm_wsp(out, max_wsp)
  assert_int(max_chars, null.ok = TRUE)
  if (isTRUE(max_chars > 0L)) {
    out <- substr(out, 1L, max_chars)
  }
  out
}

# remove NULL elements from a list
remove_null <- function(x) {
  Filter(Negate(is.null), x)
}

# remove whitespaces from strings
rm_wsp <- function(x, max_wsp = 0) {
  assert_character(x)
  assert_int(max_wsp)
  wsp <- collapse(rep(" ", max_wsp))
  gsub("[ \t\r\n]+", wsp, x, perl = TRUE)
}

# collapse a character vector
collapse <- function(..., sep = "") {
  paste(..., sep = sep, collapse = "")
}

# like 'eval' but parses characters before evaluation
eval2 <- function(expr, envir = parent.frame(), ...) {
  if (is.character(expr)) {
    expr <- parse(text = expr)
  }
  eval(expr, envir, ...)
}

# wrapper around replicate but without simplifying
repl <- function(expr, n) {
  replicate(n, expr, simplify = FALSE)
}

isNA <- function(x) {
  length(x) == 1L && is.na(x)
}

is_constant <- function(x, tol = .Machine$double.eps) {
  abs(max(x) - min(x)) < tol
}

# collapse variables via commas for pretty printing
comma <- function(...) {
  paste0("{", paste0("'", c(...), "'", collapse = ", "), "}")
}

stop_no_call <- function(...) {
  stop(..., call. = FALSE)
}

warning_no_call <- function(...) {
  warning(..., call. = FALSE)
}

SW <- function(expr) {
  base::suppressWarnings(expr)
}

# escape all special characters in character strings
escape_all <- function(x) {
  specials <- c(".", "*", "+", "?", "^", "$", "(", ")", "[", "]", "|")
  for (s in specials) {
    x <- gsub(s, paste0("\\", s), x, fixed = TRUE)
  }
  x
}

# numerically stable version of log(sum(exp(x)))
log_sum_exp <- function(x) {
  max <- max(x)
  sum <- sum(exp(x - max))
  max + log(sum)
}

# simple version of destructuring assignment
`%<-%` <- function(vars, values, envir = parent.frame()) {
  vars <- as.character(substitute(vars)[-1])
  for (i in seq_along(vars)) {
    assign(vars[[i]], values[[i]], envir = envir)
  }
  invisible(NULL)
}

# get names of all the functions of a package
package_function_names <- function(package) {
  unclass(utils::lsf.str(envir = asNamespace(package), all = TRUE))
}

# return true if there is an S3 method for the
# given function and class signature
has_s3_method <- function(f, signature) {
  for (class in signature) {
    if (!is.null(utils::getS3method(f, class, optional = TRUE))) {
      return(TRUE)
    }
  }
  FALSE
}

# Ensure that log_sum_exp agrees with expected behaviour from log(sum(exp(x)))
test_that("log_sum_exp of for x containing Inf is Inf", {
  x <- c(Inf, 1)
  expect_equal(log_sum_exp(x), Inf)
})

test_that("log_sum_exp of -Inf is -Inf", {
  x <- c(-Inf, -Inf)
  expect_equal(log_sum_exp(x), -Inf)
})

test_that("log_sum_exp of empty input is -Inf", {
  x <- numeric(0)
  expect_equal(log_sum_exp(x), -Inf)
})

test_that("log_sum_exp works", {
  set.seed(1)
  x <- log(runif(10))
  expect_equal(log_sum_exp(x), log(sum(exp(x))))
})

#' Example `draws` objects
#'
#' Objects for use in examples, vignettes, and tests.
#'
#' @param example (string) The name of the example `draws` object. See
#'   **Details** for available options.
#' @return A `draws` object.
#'
#' @details
#' The following example `draws` objects are available.
#'
#' **eight_schools**: A [`draws_array`] object with 100 iterations
#' from each of 4 Markov chains obtained by fitting the eight schools model
#' described in Gelman et al. (2013) with [Stan](https://mc-stan.org). The
#' variables are:
#' * `mu`: Overall mean of the eight schools
#' * `tau`: Standard deviation between schools
#' * `theta`: Individual means of each of the eight schools
#'
#' **multi_normal**: A [`draws_array`] object with 100 iterations from each of
#' the 4 Markov chains obtained by fitting a 3-dimensional multivariate normal
#' model to 100 simulated observations. The variables are:
#' * `mu`: Mean parameter vector of length 3
#' * `Sigma`: Covariance matrix of dimension 3 x 3
#'
#' @note These objects are only intended to be used in demonstrations and tests.
#'   They contain fewer iterations and chains than recommended for performing
#'   actual inference.
#'
#' @references
#' Andrew Gelman, John B. Carlin, Hal S. Stern, David B. Dunson, Aki Vehtari and
#' Donald B. Rubin (2013). Bayesian Data Analysis, Third Edition. Chapman and
#' Hall/CRC.
#'
#' @examples
#' draws_eight_schools <- example_draws("eight_schools")
#' summarise_draws(draws_eight_schools)
#'
#' draws_multi_normal <- example_draws("multi_normal")
#' summarise_draws(draws_multi_normal)
#'
#' @export
example_draws <- function(example = "eight_schools") {
  assert_choice(example, example_choices())
  # saved in R/sysdata.rda
  draws_name <- paste0("draws_", example)
  get(draws_name, asNamespace("posterior"))
}

# names of choices for 'example_draws'
example_choices <- function() {
  c("eight_schools", "multi_normal")
}

# since rename_variables is just a wrapper around variables() this only
# tests it on draws_matrix for now. See the tests for variables() for
# format-specific tests.
test_that("rename_variables works on draws_matrix", {
  x = as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b[1]"))))

  ref = as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("mu", "b[1]"))))
  expect_equal(rename_variables(x, mu = a), ref)
  expect_equal(rename_variables(x, mu = "a"), ref)

  ref = as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("c", "d"))))
  expect_equal(rename_variables(x, c = a, d = `b[1]`), ref)
  expect_equal(rename_variables(x, c = a, d = "b[1]"), ref)

  # renaming can be chained
  expect_equal(rename_variables(x, e = a, d = `b[1]`, c = e), ref)

  # no renaming
  expect_equal(rename_variables(x), x)
})

test_that("rename_variables works for non-scalar variables", {
  x <- example_draws()
  x <- rename_variables(x, alpha = theta)
  vars <- c("mu", "tau", paste0("alpha[", 1:8, "]"))
  expect_equal(variables(x), vars)
})

test_that("cannot rename a variable to a reserved word", {
  x = as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b[1]"))))
  err = "Variable names.*are reserved"
  expect_error(rename_variables(x, .chain = a), err)
  expect_error(rename_variables(x, .iteration = a), err)
  expect_error(rename_variables(x, .draw = a), err)
})

test_that("cannot rename a variable to an existing variable name", {
  x = as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b[1]"))))
  expect_error(rename_variables(x, a = "b[1]"), "Duplicate variable names are not allowed")
})

test_that("cannot rename a variable to an empty name", {
  x = as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b[1]"))))
  expect_error(rename_variables(x, a), "Cannot rename a variable to an empty name")
})

test_that("cannot rename a non-existent variable", {
  x = as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b[1]"))))
  expect_error(rename_variables(x, a = c), "The following variables are missing")
})

#' Random variables of arbitrary dimension
#'
#' Random variables backed by arrays of arbitrary dimension
#'
#' @name rvar
#'
#' @param x (multiple options) The object to convert to an `rvar`:
#'   * A vector of draws from a distribution.
#'   * An array where the first dimension represents draws from a distribution.
#'     The resulting [`rvar`] will have dimension `dim(x)[-1]`; that is,
#'     everything except the first dimension is used for the shape of the
#'     variable, and the first dimension is used to index draws from the
#'     distribution (see **Examples**). Optionally,
#'     if `with_chains == TRUE`, the first dimension indexes the iteration and the
#'     second dimension indexes the chain (see `with_chains`).
#'   * An `rvar`.
#' @template args-rvar-dim
#' @template args-rvar-dimnames
#' @param nchains (positive integer) The number of chains. The if `NULL` (the default),
#' `1` is used unless `x` is already an [`rvar`], in which case the number of
#' chains it has is used.
#' @param with_chains (logical) Does `x` include a dimension for chains?
#' If `FALSE` (the default), chains are not included, the first dimension of
#' the input array should index draws, and the `nchains` argument can be
#' used to determine the number of chains. If `TRUE`, the `nchains` argument
#' is ignored and the second dimension of `x` is used to index chains.
#' Internally, the array will be converted to a format without the chain index.
#' Ignored when `x` is already an [`rvar`].
#'
#' @details
#'
#' The `"rvar"` class internally represents random variables as arrays of arbitrary
#' dimension, where the first dimension is used to index draws from the distribution.
#
#' Most mathematical operators and functions are supported, including efficient matrix
#' multiplication and vector and array-style indexing. The intent is that an `rvar`
#' works as closely as possible to how a base vector/matrix/array does, with a few
#' differences:
#'
#' - The default behavior when subsetting is not to drop extra dimensions (i.e.
#'   the default `drop` argument for `[` is `FALSE`, not `TRUE`).
#' - Rather than base R-style recycling, `rvar`s use a limited form of broadcasting:
#'   if an operation is being performed on two vectors with different size of the same
#'   dimension, the smaller vector will be recycled up to the size of the larger one
#'   along that dimension so long as it has size 1.
#'
#' For functions that expect base numeric arrays and for which `rvar`s cannot be
#' used directly as arguments, you can use [rfun()] or [rdo()] to translate your
#' code into code that executes across draws from one or more random variables
#' and returns a random variable as output. Typically [rdo()] offers the most
#' straightforward translation.
#'
#' As [rfun()] and [rdo()] incur some performance cost, you can also operate directly
#' on the underlying array using the [draws_of()] function. To re-use existing
#' random number generator functions to efficiently create `rvar`s, use [rvar_rng()].
#'
#' @seealso [as_rvar()] to convert objects to `rvar`s. See [rdo()], [rfun()], and
#' [rvar_rng()] for higher-level interfaces for creating `rvar`s.
#'
#' @return An object of class `"rvar"` representing a random variable.
#'
#' @examples
#'
#' set.seed(1234)
#'
#' # To create a "scalar" `rvar`, pass a one-dimensional array or a vector
#' # whose length (here `4000`) is the desired number of draws:
#' x <- rvar(rnorm(4000, mean = 1, sd = 1))
#' x
#'
#' # Create random vectors by adding an additional dimension:
#' n <- 4   # length of output vector
#' x <- rvar(array(rnorm(4000 * n, mean = rep(1:n, each = 4000), sd = 1), dim = c(4000, n)))
#' x
#'
#' # Create a random matrix:
#' rows <- 4
#' cols <- 3
#' x <- rvar(array(rnorm(4000 * rows * cols, mean = 1, sd = 1), dim = c(4000, rows, cols)))
#' x
#'
#' # If the input sample comes from multiple chains, we can indicate that using the
#' # nchains argument (here, 1000 draws each from 4 chains):
#' x <- rvar(rnorm(4000, mean = 1, sd = 1), nchains = 4)
#' x
#'
#' # Or if the input sample has chain information as its second dimension, we can
#' # use with_chains to create the rvar
#' x <- rvar(array(rnorm(4000, mean = 1, sd = 1), dim = c(1000, 4)), with_chains = TRUE)
#' x
#'
#' @export
rvar <- function(x = double(), dim = NULL, dimnames = NULL, nchains = NULL, with_chains = FALSE) {
  if (is_rvar(x)) {
    nchains <- nchains %||% nchains(x)
    with_chains = FALSE
    x <- draws_of(x)
  }

  with_chains <- as_one_logical(with_chains)
  if (with_chains) {
    nchains <- dim(x)[[2]] %||% 1L
    x <- drop_chain_dim(x)
  } else {
    nchains <- nchains %||% 1L
  }

  out <- new_rvar(x, .nchains = nchains)

  if (!is.null(dim)) {
    dim(out) <- dim
  }
  if (!is.null(dimnames)) {
    dimnames(out) <- dimnames
  }

  out
}

#' @importFrom vctrs new_vctr
new_rvar <- function(x = double(), .nchains = 1L) {
  if (is.null(x)) {
    x <- double()
  }

  x <- cleanup_rvar_draws(x)

  .ndraws <- dim(x)[[1]]
  .nchains <- as_one_integer(.nchains)
  check_nchains_compat_with_ndraws(.nchains, .ndraws)

  structure(
    list(),
    draws = x,
    nchains = .nchains,
    class = get_rvar_class(x),
    cache = new.env(parent = emptyenv())
  )
}


# manipulating raw draws array --------------------------------------------

#' Get/set array of draws underlying a random variable
#'
#' Gets/sets the array-representation that backs an [`rvar`]. Should be used rarely.
#'
#' @param x (rvar) An [`rvar`] object.
#' @param value (array) An array of values to use as the backing array of `x`.
#' @param with_chains (logical) Should the array of draws include a dimension for chains?
#' If `FALSE` (the default), chains are not included and the array has dimension
#' `c(ndraws(x), dim(x))`. If `TRUE`, chains are included and the array has
#' dimension `c(niterations(x), nchains(x), dim(x))`.
#'
#' @details
#'
#' While [`rvar`]s implement fast versions of basic math operations (including
#' [matrix multiplication][rvar-matmult]), sometimes you may need to bypass
#' the [`rvar`] abstraction to do what you need to do more efficiently.
#' `draws_of()` allows you to get / set the underlying array of draws in
#' order to do that.
#'
#' [`rvar`]s represent draws internally using arrays of arbitrary dimension, which
#' is returned by `draws_of(x)` and can be set using `draws_of(x) <- value`.
#' The **first** dimension of these arrays is the index of the draws. If
#' `with_chains = TRUE`, then the dimensions of the returned array are modified
#' so that the first dimension is the index of the iterations and the second
#' dimension is the index of the chains.
#'
#' @return
#'
#' If `with_chains = FALSE`, an array with dimensions `c(ndraws(x), dim(x))`.
#'
#' If `with_chains = TRUE`, an array with dimensions
#' `c(niterations(x), nchains(x), dim(x))`.
#'
#'
#' @examples
#'
#' x <- rvar(1:10, nchains = 2)
#' x
#'
#' # draws_of() without arguments will return the array of draws without
#' # chain information (first dimension is draw)
#' draws_of(x)
#'
#' # draws_of() with with_chains = TRUE will reshape the returned array to
#' # include chain information in the second dimension
#' draws_of(x, with_chains = TRUE)
#'
#' # you can also set draws using draws_of(). When with_chains = FALSE the
#' # existing chain information will be retained ...
#' draws_of(x) <- 2:11
#' x
#'
#' # when with_chains = TRUE the chain information will be set by the
#' # second dimension of the assigned array
#' draws_of(x, with_chains = TRUE) <- array(2:11, dim = c(2,5))
#' x
#'
#' @export
draws_of <- function(x, with_chains = FALSE) {
  with_chains <- as_one_logical(with_chains)
  draws <- attr(x, "draws")

  if (with_chains) {
    x_dim <- dim(x)
    dim(draws) <- c(niterations(x), nchains(x), x_dim)
    x_dim_i <- seq_along(x_dim)
    draws <- copy_dimnames(x, x_dim_i, draws, x_dim_i + 2)
  }

  draws
}

#' @rdname draws_of
#' @export
`draws_of<-` <- function(x, with_chains = FALSE, value) {
  with_chains <- as_one_logical(with_chains)

  if (with_chains) {
    draws <- drop_chain_dim(value)
    nchains_rvar(x) <- dim(value)[[2]] %||% 1L
  } else {
    draws <- value
  }

  draws <- cleanup_rvar_draws(draws)
  attr(x, "draws") <- draws
  class(x) <- get_rvar_class(draws)

  x <- invalidate_rvar_cache(x)
  x
}


# misc standard methods --------------------------------------------------

#' @export
rep.rvar <- function(x, times = 1, length.out = NA, each = 1, ...) {
  # flatten before rep()ing
  dim(x) <- length(x)

  if (each != 1) {
    x = vec_restore(rep(vec_proxy(x), each = each), x)
  }

  draws = draws_of(x)
  if (is.na(length.out)) {
    # use `times`
    rep_draws = rep(draws, times)
    dim = dim(draws)
    dim[[2]] = dim[[2]] * times
    dim(rep_draws) = dim
    out <- new_rvar(rep_draws, .nchains = nchains(x))
  } else {
    # use `length.out`
    rep_draws = rep_len(draws, length.out * ndraws(x))
    dim(rep_draws) = c(ndraws(x), length(rep_draws) / ndraws(x))
    out <- new_rvar(rep_draws, .nchains = nchains(x))
  }
  out
}

#' @method rep.int rvar
#' @export
rep.int.rvar <- function(x, times) {
  rep(x, times = times)
}

#' @method rep_len rvar
#' @export
rep_len.rvar <- function(x, length.out) {
  rep(x, length.out = length.out)
}

#' @export
unique.rvar <- function(x, incomparables = FALSE, MARGIN = 1, ...) {
  draws_margin <- check_rvar_margin(x, MARGIN)
  draws_of(x) <- unique(draws_of(x), incomparables = incomparables, MARGIN = draws_margin, ...)
  x
}

#' @export
duplicated.rvar <- function(x, incomparables = FALSE, MARGIN = 1, ...) {
  draws_margin <- check_rvar_margin(x, MARGIN)
  duplicated(draws_of(x), incomparables = incomparables, MARGIN = draws_margin, ...)
}

#' @export
anyDuplicated.rvar <- function(x, incomparables = FALSE, MARGIN = 1, ...) {
  draws_margin <- check_rvar_margin(x, MARGIN)
  anyDuplicated(draws_of(x), incomparables = incomparables, MARGIN = draws_margin, ...)
}

# check that MARGIN is a valid margin for the dimensions of rvar x
# then return the corresponding margin for draws_of(x)
check_rvar_margin <- function(x, MARGIN) {
  if (!(1 <= MARGIN && MARGIN <= length(dim(x)))) {
    stop_no_call("MARGIN = ", MARGIN, " is invalid for length(dim(x)) = ", length(dim(x)))
  }
  MARGIN + 1
}

#' @export
all.equal.rvar <- function(target, current, ...) {
  result <- NULL

  if (data.class(target) != data.class(current)) {
    result <- c(result, paste0(
      "target is ", data.class(target),
      ", current is ", data.class(current)
    ))
  }

  # ignore cache in comparison
  .target <- unclass(target)
  attr(.target, "cache") <- NULL
  .current <- unclass(current)
  attr(.current, "cache") <- NULL

  object_result <- all.equal(.target, .current, ...)
  if (!isTRUE(object_result)) {
    result = c(result, object_result)
  }

  if (is.null(result)) TRUE else result
}


# match and %in% ----------------------------------------------------------

#' Value Matching
#'
#' Generic version of [base::match()]. For base vectors, returns a vector of the
#' positions of (first) matches of its first argument in its second. For [rvar]s,
#' returns an [rvar] of the matches.
#'
#' @inheritDotParams base::match
#' @param x (multiple options) the values to be matched. Can be:
#'  - A base vector: see [base::match()]
#'  - An [rvar]
#' @param table (vector) the values to be matched against.
#'
#' @details
#' For more information on how match behaves with base vectors, see [base::match()].
#'
#' When `x` is an [rvar], the draws of `x` are matched against `table` using
#' [base::match()], and the result is returned as an [rvar].
#'
#' The implementation of `%in%` here is identical to \code{base::%in%}, except
#' it uses the generic version of `match()` so that non-base vectors (such
#' as [rvar]s) are supported.
#'
#' @returns
#' When `x` is a base vector, a vector of the same length as `x`.
#'
#' When `x` is an [rvar], an [rvar] the same shape as `x`.
#' @examples
#' x <- rvar(c("a","b","b","c","d"))
#' x %in% c("b","d")
#'
#' # for additional examples, see base::match()
#' @export
match <- function(x, table, ...) UseMethod("match")

#' @rdname match
#' @export
match.default <- function(x, ...) base::match(x, ...)

#' @rdname match
#' @export
match.rvar <- function(x, ...) {
  draws_of(x) <- while_preserving_dims(base::match, draws_of(x), ...)
  x
}

#' @rdname match
#' @export
`%in%` <- function(x, table) {
  match(x, table, nomatch = 0L) > 0L
}


# ggplot2::scale_type -----------------------------------------------------

# This generic is not exported here as {ggplot2} is only in Suggests, so
# we must export it in .onLoad() for compatibility with R < 3.6.
scale_type.rvar <- function(x) {
  # ensure that rvars used in ggplot2 aesthetics (as in e.g. ggdist) are passed
  # through without modification and without raising a warning
  "identity"
}


# helpers: classes --------------------------------------------------------

get_rvar_class <- function(x) {
  UseMethod("get_rvar_class")
}

#' @export
get_rvar_class.default <- function(x) {
  c("rvar", "vctrs_vctr")
}

#' @export
get_rvar_class.factor <- function(x) {
  c("rvar_factor", NextMethod())
}

#' @export
get_rvar_class.ordered <- function(x) {
  c("rvar_ordered", NextMethod())
}

#' @importFrom methods setOldClass
setOldClass(get_rvar_class(numeric()))
setOldClass(get_rvar_class(factor()))
setOldClass(get_rvar_class(ordered(NULL)))


# helpers: validation -----------------------------------------------------------------

# Check the passed yank index (for x[[...]]) is valid
check_rvar_yank_index = function(x, i, ...) {
  index <- dots_list(i, ..., .preserve_empty = TRUE, .ignore_empty = "none")

  index_lengths <- lengths(index)
  if (any(index_lengths == 0)) {
    stop_no_call("Cannot select zero elements with `[[` in an rvar.")
  } else if (any(index_lengths > 1)) {
    stop_no_call("Cannot select more than one element per index with `[[` in an rvar.")
  } else if (any(vapply(index, function(x) is_missing(x) || is.na(x), logical(1)))) {
    stop_no_call("Missing indices not allowed with `[[` in an rvar.")
  } else if (any(vapply(index, is.logical, logical(1)))) {
    stop_no_call("Logical indices not allowed with `[[` in an rvar.")
  } else if (any(vapply(index, function(x) x < 0, logical(1)))) {
    stop_no_call("subscript out of bounds")
  }

  index
}

# Check the passed subset indices (for x[...]) do not go beyond the end
# of the valid dimensions
check_rvar_subset_indices = function(x, ...) {
  ndim = max(length(dim(x)), 1)
  if (length(substitute(list(...))) - 1 > ndim) {
    stop_no_call("Cannot index past dimension ", ndim, ".")
  }
}

# find common ndraws for the ndraws of two rvars to be broadcast to
ndraws2_common <- function(ndraws_x, ndraws_y) {
  if (ndraws_x == 1) {
    ndraws_y
  } else if (ndraws_y == 1) {
    ndraws_x
  } else if (ndraws_x == ndraws_y) {
    ndraws_x
  } else {
    stop_no_call(
      "Random variables have different number of draws (", ndraws_x,
      " and ", ndraws_y, ") and cannot be used together."
    )
  }
}

# find common nchains for the nchains of two rvars to be set to
# nchains_x or nchains_y may be NULL to indicate they are constants
# and take the nchains of the other
nchains2_common <- function(nchains_x, nchains_y) {
  # constants should give nchains of NULL for input to this function
  # so they are treated as having any number of chains
  if (is.null(nchains_x)) {
    nchains_y
  } else if (is.null(nchains_y)) {
    nchains_x
  } else if (nchains_x == nchains_y) {
    nchains_x
  } else {
    warn_merge_chains("match")
    1L
  }
}

# check that the given number of chains is compatible with the given number of draws
check_nchains_compat_with_ndraws <- function(nchains, ndraws) {
  # except with constants, nchains must divide the number of draws
  if (ndraws != 1 && isTRUE(ndraws %% nchains != 0)) {
    stop_no_call("Number of chains does not divide the number of draws.")
  }
  if (nchains < 1) {
    stop_no_call("Number of chains must be >= 1")
  }
}

# given two rvars, conform their number of chains
# so they can be used together (or throw an error if they can't be)
conform_rvar_nchains <- function(rvars) {
  # find the number of chains to use, treating constants as having any number of chains
  nchains_or_null <- lapply(rvars, function(x) if (ndraws(x) == 1) NULL else nchains(x))
  .nchains <- Reduce(nchains2_common, nchains_or_null) %||% 1L

  for (i in seq_along(rvars)) {
    nchains_rvar(rvars[[i]]) <- .nchains
  }

  rvars
}

# given two rvars, conform their number of draws and chains
# so they can be used together (or throw an error if they can't be)
# @param keep_constants keep constants as 1-draw rvars
conform_rvar_ndraws_nchains <- function(rvars, keep_constants = FALSE) {
  rvars <- conform_rvar_nchains(rvars)

  # broadcast to a common number of chains. If keep_constants = TRUE,
  # constants will not be broadcast.
  .ndraws = Reduce(ndraws2_common, lapply(rvars, ndraws))
  for (i in seq_along(rvars)) {
    rvars[[i]] <- broadcast_draws(rvars[[i]], .ndraws, keep_constants)
  }

  rvars
}

# Check that the first rvar can be conformed to the dimensions of the second,
# ignoring 1s
check_rvar_dims_first <- function(x, y) {
  x_dim <- dim(x)
  x_dim_dropped <- as.integer(x_dim[x_dim != 1])
  y_dim <- dim(y)
  y_dim_dropped <- as.integer(y_dim[y_dim != 1])

  if (length(x_dim_dropped) == 0) {
    # x can be treated as scalar, do so
    dim(x) <- rep(1, length(dim(y)))
  } else if (identical(x_dim_dropped, y_dim_dropped)) {
    dim(x) <- dim(y)
  } else {
    stop_no_call("Cannot assign an rvar with dimension ", paste0(x_dim, collapse = ","),
      " to an rvar with dimension ", paste0(y_dim, collapse = ","))
  }

  x
}


# helpers: arrays/lists -----------------------------------------------------------------

#' Get one draw from a draws_rvars as a list of base-R variables
#' @param x a draws_rvars
#' @param i a draw index
#' @returns a named list of vectors and arrays
#' @noRd
get_variables_from_one_draw <- function(x, i) {
  lapply(x, function(variable) {
    draws <- draws_of(variable)
    .dim <- dim(variable)
    ndim <- length(.dim)

    if (ndim <= 1) {
      # treat 0- and 1-dimensional arrays as vectors
      draws <- draws[i, ]
      dim(draws) <- NULL
      names(draws) <- names(variable)
    } else {
      dim(draws) <- c(NROW(draws), length(variable))
      draws <- draws[i, ]
      dim(draws) <- .dim
      dimnames(draws) <- dimnames(variable)
    }

    draws
  })
}

dim2_common <- function(dim_x, dim_y) {
  # find common dim for two arrays to be broadcast to
  ndim_x <- length(dim_x)
  ndim_y <- length(dim_y)

  if (ndim_x < ndim_y) {
    dim_x <- c(dim_x, rep(1, ndim_y - ndim_x))
  } else {
    dim_y <- c(dim_y, rep(1, ndim_x - ndim_y))
  }

  pmax(dim_x, dim_y)
}

dim_common <- function(dims) {
  Reduce(dim2_common, dims)
}

broadcast_array  <- function(x, dim, broadcast_scalars = TRUE) {
  if (!broadcast_scalars && length(x) == 1) {
    # quick exit: not broadcasting scalars; return them as vectors
    dim(x) <- NULL
    return(x)
  }

  current_dim <- dim(x)
  current_dimnames <- dimnames(x)
  current_levels <- levels(x)
  current_class <- oldClass(x)

  if (length(current_dim) < length(dim)) {
    # add dimensions of size 1 as necessary so we can broadcast those
    new_dim <- seq(length(current_dim) + 1, length(dim))
    current_dim[new_dim] <- 1
    dim(x) <- current_dim
    if (!is.null(current_dimnames)) {
      current_dimnames[new_dim] <- list(NULL)
      dimnames(x) <- current_dimnames
    }
  } else if (length(current_dim) > length(dim)) {
    stop_no_call(
      "Cannot broadcast array of shape [", paste(current_dim, collapse = ","), "] ",
      "to array of shape [", paste(dim, collapse = ","), "]:\n",
      "Desired shape has fewer dimensions than existing array."
    )
  }

  dim_to_broadcast = which(current_dim != dim)

  if (length(dim_to_broadcast) == 0) {
    # quick exit: already has desired dim or just needed extra dims on the end
    return(x)
  }

  if (any(current_dim[dim_to_broadcast] != 1)) {
    stop_no_call(
      "Cannot broadcast array of shape [", paste(current_dim, collapse = ","), "] ",
      "to array of shape [", paste(dim, collapse = ","), "]:\n",
      "All dimensions must be 1 or equal."
    )
  }

  # move the dims we aren't broadcasting to the front so they are recycled properly
  perm <- c(seq_along(dim)[-dim_to_broadcast], dim_to_broadcast)

  # broadcast the other dims
  x <- array(aperm(x, perm), dim[perm])

  # move dims back to their original order
  x <- aperm(x, order(perm))

  if (!is.null(current_dimnames)) {
    # restore any dimnames that we did not have to broadcast
    dim_to_restore <- current_dim == dim
    dimnames(x)[dim_to_restore] <- current_dimnames[dim_to_restore]
  }

  # restore class and levels
  levels(x) <- current_levels
  oldClass(x) <- current_class

  x
}

# broadcast the draws dimension of an rvar to the requested size
broadcast_draws <- function(x, .ndraws, keep_constants = FALSE) {
  ndraws_x = ndraws(x)
  if (
    (ndraws_x == 1 && keep_constants) ||
    (ndraws_x == .ndraws)
  ) {
    x
  } else {
    draws <- draws_of(x)
    new_dim <- dim(draws)
    new_dim[1] <- .ndraws

    new_rvar(broadcast_array(draws, new_dim), .nchains = nchains(x))
  }
}

# flatten dimensions and names of an array
flatten_array = function(x, x_name = NULL) {
  # determine new dimension names in the form x,y,z
  # start with numeric names
  dimname_lists <- lapply(dim(x), seq_len)
  .dimnames <- dimnames(x)
  if (!is.null(.dimnames)) {
    # where character names are provided, use those instead of the numeric names
    dimname_lists = lapply(seq_along(dimname_lists), function(i) .dimnames[[i]] %||% dimname_lists[[i]])
  }
  # expand out the dimname lists into the appropriate combinations and assemble into new names
  dimname_grid <- expand.grid(dimname_lists)
  new_names <- apply(dimname_grid, 1, paste0, collapse = ",")

  .length <- length(x)
  old_dim <- dim(x)
  dim(x) <- .length

  # update variable names
  if (is.null(x_name)) {
    # no base name for x provided, just use index names
    names(x) <- new_names
  } else if (.length == 1 && (isTRUE(old_dim == 1) || length(old_dim) == 0)) {
    # scalar, use the provided base name
    names(x) <- x_name
  } else if (.length >= 1) {
    # rename the variables with their indices in brackets
    names(x) <- paste0(x_name, "[", new_names %||% seq_along(x), "]")
  }

  x
}

#' Fast conversion of rvar draws to a flattened data frame. Equivalent to
#' as.data.frame(draws_of(flatten_array(x, name))) except it works with
#' factor rvars (as.data.frame does not work on array-like factors)
#' @noRd
flatten_rvar_draws_to_df <- function(x, x_name = NULL) {
  if (length(x) == 0) {
    out <- data.frame(row.names = draw_ids(x))
  } else {
    draws <- draws_of(flatten_array(x, x_name))
    cols <- lapply(seq_len(ncol(draws)), function(i) unname(draws[, i]))
    names(cols) <- colnames(draws)
    out <- vctrs::new_data_frame(cols)
  }
  out
}

#' copy the dimension names (and name of the dimension) from dimension src_i
#' in array src to dimension dst_i in array dst
#' @noRd
copy_dimnames <- function(src, src_i, dst, dst_i) {
  if (is.null(dimnames(dst))) {
    if (is.null(dimnames(src))) {
      return(dst)
    }
    dimnames(dst) <- list(NULL)
  }

  if (is.null(dimnames(src))) {
    dimnames(src) <- list(NULL)
  }
  if (is.null(names(dimnames(src)))) {
    names(dimnames(src)) <- rep("", length(dim(src)))
  }

  dimnames(dst)[dst_i] <- dimnames(src)[src_i]
  names(dimnames(dst))[dst_i] <- names(dimnames(src))[src_i]
  names(dimnames(dst))[is.na(names(dimnames(dst)))] <- ""

  dst_dimnames <- names(dimnames(dst))
  empty_names <- is.na(dst_dimnames) | dst_dimnames == ""
  if (all(empty_names)) {
    names(dimnames(dst)) <- NULL
  }

  dst
}

#' Drop the chain dimension from an array (presumed to be the second dim)
#' @noRd
drop_chain_dim <- function(x) {
  if (length(x) == 0) {
    # quick exit: NULL input
    dim(x) <- c(1,0)
    return(x)
  }

  x_dim <- dim(x)
  if (length(x_dim) < 2) {
    stop_no_call("Cannot use an array of dimension less than 2 when with_chains equals TRUE")
  }

  out <- x
  dim(out) <- c(x_dim[[1]] * x_dim[[2]], x_dim[-c(1,2)])
  x_dim_i <- seq_along(x_dim)[-c(1,2)]
  out <- copy_dimnames(x, x_dim_i, out, x_dim_i - 1)
  out
}

#' Clean up the data type and dimensions of an array of draws intended to back an rvar.
#' Ensures that dim and dimnames are set, that the array has at least two dimensions
#' (first one is draws), etc.
#' @noRd
cleanup_rvar_draws <- function(x) {
  if (length(x) == 0) {
    # canonical NULL rvar is at least 1 draw of nothing
    # this ensures that (e.g.) extending a null rvar
    # with x[1] = something works.
    ndraws <- max(NROW(x), 1)
    dim(x) <- c(ndraws, 0)
  }
  else if (length(dim(x)) <= 1) {
    # 1d vectors get treated as a single variable
    dim(x) <- c(length(x), 1)
  }

  # ensure dimnames is set (makes comparison easier for tests)
  if (length(dimnames(x)) == 0) {
    dimnames(x) <- list(NULL)
  }

  # ensure we have an index for draws
  if (length(rownames(x)) == 0) {
    rownames(x) <- as.character(seq_rows(x))
  }

  # if x is a character array, make it a factor
  if (is.character(x)) {
    x <- while_preserving_dims(factor, x)
  }

  x
}

#' Execute x <- f(x, ...) but preserve dimensions and dimension names of x.
#' Useful for functions that do not change the length of x but which drop
#' dimensions.
#' @noRd
while_preserving_dims <- function(f, x, ...) {
  .dim <- dim(x)
  .dimnames <- dimnames(x)
  x <- f(x, ...)
  dim(x) <- .dim
  dimnames(x) <- .dimnames
  x
}

#' Execute x <- f(x, ...) but preserve class and levels of x.
#' Useful for functions that do not change the length of x but which levels.
#' @noRd
while_preserving_levels <- function(f, x, ...) {
  .class <- oldClass(x)
  .levels <- levels(x)
  x <- f(x, ...)
  oldClass(x) <- .class
  levels(x) <- .levels
  x
}

#' a version of apply() that works on factor-like arrays
#' apply() strips classes of slices of x before passing them to FUN, which
#' breaks apply() when used with factor-like arrays; this variant ensures
#' levels and classes are preserved for slices of factors passed to FUN
#' @noRd
.apply_factor <- function(X, MARGIN, FUN, ...) {
  if (is.factor(X)) {
    .class <- oldClass(X)
    .levels <- levels(X)
    FUN <- match.fun(FUN)
    .f <- function(x, ...) {
      oldClass(x) <- .class
      levels(x) <- .levels
      FUN(x, ...)
    }
  } else {
    .f <- FUN
  }
  apply(X, MARGIN, .f, ...)
}


# helpers: applying functions over rvars ----------------------------------

# apply a summary function within each draw of the rvar (dropping other dimensions)
# summarise_rvar_within_draws_via_matrix should be used instead of this function
# if a faster implementation of .f is available in matrix form (e.g. functions
# from matrixStats); otherwise this function can be used.
summarise_rvar_within_draws <- function(x, .f, ..., .transpose = FALSE, .when_empty = .f(numeric(0))) {
  draws <- draws_of(x)
  dim <- dim(draws)
  if (!length(x)) {
    # x is a NULL rvar, need to return base value for this summary function
    as_rvar(.when_empty)
  } else {
    draws <- apply(draws, 1, .f, ...)
    if (.transpose) draws <- t(draws)
    new_rvar(draws, .nchains = nchains(x))
  }
}

#' apply a summary function within each draw of the rvar (dropping other dimensions)
#' by first collapsing dimensions into columns of the draws matrix
#' (so that .f can be a rowXXX() function)
#' @param x an rvar
#' @param name function name to use for error messages
#' @param .f a function that takes a matrix and summarises its rows, like rowMeans
#' @param ... arguments passed to `.f`
#' @param .ordered_okay can this function be applied to rvar_ordereds?
#' @noRd
summarise_rvar_within_draws_via_matrix <- function(x, .name, .f, ..., .ordered_okay = FALSE) {
  .length <- length(x)
  if (!.length) {
    x <- rvar()
  }

  dim(x) <- .length

  if (.ordered_okay && is_rvar_ordered(x)) {
    .levels <- levels(x)
    .draws <- .f(draws_of(as_rvar_numeric(x)), ...)
    .draws <- while_preserving_dims(function(.draws) ordered(.levels[round(.draws)], .levels), .draws)
  } else if (is_rvar_factor(x)) {
    stop_no_call("Cannot apply `", .name, "` function to rvar_factor objects.")
  } else {
    .draws <- .f(draws_of(x), ...)
  }

  new_rvar(.draws, .nchains = nchains(x))
}

# apply vectorized function to an rvar's draws
rvar_apply_vec_fun <- function(.f, x, ...) {
  draws_of(x) <- .f(draws_of(x), ...)
  x
}

# apply a summary function across draws of the rvar (i.e., by each element)
summarise_rvar_by_element <- function(x, .f, ...) {
  if (length(x) == 1) {
    # this ensures that scalar rvars are summarized to vectors rather than
    # to matrices with one column
    .f(draws_of(x), ...)
  } else {
    draws <- draws_of(x)
    dim <- dim(draws)
    .apply_factor(draws, seq_along(dim)[-1], .f, ...)
  }
}

#' apply a summary function across draws of the rvar (i.e., by each element)
#' by first collapsing dimensions into columns of the draws matrix, applying the
#' function, then restoring dimensions (so that .f can be a colXXX() function)
#' @param x an rvar
#' @param name function name to use for error messages
#' @param .f a function that takes a matrix and summarises its columns, like colMeans
#' @param .extra_dim extra dims added by `.f` to the output, e.g. in the case of
#' matrixStats::colRanges this is `2`
#' @param .extra_dimnames extra dimension names for dims added by `.f` to the output
#' @param .ordered_okay can this function be applied to rvar_ordereds?
#' @param .factor_okay can this function be applied to rvar_factors?
#' @param ... arguments passed to `.f`
#' @noRd
summarise_rvar_by_element_via_matrix <- function(
  x, .name, .f, .extra_dim = NULL, .extra_dimnames = NULL, .ordered_okay = TRUE, .factor_okay = FALSE, ...
) {
  .dim <- dim(x)
  .dimnames <- dimnames(x)
  .length <- length(x)
  dim(x) <- .length

  if (!is_rvar_factor(x) || .factor_okay) {
    x <- .f(draws_of(x), ...)
  } else if (.ordered_okay && is_rvar_ordered(x)) {
    .levels <- levels(x)
    x <- .f(draws_of(as_rvar_numeric(x)), ...)
    x <- ordered(.levels[round(x)], .levels)
  } else {
    stop_no_call("Cannot apply `", .name, "` function to rvar_factor objects.")
  }

  if (is.null(.extra_dim) && length(.dim) <= 1) {
    # this ensures that vector rvars are summarized to vectors rather than
    # to arrays with one dimension
    dim(x) <- NULL
    names(x) <- .dimnames[[1]]
  } else if (isTRUE(.dim == 1)) {
    # scalars with extra dimensions should just return vectors
    dim(x) <- NULL
    names(x) <- .extra_dimnames[[1]]
  } else {
    dim(x) <- c(.extra_dim, .dim)
    dimnames(x) <- c(.extra_dimnames, .dimnames)
  }

  x
}

# apply a summary function across draws of the rvar (i.e., by each element)
# including a chain dimension in the array passed to .f
summarise_rvar_by_element_with_chains <- function(x, .f, ...) {
  draws <- draws_of(x, with_chains = TRUE)
  dim <- dim(draws)
  apply(draws, seq_along(dim)[-c(1,2)], .f, ...)
}

# base apply functions ---------------------------------------------------------

test_that("base apply family functions work", {
  x_array = array(1:24, dim = c(2,3,4))
  x = rvar(x_array)

  expect_equal(lapply(x, function(x) sum(draws_of(x))), as.list(apply(draws_of(x), 2, sum)))
  expect_equal(sapply(x, function(x) sum(draws_of(x))), apply(draws_of(x), 2, sum))
  expect_equal(vapply(x, function(x) sum(draws_of(x)), numeric(1)), apply(draws_of(x), 2, sum))
  expect_equal(apply(x, 1, function(x) sum(draws_of(x))), apply(draws_of(x), 2, sum))
  expect_equal(apply(x, 1:2, function(x) sum(draws_of(x))), apply(draws_of(x), 2:3, sum))
})

# base apply functions (factors) ---------------------------------------------------------

test_that("base apply family functions work with rvar_factor", {
  x_array = array(c("a","a","a","a", letters[1:20]), dim = c(2,3,4), dimnames = list(NULL, letters[1:3], NULL))
  x = rvar(x_array)

  expect_equal(lapply(x, function(x) sum(draws_of(x) == "a")), list(a = 2, b = 2, c = 1))
  expect_equal(sapply(x, function(x) sum(draws_of(x) == "a")), c(a = 2, b = 2, c = 1))
  expect_equal(vapply(x, function(x) sum(draws_of(x) == "a"), numeric(1)), c(a = 2, b = 2, c = 1))
  expect_equal(apply(x, 1, function(x) sum(draws_of(x) == "a")), c(a = 2, b = 2, c = 1))
  expect_equal(
    apply(x, 1:2, function(x) sum(draws_of(x) == "a")),
    array(c(2, 2, 1, rep(0, 9)), dim = c(3, 4), dimnames = list(c("a","b","c"), NULL))
  )
})


# rvar_apply --------------------------------------------------------------

test_that("rvar_apply works", {
  x_array = array(1:36, dim = c(2,2,3,3),
    dimnames = list(NULL, A = paste0("a", 1:2), B = paste0("b", 1:3), C = paste0("c", 1:3))
  )
  x = rvar(x_array)

  expect_equal(rvar_apply(x, 1, rvar_mean), rvar(apply(draws_of(x), 1:2, mean)))
  expect_equal(rvar_apply(x, 2, rvar_mean), rvar(apply(draws_of(x), c(1,3), mean)))
  expect_equal(rvar_apply(x, 3, rvar_mean), rvar(apply(draws_of(x), c(1,4), mean)))
  expect_equal(rvar_apply(x, c(1,2), rvar_mean), rvar(apply(draws_of(x), c(1,2,3), mean)))
  expect_equal(rvar_apply(x, c(1,3), rvar_mean), rvar(apply(draws_of(x), c(1,2,4), mean)))

  expect_error(rvar_apply(x, c(1,3), function(x) 0))

  expect_equal(length(rvar_apply(x, c(1,3), function(x) rvar())), 0)

  # test that if the cell values are multidimensional everything is bound back
  # together properly (though with dimnames dropped)
  x1 <- x + 1
  dimnames(x1)[3] <- list(NULL)
  names(dimnames(x1))[[3]] <- ""
  expect_equal(rvar_apply(x, c(1,2), function(x) x + 1), x1)

  # test that binding results together does broadcasting: x["a1",] has a mean
  # < 18 and ref["a2",] has mean > 18 and will be replaced with 0 with this
  ref <- x
  ref["a1",,] <- ref["a1",,] + 1
  ref["a2",,] <- 0
  expect_equal(rvar_apply(x, 1, function(x) if(mean(draws_of(x)) > 18) rvar(0) else x + 1), ref)
})


# rvar_apply with factors -------------------------------------------------

test_that("rvar_apply works with rvar_factor", {
  x_array = array(rep_len(letters[1:10], 36), dim = c(2,2,3,3),
    dimnames = list(NULL, A = paste0("a", 1:2), B = paste0("b", 1:3), C = paste0("c", 1:3))
  )
  x = rvar(x_array)

  expect_equal(rvar_apply(x, 1, identity), x)
  expect_equal(rvar_apply(x, 1, `[[`, 2), x[,2,1, drop = TRUE])
  expect_equal(rvar_apply(x, 2, `[[`, 1), x[1,,1, drop = TRUE])
  expect_equal(rvar_apply(x, c(1,2), `[[`, 3), x[,,3, drop = TRUE])
  expect_equal(rvar_apply(x, c(1,3), `[[`, 1), x[,1,, drop = TRUE])

  # if the rvar is ordered is preserved
  expect_equal(rvar_apply(rvar_ordered(x), 1, identity), rvar_ordered(x))

  # test that binding results together does broadcasting: cells with "d"
  # will be replaced with all "XXX" draws with this
  ref <- x
  levels(ref) <- c(levels(ref), "XXX")
  ref["a2","b1","c1"] <- "XXX"
  ref["a1","b1","c2"] <- "XXX"
  ref["a2","b3","c2"] <- "XXX"
  ref["a1","b3","c3"] <- "XXX"
  expect_equal(rvar_apply(x, 1:3, function(x) if ("d" %in% draws_of(x)) rvar("XXX") else drop(x)), ref)
})

test_that("rdo works", {
  x_array <- array(1:24, dim = c(4,2,3))
  x <- new_rvar(x_array)
  y_array <- array(c(2:13,12:1), dim = c(4,3,2))
  y <- new_rvar(y_array)

  xy_ref <- new_rvar(abind::abind(along = 0,
    x_array[1,,] %*% y_array[1,,],
    x_array[2,,] %*% y_array[2,,],
    x_array[3,,] %*% y_array[3,,],
    x_array[4,,] %*% y_array[4,,]
  ))
  expect_equal(rdo(x %*% y), xy_ref)
})

test_that("rfun works", {
  x_array <- array(1:24, dim = c(4,2,3))
  x <- new_rvar(x_array)
  y_array <- array(c(2:13,12:1), dim = c(4,3,2))
  y <- new_rvar(y_array)

  xy_ref <- new_rvar(abind::abind(along = 0,
    x_array[1,,] %*% y_array[1,,],
    x_array[2,,] %*% y_array[2,,],
    x_array[3,,] %*% y_array[3,,],
    x_array[4,,] %*% y_array[4,,]
  ))
  expect_equal(rfun(function(a,b) a %*% b)(x, y), xy_ref)

  expect_error(rfun(function(a,b) a %*% b, rvar_args = "c"), "All arguments .* must be names of formal arguments of `.f`")
})

test_that("rfun works on functions without rvar arguments", {
  i <- 0
  f <- function(x) {
    i <<- i + 1
    i
  }
  expect_equal(rfun(f, rvar_args = NULL, ndraws = 10)(NULL), rvar(1:10))
})

test_that("rdo allows setting the dimensions of the result", {
  expect_equal(rdo(1:10, dim = c(2,5), ndraws = 1), rvar(array(1:10, dim = c(1,2,5))))
})

test_that("rvar_rng works", {
  set.seed(1234)

  mu <- rvar_rng(rnorm, 10, mean = 1:10, sd = 1)
  sigma <- rvar_rng(rgamma, 1, shape = 1, rate = 1)
  x <- rvar_rng(rnorm, 10, mu, sigma)

  expect_equal(mean(x), 1:10, tolerance = 0.1)
  expect_equal(apply(draws_of(x), 2, sd), rep(1.7, 10), tolerance = 0.1)
})

test_that("rvar_rng recycling works with numeric and rvar arguments", {
  # a fake random number generator that we can use to ensure draws are being
  # correctly reshaped and lined up
  rfake = function(n, tens, ones) {
    rep_len(tens * 10 + ones, length.out = n)
  }

  # numeric arguments
  ref <- rvar(array(c(11, 11, 22, 22, 13, 13, 24, 24), dim = c(2,4)))
  expect_equal(rvar_rng(rfake, 4, 1:2, 1:4, ndraws = 2), ref)

  # mixed numeric and rvar arguments
  ones <- rvar(array(1:8, dim = c(2,4)))
  ref <- rvar(array(c(11, 12, 23, 24, 15, 16, 27, 28), dim = c(2,4)))
  expect_equal(rvar_rng(rfake, 4, 1:2, ones), ref)

  # rvar arguments
  x <- c(15, 26, 37, 48)
  expect_equal(rvar_rng(rfake, 1, rvar(1:4), rvar(5:8)), rvar(x))
  expect_equal(rvar_rng(rfake, 2, rvar(1:4), rvar(5:8)), rvar(array(c(x, x), dim = c(4, 2))))

  tens <- rvar(array(1:4, dim = c(2,2)))
  ones <- rvar(array(1:8, dim = c(2,4)))
  ref <- rvar(array(c(11, 22, 33, 44, 15, 26, 37, 48), dim = c(2,4)))
  expect_equal(rvar_rng(rfake, 4, tens, ones), ref)
})

test_that("rvar_rng requires rvar arguments to be 1 dimensional", {
  expect_error(rvar_rng(rnorm, 1, as_rvar(matrix(1:4, nrow = 2))), "must be single-dimensional")
})

# function for making rvars from arrays that expects last index to be
# draws (for testing so that when array structure changes tests don't have to)
rvar_from_array = function(x) {
  .dim = dim(x)
  last_dim = length(.dim)
  new_rvar(aperm(x, c(last_dim, seq_len(last_dim - 1))))
}


# [[ indexing -------------------------------------------------------------

test_that("indexing with [[ works on a vector", {
  x_array <- array(1:20, dim = c(5,4), dimnames = list(NULL, A = paste0("a", 1:4)))
  x = new_rvar(x_array)

  # [[ indexing should drop names (but not indices)
  x_array_ref = x_array
  dimnames(x_array_ref) <- NULL

  expect_equal(x[[3]], new_rvar(x_array_ref[,3, drop = FALSE]))
  expect_equal(x[["a2"]], new_rvar(x_array_ref[,2, drop = FALSE]))

  expect_error(x[[]])
  expect_error(x[[NA]], "Missing indices not allowed")
  expect_error(x[[NA_integer_]], "Missing indices not allowed")
  expect_error(x[[6]], "out of bounds")
  expect_error(x[[1,1]], "out of bounds")
  expect_error(x[[1,1,1]], "out of bounds")
  expect_error(x[[NULL]], "Cannot select zero elements")
  expect_error(x[[1:2]], "Cannot select more than one element")
  expect_error(x[[-1]], "out of bounds")

  # different behavior from base vectors
  # base vectors convert these to numeric
  expect_error(x[[TRUE]], "Logical indices not allowed")
  expect_error(x[[FALSE]], "Logical indices not allowed")
})

test_that("indexing with [[ works on a matrix", {
  x_array = array(
    1:24, dim = c(2,4,3),
    dimnames = list(NULL, A = paste0("a", 1:4), B = paste0("b", 1:3))
  )
  x = new_rvar(x_array)

  x_array_ref = x_array
  dim(x_array_ref) <- c(2,12)

  expect_equal(x[[2]], new_rvar(x_array_ref[,2, drop = TRUE]))
  expect_equal(x[[12]], new_rvar(x_array_ref[,12, drop = TRUE]))
  expect_equal(x[[2,3]], new_rvar(x_array[,2,3, drop = TRUE]))

  # invalid indexing should result in errors
  expect_error(x[[1,]], "Missing indices not allowed")
  expect_error(x[[1,1,1]], "out of bounds")
  expect_error(x[[13]], "out of bounds")

  # different from base vectors
  # don't allow name-based [[ indexing on 2+D arrays
  expect_error(x[["a2"]])

  # extending a NULL rvar should work...
  x_null = rvar()
  x_null[[1]] <- 5
  expect_equal(x_null, rvar(5))
})


# [[ assignment -----------------------------------------------------------

test_that("assignment with [[ works", {
  x_array = array(
    1:24, dim = c(2,4,3),
    dimnames = list(NULL, A = paste0("a", 1:4), B = paste0("b", 1:3))
  )
  x = new_rvar(x_array)

  expect_equal(
    {x2 <- x; x2[[2]] <- 1; x2},
    new_rvar({xr <- x_array; xr[,2,1] <- 1; xr})
  )
  expect_equal(
    {x2 <- x; x2[[12]] <- 1; x2},
    new_rvar({xr <- x_array; xr[,4,3] <- 1; xr})
  )
  expect_equal(
    {x2 <- x; x2[[12]] <- new_rvar(c(1,2)); x2},
    new_rvar({xr <- x_array; xr[,4,3] <- c(1,2); xr})
  )
  expect_equal(
    {x2 <- x; x2[["a2","b3"]] <- new_rvar(c(1,2)); x2},
    new_rvar({xr <- x_array; xr[,2,3] <- c(1,2); xr})
  )

  expect_error({x2 <- x; x2[[1,1,1]] <- 1}, "out of bounds")

  # constant should have ndraws increased to value when assigned to
  x = new_rvar(array(1:2, dim = c(1,2)))
  expect_equal(
    {x[[1]] <- new_rvar(array(1:2, dim = c(2,1))); x},
    new_rvar(array(c(1,2,2,2), dim = c(2,2)))
  )

  expect_error({x2 <- x; x2[[-1]] <- 1}, "out of bounds")
  expect_error({x2 <- rvar(1:10); x2[[2]] <- c(4,5,6)})
})


# [ indexing --------------------------------------------------------------

test_that("indexing with [ works on a vector", {
  x_array = array(1:20, dim = c(4,5), dimnames = list(A = paste0("a", 1:4), NULL))
  x = rvar_from_array(x_array)

  expect_equal(x[], x)

  expect_equal(x[3], rvar_from_array(x_array[3,, drop = FALSE]))
  expect_equal(x["a2"], rvar_from_array(x_array["a2",, drop = FALSE]))
  expect_equal(x[c(1,3)], rvar_from_array(x_array[c(1,3),, drop = FALSE]))
  expect_equal(x[c("a2","a4")], rvar_from_array(x_array[c("a2","a4"),, drop = FALSE]))

  expect_equal(x[c(-1,-3)], rvar_from_array(x_array[c(-1,-3),, drop = FALSE]))

  expect_equal(x[TRUE], rvar_from_array(x_array[TRUE,, drop = FALSE]))
  expect_equal(x[c(TRUE,FALSE)], rvar_from_array(x_array[c(TRUE,FALSE),, drop = FALSE]))
  expect_equal(x[c(TRUE,FALSE,TRUE)], rvar_from_array(x_array[c(TRUE,FALSE,TRUE),, drop = FALSE]))
  expect_equal(x[c(TRUE,FALSE,FALSE,TRUE)], rvar_from_array(x_array[c(TRUE,FALSE,FALSE,TRUE),, drop = FALSE]))

  # dropping should preserve names (hence the drop = FALSE on x_array for this test)
  expect_equal(x["a1", drop = TRUE], rvar_from_array(x_array["a1",, drop = FALSE]))
  expect_equal(x[1:2, drop = TRUE], rvar_from_array(x_array[1:2,, drop = FALSE]))

  # indexing beyond the end of the array should result in NAs, to mimic normal vector indexing
  expect_equal(x[c(4,5)], rvar_from_array(x_array[c(4,NA_integer_),, drop = FALSE]))
  expect_equal(x[c(8,9)], rvar_from_array(x_array[c(NA_integer_,NA_integer_),, drop = FALSE]))

  expect_equal(x[NA], rvar_from_array(x_array[NA,, drop = FALSE]))
  expect_equal(x[NA_integer_], rvar_from_array(x_array[NA_integer_,, drop = FALSE]))
  expect_equal(x[rep(NA_integer_,7)], rvar_from_array(x_array[rep(NA_integer_,7),, drop = FALSE]))

  expect_equal(x[NULL], new_rvar(array(numeric(), dim = c(5, 0))))

  expect_error(x[1,1])

  # extending a NULL rvar should work...
  x_null = rvar()
  x_null[1] <- 5
  expect_equal(x_null, rvar(5))
})

test_that("indexing with [ works on an array", {
  x_array = array(
    1:24, dim = c(4,3,2),
    dimnames = list(A = paste0("a", 1:4), B = paste0("b", 1:3))
  )
  x = rvar_from_array(x_array)

  expect_equal(x[], x)

  expect_equal(x[2], rvar_from_array(array(x_array[2,1,], dim = c(1,2))))
  expect_equal(x[2,], rvar_from_array(x_array[2,,, drop = FALSE]))
  expect_equal(x["a2",], rvar_from_array(x_array["a2",,, drop = FALSE]))
  expect_equal(x[c(1,2)], rvar_from_array(array(x_array[c(1,2),1,], dim = c(2,2))))
  expect_equal(x[c(1,2),], rvar_from_array(x_array[c(1,2),,, drop = FALSE]))
  expect_equal(x[,c(1,3)], rvar_from_array(x_array[,c(1,3),, drop = FALSE]))
  expect_equal(x[,c("b2","b3")], rvar_from_array(x_array[,c("b2","b3"),, drop = FALSE]))

  expect_equal(x[,c(-1,-3)], rvar_from_array(x_array[,c(-1,-3),, drop = FALSE]))

  expect_equal(x[TRUE], rvar_from_array(array(x_array, dim = c(12,2))))
  expect_equal(x[c(TRUE,FALSE)], rvar_from_array(array(x_array, dim = c(12,2))[c(TRUE,FALSE),]))
  expect_equal(x[c(TRUE,FALSE,TRUE)], rvar_from_array(array(x_array, dim = c(12,2))[c(TRUE,FALSE,TRUE),]))
  expect_equal(x[c(TRUE,FALSE,FALSE,TRUE)], rvar_from_array(array(x_array, dim = c(12,2))[c(TRUE,FALSE,FALSE,TRUE),]))

  expect_equal(x[TRUE,], rvar_from_array(x_array[TRUE,,, drop = FALSE]))
  expect_equal(x[c(TRUE,FALSE),], rvar_from_array(x_array[c(TRUE,FALSE),,, drop = FALSE]))
  expect_equal(x[c(TRUE,FALSE,TRUE),], rvar_from_array(x_array[c(TRUE,FALSE,TRUE),,, drop = FALSE]))
  expect_equal(x[c(TRUE,FALSE,FALSE,TRUE),], rvar_from_array(x_array[c(TRUE,FALSE,FALSE,TRUE),,, drop = FALSE]))

  # dropping works
  expect_equal(x["a1",, drop = TRUE], rvar_from_array(x_array["a1",,, drop = TRUE]))
  expect_equal(x[1:2,, drop = TRUE], rvar_from_array(x_array[1:2,,, drop = TRUE]))
  expect_equal(x[1,2, drop = TRUE], rvar_from_array(array(x_array[1,2,], dim = c(1,2))))
  expect_equal(x[1,1:2, drop = TRUE], rvar_from_array(x_array[1,1:2,, drop = TRUE]))

  # indexing beyond the end of the array should result in NAs, to mimic normal vector indexing
  expect_equal(x[c(4,25)], rvar_from_array(array(x_array[c(4,NA_integer_),1,], dim = c(2,2))))
  expect_equal(x[c(4,5),], rvar_from_array(x_array[c(4,NA_integer_),,, drop = FALSE]))
  expect_equal(x[c(8,9),], rvar_from_array(x_array[c(NA_integer_,NA_integer_),,, drop = FALSE]))

  expect_equal(x[NA], rvar_from_array(array(x_array[NA,,], dim = c(12,2))))
  expect_equal(x[NA,], rvar_from_array(x_array[NA,,, drop = FALSE]))
  expect_equal(x[NA_integer_], rvar_from_array(array(c(NA_integer_,NA_integer_), dim = c(1,2))))
  expect_equal(x[NA_integer_,], rvar_from_array(x_array[NA_integer_,,, drop = FALSE]))
  expect_equal(x[rep(NA_integer_,7)], rvar_from_array(array(rep(NA_integer_,14), dim = c(7,2))))

  expect_equal(x[NULL], new_rvar(array(numeric(), dim = c(2, 0))))

  # logical index the length of the array works
  flat_index <- c(
    TRUE,FALSE,FALSE,
    FALSE,TRUE,FALSE,
    TRUE,TRUE, FALSE,
    FALSE,TRUE,TRUE)
  x_array_flat <- x_array
  dim(x_array_flat) <- c(12,2)
  expect_equal(x[flat_index], rvar_from_array(x_array_flat[flat_index,]))

  expect_error(x[1,1,1])

  # matrix indexing with an array
  x_array <- array(1:24, dim = c(2,2,3,2))
  x <- rvar_from_array(x_array)
  expect_equal(x[rbind(c(1,2,3),c(2,2,3),c(2,1,1))], x[c(11,12,2)])

  # indexing while leaving remaining indices to be filled in automatically
  expect_equal(x[1,], x[1,,])
})


# [ assignment ------------------------------------------------------------

test_that("assignment with [ works", {
  x_array = array(
    1:24, dim = c(2,4,3),
    dimnames = list(NULL, A = paste0("a", 1:4), B = paste0("b", 1:3))
  )
  x = new_rvar(x_array)

  expect_equal(
    {x2 <- x; x2[2,1] <- 1; x2},
    new_rvar({xr <- x_array; xr[,2,1] <- 1; xr})
  )
  expect_equal(
    {x2 <- x; x2[2,] <- 1; x2},
    new_rvar({xr <- x_array; xr[,2,] <- 1; xr})
  )
  expect_equal(
    {x2 <- x; x2[,2] <- new_rvar(c(1,2)); x2},
    new_rvar({xr <- x_array; xr[,,2] <- c(1,2); xr})
  )
  expect_equal(
    {x2 <- x; x2["a2","b3"] <- new_rvar(c(1,2)); x2},
    new_rvar({xr <- x_array; xr[,2,3] <- c(1,2); xr})
  )

  # constant should have ndraws increased to value when assigned to
  x2 = new_rvar(array(1:2, dim = c(1,2)))
  expect_equal(
    {x2[1] <- new_rvar(array(1:2, dim = c(2,1))); x2},
    new_rvar(array(c(1,2,2,2), dim = c(2,2)))
  )

  # logical index the length of the array works
  flat_index <- c(
    TRUE,FALSE,FALSE,
    FALSE,TRUE,FALSE,
    TRUE,TRUE, FALSE,
    FALSE,TRUE,TRUE)
  x_array_flat <- x_array
  dim(x_array_flat) <- c(2,12)
  x_array_flat[,flat_index] <- rep(1:6, each = 2)
  dim(x_array_flat) <- c(2,4,3)
  dimnames(x_array_flat) <- list(NULL, a = 1:4, b = 1:3)
  expect_equal(
    {x2 <- x; dimnames(x2) <- list(a = 1:4, b = 1:3); x2[flat_index] <- 1:6; x2},
    new_rvar(x_array_flat)
  )

  # matrix indexing assignment and unidimensional index assignment with an array works
  x_array <- array(
    1:24, dim = c(2,2,3,2),
    dimnames = list(A = paste0("a", 1:2), B = paste0("b", 1:2), C = paste0("c", 1:3))
  )

  x_ref <- rvar_from_array(x_array)
  x_ref[1,2,3] <- rvar(1:2)
  x_ref[2,2,3] <- rvar(3:4)
  x_ref[2,1,1] <- rvar(5:6)

  x <- rvar_from_array(x_array)
  x[rbind(c(1,2,3),c(2,2,3),c(2,1,1))] <- rvar(matrix(1:6, nrow = 2))
  expect_equal(x, x_ref)

  x <- rvar_from_array(x_array)
  x[c(11,12,2)] <- rvar(matrix(1:6, nrow = 2))
  expect_equal(x, x_ref)
})

test_that("assigning NULL dim to rvar works", {
  x <- rvar(array(1:20, dim = c(2,2,5)))
  dim(x) <- NULL
  expect_equal(x, rvar(array(1:20, dim = c(2,10))))
})


# names -------------------------------------------------------------------

test_that("unname() works", {
  x_array <- array(1:24, dim = c(2,3,4), dimnames = list(NULL, A = paste0("a", 1:3), B = paste0("b", 1:4)))
  x <- rvar(x_array)

  expect_equal(unname(x), rvar(unname(x_array)))
})


# drop --------------------------------------------------------------------

test_that("drop() works", {
  .dim <- c(2,3,4)
  .dimnames <- list(NULL, A = paste0("a", 1:3), B = paste0("b", 1:4))
  x_array <- array(1:24, dim = .dim, dimnames = .dimnames)
  x <- rvar(x_array)

  x_1 <- rvar(array(x_array, dim = c(.dim, 1), dimnames = c(.dimnames, list(NULL))))
  expect_equal(drop(x_1), x)

  x_vector <- rvar(array(x_array, dim = c(2,12)))
  names(x_vector) <- paste0("a", 1:12)
  expect_equal(drop(x_vector), x_vector)
})

test_that("adrop() works", {
  .dim <- c(2,3,4)
  .dimnames <- list(NULL, A = paste0("a", 1:3), B = paste0("b", 1:4))
  x_array <- array(1:24, dim = .dim, dimnames = .dimnames)
  x <- rvar(x_array)

  x_1 <- rvar(array(x_array, dim = c(.dim, 1), dimnames = c(.dimnames, list(NULL))))
  expect_equal(abind::adrop(x_1, drop = NULL), x_1)
  expect_equal(abind::adrop(x_1, drop = 3), x)
  expect_equal(abind::adrop(x_1, drop = c(FALSE, FALSE, TRUE)), x)

  x_vector <- rvar(array(x_array, dim = c(2,12)))
  x_vector_unnamed <- x_vector
  names(x_vector) <- paste0("a", 1:12)
  expect_equal(abind::adrop(x_vector, drop = FALSE), x_vector)
  expect_equal(abind::adrop(x_vector, drop = FALSE, named.vector = FALSE), x_vector_unnamed)
})

# basic operators ---------------------------------------------------------

test_that("math operators works", {
  x_array = array(1:24, dim = c(4,2,3), dimnames = list(NULL,letters[1:2],letters[3:5]))
  x = new_rvar(x_array)
  y_array = array(c(2:13,12:1), dim = c(4,2,3))
  y = new_rvar(y_array)

  expect_equal(log(x), new_rvar(log(x_array)))

  expect_equal(-x, new_rvar(-x_array))

  expect_equal(x + 2, new_rvar(x_array + 2))
  expect_equal(2 + x, new_rvar(x_array + 2))
  expect_equal(x + y, new_rvar(x_array + y_array))

  expect_equal(x - 2, new_rvar(x_array - 2))
  expect_equal(2 - x, new_rvar(2 - x_array))
  expect_equal(x - y, new_rvar(x_array - y_array))

  expect_equal(x * 2, new_rvar(x_array * 2))
  expect_equal(2 * x, new_rvar(x_array * 2))
  expect_equal(x * y, new_rvar(x_array * y_array))

  expect_equal(x / 2, new_rvar(x_array / 2))
  expect_equal(2 / x, new_rvar(2 / x_array))
  expect_equal(x / y, new_rvar(x_array / y_array))

  expect_equal(x ^ 2, new_rvar((x_array) ^ 2))
  expect_equal(2 ^ x, new_rvar(2 ^ (x_array)))
  expect_equal(x ^ y, new_rvar(x_array ^ y_array))

  # ensure broadcasting of constants retains shape
  z2 <- new_rvar(array(1, dim = c(1,1)))
  z4 <- new_rvar(array(2, dim = c(1,1,1,1)))
  expect_equal(z2 + z4, new_rvar(array(3, dim = c(1,1,1,1))))
})

test_that("math operators work on rvar_factors", {
  x_array = array(letters[1:24], dim = c(4,2,3), dimnames = list(NULL,letters[1:2],letters[3:5]))
  x = rvar_factor(x_array)
  x_ordered = rvar_ordered(x_array)

  expect_error(log(x), "Cannot apply `log` function to rvar_factor")
  expect_error(x + 1, "Cannot apply `\\+` operator to rvar_factor")
  expect_error(1 + x, "Cannot apply `\\+` operator to rvar_factor")

  expect_error(log(x_ordered), "Cannot apply `log` function to rvar_factor")
  expect_error(x_ordered + 1, "Cannot apply `\\+` operator to rvar_ordered")
  expect_error(1 + x_ordered, "Cannot apply `\\+` operator to rvar_ordered")
})

test_that("logical operators work", {
  x_array = c(TRUE,TRUE,FALSE,FALSE)
  y_array = c(TRUE,FALSE,TRUE,FALSE)
  x = as_rvar(x_array)
  y = as_rvar(y_array)

  expect_equal(x | y_array, as_rvar(x_array | y_array))
  expect_equal(y_array | x, as_rvar(x_array | y_array))
  expect_equal(x | y, as_rvar(x_array | y_array))

  expect_equal(x & y_array, as_rvar(x_array & y_array))
  expect_equal(y_array & x, as_rvar(x_array & y_array))
  expect_equal(x & y, as_rvar(x_array & y_array))
})

test_that("comparison operators work", {
  x_array = array(1:24, dim = c(4,2,3))
  x = new_rvar(x_array)
  y_array = array(c(2:13,12:1), dim = c(4,2,3))
  y = new_rvar(y_array)

  expect_equal(x < 5, new_rvar(x_array < 5))
  expect_equal(5 < x, new_rvar(5 < x_array))
  expect_equal(x < y, new_rvar(x_array < y_array))

  expect_equal(x <= 5, new_rvar(x_array <= 5))
  expect_equal(5 <= x, new_rvar(5 <= x_array))
  expect_equal(x <= y, new_rvar(x_array <= y_array))

  expect_equal(x > 5, new_rvar(x_array > 5))
  expect_equal(5 > x, new_rvar(5 > x_array))
  expect_equal(x > y, new_rvar(x_array > y_array))

  expect_equal(x >= 5, new_rvar(x_array >= 5))
  expect_equal(5 >= x, new_rvar(5 >= x_array))
  expect_equal(x >= y, new_rvar(x_array >= y_array))

  expect_equal(x == 5, new_rvar(x_array == 5))
  expect_equal(5 == x, new_rvar(5 == x_array))
  expect_equal(x == y, new_rvar(x_array == y_array))

  expect_equal(x != 5, new_rvar(x_array != 5))
  expect_equal(5 != x, new_rvar(5 != x_array))
  expect_equal(x != y, new_rvar(x_array != y_array))
})

test_that("comparison operators work on rvar_factors", {
  x_array = array(1:24, dim = c(4,2,3))
  x_array_letters = array(letters[1:24], dim = c(4,2,3))
  x = rvar_factor(x_array_letters, levels = letters)
  y_array = array(c(2:13,12:1), dim = c(4,2,3))
  y_array_letters = array(letters[c(2:13,12:1)], dim = c(4,2,3))
  y = rvar_factor(y_array_letters, levels = letters)

  expect_error(x < "e", "Cannot apply `<` operator to rvar_factor")
  expect_error(x > "e", "Cannot apply `>` operator to rvar_factor")
  expect_error(x <= "e", "Cannot apply `<=` operator to rvar_factor")
  expect_error(x >= "e", "Cannot apply `>=` operator to rvar_factor")

  expect_equal(x == "e", new_rvar(x_array == 5))
  expect_equal("e" == x, new_rvar(5 == x_array))
  expect_equal(x == y, new_rvar(x_array == y_array))
  expect_equal(x == "XX", new_rvar(x_array == 0))
  expect_equal("XX" == x, new_rvar(0 == x_array))

  expect_equal(x != "e", new_rvar(x_array != 5))
  expect_equal("e" != x, new_rvar(5 != x_array))
  expect_equal(x != y, new_rvar(x_array != y_array))
  expect_equal(x != "XX", new_rvar(x_array != 0))
  expect_equal("XX" != x, new_rvar(0 != x_array))
})

test_that("comparison operators work on rvar_ordereds", {
  x_array = array(1:24, dim = c(4,2,3))
  x_array_letters = array(letters[1:24], dim = c(4,2,3))
  x = rvar_ordered(x_array_letters, levels = letters)
  y_array = array(c(2:13,12:1), dim = c(4,2,3))
  y_array_letters = array(letters[c(2:13,12:1)], dim = c(4,2,3))
  y = rvar_ordered(y_array_letters, levels = letters)

  expect_equal(x < "e", new_rvar(x_array < 5))
  expect_equal("e" < x, new_rvar(5 < x_array))
  expect_equal(x < y, new_rvar(x_array < y_array))
  expect_equal(x < "XX", new_rvar(x_array < NA))
  expect_equal("XX" < x, new_rvar(NA < x_array))

  expect_equal(x <= "e", new_rvar(x_array <= 5))
  expect_equal("e" <= x, new_rvar(5 <= x_array))
  expect_equal(x <= y, new_rvar(x_array <= y_array))
  expect_equal(x <= "XX", new_rvar(x_array <= NA))
  expect_equal("XX" <= x, new_rvar(NA <= x_array))

  expect_equal(x > "e", new_rvar(x_array > 5))
  expect_equal("e" > x, new_rvar(5 > x_array))
  expect_equal(x > y, new_rvar(x_array > y_array))
  expect_equal(x > "XX", new_rvar(x_array > NA))
  expect_equal("XX" > x, new_rvar(NA > x_array))

  expect_equal(x >= "e", new_rvar(x_array >= 5))
  expect_equal("e" >= x, new_rvar(5 >= x_array))
  expect_equal(x >= y, new_rvar(x_array >= y_array))
  expect_equal(x >= "XX", new_rvar(x_array >= NA))
  expect_equal("XX" >= x, new_rvar(NA >= x_array))

  expect_equal(x == "e", new_rvar(x_array == 5))
  expect_equal("e" == x, new_rvar(5 == x_array))
  expect_equal(x == y, new_rvar(x_array == y_array))
  expect_equal(x == "XX", new_rvar(x_array == 0))
  expect_equal("XX" == x, new_rvar(0 == x_array))

  expect_equal(x != "e", new_rvar(x_array != 5))
  expect_equal("e" != x, new_rvar(5 != x_array))
  expect_equal(x != y, new_rvar(x_array != y_array))
  expect_equal(x != "XX", new_rvar(x_array != 0))
  expect_equal("XX" != x, new_rvar(0 != x_array))
})

test_that("functions in the Math generic with extra arguments work", {
  expect_equal(round(rvar(11), -1), rvar(10))
  expect_equal(signif(rvar(11), 1), rvar(10))
  expect_equal(log(rvar(c(2,4,8)), base = 2), rvar(1:3))
})

test_that("cumulative functions work", {
  x_array = array(1:12, dim = c(2,2,3))
  x = new_rvar(x_array)

  cumsum_ref = new_rvar(rbind(
    cumsum(draws_of(x)[1,,]),
    cumsum(draws_of(x)[2,,])
  ))
  expect_equal(cumsum(x), cumsum_ref)

  cumprod_ref = new_rvar(rbind(
    cumprod(draws_of(x)[1,,]),
    cumprod(draws_of(x)[2,,])
  ))
  expect_equal(cumprod(x), cumprod_ref)

  cummax_ref = new_rvar(rbind(
    cummax(draws_of(x)[1,,]),
    cummax(draws_of(x)[2,,])
  ))
  expect_equal(cummax(x), cummax_ref)

  cummin_ref = new_rvar(rbind(
    cummin(draws_of(x)[1,,]),
    cummin(draws_of(x)[2,,])
  ))
  expect_equal(cummin(x), cummin_ref)
})

# matrix stuff ------------------------------------------------------------

test_that("matrix multiplication works", {
  x_array = array(1:24, dim = c(4,2,3))
  x = new_rvar(x_array)
  y_array = array(c(2:13,12:1), dim = c(4,3,2))
  y = new_rvar(y_array)

  xy_ref = new_rvar(abind::abind(along = 0,
    x_array[1,,] %*% y_array[1,,],
    x_array[2,,] %*% y_array[2,,],
    x_array[3,,] %*% y_array[3,,],
    x_array[4,,] %*% y_array[4,,]
  ))
  expect_equal(x %**% y, xy_ref)


  x_array = array(1:6, dim = c(2,3))
  x = new_rvar(x_array)
  y_array = array(7:12, dim = c(2,3))
  y = new_rvar(y_array)

  xy_ref = new_rvar(abind::abind(along = 0,
    x_array[1,] %*% y_array[1,],
    x_array[2,] %*% y_array[2,]
  ))
  expect_equal(x %**% y, xy_ref)

  # automatic promotion to row/col vector of numeric vectors
  x_meany_ref = new_rvar(abind::abind(along = 0,
    x_array[1,] %*% colMeans(y_array),
    x_array[2,] %*% colMeans(y_array)
  ))
  expect_equal(x %**% colMeans(y_array), x_meany_ref)

  meanx_y_ref = new_rvar(abind::abind(along = 0,
    colMeans(x_array) %*% y_array[1,],
    colMeans(x_array) %*% y_array[2,]
  ))
  expect_equal(colMeans(x_array) %**% y, meanx_y_ref)

  # dimension name preservation
  m1 <- as_rvar(diag(1:3))
  dimnames(m1) <- list(a = paste0("a", 1:3), b = paste0("b", 1:3))
  m2 <- as_rvar(diag(1:3)[,1:2])
  dimnames(m2) <- list(c = paste0("c", 1:3), d = paste0("d", 1:2))
  expect_equal(dimnames(m1 %**% m2), list(a = paste0("a", 1:3), d = paste0("d", 1:2)))


  # errors
  x_array = array(1:24, dim = c(4,1,2,3))
  x = new_rvar(x_array)

  expect_error(x %**% 1, "not a vector or matrix")
  expect_error(1 %**% x, "not a vector or matrix")
  expect_error(rvar_factor("a") %**% x, "Cannot apply `%\\*\\*%` operator to rvar_factor")
  expect_error(x %**% rvar_factor("a"), "Cannot apply `%\\*\\*%` operator to rvar_factor")

})

test_that("diag works", {
  Sigma <- as_draws_rvars(example_draws("multi_normal"))$Sigma

  sigma_diag <- c(Sigma[1,1], Sigma[2,2], Sigma[3,3])
  expect_equal(diag(Sigma), sigma_diag)

  Sigma_ref <- Sigma
  Sigma_ref[1,1] <- 2
  Sigma_ref[2,2] <- 3
  Sigma_ref[3,3] <- 4

  Sigma_test <- Sigma
  diag(Sigma_test) <- 2:4
  expect_equal(Sigma_test, Sigma_ref)

  Sigma_ref <- Sigma
  Sigma_ref[1,2:3] <- 0
  Sigma_ref[2,c(1,3)] <- 0
  Sigma_ref[3,1:2] <- 0
  expect_equal(diag(sigma_diag), Sigma_ref)
  expect_equal(diag(sigma_diag, nrow = 2), Sigma_ref[1:2,1:2])
  expect_equal(diag(sigma_diag, nrow = 2, ncol = 2), Sigma_ref[1:2,1:2])
  expect_equal(diag(sigma_diag, nrow = 2, ncol = 3), Sigma_ref[1:2,])
  expect_equal(diag(sigma_diag, nrow = 1, ncol = 3), Sigma_ref[1,])
  expect_equal(diag(sigma_diag, nrow = 0, ncol = 3), rvar(array(numeric(), dim = c(400,0)), nchains = 4))
  expect_equal(diag(sigma_diag, ncol = 2), Sigma_ref[,1:2])
  expect_equal(diag(sigma_diag, ncol = 1), Sigma_ref[,1])
  expect_equal(diag(sigma_diag, ncol = 0), rvar(array(numeric(), dim = c(400,0)), nchains = 4))
  expect_equal(diag(sigma_diag, nrow = 4), cbind(rbind(Sigma_ref, 0), c(rvar(0), 0, 0, Sigma[1,1])))

  # ensure our override of base diag does not break its weird corner cases
  expect_equal(diag(matrix(1:9, nrow = 3)), c(1, 5, 9))
  expect_equal(diag(1:2), matrix(c(1, 0, 0, 2), nrow = 2))
  expect_equal(diag(3L), matrix(c(1, 0, 0, 0, 1, 0, 0, 0, 1), nrow = 3))
  expect_equal(diag(nrow = 3), matrix(c(1, 0, 0, 0, 1, 0, 0, 0, 1), nrow = 3))
})

test_that("Cholesky decomposition works", {
  Sigma <- as_draws_rvars(example_draws("multi_normal"))$Sigma

  # adding dimensions because we should expect these to be dropped
  dimnames(Sigma) <- list(
    a = paste0("a", 1:3),
    b = paste0("b", 1:3)
  )

  expect_equal(chol(Sigma), rdo(chol(Sigma)))

  expect_error(chol(rvar(1)), "must be a random matrix")
})

# array transpose and permutation -----------------------------------------

test_that("vector transpose works", {
  x_array = array(1:6, dim = c(2,3), dimnames = list(NULL, c("a","b","c")))
  x = new_rvar(x_array)
  x_array_t = array(1:6, dim = c(2,1,3), dimnames = list(NULL, NULL, c("a","b","c")))
  x_t = new_rvar(x_array_t)
  x_array_t_t = array(1:6, dim = c(2,3,1), dimnames = list(NULL, c("a","b","c"), NULL))
  x_t_t = new_rvar(x_array_t_t)

  # ensure it works with dimnames...
  expect_equal(t(x), x_t)
  expect_equal(t(t(x)), x_t_t)

  # ... and without dimnames
  dimnames(x) = NULL
  dimnames(x_t) = NULL
  dimnames(x_t_t) = NULL
  expect_equal(t(x), x_t)
  expect_equal(t(t(x)), x_t_t)
})

test_that("vector transpose works on rvar_factor", {
  x_array = array(letters[1:6], dim = c(2,3), dimnames = list(NULL, c("a","b","c")))
  x = rvar_factor(x_array)
  x_array_t = array(letters[1:6], dim = c(2,1,3), dimnames = list(NULL, NULL, c("a","b","c")))
  x_t = rvar_factor(x_array_t)
  x_array_t_t = array(letters[1:6], dim = c(2,3,1), dimnames = list(NULL, c("a","b","c"), NULL))
  x_t_t = rvar_factor(x_array_t_t)

  # ensure it works with dimnames...
  expect_equal(t(x), x_t)
  expect_equal(t(t(x)), x_t_t)
})

test_that("vector transpose works on rvar_ordered", {
  x_array = array(letters[1:6], dim = c(2,3), dimnames = list(NULL, c("a","b","c")))
  x = rvar_ordered(x_array)
  x_array_t = array(letters[1:6], dim = c(2,1,3), dimnames = list(NULL, NULL, c("a","b","c")))
  x_t = rvar_ordered(x_array_t)
  x_array_t_t = array(letters[1:6], dim = c(2,3,1), dimnames = list(NULL, c("a","b","c"), NULL))
  x_t_t = rvar_ordered(x_array_t_t)

  # ensure it works with dimnames...
  expect_equal(t(x), x_t)
  expect_equal(t(t(x)), x_t_t)
})

test_that("matrix transpose works", {
  x_array = array(1:24, dim = c(4,3,2))
  x = new_rvar(x_array)
  x_t = new_rvar(aperm(x_array, c(1,3,2)))

  expect_error(t(rvar()))
  expect_equal(t(x), x_t)
  expect_equal(t(t(x)), x)
  expect_equal(t(new_rvar(array(1:10, c(1,10)))), new_rvar(array(1:10, c(1,1,10))))
  expect_equal(t(new_rvar(array(1:10, c(2,1,5)))), new_rvar(array(1:10, c(2,5,1))))
  expect_equal(t(new_rvar(array(1:10, c(2,1,5)))), new_rvar(array(1:10, c(2,5,1))))
  expect_equal(t(new_rvar(array(1:10, c(2,5)))), new_rvar(array(1:10, c(2,1,5))))
})

test_that("matrix transpose works on rvar_factor", {
  x_array = array(letters[1:24], dim = c(4,3,2))
  x = rvar_factor(x_array)
  x_t = rvar_factor(aperm(x_array, c(1,3,2)))

  expect_equal(t(x), x_t)
  expect_equal(t(t(x)), x)
})

test_that("matrix transpose works on rvar_ordered", {
  x_array = array(letters[1:24], dim = c(4,3,2))
  x = rvar_ordered(x_array)
  x_t = rvar_ordered(aperm(x_array, c(1,3,2)))

  expect_equal(t(x), x_t)
  expect_equal(t(t(x)), x)
})

test_that("array permutation works", {
  x_array = array(
    1:24, dim = c(2,2,3,2),
    dimnames = list(NULL, A = paste0("a", 1:2), B = paste0("b", 1:3), C = paste0("c", 1:2))
  )
  x = new_rvar(x_array)
  x_perm = new_rvar(aperm(x_array, c(1,2,4,3)))

  expect_equal(aperm(x, c(1,3,2)), x_perm)
})

test_that("array permutation works on rvar_factor", {
  x_array = array(
    letters[1:24], dim = c(2,2,3,2),
    dimnames = list(NULL, A = paste0("a", 1:2), B = paste0("b", 1:3), C = paste0("c", 1:2))
  )
  x = rvar_factor(x_array)
  x_perm = rvar_factor(aperm(x_array, c(1,2,4,3)))

  expect_equal(aperm(x, c(1,3,2)), x_perm)
})

test_that("array permutation works on rvar_ordered", {
  x_array = array(
    letters[1:24], dim = c(2,2,3,2),
    dimnames = list(NULL, A = paste0("a", 1:2), B = paste0("b", 1:3), C = paste0("c", 1:2))
  )
  x = rvar_ordered(x_array)
  x_perm = rvar_ordered(aperm(x_array, c(1,2,4,3)))

  expect_equal(aperm(x, c(1,3,2)), x_perm)
})

# creating rvars ----------------------------------------------------------

test_that("rvar creation with custom dim works", {
  x_matrix <- array(1:24, dim = c(2,12))
  x_array <- array(1:24, dim = c(2,3,4))

  expect_equal(rvar(x_matrix, dim = c(3,4)), rvar(x_array))
})

test_that("rvar can be created with specified number of chains", {
  x_array <- array(1:20, dim = c(4,5))

  expect_error(rvar(x_array, nchains = 0))
  expect_equal(rvar(x_array, nchains = 1), rvar(x_array))
  expect_equal(nchains(rvar(x_array, nchains = 2)), 2)
  expect_error(rvar(x_array, nchains = 3), "Number of chains does not divide the number of draws")
})

test_that("rvar constructor using with_chains works", {
  # multidimensional rvar with chains
  x_array_nochains <- array(1:24, dim = c(6,2,2), dimnames = list(
    NULL, A = c("a1", "a2"), B = c("b1", "b2")
  ))
  x_array_chains <- array(1:24, dim = c(3,2,2,2), dimnames = list(
    NULL, NULL, A = c("a1", "a2"), B = c("b1", "b2")
  ))
  x_nochains <- rvar(x_array_nochains, nchains = 2)
  x_chains <- rvar(x_array_chains, with_chains = TRUE)
  expect_equal(x_chains, x_nochains)

  # scalar rvar with chains
  x2_array_nochains <- 1:24
  x2_array_chains <- array(1:24, dim = c(6,4))
  x2_nochains <- rvar(x2_array_nochains, nchains = 4)
  x2_chains <- rvar(x2_array_chains, with_chains = TRUE)
  expect_equal(x2_chains, x2_nochains)

  # NULL rvar
  expect_equal(rvar(with_chains = TRUE), rvar())

  # can't use with_chains when no chain dimension information provided
  expect_error(rvar(1, with_chains = TRUE))
})

test_that("NULL rvar creates a 0-length numeric rvar with 1 draw", {
  expect_equal(rvar(), rvar(numeric()))
  expect_equal(rvar(NULL), rvar(numeric()))
  expect_equal(as_rvar(NULL), rvar(numeric()))
  expect_equal(draws_of(rvar()), array(numeric(), dim = c(1, 0), dimnames = list(1, NULL)))
})

test_that("creating an rvar from an existing rvar works", {
  x_array <- array(1:20, dim = c(4,5))
  x <- rvar(x_array)
  x_4 <- rvar(x_array, nchains = 4)

  expect_equal(rvar(x), x)
  expect_equal(rvar(x_4), x_4)
  expect_equal(rvar(x, nchains = 4), x_4)
  expect_equal(rvar(x_4, nchains = 1), x)
})

# draws_of ----------------------------------------------------------------

test_that("draws_of using with_chains works", {
  # retrieving a multidimensional rvar with draws_of using with_chains
  x_array_nochains <- array(1:24, dim = c(6,2,2), dimnames = list(
    NULL, A = c("a1", "a2"), B = c("b1", "b2")
  ))
  x_array_chains <- array(1:24, dim = c(3,2,2,2), dimnames = list(
    NULL, NULL, A = c("a1", "a2"), B = c("b1", "b2")
  ))
  x <- rvar(x_array_nochains, nchains = 2)
  expect_equal(draws_of(x, with_chains = TRUE), x_array_chains)

  # setting a multidimensional rvar with draws_of using with_chains
  x2_array_nochains <- x_array_nochains + 2
  x2_array_chains <- array(1:24 + 2, dim = c(2,3,2,2), dimnames = list(
    NULL, NULL, A = c("a1", "a2"), B = c("b1", "b2")
  ))
  x2 <- x
  draws_of(x2, with_chains = TRUE) <- x2_array_chains
  expect_equal(x2, rvar(x2_array_nochains, nchains = 3))

  # retrieving a scalar rvar with draws_of using with_chains
  x2_array_nochains <- 1:24
  x2_array_chains <- array(1:24, dim = c(6,4,1), dimnames = list(NULL))
  x2 <- rvar(x2_array_nochains, nchains = 4)
  expect_equal(draws_of(x2, with_chains = TRUE), x2_array_chains)

  # setting a scalar rvar with draws_of using with_chains
  x3_array_nochains <- 1:24 + 2
  x3_array_chains <- array(1:24 + 2, dim = c(12,2), dimnames = list(NULL))
  x3 <- x2
  draws_of(x3, with_chains = TRUE) <- x3_array_chains
  expect_equal(x3, rvar(x3_array_nochains, nchains = 2))

  # NULL rvar
  expect_equal(draws_of(rvar(), with_chains = TRUE), array(numeric(), dim = c(1,1,0), dimnames = list(NULL)))

  x_null <- x
  draws_of(x_null, with_chains = TRUE) = numeric()
  expect_equal(x_null, rvar())

  # can't use with_chains when no chain dimension information provided
  expect_error(draws_of(x, with_chains = TRUE) <- 1)
})

# unique, duplicated, etc -------------------------------------------------

test_that("unique.rvar and duplicated.rvar work", {
  x <- rvar(matrix(c(1,1,3, 2,2,3, 1,1,3), nrow = 3))
  unique_x <- rvar(matrix(c(1,1,3, 2,2,3), nrow = 3))

  expect_equal(unique(x), unique_x)
  expect_equal(as.vector(duplicated(x)), c(FALSE, FALSE, TRUE))
  expect_equal(anyDuplicated(x), 3)

  x <- rvar(array(c(1,2, 2,3, 1,2, 3,3, 1,2, 2,3), dim = c(2, 2, 3)))
  unique_x <- x
  unique_x_2 <- rvar(array(c(1,2, 2,3, 1,2, 3,3), dim = c(2, 2, 2)))
  expect_equal(unique(x), unique_x)
  expect_equal(unique(x, MARGIN = 2), unique_x_2)

  expect_error(unique(x, MARGIN = 0), "MARGIN = 0 is invalid")
  expect_error(unique(x, MARGIN = 3), "MARGIN = 3 is invalid")
})

# match and %in% ----------------------------------------------------------

test_that("%in% works on rvars", {
  x <- rvar(matrix(c(1,1,3, 2,2,3, 1,2,3), nrow = 3))
  res <- rvar(matrix(c(TRUE,TRUE,TRUE, FALSE,FALSE,TRUE, TRUE,FALSE,TRUE), nrow = 3))
  expect_equal(x %in% c(1, 3), res)
})

# tibbles / dplyr --------------------------------------------------------------

test_that("rvars work in tibbles", {
  skip_if_not_installed("dplyr")
  skip_if_not_installed("tidyr")

  x_array = array(1:20, dim = c(5,4))
  x = rvar(x_array)
  df = tibble::tibble(x, y = x + 1)

  expect_equal(df$x, x)
  expect_equal(df$y, rvar(x_array + 1))
  expect_equal(dplyr::mutate(df, z = x)$z, x)

  expect_equal(dplyr::mutate(df, z = x * 2)$z, rvar(x_array * 2))
  expect_equal(
    dplyr::mutate(dplyr::group_by(df, 1:4), z = x * 2)$z,
    rvar(x_array * 2)
  )

  df = tibble::tibble(g = letters[1:4], x)
  ref = tibble::tibble(
    a = rvar(x_array[,1, drop = FALSE]),
    b = rvar(x_array[,2, drop = FALSE]),
    c = rvar(x_array[,3, drop = FALSE]),
    d = rvar(x_array[,4, drop = FALSE])
  )
  expect_equal(tidyr::pivot_wider(df, names_from = g, values_from = x), ref)
  expect_equal(tidyr::pivot_longer(ref, a:d, names_to = "g", values_to = "x"), df)

  df$y = df$x + 1
  ref2 = tibble::tibble(
    y = df$y,
    a = c(df$x[[1]], NA, NA, NA),
    b = c(rvar(NA), df$x[[2]], NA, NA),
    c = c(rvar(NA), NA, df$x[[3]], NA),
    d = c(rvar(NA), NA, NA, df$x[[4]]),
  )
  expect_equal(tidyr::pivot_wider(df, names_from = g, values_from = x), ref2)

  expect_equal(dplyr::first(df$x), x[1])

  z = cbind(a = x, b = x + 1)
  df$z = z
  expect_equal(dplyr::mutate(dplyr::rowwise(df), w = z[,"b"])$w, z[,"b"])
})


# ggplot2 -----------------------------------------------------------------

test_that("scale_type(<rvar>) works", {
  skip_if_not_installed("ggplot2")

  expect_no_condition(ggplot2::scale_type(rvar()))
  expect_equal(ggplot2::scale_type(rvar()), "identity")
})


# broadcasting ------------------------------------------------------------

test_that("broadcast_array works", {
  expect_equal(broadcast_array(5, c(1,2,3,1)), array(rep(5, 6), dim = c(1,2,3,1)))
  expect_equal(
    broadcast_array(array(1:4, c(1,4), dimnames = list("x", letters[1:4])), c(2,4)),
    array(rep(1:4, each = 2), c(2,4), dimnames = list(NULL, letters[1:4]))
  )
  expect_equal(
    broadcast_array(array(1:4, c(4,1)), c(4,2)),
    array(c(1:4, 1:4), c(4,2))
  )
  expect_equal(
    broadcast_array(array(1:2, dimnames = list(c("a","b"))), c(2,1,1,1)),
    array(1:2, c(2,1,1,1), dimnames = list(c("a","b"), NULL, NULL, NULL))
  )

  expect_error(broadcast_array(array(1:9, dim = c(3,3)), c(1,9)))
  expect_error(broadcast_array(array(1:9, dim = c(3,3)), c(9)))
})


# conforming chains / draws -----------------------------------------------

test_that("warnings for unequal draws/chains are correct", {
  options(posterior.warn_on_merge_chains = TRUE)
  expect_warning(
    expect_equal(rvar(1:10) + rvar(1:10, nchains = 2), rvar(1:10 + 1:10)),
    "Chains were dropped due to chain information not matching"
  )
  options(posterior.warn_on_merge_chains = FALSE)

  expect_error(
    draws_rvars(x = rvar(1:10), y = rvar(1:11)),
    "variables have different number of draws"
  )

  expect_error(
    rvar(1:10, nchains = 0),
    "chains must be >= 1"
  )
})

# rep ---------------------------------------------------------------------

test_that("rep works", {
  x_array = array(1:10, dim = c(5,2))
  x = rvar(x_array)

  expect_equal(rep(x, times = 3), new_rvar(cbind(x_array, x_array, x_array)))
  expect_equal(rep.int(x, 3), new_rvar(cbind(x_array, x_array, x_array)))
  each_twice = cbind(x_array[,1], x_array[,1], x_array[,2], x_array[,2])
  expect_equal(rep(x, each = 2), new_rvar(each_twice))
  expect_equal(rep(x, each = 2, times = 3), new_rvar(cbind(each_twice, each_twice, each_twice)))
  expect_equal(rep(x, length.out = 3), new_rvar(cbind(x_array, x_array[,1])))
  expect_equal(rep_len(x, 3), new_rvar(cbind(x_array, x_array[,1])))
})

# all.equal ---------------------------------------------------------------------

test_that("all.equal works", {
  x_array = array(1:10, dim = c(5,2))
  x = rvar(x_array)

  expect_true(all.equal(x, x))
  expect_true(!isTRUE(all.equal(x, x + 1)))
  expect_true(!isTRUE(all.equal(x, "a")))
})

test_that("split_chains(<draws_rvar>) works correctly", {
  x_array <- array(1:48, dim = c(4, 3, 4))
  x_rvar <- rvar(x_array, with_chains = TRUE)
  x_draws <- draws_rvars(x = x_rvar)

  x_split_array <- abind::abind(x_array[1:2,,], x_array[3:4,,], along = 2)
  x_split_rvar <- rvar(x_split_array, with_chains = TRUE)
  x_split_draws <- draws_rvars(x = x_split_rvar)

  expect_equal(split_chains(x_draws), x_split_draws)
})

test_that("for_each_draw works with unnamed variables", {
  x <- draws_rvars(
    x0 = rvar(),
    x1 = rvar(1:2),
    x2 = rvar(array(1:4, dim = c(2,2))),
    x3 = rvar(array(1:8, dim = c(2,2,2))),
    x4 = rvar(array(1:16, dim = c(2,2,2,2)))
  )

  draw_list <- numeric()
  x0_list <- list()
  x1_list <- list()
  x2_list <- list()
  x3_list <- list()
  x4_list <- list()
  for_each_draw(x, {
    # have to use <<- due to data masking
    draw_list <<- c(draw_list, .draw)
    x0_list[[.draw]] <<- x0
    x1_list[[.draw]] <<- x1
    x2_list[[.draw]] <<- x2
    x3_list[[.draw]] <<- x3
    x4_list[[.draw]] <<- x4
  })

  expect_equal(draw_list, 1:2)
  expect_equal(x0_list, list(numeric(), numeric()))
  expect_equal(x1_list, list(1, 2))
  expect_equal(x2_list, list(draws_of(x$x2)[1,], draws_of(x$x2)[2,]))
  expect_equal(x3_list, list(
    array(draws_of(x$x3)[1,,], dim = c(2,2), dimnames = list(NULL, NULL)),
    array(draws_of(x$x3)[2,,], dim = c(2,2), dimnames = list(NULL, NULL))
  ))
  expect_equal(x4_list, list(
    array(draws_of(x$x4)[1,,,], dim = c(2,2,2), dimnames = list(NULL, NULL, NULL)),
    array(draws_of(x$x4)[2,,,], dim = c(2,2,2), dimnames = list(NULL, NULL, NULL))
  ))
})

test_that("for_each_draw works with named variables", {
  x <- draws_rvars(
    x1 = rvar(array(1:2, dim = c(2,1), dimnames = list(NULL, "a"))),
    x2 = rvar(array(1:4, dim = c(2,2), dimnames = list(NULL, letters[1:2]))),
    x3 = rvar(array(1:8, dim = c(2,2,2), dimnames = list(NULL, letters[1:2], letters[3:4]))),
    x4 = rvar(array(1:16, dim = c(2,2,2,2), dimnames = list(NULL, letters[1:2], letters[3:4], letters[5:6])))
  )

  x1_list <- list()
  x2_list <- list()
  x3_list <- list()
  x4_list <- list()
  for_each_draw(x, {
    # have to use <<- due to data masking
    x1_list[[.draw]] <<- x1
    x2_list[[.draw]] <<- x2
    x3_list[[.draw]] <<- x3
    x4_list[[.draw]] <<- x4
  })

  expect_equal(x1_list, list(c(a = 1), c(a = 2)))
  expect_equal(x2_list, list(draws_of(x$x2)[1,], draws_of(x$x2)[2,]))
  expect_equal(x3_list, list(draws_of(x$x3)[1,,], draws_of(x$x3)[2,,]))
  expect_equal(x4_list, list(draws_of(x$x4)[1,,,], draws_of(x$x4)[2,,,]))
})

# creating factor rvars ----------------------------------------------------------

test_that("0-length rvar factors are constructured correctly", {
  ref <- factor()
  dim(ref) <- c(1, 0)
  dimnames(ref) <- list("1", NULL)
  expect_equal(draws_of(rvar(factor())), ref)
  expect_equal(draws_of(rvar(character())), ref)

  oldClass(ref) <- c("ordered", "factor")
  expect_equal(draws_of(rvar(ordered(NULL))), ref)
})

test_that("rvar_factor() input with a levels attribute preserves levels", {
  x_array <- matrix(1:10, nrow = 2)
  attr(x_array, "levels") <- letters[1:10]

  ref_draws <- factor(letters[1:10])
  dim(ref_draws) <- c(2, 5)
  dimnames(ref_draws) <- list(1:2, NULL)

  x <- rvar_factor(x_array)
  expect_equal(draws_of(x), ref_draws)
  expect_true(inherits(x, "rvar_factor"))
})

# unique, duplicated, etc -------------------------------------------------

test_that("unique.rvar_factor and duplicated.rvar_factor work", {
  x <- rvar(matrix(letters[c(1,1,3, 2,2,3, 1,1,3)], nrow = 3))
  unique_x <- rvar(matrix(letters[c(1,1,3, 2,2,3)], nrow = 3))

  expect_equal(unique(x), unique_x)
  expect_equal(as.vector(duplicated(x)), c(FALSE, FALSE, TRUE))
  expect_equal(anyDuplicated(x), 3)

  x <- rvar(array(letters[c(1,2, 2,3, 1,2, 3,3, 1,2, 2,3)], dim = c(2, 2, 3)))
  unique_x <- x
  unique_x_2 <- rvar(array(letters[c(1,2, 2,3, 1,2, 3,3)], dim = c(2, 2, 2)))
  expect_equal(unique(x), unique_x)
  expect_equal(unique(x, MARGIN = 2), unique_x_2)
})

# tibbles -----------------------------------------------------------------

test_that("rvar_factors work in tibbles", {
  skip_if_not_installed("dplyr")
  skip_if_not_installed("tidyr")

  x_array <- factor(letters[1:20])
  dim(x_array) <- c(5,4)
  x_array_ordered <- ordered(x_array)
  dim(x_array_ordered) <- c(5,4)
  x = rvar(x_array)
  df = tibble::tibble(x, y = as_rvar_ordered(x))

  expect_equal(df$x, x)
  expect_equal(df$y, rvar(x_array_ordered))
  expect_equal(dplyr::mutate(df, z = x)$z, x)

  expect_equal(dplyr::mutate(df, z = as_rvar_ordered(x))$z, rvar(x_array_ordered))
  expect_equal(
    dplyr::group_by(dplyr::mutate(df, z = as_rvar_ordered(x)), 1:4)$z,
    rvar(x_array_ordered)
  )

  df = tibble::tibble(g = letters[1:4], x)
  ref = tibble::tibble(
    a = rvar(x_array[,1, drop = FALSE]),
    b = rvar(x_array[,2, drop = FALSE]),
    c = rvar(x_array[,3, drop = FALSE]),
    d = rvar(x_array[,4, drop = FALSE])
  )
  expect_equal(tidyr::pivot_wider(df, names_from = g, values_from = x), ref)
  expect_equal(tidyr::pivot_longer(ref, a:d, names_to = "g", values_to = "x"), df)

  df$y = as_rvar_ordered(df$x)
  ref2 = tibble::tibble(
    y = df$y,
    a = c(df$x[[1]], NA, NA, NA),
    b = c(rvar(factor(NA)), df$x[[2]], NA, NA),
    c = c(rvar(factor(NA)), NA, df$x[[3]], NA),
    d = c(rvar(factor(NA)), NA, NA, df$x[[4]]),
  )
  expect_equal(tidyr::pivot_wider(df, names_from = g, values_from = x), ref2)
})

# rep ---------------------------------------------------------------------

test_that("rep works", {
  x_array = array(letters[1:10], dim = c(5,2))
  x = rvar(x_array)

  expect_equal(rep(x, times = 3), new_rvar(cbind(x_array, x_array, x_array)))
  expect_equal(rep.int(x, 3), new_rvar(cbind(x_array, x_array, x_array)))
  each_twice = cbind(x_array[,1], x_array[,1], x_array[,2], x_array[,2])
  expect_equal(rep(x, each = 2), new_rvar(each_twice))
  expect_equal(rep(x, each = 2, times = 3), new_rvar(cbind(each_twice, each_twice, each_twice)))
  expect_equal(rep(x, length.out = 3), new_rvar(cbind(x_array, x_array[,1])))
  expect_equal(rep_len(x, 3), new_rvar(cbind(x_array, x_array[,1])))
})

# all.equal ---------------------------------------------------------------------

test_that("all.equal works", {
  x_array <- array(letters[1:10], dim = c(5,2))
  x <- rvar(x_array)
  x_array_ordered <- ordered(x_array)
  dim(x_array_ordered) <- c(5,2)
  x_ordered <- rvar(x_array_ordered)

  expect_true(all.equal(x, x))
  expect_true(!isTRUE(all.equal(x, x_ordered)))
  expect_true(!isTRUE(all.equal(x, "a")))
})

# Copyright (C) 2012, 2013, 2014, 2015, 2016, 2017, 2018 Trustees of Columbia University
# Copyright (C) 2018, 2019 Aki Vehtari, Paul Bürkner
# See LICENSE.md for more details

#' List of available convergence diagnostics
#'
#' A list of available diagnostics and links to their individual help pages.
#'
#' @name diagnostics
#' @aliases convergence
#'
#' @details
#'
#' |**Function**|**Description**|
#' |:----------|:---------------|
#' | [ess_basic()] | Basic version of effective sample size |
#' | [ess_bulk()] | Bulk effective sample size |
#' | [ess_tail()] | Tail effective sample size |
#' | [ess_quantile()] | Effective sample sizes for quantiles |
#' | [ess_sd()] | Effective sample sizes for standard deviations |
#' | [mcse_mean()] | Monte Carlo standard error for the mean |
#' | [mcse_quantile()] | Monte Carlo standard error for quantiles |
#' | [mcse_sd()] | Monte Carlo standard error for standard deviations |
#' | [rhat_basic()] | Basic version of Rhat |
#' | [rhat()] | Improved, rank-based version of Rhat |
#' | [rstar()] | R* diagnostic |
#'
#' @return
#' See individual functions for a description of return types.
#'
NULL

#' Basic version of the Rhat convergence diagnostic
#'
#' Compute the basic Rhat convergence diagnostic for a single variable as
#' described in Gelman et al. (2013) with some changes according to Vehtari et
#' al. (2021). For practical applications, we strongly recommend the improved
#' Rhat convergence diagnostic implemented in [rhat()].
#'
#' @family diagnostics
#' @template args-conv
#' @template args-conv-split
#' @template args-methods-dots
#' @template return-conv
#' @template ref-gelman-bda-2013
#' @template ref-vehtari-rhat-2021
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' rhat_basic(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' rhat_basic(d$Sigma)
#'
#' @export
rhat_basic <- function(x, ...) UseMethod("rhat_basic")

#' @rdname rhat_basic
#' @export
rhat_basic.default <- function(x, split = TRUE, ...) {
  split <- as_one_logical(split)
  if (split) {
    x <- .split_chains(x)
  }
  .rhat(x)
}

#' @rdname rhat_basic
#' @export
rhat_basic.rvar <- function(x, split = TRUE, ...) {
  summarise_rvar_by_element_with_chains(x, rhat_basic, split, ...)
}

#' Basic version of the effective sample size
#'
#' Compute the basic effective sample size (ESS) estimate for a single variable
#' as described in Gelman et al. (2013) with some changes according to Vehtari et
#' al. (2021). For practical applications, we strongly
#' recommend the improved ESS convergence diagnostics implemented in
#' [ess_bulk()] and [ess_tail()]. See Vehtari (2021) for an in-depth
#' comparison of different effective sample size estimators.
#'
#' @family diagnostics
#' @template args-conv
#' @template args-conv-split
#' @template args-methods-dots
#' @template return-conv
#' @template ref-gelman-bda-2013
#' @template ref-vehtari-rhat-2021
#' @template ref-vehtari-ess-2021
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' ess_basic(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' ess_basic(d$Sigma)
#'
#' @export
ess_basic <- function(x, ...) UseMethod("ess_basic")

#' @rdname ess_basic
#' @export
ess_basic.default <- function(x, split = TRUE, ...) {
  split <- as_one_logical(split)
  if (split) {
    x <- .split_chains(x)
  }
  .ess(x)
}

#' @rdname ess_basic
#' @export
ess_basic.rvar <- function(x, split = TRUE, ...) {
  summarise_rvar_by_element_with_chains(x, ess_basic, split, ...)
}

#' Rhat convergence diagnostic
#'
#' Compute the Rhat convergence diagnostic for a single variable as the maximum
#' of rank normalized split-Rhat and rank normalized folded-split-Rhat as
#' proposed in Vehtari et al. (2021).
#'
#' @family diagnostics
#' @template args-conv
#' @template args-methods-dots
#' @template return-conv
#' @template ref-vehtari-rhat-2021
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' rhat(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' rhat(d$Sigma)
#'
#' @export
rhat <- function(x, ...) UseMethod("rhat")

#' @rdname rhat
#' @export
rhat.default <- function(x, ...) {
  rhat_bulk <- .rhat(z_scale(.split_chains(x)))
  rhat_tail <- .rhat(z_scale(.split_chains(fold_draws(x))))
  max(rhat_bulk, rhat_tail)
}

#' @rdname rhat
#' @export
rhat.rvar <- function(x, ...) {
  summarise_rvar_by_element_with_chains(x, rhat, ...)
}

#' Bulk effective sample size (bulk-ESS)
#'
#' Compute a bulk effective sample size estimate (bulk-ESS) for a single
#' variable. Bulk-ESS is useful as a diagnostic for the sampling efficiency in
#' the bulk of the posterior. It is defined as the effective sample size for
#' rank normalized values using split chains. For the tail effective sample size
#' see [ess_tail()]. See Vehtari (2021) for an in-depth
#' comparison of different effective sample size estimators.
#'
#' @family diagnostics
#' @template args-conv
#' @template args-methods-dots
#' @template return-conv
#' @template ref-vehtari-rhat-2021
#' @template ref-vehtari-ess-2021
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' ess_bulk(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' ess_bulk(d$Sigma)
#'
#' @export
ess_bulk <- function(x, ...) UseMethod("ess_bulk")

#' @rdname ess_bulk
#' @export
ess_bulk.default <- function(x, ...) {
  .ess(z_scale(.split_chains(x)))
}

#' @rdname ess_bulk
#' @export
ess_bulk.rvar <- function(x, ...) {
  summarise_rvar_by_element_with_chains(x, ess_bulk, ...)
}

#' Tail effective sample size (tail-ESS)
#'
#' Compute a tail effective sample size estimate (tail-ESS) for a single
#' variable. Tail-ESS is useful as a diagnostic for the sampling efficiency in
#' the tails of the posterior. It is defined as the minimum of the effective
#' sample sizes for 5% and 95% quantiles. For the bulk effective sample
#' size see [ess_bulk()]. See Vehtari (2021) for an in-depth
#' comparison of different effective sample size estimators.
#'
#' @family diagnostics
#' @template args-conv
#' @template args-methods-dots
#' @template return-conv
#' @template ref-vehtari-rhat-2021
#' @template ref-vehtari-ess-2021
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' ess_tail(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' ess_tail(d$Sigma)
#'
#' @export
ess_tail <- function(x, ...) UseMethod("ess_tail")

#' @rdname ess_tail
#' @export
ess_tail.default <- function(x, ...) {
  q05_ess <- ess_quantile(x, 0.05)
  q95_ess <- ess_quantile(x, 0.95)
  min(q05_ess, q95_ess)
}

#' @rdname ess_tail
#' @export
ess_tail.rvar <- function(x, ...) {
  summarise_rvar_by_element_with_chains(x, ess_tail, ...)
}

#' Effective sample sizes for quantiles
#'
#' Compute effective sample size estimates for quantile estimates of a single
#' variable.
#'
#' @family diagnostics
#' @template args-conv
#' @template args-conv-quantile
#' @template args-methods-dots
#' @template return-conv-quantile
#' @template ref-vehtari-rhat-2021
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' ess_quantile(mu, probs = c(0.1, 0.9))
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' ess_quantile(d$mu, probs = c(0.1, 0.9))
#'
#' @export
ess_quantile <- function(x, probs = c(0.05, 0.95), ...) {
  UseMethod("ess_quantile")
}

#' @rdname ess_quantile
#' @export
ess_quantile.default <- function(x, probs = c(0.05, 0.95), names = TRUE, ...) {
  probs <- as.numeric(probs)
  if (any(probs < 0 | probs > 1)) {
    stop_no_call("'probs' must contain values between 0 and 1.")
  }
  names <- as_one_logical(names)
  out <- ulapply(probs, .ess_quantile, x = x)
  if (names) {
    names(out) <- paste0("ess_q", probs * 100)
  }
  out
}

#' @rdname ess_quantile
#' @export
ess_quantile.rvar <- function(x, probs = c(0.05, 0.95), names = TRUE, ...) {
  summarise_rvar_by_element_with_chains(x, ess_quantile, probs, names, ...)
}

#' @rdname ess_quantile
#' @export
ess_median <- function(x, ...) {
  ess_quantile(x, probs = 0.5, names = FALSE, ...)
}

# ESS of a single quantile
.ess_quantile <- function(x, prob) {
  if (should_return_NA(x)) {
    return(NA_real_)
  }
  x <- as.matrix(x)
  if (prob == 1) {
    len <- length(x)
    prob <- (len - 0.5) / len
  }
  I <- x <= quantile(x, prob)
  .ess(.split_chains(I))
}

#' Effective sample size for the mean
#'
#' Compute an effective sample size estimate for a mean (expectation)
#' estimate of a single variable.
#'
#' @template args-conv
#' @template args-methods-dots
#' @template return-conv
#' @template ref-gelman-bda-2013
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' ess_mean(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' ess_mean(d$Sigma)
#'
#' @export
ess_mean <- function(x, ...) UseMethod("ess_mean")

#' @rdname ess_quantile
#' @export
ess_mean.default <- function(x, ...) {
  .ess(.split_chains(x))
}

#' @rdname ess_mean
#' @export
ess_mean.rvar <- function(x, ...) {
  summarise_rvar_by_element_with_chains(x, ess_mean, ...)
}

#' Effective sample size for the standard deviation
#'
#' Compute an effective sample size estimate for the standard deviation (SD)
#' estimate of a single variable. This is defined as the effective sample size
#' estimate for the absolute deviation from mean.
#'
#' @family diagnostics
#' @template args-conv
#' @template args-methods-dots
#' @template return-conv
#' @template ref-vehtari-rhat-2021
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' ess_sd(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' ess_sd(d$Sigma)
#'
#' @export
ess_sd <- function(x, ...) UseMethod("ess_sd")

#' @rdname ess_sd
#' @export
ess_sd.default <- function(x, ...) {
  .ess(.split_chains(abs(x-mean(x))))
}

#' @rdname ess_sd
#' @export
ess_sd.rvar <- function(x, ...) {
  summarise_rvar_by_element_with_chains(x, ess_sd, ...)
}

#' Monte Carlo standard error for quantiles
#'
#' Compute Monte Carlo standard errors for quantile estimates of a
#' single variable.
#'
#' @family diagnostics
#' @template args-conv
#' @template args-conv-quantile
#' @template args-methods-dots
#' @template return-conv-quantile
#' @template ref-vehtari-rhat-2021
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' mcse_quantile(mu, probs = c(0.1, 0.9))
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' mcse_quantile(d$mu)
#'
#' @export
mcse_quantile <- function(x, probs = c(0.05, 0.95), ...) {
  UseMethod("mcse_quantile")
}

#' @rdname mcse_quantile
#' @export
mcse_quantile.default <- function(x, probs = c(0.05, 0.95), names = TRUE, ...) {
  probs <- as.numeric(probs)
  if (any(probs < 0 | probs > 1)) {
    stop_no_call("'probs' must contain values between 0 and 1.")
  }
  names <- as_one_logical(names)
  out <- ulapply(probs, .mcse_quantile, x = x)
  if (names) {
    names(out) <- paste0("mcse_q", probs * 100)
  }
  out
}

#' @rdname mcse_quantile
#' @export
mcse_quantile.rvar <- function(x, probs = c(0.05, 0.95), names = TRUE, ...) {
  summarise_rvar_by_element_with_chains(x, mcse_quantile, probs, names, ...)
}

#' @rdname mcse_quantile
#' @export
mcse_median <- function(x, ...) {
  mcse_quantile(x, probs = 0.5, names = FALSE, ...)
}

# MCSE of a single quantile
.mcse_quantile <- function(x, prob) {
  ess <- ess_quantile(x, prob)
  p <- c(0.1586553, 0.8413447)
  a <- qbeta(p, ess * prob + 1, ess * (1 - prob) + 1)
  ssims <- sort(x)
  S <- length(ssims)
  th1 <- ssims[max(floor(a[1] * S), 1)]
  th2 <- ssims[min(ceiling(a[2] * S), S)]
  as.vector((th2 - th1) / 2)
}

#' Monte Carlo standard error for the mean
#'
#' Compute the Monte Carlo standard error for the mean (expectation) of a
#' single variable.
#'
#' @family diagnostics
#' @template args-conv
#' @template args-methods-dots
#' @template return-conv
#' @template ref-gelman-bda-2013
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' mcse_mean(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' mcse_mean(d$Sigma)
#'
#' @export
mcse_mean <- function(x, ...) UseMethod("mcse_mean")

#' @rdname mcse_mean
#' @export
mcse_mean.default <- function(x, ...) {
  sd(x) / sqrt(ess_mean(x))
}

#' @rdname mcse_mean
#' @export
mcse_mean.rvar <- function(x, ...) {
  summarise_rvar_by_element_with_chains(x, mcse_mean, ...)
}

#' Monte Carlo standard error for the standard deviation
#'
#' Compute the Monte Carlo standard error for the standard deviation (SD) of a
#' single variable without assuming normality using moments of moments and
#' first order Taylor series approximation (Kenney and Keeping, 1951, p. 141).
#'
#' @family diagnostics
#' @template args-conv
#' @template args-methods-dots
#' @template return-conv
#' @template ref-vehtari-rhat-2021
#' @template ref-kenney-stats-1951
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' mcse_sd(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' mcse_sd(d$Sigma)
#'
#' @export
mcse_sd <- function(x, ...) UseMethod("mcse_sd")

#' @rdname mcse_sd
#' @export
mcse_sd.default <- function(x, ...) {
  # var/sd are not a simple expectation of g(X), e.g. variance
  # has (X-E[X])^2. The following ESS is based on a relevant quantity
  # in the computation and is empirically a good choice.
  sims_c <- x - mean(x)
  ess <- ess_mean((sims_c)^2)
  # Variance of variance estimate by Kenney and Keeping (1951, p. 141),
  # which doesn't assume normality of sims.
  Evar <- mean(sims_c^2)
  varvar <- (mean(sims_c^4) - Evar^2) / ess
  # The first order Taylor series approximation of variance of sd.
  # Kenney and Keeping (1951, p. 141) write "...since fluctuations of
  # any moment are of order N^{-1/2}, squares and higher powers of
  # differentials of the moments can be neglected "
  varsd <- varvar / Evar / 4
  sqrt(varsd)
}

#' @rdname mcse_sd
#' @export
mcse_sd.rvar <- function(x, ...) {
  summarise_rvar_by_element_with_chains(x, mcse_sd, ...)
}

#' Compute Quantiles
#'
#' Compute quantiles of a sample and return them in a format consistent
#' with other summary functions in the \pkg{posterior} package.
#'
#' @template args-conv
#' @template args-conv-quantile
#' @param na.rm (logical) Should `NA` and `NaN` values be removed from `x` prior
#'   to computing quantiles? The default is `FALSE`.
#' @param ... Arguments passed to individual methods (if applicable) and then
#'   on to [stats::quantile()].
#'
#' @return
#' A numeric vector of length `length(probs)`. If `names = TRUE`, it has a
#' [names] attribute with names like `"q5"`, `"q95"`, etc, based on the values
#' of `probs`.
#'
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' quantile2(mu)
#'
#' @export
quantile2 <- function(x, probs = c(0.05, 0.95), na.rm = FALSE, ...) {
  UseMethod("quantile2")
}

#' @rdname quantile2
#' @export
quantile2.default <- function(
  x, probs = c(0.05, 0.95), na.rm = FALSE, names = TRUE, ...
) {
  names <- as_one_logical(names)
  na.rm <- as_one_logical(na.rm)
  if (!na.rm && anyNA(x)) {
    # quantile itself doesn't handle this case (#110)
    out <- rep(NA_real_, length(probs))
  } else {
    out <- quantile(x, probs = probs, na.rm = na.rm, ...)
  }
  if (names) {
    names(out) <- paste0("q", probs * 100)
  } else {
    names(out) <- NULL
  }
  out
}

#' @rdname quantile2
#' @export
quantile2.rvar <- function(
  x, probs = c(0.05, 0.95), na.rm = FALSE, names = TRUE, ...
) {
  summarise_rvar_by_element_with_chains(x, quantile2, probs, na.rm, names, ...)
}

# internal ----------------------------------------------------------------

#' Autocovariance estimates
#'
#' Compute autocovariance estimates for every lag for the specified
#' input sequence using a fast Fourier transform approach. The estimate
#' for lag t is scaled by N-t where N is the length of the sequence.
#'
#' @template args-conv-seq
#' @return A numeric vector of autocovariances at every lag (scaled by N-lag).
#' @keywords internal
#' @export
autocovariance <- function(x) {
  N <- length(x)
  varx <- var(x)
  if (varx==0) {
    # if variance is 0, then all autocovariances are 0
    return(rep(0, N))
  }
  # zero padding makes fft() much faster when N > 1000
  M <- nextn(N)
  Mt2 <- 2 * M
  yc <- x - mean(x)
  yc <- c(yc, rep.int(0, Mt2 - N))
  # FFT based unnormalized autocovariances
  ac <- Re(fft(abs(fft(yc))^2, inverse = TRUE)[1:N])
  # use "biased" estimate as recommended by Geyer (1992)
  # direct scaling with var(x) avoids need to compute "mask effect"
  ac <- ac / ac[1] * varx * (N - 1) / N
  ac
}

#' Autocorrelation estimates
#'
#' Compute autocorrelation estimates for every lag for the specified
#' input sequence using a fast Fourier transform approach. The estimate
#' for lag t is scaled by N-t where N is the length of the sequence.
#'
#' @template args-conv-seq
#' @return A numeric vector of autocorrelations at every lag (scaled by N-lag).
#' @keywords internal
#' @export
autocorrelation <- function(x) {
  ac <- autocovariance(x)
  ac <- ac / ac[1]
}

#' Rank normalization
#'
#' Compute rank normalization for a numeric array. First replace each
#' value by its rank. Average rank for ties are used to conserve the
#' number of unique values of discrete quantities. Second, normalize
#' ranks via the inverse normal transformation.
#'
#' @template args-scale
#' @template args-frac-offset
#' @return A numeric array of rank normalized values with the same size
#'   and dimension as the input.
#' @keywords internal
#' @export
z_scale <- function(x, c = 3/8) {
  r <- rank(as.array(x), ties.method = 'average')
  z <- qnorm(backtransform_ranks(r, c = c))
  z[is.na(x)] <- NA
  if (!is.null(dim(x))) {
    # output should have the input dimension
    z <- array(z, dim = dim(x), dimnames = dimnames(x))
  }
  z
}

#' Rank uniformization
#'
#' Compute rank uniformization for a numeric array. First replace each value by
#' its rank. Average rank for ties are used to conserve the number of unique
#' values of discrete quantities. Second, uniformize ranks to the scale
#' `[1/(2S), 1-1/(2S)]`, where `S` is the number of values.
#'
#' @template args-scale
#' @template args-frac-offset
#' @return A numeric array of uniformized values with the same size
#'   and dimension as the input.
#' @keywords internal
#' @export
u_scale <- function(x, c = 3/8) {
  r <- rank(as.array(x), ties.method = 'average')
  u <- backtransform_ranks(r, c = c)
  u[is.na(x)] <- NA
  if (!is.null(dim(x))) {
    # output should have the input dimension
    u <- array(u, dim = dim(x), dimnames = dimnames(x))
  }
  u
}

#' Rank values
#'
#' Compute ranks for a numeric array, that is, replace each
#' value by its rank. Average rank for ties are used to conserve the
#' number of unique values of discrete quantities.
#'
#' @template args-scale
#' @return A numeric array of ranked values with the same size
#'   and dimension as the input.
#' @keywords internal
#' @export
r_scale <- function(x) {
  r <- rank(as.array(x), ties.method = 'average')
  r[is.na(x)] <- NA
  if (!is.null(dim(x))) {
    # output should have the input dimension
    r <- array(r, dim = dim(x), dimnames = dimnames(x))
  }
  r
}

#' Back-transformation of ranks
#'
#' @param r array of ranks
#' @param c fractional offset; defaults to c = 3/8 as recommend by Blom (1958)
#' @noRd
backtransform_ranks <- function(r, c = 3/8) {
  c <- as_one_numeric(c)
  S <- length(r)
  (r - c) / (S - 2 * c + 1)
}

#' Fold draws around their median
#' @template args-conv
#' @return An array or vector of folded draws.
#' @noRd
fold_draws <- function(x) {
  abs(x - median(x))
}

#' Compute the Rhat convergence diagnostic
#' @template args-conv
#' @template return-conv
#' @noRd
.rhat <- function(x) {
  x <- as.matrix(x)
  if (should_return_NA(x)) {
    return(NA_real_)
  }
  nchains <- NCOL(x)
  niterations <- NROW(x)
  chain_mean <- matrixStats::colMeans2(x)
  chain_var <- matrixStats::colVars(x, center=chain_mean)
  var_between <- niterations * var(chain_mean)
  var_within <- mean(chain_var)
  sqrt((var_between / var_within + niterations - 1) / niterations)
}

#' Compute the effective sample size
#' @template args-conv
#' @template return-conv
#' @noRd
.ess <- function(x) {
  x <- as.matrix(x)
  nchains <- NCOL(x)
  niterations <- NROW(x)
  if (niterations < 3L || should_return_NA(x)) {
    return(NA_real_)
  }
  acov <- apply(x, 2, autocovariance)
  acov_means <- matrixStats::rowMeans2(acov)
  mean_var <- acov_means[1] * niterations / (niterations - 1)
  var_plus <- mean_var * (niterations - 1) / niterations
  if (nchains > 1) {
    var_plus <- var_plus + var(matrixStats::colMeans2(x))
  }

  # Geyer's initial positive sequence
  rho_hat_t <- rep.int(0, niterations)
  t <- 0
  rho_hat_even <- 1
  rho_hat_t[t + 1] <- rho_hat_even
  rho_hat_odd <- 1 - (mean_var - acov_means[t + 2]) / var_plus
  rho_hat_t[t + 2] <- rho_hat_odd
  while (t < NROW(acov) - 5 && !is.nan(rho_hat_even + rho_hat_odd) &&
         (rho_hat_even + rho_hat_odd > 0)) {
    t <- t + 2
    rho_hat_even = 1 - (mean_var - acov_means[t + 1]) / var_plus
    rho_hat_odd = 1 - (mean_var - acov_means[t + 2]) / var_plus
    if ((rho_hat_even + rho_hat_odd) >= 0) {
      rho_hat_t[t + 1] <- rho_hat_even
      rho_hat_t[t + 2] <- rho_hat_odd
    }
  }
  max_t <- t
  # this is used in the improved estimate
  if (rho_hat_even>0)
    rho_hat_t[max_t + 1] <- rho_hat_even

  # Geyer's initial monotone sequence
  t <- 0
  while (t <= max_t - 4) {
    t <- t + 2
    if (rho_hat_t[t + 1] + rho_hat_t[t + 2] >
        rho_hat_t[t - 1] + rho_hat_t[t]) {
      rho_hat_t[t + 1] = (rho_hat_t[t - 1] + rho_hat_t[t]) / 2;
      rho_hat_t[t + 2] = rho_hat_t[t + 1];
    }
  }
  ess <- nchains * niterations
  # Geyer's truncated estimate
  # tau_hat <- -1 + 2 * sum(rho_hat_t[1:max_t])
  # Improved estimate reduces variance in antithetic case
  tau_hat <- -1 + 2 * sum(rho_hat_t[1:max_t]) + rho_hat_t[max_t+1]
  # Safety check for negative values and with max ess equal to ess*log10(ess)
  tau_bound <- 1 / log10(ess)
  if (tau_hat < tau_bound) {
    warning_no_call("The ESS has been capped to avoid unstable estimates.")
    tau_hat <- tau_bound
  }
  ess <- ess / tau_hat
  ess
}

# should NA be returned by a convergence diagnostic?
should_return_NA <- function(x, tol = .Machine$double.eps) {
  if (anyNA(x) || checkmate::anyInfinite(x)) {
    return(TRUE)
  }
  # checking for constant input per chain is too conservative for ess_tail
  # so we don't check this for the time being until we find a better solution
  # if (is.matrix(x)) {
  #   # is_constant() vectorized over x columns
  #   if (is.logical(x)) {
  #     return(any(matrixStats::colAnys(x) == matrixStats::colAlls(x)))
  #   } else if (is.integer(x)) {
  #     return(any(matrixStats::rowDiffs(matrixStats::colRanges(x)) == 0L))
  #   } else {
  #     return(any(abs(matrixStats::rowDiffs(matrixStats::colRanges(x))) < tol))
  #   }
  # }
  is_constant(x, tol = tol)
}

test_that("thin_draws works correctly", {
  x <- as_draws_array(example_draws())
  expect_equal(niterations(thin_draws(x, 5L)), niterations(x) / 5)
  expect_equal(x, thin_draws(x, thin = 1L))
  expect_error(thin_draws(x, -1), "'thin' must be greater than or equal to 1")
  expect_error(thin_draws(x, 1000), "'thin' must be smaller than")
})

test_that("thin_draws works on rvars", {
  x <- example_draws()

  expect_equal(thin_draws(as_draws_rvars(x)$theta, 10L), as_draws_rvars(thin_draws(x, 10L))$theta)
})

test_that("automatic thinning works as expected", {
  x <- as_draws_array(example_draws())
  mu <- subset_draws(x, "mu")
  mu_1 <- subset_draws(mu, chain = 1)
  mu_2 <- subset_draws(mu, chain = 2)
  mu_3 <- subset_draws(mu, chain = 3)
  mu_4 <- subset_draws(mu, chain = 4)

  ess_mu_1 <- SW(min(ess_tail(mu_1), ess_bulk(mu_1)))
  ess_mu_2 <- SW(min(ess_tail(mu_2), ess_bulk(mu_2)))
  ess_mu_3 <- SW(min(ess_tail(mu_3), ess_bulk(mu_3)))
  ess_mu_4 <- SW(min(ess_tail(mu_4), ess_bulk(mu_4)))

  thin_by <- niterations(mu) / mean(
    c(ess_mu_1, ess_mu_2, ess_mu_3, ess_mu_4))
  
  expect_equal(thin_draws(mu), thin_draws(mu, thin = thin_by))

})

test_that("rhat diagnostics return reasonable values", {
  tau <- extract_variable_matrix(example_draws(), "tau")

  rhat <- rhat_basic(tau)
  expect_true(rhat > 0.99 & rhat < 1.05)

  rhat <- rhat(tau)
  expect_true(rhat > 0.99 & rhat < 1.05)
})

test_that("ess diagnostics return reasonable values", {
  tau <- extract_variable_matrix(example_draws(), "tau")
  mu <- extract_variable_matrix(example_draws("multi_normal"), "mu[1]")

  ess <- ess_basic(tau)
  expect_true(ess > 250 & ess < 310)

  ess <- ess_mean(tau)
  expect_true(ess > 250 & ess < 310)

  ess <- ess_sd(tau)
  expect_true(ess > 250 & ess < 310)

  ess <- ess_bulk(tau)
  expect_true(ess > 230 & ess < 280)

  # some chains are constant for the computed tail quantiles
  ess <- ess_tail(tau)
  expect_true(ess > 180 & ess < 220)
  # use a different example to obtain non-NA value
  ess <- ess_tail(mu)
  expect_true(ess > 330 & ess < 380)

  ess <- ess_quantile(tau, probs = c(0.2, 0.8))
  expect_equal(names(ess), c("ess_q20", "ess_q80"))
  expect_true(ess[1] > 150 & ess[1] < 210)
  expect_true(ess[2] > 350 & ess[2] < 420)

  ess <- ess_median(tau)
  expect_true(ess > 350 & ess < 420)
})

test_that("negative ess estimates are avoided with a warning", {
  x <- stats::arima.sim(list(ar=-0.9), 1000)
  expect_warning(ess <- ess_basic(x), "The ESS has been capped")
  expect_equal(ess, 3000)
})

test_that("mcse diagnostics return reasonable values", {
  tau <- extract_variable_matrix(example_draws(), "tau")

  mcse <- mcse_mean(tau)
  expect_true(mcse > 0.15 & mcse < 0.25)

  mcse <- mcse_sd(tau)
  expect_true(mcse > 0.15 & mcse < 0.25)

  mcse <- mcse_quantile(tau, probs = c(0.2, 0.8))
  expect_equal(names(mcse), c("mcse_q20", "mcse_q80"))
  expect_true(mcse[1] > 0.16 & mcse[1] < 0.21)
  # due to right skewness of tau the 90%ile is way more uncertain
  expect_true(mcse[2] > 0.3 & mcse[2] < 0.7)

  mcse <- mcse_median(tau)
  expect_true(mcse > 0.2 & mcse < 0.3)

  # check if issue #252 remains fixed
  se <- mcse_quantile(tau, c(0.999999, 1))
  expect_equivalent(se[1], se[2])

  se <- mcse_quantile(tau, c(0.000001, 0))
  expect_equivalent(se[1], se[2])
})

test_that("convergence diagnostics accept vectors as input", {
  set.seed(1234)
  x <- rnorm(1000)

  rhat <- rhat(x)
  expect_true(rhat > 0.99 & rhat < 1.01)

  ess <- ess_bulk(x)
  expect_true(ess > 900 & ess < 1100)

  ess <- ess_tail(x)
  expect_true(ess > 750 & ess < 850)

  mcse <- mcse_mean(x)
  expect_true(mcse > 0.02 & mcse < 0.04)
})

test_that("convergence diagnostics handle special cases correctly", {
  set.seed(1234)

  x <- c(rnorm(10), NA)
  expect_true(is.na(rhat_basic(x)))
  expect_true(is.na(ess_basic(x)))

  x <- c(rnorm(10), Inf)
  expect_true(is.na(rhat_basic(x)))
  expect_true(is.na(ess_basic(x)))

  x <- rep(1, 10)
  expect_true(is.na(rhat_basic(x)))
  expect_true(is.na(ess_basic(x)))

  # constant-per-chain checks deactivated for now
  # x <- cbind(1, rnorm(10))
  # expect_true(is.na(rhat_basic(x)))
  # expect_true(is.na(ess_basic(x)))
})

test_that("convergence diagnostics throw correct errors", {
  mu <- extract_variable_matrix(example_draws(), "mu")
  expect_error(ess_quantile(mu, probs = 1.2), "'probs' must contain values between 0 and 1")
  expect_error(mcse_quantile(mu, probs = 1.2), "'probs' must contain values between 0 and 1")
})

test_that("convergence functions work with rvars", {
  tau <- extract_variable_matrix(example_draws(), "tau")
  tau_rvar <- rvar(tau, with_chains = TRUE)

  expect_equal(ess_basic(tau_rvar), ess_basic(tau))
  expect_equal(ess_bulk(tau_rvar), ess_bulk(tau))
  expect_equal(ess_tail(tau_rvar), ess_tail(tau))
  # convergence functions with multiple return values on rvars returns a column
  # vector; should be equal to the NULL-dimension vector format when transposed
  expect_equal(t(ess_quantile(tau_rvar, probs = c(.1, .9))), t(ess_quantile(tau, probs = c(.1, .9))))
  expect_equal(ess_sd(tau_rvar), ess_sd(tau))
  expect_equal(mcse_mean(tau_rvar), mcse_mean(tau))
  expect_equal(t(mcse_quantile(tau_rvar, probs = c(.1, .9))), t(mcse_quantile(tau, probs = c(.1, .9))))
  expect_equal(mcse_sd(tau_rvar), mcse_sd(tau))
  expect_equal(rhat_basic(tau_rvar), rhat_basic(tau))
  expect_equal(rhat(tau_rvar), rhat(tau))
})

test_that("autocovariance returns correct results", {
  x <- rnorm(100)
  ac1 <- autocovariance(x)
  ac2 <- acf(x, type = "covariance", lag.max = length(x), plot = FALSE)$acf[, 1, 1]
  expect_equal(ac1, ac2)

  x <- arima.sim(list(ar = c(0.5, -0.3)), 100)
  ac1 <- autocovariance(x)
  ac2 <- acf(x, type = "covariance", lag.max = length(x), plot = FALSE)$acf[, 1, 1]
  expect_equal(ac1, ac2)
})

#' Get variable names from `draws` objects
#'
#' Get variable names from [`draws`] objects.
#'
#' @name variables
#' @template args-methods-x
#' @template args-methods-dots
#' @template args-methods-reserved
#' @template args-methods-with_indices
#'
#' @details
#' `variables()` returns a vector of all variable names, and `nvariables()`
#' returns the number of variables.
#'
#' @return
#'
#' For `variables()`, a character vector.
#'
#' For `nvariables()`, a scalar integer.
#'
#' @seealso [`variables<-`], [`rename_variables`], [`draws-index`]
#'
#' @examples
#' x <- example_draws()
#'
#' variables(x)
#' nvariables(x)
#' variables(x) <- letters[1:nvariables(x)]
#' @export
variables <- function(x, ...) {
  UseMethod("variables")
}

#' @export
variables.NULL <- function(x, ...) {
  NULL
}

#' @rdname variables
#' @export
variables.draws_matrix <- function(x, reserved = FALSE, with_indices = TRUE, ...) {
  out <- remove_reserved_variable_names(colnames(x), reserved)
  variable_names(out, with_indices)
}

#' @rdname variables
#' @export
variables.draws_array <- function(x, reserved = FALSE, with_indices = TRUE, ...) {
  out <- remove_reserved_variable_names(dimnames(x)[[3L]], reserved)
  variable_names(out, with_indices)
}

#' @rdname variables
#' @export
variables.draws_df <- function(x, reserved = FALSE, with_indices = TRUE, ...) {
  # reserved_df_variables are special data.frame columns
  # which should never be included as variables
  out <- names(x)[!names(x) %in% reserved_df_variables()]
  out <- remove_reserved_variable_names(out, reserved)
  variable_names(out, with_indices)
}

#' @rdname variables
#' @export
variables.draws_list <- function(x, reserved = FALSE, with_indices = TRUE, ...) {
  if (!length(x)) {
    return(character(0))
  }
  out <- remove_reserved_variable_names(names(x[[1]]), reserved)
  variable_names(out, with_indices)
}

#' @rdname variables
#' @export
variables.draws_rvars <- function(x, reserved = FALSE, with_indices = FALSE, ...) {
  with_indices <- as_one_logical(with_indices)
  if (with_indices) {
    out <- unlist(.mapply(flatten_indices, list(x, names(x)), NULL), recursive = FALSE, use.names = FALSE)
  } else {
    out <- names(x)
  }
  remove_reserved_variable_names(out, reserved)
}

# remove reserved variable names
remove_reserved_variable_names <- function(variables, reserved) {
  reserved <- as_one_logical(reserved)
  if (!reserved && length(variables)) {
    # can't use setdiff() here as in the edge case where someone
    # manually creates duplicate columns it will give incorrect results
    variables <- variables[!variables %in% reserved_variables()]
  }
  variables
}


#' Set variable names in `draws` objects
#'
#' Set variable names for all variables in a [`draws`] object. The
#' `set_variables()` form is useful when using pipe operators.
#'
#' @template args-methods-x
#' @template args-methods-dots
#' @template args-methods-with_indices
#' @param value,variables (character vector) new variable names.
#'
#' @details
#' `variables(x) <- value` allows you to modify the vector of variable names,
#' similar to how `names(x) <- value` works for vectors and lists. For renaming
#' specific variables, `set_variables(x, value)` works equivalently, but is more intuitive
#' when using the pipe operator.
#'
#' For renaming specific variables, [rename_variables()] may offer a more
#' convenient approach.
#'
#' @return Returns a [`draws`] object of the same format as `x`, with
#'   variables named as specified.
#'
#' @seealso [`variables`], [`rename_variables`], [`draws-index`]
#'
#' @examples
#' x <- example_draws()
#'
#' variables(x)
#' nvariables(x)
#' variables(x) <- letters[1:nvariables(x)]
#'
#' # or equivalently...
#' x <- set_variables(x, letters[1:nvariables(x)])
#'
#' @export
`variables<-` <- function(x, ..., value) {
  UseMethod("variables<-")
}

#' @rdname variables-set
#' @export
`variables<-.draws_matrix` <- function(x, with_indices = TRUE, ..., value) {
  check_new_variables(value)
  variable_names(colnames(x), with_indices) <- value
  x
}

#' @rdname variables-set
#' @export
`variables<-.draws_array` <- function(x, with_indices = TRUE, ..., value) {
  check_new_variables(value)
  variable_names(dimnames(x)[[3L]], with_indices) <- value
  x
}

#' @rdname variables-set
#' @export
`variables<-.draws_df` <- function(x, with_indices = TRUE, ..., value) {
  check_new_variables(value)
  names_i <- !names(x) %in% reserved_df_variables()
  variable_names(names(x)[names_i], with_indices) <- value
  x
}

#' @rdname variables-set
#' @export
`variables<-.draws_list` <- function(x, with_indices = TRUE, ..., value) {
  check_new_variables(value)
  for (i in seq_along(x)) {
    variable_names(names(x[[i]]), with_indices) <- value
  }
  x
}

#' @rdname variables-set
#' @export
`variables<-.draws_rvars` <- function(x, with_indices = FALSE, ..., value) {
  with_indices <- as_one_logical(with_indices)
  check_new_variables(value)
  if (with_indices) {
    # need to make sure that the provided names only change the base names of
    # the variables and that the indexes otherwise match
    vars <- split_variable_names(value)
    base_names <- unique(vars$base_name)
    base_name_i <- match(vars$base_name, base_names)

    x_indices <- ulapply(x, flatten_indices, recursive = FALSE, use.names = FALSE)
    x_base_name_i <- rep(seq_along(x), lengths(x))
    if (!identical(x_base_name_i, base_name_i) || !identical(x_indices, vars$indices)) {
      stop_no_call(
        "variables(<draws_rvars>) <- value is only allowed when the indices in `value` match ",
        "the indices in the original names. To modify the names of the indices, either modify ",
        "the dims of the underlying rvars or convert to another draws format first."
      )
    }

    names(x) <- base_names
  } else {
    names(x) <- value
  }
  x
}

#' @rdname variables-set
#' @export
set_variables <- function(x, variables, ...) {
  variables(x, ...) <- variables
  x
}

#' @rdname variables
#' @export
nvariables <- function(x, ...) {
  UseMethod("nvariables")
}
#' @export
nvariables.NULL <- function(x, ...) {
  0
}

#' @export
nvariables.draws <- function(x, ...) {
  length(variables(x, ...))
}


# internal ----------------------------------------------------------------

# check validity of existing variable names: e.g., that
# all `variables` exist in `x` and that no `variables`are reserved words
# Additionally, this returns the canonical name, so e.g. "theta" will get
# converted to c("theta[1]", "theta[2]", ...) if those variables exist.
# @param regex should 'variables' be treated as regular expressions?
# @param scalar should only scalar variables be matched?
check_existing_variables <- function(variables, x, regex = FALSE,
                                     scalar = FALSE, exclude = FALSE) {
  check_draws_object(x)
  if (is.null(variables)) {
    return(NULL)
  }

  # do not allow NA values in variables
  checkmate::assertCharacter(variables, any.missing = FALSE)

  regex <- as_one_logical(regex)
  scalar <- as_one_logical(scalar)
  exclude <- as_one_logical(exclude)
  variables <- unique(as.character(variables))
  all_variables <- variables(x, reserved = TRUE)

  if (regex) {
    tmp <- named_list(variables)
    for (i in seq_along(variables)) {
      tmp[[i]] <- grep(variables[i], all_variables)
    }
    # regular expressions are not required to match anything
    missing_variables <- NULL
    variables <- as.character(all_variables[unique(unlist(tmp))])
  } else if (!scalar) {
    # need to find variables that are matched by either a scalar or vector
    # variable in x and what the matching variable is, while keeping original
    # order of input `variables`

    # find scalar variables (1-to-1 match between all_variables and variables)
    scalar_input_ixs <- match(all_variables, variables)
    # find vector variable matches (match all_variables with the indexing stripped)
    all_variables_base <- all_variables
    # exclude already matched scalar variables
    all_variables_base[!is.na(scalar_input_ixs)] <- NA_character_
    all_variables_base <- split_variable_names(all_variables_base)$base_name
    vector_input_ixs <- match(all_variables_base, variables)
    # compose the vector of indices of matched input variables
    input_ixs <- c(scalar_input_ixs[!is.na(scalar_input_ixs)],
                   vector_input_ixs[!is.na(vector_input_ixs)])
    # compose the vector of indices of matched all_variables
    all_var_ixs <- seq_along(all_variables)
    all_var_matched_ixs <- c(all_var_ixs[!is.na(scalar_input_ixs)],
                             all_var_ixs[!is.na(vector_input_ixs)])
    # select missed input variables
    missing_vars_mask <- rep_len(TRUE, length(variables))
    missing_vars_mask[input_ixs] <- FALSE
    missing_variables <- variables[missing_vars_mask]
    # select matched all_variables maintaining the input variables order
    variables <- all_variables[all_var_matched_ixs[order(input_ixs, all_var_matched_ixs)]]
  } else {
    missing_variables <- setdiff(variables, all_variables)
  }
  variables <- check_reserved_variables(variables)
  if (length(missing_variables)) {
    stop_no_call("The following variables are missing in the draws object: ",
          comma(missing_variables))
  }

  # handle excluding variables for subset_draws
  if (exclude) {
    variables <- setdiff(all_variables, variables)
  }

  invisible(variables)
}

# check validity of new variables: e.g., that there are
# no duplicates in `variables` and that they do not use
# reserved words
check_new_variables <- function(variables) {
  # use anyDuplicated() for the check since it is faster than any(duplicated(x)) and
  # we shouldn't expect to take this branch often (since it is an error)
  if (anyDuplicated(variables)) {
    duplicates = unique(variables[duplicated(variables)])
    stop_no_call(
      "Duplicate variable names are not allowed in draws objects.\n",
      "The following variable names are duplicates:\n",
      comma(duplicates)
    )
  }
  check_reserved_variables(variables)
}

# check variables do not make use of reserved words
check_reserved_variables <- function(variables) {
  assert_character(variables)
  # for now only check reserved columns used in 'draws_df' objects
  # other reserved variables such as '.log_weight' may be overwritten
  # this has the advantage that power users can directly add such variables
  used_reserved_variables <- intersect(reserved_df_variables(), variables)
  if (length(used_reserved_variables)) {
    stop_no_call("Variable names ", comma(used_reserved_variables), " are reserved.")
  }
  invisible(variables)
}

test_that('duplicate variable names are not allowed', {
  x = matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "a")))

  err = "Duplicate variable names are not allowed"
  expect_error(as_draws_matrix(x), err)
  expect_error(as_draws_df(x), err)
  expect_error(as_draws_list(x), err)
  expect_error(as_draws_array(x), err)
})

test_that("variables() work with NULL", {
  expect_equal(variables(NULL), NULL)
  expect_equal(nvariables(NULL), 0)
})

test_that("variables() and variables<-() work on draws_matrix", {
  x <- as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b"))))

  ref <- as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("x", "y[1]"))))

  expect_equal(variables(x), c("a", "b"))
  expect_equal(variables(ref), c("x", "y[1]"))

  variables(x) <- c("x", "y[1]")
  expect_equal(x, ref)
})

test_that("variables() and variables<-() work on draws_array", {
  x <- as_draws_array(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b"))))

  ref <- as_draws_array(matrix(11:20, ncol = 2, dimnames = list(NULL, c("x", "y[1]"))))

  expect_equal(variables(x), c("a", "b"))
  expect_equal(variables(ref), c("x", "y[1]"))

  variables(x) <- c("x", "y[1]")
  expect_equal(x, ref)
})

test_that("variables() and variables<-() work on draws_list", {
  x <- as_draws_list(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b"))))

  ref <- as_draws_list(matrix(11:20, ncol = 2, dimnames = list(NULL, c("x", "y[1]"))))

  expect_equal(variables(x), c("a", "b"))
  expect_equal(variables(ref), c("x", "y[1]"))

  variables(x) <- c("x", "y[1]")
  expect_equal(x, ref)
})

test_that("variables() and variables<-() work on draws_df", {
  x <- as_draws_df(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b"))))

  ref <- as_draws_df(matrix(11:20, ncol = 2, dimnames = list(NULL, c("x", "y[1]"))))

  expect_equal(variables(x), c("a", "b"))
  expect_equal(variables(ref), c("x", "y[1]"))

  variables(x) <- c("x", "y[1]")
  expect_equal(x, ref)
})

test_that("variables() and variables<-() work on draws_rvars", {
  x <- as_draws_rvars(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b"))))

  ref <- as_draws_rvars(matrix(11:20, ncol = 2, dimnames = list(NULL, c("x", "y[1]"))))

  # variables works a bit differently for draws_rvars
  expect_equal(variables(x), c("a", "b"))
  expect_equal(variables(ref), c("x", "y"))

  variables(x) <- c("x", "y")
  expect_equal(x, ref)
})

test_that("variables() works on draws_df with duplicate columns", {
  # in the annoying case where someone manually changes a draws_df to have duplicate columns,
  # make sure that variables() returns the correct result...
  x <- as_draws_df(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b"))))
  names(x)[names(x) == "b"] = "a"

  expect_equal(variables(x), c("a", "a"))
})

test_that("variables() works on NULL", {
  expect_equal(variables(NULL), NULL)
})

# since set_variables is just a wrapper around variables() this only
# tests it on draws_matrix for now. See the tests for variables() for
# format-specific tests.
test_that("set_variables works on draws_matrix", {
  x <- as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b[1]"))))

  ref <- as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("c", "d"))))
  expect_equal(set_variables(x, c("c", "d")), ref)
})

test_that("cannot set a variable to a reserved word", {
  x <- as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b[1]"))))
  err <- "Variable names.*are reserved"
  expect_error(set_variables(x, c(".chain", "b[1]")), err)
  expect_error(set_variables(x, c(".iteration", "b[1]")), err)
  expect_error(set_variables(x, c(".draw", "b[1]")), err)
})

test_that("cannot set duplicate variable names", {
  x <- as_draws_matrix(matrix(11:20, ncol = 2, dimnames = list(NULL, c("a", "b[1]"))))
  expect_error(set_variables(x, c("a", "a")), "Duplicate variable names are not allowed")
})

test_that("with_indices works", {
  x <- example_draws()
  draws <- list(
    array = as_draws_array(x),
    df = as_draws_df(x),
    list = as_draws_list(x),
    matrix = as_draws_matrix(x),
    rvars = as_draws_rvars(x)
  )

  mu_tau_theta = c(
    "mu", "tau", "theta[1]", "theta[2]", "theta[3]", "theta[4]",
    "theta[5]", "theta[6]", "theta[7]", "theta[8]"
  )
  a_b_c = c("a", "b", "c[1]", "c[2]", "c[3]", "c[4]", "c[5]", "c[6]", "c[7]", "c[8]")
  for (type in names(draws)) {
    expect_equal(variables(draws[[!!type]], with_indices = TRUE), mu_tau_theta)
    expect_equal(variables(draws[[!!type]], with_indices = FALSE), c("mu", "tau", "theta"))
    expect_equal(nvariables(draws[[!!type]], with_indices = TRUE), 10)
    expect_equal(nvariables(draws[[!!type]], with_indices = FALSE), 3)

    expect_equal(
      variables(set_variables(draws[[!!type]], a_b_c, with_indices = TRUE), with_indices = FALSE),
      c("a", "b", "c")
    )
    expect_equal(
      variables(set_variables(draws[[!!type]], c("a","b","c"), with_indices = FALSE), with_indices = TRUE),
      a_b_c
    )

    expect_error(
      set_variables(draws[[!!type]], c("a","c","c"), with_indices = FALSE),
      "[Dd]uplicate"
    )
  }

  for (type in head(names(draws), -1)) {
    expect_error(
      set_variables(draws[[!!type]], c("a","b"), with_indices = FALSE),
      "base name.*[Ll]engths must match"
    )
  }

  expect_error(
    set_variables(
      draws$rvars,
      c("a", "b", "XX[1]", "c[2]", "c[3]", "c[4]", "c[5]", "c[6]", "c[7]", "c[8]"),
      with_indices = TRUE
    )
  )
})

#' Normalized entropy
#'
#' Normalized entropy, for measuring dispersion in draws from categorical distributions.
#'
#' @param x (multiple options) A vector to be interpreted as draws from
#' a categorical distribution, such as:
#'  - A [factor]
#'  - A [numeric] (should be [integer] or integer-like)
#'  - An [rvar], [rvar_factor], or [rvar_ordered]
#'
#' @details
#' Calculates the normalized Shannon entropy of the draws in `x`. This value is
#' the entropy of `x` divided by the maximum entropy of a distribution with `n`
#' categories, where `n` is `length(unique(x))` for numeric vectors and
#' `length(levels(x))` for factors:
#'
#' \deqn{-\frac{\sum_{i = 1}^{n} p_i \log(p_i)}{\log(n)}}
#'
#' This scales the output to be between 0 (all probability in one category)
#' and 1 (uniform). This form of normalized entropy is referred to as
#' \eqn{H_\mathrm{REL}} in Wilcox (1967).
#'
#' @returns
#' If `x` is a [factor] or [numeric], returns a length-1 numeric vector with a value
#' between 0 and 1 (inclusive) giving the normalized Shannon entropy of `x`.
#'
#' If `x` is an [rvar], returns an array of the same shape as `x`, where each
#' cell is the normalized Shannon entropy of the draws in the corresponding cell of `x`.
#' @template ref-wilcox-iqv-1967
#' @examples
#' set.seed(1234)
#'
#' levels <- c("a", "b", "c", "d", "e")
#'
#' # a uniform distribution: high normalized entropy
#' x <- factor(
#'   sample(levels, 4000, replace = TRUE, prob = c(0.2, 0.2, 0.2, 0.2, 0.2)),
#'   levels = levels
#' )
#' entropy(x)
#'
#' # a unimodal distribution: low normalized entropy
#' y <- factor(
#'   sample(levels, 4000, replace = TRUE, prob = c(0.95, 0.02, 0.015, 0.01, 0.005)),
#'   levels = levels
#' )
#' entropy(y)
#'
#' # both together, as an rvar
#' xy <- c(rvar(x), rvar(y))
#' xy
#' entropy(xy)
#' @export
entropy <- function(x) {
  UseMethod("entropy")
}
#' @rdname entropy
#' @export
entropy.default <- function(x) {
  if (anyNA(x)) return(NA_real_)
  p <- prop.table(simple_table(x)$count)
  n <- length(p)

  if (n == 1) {
    out <- 0
  } else {
    p <- p[p > 0]
    out <- -sum(p * log(p)) / log(n)
  }
  out
}
#' @rdname entropy
#' @export
entropy.rvar <- function(x) {
  summarise_rvar_by_element(x, entropy)
}


#' Dissention
#'
#' Dissention, for measuring dispersion in draws from ordinal distributions.
#'
#' @param x (multiple options) A vector to be interpreted as draws from
#' an ordinal distribution, such as:
#'  - A [factor]
#'  - A [numeric] (should be [integer] or integer-like)
#'  - An [rvar], [rvar_factor], or [rvar_ordered]
#'
#' @details
#' Calculates Tastle and Wierman's (2007) *dissention* measure:
#'
#' \deqn{-\sum_{i = 1}^{n} p_i \log_2 \left(1 - \frac{|x_i - \mathrm{E}(x)| }{\max(x) - \min(x)} \right)}
#'
#' This ranges from 0 (all probability in one category) through 0.5 (uniform) to
#' 1 (bimodal: all probability split equally between the first and last category).
#'
#' @returns
#' If `x` is a [factor] or [numeric], returns a length-1 numeric vector with a value
#' between 0 and 1 (inclusive) giving the dissention of `x`.
#'
#' If `x` is an [rvar], returns an array of the same shape as `x`, where each
#' cell is the dissention of the draws in the corresponding cell of `x`.
#' @template ref-tastle-wierman-2007
#' @examples
#' set.seed(1234)
#'
#' levels <- c("lowest", "low", "neutral", "high", "highest")
#'
#' # a bimodal distribution: high dissention
#' x <- ordered(
#'   sample(levels, 4000, replace = TRUE, prob = c(0.45, 0.04, 0.02, 0.04, 0.45)),
#'   levels = levels
#' )
#' dissent(x)
#'
#' # a unimodal distribution: low dissention
#' y <- ordered(
#'   sample(levels, 4000, replace = TRUE, prob = c(0.95, 0.02, 0.015, 0.01, 0.005)),
#'   levels = levels
#' )
#' dissent(y)
#'
#' # both together, as an rvar
#' xy <- c(rvar(x), rvar(y))
#' xy
#' dissent(xy)
#' @export
dissent <- function(x) {
  UseMethod("dissent")
}
#' @rdname dissent
#' @export
dissent.default <- function(x) {
  if (anyNA(x)) return(NA_real_)
  if (length(x) == 0) return(0)

  if (is.factor(x)) {
    d <- length(levels(x)) - 1
    x <- as.numeric(x)
  } else {
    d <- diff(range(x))
  }

  tab <- simple_table(x)
  p <- prop.table(tab$count)

  if (length(p) == 1) {
    out <- 0
  } else {
    x_i <- tab$x
    out <- -sum(p * log2(1 - abs(x_i - mean(x)) / d))
  }
  out
}
#' @rdname dissent
#' @export
dissent.rvar <- function(x) {
  summarise_rvar_by_element(x, dissent)
}


#' Modal category
#'
#' Modal category of a vector.
#'
#' @param x (multiple options) A vector to be interpreted as draws from
#' a categorical distribution, such as:
#'  - A [factor]
#'  - A [numeric] (should be [integer] or integer-like)
#'  - An [rvar], [rvar_factor], or [rvar_ordered]
#'
#' @details
#' Finds the modal category (i.e., most frequent value) in `x`. In the case of
#' ties, returns the first tie.
#'
#' @returns
#' If `x` is a [factor] or [numeric], returns a length-1 vector containing
#' the modal value.
#'
#' If `x` is an [rvar], returns an array of the same shape as `x`, where each
#' cell is the modal value of the draws in the corresponding cell of `x`.
#' @examples
#' x <- factor(c("a","b","b","c","d"))
#' modal_category(x)
#'
#' # in the case of ties, the first tie is returned
#' y <- factor(c("a","c","c","d","d"))
#' modal_category(y)
#'
#' # both together, as an rvar
#' xy <- c(rvar(x), rvar(y))
#' xy
#' modal_category(xy)
#' @export
modal_category <- function(x) {
  UseMethod("modal_category")
}
#' @rdname modal_category
#' @export
modal_category.default <- function(x) {
  if (anyNA(x)) return(NA)
  tab <- simple_table(x)
  tab$x[which.max(tab$count)]
}
#' @rdname modal_category
#' @export
modal_category.rvar <- function(x) {
  summarise_rvar_by_element(x, modal_category)
}


# helpers -----------------------------------------------------------------

#' A simpler, faster version of table() that does not mangle values into
#' character strings as names, but preserves them as their original type
#' @param x a vector (numeric, factor, character, etc)
#' @returns a list with two components of the same length
#'  - `x`: unique values from the input `x`
#'  - `count`: count of occurrences of `x`
#' @noRd
simple_table <- function(x) {
  if (is.factor(x)) {
    values <- levels(x)
    x_int <- as.integer(x)
  } else {
    values <- unique(x)
    x_int <- match(x, values)
  }
  list(
    x = values,
    count = tabulate(x_int, nbins = length(values))
  )
}

# modal_category ----------------------------------------------------------

test_that("modal_category works on vectors", {
  expect_equal(modal_category(NULL), NULL)
  expect_equal(modal_category(logical()), logical())
  expect_equal(modal_category(double()), double())
  expect_equal(modal_category(integer()), integer())
  expect_equal(modal_category(character()), character())
  expect_equal(modal_category(factor()), character())
  expect_equal(modal_category(ordered(NULL)), character())
  expect_equal(modal_category(NA), NA)

  expect_equal(modal_category(c(1,2.1,2.1,3,3)), 2.1)
  expect_equal(modal_category(c("a","b","b","c","c")), "b")
  expect_equal(modal_category(factor(c("a","b","b","c","c"))), "b")
})

test_that("modal_category works on rvars", {
  expect_equal(modal_category(rvar()), double())
  expect_equal(modal_category(rvar_factor()), character())
  expect_equal(modal_category(rvar_ordered()), character())

  expect_equal(modal_category(c(rvar(c(1,2.1,2.1,3,3)), rvar(1))), c(2.1, 1))
  expect_equal(modal_category(c(rvar(c("a","b","b","c","c")), rvar("c"))), c("b","c"))
})


# entropy -----------------------------------------------------------------

test_that("entropy works on vectors", {
  expect_equal(entropy(NULL), 0)
  expect_equal(entropy(1), 0)
  expect_equal(entropy(NA), NA_real_)

  x <- factor(c("a", "a", "b", "c"), levels = c("a", "b", "c", "d"))
  p <- c(0.5, 0.25, 0.25)
  # dividing by log(4), not log(3), to account for 0 in the "d" level
  expect_equal(entropy(x), -sum(p * log(p)) / log(4))
  # if we treat it as an integer, we divide by log(3) since it won't
  # know about the missing level
  expect_equal(entropy(as.integer(x)), -sum(p * log(p)) / log(3))
})

test_that("entropy works on rvars", {
  expect_equal(entropy(rvar()), numeric())

  x <- rvar(array(c("a","a","b","c", "d","d","d","d", "a","b","c","d"), dim = c(4, 3)))
  p <- c(0.5, 0.25, 0.25)
  # dividing by log(4), not log(3), to account for 0 in the "d" level
  expect_equal(entropy(x), c(-sum(p * log(p)) / log(4), 0, 1))

  # if we treat it as an integer, we divide by log(3) since it won't
  # know about the missing level
  expect_equal(entropy(as_rvar_numeric(x)), c(-sum(p * log(p)) / log(3), 0, 1))
})


# dissent -----------------------------------------------------------------

test_that("dissent works on vectors", {
  expect_equal(dissent(NULL), 0)
  expect_equal(dissent(1), 0)
  expect_equal(dissent(NA), NA_real_)

  x <- ordered(c("a", "a", "b", "c"), levels = c("a", "b", "c", "d"))
  p <- c(0.5, 0.25, 0.25)
  # dividing by 3, not 2, to account for the "d" level
  expect_equal(dissent(x), -sum(p * log2(1 - abs(1:3 - 1.75) / 3)))
  # if we treat it as an integer, we divide by 2 since it won't
  # know about the missing level at the end
  expect_equal(dissent(as.integer(x)), -sum(p * log2(1 - abs(1:3 - 1.75) / 2)))
})

test_that("dissent works on rvars", {
  expect_equal(dissent(rvar_ordered()), numeric())

  x <- rvar_ordered(array(c("a","a","b","c", "d","d","d","d", "a","a","d","d"), dim = c(4, 3)))
  p <- c(0.5, 0.25, 0.25)
  # dividing by 3, not 2, to account for the "d" level
  expect_equal(dissent(x), c(-sum(p * log2(1 - abs(1:3 - 1.75) / 3)), 0, 1))

  # if we treat it as an integer, we divide by 2 since it won't
  # know about the missing level at the end
  expect_equal(dissent(as_rvar_numeric(x)), c(-sum(p * log2(1 - abs(1:3 - 1.75) / 2)), 0, 1))
})

# Ops: math operators ---------------------------------------------------

#' @export
Ops.rvar <- function(e1, e2) {
  e1 <- as_rvar(e1)
  f <- get(.Generic)

  if (missing(e2)) {
    # unary operators
    rvar_apply_vec_fun(f, e1)
  } else {
    .Ops.rvar(f, e1, as_rvar(e2))
  }
}

.Ops.rvar <- function(f, e1, e2, preserve_dims = FALSE) {
  c(e1, e2) %<-% conform_rvar_nchains(list(e1, e2))
  draws_x <- draws_of(e1)
  draws_y <- draws_of(e2)

  # broadcast draws to common dimension
  new_dim <- dim2_common(dim(draws_x), dim(draws_y))
  # Most of the time we don't broadcast scalars (constant rvars of length 1).
  # With broadcast_scalars = FALSE broadcast_array will return a vector (no dims)
  # version of the input, which works unless *both* x and y are constants
  # (because then the correct output shape is lost; in this case we do need to
  # broadcast both x and y in case their dimensions are not equal; e.g. if x is
  # 1x1 and y is 1x1x1x1 we must broadcast both to 1x1x1x1)
  broadcast_scalars = length(draws_x) == 1 && length(draws_y) == 1
  draws_x <- broadcast_array(draws_x, new_dim, broadcast_scalars = broadcast_scalars)
  draws_y <- broadcast_array(draws_y, new_dim, broadcast_scalars = broadcast_scalars)

  draws <- f(draws_x, draws_y)

  # for factors, we must copy the dims back over after, because base-R factor
  # operations destroy dims (while base-R numeric array operations do not)
  if (preserve_dims) {
    # in the case where one of e1 or e2 is scalar and the other is not,
    # we need to source the dims from whichever is *not* scalar, which will be
    # the longer of the two. In any other case, we give preference to e1
    # (draws_x) to be consistent with base R behavior
    if (length(draws_y) > length(draws_x)) {
      dim_source <- draws_y
    } else {
      dim_source <- draws_x
    }
    draws <- while_preserving_dims(function(...) draws, dim_source)
  }

  new_rvar(draws, .nchains = nchains(e1))
}

#' @export
Ops.rvar_factor <- function(e1, e2) {
  if (missing(e2) || !(.Generic %in% c("==", "!="))) {
    stop_no_call("Cannot apply `", .Generic, "` operator to rvar_factor objects.")
  }

  .Ops.rvar_factor(.Generic, e1, e2)
}

.Ops.rvar_factor <- function(.Generic, e1, e2, ordered = FALSE) {
  # if one arg is a character vector, must convert it to a factor first, and
  # if the operation is comparison must make sure that it is a level in the
  # factor object we are comparing to (otherwise we end up with NAs)
  if (is.character(e1)) {
    if (.Generic %in% c("==", "!=")) {
      levels(e2) <- c(levels(e2), setdiff(e1, levels(e2)))
    }
    e1 <- as_rvar_factor(e1, levels = levels(e2), ordered = ordered)
  }
  if (is.character(e2)) {
    if (.Generic %in% c("==", "!=")) {
      levels(e1) <- c(levels(e1), setdiff(e2, levels(e1)))
    }
    e2 <- as_rvar_factor(e2, levels = levels(e1), ordered = ordered)
  }

  .Ops.rvar(get(.Generic), as_rvar(e1), as_rvar(e2), preserve_dims = TRUE)
}

#' @export
Ops.rvar_ordered <- function(e1, e2) {
  if (missing(e2) || !(.Generic %in% c("==", "!=", "<", ">", "<=", ">="))) {
    stop_no_call("Cannot apply `", .Generic, "` operator to rvar_ordered objects.")
  }

  .Ops.rvar_factor(.Generic, e1, e2, ordered = TRUE)
}

#' @export
Math.rvar <- function(x, ...) {
  f <- get(.Generic)

  if (.Generic %in% c("cumsum", "cumprod", "cummax", "cummin")) {
    # cumulative functions need to be handled differently
    # from other functions in this generic
    new_rvar(t(apply(draws_of(x), 1, f)), .nchains = nchains(x))
  } else {
    new_rvar(f(draws_of(x), ...), .nchains = nchains(x))
  }
}

#' @export
Math.rvar_factor <- function(x, ...) {
  stop_no_call("Cannot apply `", .Generic, "` function to rvar_factor objects.")
}

# matrix stuff ---------------------------------------------------

#' Matrix multiplication of random variables
#'
#' Matrix multiplication of random variables.
#'
#' @name rvar-matmult
#' @aliases %**%
#' @param x (multiple options) The object to be postmultiplied by `y`:
#'   * An [`rvar`]
#'   * A [`numeric`] vector or matrix
#'   * A [`logical`] vector or matrix
#'
#'   If a vector is used, it is treated as a *row* vector.
#'
#' @param y (multiple options) The object to be premultiplied by `x`:
#'   * An [`rvar`]
#'   * A [`numeric`] vector or matrix
#'   * A [`logical`] vector or matrix
#'
#'   If a vector is used, it is treated as a *column* vector.
#'
#' @details
#' If `x` or `y` are vectors, they are converted into matrices prior to multiplication, with `x`
#' converted to a row vector and `y` to a column vector. Numerics and logicals can be multiplied
#' by [`rvar`]s and are broadcasted across all draws of the [`rvar`] argument. Tensor multiplication
#' is used to efficiently multiply matrices across draws, so if either `x` or `y` is an [`rvar`],
#' `x %**% y` will be much faster than `rdo(x %*% y)`.
#'
#' Because [`rvar`] is an S3 class and S3 classes cannot properly override `%*%`, [`rvar`]s use
#' `%**%` for matrix multiplication.
#'
#' @return An [`rvar`] representing the matrix product of `x` and `y`.
#'
#' @examples
#'
#' # d has mu (mean vector of length 3) and Sigma (3x3 covariance matrix)
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' d$Sigma
#'
#' # trivial example: multiplication by a non-random matrix
#' d$Sigma %**% diag(1:3)
#'
#' # Decompose Sigma into R s.t. R'R = Sigma ...
#' R <- chol(d$Sigma)
#' # ... and recreate Sigma using matrix multiplication
#' t(R) %**% R
#'
#' @importFrom tensorA mul.tensor as.tensor
#' @export
`%**%` <- function(x, y) {
  # Fast version of rdo(x %*% y)

  # convert both objects into rvars if they aren't already (this will ensure
  # we have a 3d draws array for each variable)
  x <- as_rvar(x)
  y <- as_rvar(y)

  if (is_rvar_factor(x) || is_rvar_factor(y)) {
    stop_no_call("Cannot apply `%**%` operator to rvar_factor objects.")
  }

  # ensure everything is a matrix by adding dimensions as necessary to make `x`
  # a row vector and `y` a column vector
  ndim_x <- length(dim(x))
  if (ndim_x == 1) {
    dim(x) <- c(1, dim(x))
  } else if (ndim_x != 2) {
    stop_no_call("First argument (`x`) is not a vector or matrix, cannot matrix-multiply")
  }

  ndim_y <- length(dim(y))
  if (ndim_y == 1) {
    dim(y) <- c(dim(y), 1)
  } else if (ndim_y != 2) {
    stop_no_call("Second argument (`y`) is not a vector or matrix, cannot matrix-multiply")
  }

  # conform the draws dimension in both variables
  c(x, y) %<-% conform_rvar_ndraws_nchains(list(x, y))

  # drop the names of the dimensions (mul.tensor gets uppity if dimension names
  # are duplicated, but we don't care about that)
  x_tensor <- as.tensor(draws_of(x))
  y_tensor <- as.tensor(draws_of(y))
  names(dim(x_tensor)) <- NULL
  names(dim(y_tensor)) <- NULL

  # do a tensor multiplication equivalent of the requested matrix multiplication
  result <- unclass(mul.tensor(x_tensor, 3, y_tensor, 2, by = 1))

  # move draws dimension back to the front
  result <- aperm(result, c(3,1,2))

  # restore dimension names (as.tensor adds dummy names to dimensions)
  names(dim(result)) <- NULL
  result <- copy_dimnames(draws_of(x), 1:2, result, 1:2)
  result <- copy_dimnames(draws_of(y), 3, result, 3)

  new_rvar(result, .nchains = nchains(x))
}

#' Cholesky decomposition of random matrix
#'
#' Cholesky decomposition of an [`rvar`] containing a matrix.
#'
#' @param x (rvar) A 2-dimensional [`rvar`].
#' @param ... Additional parameters passed on to `chol.tensor()`
#'
#' @return An [`rvar`] containing the upper triangular factor of the Cholesky
#' decomposition, i.e., the matrix \eqn{R} such that \eqn{R'R = x}.
#'
#' @importFrom tensorA chol.tensor as.tensor
#' @export
chol.rvar <- function(x, ...) {
  # ensure x is a matrix
  if (length(dim(x)) != 2 || is_rvar_factor(x)) {
    stop_no_call("`x` must be a random matrix")
  }

  # must re-order draws dimension to the end, as chol.tensor expects it there
  x_tensor <- as.tensor(aperm(draws_of(x), c(2,3,1)))

  # do the cholesky decomp
  result <- unclass(chol.tensor(x_tensor, 1, 2, ...))

  # move draws dimension back to the front
  result <- aperm(result, c(3,1,2))

  # drop dimension names (chol.tensor screws them around)
  names(dim(result)) <- NULL

  new_rvar(result, .nchains = nchains(x))
}

#' @importFrom methods setGeneric
#' @export
setGeneric("diag")

#' Matrix diagonals (including for random variables)
#'
#' Extract the diagonal of a matrix or construct a matrix, including random
#' matrices (2-dimensional [`rvar`]s). Makes [`base::diag()`] generic.
#'
#' @inheritParams base::diag
#' @param x (numeric,rvar) a matrix, vector, 1D array, missing, or a 1- or
#' 2-dimensional [`rvar`].
#'
#' @details
#' Makes [`base::diag()`] into a generic function. See that function's documentation
#' for usage with [`numeric`]s and for usage of [`diag<-`], which is also supported
#' by [`rvar`].
#'
#' @return
#'
#' For [`rvar`]s, has two modes:
#'
#' 1. `x` is a matrix-like [`rvar`]: it returns the diagonal as a vector-like [`rvar`]
#' 2. `x` is a vector-like [`rvar`]: it returns a matrix-like [`rvar`] with `x` as
#' the diagonal and zero for off-diagonal entries.
#'
#' @seealso [`base::diag()`]
#'
#' @examples
#'
#' # Sigma is a 3x3 covariance matrix
#' Sigma <- as_draws_rvars(example_draws("multi_normal"))$Sigma
#' Sigma
#'
#' diag(Sigma)
#'
#' diag(Sigma) <- 1:3
#' Sigma
#'
#' diag(as_rvar(1:3))
#'
#' @importFrom methods setMethod callNextMethod
#' @export
setMethod("diag", signature(x = "rvar"), function(x = 1, nrow, ncol, names = TRUE) {
  if (length(dim(x)) > 1) {
    # base implementation of diag() works on rvars except when x is a vector
    callNextMethod()
  } else {
    if (missing(nrow)) {
      nrow <- length(x)
    }
    if (missing(ncol)) {
      ncol <- nrow
    }
    out <- as_rvar(matrix(rep(0, nrow * ncol), nrow = nrow, ncol = ncol))
    n <- min(nrow, ncol)
    x <- rep_len(x, n)
    i <- seq_len(n)
    out[cbind(i, i)] <- x
    out
  }
})


# transpose and permutation -----------------------------------------------

#' @export
t.rvar = function(x) {
  .draws = draws_of(x)
  ndim = length(dim(.draws))

  if (length(x) != 0 && ndim == 2) {
    # vector
    .dimnames = dimnames(.draws)
    dim(.draws) = c(dim(.draws)[1], 1, dim(.draws)[2])
    dimnames(.draws) = c(.dimnames[1], list(NULL), .dimnames[2])
    result <- new_rvar(.draws, .nchains = nchains(x))
  } else if (ndim == 3) {
    .draws <- while_preserving_levels(aperm, .draws, c(1, 3, 2))
    result <- new_rvar(.draws, .nchains = nchains(x))
  } else {
    stop_no_call("argument is not a random vector or matrix")
  }
  result
}

#' @export
aperm.rvar = function(a, perm, ...) {
  draws_of(a) <- while_preserving_levels(aperm, draws_of(a), c(1, perm + 1), ...)
  a
}

#' Order `draws` objects
#'
#' Order [`draws`] objects according to iteration and chain number. By default,
#' draws objects are ordered but subsetting or extracting parts of them may
#' leave them in an unordered state.
#'
#' @template args-methods-x
#' @template args-methods-dots
#' @template return-draws
#' @seealso [repair_draws()]
#' @examples
#' x <- as_draws_array(example_draws())
#' dimnames(x[10:5, 4:3, ])
#' dimnames(order_draws(x[10:5, 4:3, ]))
#'
#' @export
order_draws <- function(x, ...) {
  UseMethod("order_draws")
}

#' @rdname order_draws
#' @export
order_draws.draws_matrix <- function(x, ...) {
  draw_order <- order(draw_ids(x))
  if (needs_ordering(draw_order)) {
    x <- x[draw_order, ]
  }
  x
}

#' @rdname order_draws
#' @export
order_draws.draws_array <- function(x, ...) {
  iteration_order <- order(iteration_ids(x))
  chain_order <- order(chain_ids(x))
  if (needs_ordering(iteration_order, chain_order)) {
    x <- x[iteration_order, chain_order, ]
  }
  x
}

#' @rdname order_draws
#' @export
order_draws.draws_df <- function(x, ...) {
  row_order <- order(x$.chain, x$.iteration)
  if (needs_ordering(row_order)) {
    x <- x[row_order, ]
  }
  x
}

#' @rdname order_draws
#' @export
order_draws.draws_list <- function(x, ...) {
  chain_order <- order(chain_ids(x))
  if (needs_ordering(chain_order)) {
    x <- x[chain_order]
  }
  x
}

#' @rdname order_draws
#' @export
order_draws.draws_rvars <- function(x, ...) {
  for (i in seq_along(x)) {
    x[[i]] <- order_draws(x[[i]])
  }
  x
}

#' @rdname order_draws
#' @importFrom rlang missing_arg
#' @importFrom vctrs vec_slice
#' @export
order_draws.rvar <- function(x, ...) {
  draw_order <- order(draw_ids(x))
  if (needs_ordering(draw_order)) {
    # if ordering is needed, must also merge chains (as out-of-order draws
    # imply chain information is no longer meaningful)
    if (nchains(x) > 1) {
      x <- merge_chains(x)
    }
    draws_of(x) <- vec_slice(draws_of(x), draw_order)
  }
  x
}


# internal ----------------------------------------------------------------

#' Order draws if told to do so
#' @param x draws object to be ordered
#' @param order should the draws object be ordered?
#' @return potentially ordered draws object
#' @noRd
do_ordering <- function(x, order, ...) {
  order <- as_one_logical(order)
  if (order) {
    x <- order_draws(x, ...)
  }
  x
}

# are vectors in an unordered state?
needs_ordering <- function(...) {
  dots <- list(...)
  .needs_ordering <- function(x) {
    any(x != seq_along(x))
  }
  any(ulapply(dots, .needs_ordering))
}

#' Print `draws_matrix` objects
#'
#' Pretty printing for [`draws_matrix`] objects.
#'
#' @template args-methods-x
#' @template args-print-digits
#' @template args-print-max_draws
#' @template args-print-max_variables
#' @template args-methods-reserved
#' @template args-print-dots
#' @template return-draws
#'
#' @examples
#' x <- as_draws_matrix(example_draws())
#' print(x)
#'
#' @export
print.draws_matrix <- function(x, digits = 2,
                               max_draws = getOption("posterior.max_draws", 10),
                               max_variables = getOption("posterior.max_variables", 8),
                               reserved = FALSE, ...) {
  max_draws <- as_one_integer(max_draws)
  max_variables <- as_one_integer(max_variables)
  reserved <- as_one_logical(reserved)
  niterations <- niterations(x)
  nchains <- nchains(x)
  ndraws <- ndraws(x)
  nvariables <- nvariables(x)
  header <- paste0(
    "# A draws_matrix: ", niterations, " iterations, ",
    nchains, " chains, and ", nvariables, " variables\n"
  )
  cat(header)
  sel_draws <- seq_len(min(max_draws, ndraws))
  sel_variables <- seq_len(min(max_variables, nvariables))
  y <- x
  if (!reserved) {
    y <- remove_reserved_variables(y)
  }
  y <- .subset_draws(
    y, draw = sel_draws, variable = sel_variables,
    reserved = reserved
  )
  class(y) <- "matrix"
  print(y, digits = digits, ...)
  more_iterations <- ndraws -  max_draws
  more_variables <- nvariables - max_variables
  if (more_iterations > 0 || more_variables > 0) {
    comment <- character(0)
    if (more_iterations > 0) {
      comment <- c(comment, paste0(more_iterations, " more draws"))
    }
    if (more_variables > 0) {
      comment <- c(comment, paste0(more_variables, " more variables"))
    }
    comment <- paste0(comment, collapse = ", and ")
    comment <- paste0("# ... with ", comment, "\n")
    cat(comment)
  }
  reserved_variables <- reserved_variables(x)
  if (!reserved && length(reserved_variables)) {
    cat0("# ... hidden reserved variables ", comma(reserved_variables), "\n")
  }
  invisible(x)
}

#' Print `draws_array` objects
#'
#' Pretty printing for [`draws_array`] objects.
#'
#' @template args-methods-x
#' @template args-print-digits
#' @template args-print-max_iterations
#' @template args-print-max_chains
#' @template args-print-max_variables
#' @template args-methods-reserved
#' @template args-print-dots
#' @template return-draws
#'
#' @examples
#' x <- as_draws_array(example_draws())
#' print(x)
#'
#' @export
print.draws_array <- function(x, digits = 2,
                              max_iterations = getOption("posterior.max_iterations", 5),
                              max_chains = getOption("posterior.max_chains", 8),
                              max_variables = getOption("posterior.max_variables", 4),
                              reserved = FALSE, ...) {
  max_iterations <- as_one_integer(max_iterations)
  max_chains <- as_one_integer(max_chains)
  max_variables <- as_one_integer(max_variables)
  reserved <- as_one_logical(reserved)
  niterations <- niterations(x)
  nchains <- nchains(x)
  nvariables <- nvariables(x)
  header <- paste0(
    "# A draws_array: ", niterations, " iterations, ",
    nchains, " chains, and ", nvariables, " variables\n"
  )
  cat(header)
  sel_iterations <- seq_len(min(max_iterations, niterations))
  sel_chains <- seq_len(min(max_chains, nchains))
  sel_variables <- seq_len(min(max_variables, nvariables))
  y <- x
  if (!reserved) {
    y <- remove_reserved_variables(y)
  }
  y <- .subset_draws(
    y, sel_iterations, sel_chains, sel_variables,
    reserved = reserved
  )
  class(y) <- "array"
  print(y, digits = digits, ...)
  more_iterations <- niterations - max_iterations
  more_chains <- nchains - max_chains
  more_variables <- nvariables - max_variables
  if (more_iterations > 0 || more_chains > 0 || more_variables > 0) {
    comment <- character(0)
    if (more_iterations > 0) {
      comment <- c(comment, paste0(more_iterations, " more iterations"))
    }
    if (more_chains > 0) {
      comment <- c(comment, paste0(more_chains, " more chains"))
    }
    if (more_variables > 0) {
      comment <- c(comment, paste0(more_variables, " more variables"))
    }
    comment <- paste0(comment, collapse = ", and ")
    comment <- paste0("# ... with ", comment, "\n")
    cat(comment)
  }
  reserved_variables <- reserved_variables(x)
  if (!reserved && length(reserved_variables)) {
    cat0("# ... hidden reserved variables ", comma(reserved_variables), "\n")
  }
  invisible(x)
}

#' Print `draws_df` objects
#'
#' Pretty printing for [`draws_df`] objects.
#'
#' @template args-methods-x
#' @template args-print-digits
#' @template args-print-max_draws
#' @template args-print-max_variables
#' @template args-methods-reserved
#' @template args-print-dots
#' @template return-draws
#'
#' @examples
#' x <- as_draws_df(example_draws())
#' print(x)
#'
#' @export
print.draws_df <- function(x, digits = 2,
                           max_draws = getOption("posterior.max_draws", 10),
                           max_variables = getOption("posterior.max_variables", 8),
                           reserved = FALSE, ...) {
  max_draws <- as_one_integer(max_draws)
  max_variables <- as_one_integer(max_variables)
  reserved <- as_one_logical(reserved)
  niterations <- niterations(x)
  nchains <- nchains(x)
  ndraws <- ndraws(x)
  nvariables <- nvariables(x, reserved = reserved)
  header <- paste0(
    "# A draws_df: ", niterations, " iterations, ",
    nchains, " chains, and ", nvariables, " variables\n"
  )
  cat(header)
  sel_draws <- seq_len(min(max_draws, ndraws))
  sel_variables <- variables(x, reserved = reserved)
  seq_variables <- seq_len(min(max_variables, nvariables))
  sel_variables <- sel_variables[seq_variables]
  y <- .subset_draws(
    x[sel_draws,], variable = sel_variables,
    reserved = reserved
  )
  if (!reserved) {
    # reserved df variables can only be removed after subsetting
    y <- remove_reserved_df_variables(y)
  }
  class(y) <- "data.frame"
  print(y, digits = digits, ...)
  more_iterations <- ndraws - max_draws
  more_variables <- nvariables - max_variables
  if (more_iterations > 0 || more_variables > 0) {
    comment <- character(0)
    if (more_iterations > 0) {
      comment <- c(comment, paste0(more_iterations, " more draws"))
    }
    if (more_variables > 0) {
      comment <- c(comment, paste0(more_variables, " more variables"))
    }
    comment <- paste0(comment, collapse = ", and ")
    comment <- paste0("# ... with ", comment, "\n")
    cat(comment)
  }
  reserved_variables <- all_reserved_variables(x)
  if (!reserved && length(reserved_variables)) {
    cat0("# ... hidden reserved variables ", comma(reserved_variables), "\n")
  }
  invisible(x)
}

#' Print `draws_list` objects
#'
#' Pretty printing for [`draws_list`] objects.
#'
#' @template args-methods-x
#' @template args-print-digits
#' @template args-print-max_iterations
#' @template args-print-max_chains
#' @template args-print-max_variables
#' @template args-methods-reserved
#' @template args-print-dots
#' @template return-draws
#'
#' @examples
#' x <- as_draws_list(example_draws())
#' print(x)
#'
#' @export
print.draws_list <- function(x, digits = 2,
                             max_iterations = getOption("posterior.max_iterations", 10),
                             max_chains = getOption("posterior.max_chains", 2),
                             max_variables = getOption("posterior.max_variables", 4),
                             reserved = FALSE, ...) {
  max_iterations <- as_one_integer(max_iterations)
  max_chains <- as_one_integer(max_chains)
  max_variables <- as_one_integer(max_variables)
  reserved <- as_one_logical(reserved)
  niterations <- niterations(x)
  nchains <- nchains(x)
  nvariables <- nvariables(x)
  header <- paste0(
    "# A draws_list: ", niterations, " iterations, ",
    nchains, " chains, and ", nvariables, " variables\n"
  )
  cat(header)
  sel_iterations <- seq_len(min(max_iterations, niterations))
  sel_chains <- seq_len(min(max_chains, nchains))
  sel_variables <- seq_len(min(max_variables, nvariables))
  y <- x
  if (!reserved) {
    y <- remove_reserved_variables(y)
  }
  y <- .subset_draws(
    y, sel_iterations, sel_chains, sel_variables,
    reserved = reserved
  )
  for (i in seq_along(y)) {
    cat0("\n[chain = ", i, "]\n")
    print(y[[i]], digits = digits, ...)
  }
  more_iterations <- niterations - max_iterations
  more_chains <- nchains - max_chains
  more_variables <- nvariables - max_variables
  if (more_iterations > 0 || more_chains > 0 || more_variables > 0) {
    comment <- character(0)
    if (more_iterations > 0) {
      comment <- c(comment, paste0(more_iterations, " more iterations"))
    }
    if (more_chains > 0) {
      comment <- c(comment, paste0(more_chains, " more chains"))
    }
    if (more_variables > 0) {
      comment <- c(comment, paste0(more_variables, " more variables"))
    }
    comment <- paste0(comment, collapse = ", and ")
    comment <- paste0("# ... with ", comment, "\n")
    cat(comment)
  }
  reserved_variables <- reserved_variables(x)
  if (!reserved && length(reserved_variables)) {
    cat0("# ... hidden reserved variables ", comma(reserved_variables), "\n")
  }
  invisible(x)
}

#' Print `draws_rvars` objects
#'
#' Pretty printing for [`draws_rvars`] objects.
#'
#' @encoding UTF-8
#' @template args-methods-x
#' @template args-print-digits
#' @template args-print-max_variables
#' @template args-print-summary
#' @template args-methods-reserved
#' @template args-print-dots
#' @template return-draws
#'
#' @examples
#' x <- as_draws_rvars(example_draws())
#' print(x)
#'
#' @export
print.draws_rvars <- function(x,
  digits = 2,
  max_variables = getOption("posterior.max_variables", 8),
  summary = getOption("posterior.rvar_summary", "mean_sd"),
  reserved = FALSE,
  ...
) {
  max_variables <- as_one_integer(max_variables)
  niterations <- niterations(x)
  nchains <- nchains(x)
  nvariables <- nvariables(x)
  header <- paste0(
    "# A draws_rvars: ", niterations, " iterations, ",
    nchains, " chains, and ", nvariables, " variables\n"
  )
  cat(header)

  sel_variables <- seq_len(min(max_variables, nvariables))
  y <- x
  if (!reserved) {
    y <- remove_reserved_variables(y)
  }
  y <- .subset_draws(y, variable = sel_variables, reserved = reserved)
  for (i in seq_along(y)) {
    cat0("$", names(y)[[i]], ": ")
    print(y[[i]], summary = summary, digits = digits, ...)
    cat("\n")
  }

  more_variables <- nvariables - max_variables
  if (more_variables > 0) {
    comment <- paste0(more_variables, " more variables")
    comment <- paste0("# ... with ", comment, "\n")
    cat(comment)
  }
  reserved_variables <- reserved_variables(x)
  if (!reserved && length(reserved_variables)) {
    cat0("# ... hidden reserved variables ", comma(reserved_variables), "\n")
  }

  invisible(x)
}

#' Merge chains of `draws` objects
#'
#' Merge chains of [`draws`] objects into a single chain. Some operations will
#' trigger an automatic merging of chains, for example, because chains do not
#' match between two objects involved in a binary operation. By default, no
#' warning will be issued when this happens but you can activate one via
#' `options(posterior.warn_on_merge_chains = TRUE)`.
#'
#' @template args-methods-x
#' @template args-methods-dots
#' @template return-draws
#'
#' @examples
#' x <- example_draws()
#'
#' # draws_array with 4 chains, 100 iters each
#' str(x)
#'
#' # draws_array with 1 chain of 400 iterations
#' str(merge_chains(x))
#'
#' @export
merge_chains <- function(x, ...) {
  UseMethod("merge_chains")
}

#' @rdname merge_chains
#' @export
merge_chains.draws_matrix <- function(x, ...) {
  attr(x, "nchains") <- 1L
  x
}

#' @rdname merge_chains
#' @export
merge_chains.draws_array <- function(x, ...) {
  x <- merge_chains(as_draws_matrix(x))
  as_draws_array(x)
}

#' @rdname merge_chains
#' @export
merge_chains.draws_df <- function(x, ...) {
  x$.chain <- rep(1L, nrow(x))
  x$.iteration <- x$.draw
  x
}

#' @rdname merge_chains
#' @export
merge_chains.draws_list <- function(x, ...) {
  if (nchains(x) == 0) {
    return(x)
  }
  out <- empty_draws_list(variables(x), nchains = 1)
  for (v in variables(out)) {
    out[[1]][[v]] <- unname(ulapply(x, "[[", v))
  }
  out
}

#' @rdname merge_chains
#' @export
merge_chains.rvar <- function(x, ...) {
  nchains_rvar(x) <- 1L
  x
}

#' @rdname merge_chains
#' @export
merge_chains.draws_rvars <- function(x, ...) {
  for (i in seq_along(x)) {
    x[[i]] <- merge_chains(x[[i]])
  }
  x
}

# some operations lead to an automatic chain merge
# that users can choose to be warned about
warn_merge_chains <- function(type = c("match", "index")) {
  warn <- as_one_logical(getOption(
    "posterior.warn_on_merge_chains",
    default = FALSE
  ))
  if (warn) {
    type <- as_one_character(type)
    warning_no_call(
      "Chains were dropped",
      switch(type, ".",
        match = " due to chain information not matching.",
        index = " due to manually indexing draws."
      )
    )
  }
  invisible(NULL)
}

.onAttach <- function(...) {
  ver <- utils::packageVersion("posterior")
  packageStartupMessage("This is posterior version ", ver)
  # packageStartupMessage("- Online documentation and vignettes at mc-stan.org/posterior")
}

.onLoad <- function(...) {
  # S3 generics for packages in Suggests, for compatibility with R < 3.6.
  # See help("s3_register", package = "vctrs") for more information.
  vctrs::s3_register("dplyr::dplyr_reconstruct", "draws_df")
  vctrs::s3_register("ggplot2::scale_type", "rvar")

  # S3 methods for matrixOps, which is a group generic that didn't exist
  # until R 4.3, so we can't register it in NAMESPACE
  if (getRversion() >= "4.3") {
    registerS3method("matrixOps", "rvar", matrixOps.rvar, asNamespace("base"))
  }
}

#' Loop over draws
#'
#' Executes an expression once for every draw in a `draws` object. Used
#' primarily for its side effects and returns the input `x` invisibly.
#'
#' @template args-methods-x
#' @param expr (expression) A bare expression that can contain references to
#' variables in `x` by name. This expression will be executed once per draw
#' of `x`, where references to variables in `x` resolve to the value of that
#' variable in that draw. The expression supports [quasiquotation].
#'
#' @details
#' If `x` is not in the [`draws_rvars`] format, it is first converted to that
#' format. This allows the variables in `x` to include their dimensions (i.e,
#' to act as R vectors and arrays) when being referred to in `expr`.
#'
#' Within `expr`, use `.draw` to refer to the draw index, which will be a value
#' between 1 and `ndraws(x)`. `expr` is executed in the calling environment of
#' `for_each_draw()`, so it can use variables in that environment (however, due
#' to the use of data masking, to modify variables in that environment, one
#' must use `<<-`.)
#'
#' @returns
#' As `for_each_draw()` is used primarily for its side effects (the expression
#' executed for each draw of `x`), it returns the input `x` invisibly.
#'
#' @examples
#' eight_schools <- as_draws_rvars(example_draws())
#'
#'
#' # 1. A simple example --- looping over draws and printing each draw
#' # NOTE: You probably don't want to do this in practice! This example is
#' # just intended to show what for_each_draw() is doing. If you just want to
#' # print the draws of an rvar, it is probably better to use draws_of()
#' for_each_draw(eight_schools, {
#'   print(mu)
#' })
#'
#'
#' # 2. A more complex example --- building a parallel coordinates plot
#' # First, construct the plot bounds
#' plot(1, type = "n",
#'   xlim = c(1, length(eight_schools$theta)),
#'   ylim = range(range(eight_schools$theta)),
#'   xlab = "school", ylab = "theta"
#' )
#'
#' # Then, use for_each_draw() to make a parallel coordinates plot of all draws
#' # of eight_schools$theta. Use resample_draws(eight_schools, n = ...)
#' # in place of eight_schools if a smaller sample is desired for the plot.
#' for_each_draw(eight_schools, {
#'   lines(seq_along(theta), theta, col = rgb(1, 0, 0, 0.05))
#' })
#'
#' # Finally, add means and 90% intervals
#' lines(seq_along(eight_schools$theta), mean(eight_schools$theta))
#' with(summarise_draws(eight_schools$theta),
#'   segments(seq_along(eight_schools$theta), y0 = q5, y1 = q95)
#' )
#' @importFrom rlang eval_tidy
#' @export
for_each_draw <- function(x, expr) {
  x_rvars <- as_draws_rvars(x)
  expr <- enquo(expr)
  env <- parent.frame()

  for (i in seq_len(ndraws(x_rvars))) {
    variables <- get_variables_from_one_draw(x_rvars, i)
    variables$.draw = i
    eval_tidy(expr, data = variables, env = env)
  }

  invisible(x)
}

#' Extract matrix of a single variable
#'
#' Extract an iterations x chains matrix of draws of a single variable.
#' This is primarily used for convergence diagnostic functions such as [rhat()].
#'
#' @template args-methods-x
#' @template args-extract-variable
#' @template args-methods-dots
#' @return A `matrix` with dimension iterations x chains.
#' @family variable extraction methods
#' @examples
#' x <- example_draws()
#' mu <- extract_variable_matrix(x, variable = "mu")
#' dim(mu)
#' rhat(mu)
#'
#' @export
extract_variable_matrix <- function(x, variable, ...) {
  UseMethod("extract_variable_matrix")
}

#' @rdname extract_variable_matrix
#' @export
extract_variable_matrix.default <- function(x, variable, ...) {
  x <- as_draws(x)
  extract_variable_matrix(x, variable, ...)
}

#' @rdname extract_variable_matrix
#' @export
extract_variable_matrix.draws <- function(x, variable, ...) {
  variable <- as_one_character(variable)
  out <- .subset_draws(x, variable = variable, reserved = FALSE)
  out <- as_draws_array(out)
  out <- drop_dims_or_classes(out, dims = 3, reset_class = TRUE)
  class(out) <- "matrix"
  out
}

#' @rdname extract_variable_matrix
#' @export
extract_variable_matrix.draws_df <- function(x, variable, ...) {
  variable <- as_one_character(variable)
  if (is.factor(x[[variable]])) {
    x_variable_factor <- x[[variable]]
    x[[variable]] <- unclass(x[[variable]])
    out <- copy_levels(x_variable_factor, NextMethod())
  } else {
    out <- NextMethod()
  }
  out
}

#' @rdname extract_variable_matrix
#' @export
extract_variable_matrix.draws_list <- function(x, variable, ...) {
  variable <- as_one_character(variable)
  if (is.factor(x[[1]][[variable]])) {
    x_variable_factor <- x[[1]][[variable]]
    for (i in seq_along(x)) {
      x[[i]][[variable]] <- unclass(x[[i]][[variable]])
    }
    out <- copy_levels(x_variable_factor, NextMethod())
  } else {
    out <- NextMethod()
  }
  out
}

#' @rdname extract_variable_matrix
#' @export
extract_variable_matrix.draws_rvars <- function(x, variable, ...) {
  variable <- as_one_character(variable)
  parts <- split_variable_names(variable)

  .draws <- draws_of(x[[parts$base_name]])
  if (is.factor(.draws)) {
    # if x is a factor rvar, convert it to numeric before extracting and
    # then we'll add levels back at the end
    x[[parts$base_name]] <- as_rvar_integer(x[[parts$base_name]])
  }

  if (isTRUE(nzchar(parts$indices))) {
    # variable with indices in the name ("x[1]", etc) can't be subset from
    # draws_rvars directly, so we'll convert to a draws_array first
    out <- extract_variable_matrix(as_draws_array(x[parts$base_name]), variable = variable, ...)
  } else if (length(x[[variable]]) > 1) {
    stop_no_call(
      'Cannot extract non-scalar value using extract_variable_matrix():\n',
      '  "', variable, '" has dimensions: [', paste0(dim(x[[variable]]), collapse = ","), ']\n',
      '  Try including brackets ("[]") and indices in the variable name to extract a scalar value.'
    )
  } else {
    out <- NextMethod()
  }

  copy_levels(.draws, out)
}

# Summaries over array elements, within draws ---------------------------------------------------------

#' Summaries of random variables over array elements, within draws
#'
#' Compute summaries of random variables over array elements and within draws,
#' producing a new random variable of length 1 (except in the case of
#' `rvar_range()`, see **Details**).
#'
#' @param ... (rvar) One or more [`rvar`]s.
#' @template args-rvar-summaries-na.rm
#' @param constant (scalar real) For `rvar_mad()`, a scale factor for computing
#'   the median absolute deviation. See the details of `stats::mad()` for the
#'   justification for the default value.
#' @param probs (numeric vector) For `rvar_quantile()`, probabilities in `[0, 1]`.
#' @param names (logical) For `rvar_quantile()`, if `TRUE`, the result has a
#'   `names` attribute.
#'
#' @details
#'
#' These functions compute statistics within each draw of the random variable.
#' For summaries over draws (such as expectations), see [rvar-summaries-over-draws].
#'
#' Each function defined here corresponds to the base function of the same name
#' without the `rvar_` prefix (e.g., `rvar_mean()` calls `mean()` under the hood, etc).
#'
#' @return
#' An [`rvar`] of length 1 (for `range()`, length 2; for `quantile()`, length
#' equal to `length(probs)`) with the same number
#' of draws as the input rvar(s) containing the summary statistic computed within
#' each draw of the input rvar(s).
#'
#' @examples
#'
#' set.seed(5678)
#' x = rvar_rng(rnorm, 4, mean = 1:4, sd = 2)
#'
#' # These will give similar results to mean(1:4),
#' # median(1:4), sum(1:4), prod(1:4), etc
#' rvar_mean(x)
#' rvar_median(x)
#' rvar_sum(x)
#' rvar_prod(x)
#' rvar_range(x)
#' rvar_quantile(x, probs = c(0.25, 0.5, 0.75), names = TRUE)
#'
#' @seealso [rvar-summaries-over-draws] for summary functions across draws (e.g. expectations).
#' [rvar-dist] for density, CDF, and quantile functions of random variables.
#' @family rvar-summaries
#' @name rvar-summaries-within-draws
#' @export
rvar_mean <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_mean", matrixStats::rowMeans2, na.rm = na.rm
  )
}

# numeric summaries -------------------------------------------------------

#' @rdname rvar-summaries-within-draws
#' @export
rvar_median <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_median", matrixStats::rowMedians, na.rm = na.rm, .ordered_okay = TRUE
  )
}

#' @rdname rvar-summaries-within-draws
#' @export
rvar_sum <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_sum", matrixStats::rowSums2, na.rm = na.rm
  )
}

#' @rdname rvar-summaries-within-draws
#' @export
rvar_prod <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_prod", matrixStats::rowProds, na.rm = na.rm
  )
}

#' @rdname rvar-summaries-within-draws
#' @export
rvar_min <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_min", matrixStats::rowMins, na.rm = na.rm, .ordered_okay = TRUE
  )
}

#' @rdname rvar-summaries-within-draws
#' @export
rvar_max <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_max", matrixStats::rowMaxs, na.rm = na.rm, .ordered_okay = TRUE
  )
}


# spread ------------------------------------------------------------------

#' @rdname rvar-summaries-within-draws
#' @export
rvar_sd <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_sd", matrixStats::rowSds, na.rm = na.rm
  )
}

#' @rdname rvar-summaries-within-draws
#' @export
rvar_var <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_var", matrixStats::rowVars, na.rm = na.rm
  )
}

#' @rdname rvar-summaries-within-draws
#' @export
rvar_mad <- function(..., constant = 1.4826, na.rm = FALSE) {
  x <- c(...)
  if (is_rvar_ordered(x)) {
    x <- as_rvar_numeric(x)
  }
  summarise_rvar_within_draws_via_matrix(
    x, "rvar_mad", matrixStats::rowMads, constant = constant, na.rm = na.rm
  )
}


# range -------------------------------------------------------------------

#' @rdname rvar-summaries-within-draws
#' @export
rvar_range <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_range", matrixStats::rowRanges, na.rm = na.rm, .ordered_okay = TRUE
  )
}


# quantiles ---------------------------------------------------------------

#' @rdname rvar-summaries-within-draws
#' @export
rvar_quantile <- function(..., probs, names = FALSE, na.rm = FALSE) {
  names <- as_one_logical(names)
  na.rm <- as_one_logical(na.rm)
  x <- c(...)

  # use type 1 for ordered rvar because it is the inverse of the ECDF of
  # a discrete distribution (hence appropriate for ordinal samples), otherwise
  # use type 7 (the default), which is intended for continuous samples
  type <- if (is_rvar_ordered(x)) 1 else 7

  out <- summarise_rvar_within_draws_via_matrix(
    x, "rvar_quantile", matrixStats::rowQuantiles, probs = probs, type = type,
    na.rm = na.rm, drop = FALSE, .ordered_okay = TRUE
  )

  if (!names) {
    dimnames(out) <- NULL
  }

  out
}


# logical summaries -------------------------------------------------------

#' @rdname rvar-summaries-within-draws
#' @export
rvar_all <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_all", matrixStats::rowAlls, na.rm = na.rm
  )
}

#' @rdname rvar-summaries-within-draws
#' @export
rvar_any <- function(..., na.rm = FALSE) {
  summarise_rvar_within_draws_via_matrix(
    c(...), "rvar_any", matrixStats::rowAnys, na.rm = na.rm
  )
}


# special value predicates ------------------------------------------------

#' Special value predicates for random variables
#'
#' Compute special value predicates (checking for finite / infinite values, `NaN`, and `NA`)
#' on all draws within a random variable, returning a random variable.
#'
#' @param x (rvar) An [`rvar`].
#'
#' @details
#'
#' These functions return a new [`rvar`] that is the result of applying
#' `is.finite()`, `is.infinite()`, `is.nan()`, or `is.na()` to every draw
#' in the input random variable.
#'
#' @return
#' A logical [`rvar`] of the same length as the input.
#'
#' @examples
#'
#' x <- rvar(c(1, Inf, -Inf, NaN, NA))
#' x
#'
#' rvar_is_finite(x)
#' rvar_is_infinite(x)
#' rvar_is_nan(x)
#' rvar_is_na(x)
#'
#' @seealso [rvar-summaries-over-draws] for summary functions across draws, including
#' implementations of `is.finite()`, `is.infinite()`, `is.nan()`, and `is.na()` for
#' `rvar`s.
#' @family rvar-summaries
#' @name rvar_is_finite
#' @export
rvar_is_finite <- function(x) rvar_apply_vec_fun(is.finite, x)

#' @rdname rvar_is_finite
#' @export
rvar_is_infinite <- function(x) rvar_apply_vec_fun(is.infinite, x)

#' @rdname rvar_is_finite
#' @export
rvar_is_nan <- function(x) rvar_apply_vec_fun(is.nan, x)

#' @rdname rvar_is_finite
#' @export
rvar_is_na <- function(x) rvar_apply_vec_fun(is.na, x)

# numeric summaries -------------------------------------------------------

test_that("numeric summary functions work", {
  x_array <- array(1:24, dim = c(4,2,3))
  x <- new_rvar(x_array)
  x_letters <- array(letters[1:24], dim = c(4,2,3))
  x_ord <- rvar_ordered(x_letters, levels = letters)
  x_fct <- rvar_factor(x_letters, levels = letters)

  expect_equal(draws_of(rvar_mean(x)), apply(x_array, 1, mean), check.attributes = FALSE)
  expect_equal(draws_of(rvar_median(x)), apply(x_array, 1, median), check.attributes = FALSE)
  expect_equal(draws_of(rvar_sum(x)), apply(x_array, 1, sum), check.attributes = FALSE)
  expect_equal(draws_of(rvar_prod(x)), apply(x_array, 1, prod), check.attributes = FALSE)
  expect_equal(draws_of(rvar_min(x)), apply(x_array, 1, min), check.attributes = FALSE)
  expect_equal(draws_of(rvar_max(x)), apply(x_array, 1, max), check.attributes = FALSE)

  expect_error(rvar_mean(x_ord))
  expect_equal(rvar_median(x_ord), rvar_ordered(letters[apply(x_array, 1, median)], levels = letters))
  expect_error(rvar_sum(x_ord))
  expect_error(rvar_prod(x_ord))
  expect_equal(rvar_min(x_ord), rvar_ordered(letters[apply(x_array, 1, min)], levels = letters))
  expect_equal(rvar_max(x_ord), rvar_ordered(letters[apply(x_array, 1, max)], levels = letters))

  expect_error(rvar_mean(x_fct))
  expect_error(rvar_median(x_fct))
  expect_error(rvar_sum(x_fct))
  expect_error(rvar_prod(x_fct))
  expect_error(rvar_min(x_fct))
  expect_error(rvar_max(x_fct))

  # default values on empty input
  expect_equal(rvar_mean(), as_rvar(NA_real_))
  expect_equal(rvar_median(), as_rvar(NA_real_))
  expect_equal(rvar_sum(), as_rvar(0))
  expect_equal(rvar_prod(), as_rvar(1))
  expect_equal(rvar_min(), as_rvar(Inf))
  expect_equal(rvar_max(), as_rvar(-Inf))

  # test argument passing
  x[1,2] <- NA
  expect_equal(
    draws_of(rvar_mean(x, na.rm = TRUE)),
    apply(draws_of(x), 1, function(x) mean(x, na.rm = TRUE)),
    check.attributes = FALSE
  )
})


# spread ------------------------------------------------------------------

test_that("spread summary functions work", {
  x_array <- array(1:24, dim = c(4,2,3))
  x <- new_rvar(x_array)
  x_letters <- array(letters[1:24], dim = c(4,2,3))
  x_ord <- rvar_ordered(x_letters, levels = letters)
  x_fct <- rvar_factor(x_letters, levels = letters)

  expect_equal(draws_of(rvar_sd(x)), apply(x_array, 1, sd), check.attributes = FALSE)
  expect_equal(draws_of(rvar_var(x)), apply(x_array, 1, function(x) var(as.vector(x))), check.attributes = FALSE)
  expect_equal(draws_of(rvar_mad(x)), apply(x_array, 1, mad), check.attributes = FALSE)
  expect_equal(draws_of(rvar_mad(x, constant = 1)), apply(x_array, 1, mad, constant = 1), check.attributes = FALSE)

  expect_error(rvar_sd(x_ord))
  expect_error(rvar_var(x_ord))
  expect_equal(rvar_mad(x_ord, constant = 1), rvar(apply(x_array, 1, mad, constant = 1)))

  expect_error(rvar_sd(x_fct))
  expect_error(rvar_var(x_fct))
  expect_error(rvar_mad(x_fct))

  # default values on empty input
  expect_equal(rvar_sd(), as_rvar(NA_real_))
  expect_equal(rvar_var(), as_rvar(NA_real_))
  expect_equal(rvar_mad(), as_rvar(NA_real_))

  # test argument passing on var since it requires some finagling
  x[1,2] <- NA
  expect_equal(
    draws_of(rvar_var(x, na.rm = TRUE)),
    apply(draws_of(x), 1, function(x) var(as.vector(x), na.rm = TRUE)),
    check.attributes = FALSE
  )
})


# range -------------------------------------------------------------------

test_that("rvar_range works", {
  x_array <- array(1:24, dim = c(4,2,3))
  x <- new_rvar(x_array)
  x_letters <- array(letters[1:24], dim = c(4,2,3))
  x_ord <- rvar_ordered(x_letters, levels = letters)

  expect_equal(draws_of(rvar_range(x)), t(apply(x_array, 1, range)), check.attributes = FALSE)
  expect_equal(rvar_range(x_ord), rvar_ordered(array(letters[t(apply(x_array, 1, range))], dim = c(4, 2)), levels = letters))
  expect_error(rvar_range(rvar_factor("a")))

  # default values on empty input
  expect_equal(rvar_range(), as_rvar(c(Inf, -Inf)))
})


# quantiles ---------------------------------------------------------------

test_that("rvar_quantile works", {
  x_array <- array(1:24, dim = c(4,2,3))
  x <- new_rvar(x_array)
  x_letters <- array(letters[1:24], dim = c(4,2,3))
  x_ord <- rvar_ordered(x_letters, levels = letters)

  p <- c(0.25, 0.5, 0.75)
  quantiles <- t(apply(x_array, 1, quantile, probs = p, names = TRUE))
  dimnames(quantiles)[1] <- list(1:4)
  expect_equal(draws_of(rvar_quantile(x, probs = p, names = TRUE)), quantiles)

  ord_quantiles <- t(apply(x_array, 1, quantile, probs = p, type = 1, names = TRUE))
  dimnames(ord_quantiles)[1] <- list(1:4)
  expect_equal(
    rvar_quantile(x_ord, probs = p, names = TRUE),
    rvar_ordered(structure(ord_quantiles, levels = letters, class = c("ordered", "factor")))
  )

  dimnames(quantiles)[2] <- NULL
  expect_equal(draws_of(rvar_quantile(x, probs = p, names = FALSE)), quantiles)

  q50 <- array(apply(x_array, 1, quantile, probs = 0.5), dim = c(4, 1), dimnames = list(1:4, "50%"))
  expect_equal(draws_of(rvar_quantile(x, probs = 0.5, names = TRUE)), q50)

  # passing NULL should still result in a vector with length = length(probs)
  expect_equal(rvar_quantile(NULL, probs = c(0.25, 0.75)), as_rvar(c(NA_real_, NA_real_)))
})


# logical summaries -------------------------------------------------------

test_that("logical summaries work", {
  x_array <- array(1:24, dim = c(4,2,3))
  x <- new_rvar(x_array)

  expect_equal(draws_of(rvar_all(x > 6)), as.matrix(apply(x_array > 6, 1, all)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_any(x > 6)), as.matrix(apply(x_array > 6, 1, any)), check.attributes = FALSE)

  expect_error(rvar_all(rvar("a")))
  expect_error(rvar_any(rvar("a")))

  # default values on empty input
  expect_equal(rvar_all(), as_rvar(TRUE))
  expect_equal(rvar_any(), as_rvar(FALSE))
})


# special value predicates ------------------------------------------------

test_that("special value predicates work", {
  x_array <- c(1, Inf, -Inf, NaN, NA)
  x <- new_rvar(x_array)
  x_letters <- factor(letters[c(1, 2, 3, NaN, NA)])
  x_ord <- rvar_ordered(x_letters)
  x_fct <- rvar_factor(x_letters)

  expect_equal(draws_of(rvar_is_finite(x)), as.matrix(is.finite(x_array)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_infinite(x)), as.matrix(is.infinite(x_array)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_nan(x)), as.matrix(is.nan(x_array)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_na(x)), as.matrix(is.na(x_array)), check.attributes = FALSE)

  expect_equal(draws_of(rvar_is_finite(x_ord)), as.matrix(is.finite(x_letters)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_infinite(x_ord)), as.matrix(is.infinite(x_letters)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_nan(x_ord)), as.matrix(is.nan(x_letters)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_na(x_ord)), as.matrix(is.na(x_letters)), check.attributes = FALSE)

  expect_equal(draws_of(rvar_is_finite(x_fct)), as.matrix(is.finite(x_letters)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_infinite(x_fct)), as.matrix(is.infinite(x_letters)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_nan(x_fct)), as.matrix(is.nan(x_letters)), check.attributes = FALSE)
  expect_equal(draws_of(rvar_is_na(x_fct)), as.matrix(is.na(x_letters)), check.attributes = FALSE)
})

#' Create functions of random variables
#'
#' Function that create functions that can accept and/or produce [`rvar`]s.
#'
#' @param .f (multiple options) A function to turn into a function that accepts
#'   and/or produces random variables:
#'   * A function
#'   * A one-sided formula that can be parsed by [rlang::as_function()]
#' @param rvar_args (character vector) The names of the arguments of `.f` that
#'   should be allowed to accept [`rvar`]s as arguments. If `NULL` (the
#'   default), all arguments to `.f` are turned into arguments that accept
#'   [`rvar`]s.
#' @param ndraws (positive integer). The number of draws used to construct new
#'   random variables if no [`rvar`]s are supplied as arguments to the
#'   returned function. If `NULL`, `getOption("posterior.rvar_ndraws")` is used
#'   (default `4000`). If any arguments to the returned function contain
#'   [`rvar`]s, the number of draws in the provided [`rvar`]s is used instead of
#'   the value of this argument.
#'
#' @details This function wraps an existing function (`.f`) such that it returns [`rvar`]s containing
#' whatever type of data `.f` would normally return.
#'
#' The returned function, when called, executes `.f` possibly multiple times, once for each draw of
#' the [`rvar`]s passed to it, then returns a new
#' [`rvar`] representing the output of those function evaluations. If the arguments contain no [`rvar`]s,
#' then `.f` will be executed `ndraws` times and an [`rvar`] with that many draws returned.
#'
#' Functions created by `rfun()` are not necessarily *fast* (in fact in some cases they may be very slow), but
#' they have the advantage of allowing a nearly arbitrary R functions to be executed against [`rvar`]s
#' simply by wrapping them with `rfun()`. This makes it especially useful as a prototyping
#' tool. If you create code with `rfun()` and it is unacceptably slow for your application,
#' consider rewriting it using math operations directly on [`rvar`]s (which should be fast),
#' using [rvar_rng()], and/or using operations directly on the arrays that back the [`rvar`]s
#' (via [draws_of()]).
#'
#'
#' @return A function with the same argument specification as `.f`, but which can accept and return
#' [`rvar`]s.
#'
#' @examples
#'
#' rvar_norm <- rfun(rnorm)
#' rvar_gamma <- rfun(rgamma)
#'
#' mu <- rvar_norm(10, mean = 1:10, sd = 1)
#' sigma <- rvar_gamma(1, shape = 1, rate = 1)
#' x <- rvar_norm(10, mu, sigma)
#' x
#'
#' @family rfun
#' @export
rfun <- function (.f, rvar_args = NULL, ndraws = NULL) {
  # based loosely on base::Vectorize
  ndraws <- ndraws %||% getOption("posterior.rvar_ndraws", 4000)
  .f <- rlang::as_function(.f)

  arg_names <- as.list(formals(.f))
  arg_names[["..."]] <- NULL
  arg_names <- names(arg_names)
  rvar_args <- as.character(rvar_args %||% arg_names)

  if (!all(rvar_args %in% arg_names)) {
    stop_no_call("All arguments specified by `rvar_args` must be names of formal arguments of `.f`")
  }

  FUNV <- function() {
    args <- lapply(as.list(match.call())[-1L], eval, parent.frame())
    arg_names <-
      if (is.null(names(args))) character(length(args))
      else names(args)

    # convert arguments that are rvars into draws the function can be applied over
    is_rvar_arg <- (arg_names %in% rvar_args) & as.logical(lapply(args, is_rvar))
    rvar_args <- as_draws_rvars(args[is_rvar_arg])
    .nchains <- max(1, nchains(rvar_args))

    if (length(rvar_args) == 0) {
      # no rvar arguments, so just create a random variable by applying this function
      # ndraws times
      list_of_draws <- replicate(ndraws, do.call(.f, args), simplify = FALSE)
    } else {
      list_of_draws <- lapply(seq_len(ndraws(rvar_args)), function(i) {
        variables <- get_variables_from_one_draw(rvar_args, i)
        do.call(.f, c(variables, args[!is_rvar_arg]))
      })
    }
    # Need to add a first dimension before unchopping (this will be the draws dimension)
    # Doing this + list_unchop is faster than doing abind::abind(list_of_draws, along = 0)
    list_of_draws <- lapply(list_of_draws, function(x) {
      x <- as.array(x)
      dim(x) <- c(1, dim(x))
      x
    })
    new_rvar(vctrs::list_unchop(list_of_draws), .nchains = .nchains)
  }
  formals(FUNV) <- formals(.f)
  FUNV
}

#' Execute expressions of random variables
#'
#' Execute (nearly) arbitrary \R expressions that may include [`rvar`]s,
#' producing a new [`rvar`].
#'
#' @param expr (expression) A bare expression that can (optionally) contain
#'   [`rvar`]s. The expression supports [quasiquotation].
#' @param ndraws (positive integer) The number of draws used to construct new
#'   random variables if no [`rvar`]s are supplied in `expr`. If `NULL`,
#'   `getOption("posterior.rvar_ndraws")` is used (default 4000). If `expr`
#'   contains [`rvar`]s, the number of draws in the provided [`rvar`]s is used
#'   instead of the value of this argument.
#' @template args-rvar-dim
#'
#' @details This function evaluates `expr` possibly multiple times, once for each draw of
#' the [`rvar`]s it contains, then returns a new [`rvar`] representing the output of those
#' expressions. To identify [`rvar`]s, `rdo()` searches the calling environment for any variables
#' named in `expr` for which [is_rvar()] evaluates to `TRUE`. If `expr` contains no [`rvar`]s,
#' then it will be executed `ndraws` times and an [`rvar`] with that many draws returned.
#'
#' `rdo()` is not necessarily *fast* (in fact in some cases it may be very slow), but
#' it has the advantage of allowing a nearly arbitrary R expression to be executed against [`rvar`]s
#' simply by wrapping it with `rdo( ... )`. This makes it especially useful as a prototyping
#' tool. If you create code with `rdo()` and it is unacceptably slow for your application,
#' consider rewriting it using math operations directly on [`rvar`]s (which should be fast),
#' using [rvar_rng()], and/or using operations directly on the arrays that back the [`rvar`]s
#' (via [draws_of()]).
#'
#' @return An [`rvar`].
#'
#' @examples
#'
#' mu <- rdo(rnorm(10, mean = 1:10, sd = 1))
#' sigma <- rdo(rgamma(1, shape = 1, rate = 1))
#' x <- rdo(rnorm(10, mu, sigma))
#' x
#'
#' @family rfun
#' @importFrom rlang eval_tidy quo_get_env enquo missing_arg quo_get_expr
#' @export
rdo <- function(expr, dim = NULL, ndraws = NULL) {
  ndraws <- ndraws %||% getOption("posterior.rvar_ndraws", 4000)

  # basic idea here is to find all the variables that are used in the expression
  # and which are also random variables in the expression's environment, then
  # build a function that executes the expression and takes those random
  # variables as arguments, then vectorize that function using `rfun()` and execute it.
  f_expr <- enquo(expr)
  f_env <- quo_get_env(f_expr)

  rvars_in_expr <- list()
  f_alist <- alist()
  for (name in all.vars(f_expr)) {
    var <- get(name, f_env)
    if (is_rvar(var)) {
      rvars_in_expr[[name]] <- var
      f_alist[[name]] <- missing_arg()
    }
  }

  f_alist <- append(f_alist, quo_get_expr(f_expr))
  f <- rfun(as.function(f_alist, envir = f_env), ndraws = ndraws)
  result <- do.call(f, rvars_in_expr, envir = f_env)

  if (!is.null(dim)) {
    dim(result) <- dim
  }
  result
}

#' Create random variables from existing random number generators
#'
#' Specialized alternative to `rdo()` or `rfun()` for creating [`rvar`]s from
#' existing random-number generator functions (such as `rnorm()`, `rbinom()`, etc).
#'
#' @param .f (function) A function (or string naming a function) representing a
#'   random-number generating function that follows the pattern of base random
#'   number generators (like `rnorm()`, `rbinom()`, etc). It must:
#'   - Have a first argument, `n`, giving the number of draws to take from the
#'     distribution
#'   - Have vectorized parameter arguments
#'   - Return a single vector of length `n`
#' @param n (positive integer) The length of the output [`rvar`] vector (**not**
#'   the number of draws).
#' @param ... Arguments passed to `.f`. These arguments may include [`rvar`]s,
#'   so long as they are vectors only (no multidimensional [`rvar`]s are
#'   allowed).
#' @param ndraws (positive integer) The number of draws used to construct the
#'   returned random variable if no [`rvar`]s are supplied in `...`. If `NULL`,
#'   `getOption("posterior.rvar_ndraws")` is used (default 4000). If `...`
#'   contains [`rvar`]s, the number of draws in the provided [`rvar`]s is used
#'   instead of the value of this argument.
#'
#' @details This function unwraps the arrays underlying the input [`rvar`]s in
#' `...` and then passes them to `.f`, relying on the vectorization of `.f`
#' to evaluate it across draws from the input [`rvar`]s. This is why the arguments
#' of `.f` **must** be vectorized. It asks for `n` times the number of draws
#' in the input [`rvar`]s (or `ndraws` if none are given) draws from the
#' random number generator `.f`, then reshapes the output from `.f` into an
#' [`rvar`] with length `n`.
#'
#' `rvar_rng()` is a fast alternative to `rdo()` or `rfun()`, but you **must**
#' ensure that `.f` satisfies the preconditions described above for the result
#' to be correct. Most base random number generators satisfy these conditions.
#' It is advisable to test against `rdo()` or `rfun()` (which should be correct,
#' but slower) if you are uncertain.
#'
#' @return A single-dimensional [`rvar`] of length `n`.
#'
#' @examples
#'
#' mu <- rvar_rng(rnorm, 10, mean = 1:10, sd = 1)
#' sigma <- rvar_rng(rgamma, 1, shape = 1, rate = 1)
#' x <- rvar_rng(rnorm, 10, mu, sigma)
#' x
#'
#' @family rfun
#' @export
rvar_rng <- function(.f, n, ..., ndraws = NULL) {
  args <- list(...)

  is_rvar_arg <- vapply(args, is_rvar, logical(1))
  rvar_args <- conform_rvar_ndraws_nchains(args[is_rvar_arg])

  if (length(rvar_args) < 1) {
    nchains <- 1
    ndraws <- ndraws %||% getOption("posterior.rvar_ndraws", 4000)
  } else {
    # we have some arguments that are rvars. We require them to be single-dimensional
    # (vectors) so that R's vector recycling will work correctly.
    nchains <- nchains(rvar_args[[1]])
    ndraws <- ndraws(rvar_args[[1]])

    rvar_args_ndims <- lengths(lapply(rvar_args, dim))
    if (!all(rvar_args_ndims == 1)) {
      stop_no_call("All arguments to rvar_rng() that are rvars must be single-dimensional (vectors).")
    }

    args[is_rvar_arg] <- lapply(rvar_args, function(x) as.vector(draws_of(x)))
  }

  # we must manually recycle numeric vector arguments up to the desired number
  # of draws so that they can be correctly recycled along the draws of any input
  # rvars. We only convert numeric *vectors*, as (1) scalars can be recycled
  # as-is and (2) matrices and 2d+ arrays cannot be correctly recycled using R's
  # recycling rules so they are typically only used as constant arguments to
  # random number generator functions (e.g. Sigma for a multivariate normal),
  # so we don't need to worry about them.
  is_numeric_vector_arg <-
    vapply(args, function(x) is.numeric(x) && length(x) > 1 && length(dim(x)) <= 1, logical(1)) &
    !is_rvar_arg
  args[is_numeric_vector_arg] <- lapply(args[is_numeric_vector_arg], rep, each = ndraws)

  nd <- n * ndraws
  args <- c(n = nd, args)
  result <- do.call(.f, args)
  dim(result) <- c(ndraws, n)
  new_rvar(result, .nchains = nchains)
}

#' Print or format a random variable
#'
#' Printing and formatting methods for [`rvar`]s.
#'
#' @encoding UTF-8
#' @param x,object (rvar) The [`rvar`] to print.
#' @template args-print-digits
#' @template args-print-summary
#' @template args-print-dots
#' @param color (logical) Whether or not to use color when formatting the
#'   output. If `TRUE`, the [pillar::style_num()] functions may be used to
#'   produce strings containing control sequences to produce colored output on
#'   the terminal.
#' @param width The maxmimum width used to print out lists of factor levels
#'   for [`rvar_factor`]s. See [format()].
#' @param vec.len (nonnegative integer) How many 'first few' elements are
#'   displayed of each vector. If `NULL`, defaults to
#'   `getOption("str")$vec.len`, which defaults to 4.
#' @param indent.str (string) The indentation string to use.
#' @param nest.lev (nonnegative integer) Current nesting level in the recursive
#'   calls to `str()`.
#' @param give.attr (logical) If `TRUE` (default), show attributes as sub
#'   structures.
#'
#' @details
#' `print()` and `str()` print out [`rvar`] objects by summarizing each element
#' in the random variable with either its mean±sd or median±mad, depending on
#' the value of `summary`. Both functions use the `format()` implementation for
#' [`rvar`] objects under the hood, which returns a character vector in the
#' mean±sd or median±mad form.
#'
#' @return
#' For `print()`, an invisible version of the input object.
#'
#' For `str()`, nothing; i.e. `invisible(NULL)`.
#'
#' For `format()`, a character vector of the same dimensions as `x` where each
#' entry is of the form `"mean±sd"` or `"median±mad"`, depending on the value
#' of `summary`.
#'
#' @template ref-tastle-wierman-2007
#' @examples
#'
#' set.seed(5678)
#' x = rbind(
#'   cbind(rvar(rnorm(1000, 1)), rvar(rnorm(1000, 2))),
#'   cbind(rvar(rnorm(1000, 3)), rvar(rnorm(1000, 4)))
#' )
#'
#' print(x)
#' print(x, summary = "median_mad")
#'
#' str(x)
#'
#' format(x)
#'
#' @importFrom utils str lsf.str
#' @export
print.rvar <- function(x, ..., summary = NULL, digits = NULL, color = TRUE, width = getOption("width")) {
  digits <- digits %||% getOption("posterior.digits", 2)
  # \u00b1 = plus/minus sign
  summary_functions <- get_summary_functions(draws_of(x), summary)
  plus_minus <- summary_functions[[1]] != "modal_category"
  summary_string <- if (plus_minus) {
    paste0(paste(names(summary_functions), collapse = " \u00b1 "), ":")
  } else {
    paste0(paste(names(summary_functions), collapse = " <"), ">:")
  }
  if (color) {
    summary_string <- pillar::style_subtle(summary_string)
  }
  cat0(rvar_type_full(x), " ", summary_string, "\n")

  x_string <- format(x, summary = summary, digits = digits, color = FALSE, pad_right = " ")
  if (length(x_string) == 0) {
    cat0(rvar_class(x), "()\n")
  } else {
    print(x_string, quote = FALSE)
  }

  if (is_rvar_factor(x)) {
    cat0(format_levels(levels(x), is_rvar_ordered(x), width = width), "\n")
  }

  invisible(x)
}

#' @rdname print.rvar
#' @export
format.rvar <- function(x, ..., summary = NULL, digits = NULL, color = FALSE) {
  digits <- digits %||% getOption("posterior.digits", 2)
  format_rvar_draws(draws_of(x), ..., summary = summary, digits = digits, color = color)
}

#' @rdname print.rvar
#' @export
str.rvar <- function(
  object, ..., summary = NULL, vec.len = NULL,
  indent.str = paste(rep.int(" ", max(0, nest.lev + 1)), collapse = ".."),
  nest.lev = 0, give.attr = TRUE
) {

  str_next <- function(x, ...) {
    str(x, ...,
      summary = summary,
      vec.len = vec.len,
      indent.str = paste(indent.str, ".."),
      nest.lev = nest.lev + 1,
      give.attr = give.attr
    )
  }

  # HEADER
  .draws <- draws_of(object)
  vec.len <- vec.len %||% getOption("str")$vec.len %||% 4

  # flatten all the non-draws dimensions for display
  .dim <- dim(.draws)
  dim(.draws) <- c(.dim[1], prod(.dim[-1]))

  if (dim(.draws)[[2]] > vec.len) {
    .draws <- .draws[, 1:vec.len, drop = FALSE]
    ellipsis <- " ..."
  } else {
    ellipsis <- ""
  }

  cat0(" ", rvar_type_full(object), "  ",
    paste(format_rvar_draws(.draws, summary = summary, trim = TRUE), collapse = "  "),
    ellipsis, "\n"
  )

  # LEVELS
  # for factor-like rvars
  if (is_rvar_factor(object)) {
    cat0(indent.str, "- ", format_levels(levels(object), is_rvar_ordered(object), max_level = vec.len), "\n")
  }

  # ATTRIBUTES
  # we have to be a bit clever about this to hide internal structure we don't
  # want people messing with + the fact that some attributes (like dimnames)
  # are actually attributes of the draws and not the base object.
  if (give.attr) {
    .dimnames <- dimnames(object)
    if (!all(sapply(.dimnames, is.null))) {
      # only show dimnames if they aren't all NULL
      cat0(indent.str, paste0('- dimnames(*)='))
      str_next(.dimnames, ...)
    }

    str_attr <- function(a, base, exclude) {
      a_names <- names(a)
      for (i in seq_along(a)) {
        if (!a_names[[i]] %in% exclude) {
          cat0(indent.str, paste0('- attr(', base, ', "', a_names[[i]], '")='))
          str_next(a[[i]], ...)
        }
      }
    }
    str_attr(attributes(draws_of(object)), "draws_of(*)", c("names", "dim", "dimnames", "class", "levels"))
    str_attr(attributes(object), "*", c("draws", "names", "dim", "dimnames", "class", "nchains", "cache"))
  }

  invisible(NULL)
}

#' @importFrom pillar pillar_shaft new_pillar_shaft_simple
#' @export
pillar_shaft.rvar <- function(x, ...) {
  new_pillar_shaft_simple(format(x, color = TRUE, pad_left = " "), align = "right", ...)
}

#' @importFrom pillar format_glimpse
#' @export
format_glimpse.rvar <- function(x, ...) {
  .dim <- dim(x)
  if (length(.dim) > 1) {
    paste0("<rvar[", paste0(.dim, collapse = " x "), "]>")
  } else {
    format(x, ..., trim = TRUE)
  }
}

# type summaries ----------------------------------------------------------

#' @importFrom vctrs vec_ptype_abbr
#' @export
vec_ptype_abbr.rvar <- function(x, ...) {
  "rvar"
}
#' @export
vec_ptype_abbr.rvar_factor <- function(x, ...) {
  "rvar_fct"
}
#' @export
vec_ptype_abbr.rvar_ordered <- function(x, ...) {
  "rvar_ord"
}

#' @importFrom vctrs vec_ptype_full
#' @export
vec_ptype_full.rvar <- function(x, ...) {
  rvar_type_full(x, dim1 = FALSE)
}

rvar_type_full <- function(x, dim1 = TRUE) {
  .dim <- dim(draws_of(x))

  dim_str <- if (dim1) {
    paste0("[", paste0(.dim[-1], collapse = ","), "]")
  } else if (length(.dim) > 2) {
    paste0("[,", paste0(.dim[-c(1,2)], collapse = ","), "]")
  } else {
    ""
  }

  chain_str <- if (nchains(x) > 1) {
    paste0(",", nchains(x))
  }

  paste0(rvar_class(x), "<", niterations(x), chain_str, ">", dim_str)
}

rvar_class <- function(x) {
  paste0(
    "rvar",
    if (is_rvar_ordered(x)) "_ordered"
    else if (is_rvar_factor(x)) "_factor"
  )
}


# rvar draws formatting helpers ------------------------------------------------

# formats a draws array for display as individual "variables" (i.e. maintaining
# its dimensions except for the dimension representing draws)
format_rvar_draws <- function(
  draws, ..., pad_left = "", pad_right = "", summary = NULL, digits = 2, color = FALSE, trim = FALSE
) {
  if (length(draws) == 0) {
    return(character())
  }
  summary_functions <- get_summary_functions(draws, summary)
  plus_minus <- summary_functions[[1]] != "modal_category"

  summary_dimensions <- seq_len(length(dim(draws)) - 1) + 1

  # these will be mean/sd, median/mad, mode/entropy, mode/dissent depending on `summary`
  .mean <- .apply_factor(draws, summary_dimensions, summary_functions[[1]])
  .sd <- .apply_factor(draws, summary_dimensions, summary_functions[[2]])

  out <- paste0(
    pad_left,
    format_mean_sd(.mean, .sd, digits = digits, color = color, trim = trim, plus_minus = plus_minus),
    pad_right
  )

  dim(out) <- dim(draws)[summary_dimensions]
  dimnames(out) <- dimnames(draws)[summary_dimensions]
  out
}

format_mean <- function(x, digits = 2, color = FALSE, trim = FALSE) {
  format(x, justify = "right", digits = digits, scientific = 2, trim = trim)
}

format_sd <- function(x, digits = 2, color = FALSE, plus_minus = TRUE) {
  sd_formatted <- format(x, justify = "left", trim = TRUE, digits = digits, scientific = 2)
  sd_string <- if (plus_minus) {
    # \u00b1 = plus/minus sign
    paste0("\u00b1 ", sd_formatted)
  } else {
    paste0("<", sd_formatted, ">")
  }

  if (color) {
    pillar::style_subtle(sd_string)
  } else {
    sd_string
  }
}

format_mean_sd <- function(.mean, .sd, digits = 2, color = FALSE, trim = FALSE, plus_minus = TRUE) {
  format(paste0(
    format_mean(.mean, digits = digits, color = color, trim = trim), " ",
    format_sd(.sd, digits = digits, color = color, plus_minus = plus_minus)),
  justify = if (trim) "none" else "left", trim = trim)
}

format_levels <- function(levels, ordered = FALSE, max_level = NULL, width = getOption("width")) {
  levels <- encodeString(levels)
  n_levels <- length(levels)
  header <- paste(format(n_levels), "levels: ")

  sep <- if (ordered) " < " else " "
  width <- width - (nchar(header, "width") + 3L + 1L + 3L)
  levels_cumwidth <- cumsum(nchar(levels, "width") + nchar(sep, "width"))

  max_level <- max_level %||% if (n_levels <= 1L || levels_cumwidth[n_levels] <= width) {
    n_levels
  } else {
    max(1L, which.max(levels_cumwidth > width) - 1L)
  }
  drop <- n_levels > max_level

  paste0(
    header,
    paste(
      if (drop) c(levels[1L:max(1, max_level - 1)], "...", if (max_level > 1) levels[n_levels])
      else levels,
      collapse = sep
    )
  )
}

# check that summary is a valid name of the type of summary to do and
# return a vector of two elements, where the first is the point summary function
# (mean, median, mode) and the second is the uncertainty function ()
get_summary_functions <- function(draws, summary = NULL) {
  if (is.ordered(draws)) {
    summary = "mode_dissent"
  } else if (is.factor(draws)) {
    summary = "mode_entropy"
  }

  if (is.null(summary)) summary <- getOption("posterior.rvar_summary", "mean_sd")
  switch(summary,
    mean_sd = list(mean = "mean", sd = "sd"),
    median_mad = list(median = "median", mad = "mad"),
    mode_entropy = list(mode = "modal_category", entropy = "entropy"),
    mode_dissent = list(mode = "modal_category", dissent = "dissent"),
    stop_no_call('`summary` must be one of "mean_sd" or "median_mad"')
  )
}

#' Bind `draws` objects together
#'
#' Bind multiple [`draws`] objects together to form a single `draws` object.
#'
#' @param x (draws) A [`draws`] object. The draws format of `x` will define the
#'   format of the returned draws object.
#' @param ... (draws) Additional [`draws`] objects to bind to `x`.
#' @param along (string) The dimension along which draws objects should be bound
#'   together. Possible values are `"variable"` (the default), `"chain"`,
#'   `"iteration"`, and `"draw"`. Not all options are supported for all input
#'   formats.
#' @template return-draws
#'
#' @examples
#' x1 <- draws_matrix(alpha = rnorm(5), beta = rnorm(5))
#' x2 <- draws_matrix(alpha = rnorm(5), beta = rnorm(5))
#' ndraws(x1)
#' ndraws(x2)
#' x3 <- bind_draws(x1, x2, along = "draw")
#' ndraws(x3)
#'
#' x4 <- draws_matrix(theta = rexp(5))
#' x5 <- bind_draws(x1, x4, along = "variable")
#' variables(x5)
#'
#' @importFrom abind abind
#' @export
bind_draws <- function(x, ...) {
  UseMethod("bind_draws")
}

#' @rdname bind_draws
#' @export
bind_draws.draws_matrix <- function(x, ..., along = "variable") {
  along <- validate_along(along)
  dots <- list(...)
  if (!length(dots)) {
    return(as_draws_matrix(x))
  }
  dots <- c(list(x), dots)
  dots <- remove_null(dots)
  dots <- lapply(dots, as_draws_matrix)
  dots <- lapply(dots, repair_draws)
  if (along == "variable") {
    check_same_fun_output(dots, chain_ids)
    check_same_fun_output(dots, iteration_ids)
    out <- do.call(abind, c(dots, along = 2L))
    attr(out, "nchains") <- nchains(dots[[1]])
  } else if (along == "chain") {
    check_same_fun_output(dots, variables)
    check_same_fun_output(dots, iteration_ids)
    out <- do.call(abind, c(dots, along = 1L))
    attr(out, "nchains") <- sum(sapply(dots, nchains))
  } else if (along == "iteration") {
    stop_no_call("Cannot bind 'draws_matrix' objects along 'iteration'.")
  } else if (along %in% c("draw")) {
    check_same_fun_output(dots, variables)
    out <- do.call(abind, c(dots, along = 1L))
    attr(out, "nchains") <- 1L
  }
  as_draws_matrix(out)
}

#' @rdname bind_draws
#' @export
bind_draws.draws_array <- function(x, ..., along = "variable") {
  along <- validate_along(along)
  dots <- list(...)
  if (!length(dots)) {
    return(as_draws_array(x))
  }
  dots <- c(list(x), dots)
  dots <- remove_null(dots)
  dots <- lapply(dots, as_draws_array)
  dots <- lapply(dots, repair_draws)
  if (along == "variable") {
    check_same_fun_output(dots, chain_ids)
    check_same_fun_output(dots, iteration_ids)
    out <- do.call(abind, c(dots, along = 3L))
  } else if (along == "chain") {
    check_same_fun_output(dots, variables)
    check_same_fun_output(dots, iteration_ids)
    out <- do.call(abind, c(dots, along = 2L))
  } else if (along == "iteration") {
    check_same_fun_output(dots, variables)
    check_same_fun_output(dots, chain_ids)
    out <- do.call(abind, c(dots, along = 1L))
  } else if (along == "draw") {
    stop_no_call("Cannot bind 'draws_array' objects along 'draw'.")
  }
  as_draws_array(out)
}

#' @rdname bind_draws
#' @export
bind_draws.draws_df <- function(x, ..., along = "variable") {
  along <- validate_along(along)
  dots <- list(...)
  if (!length(dots)) {
    return(as_draws_df(x))
  }
  dots <- c(list(x), dots)
  dots <- remove_null(dots)
  dots <- lapply(dots, as_draws_df)
  dots <- lapply(dots, repair_draws)
  if (along == "variable") {
    check_same_fun_output(dots, chain_ids)
    check_same_fun_output(dots, iteration_ids)
    reserved_df_values <- as.data.frame(dots[[1]])[, c(".chain", ".iteration")]
    dots <- lapply(dots, remove_reserved_df_variables)
    out <- do.call(cbind, dots)
    out <- cbind(out, reserved_df_values)
  } else if (along == "chain") {
    check_same_fun_output(dots, variables)
    check_same_fun_output(dots, iteration_ids)
    cumsum_chains <- c(0, cumsum(ulapply(dots, nchains)))
    for (i in seq_along(dots)) {
      dots[[i]]$.chain <- cumsum_chains[i] + dots[[i]]$.chain
      dots[[i]]$.chain <- as.integer(dots[[i]]$.chain)
    }
    out <- do.call(rbind, dots)
  } else if (along == "iteration") {
    check_same_fun_output(dots, variables)
    check_same_fun_output(dots, chain_ids)
    cumsum_iterations <- c(0, cumsum(ulapply(dots, niterations)))
    for (i in seq_along(dots)) {
      dots[[i]]$.iteration <- cumsum_iterations[i] + dots[[i]]$.iteration
      dots[[i]]$.iteration <- as.integer(dots[[i]]$.iteration)
    }
    out <- do.call(rbind, dots)
  } else if (along == "draw") {
    check_same_fun_output(dots, variables)
    out <- do.call(rbind, dots)
    # binding along 'draw' implies dropping chain information
    out$.chain <- 1L
    out$.iteration <- seq_rows(out)
  }
  out$.draw <- NULL
  .as_draws_df(out)
}

#' @rdname bind_draws
#' @export
bind_draws.draws_list <- function(x, ..., along = "variable") {
  along <- validate_along(along)
  dots <- list(...)
  if (!length(dots)) {
    return(as_draws_list(x))
  }
  dots <- c(list(x), dots)
  dots <- remove_null(dots)
  dots <- lapply(dots, as_draws_list)
  dots <- lapply(dots, repair_draws)
  if (along == "variable") {
    check_same_fun_output(dots, chain_ids)
    check_same_fun_output(dots, iteration_ids)
    chains <- seq_along(dots[[1]])
    out <- vector("list", length(chains))
    for (i in chains) {
      out[[i]] <- ulapply(dots, "[[", i, recursive = FALSE)
    }
  } else if (along == "chain") {
    check_same_fun_output(dots, variables)
    check_same_fun_output(dots, iteration_ids)
    out <- do.call(c, dots)
  } else if (along == "iteration") {
    check_same_fun_output(dots, variables)
    check_same_fun_output(dots, chain_ids)
    chains <- seq_along(dots[[1]])
    variables <- variables(dots[[1]])
    out <- vector("list", length(chains))
    for (i in chains) {
      tmp <- lapply(dots, "[[", i)
      out[[i]] <- named_list(variables)
      for (v in variables) {
        out[[i]][[v]] <- ulapply(tmp, "[[", v)
      }
    }
  } else if (along == "draw") {
    stop_no_call("Cannot bind 'draws_list' objects along 'draw'.")
  }
  as_draws_list(out)
}

#' @rdname bind_draws
#' @importFrom abind abind
#' @export
bind_draws.draws_rvars <- function(x, ..., along = "variable") {
  along <- validate_along(along)
  dots <- list(...)
  if (!length(dots)) {
    return(as_draws_rvars(x))
  }
  dots <- c(list(x), dots)
  dots <- remove_null(dots)
  dots <- lapply(dots, as_draws_rvars)
  dots <- lapply(dots, repair_draws)
  if (along == "variable") {
    check_same_fun_output(dots, chain_ids)
    check_same_fun_output(dots, iteration_ids)
    out <- do.call(c, dots)
  } else if (along == "iteration") {
    stop_no_call("Cannot bind 'draws_rvars' objects along 'iteration'.")
  } else if (along %in% c("chain", "draw")) {
    check_same_fun_output(dots, variables)
    if (along == "chain") {
      check_same_fun_output(dots, iteration_ids)
      nchains <- sum(sapply(dots, nchains))
    } else {
      # binding along 'draw' implies dropping chain information
      dots <- lapply(dots, merge_chains)
      nchains <- 1
    }
    # bind all the corresponding variables together along draws
    out <- lapply(seq_along(dots[[1]]), function(var_i) {
      vars <- lapply(dots, `[[`, var_i)
      var_draws <- lapply(vars, draws_of)
      out <- rvar(abind(var_draws, along = 1), nchains = nchains)
      out
    })
    names(out) <- names(dots[[1]])
  }
  as_draws_rvars(out)
}

#' @export
bind_draws.NULL <- function(x, ..., along = "variable") {
  dots <- list(...)
  dots <- remove_null(dots)
  if (!length(dots)) {
    stop_no_call("All objects passed to 'bind_draws' are NULL.")
  }
  do.call("bind_draws", dots)
}

#' @export
bind_draws.list <- function(x, ..., along = "variable") {
  do.call("bind_draws", c(x, ..., along = along))
}

# check if function output is the same across objects
# @param ls list of objects
# @param fun a function to be evaluated on the objects
# @param TRUE if the function output matches for all objects
check_same_fun_output <- function(ls, fun) {
  assert_list(ls)
  if (is.function(fun)) {
    fun_name <- deparse_pretty(substitute(fun))
  } else {
    fun_name <- as_one_character(fun)
    fun <- get(fun, mode = "function")
  }
  ids <- lapply(ls, fun)
  if (sum(!duplicated(ids)) > 1L) {
    stop_no_call("'", fun_name, "' of bound objects do not match.")
  }
  invisible(TRUE)
}

# validate the 'along' argument of 'bind_draws'
validate_along <- function(along) {
  options <- c("variable", "chain", "iteration", "draw")
  match.arg(along, options)
}

#' Weight `draws` objects
#'
#' Add weights to [`draws`] objects, with one weight per draw, for use in
#' subsequent weighting operations. For reasons of numerical accuracy, weights
#' are stored in the form of unnormalized log-weights (in a variable called
#' `.log_weight`). See [weights.draws()] for details how to extract weights from
#' `draws` objects.
#'
#' @template args-methods-x
#' @param weights (numeric vector) A vector of weights of length `ndraws(x)`.
#'   Weights will be internally stored on the log scale (in a variable called
#'   `.log_weight`) and will not be normalized, but normalized (non-log) weights
#'   can be returned via the [weights.draws()] method later.
#' @param log (logicla) Are the weights passed already on the log scale? The
#'   default is `FALSE`, that is, expecting `weights` to be on the standard
#'   (non-log) scale.
#' @template args-methods-dots
#' @template return-draws
#'
#' @seealso [weights.draws()], [resample_draws()]
#'
#' @examples
#' x <- example_draws()
#'
#' # sample some random weights for illustration
#' wts <- rexp(ndraws(x))
#' head(wts)
#'
#' # add weights
#' x <- weight_draws(x, weights = wts)
#'
#' # extract weights
#' head(weights(x)) # defaults to normalized weights
#' head(weights(x, normalize=FALSE)) # recover original weights
#' head(weights(x, log=TRUE)) # get normalized log-weights
#'
#' # add weights which are already on the log scale
#' log_wts <- log(wts)
#' head(log_wts)
#'
#' x <- weight_draws(x, weights = log_wts, log = TRUE)
#' # extract weights
#' head(weights(x))
#' head(weights(x, log=TRUE, normalize = FALSE)) # recover original log_wts
#'
#' @export
weight_draws <- function(x, weights, ...) {
  UseMethod("weight_draws")
}

#' @rdname weight_draws
#' @export
weight_draws.draws_matrix <- function(x, weights, log = FALSE, ...) {
  log <- as_one_logical(log)
  log_weights <- validate_weights(weights, x, log = log)
  if (".log_weight" %in% variables(x, reserved = TRUE)) {
    # overwrite existing weights
    x[, ".log_weight"] <- log_weights
  } else {
    # add weights as a new variable
    log_weights <- draws_matrix(.log_weight = log_weights, .nchains = nchains(x))
    x <- bind_draws(x, log_weights)
  }
  x
}

#' @rdname weight_draws
#' @export
weight_draws.draws_array <- function(x, weights, log = FALSE, ...) {
  log <- as_one_logical(log)
  log_weights <- validate_weights(weights, x, log = log)
  if (".log_weight" %in% variables(x, reserved = TRUE)) {
    # overwrite existing weights
    x[, , ".log_weight"] <- log_weights
  } else {
    # add weights as a new variable
    log_weights <- draws_array(.log_weight = log_weights, .nchains = nchains(x))
    x <- bind_draws(x, log_weights)
  }
  x
}

#' @rdname weight_draws
#' @export
weight_draws.draws_df <- function(x, weights, log = FALSE, ...) {
  log <- as_one_logical(log)
  log_weights <- validate_weights(weights, x, log = log)
  x$.log_weight <- log_weights
  x
}

#' @rdname weight_draws
#' @export
weight_draws.draws_list <- function(x, weights, log = FALSE, ...) {
  log <- as_one_logical(log)
  log_weights <- validate_weights(weights, x, log = log)
  niterations <- niterations(x)
  for (i in seq_len(nchains(x))) {
    sel <- (1 + (i - 1) * niterations):(i * niterations)
    x[[i]]$.log_weight <- log_weights[sel]
  }
  x
}

#' @rdname weight_draws
#' @export
weight_draws.draws_rvars <- function(x, weights, log = FALSE, ...) {
  log <- as_one_logical(log)
  log_weights <- validate_weights(weights, x, log = log)
  x$.log_weight <- rvar(log_weights)
  x
}

#' Extract Weights from Draws Objects
#'
#' Extract weights from [`draws`] objects, with one weight per draw.
#' See [`weight_draws`] for details how to add weights to [`draws`] objects.
#'
#' @param object (draws) A [`draws`] object.
#' @param log (logical) Should the weights be returned on the log scale?
#'   Defaults to `FALSE`.
#' @param normalize (logical) Should the weights be normalized to sum to 1 on
#'   the standard scale? Defaults to `TRUE`.
#' @template args-methods-dots
#'
#' @return A vector of weights, with one weight per draw.
#'
#' @seealso [`weight_draws`], [`resample_draws`]
#'
#' @inherit weight_draws examples
#'
#' @export
weights.draws <- function(object, log = FALSE, normalize = TRUE, ...) {
  log <- as_one_logical(log)
  normalize <- as_one_logical(normalize)
  if (!".log_weight" %in% variables(object, reserved = TRUE)) {
    return(NULL)
  }
  out <- extract_variable(object, ".log_weight")
  if (normalize) {
    out <- out - log_sum_exp(out)
  }
  if (!log) {
    out <- exp(out)
  }
  out
}

# validate weights and return log weights
validate_weights <- function(weights, draws, log = FALSE) {
  checkmate::expect_numeric(weights)
  checkmate::expect_flag(log)
  if (length(weights) != ndraws(draws)) {
    stop_no_call("Number of weights must match the number of draws.")
  }
  if (!log) {
    if (any(weights < 0)) {
      stop_no_call("Weights must be non-negative.")
    }
    weights <- log(weights)
  }
  weights
}

#' Rename variables in `draws` objects
#'
#' Rename variables in a [`draws`] object.
#'
#' @param .x (draws) A [`draws`] object.
#' @param ... One or more expressions, separated by commas, indicating the
#'   variables to rename. The variable names can be unquoted
#'   (`new_name = old_name`) or quoted (`"new_name" = "old_name"`). For non-scalar
#'   variables, all elements can be renamed together (`"new_name" = "old_name"`)
#'   or they can be renamed individually (`"new_name[1]" = "old_name[1]"`).
#'
#' @return
#' Returns a [`draws`] object of the same format as `.x`, with variables renamed
#' according to the expressions provided in `...`.
#'
#' @seealso [`variables`], [`mutate_variables`]
#'
#' @examples
#' x <- as_draws_df(example_draws())
#' variables(x)
#'
#' x <- rename_variables(x, mean = mu, sigma = tau)
#' variables(x)
#'
#' x <- rename_variables(x, b = `theta[1]`) # or b  = "theta[1]"
#' variables(x)
#'
#' # rename all elements of 'theta' at once
#' x <- rename_variables(x, alpha = theta)
#' variables(x)
#'
#' @importFrom rlang quos as_name
#' @export
rename_variables <- function(.x, ...) {
  UseMethod("rename_variables")
}

#' @rdname rename_variables
#' @export
rename_variables.draws <- function(.x, ...) {
  old_names <- sapply(quos(...), as_name)
  new_names <- names(old_names)

  if (any(new_names == "")) {
    old_names_without_new_name = old_names[new_names == ""]
    stop_no_call(
      "Cannot rename a variable to an empty name.\n",
      "The following variables did not have a new name provided:\n",
      comma(old_names_without_new_name)
    )
  }

  # loop over names as every old name may correspond to multiple
  # scalar variables if the name targets a non-scalar variable
  # also this allows renaming operations to be chained
  for (i in seq_along(old_names)) {
    old_names_i <- check_existing_variables(old_names[i], .x)
    v_regex <- paste0("^", escape_all(old_names[i]))
    v_indices <- sub(v_regex, "", old_names_i)
    new_names_i <- paste0(new_names[i], v_indices)
    sel <- which(variables(.x) %in% old_names_i)
    variables(.x)[sel] <- new_names_i
  }
  check_new_variables(variables(.x))
  .x
}

#' The `draws_matrix` format
#'
#' @name draws_matrix
#' @family formats
#'
#' @templateVar draws_format draws_matrix
#' @templateVar base_class "matrix"
#' @template draws_format-skeleton
#' @template args-format-nchains
#'
#' @details Objects of class `"draws_matrix"` are matrices (2-D arrays) with
#'   dimensions `"draw"` and `"variable"`. See **Examples**.
#'
NULL

#' @rdname draws_matrix
#' @export
as_draws_matrix <- function(x, ...) {
  UseMethod("as_draws_matrix")
}

#' @rdname draws_matrix
#' @export
as_draws_matrix.default <- function(x, ...) {
  x <- as_draws(x)
  as_draws_matrix(x, ...)
}

#' @rdname draws_matrix
#' @export
as_draws_matrix.draws_matrix <- function(x, ...) {
  x
}

#' @rdname draws_matrix
#' @export
as_draws_matrix.draws_array <- function(x, ...) {
  if (ndraws(x) == 0) {
    return(empty_draws_matrix(variables(x), niterations(x)))
  }
  nchains <- nchains(x)
  old_dim <- dim(x)
  old_dimnames <- dimnames(x)
  dim(x) <- c(old_dim[1] * old_dim[2], old_dim[3])
  dimnames(x) <- list(
    draw = as.character(seq_rows(x)),
    variable = old_dimnames[[3]]
  )
  class(x) <- class_draws_matrix()
  attr(x, "nchains") <- nchains
  x
}

#' @rdname draws_matrix
#' @export
as_draws_matrix.draws_df <- function(x, ...) {
  if (ndraws(x) == 0) {
    return(empty_draws_matrix(variables(x)))
  }
  nchains <- nchains(x)
  draws <- x$.draw
  x <- remove_reserved_df_variables(x)
  class(x) <- class(x)[-1L]
  x <- check_variables_are_numeric(x, to = "draws_matrix")
  x <- .as_draws_matrix(x)
  rownames(x) <- draws
  attr(x, "nchains") <- nchains
  x
}

#' @rdname draws_matrix
#' @export
as_draws_matrix.draws_list <- function(x, ...) {
  x <- as_draws_df(x)
  as_draws_matrix(x, ...)
}

#' @rdname draws_matrix
#' @export
as_draws_matrix.draws_rvars <- function(x, ...) {
  if (ndraws(x) == 0) {
    return(empty_draws_matrix(variables(x)))
  }

  x <- check_variables_are_numeric(
    x, to = "draws_matrix", is_non_numeric = is_rvar_factor, convert = FALSE
  )

  # cbind discards class information when applied to vectors, which converts
  # the underlying factors to numeric
  out <- do.call(cbind, lapply(seq_along(x), function(i) {
    # flatten each rvar so it only has two dimensions: draws and variables
    # this also collapses indices into variable names in the format "var[i,j,k,...]"
    x_i <- flatten_array(x[[i]], names(x)[[i]])
    draws_of(x_i)
  }))
  out <- as_draws_matrix(out, ...)
  attr(out, "nchains") <- nchains(x)
  out
}

#' @rdname draws_matrix
#' @export
as_draws_matrix.mcmc <- function(x, ...) {
  class(x) <- "matrix"
  attributes(x)[c("title", "mcpar")] <- NULL
  .as_draws_matrix(x)
}

#' @rdname draws_matrix
#' @export
as_draws_matrix.mcmc.list <- function(x, ...) {
  as_draws_matrix(as_draws_array(x), ...)
}

# try to convert any R object into a 'draws_matrix' object
.as_draws_matrix <- function(x) {
  .nchains <- attr(x, "nchains") %||% 1L
  x <- as.matrix(x)
  new_dimnames <- list(draw = NULL, variable = NULL)
  if (!is.null(dimnames(x)[[2]])) {
    new_dimnames[[2]] <- dimnames(x)[[2]]
  } else {
    new_dimnames[[2]] <- default_variables(NCOL(x))
  }
  check_new_variables(new_dimnames[[2]])
  new_dimnames[[1]] <- as.character(seq_rows(x))
  dimnames(x) <- new_dimnames
  class(x) <- class_draws_matrix()
  attr(x, "nchains") <- .nchains
  x
}

#' @rdname draws_matrix
#' @export
draws_matrix <- function(..., .nchains = 1) {
  out <- validate_draws_per_variable(...)
  .nchains <- as_one_integer(.nchains)
  if (.nchains < 1) {
    stop_no_call("Number of chains must be positive.")
  }
  ndraws <- length(out[[1]])
  if (ndraws %% .nchains != 0) {
    stop_no_call("Number of chains does not divide the number of draws.")
  }
  out <- do.call(cbind, out)
  attr(out, "nchains") <- .nchains
  as_draws_matrix(out)
}

class_draws_matrix <- function() {
  c("draws_matrix", "draws", "matrix")
}

#' @rdname draws_matrix
#' @export
is_draws_matrix <- function(x) {
  inherits(x, "draws_matrix")
}

# is an object looking like a 'draws_matrix' object?
is_draws_matrix_like <- function(x) {
  is.matrix(x) || is.array(x) && length(dim(x)) == 2L
}

#' Extract parts of a `draws_matrix` object
#'
#' Extract parts of a `draws_matrix` object. They are strictly defined as
#' matrices (draws x variable) so dropping any of the
#' dimensions breaks the expected structure of the object. Accordingly, no
#' dropping of dimensions is done by default even if the extracted slices are of
#' length 1. If `drop` is manually set to `TRUE` and any of the dimensions is
#' actually dropped, this will lead to dropping the `"draws_matrix"` class as
#' well.
#'
#' @param x,i,j,...,drop Same as in the default extraction method but with
#'   `drop` being set to `FALSE` by default.
#'
#' @return An object of class `"draws_matrix"` unless any of the dimensions
#' was dropped during the extraction.
#'
#' @keywords internal
#' @export
`[.draws_matrix` <- function(x, i, j, ..., drop = FALSE) {
  # TODO: allow for argument 'reserved' as in '[.draws_df'
  #   right now this fails because NextMethod() cannot ignore arguments
  out <- NextMethod("[", drop = drop)
  if (length(dim(out)) == length(dim(x))) {
    class(out) <- class(x)
    .nchains <- nchains(x)
    if (missing(i)) {
      attr(out, "nchains") <- .nchains
    } else if (.nchains > 1L) {
      warn_merge_chains("index")
    }
  }
  out
}

#' @export
variance.draws_matrix <- function(x, ...) {
  var(as.vector(x))
}

# create an empty draws_matrix object
empty_draws_matrix <- function(variables = character(0), ndraws = 0) {
  assert_character(variables, null.ok = TRUE)
  assert_number(ndraws, lower = 0)
  out <- matrix(
    numeric(0),
    nrow = ndraws,
    ncol = length(variables),
    dimnames = list(
      draw = seq_len(ndraws),
      variable = variables
    )
  )
  class(out) <- class_draws_matrix()
  out
}

#' Extract array of a single (possibly indexed) variable
#'
#' Extract an array of draws of a single variable, including any dimensions of
#' variables with indices.
#'
#' @template args-methods-x
#' @param variable (string) The name of the variable to extract. To extract all
#' dimensions from variables with indices (e.g. `"x[1]"`), provide the base
#' variable name (e.g. `"x"`).
#' @template args-methods-dots
#' @returns
#' An `array` with dimension `niterations(x)` x `nchains(x)` x any remaining
#' dimensions determined by the indices of the variable `x`.
#' @family variable extraction methods
#' @examples
#' x <- example_draws(example = "multi_normal")
#'
#' mu <- extract_variable_array(x, variable = "mu")
#' str(mu)
#'
#' mu1 <- extract_variable_array(x, variable = "mu[1]")
#' str(mu1)
#'
#' Sigma <- extract_variable_array(x, variable = "Sigma")
#' str(Sigma)
#'
#' @export
extract_variable_array <- function(x, variable, ...) {
  UseMethod("extract_variable_array")
}

#' @rdname extract_variable_array
#' @export
extract_variable_array.default <- function(x, variable, ...) {
  x <- as_draws(x)
  extract_variable_array(x, variable, ...)
}

#' @rdname extract_variable_array
#' @export
extract_variable_array.draws <- function(x, variable, ...) {
  variable <- as_one_character(variable)

  if (isTRUE(nzchar(split_variable_names(variable)$indices))) {
    # indices provided => scalar => equivalent to extract_variable_matrix
    out <- extract_variable_matrix(x, variable, ...)
    dim(out) <- c(dim(out), 1)
    dimnames(out) <- list(NULL)
  } else {
    x <- subset_draws(x, variable = variable, reserved = FALSE)
    x <- as_draws_rvars(x)
    out <- draws_of(x[[variable]], with_chains = TRUE)
  }

  out
}

#' Extract draws of a single variable
#'
#' Extract a vector of draws of a single variable.
#'
#' @template args-methods-x
#' @template args-extract-variable
#' @template args-methods-dots
#' @return A vector of length equal to the number of draws.
#' @family variable extraction methods
#' @examples
#' x <- example_draws()
#' mu <- extract_variable(x, variable = "mu")
#' str(mu)
#'
#' @export
extract_variable <- function(x, variable, ...) {
  UseMethod("extract_variable")
}

#' @rdname extract_variable
#' @export
extract_variable.default <- function(x, variable, ...) {
  x <- as_draws(x)
  extract_variable(x, variable, ...)
}

#' @rdname extract_variable
#' @export
extract_variable.draws <- function(x, variable, ...) {
  variable <- as_one_character(variable)
  out <- .subset_draws(x, variable = variable, reserved = FALSE)
  out <- as_draws_matrix(out)
  as.vector(out)
}

#' @rdname extract_variable
#' @export
extract_variable.draws_df <- function(x, variable, ...) {
  variable <- as_one_character(variable)
  out <- .subset_draws(x, variable = variable, reserved = FALSE)
  out[[variable]]
}

#' @rdname extract_variable
#' @export
extract_variable.draws_list <- function(x, variable, ...) {
  variable <- as_one_character(variable)
  out <- .subset_draws(x, variable = variable, reserved = FALSE)
  out <- as_draws_df(out)
  out[[variable]]
}

#' @rdname extract_variable
#' @export
extract_variable.draws_rvars <- function(x, variable, ...) {
  variable <- as_one_character(variable)
  parts <- split_variable_names(variable)
  if (isTRUE(nzchar(parts$indices))) {
    # variable with indices in the name ("x[1]", etc), which can't be subset
    # from draws_rvars directly, so we'll convert to a draws_df first.
    out <- extract_variable(as_draws_df(x[parts$base_name]), variable = variable, ...)
  } else if (length(x[[variable]]) > 1) {
    stop_no_call(
      'Cannot extract non-scalar value using extract_variable():\n',
      '  "', variable, '" has dimensions: [', paste0(dim(x[[variable]]), collapse = ","), ']\n',
      '  Try including brackets ("[]") and indices in the variable name to extract a scalar value.'
    )
  } else {
    # scalar
    out <- unname(drop(draws_of(x[[variable]])))
  }
  out
}

#' The `draws_rvars` format
#'
#' @name draws_rvars
#' @family formats
#'
#' @templateVar draws_format draws_rvars
#' @templateVar base_class "list"
#' @template draws_format-skeleton
#' @template args-format-nchains
#'
#' @details Objects of class `"draws_rvars"` are lists of [`rvar`] objects.
#' See **Examples**.
#'
NULL

#' @rdname draws_rvars
#' @export
as_draws_rvars <- function(x, ...) {
  UseMethod("as_draws_rvars")
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.default <- function(x, ...) {
  x <- as_draws(x)
  as_draws_rvars(x, ...)
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.draws_rvars <- function(x, ...) {
  x
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.list <- function(x, ...) {
  if (all(vapply(x, is_rvar, logical(1)))) {
    .as_draws_rvars(x, ...)
  } else {
    NextMethod()
  }
}

#' @export
as_draws_rvars.rvar <- function(x, ...) {
  .as_draws_rvars(list(x = x), ...)
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.draws_matrix <- function(x, ...) {
  .as_draws_rvars.draws_matrix(x, ...)
}

#' Helper for as_draws_rvars.draws_matrix and as_draws_rvars.draws_df()
#' @param x_at A function taking a logical vector along variables(x) and returning a matrix of draws
#' @noRd
.as_draws_rvars.draws_matrix <- function(x, ..., x_at = function(var_i) unclass(x[, var_i, drop = FALSE])) {
  .variables <- variables(x, reserved = TRUE)
  if (ndraws(x) == 0) {
    return(empty_draws_rvars(.variables))
  }

  # split x[y,z] names into base name and indices
  #
  #                    ----- base name -> vars_indices[[i]][[2]]
  #                    ||||| lazy-matched (.*? not .*) so that indices match as much as they can
  #                    |||||
  #                    |||||      ---- optional indices -> vars_indices[[i]][[3]]
  #                    |||||      ||||
  matches <- regexec("^(.*?)(?:\\[(.*)\\])?$", .variables)
  vars_indices <- regmatches(.variables, matches)
  vars <- vapply(vars_indices, `[[`, i = 2, character(1))

  # pull out each var into its own rvar
  var_names <- unique(vars)
  rvars_list <- lapply(var_names, function(var) {
    var_i <- vars == var
    var_matrix <- x_at(var_i)
    attr(var_matrix, "nchains") <- NULL
    var_indices <- vars_indices[var_i]

    if (ncol(var_matrix) == 1 && nchar(var_indices[[1]][[3]]) == 0) {
      # single variable, no indices
      out <- rvar(var_matrix)
      dimnames(out) <- NULL
    } else {
      # variable with indices => we need to reshape the array
      # basically, we're going to do a bunch of work up front to figure out
      # a single array slice that does most of the work for us.

      # first, pull out the list of indices into a data frame
      # where each column is an index variable
      indices <- vapply(var_indices, `[[`, i = 3, character(1))
      indices <- as.data.frame(do.call(rbind, strsplit(indices, ",")),
                               stringsAsFactors = FALSE)
      unique_indices <- vector("list", length(indices))
      .dimnames <- vector("list", length(indices))
      names(unique_indices) <- names(indices)
      for (i in seq_along(indices)) {
        numeric_index <- suppressWarnings(as.numeric(indices[[i]]))
        if (!anyNA(numeric_index) && rlang::is_integerish(numeric_index)) {
          # for integer indices, we need to convert them to integers
          # so that we can sort them in numerical order (not string order)
          if (min(numeric_index) >= 1) {
            # integer indices >= 1 are forced to lower bound of 1 + no dimnames
            indices[[i]] <- as.integer(numeric_index)
            unique_indices[[i]] <- seq.int(1, max(numeric_index))
          } else {
            # indices with values < 1 are sorted but otherwise left as-is, and will create dimnames
            indices[[i]] <- numeric_index
            unique_indices[[i]] <- sort(unique(numeric_index))
            .dimnames[[i]] <- unique_indices[[i]]
          }
        } else {
          # we convert non-numeric indices to factors so that we can force them
          # to be ordered as they appear in the data (rather than in alphabetical order)
          factor_levels <- unique(indices[[i]])
          indices[[i]] <- factor(indices[[i]], levels = factor_levels)
          # these aren't sorted so they appear in original order
          unique_indices[[i]] <- factor(factor_levels, levels = factor_levels)
          .dimnames[[i]] <- unique_indices[[i]]
        }
      }

      # sort indices and fill in missing indices as NA to ensure
      # (1) even if the order of the variables is something weird (like the
      # column for x[2,2] comes before x[1,1] in the matrix) the result
      # places those columns in the correct cells of the array
      # (2) if some combination of indices is missing (say x[2,1] isn't
      # in the input) that cell in the array gets an NA

      # Use expand.grid to get all cells in output array. We reverse indices
      # here because it helps us do the sort after the merge, where
      # we need to sort in reverse order of the indices (because
      # the value of the last index should move slowest)
      all_indices <- expand.grid(rev(unique_indices))
      # merge with all.x = TRUE (left join) to fill in missing cells with NA
      indices <- merge(all_indices, cbind(indices, index = seq_len(nrow(indices))),
                       all.x = TRUE, sort = FALSE)
      # need to do the sort manually after merge because when sort = TRUE, merge
      # sorts factors as if they were strings, and we need factors to be sorted as factors
      indices <- indices[do.call(order, as.list(indices[, -ncol(indices), drop = FALSE])),]

      # re-sort the array and fill in missing cells with NA
      var_matrix <- var_matrix[, indices$index, drop = FALSE]

      # convert to rvar and adjust dimensions
      out <- rvar(var_matrix)
      dim(out) <- unname(lengths(unique_indices))
      dimnames(out) <- .dimnames
    }
    out
  })
  names(rvars_list) <- var_names
  out <- .as_draws_rvars(rvars_list, ...)
  .nchains <- nchains(x)
  for (i in seq_along(out)) {
    nchains_rvar(out[[i]]) <- .nchains
  }
  out
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.draws_array <- function(x, ...) {
  as_draws_rvars(as_draws_matrix(x), ...)
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.draws_df <- function(x, ...) {
  x_df <- as.data.frame(x, optional = TRUE)[, variables(x, reserved = TRUE), drop = FALSE]
  data_frame_to_matrix <- function(df) {
    if (any(vapply(df, is.factor, logical(1)))) {
      # as.matrix() does not convert factor columns correctly, must do this ourselves
      while_preserving_dims(
        function(df) do.call(function(...) vctrs::vec_c(..., .name_spec = rlang::zap()), df),
        df
      )
    } else {
      as.matrix(df)
    }
  }
  .as_draws_rvars.draws_matrix(
    x, ..., x_at = function(var_i) data_frame_to_matrix(x_df[, var_i, drop = FALSE])
  )
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.draws_list <- function(x, ...) {
  as_draws_rvars(as_draws_df(x), ...)
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.mcmc <- function(x, ...) {
  as_draws_rvars(as_draws_matrix(x), ...)
}

#' @rdname draws_rvars
#' @export
as_draws_rvars.mcmc.list <- function(x, ...) {
  as_draws_rvars(as_draws_array(x), ...)
}

# try to convert any R object into a 'draws_rvars' object
.as_draws_rvars <- function(x, ...) {
  x <- as.list(x)

  # convert all elements to rvars
  x <- lapply(x, as_rvar)

  # replace blank variable names with defaults
  if (is.null(names(x))) {
    names(x) <- default_variables(length(x))
  } else {
    blank_names <- nchar(names(x)) == 0
    names(x)[blank_names] <- default_variables(length(x))[blank_names]
  }

  check_new_variables(names(x))

  x <- conform_rvar_ndraws_nchains(x)

  class(x) <- class_draws_rvars()
  x
}

#' @rdname draws_rvars
#' @export
draws_rvars <- function(..., .nchains = 1) {
  out <- lapply(list(...), function(x) {
    if (is_rvar(x)) x
    else rvar(x, nchains = .nchains)
  })

  if (!rlang::is_named(out)) {
    stop_no_call("All variables must be named.")
  }

  .as_draws_rvars(out)
}

class_draws_rvars <- function() {
  c("draws_rvars", "draws", "list")
}

#' @rdname draws_rvars
#' @export
is_draws_rvars <- function(x) {
  inherits(x, "draws_rvars")
}

# is an object looking like a 'draws_rvars' object?
is_draws_rvars_like <- function(x) {
  is.list(x) && all(sapply(x, is_rvar))
}

#' @export
`[.draws_rvars` <- function(x, i, j, ..., drop = FALSE) {
  out <- NextMethod("[")
  class(out) <- class(x)
  out
}

# create an empty draws_rvars object
empty_draws_rvars <- function(variables = character(0), nchains = 0) {
  assert_character(variables, null.ok = TRUE)
  assert_number(nchains, lower = 0)
  out <- named_list(variables, rvar())
  class(out) <- class_draws_rvars()
  out
}

test_that("extract_variable works the same for different formats", {
  draws_array <- as_draws_array(example_draws())
  mu_array <- extract_variable(draws_array, "mu")

  draws_df <- as_draws_df(example_draws())
  mu_df <- extract_variable(draws_df, "mu")
  expect_equal(mu_df, mu_array)

  draws_list <- as_draws_list(example_draws())
  mu_list <- extract_variable(draws_list, "mu")
  expect_equal(mu_list, mu_array)

  draws_matrix <- as_draws_matrix(example_draws())
  mu_matrix <- extract_variable(draws_matrix, "mu")
  expect_equal(as.vector(mu_matrix), as.vector(mu_array))

  draws_rvars <- as_draws_rvars(example_draws())
  mu_matrix <- extract_variable(draws_rvars, "mu")
  expect_equal(as.vector(mu_matrix), as.vector(mu_array))
})

test_that("extract_variable works for draws_rvars on an indexed variable", {
  draws_array <- as_draws_array(example_draws())
  theta1_array <- extract_variable(draws_array, "theta[1]")

  draws_rvars <- as_draws_rvars(example_draws())
  theta1_matrix <- extract_variable(draws_rvars, "theta[1]")
  expect_equal(as.vector(theta1_matrix), as.vector(theta1_array))

  expect_error(extract_variable(draws_rvars, "theta"), "Cannot extract non-scalar value")
})

test_that("extract_variable works for factor types", {
  draws_rvars <- draws_rvars(x = rvar_factor(letters, nchains = 2))

  expect_equal(extract_variable(draws_rvars, "x"), factor(letters))
  expect_equal(extract_variable(as_draws_df(draws_rvars), "x"), factor(letters))
  expect_equal(extract_variable(as_draws_list(draws_rvars), "x"), factor(letters))
})

test_that("extract_variable default method works", {
  # it should convert matrix to draws object
  x <- matrix(1:20, nrow = 10, ncol = 2)
  colnames(x) <- c("A", "B")
  expect_equal(extract_variable(x, "A"), 1:10)
  expect_equal(extract_variable(x, "B"), 11:20)
})

test_that("extract_list_of_variable_arrays works for different formats", {
  draws <- list(
    array = as_draws_array(example_draws()),
    df = as_draws_df(example_draws()),
    list = as_draws_list(example_draws()),
    matrix = as_draws_matrix(example_draws()),
    rvars = as_draws_rvars(example_draws())
  )

  variables <- c("mu", "theta")
  
  for (type in names(draws)) {
    result <- extract_list_of_variable_arrays(draws[[type]], variables)
    
    expect_type(result, "list")
    expect_length(result, 2)
    expect_named(result, c("mu", "theta"))
    
    # Compare with individual extractions
    expect_equal(result$mu, extract_variable_array(draws[[type]], "mu"))
    expect_equal(result$theta, extract_variable_array(draws[[type]], "theta"))
  }
})

test_that("extract_list_of_variable_arrays works with indexed variables", {
  x <- example_draws(example = "multi_normal")
  
  # Test specific indexed variables
  vars <- extract_list_of_variable_arrays(x, c("mu[1]", "mu[2]", "Sigma[1,1]"))
  
  expect_type(vars, "list")
  expect_length(vars, 3)
  expect_named(vars, c("mu[1]", "mu[2]", "Sigma[1,1]"))
  
  # Compare with individual extractions
  expect_equal(vars$`mu[1]`, extract_variable_array(x, "mu[1]"))
  expect_equal(vars$`mu[2]`, extract_variable_array(x, "mu[2]"))
  expect_equal(vars$`Sigma[1,1]`, extract_variable_array(x, "Sigma[1,1]"))
})

test_that("extract_list_of_variable_arrays works with base variable names", {
  x <- example_draws(example = "multi_normal")
  
  vars <- extract_list_of_variable_arrays(x, c("mu", "Sigma"))
  
  expect_type(vars, "list")
  expect_length(vars, 2)
  expect_named(vars, c("mu", "Sigma"))
  
  # Compare with individual extractions
  expect_equal(vars$mu, extract_variable_array(x, "mu"))
  expect_equal(vars$Sigma, extract_variable_array(x, "Sigma"))
})

test_that("extract_list_of_variable_arrays handles empty input", {
  x <- example_draws()
  
  result <- extract_list_of_variable_arrays(x, character(0))
  
  expect_type(result, "list")
  expect_length(result, 0)
})

test_that("extract_list_of_variable_arrays validates input", {
  x <- example_draws()
  
  expect_error(
    extract_list_of_variable_arrays(x, 123),
    "'variables' must be a character vector."
  )
  
  expect_error(
    extract_list_of_variable_arrays(x, list("mu", "theta")),
    "'variables' must be a character vector."
  )
})

test_that("extract_list_of_variable_arrays works for factor types", {
  draws_rvars <- draws_rvars(
    y = rvar(1:26, nchains = 2), 
    x = rvar_factor(letters, nchains = 2)
  )
  
  vars <- extract_list_of_variable_arrays(draws_rvars, c("x", "y"))
  
  expect_type(vars, "list")
  expect_length(vars, 2)
  expect_named(vars, c("x", "y"))
  
  # x should be a factor array
  expect_s3_class(vars$x, "factor")
  expect_equal(levels(vars$x), letters)
  
  # Compare with individual extractions
  expect_equal(vars$x, extract_variable_array(draws_rvars, "x"))
  expect_equal(vars$y, extract_variable_array(draws_rvars, "y"))
})

test_that("extract_list_of_variable_arrays passes additional arguments", {
  # This test assumes extract_variable_array supports additional arguments
  # We'll test that the ... arguments are properly passed through
  x <- example_draws()
  
  # This should work the same whether called individually or in batch
  individual_results <- list(
    mu = extract_variable_array(x, "mu"),
    theta = extract_variable_array(x, "theta")
  )
  
  batch_results <- extract_list_of_variable_arrays(x, c("mu", "theta"))
  
  expect_equal(batch_results, individual_results)
})

test_that("extract_list_of_variable_arrays works with default method", {
  # Test that the default method properly converts and calls the draws method
  x_array <- as.array(example_draws())
  
  result <- extract_list_of_variable_arrays(x_array, c("mu", "theta"))
  
  expect_type(result, "list")
  expect_length(result, 2)
  expect_named(result, c("mu", "theta"))
})


test_that("extract_list_of_variable_arrays works with NULL variables (extract all base variables)", {
  x <- example_draws(example = "multi_normal")
  
  # Test with NULL (should extract base variables only, not indexed ones)
  all_vars <- extract_list_of_variable_arrays(x)
  
  # Get expected base variable names
  all_individual_vars <- variables(x)
  expected_base_vars <- unique(split_variable_names(all_individual_vars)$base_name)  # Fixed: base_name not base_names
  
  expect_type(all_vars, "list")
  expect_length(all_vars, length(expected_base_vars))
  expect_named(all_vars, expected_base_vars)
  
  # Should match explicit specification of base variables
  explicit_vars <- extract_list_of_variable_arrays(x, expected_base_vars)
  expect_identical(all_vars, explicit_vars)
  
  # Compare with extracting each base variable individually
  for (var in expected_base_vars) {
    expect_equal(all_vars[[var]], extract_variable_array(x, var))
  }
  
  # Test with explicit NULL
  all_vars_explicit <- extract_list_of_variable_arrays(x, variables = NULL)
  expect_equal(all_vars, all_vars_explicit)
  
  # Verify we get base variables, not indexed ones
  expect_true("mu" %in% names(all_vars))
  expect_true("Sigma" %in% names(all_vars))  
  expect_false("mu[1]" %in% names(all_vars))
  expect_false("Sigma[1,1]" %in% names(all_vars))
})

#' Coerce to a random variable
#'
#' Convert `x` to an [`rvar`] object.
#'
#' @param x (multiple options) An object that can be converted to an [`rvar`],
#'   such as a vector, array, or an [`rvar`] itself.
#' @template args-rvar-dim
#' @template args-rvar-dimnames
#' @param nchains (positive integer) The number of chains. The default is `1`.
#'
#' @details For objects that are already [`rvar`]s, returns them (with modified dimensions
#' if `dim` is not `NULL`).
#'
#' For numeric or logical vectors or arrays, returns an [`rvar`] with a single draw and
#' the same dimensions as `x`. This is in contrast to the [rvar()] constructor, which
#' treats the first dimension of `x` as the draws dimension. As a result, `as_rvar()`
#' is useful for creating constants.
#'
#' While `as_rvar()` attempts to pick the most suitable subtype of [`rvar`] based on the
#' type of `x` (possibly returning a an [`rvar_factor`] or [`rvar_ordered`]),
#' `as_rvar_numeric()` always coerces the draws of the output [`rvar`] to be [`numeric`],
#' and always returns a base [`rvar`], never a subtype.
#'
#' @seealso [rvar()] to construct [`rvar`]s directly.  See [rdo()], [rfun()], and
#' [rvar_rng()] for higher-level interfaces for creating `rvar`s.
#'
#' @return An object of class `"rvar"` (or one of its subtypes) representing a random variable.
#'
#' @examples
#'
#' # You can use as_rvar() to create "constant" rvars (having only one draw):
#' x <- as_rvar(1)
#' x
#'
#' # Such constants can be of arbitrary shape:
#' as_rvar(1:4)
#' as_rvar(matrix(1:10, nrow = 5))
#' as_rvar(array(1:12, dim = c(2, 3, 2)))
#'
#' # as_rvar_numeric() coerces subtypes of rvar to the base rvar type
#' y <- as_rvar_factor(c("a", "b", "c"))
#' y
#' as_rvar_numeric(y)
#'
#' @export
as_rvar <- function(x, dim = NULL, dimnames = NULL, nchains = NULL) {
  .as_rvar(x, dim = dim, dimnames = dimnames, nchains = nchains)
}
.as_rvar <- function(x, dim = NULL, dimnames = NULL, nchains = NULL, ptype = new_rvar()) {
  out <- x

  if (is.null(out)) {
    out <- rvar()
  } else {
    out <- vec_cast(out, ptype)
  }

  if (!is.null(dim)) {
    dim(out) <- dim
  } else if (is.null(dimnames) && is.vector(x)) {
    # for non-vector-like input (matrices, arrays, etc), vec_cast should
    # have already copied over the dimnames correctly. For vector-like input,
    # it doesn't; so as long as the `dim` argument isn't set we can copy
    # the name over
    names(out) <- names(x)
  }
  if (!is.null(dimnames)) {
    dimnames(out) <- dimnames
  }

  if (!is.null(nchains)) {
    .ndraws <- ndraws(out)
    nchains <- as_one_integer(nchains)
    check_nchains_compat_with_ndraws(nchains, .ndraws)
    nchains_rvar(out) <- nchains
  }

  out
}

#' @rdname as_rvar
#' @export
as_rvar_numeric <- function(x, dim = NULL, dimnames = NULL, nchains = NULL) {
  out <- as_rvar(x, dim = dim, dimnames = dimnames, nchains = nchains)
  draws_of(out) <- while_preserving_dims(as.numeric, draws_of(out))
  out
}


# type predicates --------------------------------------------------

#' Is `x` a random variable?
#'
#' Test if `x` is an [`rvar`].
#'
#' @param x (any object) An object to test.
#'
#' @seealso [as_rvar()] to convert objects to `rvar`s.
#'
#' @return `TRUE` if `x` is an [`rvar`], `FALSE` otherwise.
#'
#' @export
is_rvar <- function(x) {
  inherits(x, "rvar")
}

#' @export
is.matrix.rvar <- function(x) {
  length(dim(draws_of(x))) == 3
}

#' @export
is.array.rvar <- function(x) {
  length(dim(draws_of(x))) > 0
}


# type conversion ---------------------------------------------------------

#' @export
as.vector.rvar <- function(x, mode = "any") {
  dim(x) <- NULL
  names(x) <- NULL
  x
}

#' @export
as.list.rvar <- function(x, ...) {
  x_dim <- dim(x)

  if (length(x_dim) >= 2) {
    is <- seq_len(x_dim[[1]])
    names(is) <- dimnames(x)[[1]]
    out <- lapply(is, function(i) {
      out_i <- x[i,]
      .dim <- dim(out_i)
      .dimnames <- dimnames(out_i)
      dim(out_i) <- .dim[-1]
      dimnames(out_i) <- .dimnames[-1]
      out_i
    })
  } else {
    is <- seq_along(x)
    names(is) <- dimnames(x)[[1]]
    out <- lapply(is, function(i) x[[i]])
  }
  out
}

#' @importFrom rlang as_label
#' @export
as.data.frame.rvar <- function(x, ..., optional = FALSE) {
  out <- as.data.frame.array(x, ..., optional = optional)
  if (length(dim(x)) <= 1 && !optional) {
    names(out) <- as_label(substitute(x))
  }
  out
}

#' @importFrom tibble as_tibble
#' @export
as_tibble.rvar <- function(x, ...) {
  #default name for vectors is `value` with as_tibble
  value <- x
  as_tibble(as.data.frame(value, optional = FALSE), ...)
}


# vctrs proxy / restore --------------------------------------------------------

invalidate_rvar_cache = function(x) {
  attr(x, "cache") <- new.env(parent = emptyenv())
  x
}

#' @importFrom vctrs vec_proxy
#' @export
vec_proxy.rvar = function(x, ...) {
  # Using caching to help with algorithms that call vec_proxy
  # repeatedly. See https://github.com/r-lib/vctrs/issues/1411

  out <- attr(x, "cache")$vec_proxy
  if (is.null(out)) {
    # proxy is not in the cache, calculate it and store it in the cache
    out <- make_rvar_proxy(x)
    attr(x, "cache")$vec_proxy <- out
  }

  out
}

#' Make a cacheable proxy for vec_proxy.rvar
#' @noRd
make_rvar_proxy = function(x) {
  nchains <- nchains(x)
  draws <- draws_of(x)
  is <- seq_len(NROW(x))
  names(is) <- rownames(x)
  lapply(is, function(i) {
    list(
      index = i,
      nchains = nchains,
      draws = draws
    )
  })
}


#' @importFrom vctrs vec_restore
#' @export
vec_restore.rvar <- function(x, ...) {
  if (length(x) == 0) return(rvar())

  # need to handle the case of creating NAs from NULL entries so that
  # vec_init() works properly: vec_init requires vec_slice(x, NA_integer_)
  # to give you back NA values, but this breaks because we use lists as proxies.
  # When using a list as a proxy, a proxy entry in `x` that is equal to NULL
  # actually corresponds to an NA value due to the way that list indexing
  # works: when you do something like list()[c(NA_integer_,NA_integer_)]
  # you get back list(NULL, NULL), but when you do something like
  # double()[c(NA_integer_,NA_integer_)] you get back c(NA, NA).
  # So we have to make the NULL values be NA values to mimic vector indexing.
  # N.B. could potentially do this with vec_cast as well (as long as the first
  # dimension is the slicing index)
  x[lengths(x) == 0] <- make_rvar_proxy(new_rvar(NA_real_))

  # find runs where the same underlying draws are in the proxy
  different_draws_from_previous <- vapply(seq_along(x)[-1], FUN.VALUE = logical(1), function(i) {
    !identical(x[[i]]$draws, x[[i - 1]]$draws) || !identical(x[[i]]$nchains, x[[i - 1]]$nchains)
  })
  draws_groups <- cumsum(c(TRUE, different_draws_from_previous))

  # convert each run into a slice on an rvar and bind the resulting rvars together
  groups <- split(x, draws_groups)
  rvars <- lapply(groups, function(x) {
    i <- vapply(x, `[[`, "index", FUN.VALUE = numeric(1))
    rvar <- new_rvar(x[[1]]$draws, .nchains = x[[1]]$nchains)
    if (length(dim(rvar)) > 1) {
      rvar[i, ]
    } else {
      rvar <- rvar[i]
      .dimnames <- dimnames(rvar)
      dim(rvar) <- c(length(rvar), 1)
      dimnames(rvar) <- c(.dimnames, NULL)
      rvar
    }
  })
  out <- bind_rvars(rvars, arg_exprs = NULL, deparse.level = 0, axis = 1)

  if (all(lengths(lapply(groups, function(x) dim(x[[1]]$draws))) <= 2)) {
    # input was a bunch of vectors, ensure output is also a vector
    .dimnames <- dimnames(out)
    dim(out) <- length(out)
    dimnames(out) <- .dimnames[1]
  }

  # since we've already spent time calculating it, save the proxy in the cache
  # - but only if the proxy only has one group (else we'd have to recalculate
  # the bind above again, which is usually more expensive than generating the
  # proxy itself)
  if (length(groups) == 1) {
    attr(out, "cache")$vec_proxy <- x
  }

  out
}

#' @export
vec_restore.rvar_factor = function(x, to, ...) {
  x[lengths(x) == 0] <- make_rvar_proxy(rvar_factor(NA_integer_))
  vec_restore.rvar(x, ...)
}

#' @export
vec_restore.rvar_ordered = function(x, to, ...) {
  x[lengths(x) == 0] <- make_rvar_proxy(rvar_ordered(NA_integer_))
  vec_restore.rvar(x, ...)
}


# vec_ptype performance generics -------------------------------------------

#' @importFrom vctrs vec_ptype
#' @export
vec_ptype.rvar <- function(x, ..., x_arg = "") new_rvar()
#' @export
vec_ptype.rvar_factor <- function(x, ..., x_arg = "") new_rvar(factor())
#' @export
vec_ptype.rvar_ordered <- function(x, ..., x_arg = "") new_rvar(ordered(NULL, levels = levels(x)))


# identity casts -----------------------------------------------------------

#' @importFrom vctrs vec_ptype2
#' @importFrom vctrs vec_cast
#' @export
vec_ptype2.rvar.rvar <- function(x, y, ...) new_rvar()
#' @export
vec_cast.rvar.rvar <- function(x, to, ...) x

#' @export
vec_ptype2.rvar_factor.rvar_factor <- function(x, y, ...) new_rvar(factor())
#' @export
vec_cast.rvar_factor.rvar_factor <- function(x, to, ...) x

#' @export
vec_ptype2.rvar_ordered.rvar_ordered <- function(x, y, ...) new_rvar(ordered(NULL))
#' @export
vec_cast.rvar_ordered.rvar_ordered <- function(x, to, ...) x


# numeric and logical casts -----------------------------------------------

# double -> rvar
#' @export
vec_ptype2.double.rvar <- function(x, y, ...) new_rvar()
#' @export
vec_ptype2.rvar.double <- function(x, y, ...) new_rvar()
#' @export
vec_cast.rvar.double <- function(x, to, ...) new_constant_rvar(x)

# double -> rvar_factor
#' @export
vec_cast.rvar_factor.double <- function(x, to, ...) new_constant_rvar(while_preserving_dims(as.factor, x))

# double -> rvar_ordered
#' @export
vec_cast.rvar_ordered.double <- function(x, to, ...) new_constant_rvar(while_preserving_dims(as.ordered, x))

# integer -> rvar
#' @export
vec_ptype2.integer.rvar <- function(x, y, ...) new_rvar()
#' @export
vec_ptype2.rvar.integer <- function(x, y, ...) new_rvar()
#' @export
vec_cast.rvar.integer <- function(x, to, ...) new_constant_rvar(x)

# integer -> rvar_factor
#' @export
vec_cast.rvar_factor.integer <- function(x, to, ...) new_constant_rvar(while_preserving_dims(as.factor, x))

# integer -> rvar_ordered
#' @export
vec_cast.rvar_ordered.integer <- function(x, to, ...) new_constant_rvar(while_preserving_dims(as.ordered, x))

# logical -> rvar
#' @export
vec_ptype2.logical.rvar <- function(x, y, ...) new_rvar()
#' @export
vec_ptype2.rvar.logical <- function(x, y, ...) new_rvar()
#' @export
vec_cast.rvar.logical <- function(x, to, ...) new_constant_rvar(x)

# logical -> rvar_factor
#' @export
vec_cast.rvar_factor.logical <- function(x, to, ...) new_constant_rvar(while_preserving_dims(as.factor, x))

# logical -> rvar_ordered
#' @export
vec_cast.rvar_ordered.logical <- function(x, to, ...) new_constant_rvar(while_preserving_dims(as.ordered, x))


# character casts ---------------------------------------------------------

# rvar_[factor|ordered] -> character
#' @export
vec_cast.character.rvar <- function(x, to, ...) format(x)
#' @export
vec_cast.character.rvar_factor <- function(x, to, ...) format(x)
#' @export
vec_cast.character.rvar_ordered <- function(x, to, ...) format(x)

# character -> rvar
#' @export
vec_cast.rvar.character <- function(x, to, ...) new_constant_rvar(while_preserving_dims(as.factor, x))

# character -> rvar_factor
#' @export
vec_ptype2.character.rvar_factor <- function(x, y, ...) new_rvar(factor())
#' @export
vec_ptype2.rvar_factor.character <- function(x, y, ...) new_rvar(factor())
#' @export
vec_cast.rvar_factor.character <- function(x, to, ...) new_constant_rvar(while_preserving_dims(as.factor, x))

# character -> rvar_ordered
#' @export
vec_ptype2.character.rvar_ordered <- function(x, y, ...) rvar_ordered(levels = levels(y))
#' @export
vec_ptype2.rvar_ordered.character <- function(x, y, ...) rvar_ordered(levels = levels(x))
#' @export
vec_cast.rvar_ordered.character <- function(x, to, ...) {
  old_levels <- levels(to)
  new_levels <- sort(setdiff(x, levels(to)))
  levels <- c(old_levels, new_levels)
  ordered <- length(new_levels) == 0
  new_constant_rvar(while_preserving_dims(factor, x, levels = levels, ordered = ordered))
}


# factor casts ---------------------------------------------------------

# factor -> rvar
#' @export
vec_cast.rvar.factor <- function(x, to, ...) new_constant_rvar(x)

# factor -> rvar_factor
#' @export
vec_ptype2.factor.rvar_factor <- function(x, y, ...) new_rvar(factor())
#' @export
vec_ptype2.rvar_factor.factor <- function(x, y, ...) new_rvar(factor())
#' @export
vec_cast.rvar_factor.factor <- function(x, to, ...) new_constant_rvar(x)

# factor -> rvar_ordered
#' @export
vec_ptype2.factor.rvar_ordered <- function(x, y, ...) new_rvar(factor())
#' @export
vec_ptype2.rvar_ordered.factor <- function(x, y, ...) new_rvar(factor())
#' @export
vec_cast.rvar_ordered.factor <- function(x, to, ...) new_constant_rvar(x)

# ordered -> rvar
#' @export
vec_cast.rvar.ordered <- function(x, to, ...) new_constant_rvar(x)

# ordered -> rvar_factor
#' @export
vec_ptype2.ordered.rvar_factor <- function(x, y, ...) new_rvar(ordered(NULL))
#' @export
vec_ptype2.rvar_factor.ordered <- function(x, y, ...) new_rvar(ordered(NULL))
#' @export
vec_cast.rvar_factor.ordered <- function(x, to, ...) new_constant_rvar(x)

# ordered -> rvar_ordered
#' @export
vec_ptype2.ordered.rvar_ordered <- function(x, y, ...) new_rvar(ordered(NULL))
#' @export
vec_ptype2.rvar_ordered.ordered <- function(x, y, ...) new_rvar(ordered(NULL))
#' @export
vec_cast.rvar_ordered.ordered <- function(x, to, ...) new_constant_rvar(x)


# subtype casts -----------------------------------------------------------

#' @export
vec_cast.rvar_factor.rvar <- function(x, to, ...) .rvar_to_rvar_factor(x)
#' @export
vec_cast.rvar_ordered.rvar <- function(x, to, ...) .rvar_to_rvar_factor(x, ordered = TRUE)
#' @export
vec_cast.rvar_factor.rvar_ordered <- function(x, to, ...) .rvar_to_rvar_factor(x)
#' @export
vec_cast.rvar_ordered.rvar_factor <- function(x, to, ...) .rvar_to_rvar_factor(x, ordered = TRUE)
#' @export
vec_cast.rvar.rvar_ordered <- function(x, to, ...) x
#' @export
vec_cast.rvar.rvar_factor <- function(x, to, ...) x

.rvar_to_rvar_factor <- function(x, ordered = FALSE, ...) {
  if (
    ...length() == 0 &&
    ((ordered && is_rvar_ordered(x)) || (!ordered && is_rvar_factor(x)))
  ) {
    # already correct type and nothing is being passed to factor() to change it
    return(x)
  }

  draws_of(x) <- while_preserving_dims(factor, draws_of(x), ordered = ordered, ...)
  x
}


# casting between rvar and distribution objects ---------------------------

#' @export
vec_ptype2.distribution.rvar <- function(x, y, ...) x

#' @export
vec_ptype2.rvar.distribution <- function(x, y, ...) x

#' @export
vec_cast.rvar.distribution <- function(x, to, ..., x_arg = "", to_arg = "") {
  x_list <- vctrs::vec_data(x)
  if (length(dim(to)) > 1 || !is_dist_sample_list(x_list)) {
    vctrs::stop_incompatible_cast(x, to, x_arg = x_arg, to_arg = to_arg)
  }
  x_rvar_list <- lapply(x_list, function(x) rvar(vctrs::field(x, 1)))
  do.call(c, x_rvar_list)
}

#' @export
vec_cast.distribution.rvar <- function(x, to, ..., x_arg = "", to_arg = "") {
  if (length(dim(x)) > 1) {
    vctrs::stop_incompatible_cast(x, to, x_arg = x_arg, to_arg = to_arg)
  }
  .draws <- draws_of(x)
  x_array_list <- vctrs::vec_chop(aperm(.draws, c(2, 1, seq_along(dim(.draws))[c(-1,-2)])))
  x_vector_list <- lapply(x_array_list, as.vector)
  names(x_vector_list) <- names(x)
  distributional::dist_sample(x_vector_list)
}


# helpers: casting --------------------------------------------------------

# create a constant rvar based on x (a double, logical, or integer)
new_constant_rvar <- function(x) {
  out <- x
  dim_x <- dim(x)
  if (length(dim_x) == 0) {
    dim(out) <- c(1, length(x))
  } else {
    dim(out) <- c(1, dim_x)
    dim_i <- seq_along(dim_x)
    out <- copy_dimnames(x, dim_i, out, dim_i + 1)
  }
  new_rvar(out)
}

# is this a list of dist_sample()s?
is_dist_sample_list <- function(x) {
  all(vapply(x, inherits, logical(1), "dist_sample"))
}

#' The `draws_list` format
#'
#' @name draws_list
#' @family formats
#'
#' @templateVar draws_format draws_list
#' @templateVar base_class "list"
#' @template draws_format-skeleton
#' @template args-format-nchains
#'
#' @details Objects of class `"draws_list"` are lists with one element per MCMC
#'   chain. Each of these elements is itself a named list of numeric vectors
#'   with one vector per variable. The length of each vector is equal to the
#'   number of saved iterations per chain. See **Examples**.
#'
NULL

#' @rdname draws_list
#' @export
as_draws_list <- function(x, ...) {
  UseMethod("as_draws_list")
}

#' @rdname draws_list
#' @export
as_draws_list.default <- function(x, ...) {
  x <- as_draws(x)
  as_draws_list(x, ...)
}

#' @rdname draws_list
#' @export
as_draws_list.draws_list <- function(x, ...) {
  x
}

#' @rdname draws_list
#' @export
as_draws_list.draws_matrix <- function(x, ...) {
  x <- as_draws_df(x)
  as_draws_list(x, ...)
}

#' @rdname draws_list
#' @export
as_draws_list.draws_array <- function(x, ...) {
  x <- as_draws_df(x)
  as_draws_list(x, ...)
}

#' @rdname draws_list
#' @export
as_draws_list.draws_df <- function(x, ...) {
  if (ndraws(x) == 0) {
    return(empty_draws_list(variables(x)))
  }
  out <- named_list(chain_ids(x))
  x <- x[order(x$.draw), ]
  for (i in seq_along(out)) {
    out[[i]] <- subset(x, chain = i)
    out[[i]] <- remove_reserved_df_variables(out[[i]])
    out[[i]] <- as.list(out[[i]])
  }
  class(out) <- class_draws_list()
  out
}

#' @rdname draws_list
#' @export
as_draws_list.draws_rvars <- function(x, ...) {
  as_draws_list(as_draws_df(x), ...)
}

#' @rdname draws_list
#' @export
as_draws_list.mcmc <- function(x, ...) {
  as_draws_list(as_draws_matrix(x), ...)
}

#' @rdname draws_list
#' @export
as_draws_list.mcmc.list <- function(x, ...) {
  as_draws_list(as_draws_array(x), ...)
}

# try to convert any R object into a 'draws_list' object
.as_draws_list <- function(x) {
  x <- as.list(x)
  # check heuristically if a list of a single chain is supplied
  if (is.numeric(x[[1]])) {
    x <- list(x)
  }
  if (any(!ulapply(x, is.list))) {
    stop_no_call("All list elements must be lists themselves.")
  }
  if (length(unique(lengths(x))) != 1L) {
    stop_no_call("All list elements must have the same length.")
  }
  if (is.null(names(x[[1]]))) {
    # no variable names provided; using default names
    variables <- default_variables(length(x[[1]]))
    for (i in seq_along(x)) {
      names(x[[i]]) <- variables
    }
  }
  variables <- names(x[[1]])
  check_new_variables(variables)
  niterations <- length(x[[1]][[1]])
  for (i in seq_along(x)) {
    if (!all(names(x[[i]]) == variables)) {
      stop_no_call("Variables in all chains must have the same names.")
    }
    for (j in seq_along(x[[i]])) {
      if (length(x[[i]][[j]]) != niterations) {
        stop_no_call("All variables in all chains must have the same length.")
      }
    }
    # data.frames pass as lists but may not have the same behavior later on
    if (is.data.frame(x[[i]])) {
      x[[i]] <- as.list(x[[i]])
    }
  }
  names(x) <- as.character(seq_along(x))
  class(x) <- class_draws_list()
  x
}

#' @rdname draws_list
#' @export
draws_list <- function(..., .nchains = 1) {
  out <- draws_df(..., .nchains = .nchains)
  as_draws_list(out)
}

class_draws_list <- function() {
  c("draws_list", "draws", "list")
}

#' @rdname draws_list
#' @export
is_draws_list <- function(x) {
  inherits(x, "draws_list")
}

# is an object looking like a 'draws_list' object?
is_draws_list_like <- function(x) {
  # TODO: add more sophisticated checks
  is.list(x)
}

#' @export
`[.draws_list` <- function(x, i) {
  out <- NextMethod("[")
  class(out) <- class(x)
  out
}

# create an empty draws_list object
empty_draws_list <- function(variables = character(0), nchains = 0) {
  assert_character(variables, null.ok = TRUE)
  assert_number(nchains, lower = 0)
  out <- named_list(seq_len(nchains))
  for (i in seq_along(out)) {
    out[[i]] <- named_list(variables, numeric(0))
  }
  class(out) <- class_draws_list()
  out
}

# helpers for manipulating array indices embedded in variable names


# flattening dimensions into variable names with indices -----------------------

#' Flatten the indices of an array-like object.
#' @param x an array or array-like object (e.g. an rvar)
#' @param base_name a base name which, if provided, is prepended to the
#'   resulting index names.
#' @returns a character vector of the same length as `x`, giving index names
#'   names of the form `"base_name[i,j,k,...]"` where `i`, `j`, `k`, ... are indices
#'   (either numeric indices or strings if x has `dimnames`) that correspond to
#'   the indices of that element in the input `x`.
#' @noRd
flatten_indices <- function(x, base_name = "") {
  # determine new dimension names in the form x,y,z
  # start with numeric names
  .length <- length(x)
  .dim <- dim(x) %||% .length
  .dimnames <- dimnames(x)
  dimname_lists <- lapply(.dim, seq_len)
  if (!is.null(.dimnames)) {
    # where character names are provided, use those instead of the numeric names
    dimname_lists = lapply(seq_along(dimname_lists), function(i) .dimnames[[i]] %||% dimname_lists[[i]])
  }
  # expand out the dimname lists into the appropriate combinations and assemble into new names
  dimname_grid <- expand.grid(dimname_lists, KEEP.OUT.ATTRS = FALSE, stringsAsFactors = FALSE)
  new_names <- do.call(paste, c(list(sep = ","), dimname_grid))

  # prepend the base_name
  if (isTRUE(.dim == 1)) {
    # scalar, just use the base name
    new_names <- base_name
  } else if (.length >= 1) {
    # non-zero length, put the indices in brackets
    new_names <- paste0(base_name, "[", new_names, "]")
  }

  new_names
}

#' Flatten an array so that it is one-dimensional, setting the name of each
#' element of the output array to match its indices in the input array.
#' @param x an array or array-like object (e.g. an rvar)
#' @param base_name a base name which, if provided, is prepended to the
#'   resulting dimension names.
#' @returns a modified array-like object with `dim(out) == length(x)` and, so
#'   long as the input is not scalar, names of the form `"i,j,k,..."` where
#'   `i`, `j`, `k`, ... are indices from the input `x`. If `base_name` is provided,
#'   names of `out` will have the form `"base_name[i,j,k,...]"`.
#' @noRd
flatten_array <- function(x, base_name = "") {
  new_names <- flatten_indices(x, base_name)
  dim(x) <- length(x)
  names(x) <- new_names
  x
}

#' Fast conversion of rvar draws to a flattened data frame. Equivalent to
#' as.data.frame(draws_of(flatten_array(x, name))) except it works with
#' factor rvars (as.data.frame does not work on array-like factors)
#' @noRd
flatten_rvar_draws_to_df <- function(x, base_name = "") {
  if (length(x) == 0) {
    out <- data.frame(row.names = draw_ids(x))
  } else {
    draws <- draws_of(flatten_array(x, base_name))
    cols <- lapply(seq_len(ncol(draws)), function(i) unname(draws[, i]))
    names(cols) <- colnames(draws)
    out <- vctrs::new_data_frame(cols)
  }
  out
}


# splitting flattened variable names into base names and indices ---------------

#' Given a vector of variable names possibly containing indices, split them into
#' a structure representing the base name and the indices
#' @param x a character vector of names, possibly containing indices;
#' e.g. `c("x", "y[1,2]", NA)`
#' @returns a data frame with `length(x)` rows and these columns:
#'  - `base_name`: character vector of base names;
#'     e.g. `c("x", "y", NA)`
#'  - `indices`: character vector of the index strings if present;
#'     e.g. `c("", "[1,2]", NA)`
#' If an input element of `x` is `NA`, the corresponding entries in the returned
#' data frame will also be `NA`.
#' @noRd
split_variable_names <- function(x) {
  # split x[y,z] names into base name and indices
  # we do this with a regex matching just the indices at the end instead of a
  # single regex matching the entire string with multiple capture groups
  # --- something like regexec("^(.*?)(\\[(.*)\\])?$", x) --- because
  # regexec() + regmatches() is slow on a very large number of variables
  # (because it uses lists), but regexpr() + substr() is vectorized
  indices_matches <- regexpr("(\\[.*\\])?$", x)

  vctrs::new_data_frame(list(
    base_name = substr(x, 1L, indices_matches - 1L),
    indices = substr(x, indices_matches, .Machine$integer.max)
  ))
}

#' Given a vector of index strings (such as returned by
#' `split_variable_names(x)$indices`), split each index string into a character
#' vector of indices.
#' @param x a character vector of index strings;
#' e.g. `c("", "[1,2]", NA)`
#' @returns a list of character vectors of indices;
#' e.g. list(character(), c("1", "2"), NA)
#' @noRd
split_indices <- function(x) {
  strsplit(substr(x, 2, nchar(x) - 1), ",", fixed = TRUE)
}


# manipulating flattened variable names -----------------------------------

#' Given a vector of names possibly containing indices, return the base names
#' @param x a character vector of names, possibly containing indices (e.g. `"x[1,2]"`)
#' @returns a character vector of all unique base names in `x`
#' @noRd
base_names <- function(x) {
  unique(split_variable_names(x)$base_name)
}

#' Given a vector of names possibly containing indices, modify the base names
#' @param x a character vector of names, possibly containing indices (e.g. `"x[1,2]"`)
#' @param value a character vector of replacement base names of `length(base_names(x))`.
#' @returns a character vector of `length(x)` with the base names in the original
#'   strings replaced with `value`.
#' @noRd
`base_names<-` <- function(x, value) {
  vars <- split_variable_names(x)
  base_names <- unique(vars$base_name)
  if (length(value) != length(base_names)) {
    stop_no_call(
      "Attempting to replace variables with ", length(base_names), " unique base names",
      " with ", length(value), " new base names. Lengths must match.\n",
      " - Existing base names: ", toString(base_names, width = 40), "\n",
      " - Replacements:        ", toString(value, width = 40)
    )
  }
  base_name_i <- match(vars$base_name, base_names)
  paste0(value[base_name_i], vars$indices)
}

#' Given a vector of variable names, either return the variable names or the
#' base names of the variables, depending on the value of `with_indices`
#' @noRd
variable_names <- function(x, with_indices = TRUE) {
  with_indices <- as_one_logical(with_indices)
  if (with_indices) {
    x
  } else {
    base_names(x)
  }
}

#' Given a vector of variable names, either replace the variable names or the
#' base names of the variables, depending on the value of `with_indices`
#' @noRd
`variable_names<-` <- function(x, with_indices = TRUE, value) {
  with_indices <- as_one_logical(with_indices)
  if (with_indices) {
    value
  } else {
    base_names(x) <- value
    x
  }
}

#' Split Chains
#'
#' Split chains by halving the number of iterations per chain and doubling the
#' number of chains.
#'
#' @template args-methods-x
#' @template args-methods-dots
#' @template return-draws
#'
#' @examples
#' x <- example_draws()
#' niterations(x)
#' nchains(x)
#'
#' x <- split_chains(x)
#' niterations(x)
#' nchains(x)
#'
#' @export
split_chains <- function(x, ...) {
  UseMethod("split_chains")
}

#' @export
split_chains.draws <- function(x, ...) {
  niter <- niterations(x)
  if (niter %% 2 != 0) {
    warning_no_call(
      "Number of iterations is not even. Removing the last iteration ",
      "in order to split chains into two parts of equal length."
    )
    niter <- niter - 1
  }
  iter_first_half <- seq_len(floor(niter / 2))
  iter_second_half <- seq_len(niter)[-iter_first_half]
  x_first_half <- .subset_draws(x, iteration = iter_first_half)
  x_second_half <- .subset_draws(x, iteration = iter_second_half)
  bind_draws(x_first_half, x_second_half, along = "chain")
}


# internal ----------------------------------------------------------------

# split chains for use in convergence diagnostics
# @param x matrix of draws (iterations x chains)
# @return matrix of draws with split chains
.split_chains <- function(x) {
  x <- as.matrix(x)
  niter <- NROW(x)
  if (niter == 1L) {
    return(x)
  }
  half <- niter / 2
  cbind(x[1:floor(half), ], x[ceiling(half + 1):niter, ])
}

#' Tools for working with posterior (and prior) distributions
#'
#' @docType package
#' @name posterior-package
#' @aliases posterior
#'
#' @import checkmate
#' @import stats
#'
#' @description
#' \if{html}{
#'    \figure{stanlogo.png}{options: width="50"}
#'    \url{https://mc-stan.org/posterior/}
#' }
#'
#' The \pkg{posterior} package is intended to provide useful tools
#' for both users and developers of packages for fitting Bayesian models or
#' working with output from Bayesian models. The primary goals of the package
#' are to:
#' * Efficiently convert between many different useful formats of
#' draws (samples) from posterior or prior distributions.
#' * Provide consistent methods for operations commonly performed on draws,
#' for example, subsetting, binding, or mutating draws.
#' * Provide various summaries of draws in convenient formats.
#' * Provide lightweight implementations of state of the art posterior inference
#' diagnostics.
#'
#' @section Package options:
#'
#' The following options are used to format and print [`draws`] objects,
#' as in `print.draws_array()`, `print.draws_df()`, `print.draws_list()`,
#' `print.draws_matrix()`, and `print.draws_rvars()`:
#'
#' * `posterior.max_draws`: Maximum number of draws to print.
#' * `posterior.max_iterations`: Maximum number of iterations to print.
#' * `posterior.max_chains`: Maximum number of chains to print.
#' * `posterior.max_variables`: Maximum number of variables to print.
#'
#' The following options are used to format and print [`rvar`] objects,
#' as in `print.rvar()` and `print.draws_rvars()`:
#'
#' * `posterior.rvar_summary`: What style of summary to display:
#' `"mean_sd"` displays `mean ± sd`, `"median_mad"` displays `median ± mad`.
#' * `posterior.digits`: How many significant digits are displayed. This
#' defaults to a smaller value (`2`) than `getOption("digits")` because
#' [`rvar`]s print two numbers (point summary and uncertainty) next to
#' each other.
#'
#' The following option is used to construct new [`rvar`] objects,
#' as in `rfun()` and `rdo()`:
#'
#' * `posterior.rvar_ndraws`: The number of draws used to construct
#' new random variables when this number cannot be determined
#' from existing arguments (e.g., other [`rvar`]s passed to a function).
#'
#' The following options are used to control warning messages:
#'
#' * `posterior.warn_on_merge_chains`: (logical) Some operations will
#' trigger an automatic merging of chains, for example, because chains do not
#' match between two objects involved in a binary operation. Whether this
#' causes a warning can be controlled by this option.
#'
NULL

#' Extract arrays of multiple variables
#'
#' Extract arrays of draws for multiple variables, returning them as a named list
#' of arrays. Each array has the same structure as returned by 
#' \code{\link{extract_variable_array}}.
#'
#' @template args-methods-x
#' @param variables A character vector of variable names to extract, or NULL to
#'   extract all variables. To extract all dimensions from variables with 
#'   indices (e.g. \code{"x[1]"}), provide the base variable names 
#'   (e.g. \code{"x"}).
#' @template args-methods-dots
#' @returns
#' A named list of arrays, where each array has dimension
#' `niterations(x)` x `nchains(x)` x any remaining
#' dimensions determined by the indices of the variable `x`
#' (if `with_chains = TRUE`) or dimension
#' `niterations(x) * nchains(x)` x any remaining dimensions
#' (if `with_chains = FALSE`).
#' @family variable extraction methods
#' @examples
#' x <- example_draws(example = "multi_normal")
#'
#' # Extract multiple variables at once
#' vars <- extract_list_of_variable_arrays(x, c("mu", "Sigma"))
#' str(vars)
#'
#' # Extract all variables (uses base variable names)
#' all_vars <- extract_list_of_variable_arrays(x)
#' str(all_vars)
#'
#' # Extract specific indexed variables
#' vars2 <- extract_list_of_variable_arrays(x, c("mu[1]", "mu[2]"))
#' str(vars2)
#'
#' @export
extract_list_of_variable_arrays <- function(x, variables = NULL, ...) {
  UseMethod("extract_list_of_variable_arrays")
}

#' @rdname extract_list_of_variable_arrays
#' @export
extract_list_of_variable_arrays.default <- function(x, variables = NULL, ...) {
  x <- as_draws(x)
  extract_list_of_variable_arrays(x, variables, ...)
}

#' @rdname extract_list_of_variable_arrays
#' @export
extract_list_of_variable_arrays.draws <- function(x, variables = NULL, ...) {
  if (is.null(variables)) {
    # Get unique base variable names (without indices) when variables = NULL
    all_vars <- variables(x)
    variables <- unique(split_variable_names(all_vars)$base_name)
  }
  
  if (!is.character(variables)) {
    stop_no_call("'variables' must be a character vector.")
  }
  
  if (length(variables) == 0) {
    return(list())
  }
  
  out <- lapply(variables, function(var) {
    extract_variable_array(x, variable = var, ...)
  })
  
  names(out) <- variables
  out
}

#' Transform to `draws` objects
#'
#' Try to transform an \R object to a format supported by the \pkg{posterior}
#' package.
#'
#' @name draws
#' @family formats
#'
#' @template args-methods-x
#' @template args-methods-dots
#'
#' @details The class `"draws"` is the parent class of all supported formats,
#'   which also have their own subclasses of the form `"draws_{format}"` (e.g.
#'   `"draws_array"`).
#'
#' @return If possible, a `draws` object in the closest supported format to `x`.
#'   The formats are linked to in the **See Also** section below.
#'
#' @examples
#' # create some random draws
#' x <- matrix(rnorm(30), nrow = 10)
#' colnames(x) <- c("a", "b", "c")
#' str(x)
#'
#' # transform to a draws object
#' y <- as_draws(x)
#' str(y)
#'
#' # remove the draws classes from the object
#' class(y) <- class(y)[-(1:2)]
#' str(y)
#'
NULL

#' @rdname draws
#' @export
as_draws <- function(x, ...) {
  UseMethod("as_draws")
}

#' @export
as_draws.draws <- function(x, ...) {
  x
}

#' @export
as_draws.default <- function(x, ...) {
  # transform an object to the closest supported draws format
  format <- closest_draws_format(x)
  fun_name <- paste0("as_", format)
  if (!has_s3_method(fun_name, class(x))) {
    # if there is no implementation of as_draws_XXX() for this class, then
    # we can't call as_draws_XXX() here as it will end up in as_draws_XXX.default()
    # which will call back to as_draws.default(), creating an infinite loop. So
    # we call down to .as_draws_XXX() instead.
    fun_name <- paste0(".", fun_name)
  }
  fun <- get(fun_name, asNamespace("posterior"))
  fun(x, ...)
}

#' @export
as_draws.rvar <- function(x, ...) {
  as_draws_rvars(x, ...)
}

# detect the supported format closest to the format of the input
closest_draws_format <- function(x) {
  if (is_draws_matrix_like(x)) {
    out <- "matrix"
  } else if (is_draws_array_like(x)) {
    out <- "array"
  } else if (is_draws_df_like(x)) {
    out <- "df"
  } else if (is_draws_rvars_like(x)) {
    out <- "rvars"
  } else if (is_draws_list_like(x)) {
    out <- "list"
  }
  else {
    stop_no_call("Don't know how to transform an object of class ",
          "'", class(x)[1L], "' to any supported draws format.")
  }
  paste0("draws_", out)
}

#' @rdname draws
#' @export
is_draws <- function(x) {
  inherits(x, "draws")
}

# check if an object is supported by the posterior package
# the name 'check_draws' is already in use for checking
# the validity of the 'draw' argument in 'subset'
check_draws_object <- function(x) {
  if (!is_draws(x)) {
    stop_no_call("The object is not in a format supported by posterior.")
  }
  x
}

#' check all variables in an object are numeric, converting non-numeric
#' to numeric with a warning. Used when converting to draws_array or
#' draws_matrix formats (which don't support non-numeric variables).
#' @param x A draws_df, draws_list, or draws_rvars object
#' @param is_non_numeric function that checks if a variable is non-numeric
#' @param convert convert non-numeric variables to numeric?
#' @noRd
check_variables_are_numeric <- function(
  x, to = "draws_array",
  is_non_numeric = function(x_i) !is.numeric(x_i) && !is.logical(x_i),
  convert = TRUE
) {

  non_numeric_cols <- vapply(x, is_non_numeric, logical(1))
  if (any(non_numeric_cols)) {
    warning_no_call(
      to,
      " does not support non-numeric variables (e.g., factors). Converting non-numeric variables to numeric."
    )
  }
  if (convert) {
    x[, non_numeric_cols] <- lapply(unclass(x)[non_numeric_cols], as.numeric)
  }
  x
}

# define default variable names
# use the 'unique' naming strategy of tibble
# @param nvariables number of variables
default_variables <- function(nvariables) {
  sprintf("...%s", seq_len(nvariables))
}

# validate draws vectors per variable
# @param ... Named arguments containing numeric vector
# @return a named list of numeric vectors
validate_draws_per_variable <- function(...) {
  out <- list(...)
  if (!rlang::is_named(out)) {
    stop_no_call("All variables must be named.")
  }
  if (".nchains" %in% names(out)) {
    # '.nchains' is an additional argument in chain supporting formats
    stop_no_call("'.nchains' is not supported for this format.")
  }
  out <- lapply(out, as.numeric)
  ndraws_per_variable <- lengths(out)
  ndraws <- max(ndraws_per_variable)
  if (!all(ndraws_per_variable %in% c(1, ndraws))) {
    stop_no_call("Number of draws per variable needs to be 1 or ", ndraws, ".")
  }
  for (i in which(ndraws_per_variable == 1)) {
    out[[i]] <- rep(out[[i]], ndraws)
  }
  out
}

#' Random variable slicing
#'
#' Operations for slicing [`rvar`]s and replacing parts of [`rvar`]s.
#'
#' @param x an [`rvar`].
#' @param i,... indices; see *Details*.
#' @param drop (logical) Should singular dimensions be dropped when slicing
#' array [`rvar`]s? Unlike base array slicing operations, defaults to `FALSE`.
#' @param value (`rvar` or coercable to `rvar`) Value to insert into
#' `x` at the location determined by the indices.
#'
#' @details
#' The [`rvar`] slicing operators (`[` and `[[`) attempt to implement the same
#' semantics as the [base array slicing operators][base::Extract]. There are some
#' exceptions; most notably, [`rvar`] slicing defaults to `drop = FALSE` instead
#' of `drop = TRUE`.
#'
#' @section Extracting or replacing single elements with `[[`:
#'
#' The `[[` operator extracts (or replaces) single elements. It always
#' returns (or replaces) a scalar (length-1) [`rvar`].
#'
#' The `x[[i,...]]` operator can be used as follows:
#'
#' - `x[[<numeric>]]` for scalar numeric `i`: gives the `i`th element of `x`. If `x` is
#' multidimensional (i.e. `length(dim(x)) > 1`), extra dimensions are ignored
#' when indexing. For example, if `x` is a \eqn{6 \times 2} [`rvar`] array, the
#' 7th element, `x[[7]]`, will be the first element of the second column, `x[1,2]`.
#'
#' - `x[[<numeric rvar>]]` for scalar numeric [`rvar`] `i`: a generalization of indexing when
#' `i` is a scalar numeric. Within each draw of `x`, selects the element
#' corresponding to the value of `i` within that same draw.
#'
#' - `x[[<character>]]` for scalar character `i`: gives the element of `x` with name
#' equal to `i`. **Unlike with base arrays**, does not work with
#' multidimensional [`rvar`]s.
#'
#' - `x[[i_1,i_2,...,i_n]]` for scalar numeric or character `i_1`, `i_2`, etc.
#' Must provide exactly the same number of indices as dimensions in `x`. Selects
#' the element at the corresponding position in the [`rvar`] by number and/or
#' dimname (as a string).
#'
#' @section Extracting or replacing multiple elements with `[`:
#'
#' The `[` operator extracts (or replaces) multiple elements. It always returns
#' (or replaces) a possibly-multidimensional [`rvar`].
#'
#' The `x[i,...]` operator can be used as follows:
#'
#' - `x[<logical>]` for vector logical `i`: `i` is recycled to the same length as `x`,
#' ignoring multiple dimensions in `x`, then an [`rvar`] vector is returned
#' containing the elements in `x` where `i` is `TRUE`.
#'
#' - `x[<logical rvar>]` for scalar logical [`rvar`] `i`: returns an [`rvar`] the same shape
#' as `x` containing only those draws where `i` is `TRUE`.
#'
#' - `x[<numeric>]` for vector numeric `i`: an [`rvar`] vector is returned
#' containing the `i`th elements of `x`, ignoring dimensions.
#'
#' - `x[<matrix>]` for numeric matrix `i`, where `ncol(i) == length(dim(x))`: each row
#' of `i` should give the multidimensional index for a single element in `x`. The
#' result is an [`rvar`] vector of length `nrow(i)` containing elements of `x`
#' selected by each row of `i`.
#'
#' - `x[i_1,i_2,...,i_n]` for vector numeric, character, or logical `i_1`,
#' `i_2`, etc. Returns a slice of `x` containing all elements from the dimensions
#' specified in `i_1`, `i_2`, etc. If an argument is left empty, all elements
#' from that dimension are included. Unlike base arrays, trailing dimensions
#' can be omitted entirely and will still be selected; for example, if `x` has
#' three dimensions, both `x[1,,]` and `x[1,]` can be used to create a
#' slice that includes all elements from the last two dimensions. Unlike base
#' arrays, `[` defaults to `drop = FALSE`, so results retain the same number of
#' dimensions as `x`.
#'
#' @examples
#' x <- rvar(array(1:24, dim = c(4,2,3)))
#' dimnames(x) <- list(c("a","b"), c("d","e","f"))
#' x
#'
#' ## Slicing single elements
#' # x[[<numeric>]]
#' x[[2]]
#'
#' # x[[<numeric rvar>]]
#' # notice the draws of x[1:4]...
#' draws_of(x[1:4])
#' x[[rvar(c(1,3,4,4))]]
#' # ... x[[rvar(c(1,3,4,4))]] creates a mixures of those draws
#' draws_of(x[[rvar(c(1,3,4,4))]])
#'
#' # x[[i_1,i_2,...]]
#' x[[2,"e"]]
#'
#'
#' ## Slicing multiple elements
#' # x[<logical>]
#' x[c(TRUE,TRUE,FALSE)]
#'
#' # x[<logical rvar>]
#' # select every other draw
#' x[rvar(c(TRUE,FALSE,TRUE,FALSE))]
#'
#' # x[<numeric>]
#' x[1:3]
#'
#' # x[<matrix>]
#' x[rbind(
#'   c(1,2),
#'   c(1,3),
#'   c(2,2)
#' )]
#'
#' # x[i_1,i_2,...,i_n]
#' x[1,]
#' x[1,2:3]
#' x[,2:3]
#' @name rvar-slice
NULL

#' @rdname rvar-slice
#' @importFrom rlang inject is_missing missing_arg dots_list
#' @export
`[[.rvar` <- function(x, i, ...) {
  index <- check_rvar_yank_index(x, i, ...)

  if (length(index) == 1) {
    # single element selection => collapse the dims so we can select directly using i
    .dim = dim(x)
    if (length(.dim) != 1) {
      # we only collapse dims if necessary since this will drop dimnames (which
      # would prevent single-element by-name selection for 1d rvars)
      dim(x) <- prod(.dim)
    }
    if (is_rvar(index[[1]])) {
      # x[[i]] where i is a scalar numeric rvar
      i_rvar <- index[[1]]
      c(i, x) %<-% scalar_numeric_rvar_to_index(i_rvar, x)
      .draws <- draws_of(x)[i]
    } else {
      # x[[i]] where i is a scalar numeric or character
      .draws <- draws_of(x)[, i, drop = FALSE]
    }
    dimnames(.draws) <- NULL
    out <- new_rvar(.draws, .nchains = nchains(x))
  } else if (length(index) == length(dim(x))) {
    # multiple element selection => must have exactly the right number of dims
    .draws <- inject(draws_of(x)[, !!!index, drop = FALSE])
    # must do drop manually in case the draws dimension has only 1 draw
    dim(.draws) <- c(ndraws(x), 1)
    out <- new_rvar(.draws, .nchains = nchains(x))
  } else {
    stop_no_call("subscript out of bounds")
  }
  out
}

#' @rdname rvar-slice
#' @export
`[[<-.rvar` <- function(x, i, ..., value) {
  value <- vec_cast(value, x)
  c(x, value) %<-% conform_rvar_ndraws_nchains(list(x, value))
  value <- check_rvar_dims_first(value, new_rvar(0))
  index <- check_rvar_yank_index(x, i, ...)

  if (length(index) == 1) {
    .dim = dim(x)

    if (length(.dim) == 1 && isTRUE(i > length(x))) {
      # unidimensional indexing allows array extension; extend the array
      # then do the assignment
      x <- x[seq_len(max(i, na.rm = TRUE))]
      draws_of(x)[, i] <- draws_of(value)
    } else {
      # single element selection => collapse the dims so we can select directly using i
      .dimnames = dimnames(draws_of(x)) # to restore later
      if (length(.dim) != 1) {
        # we only collapse dims if necessary since this will drop dimnames (which
        # would prevent single-element by-name selection for 1d rvars)
        dim(x) <- prod(.dim)
      }
      if (is_rvar(i)) {
        # x[[i]] <- value where i is a scalar numeric rvar
        c(i, x, value) %<-% scalar_numeric_rvar_to_index(i, x, value)
        draws_of(x)[i] <- draws_of(value)
      } else {
        # x[[i]] <- value where i is a scalar numeric or character
        draws_of(x)[, i] <- draws_of(value)
      }
      dim(x) <- .dim
      dimnames(draws_of(x)) <- .dimnames
    }
  } else if (length(index) == length(dim(x))) {
    # multiple element selection => must have exactly the right number of dims
    x <- inject({
      draws_of(x)[, !!!index] <- draws_of(value)
      x
    })
  } else {
    stop_no_call("subscript out of bounds")
  }
  x
}

#' @rdname rvar-slice
#' @importFrom rlang dots_list inject is_missing missing_arg expr
#' @export
`[.rvar` <- function(x, ..., drop = FALSE) {
  check_rvar_subset_indices(x, ...)
  index <- dots_list(..., .ignore_empty = "none", .preserve_empty = TRUE)

  # check for indexing of draws
  if (length(index) == 1 && is_rvar(index[[1]])) {
    i <- index[[1]]
    if (!is.logical(draws_of(i)) || length(i) != 1) {
      stop_no_call("`x[i]` for rvars `x` and `i` is only supported when `i` is a scalar logical rvar.")
    }

    # indexing by a scalar rvar, use it as a draws index
    # this kind of indexing must ignore chains
    nchains_rvar(x) <- 1L
    nchains_rvar(i) <- 1L
    c(x, i) %<-% conform_rvar_ndraws(list(x, i))
    index <- list()
    draws_index <- list(draws_of(i))
  } else {
    # not indexing draws, so draws_index is empty
    # must store draws_index as a list to safely pass around the missing argument
    draws_index <- list(missing_arg())
  }

  .draws <- draws_of(x)
  .dim <- dim(.draws)

  # clean up the indices: because we have to index using the multi-index
  # notation x[,...] (to account for draws) and this notation has slightly
  # different semantics for NAs from the way that x[i] works, have to do a bit
  # of translation here
  for (i in seq_along(index)) {
    if (is.numeric(index[[i]])) {
      # numeric indices outside the range of the corresponding dimension
      # should create NAs; but array indexing doesn't do this (it throws
      # an error), so we adjust the indices to do so.

      dim_i_length <- if (i == 1 && length(index) == 1) {
        # for x[i] style indexing of multidimensional arrays we will flatten
        # the array before indexing, so the max of the dim length will be
        # the length of x
        length(x)
      } else {
        .dim[[i + 1]]
      }

      index[[i]][index[[i]] > dim_i_length] <- NA_integer_
    }
  }

  if (length(index) == 1) {
    # indexing by a single dimension, (could be numerical indexing along one
    # dimension --- even for multidimensional arrays; logical indexing; or matrix indexing)

    if (is.matrix(index[[1]]) && ncol(index[[1]]) == length(.dim) - 1) {
      # matrix-based indexing, like x[cbind(2,1,2)]
      # => translate matrix-based indices into unidimensional indices
      index[[1]] <- matrix_to_index(index[[1]], .dim[-1])
    }

    if (is_missing(index[[1]])) {
      # if we only have one index and it is missing entirely, the call must
      # have been for x[], which actually means no index at all
      index <- list()
    } else if (length(dim(.draws)) > 2) {
      # draws have > 2 dims => array has > 1 dimension => must flatten array so
      # that the index can be applied along all the elements of the array
      dim(.draws) <- c(dim(.draws)[[1]], length(x))
    }
  } else if (length(index) < length(dim(.draws)) - 1) {
    # fill in final indices with missing arguments
    index[seq(length(index) + 1, length(dim(.draws)) - 1)] = list(missing_arg())
  }

  .draws <- inject(.draws[!!!draws_index, !!!index, drop = FALSE])

  if (!is_missing(draws_index[[1]])) {
    # if we subsetted draws, replace draw ids with sequential ids
    rownames(.draws) <- seq_len(NROW(.draws))
  }

  x <- new_rvar(.draws, .nchains = nchains(x))

  if (drop) {
    x <- drop(x)
  }

  x
}

#' @rdname rvar-slice
#' @export
`[<-.rvar` <- function(x, i, ..., value) {
  if (missing(i)) i = missing_arg()
  if (length(dim(x)) == 1 && !missing(i) && is.numeric(i) && any(i > length(x), na.rm = TRUE)) {
    # unidimensional indexing allows array extension; extend the array
    # before we do the assignment
    x <- x[seq_len(max(i, na.rm = TRUE))]
  }

  value <- vec_cast(value, x)

  # check to see if we're doing any draw indexing
  if (!missing(i) && is_rvar(i)) {
    if (!is.logical(draws_of(i)) || length(i) != 1L || !missing(...)) {
      stop_no_call("`x[i]` for rvars `x` and `i` is only supported when `i` is a single, scalar logical rvar.")
    }

    # for the purposes of this kind of assignment, we check draws only, not chains,
    # as chain information is irrelevant when subsetting by draw
    c(x, i) %<-% conform_rvar_ndraws(list(x, i))
    draws_index <- draws_of(i)

    # necessary number of draws in `value` is determined by whether or not
    # we're doing logical indexing
    value_ndraws <- sum(draws_index, na.rm = TRUE)
    draws_of(value) <- broadcast_array(draws_of(value), c(value_ndraws, dim(x)), broadcast_scalars = FALSE)
    i <- missing_arg()
  } else {
    c(x, value) %<-% conform_rvar_ndraws_nchains(list(x, value))
    draws_index <- missing_arg()
  }

  if (missing(...)) {
    # index over entire array: flatten array so that the  index
    # can be applied along all the elements of the array, then invert after assignment
    original_dim <- dim(draws_of(x))
    original_dimnames <- dimnames(draws_of(x))

    if (!missing(i) && is.matrix(i) && ncol(i) == length(original_dim) - 1) {
      # matrix-based indexing, like x[cbind(2,1,2)] <- y
      # => translate matrix-based indices into unidimensional indices
      i <- matrix_to_index(i, original_dim[-1])
    }

    #flatten and assign
    dim(x) <- length(x)
    draws_of(x)[draws_index,i] <- draws_of(value)

    # unflatten and restore dimnames
    dim(draws_of(x)) <- original_dim
    dimnames(draws_of(x)) <- original_dimnames
  } else {
    draws_of(x)[,i,...] <- draws_of(value)
  }

  x
}



# slicing helpers ---------------------------------------------------------

# Given m, a matrix with ncol(m) == length(dim), where each row specifies the
# index of a single cell from an array of shape == dim, return a vector of
# length nrow(m) giving the unidimensional indices of that array corresponding
# to the cells specified by each row in m
# e.g. if we have an array x with dim(x) = c(2,3,4), we might want to
# translate an index of style x[cbind(2,1,2)] into x[i] so that we can index
# index x using the single index i instead of cbind(2,1,2).
# matrix_to_index(cbind(2,1,2), dim(x)) does that translation.
matrix_to_index <- function(m, dim) {
  cumdim <- cumprod(c(1, dim[-length(dim)]))
  as.vector((m - 1) %*% cumdim + 1)
}

# Given i_rvar, a scalar numeric rvar acting as an index on x, return a list
# whose first element is i, an index on draws_of(x) that corresponds to i_rvar.
# remaining elements are x and any other rvars in `...`, all of which will
# have chains and draws conformed to each other.
scalar_numeric_rvar_to_index <- function(i_rvar, x, ...) {
  if (!is.numeric(draws_of(i_rvar)) || length(i_rvar) != 1) {
    stop_no_call("`x[[i]]` for rvars `x` and `i` is only supported when `i` is a scalar numeric rvar.")
  }
  out <- conform_rvar_ndraws_nchains(list(i_rvar, x, ...))
  c(i_rvar, x) %<-% out[1:2]
  out[[1]] <- matrix_to_index(cbind(seq_len(ndraws(x)), draws_of(i_rvar)), c(ndraws(x), length(x)))
  out
}

# remove variables from a draws object
remove_variables <- function(x, ...) {
  UseMethod("remove_variables")
}

#' @export
remove_variables.list <- function(x, variables, ...) {
  variables <- as.character(variables)
  variables  <- intersect(variables, names(x))
  for (v in variables) {
    x[[v]] <- NULL
  }
  x
}

#' @export
remove_variables.draws_matrix <- function(x, variables, ...) {
  variables <- as.character(variables)
  if (!length(variables)) {
    return(x)
  }
  x[, !colnames(x) %in% variables]
}

#' @export
remove_variables.draws_array <- function(x, variables, ...) {
  variables <- as.character(variables)
  if (!length(variables)) {
    return(x)
  }
  x[, , !dimnames(x)[[3]] %in% variables]
}

#' @export
remove_variables.draws_df <- function(x, variables, ...) {
  variables <- as.character(variables)
  for (v in variables) {
    x[[v]] <- NULL
  }
  x
}

#' @export
remove_variables.draws_list <- function(x, variables, ...) {
  variables <- as.character(variables)
  if (!length(variables)) {
    return(x)
  }
  for (i in seq_along(x)) {
    for (v in variables) {
      x[[i]][[v]] <- NULL
    }
  }
  x
}

#' @export
remove_variables.draws_rvars <- function(x, variables, ...) {
  variables <- as.character(variables)
  if (!length(variables)) {
    return(x)
  }
  x[!names(x) %in% variables]
}


# internal ----------------------------------------------------------------

# remove all reserved variables
remove_reserved_variables <- function(x, ...) {
  remove_variables(x, reserved_variables(x))
}

# remove reserved variables specific for the 'draws_df' format
remove_reserved_df_variables <- function(x, ...) {
  remove_variables(x, reserved_df_variables())
}

#' Random variable resulting from a function applied over margins of an array or random variable
#'
#' Returns an [`rvar`] obtained by applying a function to margins of an array or [`rvar`].
#' Acts like `apply()`, except that the function supplied (`.f`) should return an [`rvar`],
#' and the final result is always an [`rvar`].
#'
#' @param .x An array or an [`rvar`].
#' @param .margin (multiple options) The subscripts which the function will be applied over:
#'   * An integer vector. E.g., for a matrix `1` indicates rows, `2` indicates
#'     columns, `c(1, 2)` indicates rows and columns.
#'   * A character vector of dimension names if `.x` has named dimensions.
#' @param .f (function) The function to be applied. The function `.f` must
#'   return an [`rvar`] and the dimensions of the result of `.f` applied to each
#'   margin of `.x` must be able to be broadcasted to a common shape (otherwise
#'   the resulting [`rvar`] cannot be simplified). See **Details**.
#' @param ... Optional arguments passed to `.f`.
#'
#' @details
#'
#' This function acts much like `apply()`, except that the function passed to it (`.f`)
#' must return [`rvar`]s, and the result is simplified into an [`rvar`]. Unlike
#' `apply()`, it also keeps the dimensions of the returned values along each margin,
#' rather than simplifying each margin to a vector, and if the results of `.f` do
#' not all have the same dimensions, it applies the [`rvar`] broadcasting rules to
#' bind results together rather than using vector recycling.
#'
#' If you wish to apply functions over [`rvar`]s where the result is not intended to
#' be simplified into an [`rvar`], you can use the standard `apply()`, `lapply()`,
#' `sapply()`, or `vapply()` functions.
#'
#' @seealso [as_rvar()] to convert objects to `rvar`s. See [rdo()], [rfun()], and
#' [rvar_rng()] for higher-level interfaces for creating `rvar`s.
#'
#' @return An [`rvar`].
#'
#' If the result of each call to `.f` returns an [`rvar`] of dimension `d` after
#' being broadcast to a common shape, then `rvar_apply()` returns an [`rvar`] of
#' dimension `c(d, dim(.x)[.margin])`. If the last dimension of the result would
#' be `1`, it is dropped (other dimensions equal to `1` are retained). If `d` is
#' `0`, the result has length `0` but not necessarily the 'correct' dimension.
#'
#' @examples
#'
#' set.seed(3456)
#' x <- rvar_rng(rnorm, 24, mean = 1:24)
#' dim(x) <- c(2,3,4)
#'
#' # we can find the distributions of marginal means of the above array
#' # using rvar_mean along with rvar_apply
#' rvar_apply(x, 1, rvar_mean)
#' rvar_apply(x, 2:3, rvar_mean)
#'
#' @export
rvar_apply <- function(.x, .margin, .f, ...) {
  # this should return a list of rvars
  rvar_list <- apply(.x, .margin, .f, ...)
  if (!is.list(rvar_list) || !all(sapply(rvar_list, is_rvar))) {
    stop_no_call("The function passed to rvar_apply() must return rvars.")
  }

  # get the dimensions of the results in each cell
  cell_dim <- dim(rvar_list[[1]])
  # get the dimensions from the original array that are being kept
  marginal_dim <- dim(rvar_list) %||% length(rvar_list)

  # bind the output into an rvar along the first dimension (which the
  # marginal dimensions flattened into a single dim)
  out <- rvar_list[[1]]
  .dimnames <- dimnames(out)
  dim(out) <- c(1, dim(out))
  if (length(out)) dimnames(out) <- c(list(NULL), .dimnames)
  # process remaining rvars in succession, binding them to the output
  for (i in seq_along(rvar_list)[-1]) {
    rvar_i <- rvar_list[[i]]
    .dimnames <- dimnames(rvar_i)
    dim(rvar_i) <- c(1, dim(rvar_i))
    if (length(rvar_i)) dimnames(rvar_i) <- c(list(NULL), .dimnames)
    out <- broadcast_and_bind_rvars(out, rvar_i, 1)
  }

  if (length(out) > 0) {
    # restore the shape of the marginal dimensions
    .dimnames <- dimnames(out)
    dim(out) <- c(marginal_dim, dim(out)[-1])
    dimnames(out) <- c(list(NULL), .dimnames[-1])
    # if the last dimension is 1, drop it
    n_dim <- length(dim(out))
    if (dim(out)[[n_dim]] == 1) {
      dim(out) <- dim(out)[-n_dim]
    }
    # restore marginal dimnames
    marginal_dim_i <- seq_along(marginal_dim)
    out <- copy_dimnames(.x, .margin, out, marginal_dim_i)
  }

  out
}

#' Density, CDF, and quantile functions of random variables
#'
#' The probability density function (`density()`), cumulative distribution
#' function (`cdf()`), and quantile function / inverse CDF (`quantile()`) of
#' an [`rvar`].
#'
#' @param x (rvar) An [`rvar`] object.
#' @param q,at (numeric vector) One or more quantiles.
#' @param probs (numeric vector) One or more probabilities in `[0,1]`.
#' @param ... Additional arguments passed onto underlying methods:
#'   - For `density()`, these are passed to [stats::density()].
#'   - For `cdf()`, these are ignored.
#'   - For `quantile()`, these are passed to [stats::quantile()].
#'
#' @return
#'
#' If `x` is a scalar [`rvar`], returns a vector of the same length as the input
#' (`q`, `at`, or `probs`) containing values from the corresponding function
#' of the given [`rvar`].
#'
#' If `x` has length greater than 1, returns an array with dimensions
#' `c(length(y), dim(x))` where `y` is `q`, `at`, or `probs`, where each
#' `result[i,...]` is the value of the corresponding function,`f(y[i])`, for
#' the corresponding cell in the input array, `x[...]`.
#'
#' @examples
#'
#' set.seed(1234)
#' x = rvar(rnorm(100))
#'
#' density(x, seq(-2, 2, length.out = 10))
#' cdf(x, seq(-2, 2, length.out = 10))
#' quantile(x, ppoints(10))
#'
#' x2 = c(rvar(rnorm(100, mean = -0.5)), rvar(rnorm(100, mean = 0.5)))
#' density(x2, seq(-2, 2, length.out = 10))
#' cdf(x2, seq(-2, 2, length.out = 10))
#' quantile(x2, ppoints(10))
#'
#' @name rvar-dist
#' @export
density.rvar <- function(x, at, ...) {
  summarise_rvar_by_element(x, function(draws) {
    d <- density(draws, cut = 0, ...)
    f <- approxfun(d$x, d$y, yleft = 0, yright = 0)
    f(at)
  })
}

#' @rdname rvar-dist
#' @export
density.rvar_factor <- function(x, at, ...) {
  at <- as.numeric(factor(at, levels = levels(x)))
  nbins <- nlevels(x)

  summarise_rvar_by_element(x, function(draws) {
    props <- prop.table(tabulate(draws, nbins = nbins))[at]
    props
  })
}

#' @importFrom distributional cdf
#' @export
distributional::cdf

#' @rdname rvar-dist
#' @export
cdf.rvar <- function(x, q, ...) {
  summarise_rvar_by_element(x, function(draws) {
    ecdf(draws)(q)
  })
}

#' @rdname rvar-dist
#' @export
cdf.rvar_factor <- function(x, q, ...) {
  # CDF is not defined for unordered distributions
  # generate an all-NA array of the appropriate shape
  out <- rep_len(NA, length(x) * length(q))
  if (length(x) > 1) dim(out) <- c(length(q), dim(x))
  out
}

#' @rdname rvar-dist
#' @export
cdf.rvar_ordered <- function(x, q, ...) {
  q <- as.numeric(factor(q, levels = levels(x)))
  cdf.rvar(x, q, ...)
}

#' @rdname rvar-dist
#' @export
quantile.rvar <- function(x, probs, ...) {
  summarise_rvar_by_element_via_matrix(x,
    "quantile",
    function(draws) {
      t(matrixStats::colQuantiles(draws, probs = probs, useNames = TRUE, ...))
    },
    .extra_dim = length(probs),
    .extra_dimnames = list(NULL)
  )
}

#' @rdname rvar-dist
#' @export
quantile.rvar_factor <- function(x, probs, ...) {
  # quantile function is not defined for unordered distributions
  # generate an all-NA array of the appropriate shape (CDF has the same output)
  cdf.rvar_factor(x, probs)
}

#' @rdname rvar-dist
#' @export
quantile.rvar_ordered <- function(x, probs, ...) {
  # `type` must be in 1:3 because x is discrete
  out <- quantile(as_rvar_numeric(x), probs, type = 1, ...)
  while_preserving_dims(function(out) levels(x)[out], out)
}

#' Reserved variables
#'
#' Get names of reserved variables from objects in the \pkg{posterior} package.
#'
#' @name reserved_variables
#' @template args-methods-x
#' @template args-methods-dots
#' @details
#'
#' `reserved_variables()` returns the names of reserved variables in use by
#' an object.
#'
#' The following variables names are currently reserved for special use cases
#' in all [`draws`] formats:
#' * `.log_weight`: Log weights per draw (see [`weight_draws`]).
#'
#' Further, specific for the [`draws_df`] format, there are three additional
#' reserved variables:
#' * `.chain`: Chain index per draw
#' * `.iteration`: Iteration index within each chain
#' * `.draw`: Draw index across chains
#'
#' More reserved variables may be added in the future.
#'
#' @return
#'
#' A character vector of reserved variables used in `x`.
#'
#' @examples
#'
#' x <- example_draws()
#' reserved_variables(x)
#'
#' # if we add weights, the `.log_weight` reserved variable is used
#' x <- weight_draws(x, rexp(ndraws(x)))
#' reserved_variables(x)
#'
#' @export
reserved_variables <- function(x, ...) {
  UseMethod("reserved_variables")
}

#' @rdname reserved_variables
#' @export
reserved_variables.default <- function(x, ...) {
  c(".log_weight")
}

#' @rdname reserved_variables
#' @export
reserved_variables.draws_matrix <- function(x, ...) {
  intersect(reserved_variables(), colnames(x))
}

#' @rdname reserved_variables
#' @export
reserved_variables.draws_array <- function(x, ...) {
  intersect(reserved_variables(), dimnames(x)[[3]])
}

#' @rdname reserved_variables
#' @export
reserved_variables.draws_df <- function(x, ...) {
  intersect(reserved_variables(), names(x))
}

#' @rdname reserved_variables
#' @export
reserved_variables.draws_list <- function(x, ...) {
  intersect(reserved_variables(), names(x[[1]]))
}

#' @rdname reserved_variables
#' @export
reserved_variables.draws_rvars <- function(x, ...) {
  intersect(reserved_variables(), names(x))
}


# internal ----------------------------------------------------------------

# reserved variables specific for the 'draws_df' format
reserved_df_variables <- function() {
  c(".chain", ".iteration", ".draw")
}

all_reserved_variables <- function(x = NULL) {
  c(reserved_variables(x), reserved_df_variables())
}

#' Repair indices of `draws` objects
#'
#' Repair indices of `draws` objects so that iterations, chains, and draws
#' are continuously and consistently numbered.
#'
#' @template args-methods-x
#' @template args-methods-dots
#' @param order (logical) Should draws be ordered (via [`order_draws()`]) before
#'   repairing indices? Defaults to `TRUE`.
#' @template return-draws
#' @seealso [order_draws()]
#' @examples
#' x <- as_draws_array(example_draws())
#' (x <- x[10:5, 3:4, ])
#' repair_draws(x)
#'
#' @export
repair_draws <- function(x, order = TRUE, ...) {
  UseMethod("repair_draws")
}

#' @rdname repair_draws
#' @export
repair_draws.draws_matrix <- function(x, order = TRUE, ...) {
  # ensure integer instead of character ordering
  rownames(x) <- repair_ids(rownames(x))
  x <- do_ordering(x, order)
  rownames(x) <- as.character(seq_rows(x))
  x
}

#' @rdname repair_draws
#' @export
repair_draws.draws_array <- function(x, order = TRUE, ...) {
  # ensure integer instead of character ordering
  rownames(x) <- repair_ids(rownames(x))
  colnames(x) <- repair_ids(colnames(x))
  x <- do_ordering(x, order)
  rownames(x) <- as.character(seq_rows(x))
  colnames(x) <- as.character(seq_cols(x))
  x
}

#' @rdname repair_draws
#' @export
repair_draws.draws_df <- function(x, order = TRUE, ...) {
  x <- do_ordering(x, order)
  x$.chain <- repair_chain_ids(x$.chain)
  x$.iteration <- repair_iteration_ids(x$.iteration, x$.chain)
  x$.draw <- compute_draw_ids(x$.chain, x$.iteration)
  x
}

#' @rdname repair_draws
#' @export
repair_draws.draws_list <- function(x, order = TRUE, ...) {
  # ensure integer instead of character ordering
  names(x) <- repair_ids(names(x))
  x <- do_ordering(x, order)
  names(x) <- seq_along(x)
  x
}

#' @rdname repair_draws
#' @export
repair_draws.draws_rvars <- function(x, order = TRUE, ...) {
  for (i in seq_along(x)) {
    x[[i]] <- repair_draws(x[[i]], order = order, ...)
  }
  x
}

#' @rdname repair_draws
#' @export
repair_draws.rvar <- function(x, order = TRUE, ...) {
  rownames(draws_of(x)) <- repair_ids(rownames(draws_of(x)))
  order <- as_one_logical(order)
  if (order) {
    x <- order_draws(x)
  } else if (nchains(x) > 1) {
    # even if we aren't asked to order draws, still need to check on the order
    # because if the draws are out of order we have to drop chain info (it would
    # no longer be guaranteed to be correct)
    draw_order <- order(draw_ids(x))
    if (needs_ordering(draw_order)) {
      x <- merge_chains(x)
    }
  }
  rownames(draws_of(x)) <- as.character(seq_rows(draws_of(x)))
  x
}


# internal ----------------------------------------------------------------

#' Repair indices to be continuously numbered integers starting from one
#' @param x vector of values
#' @noRd
repair_ids <- function(x) {
  out <- SW(as.integer(x))
  if (anyNA(out)) {
    # use character instead of integer ordering if
    # some values cannot be converted to integers
    out <- factor(x)
  } else {
    out <- factor(out)
  }
  as.integer(out)
}

#' Repair iteration indices
#' @param iteration_ids A vector of iteration indices
#' @param chain_ids A vector of chain indices
#' @noRd
repair_iteration_ids <- function(iteration_ids, chain_ids = NULL) {
  .repair_iteration_ids <- function(x) {
    match(seq_along(x), order(x))
  }
  if (is.null(chain_ids)) {
    out <- .repair_iteration_ids(iteration_ids)
  } else {
    check_true(length(iteration_ids) == length(chain_ids))
    unique_chain_ids <- unique(chain_ids)
    out <- rep(NA, length(iteration_ids))
    for (u in unique(chain_ids)) {
      sel <- chain_ids == u
      out[sel] <- .repair_iteration_ids(iteration_ids[sel])
    }
  }
  as.integer(out)
}

#' Repair chain indices
#' @param chain_ids A vector of chain indices
#' @noRd
repair_chain_ids <- function(chain_ids) {
  repair_ids(chain_ids)
}

#' Compute draw indices from iteration and chain indices
#' @param chain_ids A vector of chain indices
#' @param iteration_ids A vector of iteration indices
#' @noRd
compute_draw_ids <- function(chain_ids, iteration_ids) {
  assert_true(length(chain_ids) == length(iteration_ids))
  niterations <- SW(max(iteration_ids))
  out <- (chain_ids - 1L) * niterations + iteration_ids
  as.integer(out)
}

# Copyright (C) 2020 Ben Lambert, Aki Vehtari
# See LICENSE.md for more details

#' Calculate R* convergence diagnostic
#'
#' The `rstar()` function generates a measure of convergence for MCMC draws
#' based on whether it is possible to determine the Markov chain that generated
#' a draw with probability greater than chance. To do so, it fits a machine
#' learning classifier to a training set of MCMC draws and evaluates its
#' predictive accuracy on a testing set: giving the ratio of accuracy to
#' predicting a chain uniformly at random.
#'
#' @family diagnostics
#' @param x (draws) A [`draws_df`] object or one coercible to a `draws_df` object.
#' @template args-conv-split
#'
#' @param uncertainty (logical). Indicates whether to provide a vector of R*
#'   values representing uncertainty in the calculated value (if `TRUE`) or a
#'   single value (if `FALSE`). The default is `TRUE.`
#'
#' @param method (string) The machine learning classifier to use (must be
#'   available in the \pkg{caret} package). The default is `"rf"`, which calls
#'   the random forest classifier.
#'
#' @param hyperparameters (named list) Hyperparameter settings passed to the classifier.
#'   The default for the random forest classifier (`method = "rf"`) is
#'   `list(mtry = floor(sqt(nvariables(x))))`.
#'   The default for the gradient-based model (`method = "gbm"`) is
#'   `list(interaction.depth = 3, n.trees = 50, shrinkage = 0.1, n.minobsinnode = 10)`.
#'
#' @param nsimulations (positive integer) The number of R* values in the
#'   returned vector if `uncertainty` is `TRUE`. The default is `1000.`
#'
#' @param training_proportion (positive real) The proportion (in `(0,1)`) of
#'   iterations in used to train the classifier. The default is `0.7`.
#'
#' @param ... Other arguments passed to `caret::train()`.
#'
#' @details The `rstar()` function provides a measure of MCMC convergence based
#'   on whether it is possible to determine the chain that generated a
#'   particular draw with a probability greater than chance. To do so, it fits a
#'   machine learning classifier to a subset of the original MCMC draws (the
#'   training set) and evaluates its predictive accuracy on the remaining draws
#'   (the testing set). If predictive accuracy exceeds chance (i.e. predicting
#'   the chain that generated a draw uniformly at random), the diagnostic
#'   measure R* will be above 1, indicating that convergence has yet to occur.
#'   This statistic is recently developed, and it is currently unclear what is a
#'   reasonable threshold for diagnosing convergence.
#'
#'   The statistic, R*, is stochastic, meaning that each time the test is run,
#'   unless the random seed is fixed, it will generally produce a different
#'   result. To minimize the implications of this stochasticity, it is
#'   recommended to repeatedly run this function to calculate a distribution of
#'   R*; alternatively, an approximation to this distribution can be obtained by
#'   setting `uncertainty = TRUE`, although this approximation of uncertainty
#'   will generally have a lower mean.
#'
#'   By default, a random forest classifier is used (`method = "rf"`), which tends
#'   to perform best for target distributions of around 4 dimensions and above.
#'   For lower dimensional targets, gradient boosted models (called via
#'   `method = "gbm"`) tend to have a higher classification accuracy. On a given
#'   MCMC sample, it is recommended to try both of these classifiers.
#'
#' @return A numeric vector of length 1 (by default) or length `nsimulations`
#'   (if `uncertainty = TRUE`).
#'
#' @references Ben Lambert, Aki Vehtari (2020) R*: A robust MCMC convergence
#'   diagnostic with uncertainty using gradient-boosted machines.
#'   *arXiv preprint* `arXiv:2003.07900`.
#'
#' @examples
#' \donttest{
#' if (require("caret", quietly = TRUE) && require("randomForest", quietly = TRUE)) {
#'   x <- example_draws("eight_schools")
#'   print(rstar(x))
#'   print(rstar(x, split = FALSE))
#'   print(rstar(x, method = "gbm"))
#'   # can pass additional arguments to methods
#'   print(rstar(x, method = "gbm", verbose = FALSE))
#'
#'   # with uncertainty, returns a vector of R* values
#'   hist(rstar(x, uncertainty = TRUE))
#'   hist(rstar(x, uncertainty = TRUE, nsimulations = 100))
#'
#'   # can use other classification methods in caret library
#'   print(rstar(x, method = "knn"))
#' }
#' }
#' @export
rstar <- function(x, split = TRUE,
                  uncertainty = FALSE,
                  method = "rf",
                  hyperparameters = NULL,
                  training_proportion = 0.7,
                  nsimulations = 1000,
                  ...) {

  # caret requires itself to be attached to the search list (not just loaded).
  # To avoid polluting the user's namespace we manually attach caret and its
  # two hard dependencies (ggplot2 and lattice) if they aren't already attached,
  # and unload any of them we had to attach when this function exits.
  loaded_packages <- character()
  on.exit(for (package in loaded_packages) {
    detach(paste0("package:", package), character.only = TRUE)
  })
  for (package in setdiff(c("ggplot2", "lattice", "caret"), .packages())) {
    if (!suppressPackageStartupMessages(get("require")(package, character.only = TRUE, quietly = TRUE))) {
      stop_no_call("Package '", package, "' is required for 'rstar' to work.")
    }
    # store in reverse order since we'll unload in reverse
    loaded_packages <- c(package, loaded_packages)
  }

  split <- as_one_logical(split)
  uncertainty <- as_one_logical(uncertainty)
  method <- as_one_character(method)
  nsimulations <- as_one_integer(nsimulations)
  if (nsimulations < 1) {
    stop_no_call("'nsimulations' must be greater than or equal to 1.")
  }
  training_proportion <- as_one_numeric(training_proportion)
  if (training_proportion <= 0 || training_proportion >= 1) {
    stop_no_call("'training_proportion' must be greater than 0 and less than 1.")
  }

  # caret requires data.frame like objects
  x <- as_draws_df(x)
  if (split) {
    x <- split_chains(x)
  }

  # caret requires at least two variables to work
  if (nvariables(x) == 1) {
    x$.random <- rnorm(ndraws(x))
  }

  # choose hyperparameters
  if (method == "rf" && is.null(hyperparameters)) {
    caret_grid <- data.frame(mtry = floor(sqrt(nvariables(x))))
  } else if (method == "gbm" && is.null(hyperparameters)) {
    caret_grid <- data.frame(
      interaction.depth = 3,
      n.trees = 50,
      shrinkage = 0.1,
      n.minobsinnode = 10
    )
  } else {
    if (!(is.null(hyperparameters) || is.list(hyperparameters))) {
      stop_no_call("'hyperparameters' must be a list or NULL.")
    }
    caret_grid <- hyperparameters
  }

  # reserved variables should not be used for classification
  reserved_vars <- all_reserved_variables(x)
  reserved_vars <- setdiff(reserved_vars, ".chain")
  x <- remove_variables(x, reserved_vars)
  ndraws <- ndraws(x)
  class(x) <- "data.frame"
  x$.chain <- as.factor(x$.chain)

  # create training / testing sets
  random_draws <- sample(seq_len(ndraws), training_proportion * ndraws)
  training_data <- x[random_draws, ]
  testing_data <- x[-random_draws, ]

  # predict chain index using all other variables
  fit <- caret::train(
    .chain ~ .,
    data = training_data,
    method = method,
    trControl = caret::trainControl(method = 'none'),
    tuneGrid = caret_grid,
    ...
  )

  # calculate classification accuracy and then R*
  # not all chains may be represented in testing_data
  nchains <- length(unique(testing_data$.chain))
  if (uncertainty) {
    probs <- predict(object = fit, newdata = testing_data, type = "prob")
    m_accuracy <- matrix(nrow = nrow(probs), ncol = nsimulations)
    for (j in seq_len(NROW(probs))) {
      vals <- rmultinom(nsimulations, 1, prob = probs[j, ])
      test <- apply(vals, 2, function(v) which(v == 1))
      m_accuracy[j, ] <- ifelse(test == testing_data$.chain[j], 1, 0)
    }
    out <- colMeans(m_accuracy) * nchains
  } else {
    plda <- predict(object = fit, newdata = testing_data)
    res <- data.frame(predicted = plda, actual = testing_data$.chain)
    accuracy <- mean(res$predicted == res$actual)
    out <- accuracy * nchains
  }
  out
}

# Summaries within array elements, over draws  --------------------------

#' Summaries of random variables within array elements, over draws
#'
#' Compute summaries within elements of an [`rvar`] and over draws of each element,
#' producing an array of the same shape as the input random variable (except in
#' the case of `range()`, see **Details**).
#'
#' @param x (rvar) An [`rvar`].
#' @param ... Further arguments passed to underlying functions (e.g.,
#'   `base::mean()` or `base::median()`), such as `na.rm`.
#'
#' @details
#'
#' Summaries include expectations (`E()` or `mean()`), probabilities (`Pr()`),
#' medians (`median()`), spread (`var()`, `variance()`, `sd()`, `mad()`), sums and
#' products (`sum()`, `prod()`), extrema and ranges (`min()`, `max()`, `range()`),
#' logical summaries (`all()`, `any()`), and special value predicates (`is.finite()`,
#' `is.infinite()`, `is.nan()`, `is.na()`).
#'
#' Unless otherwise stated, these functions return a numeric array with the same shape
#' (same dimensions) as the input [`rvar`], `x`.
#'
#' `range(x)` returns an array with dimensions `c(2, dim(x))`, where the last
#' dimension contains the minimum and maximum values.
#'
#' `is.infinite(x)`, `is.nan(x)`, and `is.na(x)` return logical arrays, where each
#' element is `TRUE` if **any** draws in its corresponding element in `x` match
#' the predicate. Each elements in the result of `is.finite(x)` is `TRUE` if
#' **all** draws in the corresponding element in `x` are finite.
#'
#' Both `E()`, `mean()`, and `Pr()` return the means of each element in the input.
#' `Pr()` additionally checks that the provided [`rvar`]
#' is a logical variable (hence, taking its expectation results in a probability).
#'
#' For consistency, `E()` and `Pr()` are also defined for base arrays so that
#' they can be used as summary functions in `summarise_draws()`.
#'
#' @return
#' A numeric or logical vector with the same dimensions as the given random variable, where
#' each entry in the vector is the mean, median, or variance of the corresponding entry in `x`.
#'
#' @examples
#'
#' set.seed(5678)
#' x = rvar_rng(rnorm, 4, mean = 1:4, sd = 2)
#'
#' # These should all be ~= c(1, 2, 3, 4)
#' E(x)
#' mean(x)
#' median(x)
#'
#' # This ...
#' Pr(x < 1.5)
#' # ... should be about the same as this:
#' pnorm(1.5, mean = 1:4, sd = 2)
#'
#' @name rvar-summaries-over-draws
#' @seealso [rvar-summaries-within-draws] for summary functions within draws.
#' [rvar-dist] for density, CDF, and quantile functions of random variables.
#' @family rvar-summaries
#' @export
E <- function(x, ...) {
  mean(x, ...)
}

#' @rdname rvar-summaries-over-draws
#' @export
mean.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "mean", matrixStats::colMeans2, useNames = FALSE, .ordered_okay = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
Pr <- function(x, ...) UseMethod("Pr")
#' @rdname rvar-summaries-over-draws
#' @export
Pr.default <- function(x, ...) {
  stop_no_call("Can only use `Pr()` on logical variables.")
}
#' @rdname rvar-summaries-over-draws
#' @export
Pr.logical <- function(x, ...) {
  mean(x, ...)
}
#' @rdname rvar-summaries-over-draws
#' @export
Pr.rvar <- function(x, ...) {
  if (!is.logical(draws_of(x))) {
    stop_no_call("Can only use `Pr()` on logical random variables.")
  }
  mean(x, ...)
}


# numeric summaries -------------------------------------------------------

#' @rdname rvar-summaries-over-draws
#' @export
median.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "median", matrixStats::colMedians, useNames = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
min.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "min", matrixStats::colMins, useNames = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
max.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "max", matrixStats::colMaxs, useNames = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
sum.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "sum", matrixStats::colSums2, useNames = FALSE, .ordered_okay = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
prod.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "prod", matrixStats::colProds, useNames = FALSE, .ordered_okay = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
all.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "all", matrixStats::colAlls, useNames = FALSE, .ordered_okay = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
any.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "any", matrixStats::colAnys, useNames = FALSE, .ordered_okay = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
Summary.rvar <- function(...) {
  # min, max, sum, prod, all, any, range --- these are all defined by more
  # specific functions to be faster, but I left the generic implementation here
  # on the off chance anything gets added to this group generic in the future
  f <- get(.Generic)
  summarise_rvar_by_element(.f = f, ...)
}


# spread ------------------------------------------------------------------

#' @importFrom distributional variance
#' @export
distributional::variance
#' @rdname rvar-summaries-over-draws
#' @export
variance.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "variance", matrixStats::colVars, useNames = FALSE, .ordered_okay = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
var <- function(x, ...) UseMethod("var")
#' @rdname rvar-summaries-over-draws
#' @export
var.default <- function(x, ...) stats::var(x, ...)
#' @rdname rvar-summaries-over-draws
#' @export
var.rvar <- variance.rvar

#' @rdname rvar-summaries-over-draws
#' @export
sd <- function(x, ...) UseMethod("sd")
#' @rdname rvar-summaries-over-draws
#' @export
sd.default <- function(x, ...) stats::sd(x, ...)
#' @rdname rvar-summaries-over-draws
#' @export
sd.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "sd", matrixStats::colSds, useNames = FALSE, .ordered_okay = FALSE, ...
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
mad <- function(x, ...) UseMethod("mad")
#' @rdname rvar-summaries-over-draws
#' @export
mad.default <- function(x, ...) stats::mad(x, ...)
#' @rdname rvar-summaries-over-draws
#' @export
mad.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "mad", matrixStats::colMads, useNames = FALSE, .ordered_okay = FALSE, ...
  )
}
#' @rdname rvar-summaries-over-draws
#' @export
mad.rvar_ordered <- function(x, ...) {
  mad(as_rvar_numeric(x), ...)
}


# range -------------------------------------------------------------------

#' @rdname rvar-summaries-over-draws
#' @export
range.rvar <- function(x, ...) {
  summarise_rvar_by_element_via_matrix(
    x, "range", function(...) t(matrixStats::colRanges(...)),
    useNames = FALSE, .extra_dim = 2, .extra_dimnames = list(NULL), ...
  )
}


# special value predicates ---------------------------------------------------------------

#' @rdname rvar-summaries-over-draws
#' @export
is.finite.rvar <- function(x) {
  summarise_rvar_by_element_via_matrix(
    x, "is.finite", function(x) matrixStats::colAlls(is.finite(x), useNames = FALSE), .factor_okay = TRUE
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
is.infinite.rvar <- function(x) {
  summarise_rvar_by_element_via_matrix(
    x, "is.inifite", function(x) matrixStats::colAnys(is.infinite(x), useNames = FALSE), .factor_okay = TRUE
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
is.nan.rvar <- function(x) {
  summarise_rvar_by_element_via_matrix(
    x, "is.nan", function(x) matrixStats::colAnys(is.nan(x), useNames = FALSE), .factor_okay = TRUE
  )
}

#' @rdname rvar-summaries-over-draws
#' @export
is.na.rvar <- function(x) {
  summarise_rvar_by_element_via_matrix(
    x, "is.na", matrixStats::colAnyNAs, useNames = FALSE, .factor_okay = TRUE
  )
}

#' @export
anyNA.rvar <- function(x, ...) anyNA(draws_of(x), ...)

#' Summaries of `draws` objects
#'
#' The `summarise_draws()` (and `summarize_draws()`) methods provide a quick way
#' to get a table of summary statistics and diagnostics. These methods will
#' convert an object to a `draws` object if it isn't already. For convenience, a
#' [summary()][base::summary] method for `draws` and `rvar` objects are also
#' provided as an alias for `summarise_draws()` if the input object is a `draws`
#' or `rvar` object.
#'
#' @name draws_summary
#'
#' @param .x,object (draws) A `draws` object or one coercible to a `draws` object.
#' @param ... Name-value pairs of summary or [diagnostic][diagnostics]
#'   functions. The provided names will be used as the names of the columns in
#'   the result *unless* the function returns a named vector, in which case the
#'   latter names are used. The functions can be specified in any format
#'   supported by [as_function()][rlang::as_function]. See **Examples**.
#' @param .args (named list) Optional arguments passed to the summary functions.
#' @param .num_args (named list) Optional arguments passed to
#'   [num()][tibble::num] for pretty printing of summaries. Can be controlled
#'   globally via the `posterior.num_args` [option][base::options].
#' @param .cores (positive integer) The number of cores to use for computing
#'   summaries for different variables in parallel. Coerced to integer if
#'   possible, otherwise errors. The default is `.cores = 1`, in which case no
#'   parallelization is implemented. By default, a socket cluster is used on
#'   Windows and forks otherwise.
#'
#' @return
#' The `summarise_draws()` methods return a [tibble][tibble::tibble] data frame.
#' The first column (`"variable"`) contains the variable names and the remaining
#' columns contain summary statistics and diagnostics.
#'
#' The functions `default_summary_measures()`, `default_convergence_measures()`,
#' and `default_mcse_measures()` return character vectors of names of the
#' default measures.
#'
#' @details
#' The default summary functions used are the ones specified by
#' `default_summary_measures()` and `default_convergence_measures()`:
#'
#' `default_summary_measures()`
#' * [mean()]
#' * [median()]
#' * [sd()]
#' * [mad()]
#' * [quantile2()]
#'
#' `default_convergence_measures()`
#' * [rhat()]
#' * [ess_bulk()]
#' * [ess_tail()]
#'
#' The `var()` function should not be used to compute variances due
#' to its inconsistent behavior with matrices. Instead, please use
#' `distributional::variance()`.
#'
#' @seealso [`diagnostics`] for a list of available diagnostics and links to
#'   their individual help pages.
#'
#' @examples
#' x <- example_draws("eight_schools")
#' class(x)
#' str(x)
#'
#' summarise_draws(x)
#' summarise_draws(x, "mean", "median")
#' summarise_draws(x, mean, mcse = mcse_mean)
#' summarise_draws(x, ~quantile(.x, probs = c(0.4, 0.6)))
#'
#' # using default_*_meaures()
#' summarise_draws(x, default_summary_measures())
#' summarise_draws(x, default_convergence_measures())
#' summarise_draws(x, default_mcse_measures())
#'
#' # compute variance of variables
#' summarise_draws(x, var = distributional::variance)
#'
#' # illustrate use of '.args'
#' ws <- rexp(ndraws(x))
#' summarise_draws(x, weighted.mean, .args = list(w = ws))
#'
#' # adjust how numerical summaries are printed
#' summarise_draws(x, .num_args = list(sigfig = 2, notation = "dec"))
#'
NULL

#' @rdname draws_summary
#' @export
summarise_draws <- function(.x, ...) {
  UseMethod("summarise_draws")
}

#' @rdname draws_summary
#' @export
summarize_draws <- summarise_draws


#' @export
summarise_draws.default <- function(.x, ...) {
  .x <- as_draws(.x)
  summarise_draws(.x, ...)
}

#' @rdname draws_summary
#' @export
summarise_draws.draws <- function(
    .x, ..., .args = list(),
    .num_args = getOption("posterior.num_args", list()),
    .cores = 1
  ) {

  if (ndraws(.x) == 0L) {
    return(empty_draws_summary())
  }

  .cores <- as_one_integer(.cores)
  if (.cores <= 0) {
    stop_no_call("'.cores' must be a positive integer.")
  }
  funs <- as.list(c(...))
  .args <- as.list(.args)
  if (length(funs)) {
    if (is.null(names(funs))) {
      # ensure names are initialized properly
      names(funs) <- rep("", length(funs))
    }
    calls <- substitute(list(...))[-1]
    calls <- ulapply(calls, deparse_pretty)
    for (i in seq_along(funs)) {
      fname <- NULL
      if (is.character(funs[[i]])) {
        fname <- as_one_character(funs[[i]])
      }
      # label unnamed arguments via their calls
      if (!nzchar(names(funs)[i])) {
        if (!is.null(fname)) {
          names(funs)[i] <- fname
        } else {
          names(funs)[i] <- calls[i]
        }
      }
      # get functions passed as strings from the right environments
      if (!is.null(fname)) {
        if (exists(fname, envir = caller_env())) {
          env <- caller_env()
        } else if (fname %in% getNamespaceExports("posterior")) {
          env <- asNamespace("posterior")
        } else {
          stop_no_call("Cannot find function '", fname, "'.")
        }
      }
      funs[[i]] <- rlang::as_function(funs[[i]], env = env)
    }
  } else {
    # default functions
    funs <- list(
      mean = base::mean,
      median = stats::median,
      sd = stats::sd,
      mad = stats::mad,
      quantile = quantile2,
      rhat = rhat,
      ess_bulk = ess_bulk,
      ess_tail = ess_tail
    )
  }

  # it is more efficient to repair and transform objects for all variables
  # at once instead of doing it within the loop for each variable separately
  .x <- repair_draws(.x)
  .x <- as_draws_array(.x)
  variables_x <- variables(.x)

  if (!length(variables_x)) {
    warning_no_call(
      "The draws object contained no variables with unreserved names. ",
      "No summaries were computed."
    )
    return(tibble::tibble(character()))
  }

  if (.cores == 1) {
    out <- summarise_draws_helper(.x, funs, .args)
  } else {
    .x <- .x[, , variables_x]
    n_vars <- length(variables_x)
    chunk_size <- ceiling(n_vars / .cores)
    n_chunks <- ceiling(n_vars / chunk_size)
    chunk_list <- vector(length = n_chunks, mode = "list")
    for (i in seq_len(n_chunks)) {
      if ((chunk_size * (i - 1) + 1) <= n_vars) {
        chunk <- (chunk_size * (i - 1) + 1):(min(c(chunk_size * i, n_vars)))
        chunk_list[[i]] <- .x[, , chunk]
      }
    }
    if (checkmate::test_os("windows")) {
      cl <- parallel::makePSOCKcluster(.cores)
      on.exit(parallel::stopCluster(cl))
      # exporting all these functions seems to be required to
      # pass GitHub actions checks on Windows
      parallel::clusterExport(
        cl,
        varlist = package_function_names("posterior"),
        envir = as.environment(asNamespace("posterior"))
      )
      parallel::clusterExport(
        cl,
        varlist = package_function_names("checkmate"),
        envir = as.environment(asNamespace("checkmate"))
      )
      parallel::clusterExport(
        cl,
        varlist = package_function_names("rlang"),
        envir = as.environment(asNamespace("rlang"))
      )
      summary_list <- parallel::parLapply(
        cl,
        X = chunk_list,
        fun = summarise_draws_helper,
        funs = funs,
        .args = .args
      )
    } else {
      summary_list <- parallel::mclapply(
        X = chunk_list,
        FUN = summarise_draws_helper,
        mc.cores = .cores,
        funs = funs,
        .args = .args
      )
    }
    out <- do.call("rbind", summary_list)
  }
  for (i in seq_cols(out)) {
    if (is.numeric(out[[i]])) {
      out[[i]] <- do.call(tibble::num, c(list(out[[i]]), .num_args))
    }
  }
  out
}

#' @rdname draws_summary
#' @export
summary.draws <- function(object, ...) {
  summarise_draws(object, ...)
}

#' @rdname draws_summary
#' @export
summarise_draws.rvar <- function(.x, ...) {
  x <- draws_rvars(x = .x)
  names(x) <- deparse_pretty(substitute(.x))
  summarise_draws(x, ...)
}

#' @rdname draws_summary
#' @export
summary.rvar <- function(object, ...) {
  .x <- draws_rvars(x = object)
  names(.x) <- deparse_pretty(substitute(object))
  summarise_draws(.x, ...)
}

#' @rdname draws_summary
#' @export
default_summary_measures <- function() {
  c("mean", "median", "sd", "mad", "quantile2")
}

#' @rdname draws_summary
#' @export
default_convergence_measures <- function() {
  c("rhat", "ess_bulk", "ess_tail")
}

#' @rdname draws_summary
#' @export
default_mcse_measures <- function() {
  c("mcse_mean", "mcse_median", "mcse_sd", "mcse_quantile")
}

class_draws_summary <- function() {
  c("draws_summary", "tbl_df", "tbl", "data.frame")
}

# empty draws_summary object
# @param dimensions names of dimensions to be added as empty columns
empty_draws_summary <- function(dimensions = "variable") {
  assert_character(dimensions, null.ok = TRUE)
  out <- tibble::tibble()
  for (d in dimensions) {
    out[[d]] <- character(0)
  }
  class(out) <- class_draws_summary()
  out
}


create_summary_list <- function(x, v, funs, .args) {
  draws <- drop_dims_or_classes(x[, , v], dims = 3, reset_class = FALSE)
  args <- c(list(draws), .args)
  v_summary <- named_list(names(funs))
  for (m in names(funs)) {
    v_summary[[m]] <- do.call(funs[[m]], args)
  }
  v_summary
}

summarise_draws_helper <- function(x, funs, .args) {
  variables_x <- variables(x)
  # get length and output names, calculated on the first variable
  out_1 <- create_summary_list(x, variables_x[1], funs, .args)
  the_names <- vector(mode = "list", length = length(funs))
  for (i in seq_along(out_1)){
    if (rlang::is_named(out_1[[i]])) {
      the_names[[i]] <- names(out_1[[i]])
    } else if (length(out_1[[i]]) > 1) {
      the_names[[i]] <- paste0(names(out_1)[i], ".", c(1:length(out_1[[i]])))
    } else {
      the_names[[i]] <- names(out_1)[i]
    }
  }
  the_names <- unlist(the_names)
  # Check for naming issues prior do doing lengthy computation
  if ("variable" %in% the_names) {
    stop_no_call("Name 'variable' is reserved in 'summarise_draws'.")
  }
  # Pre-allocate matrix to store output
  out <- matrix(NA, nrow = length(variables_x), ncol = length(the_names))
  colnames(out) <- the_names
  out[1, ] <- unlist(out_1)
  # Do the computation for all remaining variables
  if (length(variables_x) > 1L) {
    for (v_ind in 2:length(variables_x)) {
      out_v <- create_summary_list(x, variables_x[v_ind], funs, .args)
      out[v_ind, ] <- unlist(out_v)
    }
  }
  out <- tibble::as_tibble(out)
  out$variable <- variables_x
  out <- move_to_start(out, "variable")
  class(out) <- class_draws_summary()
  out
}

test_that("summarise_draws works correctly", {
  x <- as_draws_df(example_draws())
  sum_x <- summarise_draws(x)
  expect_true(all(default_convergence_measures() %in% names(sum_x)))
  expect_true(all(c("q5", "q95") %in% names(sum_x)))
  expect_equal(sum_x$variable, variables(x))
  expect_equal(mean(x$mu), as.numeric(sum_x$mean[sum_x$variable == "mu"]))

  sum_x <- summarise_draws(x, mean, median)
  expect_true(all(c("mean", "median") %in% names(sum_x)))

  sum_x <- summarise_draws(x, default_mcse_measures())
  expect_true(all(c("mcse_q5", "mcse_q95") %in% names(sum_x)))

  sum_x <- summarise_draws(x, ~quantile(.x, probs = c(0.4, 0.6)))
  expect_true(all(c("40%", "60%") %in% names(sum_x)))

  x[1, 1] <- NA
  sum_x <- summarise_draws(x)
  expect_true(is.na(sum_x[1, "q5"]))
  expect_true(all(c("q5", "q95") %in% names(sum_x)))
})

test_that("aliases of summarise_draws work", {
  x <- as_draws_array(example_draws())
  sum_x <- summarise_draws(x)
  sum_x2 <- summarize_draws(x)
  expect_equal(sum_x, sum_x2)
  sum_x3 <- summary(x)
  expect_equal(sum_x, sum_x3)
})

test_that("summarise_draws errors if name 'variable' is used", {
  x <- example_draws()
  variable <- function(x) mean(x)
  expect_error(
    summarise_draws(x, "variable"),
    "Name 'variable' is reserved in 'summarise_draws'"
  )
})

test_that("summarise_draws default method works", {
  expect_identical(
    summarise_draws(matrix(1:20, 10, 2)),
    summarise_draws(as_draws_matrix(matrix(1:20, 10, 2)))
  )
})

test_that("summarise_draws doesn't error for empty draws", {
  expect_identical(
    summarise_draws(empty_draws_array()),
    empty_draws_summary()
  )
})

test_that("summarise_draws and summary work for rvars", {
  d <- as_draws_rvars(example_draws())
  d_theta <- draws_rvars(x = d$theta)
  names(d_theta) <- "d$theta"
  ref <- summarise_draws(d_theta)

  expect_identical(summarise_draws(d$theta), ref)
  expect_identical(summary(d$theta), ref)
})

test_that("summarise_draws warns if all variable names are reserved", {
  x <- subset_draws(as_draws_df(example_draws()), variable = "mu")
  variables(x) <- ".log_weight"
  expect_warning(summarize_draws(x), "no variables with unreserved names")
})

test_that(paste(
  "multicore summarise_draws is identical to single-core summarise_draws",
  "including if some chunks contain no variables"
), {
  set.seed(1)
  cores <- 2
  nc <- 4

  n <- 20
  test_array <- array(data = rnorm(1000*nc*n), dim = c(1000,nc,n))
  x <- as_draws_array(test_array)
  sum_x <- summarise_draws(x)
  parsum_x <- summarise_draws(x, .cores = cores)
  expect_equal(sum_x, parsum_x)

  dimnames(x)$variable[2] <- reserved_variables()[1]
  sum_x <- summarise_draws(x)
  parsum_x <- summarise_draws(x, .cores = cores)
  expect_equal(sum_x, parsum_x)

  # test that externally defined summary functions can be found
  mean2 <- function(x) sum(x) / length(x)
  sum_x <- summarise_draws(x, mean2)
  parsum_x <- summarise_draws(x, mean2, .cores = cores)
  expect_identical(sum_x, parsum_x)

  n <- 2
  test_array <- array(data = rnorm(1000*nc*n), dim = c(1000,nc,n))
  x <- as_draws_array(test_array)
  sum_x <- summarise_draws(x)
  parsum_x <- summarise_draws(x, .cores = cores)
  expect_identical(sum_x, parsum_x)

  dimnames(x)$variable[2] <- reserved_variables()[1]
  sum_x <- summarise_draws(x)
  parsum_x <- summarise_draws(x, .cores = cores)
  expect_identical(sum_x, parsum_x)

  n <- 1
  test_array <- array(data = rnorm(1000*nc*n), dim = c(1000,nc,n))
  x <- as_draws_array(test_array)
  sum_x <- summarise_draws(x)
  parsum_x <- summarise_draws(x, .cores = cores)
  expect_identical(sum_x, parsum_x)

  dimnames(x)$variable[1] <- reserved_variables()[1]
  suppressWarnings(sum_x <- summarise_draws(x))
  suppressWarnings(parsum_x <- summarise_draws(x, .cores = cores))
  expect_identical(sum_x, parsum_x)
})

test_that("summarise_draws supports tibble::num correctly", {
  x <- example_draws()
  expect_output(
    print(summarise_draws(x, .num_args = list(sigfig = 2, notation="dec"))),
    "<dec:2>"
  )
})

test_that("summarise_draws errors for invalid cores specification", {
  x <- example_draws()
  expect_error(
    summarise_draws(x, .cores = -1),
    "'.cores' must be a positive integer"
  )
  expect_error(
    summarise_draws(x, .cores = NULL),
    "Cannot coerce '.cores' to a single integer value"
  )
})


test_that("summarise_draws works with variance()", {
  draws_array <- as_draws_array(example_draws())
  draws_matrix <- as_draws_matrix(draws_array)
  draws_df <- as_draws_df(draws_array)
  draws_list <- as_draws_list(draws_array)
  draws_rvars <- as_draws_rvars(draws_array)

  ref <- data.frame(
    variable = variables(draws_array),
    variance = tibble::num(as.vector(
      apply(draws_array, 3, function(x) var(as.vector(x)))
    )),
    stringsAsFactors = FALSE
  )
  class(ref) <- posterior:::class_draws_summary()

  expect_equal(summarise_draws(draws_array, variance), ref)
  expect_equal(summarise_draws(draws_matrix, variance), ref)
  expect_equal(summarise_draws(draws_df, variance), ref)
  expect_equal(summarise_draws(draws_list, variance), ref)
  expect_equal(summarise_draws(draws_rvars, variance), ref)

  # for consistency, draws_matrix and draws_array
  # have the same implementation of variance()
  expect_equal(variance(draws_array), var(as.vector(draws_array)))
  expect_equal(variance(draws_matrix), var(as.vector(draws_matrix)))
})

#' Probability integral transform
#'
#' Probability integral transform (PIT). LOO-PIT is given by a weighted sample.
#'
#' @name pit
#' 
#' @param x (draws) A [`draws_matrix`] object or one coercible to a
#'   `draws_matrix` object, or an [`rvar`] object.
#'
#' @param y (observations) A 1D vector, or an array of dim(x), if x is `rvar`.
#'   Each element of `y` corresponds to a variable in `x`.
#'
#' @param weights A matrix of weights for each draw and variable. `weights`
#'   should have one column per variable in `x`, and `ndraws(x)` rows.
#'
#' @param log (logical) Are the weights passed already on the log scale? The
#'   default is `FALSE`, that is, expecting `weights` to be on the standard
#'   (non-log) scale.
#'
#' @template args-methods-dots
#'
#' @details The `pit()` function computes the probability integral transform of
#'   `y` using the empirical cumulative distribution computed from the draws
#'   in `x`. For continuous valued `y` and `x`, the PIT for the elements of `y`
#'   is computed as the empirical cumulative distribution value:
#'
#'     PIT(y_i) = Pr(x_i < y_i),
#'
#'   where x_i, is the corresponding set of draws in `x`. For `draws` objects,
#'   this corresponds to the draws of the *i*th variable, and for `rvar`
#'   the elements of `y` and `x` are matched.
#'
#'   The draws in `x` can further be provided (log-)weights in
#    `weights`, which enables for example the computation of LOO-PITs.
#'
#'   If `y` and `x` are discrete, randomization is used to obtain continuous PIT
#'   values. (see, e.g., Czado, C., Gneiting, T., Held, L.: Predictive model
#'   assessment for count data.  Biometrics 65(4), 1254–1261 (2009).)
#'
#' @return A numeric vector of length `length(y)` containing the PIT values, or
#'   an array of shape `dim(y)`, if `x` is an `rvar`.

#' @examples
#' # PIT for a draws object
#' x <- example_draws()
#' # Create a vector of observations
#' y <- rnorm(nvariables(x), 5, 5)
#' pit(x, y)
#'
#' # Compute weighted PIT (for example LOO-PIT)
#' weights <- matrix(runif(length(x)), ncol = nvariables(x))
#'
#' pit(x, y, weights)
#'
#' # PIT for an rvar
#' x <- rvar(example_draws())
#' # Create an array of observations with the same dimensions as x.
#' y_arr <- array(rnorm(length(x), 5, 5), dim = dim(x))
#' pit(x, y_arr)
#'
NULL

#' @rdname pit
#' @export
pit <- function(x, y, ...) UseMethod("pit")

#' @rdname pit
#' @export
pit.default <- function(x, y, weights = NULL, log = FALSE, ...) {
  x <- as_draws_matrix(x)
  if (!is.null(weights)) {
    weights <- as_draws_matrix(weights)
  }
  pit(x, y, weights, log)
}

#' @rdname pit
#' @export
pit.draws_matrix <- function(x, y, weights = NULL, log = FALSE, ...) {
  y <- validate_y(y, x)
  if (!is.null(weights)) {
    weights <- sapply(seq_len(nvariables(x)), function(var_idx) {
      validate_weights(weights[, var_idx], x[, var_idx], log)
    })
    weights <- normalize_log_weights(weights)
  }
  pit <- vapply(seq_len(ncol(x)), function(j) {
    sel_min <- x[, j] < y[j]
    if (!any(sel_min)) {
      pit <- 0
    } else {
      if (is.null(weights)) {
        pit <- mean(sel_min)
      } else {
        pit <- exp(log_sum_exp(weights[sel_min, j]))
      }
    }

    sel_sup <- x[, j] == y[j]
    if (any(sel_sup)) {
      # randomized PIT for discrete y (see, e.g., Czado, C., Gneiting, T.,
      # Held, L.: Predictive model assessment for count data.
      # Biometrics 65(4), 1254–1261 (2009).)
      if (is.null(weights)) {
        pit_sup <- pit + mean(sel_sup)
      } else {
        pit_sup <- pit + exp(log_sum_exp(weights[sel_sup, j]))
      }

      pit <- runif(1, pit, pit_sup)
    }
    pit
  }, FUN.VALUE = 1.0)

  if (any(pit > 1 + 1e-10)) {
    warning_no_call(
      paste(
        "Some PIT values larger than 1. ",
        "This is usually due to numerical inaccuracies. ",
        "Largest value: ",
        max(pit),
        "\nRounding PIT > 1 to 1.",
        sep = ""
      )
    )
  }

  setNames(pmin(1, pit), variables(x))
}

#' @rdname pit
#' @export
pit.rvar <- function(x, y, weights = NULL, log = FALSE, ...) {
  y <- validate_y(y, x)
  if (is.null(weights)) {
    out <- array(
      runif(length(y), Pr(x < y), Pr(x <= y)),
      dim(x),
      dimnames(x)
    )
  } else {
    out <- array(
      data = pit(
        x = as_draws_matrix(c(x)),
        y = c(y),
        weights = weights,
        log = log
      ),
      dim = dim(x),
      dimnames = dimnames(x)
    )
  }
  out
}

#' Pareto-smoothed probability integral transform
#'
#' Compute PIT values using the empirical CDF, then refine values in
#' the tails by fitting a generalized Pareto distribution (GPD) to
#' the tail draws. This gives smoother, more accurate PIT values in
#' the tails where the ECDF is coarse, and avoids PIT values of 0 and 1.
#' Due to use of generalized Pareto distribution CDF in tails, the
#' PIT values are not anymore rank based and continuous uniformity
#' test is appropriate.
#'
#' @name pareto_pit
#'
#' @param x (draws) A [`draws_matrix`] object or one coercible to a
#'   `draws_matrix` object, or an [`rvar`] object.
#'
#' @param y (observations) A 1D vector, or an array of dim(x), if x is `rvar`.
#'   Each element of `y` corresponds to a variable in `x`.
#'
#' @param weights A matrix of weights for each draw and variable. `weights`
#'   should have one column per variable in `x`, and `ndraws(x)` rows.
#'
#' @param log (logical) Are the weights passed already on the log scale? The
#'   default is `FALSE`, that is, expecting `weights` to be on the standard
#'   (non-log) scale.
#'
#' @param ndraws_tail (integer) Number of tail draws to use for GPD
#'   fitting. If `NULL` (the default), computed using [ps_tail_length()].
#'
#' @template args-methods-dots
#'
#' @details The function first computes raw PIT values identically to
#'   [pit()] (including support for weighted draws). It then fits a
#'   GPD to both tails of the draws (using the same approach as
#'   [pareto_smooth()]) and replaces PIT values for observations falling in
#'   the tail regions:
#'
#'   For a right-tail observation \eqn{y_i > c_R} (where \eqn{c_R} is
#'   the right-tail cutoff):
#'
#'   \deqn{PIT(y_i) = 1 - p_{tail}(1 - F_{GPD}(y_i; c_R, \sigma_R, k_R))}
#'
#'   For a left-tail observation \eqn{y_i < c_L}:
#'
#'   \deqn{PIT(y_i) = p_{tail}(1 - F_{GPD}(-y_i; -c_L, \sigma_L, k_L))}
#'
#'   where \eqn{p_{tail}} is the proportion of (weighted) mass in the
#'   tail.
#'
#'   When (log-)weights in `weights` are provided, they are used for
#'   the raw PIT computation (as in [pit()]) and for GPD fit.
#'
#' @return A numeric vector of length `length(y)` containing the PIT values, or
#'   an array of shape `dim(y)`, if `x` is an `rvar`.
#'
#' @seealso [pit()] for the unsmoothed version, [pareto_smooth()] for
#'   Pareto smoothing of draws.
#'
#' @examples
#' x <- example_draws()
#' y <- rnorm(nvariables(x), 5, 5)
#' pareto_pit(x, y)
#'
NULL

#' @rdname pareto_pit
#' @export
pareto_pit <- function(x, y, ...) UseMethod("pareto_pit")

#' @rdname pareto_pit
#' @export
pareto_pit.default <- function(x, y, weights = NULL, log = FALSE,
                               ndraws_tail = NULL, ...) {
  x <- as_draws_matrix(x)
  if (!is.null(weights)) {
    weights <- as_draws_matrix(weights)
  }
  pareto_pit(x, y, weights = weights, log = log,
             ndraws_tail = ndraws_tail, ...)
}

#' @rdname pareto_pit
#' @export
pareto_pit.draws_matrix <- function(x, y, weights = NULL, log = FALSE,
                                    ndraws_tail = NULL, ...) {
  y <- validate_y(y, x)

  # validate and normalize weights to log scale (same as pit.draws_matrix)
  if (!is.null(weights)) {
    weights <- sapply(seq_len(nvariables(x)), function(var_idx) {
      validate_weights(weights[, var_idx], x[, var_idx], log)
    })
    weights <- normalize_log_weights(weights)
  }

  ndraws <- ndraws(x)

  if (is.null(ndraws_tail)) {
    ndraws_tail <- ps_tail_length(ndraws, 1)
  } else {
    ndraws_tail <- as_one_integer(ndraws_tail)
  }

  # validate ndraws_tail once for all variables
  gpd_ok <- !is.na(ndraws_tail) && ndraws_tail >= 5
  if (gpd_ok && ndraws_tail > ndraws / 2) {
    ndraws_tail <- floor(ndraws / 2)
  }
  if (gpd_ok && ndraws_tail >= ndraws) {
    gpd_ok <- FALSE
  }

  # precompute tail indices (shared across all variables)
  if (gpd_ok) {
    tail_ids <- seq(ndraws - ndraws_tail + 1, ndraws)
  }

  pit_values <- vapply(seq_len(ncol(x)), function(j) {
    draws <- x[, j]

    # --- raw PIT (same logic as pit.draws_matrix) ---
    sel_min <- draws < y[j]
    if (!any(sel_min)) {
      raw_pit <- 0
    } else {
      if (is.null(weights)) {
        raw_pit <- mean(sel_min)
      } else {
        raw_pit <- exp(log_sum_exp(weights[sel_min, j]))
      }
    }

    sel_sup <- draws == y[j]
    if (any(sel_sup)) {
      if (is.null(weights)) {
        pit_sup <- raw_pit + mean(sel_sup)
      } else {
        pit_sup <- raw_pit + exp(log_sum_exp(weights[sel_sup, j]))
      }
      raw_pit <- runif(1, raw_pit, pit_sup)
    }

    # --- GPD tail refinement ---
    if (!gpd_ok || should_return_NA(draws)) {
      return(raw_pit)
    }

    # sort draws and carry weights along
    ord <- sort.int(draws, index.return = TRUE)
    sorted <- ord$x
    log_wt_sorted <- if (!is.null(weights)) weights[ord$ix, j] else NULL

    # tail proportion: sum of (normalized) weights in the tail
    if (!is.null(log_wt_sorted)) {
      tail_proportion <- exp(log_sum_exp(log_wt_sorted[tail_ids]))
    } else {
      tail_proportion <- ndraws_tail / ndraws
    }

    # --- right tail ---
    right_replaced <- FALSE
    right_tail <- sorted[tail_ids]
    if (!is_constant(right_tail)) {
      right_cutoff <- sorted[min(tail_ids) - 1]
      if (right_cutoff == right_tail[1]) {
        right_cutoff <- right_cutoff - .Machine$double.eps
      }
      right_tail_wt <- if (!is.null(log_wt_sorted)) {
        exp(log_wt_sorted[tail_ids])
      } else {
        NULL
      }
      right_fit <- gpdfit(right_tail - right_cutoff, sort_x = FALSE,
                          weights = right_tail_wt)
      if (is.finite(right_fit$k) && !is.na(right_fit$sigma)) {
        if (y[j] > right_cutoff) {
          gpd_cdf <- pgeneralized_pareto(
            y[j], mu = right_cutoff, sigma = right_fit$sigma, k = right_fit$k
          )
          raw_pit <- 1 - tail_proportion * (1 - gpd_cdf)
          right_replaced <- TRUE
        }
      }
    }

    # --- left tail (negate trick, same as ps_tail) ---
    if (!right_replaced) {
      left_draws <- -draws
      left_ord <- sort.int(left_draws, index.return = TRUE)
      left_sorted <- left_ord$x
      log_wt_left_sorted <- if (!is.null(weights)) weights[left_ord$ix, j] else NULL

      left_tail <- left_sorted[tail_ids]
      if (!is_constant(left_tail)) {
        left_cutoff <- left_sorted[min(tail_ids) - 1]
        if (left_cutoff == left_tail[1]) {
          left_cutoff <- left_cutoff - .Machine$double.eps
        }
        left_tail_wt <- if (!is.null(log_wt_left_sorted)) {
          exp(log_wt_left_sorted[tail_ids])
        } else {
          NULL
        }
        left_fit <- gpdfit(left_tail - left_cutoff, sort_x = FALSE,
                           weights = left_tail_wt)
        if (is.finite(left_fit$k) && !is.na(left_fit$sigma)) {
          if (-y[j] > left_cutoff) {
            gpd_cdf <- pgeneralized_pareto(
              -y[j], mu = left_cutoff, sigma = left_fit$sigma, k = left_fit$k
            )
            if (!is.null(log_wt_left_sorted)) {
              left_tail_proportion <- exp(log_sum_exp(log_wt_left_sorted[tail_ids]))
            } else {
              left_tail_proportion <- tail_proportion
            }
            raw_pit <- left_tail_proportion * (1 - gpd_cdf)
          }
        }
      }
    }

    raw_pit
  }, FUN.VALUE = 1.0)

  min_tail_prob <- 1/ndraws/1e4
  pit_values <- pmin(pmax(pit_values, min_tail_prob), 1-min_tail_prob)
  setNames(pit_values, variables(x))
}

#' @rdname pareto_pit
#' @export
pareto_pit.rvar <- function(x, y, weights = NULL, log = FALSE,
                            ndraws_tail = NULL, ...) {
  y <- validate_y(y, x)
  out <- array(
    data = pareto_pit(
      x = as_draws_matrix(c(x)),
      y = c(y),
      weights = weights,
      log = log,
      ndraws_tail = ndraws_tail,
      ...
    ),
    dim = dim(x),
    dimnames = dimnames(x)
  )
  out
}

# internal ----------------------------------------------------------------

validate_y <- function(y, x = NULL) {
  if (!is.numeric(y)) {
    stop_no_call("`y` must be numeric.")
  }
  if (anyNA(y)) {
    stop_no_call("NAs not allowed in `y`.")
  }
  if (is_rvar(x)) {
    if (length(x) != length(y) || any(dim(y) != dim(x))) {
      stop_no_call("`dim(y)` must match `dim(x)`.")
    }
  } else if (is_draws(x)) {
    if (!is.vector(y, mode = "numeric") || length(y) != nvariables(x)) {
      stop_no_call("`y` must be a vector of length `nvariables(x)`.")
    }
  }
  y
}

normalize_log_weights <- function(log_weights) {
  apply(log_weights, 2, function(col) col - log_sum_exp(col))
}

test_that("validate_y handles numeric inputs correctly", {
  expect_equal(
    validate_y(c(1, 2, 3), as_draws_matrix(matrix(1, ncol = 3))),
    c(1, 2, 3)
  )
})

test_that("validate_y raises an error for non-numeric 'y'", {
  expect_error(validate_y(c("a", "b", "c")), "`y` must be numeric.")
})

test_that("validate_y raises an error for NAs in 'y'", {
  expect_error(validate_y(c(1, NA, 3)), "NAs not allowed in `y`.")
})

test_that("validate_y checks 'y' length against nvariables(x)", {
  x <- as_draws_matrix(example_draws())
  expect_error(
    validate_y(1:3, x),
    "`y` must be a vector of length `nvariables(x)`.",
    fixed = TRUE
  )
})

test_that("validate_y verifies dimensions for array inputs", {
  x <- rvar(posterior::example_draws())
  y <- mean(rvar(posterior::example_draws()))
  expect_equal(validate_y(y, x), y)

  y_mismatched <- array(1:8, dim = c(2, 2, 2))
  expect_error(
    validate_y(y_mismatched, x),
    "`dim(y)` must match `dim(x)`.",
    fixed = TRUE
  )
})

# test normalize_log_weights
test_that("normalize_log_weights returns log-normalized columns", {
  set.seed(1)
  x <- matrix(log(runif(200)), ncol = 10)
  result <- colSums(exp(normalize_log_weights(x)))
  expect_equal(result, rep(1, 10))
})

# tests for pit.default
test_that("pit works without weights", {
  x <- matrix(c(1, 2, 3, 5), nrow = 2)
  y <- c(3, 4)
  result <- pit(x, y)

  expect_length(result, ncol(x))
  expect_equal(unname(result), c(1, 0.5))
})

test_that("pit.default works with log-weights", {
  x <- matrix(c(1, 2, 3, 5), nrow = 2)
  y <- c(3, 4)
  weights <- matrix(log(c(1, 1, 1, 1)), nrow = 2)
  result <- pit.default(x, y, weights = weights, log = TRUE)

  expect_length(result, ncol(x))
  expect_equal(unname(result), c(1, 0.5))
})

test_that("pit.default works with non-log weights", {
  x <- matrix(c(1, 2, 3, 5), nrow = 2)
  y <- c(3, 4)
  weights <- matrix(c(1, 1, 1, 1), nrow = 2)
  result <- pit.default(x, y, weights = weights)

  expect_length(result, ncol(x))
  expect_equal(unname(result), c(1, 0.5))
})

test_that("pit.default handles randomized PIT with no weights", {
  x <- matrix(c(rep(c(0, 1), 1000), rep(c(3, 3), 1000)), nrow = 2)
  y <- rep(c(1, 3), each = 1000)
  result <- pit.default(x, y)

  expect_length(result, ncol(x))
  expect_true(all(result[1:1000] >= .5 & result[1:1000] <= 1))
  expect_true(all(result[1001:2000] >= 0 & result[1001:2000] <= 1))
})

test_that("pit.default handles randomized PIT with weights", {
  x <- matrix(c(rep(c(0, 1), 1000), rep(c(3, 3), 1000)), nrow = 2)
  y <- rep(c(1, 3), each = 1000)
  weights <- matrix(.5, nrow = 2, ncol = 2000)
  result <- pit.default(x, y, weights)

  expect_length(result, ncol(x))
  expect_true(all(result[1:1000] >= .5 & result[1:1000] <= 1))
  expect_true(all(result[1001:2000] >= 0 & result[1001:2000] <= 1))
})

test_that("pit.default handles randomized PIT with log-weights", {
  x <- matrix(c(rep(c(0, 1), 1000), rep(c(3, 3), 1000)), nrow = 2)
  y <- rep(c(1, 3), each = 1000)
  weights <- matrix(log(.5), nrow = 2, ncol = 2000)
  result <- pit.default(x, y, weights, log = TRUE)

  expect_length(result, ncol(x))
  expect_true(all(result[1:1000] >= .5 & result[1:1000] <= 1))
  expect_true(all(result[1001:2000] >= 0 & result[1001:2000] <= 1))
})

test_that("pit works with draws objects", {
  set.seed(1)
  n_chains <- 4
  n_vars <- 3
  n_draws <- 100

  test_array <- array(
    rnorm(n_draws * n_vars * n_chains),
    dim = c(n_draws, n_chains, n_vars)
  )

  y <- rnorm(n_vars)

  pit_true <- sapply(1:n_vars, function(v) mean(test_array[,, v] < y[v]))

  x <- as_draws(test_array)

  expect_equal(unname(pit(x, y)), pit_true)
  expect_equal(unname(pit(as_draws_df(test_array), y)), pit_true)
  expect_equal(unname(pit(as_draws_matrix(test_array), y)), pit_true)
  expect_equal(unname(pit(as_draws_list(test_array), y)), pit_true)
  expect_equal(unname(pit(as_draws_rvars(test_array), y)), pit_true)
})

test_that("pit works with rvars", {
  set.seed(1)
  n_col <- 4
  n_row <- 3
  n_draws <- 100

  test_array <- array(
    rnorm(n_draws * n_col * n_row),
    dim = c(n_draws, n_row, n_col)
  )

  weights <- matrix(
    runif(n_draws * n_col * n_row),
    ncol = n_row * n_col
  )

  y <- array(rnorm(n_col * n_row), dim = c(n_row, n_col))

  result <- pit(rvar(test_array), y)
  expect_equal(dim(result), dim(y))
  expect_true(all(
    result ==
      array(
        pit(array(test_array, dim = c(n_draws, n_col * n_row)), c(y)),
        dim(y)
      )
  ))

  result <- pit(rvar(test_array), y, weights)
  expect_equal(dim(result), dim(y))
  expect_true(all(
    result ==
      array(
        pit(
          array(test_array, dim = c(n_draws, n_col * n_row)),
          c(y),
          array(weights, dim = c(n_draws, n_col * n_row))
        ),
        dim(y)
      )
  ))
})

test_that("pit doesn't error for empty draws", {
  expect_numeric(pit(empty_draws_array(), numeric(0)))
  expect_numeric(pit(rvar(), numeric(0)))
})

# ---- basic behavior ----

test_that("pareto_pit returns named vector of correct length", {
  set.seed(1)
  x <- as_draws_matrix(matrix(rnorm(2000), ncol = 4))
  y <- c(-1, 0, 1, 2)
  result <- pareto_pit(x, y)

  expect_length(result, 4)
  expect_named(result, variables(x))
  expect_true(all(result >= 0 & result <= 1))
})

test_that("pareto_pit bulk values match pit", {
  # For observations well inside the bulk, pareto_pit and pit should agree
  set.seed(42)
  ndraws <- 1000
  x <- as_draws_matrix(matrix(rnorm(ndraws * 3), ncol = 3))
  # observations near the median (well within bulk)
  y <- c(0, 0.1, -0.1)

  raw <- pit(x, y)
  refined <- pareto_pit(x, y)

  expect_equal(unname(refined), unname(raw))
})

test_that("pareto_pit differs from pit in tails", {
  set.seed(42)
  ndraws <- 1000
  x <- as_draws_matrix(matrix(rnorm(ndraws * 2), ncol = 2))
  # extreme observations in the tails
  y <- c(-4, 4)

  raw <- pit(x, y)
  refined <- pareto_pit(x, y)

  # refined values should differ from raw ECDF in the tails
  expect_false(isTRUE(all.equal(unname(refined), unname(raw))))
  expect_true(all(refined >= 0 & refined <= 1))
})

# ---- known-distribution round-trip ----

test_that("pareto_pit gives values closer to true CDF than raw pit for right tail", {
  set.seed(123)
  ndraws <- 500
  nvars <- 200

  # t-distribution with 3 df (heavy tails where GPD fits well)
  x <- as_draws_matrix(matrix(rt(ndraws * nvars, df = 3), ncol = nvars))

  # Observations far in the right tail
  y_right <- rep(qt(0.999, df = 3), nvars)
  true_pit <- rep(0.999, nvars)

  raw <- pit(x, y_right)
  refined <- pareto_pit(x, y_right)

  # On average, pareto_pit should be closer to the true CDF value
  raw_mse <- mean((raw - true_pit)^2)
  refined_mse <- mean((refined - true_pit)^2)

  expect_lt(refined_mse, raw_mse)
})

test_that("pareto_pit gives values closer to true CDF in left tail", {
  set.seed(123)
  ndraws <- 500
  nvars <- 200

  # t-distribution with 3 df
  x <- as_draws_matrix(matrix(rt(ndraws * nvars, df = 3), ncol = nvars))

  # Observations far in the left tail
  y_left <- rep(qt(0.001, df = 3), nvars)
  true_pit <- rep(0.001, nvars)

  raw <- pit(x, y_left)
  refined <- pareto_pit(x, y_left)

  raw_mse <- mean((raw - true_pit)^2)
  refined_mse <- mean((refined - true_pit)^2)

  expect_lt(refined_mse, raw_mse)
})

test_that("pareto_pit provides more varied estimates than ECDF in extreme tails", {
  # When observations are beyond most draws, ECDF gives 0 or 1 for many,
  # while GPD can extrapolate and give more varied (and accurate) estimates
  set.seed(123)
  ndraws <- 500
  nvars <- 200

  x <- as_draws_matrix(matrix(rt(ndraws * nvars, df = 3), ncol = nvars))
  y_right <- rep(qt(0.9995, df = 3), nvars)

  raw <- pit(x, y_right)
  refined <- pareto_pit(x, y_right)

  # Raw ECDF should have many values exactly at 1
  # Refined should have more varied values
  expect_gt(length(unique(refined)), length(unique(raw)))
})

# ---- discrete randomization ----

test_that("pareto_pit uses randomization for discrete observations in bulk", {
  set.seed(1)
  x <- as_draws_matrix(matrix(rep(c(0, 1, 2, 3), each = 250), ncol = 1))
  y <- 2  # matches some draws exactly

  results <- replicate(100, {
    pareto_pit(x, y)
  })

  # With randomization, results should vary
  expect_gt(sd(results), 0)
  # All results should be in valid range
  expect_true(all(results >= 0 & results <= 1))
})

# ---- ndraws_tail argument ----

test_that("pareto_pit respects ndraws_tail argument", {
  set.seed(42)
  ndraws <- 1000
  x <- as_draws_matrix(matrix(rnorm(ndraws), ncol = 1))
  y <- 3  # right tail

  result_small <- pareto_pit(x, y, ndraws_tail = 10)
  result_large <- pareto_pit(x, y, ndraws_tail = 100)

  # Different ndraws_tail should give different results
  expect_false(isTRUE(all.equal(unname(result_small), unname(result_large))))
})

# ---- edge cases ----

test_that("pareto_pit handles constant draws gracefully", {
  x <- as_draws_matrix(matrix(rep(5, 200), ncol = 1))
  y <- 5

  # Should not error; falls back to raw PIT
  result <- pareto_pit(x, y)
  expect_length(result, 1)
  expect_true(result >= 0 & result <= 1)
})

test_that("pareto_pit handles NA/Inf draws gracefully", {
  x <- as_draws_matrix(matrix(c(1:99, NA), ncol = 1))
  y <- 50

  # should_return_NA triggers, falls back to raw PIT
  expect_warning(result <- pareto_pit(x, y))
  expect_length(result, 1)
})

test_that("pareto_pit works when all observations are in bulk", {
  set.seed(42)
  ndraws <- 1000
  x <- as_draws_matrix(matrix(rnorm(ndraws * 3), ncol = 3))
  y <- c(0, 0, 0)

  result <- pareto_pit(x, y)
  raw <- pit(x, y)

  expect_equal(unname(result), unname(raw))
})

# ---- S3 dispatch ----

test_that("pareto_pit.default coerces to draws_matrix", {
  set.seed(42)
  x_mat <- matrix(rnorm(2000), ncol = 2)
  y <- c(-3, 3)

  result <- pareto_pit(x_mat, y)
  expect_length(result, 2)
  expect_true(all(result >= 0 & result <= 1))
})

test_that("pareto_pit works with different draws formats", {
  set.seed(42)
  n_chains <- 4
  n_vars <- 3
  n_draws <- 100

  test_array <- array(
    rnorm(n_draws * n_vars * n_chains),
    dim = c(n_draws, n_chains, n_vars)
  )
  y <- c(0, 0, 0)

  result_matrix <- pareto_pit(as_draws_matrix(test_array), y)
  result_df <- pareto_pit(as_draws_df(test_array), y)
  result_list <- pareto_pit(as_draws_list(test_array), y)

  expect_equal(unname(result_matrix), unname(result_df))
  expect_equal(unname(result_matrix), unname(result_list))
})

# ---- rvar method ----

test_that("pareto_pit.rvar returns correct shape", {
  set.seed(42)
  x_rvar <- rvar(matrix(rnorm(1000 * 6), nrow = 1000, ncol = 6))
  dim(x_rvar) <- c(2, 3)
  y_arr <- array(c(0, 0, 0, 0, 0, 0), dim = c(2, 3))

  result <- pareto_pit(x_rvar, y_arr)

  expect_equal(dim(result), c(2, 3))
  expect_true(all(result >= 0 & result <= 1))
})

test_that("pareto_pit.rvar matches draws_matrix result", {
  set.seed(42)
  x_rvar <- rvar(matrix(rnorm(1000 * 4), nrow = 1000, ncol = 4))
  y_vec <- c(-3, 0, 1, 3)

  result_rvar <- c(pareto_pit(x_rvar, y_vec))
  result_matrix <- unname(pareto_pit(as_draws_matrix(c(x_rvar)), y_vec))

  expect_equal(result_rvar, result_matrix)
})

# ---- values are in valid range ----

test_that("pareto_pit values are always in [0, 1]", {
  set.seed(42)
  ndraws <- 500
  x <- as_draws_matrix(matrix(rnorm(ndraws * 10), ncol = 10))
  y <- qnorm(c(0.001, 0.01, 0.05, 0.1, 0.3, 0.7, 0.9, 0.95, 0.99, 0.999))

  result <- pareto_pit(x, y)
  expect_true(all(result >= 0 & result <= 1))
})

test_that("pareto_pit handles observations beyond all draws", {
  set.seed(42)
  ndraws <- 500
  x <- as_draws_matrix(matrix(rnorm(ndraws), ncol = 1))
  y_extreme_right <- 100
  y_extreme_left <- -100

  result_right <- pareto_pit(x, y_extreme_right)
  result_left <- pareto_pit(x, y_extreme_left)

  expect_true(result_right > 0.99)
  expect_true(result_left < 0.01)
  expect_true(result_right >= 0 & result_right <= 1)
  expect_true(result_left >= 0 & result_left <= 1)
})

# ---- monotonicity in tails ----

test_that("pareto_pit is monotonic in the right tail", {
  set.seed(42)
  ndraws <- 1000
  x <- as_draws_matrix(matrix(rnorm(ndraws), ncol = 1))
  y_seq <- seq(2, 4, by = 0.2)

  results <- sapply(y_seq, function(yi) pareto_pit(x, yi))
  # PIT values should be non-decreasing
  expect_true(all(diff(results) >= 0))
})

test_that("pareto_pit is monotonic in the left tail", {
  set.seed(42)
  ndraws <- 1000
  x <- as_draws_matrix(matrix(rnorm(ndraws), ncol = 1))
  y_seq <- seq(-4, -2, by = 0.2)

  results <- sapply(y_seq, function(yi) pareto_pit(x, yi))
  # PIT values should be non-decreasing
  expect_true(all(diff(results) >= 0))
})

# ---- weighted pareto_pit ----

test_that("pareto_pit with uniform weights matches unweighted", {
  set.seed(42)
  ndraws <- 500
  nvars <- 4
  x <- as_draws_matrix(matrix(rnorm(ndraws * nvars), ncol = nvars))
  y <- c(-3, 0, 1, 3)
  w_uniform <- matrix(1, nrow = ndraws, ncol = nvars)

  result_none <- pareto_pit(x, y)
  result_unif <- pareto_pit(x, y, weights = w_uniform)

  expect_equal(result_none, result_unif)
})

test_that("pareto_pit with log weights matches standard weights", {
  set.seed(42)
  ndraws <- 500
  nvars <- 4
  x <- as_draws_matrix(matrix(rnorm(ndraws * nvars), ncol = nvars))
  y <- c(-3, 0, 1, 3)
  w <- matrix(runif(ndraws * nvars, 0.5, 2), nrow = ndraws, ncol = nvars)

  result_std <- pareto_pit(x, y, weights = w)
  result_log <- pareto_pit(x, y, weights = log(w), log = TRUE)

  expect_equal(result_std, result_log)
})

test_that("pareto_pit with non-uniform weights differs from unweighted", {
  set.seed(42)
  ndraws <- 500
  nvars <- 2
  x <- as_draws_matrix(matrix(rnorm(ndraws * nvars), ncol = nvars))
  y <- c(-3, 3)  # tail observations to ensure GPD refinement kicks in
  # strongly non-uniform weights: heavily upweight positive draws
  w <- matrix(
    ifelse(x > 0, 10, 1),
    nrow = ndraws, ncol = nvars
  )

  result_none <- pareto_pit(x, y)
  result_wt <- pareto_pit(x, y, weights = w)

  expect_false(isTRUE(all.equal(unname(result_none), unname(result_wt))))
  expect_true(all(result_wt >= 0 & result_wt <= 1))
})

test_that("pareto_pit weighted values are in [0, 1]", {
  set.seed(42)
  ndraws <- 500
  nvars <- 10
  x <- as_draws_matrix(matrix(rt(ndraws * nvars, df = 3), ncol = nvars))
  y <- qt(c(0.001, 0.01, 0.05, 0.1, 0.3, 0.7, 0.9, 0.95, 0.99, 0.999), df = 3)
  w <- matrix(runif(ndraws * nvars, 0.1, 5), nrow = ndraws, ncol = nvars)

  result <- pareto_pit(x, y, weights = w)
  expect_true(all(result >= 0 & result <= 1))
})

test_that("pareto_pit weighted bulk matches pit weighted bulk", {
  set.seed(42)
  ndraws <- 500
  nvars <- 3
  x <- as_draws_matrix(matrix(rnorm(ndraws * nvars), ncol = nvars))
  y <- c(0, 0.1, -0.1)  # well in the bulk
  w <- matrix(runif(ndraws * nvars, 0.5, 2), nrow = ndraws, ncol = nvars)

  raw <- pit(x, y, weights = w)
  refined <- pareto_pit(x, y, weights = w)

  expect_equal(unname(refined), unname(raw))
})

test_that("pareto_pit.rvar passes weights through", {
  set.seed(42)
  ndraws <- 500
  nvars <- 4
  x_rvar <- rvar(matrix(rnorm(ndraws * nvars), nrow = ndraws, ncol = nvars))
  y_vec <- c(-3, 0, 1, 3)
  w <- matrix(runif(ndraws * nvars, 0.5, 2), nrow = ndraws, ncol = nvars)

  result_rvar <- c(pareto_pit(x_rvar, y_vec, weights = w))
  result_matrix <- unname(pareto_pit(as_draws_matrix(c(x_rvar)), y_vec, weights = w))

  expect_equal(result_rvar, result_matrix)
})

test_that("pareto_pit.default passes weights through", {
  set.seed(42)
  ndraws <- 500
  nvars <- 2
  x_mat <- matrix(rnorm(ndraws * nvars), ncol = nvars)
  y <- c(-3, 3)
  w <- matrix(runif(ndraws * nvars, 0.5, 2), nrow = ndraws, ncol = nvars)

  result <- pareto_pit(x_mat, y, weights = w)
  expect_length(result, 2)
  expect_true(all(result >= 0 & result <= 1))
})

#' Uniformity test for PIT values
#'
#' Tests whether PIT (probability integral transform) values deviate from
#' uniformity while accounting for dependence among them. LOO-PIT values are
#' not independent because each LOO predictive distribution is conditioned on
#' nearly the same data (all but one observation), which induces correlation.
#' Standard uniformity tests that assume independence can have inflated
#' Type I error or reduced power. This function implements dependence-aware
#' procedures from Tesso and Vehtari (2026).
#'
#' Three test variants are available:
#'
#' * **POT** (Pointwise Order Tests): Uses beta distributions for order
#'   statistics. Best for continuous PIT values; recommended as default for
#'   LOO-PIT model checking. Good power against diverse departures from
#'   uniformity.
#' * **PIET** (Pointwise Inverse-CDF Evaluation Tests): Uses a continuous
#'   reference distribution (exponential) via inverse-CDF transformation.
#'   Best for detecting tail deviations; maintains valid Type I error control
#'   under strong dependence.
#' * **PRIT** (Pointwise Rank-based Individual Tests): Uses binomial
#'   distributions on scaled ECDF (ranks). Intended for discrete or
#'   rank-based PIT values.
#'
#' All procedures compute pointwise p-values, aggregate them via the (truncated)
#' Cauchy combination test (Liu and Xie, 2020; Chen et al., 2025), and derive
#' Shapley values to quantify each point's contribution to the overall test.
#'
#' @param pit Numeric vector of PIT values in `[0, 1]`.
#' @param test Character string. One of `"POT"`, `"PIET"`, or `"PRIT"`.
#'   See details above.
#'
#' @return A list with components:
#'   * `pvalue`: Global p-value from the Cauchy combination test.
#'   * `pointwise`: Shapley values (contributions of each PIT value to the
#'     test statistic). Non-negative values indicate points contributing to
#'     evidence against uniformity.
#'
#' @references
#'   Tesso, H., and Vehtari, A. (2026). LOO-PIT predictive model checking.
#'   arXiv preprint arXiv:2603.02928.
#'
#'   Liu, Y., and Xie, J. (2020). Cauchy combination test: a powerful test
#'   with analytic p-value calculation under arbitrary dependency structures.
#'   *Journal of the American Statistical Association*, 115(529), 393-402.
#'
#'   Chen, B., Xu, W., and Gao, X. (2025). Truncated Cauchy combination test:
#'   a robust and powerful p-value combination method with arbitrary
#'   correlations. arXiv preprint arXiv:2506.12489.
#'
#' @export
uniformity_test <- function(pit, test) {

  choices <- c("POT", "PRIT", "PIET")
  if (!test %in% choices) {
    stop("'test' should be one of ", paste(dQuote(choices), collapse = ", "), call. = FALSE)
  }
  test <- match.arg(test, choices = choices)

  test_fn <- switch(test,
      POT  = .pot_test,
      PIET = .piet_test,
      PRIT = .prit_test
    )

    pit_sorted           <- sort(pit)
    std_cauchy_values    <- .compute_cauchy(test_fn(pit_sorted))
    p_value_CCT          <- .cauchy_combination_test(
      test_fn(pit),
      truncate = test != "PIET"
    )
    pointwise_contributions  <- .compute_shapley_values(std_cauchy_values)

    res <- list()
    res$pvalue      <- p_value_CCT
    res$pointwise   <- pointwise_contributions

    return(res)
}


# internal ----------------------------------------------------------------

#' Compute Shapley values
#'
#' Calculates the average marginal contribution of players across 
#' all random arrival orders in a cooperative game.
#' Used to provide a principled approach for quantifying 
#' point-specific influences in a way that reflects local miscalibration.
#'
#' @param x Numeric vector of Cauchy-transformed PIT values.
#' @return Numeric vector of Shapley values with the same length as `x`.
#' @noRd
.compute_shapley_values <- function(x) {
  n <- length(x)
  if (n == 0) {
    return(numeric(0))
  }
  if (n == 1) {
    return(0)
  }
  
  # Harmonic number
  # H_n = sum(1/i) for i = 1 to n
  harmonic_number <- sum(1 / seq_len(n))
  
  shapley_values <- numeric(n)
  for (i in seq_len(n)) {
    mean_others <- sum(x[-i]) / (n - 1)
    # Applies the closed-form formula to assign player i their fair share.
    shapley_values[i] <- (1 / n) * x[i] + ((harmonic_number - 1) / n) * (x[i] - mean_others)
  }

  return(shapley_values)
}

#' Pointwise Inverse-CDF Evaluation Tests Combination (PIET)
#'
#' Uniformity test with respect to any continuous distribution.
#' H0: The value obtained via the inverse CDF transformation F^(-1)(x_i)
#' follows the distribution of X under uniformity.
#' HA: The p-value p_(i) provides evidence against uniformity.
#' 
#' @param x Numeric vector of PIT values in `[0, 1]`.
#' @return Numeric vector of p-values.
#' @noRd
.piet_test <- function(x) {
  cdf_exp <- pexp(-log(x), rate = 1) # same as 1-x but numerically more stable
  p_values <- 2 * pmin(cdf_exp, 1 - cdf_exp)

  return(p_values)
}

#' Pointwise Order Tests Combination (POT)
#'
#' Uniformity test based on a beta distribution.
#' H0: The i-th order statistic u_(i) follows a beta distribution
#' under uniformity.
#' HA: The p-value p_(i) provides evidence against uniformity
#' at the i-th order statistic u_(i).
#' 
#' @param x Numeric vector of PIT values in `[0, 1]`.
#' @return Numeric vector of p-values.
#' @noRd
.pot_test <- function(x) {
  n <- length(x)
  # keep NA values instead of silent recycling via sort()
  # TODO: Can PIT values be NAN at this point?
  cdf_beta <- pbeta(sort(x, na.last = TRUE), 1:n, seq(n, 1, by = -1))
  p_values <- 2 * pmin(cdf_beta, 1 - cdf_beta)

  return(p_values)
}

#' Pointwise Rank-based Individual Tests Combination (PRIT)
#' 
#' Uniformity test based on a binomial distribution.
#' H0: The number of observations falling at or below x_i
#' follows a binomial distribution under uniformity.
#' HA: The p-value p_i provides evidence against uniformity.
#' 
#' @param x Numeric vector of PIT values in `[0, 1]`.
#' @return Numeric vector of p-values.
#' @noRd
.prit_test <- function(x) {
  n <- length(x)
  scaled_ecdf <- n * ecdf(x)(x)
  probs1 <- pbinom(scaled_ecdf - 1, n, x)
  probs2 <- pbinom(scaled_ecdf, n, x)
  p_values <- 2 * pmin(1 - probs1, probs2)
  
  return(p_values)
}

#' Truncated Cauchy combination test
#'
#' Combines dependent p-values using the Cauchy combination method.
#' If truncate, only p-values less than 0.5 are included.
#'
#' @param x Numeric vector of p-values transformed to follow a standard
#' Cauchy distribution.
#' @param truncate Boolean; If TRUE only p-values less than 0.5 are
#' included.
#' @return p-value of the Cauchy combination method.
#' @noRd
.cauchy_combination_test <- function(x, truncate = NULL) {
  if (truncate) {
    idx <- which(x < 0.5)
    if (length(idx) == 0) {
      stop("Cannot compute truncated Cauchy combination test. ",
           "No p-values below 0.5 found.")
    }
    1 - pcauchy(mean(-qcauchy(x[idx])))
  } else {
    1 - pcauchy(mean(-qcauchy(x)))
  }
}

#' Compute Cauchy transformation
#'
#' Transforms PIT values to follow a standard Cauchy distribution.
#'
#' @param x Numeric vector of PIT values in `[0, 1]`.
#' @return Numeric vector of Cauchy-transformed values.
#' @noRd
.compute_cauchy <- function(x) {
  tan((0.5 - x) * pi)
}

# Tests for uniformity_test (main entry point) ---------------------------------
set.seed(123)

test_that("uniformity_test returns list with pvalue and pointwise", {
  pit <- runif(20)
  for (test in c("POT", "PIET", "PRIT")) {
    res <- posterior:::uniformity_test(pit, test)
    expect_type(res, "list")
    expect_named(res, c("pvalue", "pointwise"))
    expect_length(res$pvalue, 1)
    expect_true(res$pvalue >= 0 && res$pvalue <= 1)
    expect_equal(length(res$pointwise), length(pit))
  }
})

test_that("uniformity_test is deterministic for same inputs", {
  pit <- runif(15)
  for (test in c("POT", "PIET", "PRIT")) {
    res1 <- posterior:::uniformity_test(pit, test)
    res2 <- posterior:::uniformity_test(pit, test)
    expect_equal(res1$pvalue, res2$pvalue)
    expect_equal(res1$pointwise, res2$pointwise)
  }
})

test_that("uniformity_test handles single-element pit", {
  pit <- 0.4
  res <- posterior:::uniformity_test(pit, "PIET")
  expect_equal(length(res$pointwise), 1)
  expect_equal(res$pointwise, 0)  # Shapley for n=1 is 0
  expect_true(is.finite(res$pvalue))
  
})

test_that("uniformity_test errors on invalid test", {
  pit <- runif(5)
  expect_error(posterior:::uniformity_test(pit, "INVALID"),
               "'test' should be one of")
})

# Tests for the dependence-aware uniformity tests ------------------------------
test_that("piet_test computes correct p-values", {
  x <- c(0.1, 0.25, 0.5, 0.75, 0.9)
  # Equivalent to 2 * min(1-x, x)
  expected <- c(0.2, 0.5, 1.0, 0.5, 0.2)
  
  expect_equal(.piet_test(x), expected, tolerance = 1e-7)
})

test_that("piet_test handles boundary values of 0 and 1", {
  x <- c(0, 1)
  expected <- c(0, 0)
  
  expect_equal(.piet_test(x), expected)
})

test_that("piet_test handles extreme values stably", {
  # Testing values very close to 0 and 1
  x <- c(1e-17, 1 - 1e-17)
  expected <- c(2e-17, 2e-17)
  
  # Tolerance needs to be adjusted for very small numbers
  expect_equal(.piet_test(x), expected, tolerance = 1e-16)
})

test_that("piet_test handles NA, NaN, and empty inputs correctly", {
  # NA and NaN propagation
  x_na <- c(0.5, NA, NaN)
  res_na <- .piet_test(x_na)
  
  expect_equal(res_na[1], 1.0)
  expect_true(is.na(res_na[2]))
  expect_true(is.nan(res_na[3]))
})

test_that("pot_test calculates correct p-values", {
  # Manually calculating expected values for x = c(0.2, 0.8), n = 2. 
  # Note: If X ~ Beta(1, b) then X ~ Kumaraswamy(1, b) with CDF 1 - (1 - x)^b 
  # and if X ~ Beta(a, 1) then X ~ Kumaraswamy(a, 1) with CDF to x^a
  # Beta(1, 2) CDF at 0.2 is 1 - (1 - 0.2)^2 = 0.36
  # Beta(2, 1) CDF at 0.8 is 0.8^2 = 0.64
  # p-values: 2 * min(0.36, 1-0.36) = 0.72; 2 * min(0.64, 1-0.64) = 0.72
  x <- c(0.8, 0.2)
  expected <- c(0.72, 0.72)
  
  expect_equal(.pot_test(x), expected)
})

test_that("pot_test handles boundary values correctly", {
  x <- c(0, 1)
  expected <- c(0, 0)
  
  expect_equal(.pot_test(x), expected)
})

test_that("pot_test bounds p-values between 0 and 1 for extreme out-of-bounds inputs", {
  # pbeta handles values outside [0, 1] by returning 0
  x <- c(-0.5, 1.5)
  expected <- c(0, 0)
  
  expect_equal(.pot_test(x), expected)
})

test_that("pot_test handles NAs", {
  x_na <- c(0.5, NA) # resulting in a = 2, b = 1
  # Beta(2, 1) at 0.5 = 0.25 -> 2 * min(0.25, 0.75) = 0.5
  expected <- c(0.5, NA)

  expect_equal(.pot_test(x_na), expected)
})

test_that("prit_test computes correct p-values", {
  # Let n = 2, x = c(0.5, 0.5) 
  # scaled_ecdf = 2 * c(1, 1) = c(2, 2)
  # probs1 = pbinom(1, 2, 0.5) = 0.75
  # probs2 = pbinom(2, 2, 0.5) = 1.00
  # p_val = 2 * min(1 - 0.75, 1.00) = 2 * 0.25 = 0.5
  
  x <- c(0.5, 0.5)
  expect_equal(.prit_test(x), c(0.5, 0.5))
})

# Test for computation of Shapley values -------------------------------------

test_that("compute_shapley_values handles empty vector and single element", {
  result_empty <- .compute_shapley_values(numeric(0))
  result_single <- .compute_shapley_values(5)

  expect_equal(result_empty, numeric(0))
  expect_equal(length(result_empty), 0)
  expect_equal(result_single, 0)
  expect_equal(length(result_single), 1)
})

test_that("compute_shapley_values for simple case", {
  x <- c(1, 2)
  result <- .compute_shapley_values(x)
  
  # Manual calculation for n=2:
  # harmonic_number = 1 + 1/2 = 1.5
  # For i=1: mean_others = 2/1 = 2
  # shapley[1] = (1/2)*1 + ((1.5-1)/2)*(1-2) = 0.5 - 0.25 = 0.25
  # For i=2: mean_others = 1/1 = 1
  # shapley[2] = (1/2)*2 + ((1.5-1)/2)*(2-1) = 1 + 0.25 = 1.25
  
  expected <- c(0.25, 1.25)
  expect_equal(result, expected, tolerance = 1e-10)
  expect_equal(length(result), 2)
})

test_that("compute_shapley_values handles mixed input values", {
  x <- c(-0.2, 0, 2, 3.1, 4.2)
  result <- .compute_shapley_values(x)
  expect_equal(length(result), 5)
  expect_true(all(is.finite(result)))
})

# Test for (truncated) Cauchy combination test ----------------------------------
test_that("cauchy_combination_test handles truncate = FALSE", {
  # avg = mean(-qcauchy(x))
  # result = 1 - pcauchy(avg)
  
  x <- c(0.1, 0.2, 0.3)
  result <- .cauchy_combination_test(x, truncate = FALSE)
  expected <- 1 - pcauchy(mean(-qcauchy(x)))
  
  expect_equal(result, expected, tolerance = 1e-10)
  expect_true(is.finite(result))
  expect_true(result >= 0 && result <= 1)
})

test_that("cauchy_combination_test handles truncate = TRUE", {
  # avg = mean(-qcauchy(x))
  # result = 1 - pcauchy(avg)
  
  x <- c(0.1, 0.2, 0.3, 0.4, 0.7, 0.8)
  result <- .cauchy_combination_test(x, truncate = TRUE)
  expected <- 1 - pcauchy(mean(-qcauchy(c(0.1, 0.2, 0.3, 0.4))))
  
  expect_equal(result, expected, tolerance = 1e-10)
  expect_true(is.finite(result))
  expect_true(result >= 0 && result <= 1)
})

test_that("cauchy_combination_test handles boundary values", {
  # x = 0: -qdf(0) = Inf and cdf(Inf) = 1 -> 1 - 1 = 0
  # x = 1: -qdf(1) = -Inf and cdf(-Inf) = 0 -> 1 - 0 = 1 
  
  expect_equal(.cauchy_combination_test(0, truncate = FALSE), 0)
  expect_equal(.cauchy_combination_test(1, truncate = FALSE), 1)
  expect_true(is.nan(.cauchy_combination_test(c(0, 1), truncate = FALSE)))
  # TODO: if 1 included in vector, CCT will always evaluate to 0 
  # as the mean evaluates to Inf and 1 - cdf(Inf) = 1 - 1 = 0
  expect_equal(.cauchy_combination_test(c(0, 0.3, 0.4, 1), truncate = TRUE), 0)
})

# Test for compute_cauchy -----------------------------------------------------

test_that("compute_cauchy computes correct transformations", {
  # For x = 0.5: tan((0.5 - 0.5) * pi) = tan(0) = 0
  expect_equal(.compute_cauchy(0.5), 0, tolerance = 1e-10)
  
  # For x = 0.25: tan((0.5 - 0.25) * pi) = tan(0.25 * pi) = 1
  expect_equal(.compute_cauchy(0.25), 1, tolerance = 1e-10)
  
  # For x = 0.75: tan((0.5 - 0.75) * pi) = tan(-0.25 * pi) = -1
  expect_equal(.compute_cauchy(0.75), -1, tolerance = 1e-10)
})

#' Thin `draws` objects
#'
#' Thin [`draws`] objects to reduce their size and autocorrelation in
#' the chains.
#'
#' @aliases thin
#' @template args-methods-x
#' @param thin (positive numeric) The period for selecting draws. Must
#'   be between 1 and the number of iterations. If the value is not an
#'   integer, the draws will be selected such that the number of draws
#'   returned is equal to round(ndraws(x) / thin). Intervals between
#'   selected draws will be either ceiling(thin) or floor(thin), such
#'   that the average interval will be close to the thin value. If
#'   `NULL`, it will be automatically calculated based on bulk and
#'   tail effective sample size as suggested by Säilynoja et
#'   al. (2022).
#' @template args-methods-dots
#' @template ref-sailynoja-ecdf-2022
#' @template return-draws
#'
#' @examples
#' x <- example_draws()
#' niterations(x)
#'
#' x <- thin_draws(x, thin = 5)
#' niterations(x)
#'
#' @export
thin_draws <- function(x, thin = NULL, ...) {
  UseMethod("thin_draws")
}

#' @rdname thin_draws
#' @export
thin_draws.draws <- function(x, thin = NULL, ...) {
  if (is.null(thin)) {
    thin <- ess_based_thinning_all_vars(x)
    message("Automatically thinned by ", round(thin, 1), " based on ESS.")
  }

  thin <- as_one_numeric(thin)
  if (thin == 1L) {
    # no thinning requested
    return(x)
  }
  if (thin <= 1L) {
    stop_no_call("'thin' must be greater than or equal to 1")
  }
  niterations <- niterations(x)
  if (thin > niterations) {
    stop_no_call("'thin' must be smaller than the total number of iterations.")
  }
  iteration_ids <- round(seq(1, niterations, by = thin))
  subset_draws(x, iteration = iteration_ids)
}

#' @rdname thin_draws
#' @export
thin_draws.rvar <- function(x, thin = NULL, ...) {
  thin_draws(draws_rvars(x = x), thin, ...)$x
}

ess_based_thinning_all_vars <- function(x, ...) {
  max(summarise_draws(x, thin = ess_based_thinning)$thin)
}

ess_based_thinning <- function(x, ...) {
  # thin based on mean (over chains) of minimum of tail and bulk ess
  x <- as.matrix(x)
  ess_tailbulk_chains <- apply(x,
    MARGIN = 2,
    FUN = function(x) min(SW(ess_tail(x)), SW(ess_bulk(x)))
    )
  nrow(x) / mean(ess_tailbulk_chains)
}

#' Quantile function for the generalized Pareto distribution
#'
#' Computes the quantile function for a generalized Pareto distribution
#' with location `mu`, scale `sigma`, and shape `k`.
#' 
#' @param p Numeric vector of probabilities.
#' @param mu Location parameter.
#' @param sigma Scale parameter (must be positive).
#' @param k Shape parameter.
#' @param lower.tail Logical; if `TRUE` (default), probabilities are `P[X <= x]`.
#' @param log.p Logical; if `TRUE`, probabilities are given as `log(p)`.
#' @return A numeric vector of quantiles.
#' @keywords internal
#' @export
#' @examples
#' qgeneralized_pareto(p = c(0.1, 0.5, 0.9), mu = 0, sigma = 1, k = 0.2)
qgeneralized_pareto <- function(p, mu = 0, sigma = 1, k = 0, lower.tail = TRUE, log.p = FALSE) {
  stopifnot(length(mu) == 1 && length(sigma) == 1 && length(k) == 1)
  if (is.na(sigma) || sigma <= 0) {
    return(rep(NaN, length(p)))
  }
  if (log.p) {
    p <- exp(p)
  }
  if (!lower.tail) {
    p <- 1 - p
  }
  if (k == 0) {
    q <-  mu - sigma * log1p(-p)
  } else {
    q <- mu + sigma * expm1(-k * log1p(-p)) / k
  }
  q
}

#' Distribution function for the generalized Pareto distribution
#'
#' Computes the cumulative distribution function (CDF) for a generalized
#' Pareto distribution with location `mu`, scale `sigma`, and shape `k`.
#'
#' @param q Numeric vector of quantiles.
#' @param mu Location parameter.
#' @param sigma Scale parameter (must be positive).
#' @param k Shape parameter.
#' @param lower.tail Logical; if `TRUE` (default), probabilities are `P[X <= x]`.
#' @param log.p Logical; if `TRUE`, probabilities are returned as `log(p)`.
#' @return A numeric vector of probabilities.
#' @keywords internal
#' @export
#' @examples
#' pgeneralized_pareto(q = c(1, 2, 5), mu = 0, sigma = 1, k = 0.2)
pgeneralized_pareto <- function(q, mu = 0, sigma = 1, k = 0, lower.tail = TRUE, log.p = FALSE) {
  stopifnot(length(mu) == 1 && length(sigma) == 1 && length(k) == 1)
  if (is.na(sigma) || sigma <= 0) {
    return(rep(NaN, length(q)))
  }
  z <- (q - mu) / sigma
  if (abs(k) < 1e-15) {
    # for very small values of indistinguishable in floating point accuracy from the case k=0
    p <- -expm1(-z)
  } else {
    # pmax handles values outside the support
    p <- -expm1(log1p(pmax(k * z, -1)) / -k)
  }
  # force to [0, 1] for values outside the support
  p <- pmin(pmax(p, 0), 1)
  if (!lower.tail) {
    p <- 1 - p
  }
  if (log.p) {
    p <- log(p)
  }
  p
}

#' Estimate parameters of the Generalized Pareto distribution
#'
#' Given a sample \eqn{x}, Estimate the parameters \eqn{k} and
#' \eqn{\sigma} of the generalized Pareto distribution (GPD), assuming
#' the location parameter is 0. By default the fit uses a prior for
#' \eqn{k} (this is in addition to the prior described by Zhang and
#' Stephens, 2009), which will stabilize estimates for very small
#' sample sizes (and low effective sample sizes in the case of MCMC
#' samples). The weakly informative prior is a Gaussian prior centered
#' at 0.5 (see details in Vehtari et al., 2024). This is used
#' internally but is exported for use by other packages.
#' @family helper-functions
#' @param x A numeric vector. The sample from which to estimate the
#'   parameters.
#' @param wip Logical indicating whether to adjust \eqn{k} based on a
#'   weakly informative Gaussian prior centered on 0.5. Defaults to
#'   `TRUE`.
#' @param min_grid_pts The minimum number of grid points used in the
#'   fitting algorithm. The actual number used is `min_grid_pts +
#'   floor(sqrt(length(x)))`.
#' @param sort_x If `TRUE` (the default), the first step in the
#'   fitting algorithm is to sort the elements of `x`. If `x` is
#'   already sorted in ascending order then `sort_x` can be set to
#'   `FALSE` to skip the initial sorting step.
#' @param weights An optional numeric vector of positive weights the same
#'   length as `x`. If `NULL` (the default), all observations are
#'   weighted equally and the result is identical to the unweighted fit.
#'   Weights are normalized internally to sum to `length(x)`.
#' @return A named list with components `k` and `sigma`.
#'
#' @details Here the parameter \eqn{k} is the negative of \eqn{k} in Zhang &
#'   Stephens (2009).
#'
#'
#' @references
#' Zhang, J., and Stephens, M. A. (2009). A new and efficient estimation method
#' for the generalized Pareto distribution. *Technometrics* **51**, 316-325.
#'
#' @keywords internal
#' @export
gpdfit <- function(x, wip = TRUE, min_grid_pts = 30, sort_x = TRUE,
                   weights = NULL) {
  # see section 4 of Zhang and Stephens (2009)
  if (sort_x) {
    if (!is.null(weights)) {
      ord <- sort.int(x, index.return = TRUE)
      x <- ord$x
      weights <- weights[ord$ix]
    } else {
      x <- sort.int(x)
    }
  }
  N <- length(x)

  # normalize weights to sum to N so the log-likelihood scale is preserved
  if (!is.null(weights)) {
    weights <- weights / sum(weights) * N
  }

  prior <- 3
  M <- min_grid_pts + floor(sqrt(N))
  jj <- seq_len(M)
  xstar <- x[floor(N / 4 + 0.5)] # first quartile of sample
  if (xstar > x[1])  {
    # first quantile is bigger than the minimum
    theta <- 1 / x[N] + (1 - sqrt(M / (jj - 0.5))) / prior / xstar

    # log1p(-theta %o% x) is M x N matrix
    log1p_mat <- log1p(-theta %o% x)

    if (!is.null(weights)) {
      # weighted mean across observations for each theta value
      k <- drop(log1p_mat %*% weights) / N
    } else {
      k <- matrixStats::rowMeans2(log1p_mat)
    }

    l_theta <- N * (log(-theta / k) - k - 1) # profile log-lik
    w_theta <- exp(l_theta - matrixStats::logSumExp(l_theta)) # normalize
    theta_hat <- sum(theta * w_theta)

    if (!is.null(weights)) {
      k_hat <- sum(weights * log1p(-theta_hat * x)) / N
    } else {
      k_hat <- mean.default(log1p(-theta_hat * x))
    }
    sigma_hat <- -k_hat / theta_hat

    # adjust k_hat based on weakly informative prior, Gaussian centered on 0.5.
    # this stabilizes estimates for very small Monte Carlo sample sizes and low ess
    # (see Vehtari et al., 2024 for details)
    if (wip) {
      k_hat <- (k_hat * N + 0.5 * 10) / (N + 10)
    }
    if (is.na(k_hat)) {
      k_hat <- Inf
      sigma_hat <- NaN
    }
  } else {
    # first quantile is not bigger than the minimum, which indicates
    # that the distribution is far from a generalized Pareto
    # distribution
    k_hat <- NA
    sigma_hat <- NA
  }

  list(k = k_hat, sigma = sigma_hat)
}

# --- qgeneralized_pareto tests ------------------------------------------------

test_that("qgeneralized_pareto returns correct quantiles for k != 0", {
  # Known values for GPD(mu=0, sigma=1, k=0.5)
  # Q(p) = (1/k) * ((1-p)^(-k) - 1) = 2 * ((1-p)^(-0.5) - 1)
  p <- c(0, 0.25, 0.5, 0.75, 1)
  expected <- 2 * ((1 - p)^(-0.5) - 1)
  result <- qgeneralized_pareto(p, mu = 0, sigma = 1, k = 0.5)
  expect_equal(result, expected)
})

test_that("qgeneralized_pareto returns correct quantiles for k == 0 (exponential)", {
  # When k == 0, GPD reduces to exponential: Q(p) = mu - sigma * log(1-p)
  p <- c(0.1, 0.5, 0.9)
  expected <- -log(1 - p)
  result <- qgeneralized_pareto(p, mu = 0, sigma = 1, k = 0)
  expect_equal(result, expected)
})

test_that("qgeneralized_pareto respects mu and sigma", {
  p <- c(0.1, 0.5, 0.9)
  base <- qgeneralized_pareto(p, mu = 0, sigma = 1, k = 0.2)

  # Shifting mu shifts quantiles
  shifted <- qgeneralized_pareto(p, mu = 5, sigma = 1, k = 0.2)
  expect_equal(shifted, base + 5)

  # Scaling sigma scales quantiles (relative to mu)
  scaled <- qgeneralized_pareto(p, mu = 0, sigma = 3, k = 0.2)
  expect_equal(scaled, base * 3)
})

test_that("qgeneralized_pareto handles lower.tail = FALSE", {
  p <- c(0.1, 0.5, 0.9)
  result_lower <- qgeneralized_pareto(p, mu = 0, sigma = 1, k = 0.2)
  result_upper <- qgeneralized_pareto(1 - p, mu = 0, sigma = 1, k = 0.2, lower.tail = FALSE)
  expect_equal(result_lower, result_upper)
})

test_that("qgeneralized_pareto handles log.p = TRUE", {
  p <- c(0.1, 0.5, 0.9)
  result <- qgeneralized_pareto(p, mu = 0, sigma = 1, k = 0.2)
  result_log <- qgeneralized_pareto(log(p), mu = 0, sigma = 1, k = 0.2, log.p = TRUE)
  expect_equal(result, result_log)
})

test_that("qgeneralized_pareto returns NaN for invalid sigma", {
  result <- qgeneralized_pareto(0.5, mu = 0, sigma = -1, k = 0.2)
  expect_true(is.nan(result))

  result <- qgeneralized_pareto(0.5, mu = 0, sigma = 0, k = 0.2)
  expect_true(is.nan(result))

  result <- qgeneralized_pareto(c(0.1, 0.5), mu = 0, sigma = NA, k = 0.2)
  expect_length(result, 2)
  expect_true(all(is.nan(result)))
})

test_that("qgeneralized_pareto errors on vector parameters", {
  expect_error(qgeneralized_pareto(0.5, mu = c(0, 1), sigma = 1, k = 0.2))
  expect_error(qgeneralized_pareto(0.5, mu = 0, sigma = c(1, 2), k = 0.2))
  expect_error(qgeneralized_pareto(0.5, mu = 0, sigma = 1, k = c(0.1, 0.2)))
})

# --- pgeneralized_pareto tests ------------------------------------------------

test_that("pgeneralized_pareto returns correct probabilities for k != 0", {
  # CDF: p = 1 - (1 + k*z)^(-1/k), z = (q-mu)/sigma
  q <- c(0, 1, 2, 5)
  k <- 0.5
  expected <- 1 - (1 + k * q)^(-1 / k)
  result <- pgeneralized_pareto(q, mu = 0, sigma = 1, k = k)
  expect_equal(result, expected)
})

test_that("pgeneralized_pareto returns correct probabilities for k == 0 (exponential)", {
  q <- c(0, 0.5, 1, 2, 5)
  expected <- 1 - exp(-q)
  result <- pgeneralized_pareto(q, mu = 0, sigma = 1, k = 0)
  expect_equal(result, expected)
})

test_that("pgeneralized_pareto and qgeneralized_pareto are inverses (k != 0)", {
  p <- seq(0.05, 0.95, by = 0.05)
  params <- list(
    list(mu = 0, sigma = 1, k = 0.2),
    list(mu = 5, sigma = 2, k = -0.3),
    list(mu = -1, sigma = 0.5, k = 0.5)
  )
  for (par in params) {
    q <- qgeneralized_pareto(p, mu = par$mu, sigma = par$sigma, k = par$k)
    p_back <- pgeneralized_pareto(q, mu = par$mu, sigma = par$sigma, k = par$k)
    expect_equal(p_back, p, tolerance = 1e-12)
  }
})

test_that("pgeneralized_pareto and qgeneralized_pareto are inverses (k == 0)", {
  p <- seq(0.05, 0.95, by = 0.05)
  q <- qgeneralized_pareto(p, mu = 0, sigma = 1, k = 0)
  p_back <- pgeneralized_pareto(q, mu = 0, sigma = 1, k = 0)
  expect_equal(p_back, p, tolerance = 1e-12)
})

test_that("pgeneralized_pareto clamps values outside support", {
  # Below the location parameter, CDF should be 0
  result <- pgeneralized_pareto(-1, mu = 0, sigma = 1, k = 0.2)
  expect_equal(result, 0)

  # For k < 0 there is an upper bound at mu - sigma/k
  # k = -0.5, sigma = 1 => upper bound = 0 - 1/(-0.5) = 2
  result <- pgeneralized_pareto(3, mu = 0, sigma = 1, k = -0.5)
  expect_equal(result, 1)
})

test_that("pgeneralized_pareto handles lower.tail = FALSE", {
  q <- c(0.5, 1, 2)
  result_lower <- pgeneralized_pareto(q, mu = 0, sigma = 1, k = 0.2)
  result_upper <- pgeneralized_pareto(q, mu = 0, sigma = 1, k = 0.2, lower.tail = FALSE)
  expect_equal(result_lower + result_upper, rep(1, 3))
})

test_that("pgeneralized_pareto handles log.p = TRUE", {
  q <- c(0.5, 1, 2)
  result <- pgeneralized_pareto(q, mu = 0, sigma = 1, k = 0.2)
  result_log <- pgeneralized_pareto(q, mu = 0, sigma = 1, k = 0.2, log.p = TRUE)
  expect_equal(result_log, log(result))
})

test_that("pgeneralized_pareto returns NaN for invalid sigma", {
  result <- pgeneralized_pareto(1, mu = 0, sigma = -1, k = 0.2)
  expect_true(is.nan(result))

  result <- pgeneralized_pareto(1, mu = 0, sigma = 0, k = 0.2)
  expect_true(is.nan(result))

  result <- pgeneralized_pareto(c(1, 2), mu = 0, sigma = NA, k = 0.2)
  expect_length(result, 2)
  expect_true(all(is.nan(result)))
})

test_that("pgeneralized_pareto errors on vector parameters", {
  expect_error(pgeneralized_pareto(1, mu = c(0, 1), sigma = 1, k = 0.2))
  expect_error(pgeneralized_pareto(1, mu = 0, sigma = c(1, 2), k = 0.2))
  expect_error(pgeneralized_pareto(1, mu = 0, sigma = 1, k = c(0.1, 0.2)))
})

test_that("pgeneralized_pareto at mu equals 0", {
  # CDF at the location parameter should be 0
  expect_equal(pgeneralized_pareto(0, mu = 0, sigma = 1, k = 0), 0)
  expect_equal(pgeneralized_pareto(0, mu = 0, sigma = 1, k = 0.5), 0)
  expect_equal(pgeneralized_pareto(5, mu = 5, sigma = 2, k = 0.3), 0)
})

# --- gpdfit tests -------------------------------------------------------------

test_that("gpdfit with uniform weights matches unweighted fit", {
  set.seed(42)
  x <- sort(rexp(200))
  fit_no_wt <- gpdfit(x, sort_x = FALSE)
  fit_uniform <- gpdfit(x, sort_x = FALSE, weights = rep(1, 200))
  expect_equal(fit_uniform$k, fit_no_wt$k)
  expect_equal(fit_uniform$sigma, fit_no_wt$sigma)
})

test_that("gpdfit with scaled uniform weights matches unweighted fit", {
  set.seed(42)
  x <- sort(rexp(200))
  fit_no_wt <- gpdfit(x, sort_x = FALSE)
  fit_scaled <- gpdfit(x, sort_x = FALSE, weights = rep(7.3, 200))
  expect_equal(fit_scaled$k, fit_no_wt$k)
  expect_equal(fit_scaled$sigma, fit_no_wt$sigma)
})

test_that("gpdfit with non-uniform weights differs from unweighted fit", {
  set.seed(42)
  x <- sort(rexp(200))
  w <- c(rep(1, 150), rep(5, 50))
  fit_unif <- gpdfit(x, sort_x = FALSE)
  fit_wt <- gpdfit(x, sort_x = FALSE, weights = w)
  expect_false(isTRUE(all.equal(fit_unif$k, fit_wt$k)))
  expect_false(isTRUE(all.equal(fit_unif$sigma, fit_wt$sigma)))
})

test_that("gpdfit sort_x = TRUE reorders weights correctly", {
  set.seed(42)
  x <- rexp(200)
  w <- runif(200, 0.5, 2)

  fit_auto <- gpdfit(x, sort_x = TRUE, weights = w)

  ord <- order(x)
  fit_manual <- gpdfit(x[ord], sort_x = FALSE, weights = w[ord])

  expect_equal(fit_auto$k, fit_manual$k)
  expect_equal(fit_auto$sigma, fit_manual$sigma)
})

test_that("gpdfit with zero-weight observations effectively ignores them", {
  set.seed(42)
  x <- rexp(100)

  # Fit on just the original 100 observations
  fit_base <- gpdfit(x, sort_x = TRUE)

  # Add 50 outlier observations with zero weight
  x_extra <- c(x, rexp(50, rate = 0.01))
  w_extra <- c(rep(1, 100), rep(0, 50))
  fit_zero <- gpdfit(x_extra, sort_x = TRUE, weights = w_extra)

  # Not exactly equal (N changes the grid and WIP prior), but close
  expect_equal(fit_zero$k, fit_base$k, tolerance = 0.15)
  expect_equal(fit_zero$sigma, fit_base$sigma, tolerance = 0.15)
})

test_that("gpdfit upweighting tail draws changes the fit", {
  set.seed(42)
  x <- sort(rexp(300))

  fit_unif <- gpdfit(x, sort_x = FALSE)

  # Upweight the largest values
  w <- rep(1, 300)
  w[251:300] <- 3
  fit_heavy <- gpdfit(x, sort_x = FALSE, weights = w)

  expect_false(isTRUE(all.equal(fit_unif$sigma, fit_heavy$sigma)))
})

test_that("gpdfit with weights = NULL is backward compatible", {
  set.seed(42)
  x <- rexp(150)

  fit1 <- gpdfit(x)
  fit2 <- gpdfit(x, weights = NULL)
  expect_identical(fit1, fit2)
})

test_that("gpdfit returns NA when first quartile equals minimum", {
  # gpdfit should return NA if xstar == x[1]
  # This can happen when the distribution is far from a generalized Pareto distribution

  # Case 1: All values are the same (constant)
  x_constant <- rep(5, 100)
  result <- gpdfit(x_constant)
  expect_true(is.na(result$k))
  expect_true(is.na(result$sigma))

  # Case 2: First quartile equals minimum (many repeated minimum values)
  # More than 25% of values are at the minimum
  x_many_min <- c(rep(1, 30), seq(2, 10, length.out = 70))
  result <- gpdfit(x_many_min)
  expect_true(is.na(result$k))
  expect_true(is.na(result$sigma))

  # Case 3: Exactly 25% at minimum (edge case where xstar == x[1])
  x_quarter_min <- c(rep(0, 25), seq(1, 10, length.out = 75))
  result <- gpdfit(x_quarter_min)
  expect_true(is.na(result$k))
  expect_true(is.na(result$sigma))

})

test_that("gpdfit returns valid estimates for exponential-like tails", {
  set.seed(1234)

  # Exponential distribution (k should be close to 0)
  x <- sort(rexp(1000))
  result <- gpdfit(x)

  expect_false(is.na(result$k))
  expect_false(is.na(result$sigma))
  expect_true(is.finite(result$k))
  expect_true(result$sigma > 0)
})

test_that("gpdfit handles sort_x argument correctly", {
  set.seed(1234)

  x <- rexp(100)
  x_sorted <- sort(x)

  # Results should be the same whether we sort or provide pre-sorted data
  result_sort <- gpdfit(x, sort_x = TRUE)
  result_presorted <- gpdfit(x_sorted, sort_x = FALSE)

  expect_equal(result_sort$k, result_presorted$k)
  expect_equal(result_sort$sigma, result_presorted$sigma)
})

test_that("gpdfit handles wip argument correctly", {
  set.seed(1234)

  x <- sort(rexp(100))

  result_wip <- gpdfit(x, wip = TRUE)
  result_no_wip <- gpdfit(x, wip = FALSE)

  # Both should give valid results
  expect_false(is.na(result_wip$k))
  expect_false(is.na(result_no_wip$k))

  # With WIP prior, k should be pulled toward 0.5
  expect_true(abs(0.5 - result_wip$k) < abs(0.5 - result_no_wip$k))
})

#' Pareto khat diagnostic
#'
#' Estimate Pareto k value by fitting a Generalized Pareto
#' Distribution to one or two tails of x. This can be used to estimate
#' the number of fractional moments that is useful for convergence
#' diagnostics. For further details see Vehtari et al. (2024).
#'
#' @family diagnostics
#' @template args-pareto
#' @template args-methods-dots
#' @template ref-vehtari-paretosmooth-2022
#' @template return-conv
#'
#' @seealso [`pareto_diags`] for additional related diagnostics, and
#'   [`pareto_smooth`] for Pareto smoothed draws.
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' pareto_khat(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' pareto_khat(d$Sigma)
#' @export
pareto_khat <- function(x, ...) UseMethod("pareto_khat")

#' @rdname pareto_khat
#' @export
pareto_khat.default <- function(x,
                                tail = c("both", "right", "left"),
                                r_eff = NULL,
                                ndraws_tail = NULL,
                                verbose = FALSE,
                                are_log_weights = FALSE,
                                ...) {
  smoothed <- pareto_smooth.default(
    x,
    tail = tail,
    r_eff = r_eff,
    ndraws_tail = ndraws_tail,
    verbose = verbose,
    return_k = TRUE,
    smooth_draws = FALSE,
    are_log_weights = are_log_weights
  )
  return(smoothed$diagnostics$khat)
}

#' @rdname pareto_khat
#' @export
pareto_khat.rvar <- function(x, ...) {
  draws_diags <- summarise_rvar_by_element_with_chains(
    x,
    pareto_khat.default,
    ...
  )
  dim(draws_diags) <- dim(draws_diags) %||% length(draws_diags)
  margins <- seq_along(dim(draws_diags))

  diags <- list(
    khat = apply(draws_diags, margins, function(x) x[[1]])
  )

  diags$khat
}


#' Pareto smoothing diagnostics
#'
#' Compute diagnostics for Pareto smoothing the tail draws of x by
#' replacing tail draws by order statistics of a generalized Pareto
#' distribution fit to the tail(s).
#'
#' @family diagnostics
#' @template args-pareto
#' @template args-methods-dots
#' @template ref-vehtari-paretosmooth-2022
#' @return List of Pareto smoothing diagnostics:
#'  * `khat`: estimated Pareto k shape parameter,
#'  * `min_ss`: minimum sample size for reliable Pareto smoothed estimate,
#'  * `khat_threshold`: khat-threshold for reliable Pareto smoothed estimate,
#'  * `convergence_rate`: Pareto smoothed estimate RMSE convergence rate.
#'
#' @details When the fitted Generalized Pareto Distribution is used to
#'   smooth the tail values and these smoothed values are used to
#'   compute expectations, the following diagnostics can give further
#'   information about the reliability of these estimates.
#'
#'  * `min_ss`: Minimum sample size for reliable Pareto smoothed
#' estimate. If the actual sample size is greater than `min_ss`, then
#' Pareto smoothed estimates can be considered reliable. If the actual
#' sample size is lower than `min_ss`, increasing the sample size
#' might result in more reliable estimates. For further details, see
#' Section 3.2.3, Equation 11 in Vehtari et al. (2024).
#'
#'  * `khat_threshold`: Threshold below which k-hat values result in
#' reliable Pareto smoothed estimates. The threshold is lower for
#' smaller effective sample sizes. If k-hat is larger than the
#' threshold, increasing the total sample size may improve reliability
#' of estimates. For further details, see Section 3.2.4, Equation 13
#' in Vehtari et al. (2024).
#'
#'  * `convergence_rate`: Relative convergence rate compared to the
#'  central limit theorem. Applicable only if the actual sample size
#'  is sufficiently large (greater than `min_ss`). The convergence
#'  rate tells the rate at which the variance of an estimate reduces
#'  when the sample size is increased, compared to the central limit
#'  theorem convergence rate. See Appendix B in Vehtari et al. (2024).
#'
#' @seealso [`pareto_khat`], [`pareto_min_ss`],
#'   [`pareto_khat_threshold`], and [`pareto_convergence_rate`] for
#'   individual diagnostics; and [`pareto_smooth`] for Pareto smoothing
#'   draws.
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' pareto_diags(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' pareto_diags(d$Sigma)
#' @export
pareto_diags <- function(x, ...) UseMethod("pareto_diags")


#' @rdname pareto_diags
#' @export
pareto_diags.default <- function(x,
                                 tail = c("both", "right", "left"),
                                 r_eff = NULL,
                                 ndraws_tail = NULL,
                                 verbose = FALSE,
                                 are_log_weights = FALSE,
                                 ...) {

  smoothed <- pareto_smooth.default(
    x,
    tail = tail,
    r_eff = r_eff,
    ndraws_tail = ndraws_tail,
    return_k = TRUE,
    extra_diags = TRUE,
    verbose = verbose,
    smooth_draws = FALSE,
    are_log_weights = are_log_weights,
    ...)

  return(smoothed$diagnostics)

}


#' @rdname pareto_diags
#' @export
pareto_diags.rvar <- function(x, ...) {
  draws_diags <- summarise_rvar_by_element_with_chains(
    x,
    pareto_diags.default,
    ...
  )

  dim(draws_diags) <- dim(draws_diags) %||% length(draws_diags)
  margins <- seq_along(dim(draws_diags))

  diags <- list(
    khat = apply(draws_diags, margins, function(x) x[[1]]$khat),
    min_ss = apply(draws_diags, margins, function(x) x[[1]]$min_ss),
    khat_threshold = apply(draws_diags, margins, function(x) x[[1]]$khat_threshold),
    convergence_rate = apply(draws_diags, margins, function(x) x[[1]]$convergence_rate)
  )

  diags
}



#' Pareto smoothing
#'
#' Smooth the tail draws of x by replacing tail draws by order
#' statistics of a generalized Pareto distribution fit to the
#' tail(s). For further details see Vehtari et al. (2024).
#'
#' @template args-pareto
#' @param return_k (logical) Should the Pareto khat be included in
#'   output? If `TRUE`, output will be a list containing smoothed
#'   draws and diagnostics, otherwise it will be a numeric of the
#'   smoothed draws. Default is `FALSE`.
#' @param extra_diags (logical) Should extra Pareto khat diagnostics
#'   be included in output? If `TRUE`, `min_ss`, `khat_threshold` and
#'   `convergence_rate` for the estimated k value will be
#'   returned. Default is `FALSE`.
#' @template args-methods-dots
#' @template ref-vehtari-paretosmooth-2022
#' @return Either a vector `x` of smoothed values or a named list
#'   containing the vector `x` and a named list `diagnostics`
#'   containing numeric values:
#'
#' * `khat`: estimated Pareto k shape parameter, and optionally
#' * `min_ss`: minimum sample size for reliable Pareto smoothed
#'   estimate
#' * `khat_threshold`: sample size specific khat threshold for
#'   reliable Pareto smoothed estimates
#' * `convergence_rate`: Relative convergence rate for
#'   Pareto smoothed estimates
#'
#' If any of the draws is non-finite, that is, `NA`, `NaN`, `Inf`, or
#' `-Inf`, Pareto smoothing will not be performed, and the original
#' draws will be returned and and diagnostics will be `NA` (numeric).
#'
#' @seealso [`pareto_khat`] for only calculating khat, and
#'   [`pareto_diags`] for additional diagnostics.
#' @examples
#' mu <- extract_variable_matrix(example_draws(), "mu")
#' pareto_smooth(mu)
#'
#' d <- as_draws_rvars(example_draws("multi_normal"))
#' pareto_smooth(d$Sigma)
#' @export
pareto_smooth <- function(x, ...) UseMethod("pareto_smooth")

#' @rdname pareto_smooth
#' @export
pareto_smooth.rvar <- function(x, return_k = FALSE, extra_diags = FALSE, ...) {

  if (extra_diags) {
    return_k <- TRUE
  }

  draws_diags <- summarise_rvar_by_element_with_chains(x, pareto_smooth.default, return_k = return_k, extra_diags = extra_diags, ...)
  dim(draws_diags) <- dim(draws_diags) %||% length(draws_diags)
  margins <- seq_along(dim(draws_diags))

  if (return_k) {
    if (extra_diags) {

      diags <- list(
        khat = apply(draws_diags, margins, function(x) x[[1]]$diagnostics$khat),
        min_ss = apply(draws_diags, margins, function(x) x[[1]]$diagnostics$min_ss),
        khat_threshold = apply(draws_diags, margins, function(x) x[[1]]$diagnostics$khat_threshold),
        convergence_rate = apply(draws_diags, margins, function(x) x[[1]]$diagnostics$convergence_rate)
      )
    } else {
      diags <- list(
        khat = apply(draws_diags, margins, function(x) x[[1]]$diagnostics$khat)
      )
    }
    out <- list(
      x = rvar(apply(draws_diags, margins, function(x) x[[1]]$x), nchains = nchains(x)),
      diagnostics = diags
    )
  } else {
    out <- rvar(apply(draws_diags, margins, function(x) x[[1]]), nchains = nchains(x))
  }
  out
}

#' @rdname pareto_smooth
#' @export
pareto_smooth.default <- function(x,
                                  tail = c("both", "right", "left"),
                                  r_eff = NULL,
                                  ndraws_tail = NULL,
                                  return_k = FALSE,
                                  extra_diags = FALSE,
                                  verbose = TRUE,
                                  are_log_weights = FALSE,
                                  ...) {

  if (!is.null(r_eff)) {
    r_eff <- as_one_numeric(r_eff)
  }
  if (!is.null(ndraws_tail)) {
    ndraws_tail <- as_one_integer(ndraws_tail)
  }
  extra_diags <- as_one_logical(extra_diags)
  return_k <- as_one_logical(return_k)
  verbose <- as_one_logical(verbose)
  are_log_weights <- as_one_logical(are_log_weights)

  if (extra_diags) {
    return_k <- TRUE
  }

  # check for infinite or na values
  if (should_return_NA(x)) {
    warning_no_call("Input contains infinite or NA values, is constant or has constant tail. Fitting of generalized Pareto distribution not performed.")
    return(pareto_diags_na(x, return_k, extra_diags))
  }

  if (are_log_weights) {
    tail <- "right"
  }

  tail <- match.arg(tail)
  ndraws <- length(x)

  # automatically calculate relative efficiency
  if (is.null(r_eff)) {
    r_eff <- ess_tail(x) / ndraws
  }

  # automatically calculate tail length
  if (is.null(ndraws_tail)) {
    ndraws_tail <- ps_tail_length(ndraws, r_eff)
  }

  if (is.na(ndraws_tail)) {
    warning_no_call("Input contains infinite or NA values, is constant, or has constant tail. Fitting of generalized Pareto distribution not performed.")
    return(pareto_diags_na(x, return_k, extra_diags))
  }

  if (tail == "both") {

    if (ndraws_tail > ndraws / 2) {
      warning("Number of tail draws cannot be more than half ",
              "the total number of draws if both tails are fit, ",
              "changing to ", ndraws / 2, ".")
      ndraws_tail <- ndraws / 2
    }

    if (ndraws_tail < 5) {
      warning("Number of tail draws cannot be less than 5. ",
              "Changing to ", 5, ".")
      ndraws_tail <- 5
    }

    # left tail
    smoothed <- .pareto_smooth_tail(
      x,
      ndraws_tail = ndraws_tail,
      tail = "left",
      are_log_weights = are_log_weights,
      ...
    )
    left_k <- smoothed$k

    # right tail
    smoothed <-.pareto_smooth_tail(
      x = smoothed$x,
      ndraws_tail = ndraws_tail,
      tail = "right",
      are_log_weights = are_log_weights,
      ...
    )
    right_k <- smoothed$k

    k <- max(left_k, right_k, na.rm = TRUE)
    x <- smoothed$x

  } else {

    smoothed <- .pareto_smooth_tail(
      x,
      ndraws_tail = ndraws_tail,
      tail = tail,
      are_log_weights = are_log_weights,
      ...
    )
    k <- smoothed$k
    x <- smoothed$x
  }

  diags_list <- list(khat = k)

  if (extra_diags) {
    ext_diags <- .pareto_smooth_extra_diags(k, ndraws)
    diags_list <- c(diags_list, ext_diags)
  }

  if (verbose) {
    if (!extra_diags) {
      diags_list <- c(diags_list, .pareto_smooth_extra_diags(diags_list$khat, length(x)))
    }
    pareto_k_diagmsg(
      diags = diags_list,
      are_weights = are_log_weights
    )
  }

  if (return_k) {
    out <- list(x = x, diagnostics = diags_list)
  } else {
    out <- x
  }

  return(out)
}

#' @rdname pareto_diags
#' @export
pareto_khat_threshold <- function(x, ...) {
  UseMethod("pareto_khat_threshold")
}

#' @rdname pareto_diags
#' @export
pareto_khat_threshold.default <- function(x, ...) {
  ps_khat_threshold(length(x))
}

#' @rdname pareto_diags
#' @export
pareto_khat_threshold.rvar <- function(x, ...) {
  ps_khat_threshold(ndraws(x))
}

#' @rdname pareto_diags
#' @export
pareto_min_ss <- function(x, ...) {
  UseMethod("pareto_min_ss")
}

#' @rdname pareto_diags
#' @export
pareto_min_ss.default <- function(x, ...) {
  k <- pareto_khat(x)
  ps_min_ss(k)
}

#' @rdname pareto_diags
#' @export
pareto_min_ss.rvar <- function(x, ...) {
  k <- pareto_khat(x)
  ps_min_ss(k)
}

#' @rdname pareto_diags
#' @export
pareto_convergence_rate <- function(x, ...) {
  UseMethod("pareto_convergence_rate")
}

#' @rdname pareto_diags
#' @export
pareto_convergence_rate.default <- function(x, ...) {
  k <- pareto_khat(x)
  ps_convergence_rate(k, length(x))
}

#' @rdname pareto_diags
#' @export
pareto_convergence_rate.rvar <- function(x, ...) {
  k <- pareto_khat(x)
  ps_convergence_rate(k, ndraws(x))
}


#' Pareto smooth tail
#' internal function to pareto smooth the tail of a vector
#' @noRd
.pareto_smooth_tail <- function(x,
                                ndraws_tail,
                                smooth_draws = TRUE,
                                tail = c("right", "left"),
                                are_log_weights = FALSE,
                                ...
                                ) {

  if (are_log_weights) {
    # shift log values for safe exponentiation
    x <- x - max(x)
  }

  tail <- match.arg(tail)

  ndraws <- length(x)
  tail_ids <- seq(ndraws - ndraws_tail + 1, ndraws)

  if (tail == "left") {
    x <- -x
  }

  ord <- sort.int(x, index.return = TRUE)
  draws_tail <- ord$x[tail_ids]

  if (is_constant(draws_tail)) {

    if (tail == "left") {
      x <- -x
    }

    out <- list(x = x, k = NA)
    return(out)
  }

  cutoff <- ord$x[min(tail_ids) - 1] # largest value smaller than tail values

  max_tail <- max(draws_tail)
  min_tail <- min(draws_tail)

  if (ndraws_tail >= 5) {
    ord <- sort.int(x, index.return = TRUE)
    if (abs(max_tail - min_tail) < .Machine$double.eps / 100) {
      warning_no_call(
        "Can't fit generalized Pareto distribution ",
        "because all tail values are the same."
      )
      smoothed <- NULL
      k <- NA
    } else {
      # save time not sorting since x already sorted
      if (are_log_weights) {
        draws_tail <- exp(draws_tail)
        cutoff <- exp(cutoff)
      }
      fit <- gpdfit(draws_tail - cutoff, sort_x = FALSE, ...)
      k <- fit$k
      sigma <- fit$sigma
      if (is.finite(k) && smooth_draws) {
        p <- (seq_len(ndraws_tail) - 0.5) / ndraws_tail
        smoothed <- qgeneralized_pareto(p = p, mu = cutoff, k = k, sigma = sigma)
        if (are_log_weights) {
          smoothed <- log(smoothed)
        }
      } else {
        smoothed <- NULL
      }
    }
  } else {
    warning_no_call(
      "Can't fit generalized Pareto distribution ",
      "because ndraws_tail is less than 5."
    )
    smoothed <- NULL
    k <- NA
  }

  # truncate at max of raw draws
  if (!is.null(smoothed)) {
    smoothed[smoothed > max_tail] <- max_tail
    x[ord$ix[tail_ids]] <- smoothed
  }

  if (tail == "left") {
    x <- -x
  }

  out <- list(x = x, k = k)

  return(out)
}

#' Extra pareto-k diagnostics
#'
#' internal function to calculate the extra diagnostics for a given
#' pareto k and number of draws ndraws
#' @noRd
.pareto_smooth_extra_diags <- function(k, ndraws, ...) {

  min_ss <- ps_min_ss(k)

  khat_threshold <- ps_khat_threshold(ndraws)

  convergence_rate <- ps_convergence_rate(k, ndraws)

  other_diags <- list(
    min_ss = min_ss,
    khat_threshold = khat_threshold,
    convergence_rate = convergence_rate
  )
}

#' Pareto-smoothing minimum sample-size
#'
#' Given Pareto-k computes the minimum sample size for reliable Pareto
#' smoothed estimate (to have small probability of large error). See
#' section 3.2.3 in Vehtari et al. (2024). This function is exported
#' to be usable by other packages. For user-facing diagnostic functions, see
#' [`pareto_min_ss`] and [`pareto_diags`].
#' @family helper-functions
#' @param k pareto k value
#' @param ... unused
#' @return minimum sample size
#' @export
ps_min_ss <- function(k, ...) {
  if (k < 1) {
    out <- 10^(1 / (1 - max(0, k)))
  } else {
    out <- Inf
  }
  out
}

#' Pareto k-hat threshold
#'
#' Given number of draws, computes khat threshold for reliable Pareto
#' smoothed estimate (to have small probability of large error). See
#' section 3.2.4, equation (13) of Vehtari et al. (2024). This
#' function is exported to be usable by other packages. For
#' user-facing diagnostic functions, see [`pareto_khat_threshold`] and
#' [`pareto_diags`].
#' @family helper-functions
#' @param ndraws number of draws
#' @param ... unused
#' @return threshold
#' @export
ps_khat_threshold <- function(ndraws, ...) {
  1 - 1 / log10(ndraws)
}

#' Pareto convergence rate
#'
#' Given number of draws and scalar or array of k's, compute the
#' relative convergence rate of PSIS estimate RMSE. See Appendix B of
#' Vehtari et al. (2024). This function is exported to be usable by
#' other packages. For user-facing diagnostic functions, see
#' [`pareto_convergence_rate`] and [`pareto_diags`].
#' @family helper-functions
#' @param k pareto-k values
#' @param ndraws number of draws
#' @param ... unused
#' @return convergence rate
#' @export
ps_convergence_rate <- function(k, ndraws, ...) {
  # allow array of k's
  rate <- numeric(length(k))
  # k<0 bounded distribution
  rate[k < 0] <- 1
  # k>0 non-finite mean
  rate[k > 1] <- 0
  # limit value at k=1/2
  rate[k == 0.5] <- 1 - 1 / log(ndraws)
  # smooth approximation for the rest (see Appendix B of PSIS paper)
  ki <- (k > 0 & k < 1 & k != 0.5)
  kk <- k[ki]
  rate[ki] <- pmax(
    0,
    (2 * (kk - 1) * ndraws^(2 * kk + 1) + (1 - 2 * kk) * ndraws^(2 * kk) + ndraws^2) /
      ((ndraws - 1) * (ndraws - ndraws^(2 * kk)))
  )
  rate
}

#' Pareto tail length
#'
#' Calculate the tail length from number of draws and relative efficiency
#' r_eff. See Appendix H in Vehtari et al. (2024). This function is
#' used internally and is exported to be available for other packages.
#' @family helper-functions
#' @param ndraws number of draws
#' @param r_eff relative efficiency
#' @param ... unused
#' @return tail length
#' @export
ps_tail_length <- function(ndraws, r_eff, ...) {
  ceiling(ifelse(ndraws > 225, 3 * sqrt(ndraws / r_eff), ndraws / 5))
}

#' Pareto-k diagnostic message
#'
#' Given number of draws and k, form a diagnostic message string.
#' @param diags (numeric) named vector of diagnostic values
#' @param are_weights (logical) are the diagnostics for weights
#' @param ... unused
#' @return diagnostic message
#' @noRd
pareto_k_diagmsg <- function(diags, are_weights = FALSE, ...) {
  khat <- diags$khat
  min_ss <- diags$min_ss
  khat_threshold <- diags$khat_threshold
  convergence_rate <- diags$convergence_rate
  msg <- NULL

  if (!are_weights) {

    if (khat > 1) {
      msg <- paste0(msg, " Mean does not exist, making empirical mean estimate of the draws not applicable.")
    } else {
      if (khat > khat_threshold) {
        msg <- paste0(msg, " Sample size is too small, for given Pareto k-hat. Sample size larger than ", round(min_ss, 0), " is needed for reliable results.\n")
      }
      if (khat > 0.7) {
        msg <- paste0(msg, " Bias dominates when k-hat > 0.7, making empirical mean estimate of the Pareto-smoothed draws unreliable.\n")
      }
    }
  } else {
    if (khat > khat_threshold || khat > 0.7) {
        msg <- paste0(msg, " Pareto khat for weights is high (", round(khat, 1) ,"). This indicates a single or few weights dominate.\n", "Inference based on weighted draws will be unreliable.\n")
    }
  }
  message("Pareto k-hat = ", round(khat, 2), ".", msg)
  invisible(diags)
}


pareto_diags_na <- function(x, return_k, extra_diags) {
    if (!return_k) {
      out <- x
    } else if (!extra_diags) {
      out <- list(x = x, diagnostics = list(khat = NA_real_))
    } else {
      out <- list(
        x = x,
        diagnostics = list(
          khat = NA_real_,
          min_ss = NA_real_,
          khat_threshold = NA_real_,
          convergence_rate = NA_real_
        )
      )
    }
    return(out)
}

test_that("pareto_khat handles constant tail correctly", {

  # left tail is constant, so khat should be NA, but for "both" it
  # should be the same as the right tail
  x <- c(rep(-100, 10), sort(rnorm(100)))

  expect_true(is.na(pareto_khat(x, tail = "left", ndraws_tail = 10)))
  expect_equal(
    pareto_khat(x, tail = "right", ndraws_tail = 10),
    pareto_khat(x, tail = "both", ndraws_tail = 10)
  )

})


test_that("pareto_khat handles tail argument", {

  # as tau is bounded (0, Inf) the left pareto k should be lower than
  # right
  tau <- extract_variable_matrix(example_draws(), "tau")
  pkl <- pareto_khat(tau, tail = "left")
  pkr <- pareto_khat(tau, tail = "right")
  pkb <- pareto_khat(tau)
  expect_true(pkl < pkr)
  expect_equal(pkr, pkb)
})

test_that("pareto_khat handles ndraws_tail argument", {

  tau <- extract_variable_matrix(example_draws(), "tau")
  pk10 <- pareto_khat(tau, tail = "right", ndraws_tail = 10)
  pk25 <- pareto_khat(tau, tail = "right", ndraws_tail = 25)
  expect_true(pk10 > pk25)

  expect_warning(pareto_khat(tau, tail = "both", ndraws_tail = 201),
                 paste0("Number of tail draws cannot be more than half ",
                 "the total number of draws if both tails are fit, ",
                 "changing to 200."))

  expect_warning(pareto_khat(tau, tail = "both", ndraws_tail = 4),
                 "Number of tail draws cannot be less than 5. Changing to 5.")

})


test_that("pareto_khat handles r_eff argument", {

  tau <- extract_variable_matrix(example_draws(), "tau")
  pk1 <- pareto_khat(tau, r_eff = 1)
  pk0.6 <- pareto_khat(tau, r_eff = 0.6)
  expect_true(pk1 < pk0.6)

})


test_that("pareto_khat diagnostics messages are as expected", {

  diags <- list(
    khat = 0.5,
    min_ss = 10,
    khat_threshold = 0.55,
    convergence_rate = 0.99
  )

  expect_message(pareto_k_diagmsg(diags),
                   paste0("Pareto k-hat = 0.5.\n"))

  diags$khat <- 0.6

    expect_message(pareto_k_diagmsg(diags),
                   paste0("Pareto k-hat = 0.6. Sample size is too small, for given Pareto k-hat. Sample size larger than 10 is needed for reliable results.\n"))

    diags$khat <- 0.71
    diags$khat_threshold <- 0.8

    expect_message(pareto_k_diagmsg(diags),
                   paste0("Pareto k-hat = 0.71. Bias dominates when k-hat > 0.7, making empirical mean estimate of the Pareto-smoothed draws unreliable.\n"))


    diags$khat <- 1.1

    expect_message(pareto_k_diagmsg(diags),
                   paste0("Pareto k-hat = 1.1. Mean does not exist, making empirical mean estimate of the draws not applicable.\n"))

})

test_that("pareto_diags returns expected diagnostics", {

  tau <- extract_variable_matrix(example_draws(), "tau")

  pk <- pareto_diags(tau)
  expect_true(all(names(pk) == c("khat", "min_ss", "khat_threshold", "convergence_rate")))
})


test_that("pareto_khat diagnostics handles verbose argument", {
  tau <- extract_variable_matrix(example_draws(), "tau")

  expect_message(pareto_khat(tau, extra_diags = TRUE, verbose = TRUE))
  expect_message(pareto_khat(tau, extra_diags = FALSE, verbose = TRUE))
  expect_no_message(pareto_khat(tau, extra_diags = TRUE, verbose = FALSE))

})

test_that("pareto_khat diagnostics accept vectors as input", {
  set.seed(1234)
  x <- rnorm(1000)

  pk <- pareto_khat(x)
  expect_true(pk < 1)
})

test_that("pareto_khat diagnostics handle special cases correctly", {
  set.seed(1234)

  x <- c(rnorm(50), NA)
  expect_warning(pareto_khat(x))
  expect_true(is.na(suppressWarnings(pareto_khat(x))))

  x <- c(rnorm(50), Inf)
  expect_warning(pareto_khat(x))
  expect_true(is.na(suppressWarnings(pareto_khat(x))))

  x <- rep(1, 50)
  expect_warning(pareto_khat(x))
  expect_true(is.na(suppressWarnings(pareto_khat(x))))

})


test_that("pareto_khat functions work with matrix with chains", {

  tau_chains <- extract_variable_matrix(example_draws(), "tau")
  tau_nochains <- extract_variable(example_draws(), "tau")

  expect_equal(pareto_khat(tau_chains, ndraws_tail = 20),
               pareto_khat(tau_nochains, ndraws_tail = 20))

  ps_chains <- pareto_smooth(tau_chains, ndraws_tail = 20, return_k = TRUE)
  ps_nochains <- pareto_smooth(tau_nochains, ndraws_tail = 20, return_k = TRUE)

  expect_equal(as.numeric(ps_chains$x), as.numeric(ps_nochains$x))

})

test_that("pareto_khat functions work with rvars with and without chains", {

  tau_chains <- extract_variable_matrix(example_draws(), "tau")
  tau_nochains <- extract_variable(example_draws(), "tau")

  tau_rvar_chains <- rvar(tau_chains, with_chains = TRUE)
  tau_rvar_nochains <- rvar(tau_nochains)

  expect_equal(pareto_khat(tau_chains, ndraws_tail = 20),
               pareto_khat(tau_rvar_chains, ndraws_tail = 20))

  expect_equal(pareto_khat(tau_rvar_chains, ndraws_tail = 20),
               pareto_khat(tau_rvar_nochains, ndraws_tail = 20))


  expect_equal(pareto_diags(tau_chains, ndraws_tail = 20),
               pareto_diags(tau_rvar_chains, ndraws_tail = 20))

  expect_equal(pareto_diags(tau_rvar_chains, ndraws_tail = 20),
               pareto_diags(tau_rvar_nochains, ndraws_tail = 20))

  ps_chains <- pareto_smooth(tau_chains, ndraws_tail = 20, return_k = TRUE)
  ps_rvar_chains <- pareto_smooth(tau_rvar_chains, ndraws_tail = 20, return_k = TRUE)

  ps_nochains <- pareto_smooth(tau_nochains, ndraws_tail = 20, return_k = TRUE)
  ps_rvar_nochains <- pareto_smooth(tau_rvar_nochains, ndraws_tail = 20, return_k = TRUE)

  expect_equal(ps_rvar_chains$x, rvar(ps_chains$x, with_chains = TRUE))

  expect_equal(ps_rvar_nochains$x, rvar(ps_nochains$x))


  ps_chains <- pareto_smooth(tau_chains, ndraws_tail = 20, extra_diags = TRUE, return_k = TRUE)
  ps_rvar_chains <- pareto_smooth(tau_rvar_chains, ndraws_tail = 20, extra_diags = TRUE, return_k = TRUE)

  ps_nochains <- pareto_smooth(tau_nochains, ndraws_tail = 20, extra_diags = TRUE, return_k = TRUE)
  ps_rvar_nochains <- pareto_smooth(tau_rvar_nochains, ndraws_tail = 20, extra_diags = TRUE, return_k = TRUE)

  expect_equal(ps_rvar_chains$x, rvar(ps_chains$x, with_chains = TRUE))

  expect_equal(ps_rvar_nochains$x, rvar(ps_nochains$x))

})

test_that("pareto_smooth returns x with smoothed tail(s)", {
  mu <- extract_variable_matrix(example_draws(), "mu")

  mu_smoothed_right <- pareto_smooth(mu, ndraws_tail = 10, tail = "right", return_k = TRUE)$x

  mu_smoothed_left <- pareto_smooth(mu, ndraws_tail = 10, tail = "left", return_k = TRUE)$x

  mu_smoothed_both <- pareto_smooth(mu, ndraws_tail = 10, tail = "both", return_k = TRUE)$x

  expect_equal(sort(mu)[1:390], sort(mu_smoothed_right)[1:390])
  expect_equal(sort(mu_smoothed_both)[11:400], sort(mu_smoothed_right)[11:400])

  expect_equal(sort(mu)[11:400], sort(mu_smoothed_left)[11:400])
  expect_equal(sort(mu_smoothed_both)[1:390], sort(mu_smoothed_left)[1:390])

  expect_false(isTRUE(all.equal(sort(mu), sort(mu_smoothed_left))))
  expect_false(isTRUE(all.equal(sort(mu), sort(mu_smoothed_right))))
  expect_false(isTRUE(all.equal(sort(mu), sort(mu_smoothed_both))))

  expect_false(isTRUE(all.equal(sort(mu_smoothed_both), sort(mu_smoothed_left))))
  expect_false(isTRUE(all.equal(sort(mu_smoothed_both), sort(mu_smoothed_right))))

})

test_that("pareto_smooth works for log_weights", {
  w <- c(1:25, 1e3, 1e3, 1e3)
  lw <- log(w)

  ps <- pareto_smooth(lw, are_log_weights = TRUE, verbose = FALSE, ndraws_tail = 10, return_k = TRUE)

  # only right tail is smoothed
  expect_equal(ps$x[1:15] + max(lw), lw[1:15])

  expect_true(ps$diagnostics$khat > 0.7)

})

#' Index `draws` objects
#'
#' Index iterations, chains, and draws of [`draws`] objects.
#'
#' @name draws-index
#' @template args-methods-x
#'
#' @details
#' The methods `iteration_ids()`, `chain_ids()`, and `draw_ids()` return
#' vectors of all iterations, chains, and draws, respectively. In
#' contrast, the methods `niterations()`, `nchains()`, and
#' `ndraws()` return the number of variables, iterations, chains, and draws,
#' respectively.
#'
#' @return
#'
#' For `iteration_ids()`, `chain_ids()`, and `draw_ids()`, an integer vector.
#'
#' For `niterations()`, `nchains()`, and `ndraws()`, a scalar integer.
#'
#' @seealso [`variables`], [`rename_variables`]
#'
#' @examples
#' x <- example_draws()
#'
#' iteration_ids(x)
#' niterations(x)
#'
#' chain_ids(x)
#' nchains(x)
#'
#' draw_ids(x)
#' ndraws(x)
#'
NULL

#' @rdname draws-index
#' @export
iteration_ids <- function(x) {
  UseMethod("iteration_ids")
}

#' @export
iteration_ids.NULL <- function(x) {
  NULL
}

#' @export
iteration_ids.draws_matrix <- function(x) {
  if (nchains(x) > 1) {
    out <- seq_len(niterations(x))
  } else {
    out <- draw_ids(x)
  }
  out
}

#' @export
iteration_ids.draws_array <- function(x) {
  out <- rownames(x) %||% seq_rows(x)
  as.integer(out)
}

#' @export
iteration_ids.draws_df <- function(x) {
  as.integer(unique(x$.iteration))
}

#' @export
iteration_ids.draws_list <- function(x) {
  if (!length(x) || !length(x[[1]])) {
    # no chains or no variables within chains
    return(integer(0))
  }
  seq_along(x[[1]][[1]])
}

#' @export
iteration_ids.draws_rvars <- function(x) {
  iteration_ids(x[[1]])
}

#' @export
iteration_ids.rvar <- function(x) {
  if (nchains(x) > 1) {
    out <- seq_len(niterations(x))
  } else {
    out <- draw_ids(x)
  }
  out
}

#' @rdname draws-index
#' @export
chain_ids <- function(x) {
  UseMethod("chain_ids")
}

#' @export
chain_ids.NULL <- function(x) {
  NULL
}

#' @export
chain_ids.draws_matrix <- function(x) {
  seq_len(nchains(x))
}

#' @export
chain_ids.draws_array <- function(x) {
  out <- colnames(x) %||% seq_cols(x)
  as.integer(out)
}

#' @export
chain_ids.draws_df <- function(x) {
  as.integer(unique(x$.chain))
}

#' @export
chain_ids.draws_list <- function(x) {
  out <- names(x) %||% seq_rows(x)
  as.integer(out)
}

#' @export
chain_ids.draws_rvars <- function(x) {
  chain_ids(x[[1]])
}

#' @export
chain_ids.draws_rvars <- function(x) {
  seq_len(nchains(x))
}

#' @rdname draws-index
#' @export
draw_ids <- function(x) {
  UseMethod("draw_ids")
}

#' @export
draw_ids.NULL <- function(x) {
  NULL
}

#' @export
draw_ids.draws_matrix <- function(x) {
  as.integer(rownames(x) %||% seq_rows(x))
}

#' @export
draw_ids.draws_array <- function(x) {
  iteration_ids <- iteration_ids(x)
  niterations <- niterations(x)
  chain_ids <- chain_ids(x)
  ulapply(chain_ids, function(c) niterations * (c - 1L) + iteration_ids)
}

#' @export
draw_ids.draws_df <- function(x) {
  as.integer(unique(x$.draw))
}

#' @export
draw_ids.draws_list <- function(x) {
  iteration_ids <- iteration_ids(x)
  niterations <- niterations(x)
  chain_ids <- chain_ids(x)
  ulapply(chain_ids, function(c) niterations * (c - 1L) + iteration_ids)
}

#' @export
draw_ids.draws_rvars <- function(x) {
  draw_ids(x[[1]])
}

#' @export
draw_ids.rvar <- function(x) {
  draws <- draws_of(x)
  out <- rownames(draws) %||% seq_rows(draws)
  as.integer(out)
}

#' @rdname draws-index
#' @export
niterations <- function(x) {
  UseMethod("niterations")
}

#' @export
niterations.NULL <- function(x) {
  0
}

#' @export
niterations.draws_matrix <- function(x) {
  NROW(x) / nchains(x)
}

#' @export
niterations.draws_array <- function(x) {
  NROW(x)
}

#' @export
niterations.draws_df <- function(x) {
  length(iteration_ids(x))
}

#' @export
niterations.draws_list <- function(x) {
  if (!length(x) || !length(x[[1]])) {
    return(0)
  }
  length(x[[1]][[1]])
}

#' @export
niterations.draws_rvars <- function(x) {
  if (!length(x)) 0 else niterations(x[[1]])
}

#' @export
niterations.rvar <- function(x) {
  ndraws(x) / nchains(x)
}

#' @rdname draws-index
#' @export
nchains <- function(x) {
  UseMethod("nchains")
}

#' @export
nchains.NULL <- function(x) {
  0
}

#' @export
nchains.draws_matrix <- function(x) {
  attr(x, "nchains") %||% 1L
}

#' @export
nchains.draws_array <- function(x) {
  NCOL(x)
}

#' @export
nchains.draws_df <- function(x) {
  length(chain_ids(x))
}

#' @export
nchains.draws_list <- function(x) {
  length(x)
}

#' @export
nchains.draws_rvars <- function(x) {
  if (!length(x)) 0 else nchains(x[[1]])
}

#' @export
nchains.rvar <- function(x) {
  attr(x, "nchains") %||% 1L
}
# for internal use only currently: if you are setting the nchains
# attribute on an rvar, ALWAYS use this function so that the proxy
# cache is invalidated
`nchains_rvar<-` <- function(x, value) {
  attr(x, "nchains") <- value
  invalidate_rvar_cache(x)
}


#' @rdname draws-index
#' @export
ndraws <- function(x) {
  UseMethod("ndraws")
}

#' @export
ndraws.NULL <- function(x) {
  0
}

#' @export
ndraws.draws_matrix <- function(x) {
  NROW(x)
}

#' @export
ndraws.draws_array <- function(x) {
  niterations(x) * nchains(x)
}

#' @export
ndraws.draws_df <- function(x) {
  NROW(x)
}

#' @export
ndraws.draws_list <- function(x) {
  niterations(x) * nchains(x)
}

#' @export
ndraws.draws_rvars <- function(x) {
  if (!length(x)) 0 else ndraws(x[[1]])
}

#' @export
ndraws.rvar <- function(x) {
  # as.vector() to drop names in case there are index names
  as.vector(NROW(draws_of(x)))
}


# internal ----------------------------------------------------------------

# check validity of iteration indices
# @param unique should the returned IDs be unique?
check_iteration_ids <- function(iteration_ids, x, unique = TRUE, exclude = FALSE) {
  check_draws_object(x)
  if (is.null(iteration_ids)) {
    return(NULL)
  }
  unique <- as_one_logical(unique)
  exclude <- as_one_logical(exclude)
  iteration_ids <- as.integer(iteration_ids)
  if (unique) {
    iteration_ids <- unique(iteration_ids)
  }
  iteration_ids <- sort(iteration_ids)
  if (any(iteration_ids < 1L)) {
    stop_no_call("Iteration indices should be positive.")
  }
  niterations <- niterations(x)
  max_iteration <- SW(max(iteration_ids))
  if (max_iteration > niterations) {
    stop_no_call("Tried to subset iterations up to '", max_iteration, "' ",
          "but the object only has '", niterations, "' iterations.")
  }

  # handle exclude iterations in subset_draws
  if (exclude) {
    iteration_ids <- setdiff(iteration_ids(x), iteration_ids)
  }

  invisible(iteration_ids)
}

# check validity of chain indices
# @param unique should the returned IDs be unique?
check_chain_ids <- function(chain_ids, x, unique = TRUE, exclude = FALSE) {
  check_draws_object(x)
  if (is.null(chain_ids)) {
    return(NULL)
  }
  unique <- as_one_logical(unique)
  exclude <- as_one_logical(exclude)
  chain_ids <- as.integer(chain_ids)
  if (unique) {
    chain_ids <- unique(chain_ids)
  }
  chain_ids <- sort(chain_ids)
  if (any(chain_ids < 1L)) {
    stop_no_call("Chain indices should be positive.")
  }
  nchains <- nchains(x)
  max_chain <- SW(max(chain_ids))
  if (max_chain > nchains) {
    stop_no_call("Tried to subset chains up to '", max_chain, "' ",
          "but the object only has '", nchains, "' chains.")
  }

  if (exclude) {
    chain_ids <- setdiff(chain_ids(x), chain_ids)
  }

  invisible(chain_ids)
}

# check validity of draw indices
# @param unique should the returned IDs be unique?
check_draw_ids <- function(draw_ids, x, unique = TRUE, exclude = FALSE) {
  check_draws_object(x)
  if (is.null(draw_ids)) {
    return(NULL)
  }
  unique <- as_one_logical(unique)
  exclude <- as_one_logical(exclude)
  draw_ids <- as.integer(draw_ids)
  if (unique) {
    draw_ids <- unique(draw_ids)
  }
  draw_ids <- sort(draw_ids)
  if (any(draw_ids < 1L)) {
    stop_no_call("Draw indices should be positive.")
  }
  ndraws <- ndraws(x)
  max_draw <- SW(max(draw_ids))
  if (max_draw > ndraws) {
    stop_no_call("Tried to subset draws up to '", max_draw, "' ",
          "but the object only has '", ndraws, "' draws.")
  }

  if (exclude) {
    draw_ids <- setdiff(draw_ids(x), draw_ids)
  }

  invisible(draw_ids)
}

test_that("bind_draws works for draws_matrix objects", {
  draws1 <- as_draws_matrix(example_draws())
  draws2 <- subset_draws(draws1, chain = 2)
  draws3 <- subset_draws(draws1, iteration = 10:20)
  draws4 <- as_draws_df(data.frame(
    nu = rnorm(ndraws(draws1)),
    .chain = rep(chain_ids(draws1), each = niterations(draws1))
  ))

  draws_new <- bind_draws(draws1, draws4, along = "variable")
  expect_equal(
    variables(draws_new),
    c(variables(draws1), variables(draws4))
  )
  expect_equal(ndraws(draws_new), ndraws(draws1))

  expect_error(bind_draws(draws1, draws3, along = "iteration"),
               "Cannot bind 'draws_matrix' objects along 'iteration'")

  draws_new <- bind_draws(draws1, draws2, along = "chain")
  expect_equal(
    nchains(draws_new),
    nchains(draws1) + nchains(draws2)
  )
  expect_equal(variables(draws_new), variables(draws1))

  draws_new <- bind_draws(draws1, draws3, along = "draw")
  expect_equal(
    ndraws(draws_new),
    ndraws(draws1) + ndraws(draws3)
  )
  expect_equal(variables(draws_new), variables(draws1))
})

test_that("bind_draws works for draws_array objects", {
  draws1 <- as_draws_array(example_draws())
  draws2 <- subset_draws(draws1, chain = 2)
  draws3 <- subset_draws(draws1, iteration = 10:20)
  draws4 <- as_draws_df(data.frame(
    nu = rnorm(ndraws(draws1)),
    .chain = rep(chain_ids(draws1), each = niterations(draws1))
  ))

  draws_new <- bind_draws(draws1, draws4, along = "variable")
  expect_equal(
    variables(draws_new),
    c(variables(draws1), variables(draws4))
  )
  expect_equal(ndraws(draws_new), ndraws(draws1))

  draws_new <- bind_draws(draws1, draws2, along = "chain")
  expect_equal(nchains(draws_new), nchains(draws1) + nchains(draws2))
  expect_equal(variables(draws_new), variables(draws1))

  draws_new <- bind_draws(draws1, draws3, along = "iteration")
  expect_equal(
    niterations(draws_new),
    niterations(draws1) + niterations(draws3)
  )
  expect_equal(variables(draws_new), variables(draws1))

  expect_error(bind_draws(draws1, draws3, along = "draw"),
               "Cannot bind 'draws_array' objects along 'draw'")
})

test_that("bind_draws works for draws_df objects", {
  draws1 <- as_draws_df(example_draws())
  draws2 <- subset_draws(draws1, chain = 2)
  draws3 <- subset_draws(draws1, iteration = 10:20)
  draws4 <- as_draws_list(data.frame(
    nu = rnorm(ndraws(draws1)),
    .chain = rep(chain_ids(draws1), each = niterations(draws1))
  ))

  draws_new <- bind_draws(draws1, draws4, along = "variable")
  expect_equal(
    variables(draws_new),
    c(variables(draws1), variables(draws4))
  )
  expect_equal(ndraws(draws_new), ndraws(draws1))

  draws_new <- bind_draws(draws1, draws2, along = "chain")
  expect_equal(nchains(draws_new), nchains(draws1) + nchains(draws2))
  expect_equal(variables(draws_new), variables(draws1))

  draws_new <- bind_draws(draws1, draws3, along = "iteration")
  expect_equal(
    niterations(draws_new),
    niterations(draws1) + niterations(draws3)
  )
  expect_equal(variables(draws_new), variables(draws1))

  draws_new <- bind_draws(draws1, draws3, along = "draw")
  expect_equal(ndraws(draws_new), ndraws(draws1) + ndraws(draws3))
  expect_equal(nchains(draws_new), 1L)

  draws_new <- bind_draws(NULL, draws1)
  expect_equal(draws_new, draws1)
})


test_that("bind_draws works for multiple draws_df objects", {

  draws1 <- as_draws_df(example_draws())
  draws2 <- subset_draws(draws1, chain = 2)
  draws3 <-  subset_draws(draws1, chain = 3)

  draws12 <- bind_draws(draws1, draws2, along = "chain")
  draws123 <- bind_draws(draws12, draws3, along = "chain")
  draws_all <- bind_draws(draws1, draws2, draws3, along = "chain")
  expect_equal(draws123, draws_all)
  expect_equal(nchains(draws_all),
    nchains(draws1) + nchains(draws2) + nchains(draws3))
  expect_equal(ndraws(draws_all),
    ndraws(draws1) + ndraws(draws2) + ndraws(draws3))
  expect_equal(niterations(draws_all), niterations(draws1))
  expect_equal(variables(draws_all), variables(draws1))

  draws4 <-  subset_draws(draws1, chain = 4)
  draws23 <- bind_draws(draws2, draws3, along = "iteration")
  draws234 <- bind_draws(draws23, draws4, along = "iteration")
  draws_all <- bind_draws(draws2, draws3, draws4, along = "iteration")
  expect_equal(draws234, draws_all)

  expect_equal(nchains(draws_all), 1L)
  expect_equal(niterations(draws_all),
    niterations(draws2) + niterations(draws3) + niterations(draws4))
  expect_equal(niterations(draws_all), ndraws(draws_all))

  expect_equal(variables(draws_all), variables(draws2))

})

test_that("bind_draws works for draws_list objects", {
  draws1 <- as_draws_list(example_draws())
  draws2 <- subset_draws(draws1, chain = 2)
  draws3 <- subset_draws(draws1, iteration = 10:20)
  draws4 <- as_draws_df(data.frame(
    nu = rnorm(ndraws(draws1)),
    .chain = rep(chain_ids(draws1), each = niterations(draws1))
  ))

  draws_new <- bind_draws(draws1, draws4, along = "variable")
  expect_equal(
    variables(draws_new),
    c(variables(draws1), variables(draws4))
  )
  expect_equal(ndraws(draws_new), ndraws(draws1))

  draws_new <- bind_draws(draws1, draws2, along = "chain")
  expect_equal(nchains(draws_new), nchains(draws1) + nchains(draws2))
  expect_equal(variables(draws_new), variables(draws1))

  draws_new <- bind_draws(draws1, draws3, along = "iteration")
  expect_equal(
    niterations(draws_new),
    niterations(draws1) + niterations(draws3)
  )
  expect_equal(variables(draws_new), variables(draws1))

  expect_error(bind_draws(draws1, draws3, along = "draw"),
               "Cannot bind 'draws_list' objects along 'draw'")
})

test_that("bind_draws works for draws_rvars objects", {
  draws1 <- as_draws_rvars(example_draws())
  draws2 <- subset_draws(draws1, chain = 2)
  draws3 <- subset_draws(draws1, iteration = 10:20)
  draws4 <- as_draws_list(data.frame(
    nu = rnorm(ndraws(draws1)),
    .chain = rep(chain_ids(draws1), each = niterations(draws1))
  ))

  draws_new <- bind_draws(draws1, draws4, along = "variable")
  expect_equal(
    variables(draws_new),
    c(variables(draws1), variables(draws4))
  )
  expect_equal(ndraws(draws_new), ndraws(draws1))

  draws_new <- bind_draws(draws1, draws2, along = "chain")
  expect_equal(nchains(draws_new), nchains(draws1) + nchains(draws2))
  expect_equal(variables(draws_new), variables(draws1))

  expect_error(bind_draws(draws1, draws3, along = "iteration"),
    "Cannot bind 'draws_rvars' objects along 'iteration'")

  draws_new <- bind_draws(draws1, draws3, along = "draw")
  expect_equal(ndraws(draws_new), ndraws(draws1) + ndraws(draws3))
  expect_equal(nchains(draws_new), 1L)

  draws_new <- bind_draws(NULL, draws1)
  expect_equal(draws_new, draws1)
})

test_that("bind_draws works for list objects", {
  draws1 <- as_draws_df(example_draws())
  draws2 <- subset_draws(draws1, chain = 2)
  draws3 <-  subset_draws(draws1, chain = 3)

  draws12 <- bind_draws(draws1, draws2, along = "chain")
  draws_all <- bind_draws(draws1, draws2, draws3, along = "chain")
  expect_equal(bind_draws(list(draws1, draws2), along = "chain"), draws12)
  expect_equal(bind_draws(list(draws1, draws2, draws3), along = "chain"), draws_all)

  draws4 <-  subset_draws(draws1, chain = 4)
  draws_all <- bind_draws(draws2, draws3, draws4, along = "iteration")
  expect_equal(bind_draws(list(draws2, draws3, draws4), along = "iteration"), draws_all)
})

test_that("bind_draws errors if all NULL", {
  expect_error(bind_draws(NULL, NULL), "All objects passed to 'bind_draws' are NULL")
})

test_that("bind_draws errors for inputs with incompatible variables", {
  x1 <- x2 <- example_draws()
  variables(x1) <- gsub("theta", "beta", variables(x1))
  expect_error(bind_draws(x1, x2, along = "chain"), "'variables' of bound objects do not match")
})

test_that("repair_draws works correctly on draws_matrix objects", {
  x <- as_draws_matrix(example_draws())
  x <- x[10:6, ]

  x_rep <- repair_draws(x, order = TRUE)
  expect_equal(x[5:1, ], x_rep, check.attributes = FALSE)
  expect_equal(rownames(x_rep), as.character(1:5))

  x_rep <- repair_draws(x, order = FALSE)
  expect_equal(x, x_rep, check.attributes = FALSE)
  expect_equal(rownames(x_rep), as.character(1:5))
})

test_that("repair_draws works correctly on draws_array objects", {
  x <- as_draws_array(example_draws())
  x <- x[10:6, c(4, 1), ]

  x_rep <- repair_draws(x, order = TRUE)
  expect_equal(x[5:1, 2:1, ], x_rep, check.attributes = FALSE)
  expect_equal(rownames(x_rep), as.character(1:5))
  expect_equal(colnames(x_rep), as.character(1:2))

  x_rep <- repair_draws(x, order = FALSE)
  expect_equal(x, x_rep, check.attributes = FALSE)
  expect_equal(rownames(x_rep), as.character(1:5))
  expect_equal(colnames(x_rep), as.character(1:2))
})

test_that("repair_draws works correctly on draws_df objects", {
  x <- as_draws_df(example_draws())
  x <- subset(x, iteration = 1:5)
  x <- x[c(16, 11, 8, 2), ]

  x_rep <- repair_draws(x, order = TRUE)
  expect_equal(x$mu[order(x$.chain, x$.iteration)], x_rep$mu)
  expect_equal(x_rep$.iteration, rep(1, 4))
  expect_equal(x_rep$.chain, 1:4)

  x_rep <- repair_draws(x, order = FALSE)
  expect_equal(x$mu, x_rep$mu, check.attributes = FALSE)
  expect_equal(x_rep$.iteration, rep(1, 4))
  expect_equal(x_rep$.chain, 4:1)
})

test_that("repair_draws works correctly on draws_list objects", {
  x <- as_draws_list(example_draws())
  x <- x[c(4, 2)]

  x_rep <- repair_draws(x, order = TRUE)
  expect_equal(x[2], x_rep[1], check.attributes = FALSE)
  expect_equal(names(x_rep), as.character(1:2))

  x_rep <- repair_draws(x, order = FALSE)
  expect_equal(x, x_rep, check.attributes = FALSE)
  expect_equal(names(x_rep), as.character(1:2))
})

test_that("repair_draws works correctly on draws_rvars objects", {
  x <- as_draws_rvars(example_draws())
  draws_of(x$mu) <- draws_of(x$mu)[c(16, 11, 8, 2),, drop = FALSE]
  draws_of(x$tau) <- draws_of(x$tau)[c(16, 11, 8, 2),, drop = FALSE]
  draws_of(x$theta) <- draws_of(x$theta)[c(16, 11, 8, 2),, drop = FALSE]

  x_rep <- repair_draws(x, order = TRUE)
  expect_equal(
    draws_of(x_rep$mu),
    draws_of(x$mu)[order(as.integer(rownames(draws_of(x$mu)))),],
    check.attributes = FALSE
  )
  expect_equal(niterations(x_rep), 4)
  expect_equal(nchains(x_rep), 1)

  x_rep <- repair_draws(x, order = FALSE)
  expect_equal(draws_of(x$mu), draws_of(x_rep$mu), check.attributes = FALSE)
  expect_equal(niterations(x_rep), 4)
  expect_equal(nchains(x_rep), 1)

  x_ord <- order_draws(x)
  expect_equal(
    draws_of(x_ord$mu),
    draws_of(x$mu)[order(as.integer(rownames(draws_of(x$mu)))),],
    check.attributes = FALSE
  )
  expect_equal(
    draws_of(x_ord$theta),
    draws_of(x$theta)[order(as.integer(rownames(draws_of(x$theta)))),],
    check.attributes = FALSE
  )
})

test_that("indices work for NULL", {
   x <- NULL

   expect_equal(iteration_ids(NULL), NULL)
   expect_equal(chain_ids(NULL), NULL)
   expect_equal(draw_ids(NULL), NULL)

   expect_equal(niterations(NULL), 0)
   expect_equal(nchains(NULL), 0)
   expect_equal(ndraws(NULL), 0)
})

test_that("indices of draws_matrix objects are correct", {
   x <- as_draws_matrix(example_draws())

   expect_equal(iteration_ids(x), 1:100)
   expect_equal(chain_ids(x), 1:4)
   expect_equal(draw_ids(x), 1:NROW(x))

   expect_equal(niterations(x), NROW(x) / 4)
   expect_equal(nchains(x), 4)
   expect_equal(ndraws(x), NROW(x))

   rownames(x) <- NULL
   expect_equal(draw_ids(x), 1:NROW(x))
})

test_that("indices of draws_array objects are correct", {
  x <- as_draws_array(example_draws())

  expect_equal(iteration_ids(x), 1:NROW(x))
  expect_equal(chain_ids(x), 1:NCOL(x))
  expect_equal(draw_ids(x), 1:(NROW(x) * NCOL(x)))

  expect_equal(niterations(x), NROW(x))
  expect_equal(nchains(x), NCOL(x))
  expect_equal(ndraws(x), NROW(x) * NCOL(x))

  colnames(x) <- NULL
  rownames(x) <- NULL
  expect_equal(iteration_ids(x), 1:NROW(x))
  expect_equal(chain_ids(x), 1:NCOL(x))
})

test_that("indices of draws_df objects are correct", {
  x <- as_draws_df(example_draws())

  expect_equal(iteration_ids(x), unique(x$.iteration))
  expect_equal(chain_ids(x), unique(x$.chain))
  expect_equal(draw_ids(x), unique(x$.draw))

  expect_equal(niterations(x), length(unique(x$.iteration)))
  expect_equal(nchains(x), length(unique(x$.chain)))
  expect_equal(ndraws(x), length(unique(x$.draw)))
})

test_that("indices of draws_list objects are correct", {
  x <- as_draws_list(example_draws())

  expect_equal(iteration_ids(x), 1:length(x[[1]][[1]]))
  expect_equal(chain_ids(x), 1:length(x))
  expect_equal(draw_ids(x), 1:(length(x[[1]][[1]]) * length(x)))

  expect_equal(niterations(x), length(x[[1]][[1]]))
  expect_equal(nchains(x), length(x))
  expect_equal(ndraws(x), length(x[[1]][[1]]) * length(x))

  names(x) <- NULL
  expect_equal(chain_ids(x), 1:length(x))
})

test_that("indices of draws_rvars objects are correct", {
  x <- as_draws_rvars(example_draws())

  expect_equal(iteration_ids(x), 1:(length(draws_of(x[[1]]))/4))
  expect_equal(chain_ids(x), 1:4)
  expect_equal(draw_ids(x), 1:length(draws_of(x[[1]])))

  expect_equal(niterations(x), length(draws_of(x[[1]]))/4)
  expect_equal(nchains(x), 4)
  expect_equal(ndraws(x), length(draws_of(x[[1]])))
})

test_that("indexing draws_array with [ and drop works correctly", {
  x <- example_draws()
  x1 <- x[,,1]
  x2 <- x[,,1, drop=TRUE]
  expect_s3_class(x1, "draws_array")
  if (R.version$major >= "4") {
    expect_equal(class(x2), c("matrix", "array"))
  } else {
    expect_equal(class(x2), "matrix")
  }
  expect_length(dim(x1), 3)
  expect_length(dim(x2), 2)
  expect_equal(x2, extract_variable_matrix(x, "mu"))

  # drop=TRUE shouldn't do anything if multiple parameters selected
  x3 <- x[,,1:2, drop=TRUE]
  expect_s3_class(x3, "draws_array")
})

test_that("indexing draws_matrix with [ and drop works correctly", {
  x <- as_draws_matrix(example_draws())
  x1 <- x[,1]
  x2 <- x[,1, drop=TRUE]
  expect_s3_class(x1, "draws_matrix")
  expect_equal(class(x2), "numeric")
  expect_length(dim(x1), 2)
  expect_null(dim(x2))

  # drop=TRUE shouldn't do anything if multiple parameters selected
  x3 <- x[,1:2, drop=TRUE]
  expect_s3_class(x3, "draws_matrix")
})

test_that("indexing draws dimension draws_matrix triggers a warning", {
  options(posterior.warn_on_merge_chains = TRUE)
  x <- as_draws_matrix(example_draws())
  expect_warning(
    x1 <- x[1:3, ],
    "Chains were dropped due to manually indexing draws"
  )
  x <- merge_chains(x)
  x2 <- x[1:3, ]
  expect_equal(x1, x2)
  options(posterior.warn_on_merge_chains = FALSE)
})

test_that("merge_chains works correctly", {
   x <- example_draws()

   xm <- merge_chains(as_draws_matrix(x))
   expect_equal(nchains(xm), 1)
   expect_equal(ndraws(xm), ndraws(x))
   expect_equal(variables(xm), variables(x))

   xm <- merge_chains(as_draws_array(x))
   expect_equal(nchains(xm), 1)
   expect_equal(ndraws(xm), ndraws(x))
   expect_equal(variables(xm), variables(x))

   xm <- merge_chains(as_draws_df(x))
   expect_equal(nchains(xm), 1)
   expect_equal(ndraws(xm), ndraws(x))
   expect_equal(variables(xm), variables(x))

   xm <- merge_chains(as_draws_list(x))
   expect_equal(nchains(xm), 1)
   expect_equal(ndraws(xm), ndraws(x))
   expect_equal(variables(xm), variables(x))

   # need to keep x_rvar around for this test because draws_rvars counts
   # variables differently from the other formats
   x_rvar <- as_draws_rvars(x)
   xm <- merge_chains(x_rvar)
   expect_equal(nchains(xm), 1)
   expect_equal(ndraws(xm), ndraws(x))
   expect_equal(variables(xm), variables(x_rvar))
})

test_that("subset_draws works correctly for draws_matrix objects", {
  x <- as_draws_matrix(example_draws())
  x_sub <- subset_draws(x, variable = c("mu", "tau"), iteration = 5:10)
  x_sub2 <- x[c(5:10, 105:110, 205:210, 305:310), c("mu", "tau")]
  expect_equal(x_sub, x_sub2, check.attributes = FALSE)
  expect_equal(iteration_ids(x_sub), 1:6)

  x_sub <- subset_draws(x, draw = c(2, 2, 4, 4), unique = FALSE)
  expect_equal(niterations(x_sub), 4)
  expect_equivalent(x_sub[1, ], x_sub[2, ])

  x <- weight_draws(x, rep(1, ndraws(x)))
  x_sub <- subset_draws(x, variable = "mu")
  expect_equal(variables(x_sub, reserved = TRUE), c("mu", ".log_weight"))
})

test_that("subset_draws works correctly for draws_array objects", {
  x <- as_draws_array(example_draws())

  x_sub <- subset_draws(x, variable = c("mu", "tau"), iteration = 5:10, chain = 3:4)
  expect_equal(x[5:10, 3:4, c("mu", "tau")], x_sub, check.attributes = FALSE)
  expect_equal(iteration_ids(x_sub), 1:6)
  expect_equal(chain_ids(x_sub), 1:2)

  x_sub <- subset_draws(x, chain = c(1, 1), unique = FALSE)
  expect_equal(nchains(x_sub), 2)
  expect_equivalent(x_sub[, 1, ], x_sub[, 2, ])

  expect_message(
    x_sub <- subset_draws(x, draw = c(1, 200, 10)),
    "Merging chains in order to subset via 'draw'"
  )
  expect_equal(niterations(x_sub), 3)

  x <- weight_draws(x, rep(1, ndraws(x)))
  x_sub <- subset_draws(x, variable = "mu")
  expect_equal(variables(x_sub, reserved = TRUE), c("mu", ".log_weight"))
})

test_that("subset_draws works correctly for draws_df objects", {
  x <- as_draws_df(example_draws())
  x_sub <- subset_draws(x, variable = c("mu", "tau"), iteration = 5:10, chain = 3:4)
  expect_equal(x$mu[x$.iteration %in% 5:10 & x$.chain %in% 3:4], x_sub$mu)
  expect_equal(iteration_ids(x_sub), 1:6)
  expect_equal(chain_ids(x_sub), 1:2)

  x_sub <- subset_draws(x, draw = c(5, 5, 6), unique = FALSE)
  expect_equal(ndraws(x_sub), 3)
  expect_equal(x_sub$mu[1], x_sub$mu[2])

  x <- weight_draws(x, rep(1, ndraws(x)))
  x_sub <- subset_draws(x, variable = "mu")
  expect_equal(names(x_sub), c("mu", ".log_weight", ".chain", ".iteration", ".draw"))
})

test_that("subset_draws works correctly for draws_list objects", {
  x <- as_draws_list(example_draws())
  x_sub <- subset_draws(x, variable = c("theta[1]"), iteration = 5:10, chain = 3:4)
  expect_equal(variables(x_sub), "theta[1]")
  expect_equal(iteration_ids(x_sub), 1:6)
  expect_equal(chain_ids(x_sub), 1:2)

  x_sub <- subset_draws(x, iteration = c(1, 1, 2), unique = FALSE)
  expect_equal(niterations(x_sub), 3)
  expect_equal(x_sub[[1]]$mu[1], x_sub[[1]]$mu[2])

  expect_message(
    x_sub <- subset_draws(x, draw = c(1, 200, 10)),
    "Merging chains in order to subset via 'draw'"
  )
  expect_equal(niterations(x_sub), 3)

  x <- weight_draws(x, rep(1, ndraws(x)))
  x_sub <- subset_draws(x, variable = "mu")
  expect_equal(variables(x_sub, reserved = TRUE), c("mu", ".log_weight"))
})

test_that("subset_draws works correctly for draws_rvars objects", {
  x <- as_draws_rvars(example_draws())
  x_sub <- subset_draws(x, variable = c("mu"), iteration = 5:10, chain = 3:4)
  expect_equal(variables(x_sub), "mu")
  expect_equal(iteration_ids(x_sub), 1:6)
  expect_equal(chain_ids(x_sub), 1:2)

  x_sub <- subset_draws(x, iteration = c(1, 1, 2), unique = FALSE)
  expect_equal(niterations(x_sub), 3)
  expect_equal(draws_of(x_sub[[1]]$mu)[1], draws_of(x_sub[[1]]$mu)[2])

  expect_message(
    x_sub <- subset_draws(x, draw = c(1, 200, 10)),
    "Merging chains in order to subset via 'draw'"
  )
  expect_equal(niterations(x_sub), 3)

  x <- weight_draws(x, rep(1, ndraws(x)))
  x_sub <- subset_draws(x, variable = "mu")
  expect_equal(variables(x_sub, reserved = TRUE), c("mu", ".log_weight"))
})

test_that("subset_draws works correctly for rvar objects", {
  x <- as_draws_rvars(example_draws())$theta

  expect_error(subset_draws(x, variable = "mu"), "Cannot subset an rvar by variable")

  x_sub <- subset_draws(x, iteration = c(1, 1, 2), unique = FALSE)
  expect_equal(niterations(x_sub), 3)
  expect_equal(draws_of(x_sub)[1], draws_of(x_sub)[2])

  expect_message(
    x_sub <- subset_draws(x, draw = c(1, 200, 10)),
    "Merging chains in order to subset via 'draw'"
  )
  expect_equal(niterations(x_sub), 3)
})

test_that("variables can be subsetted via regular expressions", {
  x <- as_draws_df(example_draws())
  x_sub <- subset_draws(x, variable = c("theta\\[", "m"), regex = TRUE)
  expect_equal(variables(x_sub), c(paste0("theta[", 1:8, "]"), "mu"))

  # do the same thing using the 'subset' alias
  x_sub <- subset(x, variable = c("theta\\[", "m"), regex = TRUE)
  expect_equal(variables(x_sub), c(paste0("theta[", 1:8, "]"), "mu"))
})

test_that("variables can be subsetted via non-scalar selection", {
  x <- as_draws_df(example_draws())
  x_sub <- subset_draws(x, variable = "theta")
  expect_equal(variables(x_sub), c(paste0("theta[", 1:8, "]")))
})

test_that("subset_draws speed is tolerable with many variables", {
  # some machines will be slower and so this test is unreliable on CRAN
  skip_on_cran()
  x <- as_draws_matrix(matrix(rnorm(10 * 300000), nrow = 10))
  tt <- system.time(x2 <- subset_draws(x, colnames(x)))
  expect_equal(x, x2)
  expect_lt(tt[["elapsed"]], 1)
})

test_that("subset_draws errors if selecting missing variables", {
  x <- as_draws_matrix(example_draws())
  expect_error(
    subset_draws(x, variable = c("theta[2]", "X", "theta[3]", "Y")),
    "The following variables are missing in the draws object: {'X', 'Y'}",
    fixed = TRUE
  )
})

test_that("subset_draws preserves variable order", {
  x <- as_draws_matrix(example_draws())
  x <- subset_draws(x, variable = c("theta[2]", "theta[1]"))
  expect_equal(variables(x), c("theta[2]", "theta[1]"))
})

test_that("subset_draws preserves variable order with vectors", {
  x <- as_draws_matrix(example_draws())
  theta_names <- paste0("theta[", 1:8, "]")
  # Expect variables to be returned in the order listed:
  v1 <- variables(subset_draws(x, variable = c("theta", "mu")))
  expect_equal(v1, c(theta_names, "mu"))
  v2 <- variables(subset_draws(x, variable = c("mu", "theta")))
  expect_equal(v2, c("mu", theta_names))
  v3 <- variables(subset_draws(x, variable = c("mu", "theta", "tau")))
  expect_equal(v3, c("mu", theta_names, "tau"))
  # No duplication:
  v4 <- variables(subset_draws(x, variable = c("mu", "mu", "theta")))
  expect_equal(v4, c("mu", theta_names))
  v5 <- variables(subset_draws(x, variable = c("mu", "theta", "theta")))
  expect_equal(v5, c("mu", theta_names))
  v6 <- variables(subset_draws(x, variable = c("theta", "mu", "theta")))
  expect_equal(v6, c(theta_names, "mu"))

  # Output is sorted numerically, not alphabetically
  x2 <- as_draws_matrix(matrix(rep.int(1, 11 * 3), ncol = 11))
  colnames(x2) <- paste0("a[",1:11,"]")
  v7 <- variables(subset_draws(x2, variable = "a"))
  expect_equal(v7, colnames(x2))
})

test_that("non-unique subsetting for draws_df same as doing it with draws_list", {
  x_df <- as_draws_df(example_draws())
  x_list <- as_draws_list(example_draws())

  x_df_sub <- subset_draws(x_df, chain = c(1,1,2), iteration = c(1:2, 1:50),
                           unique = FALSE)
  x_list_sub <- subset_draws(x_list, chain = c(1,1,2), iteration = c(1:2, 1:50),
                             unique = FALSE)
  expect_equal(x_df_sub, as_draws_df(x_list_sub))
})

test_that("extract_variable_array works the same for different formats", {
  draws <- list(
    array = as_draws_array(example_draws()),
    df = as_draws_df(example_draws()),
    list = as_draws_list(example_draws()),
    matrix = as_draws_matrix(example_draws()),
    rvars = as_draws_rvars(example_draws())
  )
  
  for (type in names(draws)) {
    mu <- extract_variable_array(draws[[type]], variable = "mu")
    expect_equal(length(dim(mu)), 3)
    expect_equal(dim(mu)[1], niterations(draws[[type]]))
    expect_equal(dim(mu)[2], nchains(draws[[type]]))
    expect_equal(dim(mu)[3], 1)
    
    theta <- extract_variable_array(draws[[type]], variable = "theta")
    expect_equal(length(dim(theta)), 3)
    expect_equal(dim(theta)[1], niterations(draws[[type]]))
    expect_equal(dim(theta)[2], nchains(draws[[type]]))
    expect_equal(dim(theta)[3], 8)
  }
})

test_that("extract_variable_array with_chains parameter works", {
  x <- example_draws()
  
  # Test with_chains = TRUE (default)
  mu_with_chains <- extract_variable_array(x, "mu", with_chains = TRUE)
  expect_equal(dim(mu_with_chains)[1], niterations(x))
  expect_equal(dim(mu_with_chains)[2], nchains(x))
  
  # Test with_chains = FALSE
  mu_no_chains <- extract_variable_array(x, "mu", with_chains = FALSE)
  expect_equal(dim(mu_no_chains)[1], ndraws(x))
  expect_equal(dim(mu_no_chains)[2], 1)  # scalar variable
  
  # Test that the draws are the same, just reshaped
  expect_equal(c(mu_with_chains), c(mu_no_chains))
})

test_that("extract_variable_array with_chains works for different formats", {
  draws <- list(
    array = as_draws_array(example_draws()),
    df = as_draws_df(example_draws()),
    list = as_draws_list(example_draws()),
    matrix = as_draws_matrix(example_draws()),
    rvars = as_draws_rvars(example_draws())
  )
  
  for (type in names(draws)) {
    # Test with chains
    result_with <- extract_variable_array(draws[[type]], "mu", with_chains = TRUE)
    expect_equal(dim(result_with)[1], niterations(draws[[type]]))
    expect_equal(dim(result_with)[2], nchains(draws[[type]]))
    
    # Test without chains 
    result_without <- extract_variable_array(draws[[type]], "mu", with_chains = FALSE)
    expect_equal(dim(result_without)[1], ndraws(draws[[type]]))
    
    # Values should be the same
    expect_equal(c(result_with), c(result_without))
  }
})

test_that("extract_variable_array with_chains works with indexed variables", {
  x <- example_draws(example = "multi_normal")
  
  # Test specific indexed variable
  mu1_with <- extract_variable_array(x, "mu[1]", with_chains = TRUE)
  mu1_without <- extract_variable_array(x, "mu[1]", with_chains = FALSE)
  
  # With chains: should be 3D array (niterations, nchains, 1)
  expect_equal(length(dim(mu1_with)), 3)
  expect_equal(dim(mu1_with)[3], 1)
  
  # Without chains: should be 2D array (ndraws, 1)  
  expect_equal(length(dim(mu1_without)), 2)
  expect_equal(dim(mu1_without)[2], 1)
  
  # Values should match
  expect_equal(c(mu1_with), c(mu1_without))
})

test_that("extract_list_of_variable_arrays passes with_chains parameter", {
  x <- example_draws()
  
  # Test that with_chains is passed through to extract_variable_array
  vars_with <- extract_list_of_variable_arrays(x, c("mu", "tau"), with_chains = TRUE)
  vars_without <- extract_list_of_variable_arrays(x, c("mu", "tau"), with_chains = FALSE)
  
  # Check dimensions
  expect_equal(dim(vars_with$mu)[1], niterations(x))
  expect_equal(dim(vars_with$mu)[2], nchains(x))
  expect_equal(dim(vars_without$mu)[1], ndraws(x))
  
  # Values should be the same
  expect_equal(c(vars_with$mu), c(vars_without$mu))
  expect_equal(c(vars_with$tau), c(vars_without$tau))
  
  # Compare with individual extractions
  expect_equal(vars_with$mu, extract_variable_array(x, "mu", with_chains = TRUE))
  expect_equal(vars_without$mu, extract_variable_array(x, "mu", with_chains = FALSE))
})

test_that("extract_variable_array preserves factor levels with indexed variables", {
  # Create rvar_factor with known levels (vector, not matrix)
  mu <- rvar_factor(
    array(sample(c("low", "med", "high"), 200, replace = TRUE), dim = c(100, 2))
  )
  draws <- as_draws_rvars(list(mu = mu))
  
  # Extract indexed variable with both with_chains settings
  mu1_chains <- extract_variable_array(draws, "mu[1]", with_chains = TRUE)
  mu1_nochains <- extract_variable_array(draws, "mu[1]", with_chains = FALSE)
  
  # Both should be factors
  expect_s3_class(mu1_chains, "factor")
  expect_s3_class(mu1_nochains, "factor")
  
  # Both should have the same levels
  expect_equal(levels(mu1_chains), levels(mu1_nochains))
  expect_equal(levels(mu1_chains), c("high", "low", "med"))
  
  # Values should be the same
  expect_equal(c(mu1_chains), c(mu1_nochains))
})

---
title: "The posterior R package"
author: "Paul Bürkner, Jonah Gabry, Matthew Kay, and Aki Vehtari"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{The posterior R package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction 

The posterior R package is intended to provide useful tools for both users
and developers of packages for fitting Bayesian models or working with output
from Bayesian models. The primary goals of the package are to:

* Efficiently convert between many different useful formats of
  draws (samples) from posterior or prior distributions.
* Provide consistent methods for operations commonly performed on draws, 
  for example, subsetting, binding, or mutating draws.
* Provide various summaries of draws in convenient formats.
* Provide lightweight implementations of state of the art posterior inference 
  diagnostics.

## Installation

You can install the latest official release version via

```{r install, eval=FALSE}
install.packages("posterior")
```

or the latest development version from GitHub via 

```{r install_github, eval=FALSE}
# install.packages("remotes")
remotes::install_github("stan-dev/posterior")
```

## Example 

```{r setup}
library("posterior")
```

To demonstrate how to work with the posterior package, throughout the rest of
this vignette we will use example posterior draws obtained from the eight
schools hierarchical meta-analysis model described in Gelman et al. (2013). The
variables are an estimate per school (`theta[1]` through `theta[8]`) as well as
an overall mean (`mu`) and standard deviation across schools (`tau`).

```{r example-drawss}
eight_schools_array <- example_draws("eight_schools")
print(eight_schools_array, max_variables = 3)
```

The structure of this object is explained in the next section.

## Draws formats 

### Available formats 

Because different formats are preferable in different situations, posterior
supports multiple formats and easy conversion between them. The currently
supported formats are:

* `draws_array`: An iterations by chains by variables array.
* `draws_matrix`: A draws (iterations x chains) by variables array.
* `draws_df`: A draws by variables data frame with addition meta columns
`.chain`, `.iteration`, `.draw`.
* `draws_list`: A list with one sublist per chain. Each sublist is a named list
with one vector of iterations per variable.
* `draws_rvars`: A list of random variable `rvar` objects, one per variable. See
`vignette("rvar")` for an introduction to this new data type.

These formats are essentially base R object classes and can be used as such. For
example, a `draws_matrix` object is just a `matrix` with a little more
consistency (e.g., no dropping of dimensions with one level when indexing) and
additional methods. The exception to this is the `draws_rvars` format, which
contains `rvar` objects that behave somewhat like arrays but are really a unique
data type. See the separate vignette on the `rvar` and `draws_rvars` data types
for details.

The draws for our example come as a `draws_array` object with 
`r niterations(eight_schools_array)` iterations, 
`r nchains(eight_schools_array)` chains, and 
`r nvariables(eight_schools_array)` variables:

```{r draws_array-structure}
str(eight_schools_array)
```

### Converting between formats

Each of the formats has a method `as_draws_<format>` (e.g., `as_draws_list()`) 
for creating an object  of the class from any of the other formats. As a 
demonstration we can convert the example `draws_array` to a `draws_df`, 
a data frame with additional meta information. To convert to a `draws_df` we
use `as_draws_df()`.

```{r draws_df}
eight_schools_df <- as_draws_df(eight_schools_array)
str(eight_schools_df)
print(eight_schools_df)
```

### Converting regular R objects to `draws` formats

The example draws already come in a format natively supported by posterior, but
we can of course also import the draws from other sources like common base R
objects.

#### Example: create draws_matrix from a matrix

In addition to converting other `draws` objects to the `draws_matrix` format,
the `as_draws_matrix()` function will convert a regular matrix to a
`draws_matrix`.

```{r draws_matrix-from-matrix}
x <- matrix(rnorm(50), nrow = 10, ncol = 5)
colnames(x) <- paste0("V", 1:5)
x <- as_draws_matrix(x)
print(x)
```
Because the matrix was converted to a `draws_matrix`, all of the methods for
working with `draws` objects described in subsequent sections of this vignette
will now be available.

Instead of `as_draws_matrix()` we also could have just used `as_draws()`, which
attempts to find the closest available format to the input object. In this case
the result would be a `draws_matrix` object either way.

#### Example: create draws_matrix from multiple vectors

In addition to the `as_draws_matrix()` converter function there is also a
`draws_matrix()` constructor function that can be used to create draws matrix
from multiple vectors.

```{r draws_matrix-from-vectors}
x <- draws_matrix(alpha = rnorm(50), beta = rnorm(50))
print(x)
```

Analogous functions exist for the other draws formats and are used similarly.


## Manipulating `draws` objects

The posterior package provides many methods for manipulating draws objects in
useful ways. In this section we demonstrate several of the most commonly used
methods. These methods, like the other methods in posterior, are available for
every supported draws format.


### Subsetting 

Subsetting `draws` objects can be done according to various aspects of the draws
(iterations, chains, or variables). The posterior package provides a convenient
interface for this purpose via `subset_draws()`. For example, here is the
code to extract the first five iterations of the first two chains of the
variable `mu`.

```{r subset-df}
sub_df <- subset_draws(eight_schools_df, variable = "mu", chain = 1:2, iteration = 1:5)
str(sub_df)
```

The same call to `subset_draws()` can be used regardless of the draws format.
For example, here is the same code except replacing the `draws_df` object with 
the `draws_array` object. 

```{r subset-array}
sub_arr <- subset_draws(eight_schools_array, variable = "mu", chain = 1:2, iteration = 1:5)
str(sub_arr)
```

We can check that these two calls to `subset_draws()` (the first with the data
frame, the second with the array) produce the same result.

```{r subset-compare, results='hold'}
identical(sub_df, as_draws_df(sub_arr))
identical(as_draws_array(sub_df), sub_arr)
```

It is also possible to use standard R subsetting syntax with `draws` objects.
The following is equivalent to the use of `subset_draws()` with the array above.

```{r subset-standard}
eight_schools_array[1:5, 1:2, "mu"]
```

The major difference between how posterior behaves when indexing and how base R
behaves is that posterior will _not_ drop dimensions with only one level. That
is, even though there is only one variable left after subsetting, the result of
the subsetting above is still a `draws_array` and not a `draws_matrix`.

### Mutating (transformations of variables)

The magic of having obtained draws from the joint posterior (or prior)
distribution of a set of variables is that these draws can also be used to
obtain draws from any other variable that is a function of the original
variables.
That is, if we are interested in the posterior distribution of, say,
`phi = (mu + tau)^2` all we have to do is to perform the transformation for each
of the individual draws to obtain draws from the posterior distribution of the
transformed variable. This procedure is handled by `mutate_variables()`.

```{r mutate}
x <- mutate_variables(eight_schools_df, phi = (mu + tau)^2)
x <- subset_draws(x, c("mu", "tau", "phi"))
print(x)
```

### Renaming

To rename variables use `rename_variables()`. Here we rename the scalar `mu` to
`mean` and the vector `theta` to `alpha`.

```{r rename}
# mu is a scalar, theta is a vector
x <- rename_variables(eight_schools_df, mean = mu, alpha = theta)
variables(x)
```
In the call to `rename_variables()` above, `mu` and `theta` can be quoted or
unquoted.

It is also possible to rename individual elements of non-scalar parameters, 
for example we can rename just the first element of `alpha`:

```{r rename-element}
x <- rename_variables(x, a1 = `alpha[1]`)
variables(x)
```

### Binding

The `bind_draws()` method can be used to combine `draws` objects along different
dimensions. As an example, suppose we have several different `draws_matrix`
objects:

```{r objects-to-bind}
x1 <- draws_matrix(alpha = rnorm(5), beta = rnorm(5))
x2 <- draws_matrix(alpha = rnorm(5), beta = rnorm(5))
x3 <- draws_matrix(theta = rexp(5))
```

We can bind `x1` and `x3` together along the `'variable'` dimension to get 
a single `draws_matrix` with the variables from both `x1` and `x3`: 

```{r bind-variable}
x4 <- bind_draws(x1, x3, along = "variable")
print(x4)
```

Because `x1` and `x2` have the same variables, we can bind them along the
`'draw'` dimension to create a single `draws_matrix` with more draws:

```{r bind-draw}
x5 <- bind_draws(x1, x2, along = "draw")
print(x5)
```

As with all posterior methods, `bind_draws()` can be used with all draws formats
and depending on the format different dimensions are available to bind on. For
example, we can bind `draws_array` objects together by `iteration`, `chain`, or
`variable`, but a 2-D `draws_matrix` with the chains combined can only by bound
by `draw` and `variable`.


## Summaries and diagnostics

### summarise_draws() basic usage

Computing summaries of posterior or prior draws and convergence diagnostics for
posterior draws are some of the most common tasks when working with Bayesian
models fit using Markov Chain Monte Carlo (MCMC) methods. The posterior
package provides a flexible interface for this purpose via `summarise_draws()`
(or `summarize_draws()`), which can be passed any of the formats supported by
the package.

```{r summary}
# summarise_draws or summarize_draws
summarise_draws(eight_schools_df)
```

The result is a data frame with one row per variable and one column per
summary statistic or convergence diagnostic. The summaries `rhat`, `ess_bulk`,
and `ess_tail` are described in Vehtari et al. (2020). We can choose which
summaries to compute by passing additional arguments, either functions or names
of functions. For instance, if we only wanted the mean and its corresponding
Monte Carlo Standard Error (MCSE) we could use either of these options:

```{r summary-with-measures}
# the function mcse_mean is provided by the posterior package
s1 <- summarise_draws(eight_schools_df, "mean", "mcse_mean") 
s2 <- summarise_draws(eight_schools_df, mean, mcse_mean) 
identical(s1, s2)
print(s1)
```

### Changing column names

The column names in the output can be changed by providing the functions 
as name-value pairs, where the name is the name to use in the output and the 
value is a function name or definition. For example, here we change the names
`mean` and `sd` to `posterior_mean` and `posterior_sd`.

```{r change-summary-names}
summarise_draws(eight_schools_df, posterior_mean = mean, posterior_sd = sd)
```

### Using custom functions 

For a function to work with `summarise_draws()`, it needs to take a vector or
matrix of numeric values and return a single numeric value or a named vector of
numeric values. Additional arguments to the function can be specified in a list
passed to the `.args` argument.

```{r summary-.args}
weighted_mean <- function(x, wts) {
  sum(x * wts)/sum(wts)
}
summarise_draws(
  eight_schools_df, 
  weighted_mean, 
  .args = list(wts = rexp(ndraws(eight_schools_df)))
)
```

### Specifying functions using lambda-like syntax

It is also possible to specify a summary function using a one-sided formula that
follows the conventions supported by `rlang::as_function()`. For example, the
function

```{r standard-quantile, eval = FALSE}
function(x) quantile(x, probs = c(0.4, 0.6))
```

can be simplified to 

```{r lambda-quantile, eval = FALSE}
# for multiple arguments `.x` and `.y` can be used, see ?rlang::as_function
~quantile(., probs = c(0.4, 0.6))
```

Both can be used with `summarise_draws()` and produce the same output:

```{r lambda-syntax}
summarise_draws(eight_schools_df, function(x) quantile(x, probs = c(0.4, 0.6)))

summarise_draws(eight_schools_df, ~quantile(.x, probs = c(0.4, 0.6)))
```

See `help("as_function", "rlang")` for details on specifying these functions.

### Other diagnostics

In addition to the default diagnostic functions used by `summarise_draws()`
(`rhat()`, `ess_bulk()`, `ess_tail()`), posterior also provides additional
diagnostics like effective sample sizes and Monte Carlo standard errors for
quantiles and standard deviations, an experimental new diagnostic called R*, and
others. For a list of available diagnostics and links to their individual help
pages see `help("diagnostics", "posterior")`.

If you have suggestions for additional diagnostics that should be implemented in
posterior, please open an issue at
<https://github.com/stan-dev/posterior/issues>.


## Other methods for working with `draws` objects

In addition to the methods demonstrated in this vignette, posterior has various
other methods available for working with `draws` objects. The following is a
(potentially incomplete) list.

|**Method**|**Description**|
|:----------|:---------------|
| `order_draws()` | Order `draws` objects according to iteration and chain number |
| `repair_draws()`| Repair indices of `draws` objects so that iterations chains, and draws are continuously and consistently numbered |
|`resample_draws()` | Resample `draws` objects according to provided weights |
| `thin_draws()` | Thin `draws` objects to reduce size and autocorrelation |
| `weight_draws()`| Add weights to draws objects, with one weight per draw, for use in subsequent weighting operations |
| `extract_variable()` | Extract a vector of draws of a single variable  |
| `extract_variable_matrix()` | Extract an iterations x chains matrix of draws of a single variable |
| `merge_chains()` | Merge chains of `draws` objects into a single chain. | 
| `split_chains()` | Split chains of `draws` objects by halving the number of iterations per chain and doubling the number of chains. |

If you have suggestions for additional methods that would be useful for working
with `draws` objects, please open an issue at
<https://github.com/stan-dev/posterior/issues>.


## References

Gelman A., Carlin J. B., Stern H. S., David B. Dunson D. B., Vehtari, A.,
& Rubin D. B. (2013). *Bayesian Data Analysis, Third Edition*. Chapman and
Hall/CRC.

Vehtari A., Gelman A., Simpson D., Carpenter B., & Bürkner P. C. (2020).
Rank-normalization, folding, and localization: An improved Rhat for assessing
convergence of MCMC. *Bayesian Analysis*, 16(2):667-718.

---
title: "The posterior R package"
author: "Paul Bürkner, Jonah Gabry, Matthew Kay, and Aki Vehtari"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{The posterior R package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction 

The posterior R package is intended to provide useful tools for both users
and developers of packages for fitting Bayesian models or working with output
from Bayesian models. The primary goals of the package are to:

* Efficiently convert between many different useful formats of
  draws (samples) from posterior or prior distributions.
* Provide consistent methods for operations commonly performed on draws, 
  for example, subsetting, binding, or mutating draws.
* Provide various summaries of draws in convenient formats.
* Provide lightweight implementations of state of the art posterior inference 
  diagnostics.

## Installation

You can install the latest official release version via

```{r install, eval=FALSE}
install.packages("posterior")
```

or the latest development version from GitHub via 

```{r install_github, eval=FALSE}
# install.packages("remotes")
remotes::install_github("stan-dev/posterior")
```

## Example 

```{r setup}
library("posterior")
```

To demonstrate how to work with the posterior package, throughout the rest of
this vignette we will use example posterior draws obtained from the eight
schools hierarchical meta-analysis model described in Gelman et al. (2013). The
variables are an estimate per school (`theta[1]` through `theta[8]`) as well as
an overall mean (`mu`) and standard deviation across schools (`tau`).

```{r example-drawss}
eight_schools_array <- example_draws("eight_schools")
print(eight_schools_array, max_variables = 3)
```

The structure of this object is explained in the next section.

## Draws formats 

### Available formats 

Because different formats are preferable in different situations, posterior
supports multiple formats and easy conversion between them. The currently
supported formats are:

* `draws_array`: An iterations by chains by variables array.
* `draws_matrix`: A draws (iterations x chains) by variables array.
* `draws_df`: A draws by variables data frame with addition meta columns
`.chain`, `.iteration`, `.draw`.
* `draws_list`: A list with one sublist per chain. Each sublist is a named list
with one vector of iterations per variable.
* `draws_rvars`: A list of random variable `rvar` objects, one per variable. See
`vignette("rvar")` for an introduction to this new data type.

These formats are essentially base R object classes and can be used as such. For
example, a `draws_matrix` object is just a `matrix` with a little more
consistency (e.g., no dropping of dimensions with one level when indexing) and
additional methods. The exception to this is the `draws_rvars` format, which
contains `rvar` objects that behave somewhat like arrays but are really a unique
data type. See the separate vignette on the `rvar` and `draws_rvars` data types
for details.

The draws for our example come as a `draws_array` object with 
`r niterations(eight_schools_array)` iterations, 
`r nchains(eight_schools_array)` chains, and 
`r nvariables(eight_schools_array)` variables:

```{r draws_array-structure}
str(eight_schools_array)
```

### Converting between formats

Each of the formats has a method `as_draws_<format>` (e.g., `as_draws_list()`) 
for creating an object  of the class from any of the other formats. As a 
demonstration we can convert the example `draws_array` to a `draws_df`, 
a data frame with additional meta information. To convert to a `draws_df` we
use `as_draws_df()`.

```{r draws_df}
eight_schools_df <- as_draws_df(eight_schools_array)
str(eight_schools_df)
print(eight_schools_df)
```

### Converting regular R objects to `draws` formats

The example draws already come in a format natively supported by posterior, but
we can of course also import the draws from other sources like common base R
objects.

#### Example: create draws_matrix from a matrix

In addition to converting other `draws` objects to the `draws_matrix` format,
the `as_draws_matrix()` function will convert a regular matrix to a
`draws_matrix`.

```{r draws_matrix-from-matrix}
x <- matrix(rnorm(50), nrow = 10, ncol = 5)
colnames(x) <- paste0("V", 1:5)
x <- as_draws_matrix(x)
print(x)
```
Because the matrix was converted to a `draws_matrix`, all of the methods for
working with `draws` objects described in subsequent sections of this vignette
will now be available.

Instead of `as_draws_matrix()` we also could have just used `as_draws()`, which
attempts to find the closest available format to the input object. In this case
the result would be a `draws_matrix` object either way.

#### Example: create draws_matrix from multiple vectors

In addition to the `as_draws_matrix()` converter function there is also a
`draws_matrix()` constructor function that can be used to create draws matrix
from multiple vectors.

```{r draws_matrix-from-vectors}
x <- draws_matrix(alpha = rnorm(50), beta = rnorm(50))
print(x)
```

Analogous functions exist for the other draws formats and are used similarly.


## Manipulating `draws` objects

The posterior package provides many methods for manipulating draws objects in
useful ways. In this section we demonstrate several of the most commonly used
methods. These methods, like the other methods in posterior, are available for
every supported draws format.


### Subsetting 

Subsetting `draws` objects can be done according to various aspects of the draws
(iterations, chains, or variables). The posterior package provides a convenient
interface for this purpose via `subset_draws()`. For example, here is the
code to extract the first five iterations of the first two chains of the
variable `mu`.

```{r subset-df}
sub_df <- subset_draws(eight_schools_df, variable = "mu", chain = 1:2, iteration = 1:5)
str(sub_df)
```

The same call to `subset_draws()` can be used regardless of the draws format.
For example, here is the same code except replacing the `draws_df` object with 
the `draws_array` object. 

```{r subset-array}
sub_arr <- subset_draws(eight_schools_array, variable = "mu", chain = 1:2, iteration = 1:5)
str(sub_arr)
```

We can check that these two calls to `subset_draws()` (the first with the data
frame, the second with the array) produce the same result.

```{r subset-compare, results='hold'}
identical(sub_df, as_draws_df(sub_arr))
identical(as_draws_array(sub_df), sub_arr)
```

It is also possible to use standard R subsetting syntax with `draws` objects.
The following is equivalent to the use of `subset_draws()` with the array above.

```{r subset-standard}
eight_schools_array[1:5, 1:2, "mu"]
```

The major difference between how posterior behaves when indexing and how base R
behaves is that posterior will _not_ drop dimensions with only one level. That
is, even though there is only one variable left after subsetting, the result of
the subsetting above is still a `draws_array` and not a `draws_matrix`.

### Mutating (transformations of variables)

The magic of having obtained draws from the joint posterior (or prior)
distribution of a set of variables is that these draws can also be used to
obtain draws from any other variable that is a function of the original
variables.
That is, if we are interested in the posterior distribution of, say,
`phi = (mu + tau)^2` all we have to do is to perform the transformation for each
of the individual draws to obtain draws from the posterior distribution of the
transformed variable. This procedure is handled by `mutate_variables()`.

```{r mutate}
x <- mutate_variables(eight_schools_df, phi = (mu + tau)^2)
x <- subset_draws(x, c("mu", "tau", "phi"))
print(x)
```

### Renaming

To rename variables use `rename_variables()`. Here we rename the scalar `mu` to
`mean` and the vector `theta` to `alpha`.

```{r rename}
# mu is a scalar, theta is a vector
x <- rename_variables(eight_schools_df, mean = mu, alpha = theta)
variables(x)
```
In the call to `rename_variables()` above, `mu` and `theta` can be quoted or
unquoted.

It is also possible to rename individual elements of non-scalar parameters, 
for example we can rename just the first element of `alpha`:

```{r rename-element}
x <- rename_variables(x, a1 = `alpha[1]`)
variables(x)
```

### Binding

The `bind_draws()` method can be used to combine `draws` objects along different
dimensions. As an example, suppose we have several different `draws_matrix`
objects:

```{r objects-to-bind}
x1 <- draws_matrix(alpha = rnorm(5), beta = rnorm(5))
x2 <- draws_matrix(alpha = rnorm(5), beta = rnorm(5))
x3 <- draws_matrix(theta = rexp(5))
```

We can bind `x1` and `x3` together along the `'variable'` dimension to get 
a single `draws_matrix` with the variables from both `x1` and `x3`: 

```{r bind-variable}
x4 <- bind_draws(x1, x3, along = "variable")
print(x4)
```

Because `x1` and `x2` have the same variables, we can bind them along the
`'draw'` dimension to create a single `draws_matrix` with more draws:

```{r bind-draw}
x5 <- bind_draws(x1, x2, along = "draw")
print(x5)
```

As with all posterior methods, `bind_draws()` can be used with all draws formats
and depending on the format different dimensions are available to bind on. For
example, we can bind `draws_array` objects together by `iteration`, `chain`, or
`variable`, but a 2-D `draws_matrix` with the chains combined can only by bound
by `draw` and `variable`.


## Summaries and diagnostics

### summarise_draws() basic usage

Computing summaries of posterior or prior draws and convergence diagnostics for
posterior draws are some of the most common tasks when working with Bayesian
models fit using Markov Chain Monte Carlo (MCMC) methods. The posterior
package provides a flexible interface for this purpose via `summarise_draws()`
(or `summarize_draws()`), which can be passed any of the formats supported by
the package.

```{r summary}
# summarise_draws or summarize_draws
summarise_draws(eight_schools_df)
```

The result is a data frame with one row per variable and one column per
summary statistic or convergence diagnostic. The summaries `rhat`, `ess_bulk`,
and `ess_tail` are described in Vehtari et al. (2020). We can choose which
summaries to compute by passing additional arguments, either functions or names
of functions. For instance, if we only wanted the mean and its corresponding
Monte Carlo Standard Error (MCSE) we could use either of these options:

```{r summary-with-measures}
# the function mcse_mean is provided by the posterior package
s1 <- summarise_draws(eight_schools_df, "mean", "mcse_mean") 
s2 <- summarise_draws(eight_schools_df, mean, mcse_mean) 
identical(s1, s2)
print(s1)
```

### Changing column names

The column names in the output can be changed by providing the functions 
as name-value pairs, where the name is the name to use in the output and the 
value is a function name or definition. For example, here we change the names
`mean` and `sd` to `posterior_mean` and `posterior_sd`.

```{r change-summary-names}
summarise_draws(eight_schools_df, posterior_mean = mean, posterior_sd = sd)
```

### Using custom functions 

For a function to work with `summarise_draws()`, it needs to take a vector or
matrix of numeric values and return a single numeric value or a named vector of
numeric values. Additional arguments to the function can be specified in a list
passed to the `.args` argument.

```{r summary-.args}
weighted_mean <- function(x, wts) {
  sum(x * wts)/sum(wts)
}
summarise_draws(
  eight_schools_df, 
  weighted_mean, 
  .args = list(wts = rexp(ndraws(eight_schools_df)))
)
```

### Specifying functions using lambda-like syntax

It is also possible to specify a summary function using a one-sided formula that
follows the conventions supported by `rlang::as_function()`. For example, the
function

```{r standard-quantile, eval = FALSE}
function(x) quantile(x, probs = c(0.4, 0.6))
```

can be simplified to 

```{r lambda-quantile, eval = FALSE}
# for multiple arguments `.x` and `.y` can be used, see ?rlang::as_function
~quantile(., probs = c(0.4, 0.6))
```

Both can be used with `summarise_draws()` and produce the same output:

```{r lambda-syntax}
summarise_draws(eight_schools_df, function(x) quantile(x, probs = c(0.4, 0.6)))

summarise_draws(eight_schools_df, ~quantile(.x, probs = c(0.4, 0.6)))
```

See `help("as_function", "rlang")` for details on specifying these functions.

### Other diagnostics

In addition to the default diagnostic functions used by `summarise_draws()`
(`rhat()`, `ess_bulk()`, `ess_tail()`), posterior also provides additional
diagnostics like effective sample sizes and Monte Carlo standard errors for
quantiles and standard deviations, an experimental new diagnostic called R*, and
others. For a list of available diagnostics and links to their individual help
pages see `help("diagnostics", "posterior")`.

If you have suggestions for additional diagnostics that should be implemented in
posterior, please open an issue at
<https://github.com/stan-dev/posterior/issues>.


## Other methods for working with `draws` objects

In addition to the methods demonstrated in this vignette, posterior has various
other methods available for working with `draws` objects. The following is a
(potentially incomplete) list.

|**Method**|**Description**|
|:----------|:---------------|
| `order_draws()` | Order `draws` objects according to iteration and chain number |
| `repair_draws()`| Repair indices of `draws` objects so that iterations chains, and draws are continuously and consistently numbered |
|`resample_draws()` | Resample `draws` objects according to provided weights |
| `thin_draws()` | Thin `draws` objects to reduce size and autocorrelation |
| `weight_draws()`| Add weights to draws objects, with one weight per draw, for use in subsequent weighting operations |
| `extract_variable()` | Extract a vector of draws of a single variable  |
| `extract_variable_matrix()` | Extract an iterations x chains matrix of draws of a single variable |
| `merge_chains()` | Merge chains of `draws` objects into a single chain. | 
| `split_chains()` | Split chains of `draws` objects by halving the number of iterations per chain and doubling the number of chains. |

If you have suggestions for additional methods that would be useful for working
with `draws` objects, please open an issue at
<https://github.com/stan-dev/posterior/issues>.


## References

Gelman A., Carlin J. B., Stern H. S., David B. Dunson D. B., Aki Vehtari A.,
& Rubin D. B. (2013). *Bayesian Data Analysis, Third Edition*. Chapman and
Hall/CRC.

Vehtari A., Gelman A., Simpson D., Carpenter B., & Bürkner P. C. (2020).
Rank-normalization, folding, and localization: An improved Rhat for assessing
convergence of MCMC. *Bayesian Analysis*.

---
title: "rvar: The Random Variable Datatype"
author: "Matthew Kay"
date: "`r Sys.Date()`"
output: 
  html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{rvar: The Random Variable Datatype}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

This vignette describes the `rvar()` datatype, a multidimensional, sample-based
representation of random variables designed to act as much like base R arrays
as possible (e.g., by supporting many math operators and functions). This format
is also the basis of the `draws_rvars()` format.

The `rvar()` datatype is inspired by the [rv](https://cran.r-project.org/package=rv)
package and [Kerman and Gelman (2007)](https://doi.org/10.1007%2Fs11222-007-9020-4),
though with a slightly different backing format (multidimensional arrays). It is
also designed to interoperate with vectorized distributions in the
[distributional](https://pkg.mitchelloharawild.com/distributional/) package,
to be able to be used inside `data.frame()`s and `tibble()`s, and to be used
with distribution visualizations in the [ggdist](https://mjskay.github.io/ggdist/) package.

```{r setup, include = FALSE}
library(posterior)

set.seed(1234)
```

## The `rvars` datatype

The `rvar()` datatype is a wrapper around a multidimensional array where the
first dimension is the number of draws in the random variable. The most direct
way to create a random variable is to pass such an array to the `rvar()` function.

For example, to create a "scalar" `rvar`, one would pass a one-dimensional array
or a vector whose length (here `4000`) is the desired number of draws:

```{r x_rvar_rnorm}
x <- rvar(rnorm(4000, mean = 1, sd = 1))
x
```

The default display of an `rvar` shows the mean and standard deviation of
each element of the array.

We can create random vectors by adding an additional dimension beyond just the
draws dimension to the input array:

```{r x_rvar_array}
n <- 4   # length of output vector
x <- rvar(array(rnorm(4000*n, mean = 1, sd = 1), dim = c(4000, n)))
x
```

Or we can create a random matrix:

```{r x_matrix}
rows <- 4
cols <- 3
x <- rvar(array(rnorm(4000 * rows * cols, mean = 1, sd = 1), dim = c(4000, rows, cols)))
x
```

Or any array up to an arbitrary number of dimensions. The array backing an `rvar`
can be accessed (and modified, with caution) via `draws_of()`:

```{r str_draws}
str(draws_of(x))
```

While the above examples assume all draws come from a single chain, `rvar`s
can also contain samples from multiple chains. For example, if your array of
draws has iterations as the first dimension and chains as the second dimension,
you can use `with_chains = TRUE` to create an `rvar` that includes chain
information:

```{r x_matrix_with_chains}
iterations <- 1000
chains <- 4
rows <- 4
cols <- 3
x_array <- array(
  rnorm(iterations * chains * rows * cols, mean = 1, sd = 1),
  dim = c(iterations, chains, rows, cols)
)
x <- rvar(x_array, with_chains = TRUE)
x
```

Manual construction and modification of `rvar`s in this way is not always recommended
unless you need it for performance reasons: several other higher-level interfaces
to constructing and manipulating `rvar`s are described below.

### `rvar_factor` and `rvar_ordered` subtypes {#rvar_factor}

You can also use `rvar`s to represent discrete distributions, using the `rvar_factor()`
and `rvar_ordered()` subtypes. If you attempt to create an `rvar` using character
values or a `factor`, it will automatically be treated as an `rvar_factor`:

```{r rvar_factor}
x <- rvar(sample(c("a","b","c"), 4000, prob = c(0.7, 0.2, 0.1), replace = TRUE))
x
```

Numeric arrays with a `"levels"` attribute can also be passed to `rvar_factor()`.
This (along with conversion of character values) means output from 
`rstanarm::posterior_predict()` and `brms::posterior_predict()` on categorical
models can be passed directly to `rvar_factor()`.

The default display shows the mode (as returned by `modal_category()`) and
normalized entropy (`entropy()`), which is Shannon entropy scaled by the maximum
possible entropy for a distribution with the same number of levels: thus 0
means all probability is concentrated in one category, and 1 means the distribution
is uniform.

You can construct an ordered factor using `rvar_ordered()` (or by passing an
`ordered()` vector to `rvar()`):

```{r rvar_ordered}
x <- rvar_ordered(sample(c("a","b","c"), 4000, prob = c(0.7, 0.2, 0.1), replace = TRUE))
x
```

For `rvar_ordered()`, the default display is mode and dissention (`dissent()`),
which is 0 when all probability is concentrated in one category, and 1 when
the distribution is bimodal at opposite ends of the scale.

`rvar_factor`s attempt to mimic `factor()` and `rvar_ordered`s attempt to mimic
`ordered()`s, by implementing factor-specific functions like `levels()`. Comparison
operations are also implemented where valid. For example, in `x` as defined above, 
approximately 90% of draws should be less than `"b"` (which means the `"a"` and
`"b"` levels):

```{r x_lte_b}
x <= "b"
```

`rvar`s also supply an implementation of `match()` and `%in%`, which can be
especially useful with `rvar_factor`s:

```{r x_in_ac}
x %in% c("a", "c")
```

## The `draws_rvars` datatype

The `draws_rvars()` datatype, like all `draws` datatypes in posterior, contains
multiple variables in a joint sample from some distribution (e.g. a posterior or 
prior distribution).

You can construct `draws_rvars()` objects directly using the `draws_rvars()`
function. The input `rvar`s must have the same number of chains and iterations,
but can otherwise have different shapes:

```{r draws_rvars}
d <- draws_rvars(x = x, y = rvar(rnorm(iterations * chains), nchains = 4))
d
```

Existing objects can also be converted to the `draws_rvars()` format
using `as_draws_rvars()`. Below is the `example_draws("multi_normal")` dataset
converted into the `draws_rvars()` format. This dataset has 100 iterations from 
4 chains from the posterior of a a 3-dimensional multivariate normal model. The 
`mu` variable is a mean vector of length 3 and the `Sigma` variable is a $3 \times 3$
covariance matrix:

```{r post}
post <- as_draws_rvars(example_draws("multi_normal"))
post
```

The `draws_rvars()` datatype works much the same way that other `draws` formats
do; see the main package vignette at `vignette("posterior")` for an introduction 
to `draws` objects. One difference is that `draws_rvars` counts variables
differently, because it allows variables to be multidimensional. For example,
the `post` object above contains two variables, `mu` and `Sigma`:

```{r variables_draws_rvars}
variables(post)
```

But converted to a `draws_list()`, it contains one variable for each
combination of the dimensions of its variables:

```{r variables_draws_list}
variables(as_draws_list(post))
```

## Math with `rvar`s

The `rvar()` datatype implements most math operations, including basic arithmetic,
functions in the *Math* and *Summary* groups, like `log()` and `exp()` (see 
`help("groupGeneric")` for a list), and more. Binary operators can be performed
between multiple `rvar`s or between `rvar`s and `numeric`s. A simple example:

```{r mu_plus_1}
mu <- post$mu
Sigma <- post$Sigma

mu + 1
```

Matrix multiplication is also implemented (using a tensor product under the hood).
In R < 4.3, the normal matrix multiplication operator (`%*%`) cannot be properly
implemented for S3 datatypes, so `rvar` uses `%**%` instead. In R ≥ 4.3, which 
does support matrix multiplication for S3 datatypes, you can use `%*%` to 
matrix-multiply `rvar`s.

A trivial example:

```{r matrix_mult}
Sigma %**% diag(1:3)
```

The set of mathematical functions and operators supported by `rvar`s
includes:

| Group                       | Functions and operators |
|:----------------------------|:------------------------|
| Arithmetic operators        | `+`, `-`, `*`, `/`, `^`, `%%`, `%/%` |
| Logical operators           | `&`, `|`, `!` |
| Comparison operators        | `==`, `!=`, `<`, `<=`, `>=`, `>` |
| Value matching              | `match()`, `%in%` |
| Matrix multiplication       | `%**%`, `%*%` (R ≥ 4.3 only) |
| Basic functions             | `abs()`, `sign()`<br>`sqrt()`<br>`floor()`, `ceiling()`, `trunc()`, `round()`, `signif()` |
| Logarithms and exponentials | `exp()`, `expm1()`<br>`log()`, `log10()`, `log2()`, `log1p()` |
| Trigonometric functions     | `cos()`, `sin()`, `tan()`<br>`cospi()`, `sinpi()`, `tanpi()`<br>`acos()`, `asin()`, `atan()`|
| Hyperbolic functions        | `cosh()`, `sinh()`, `tanh()`<br>`acosh()`, `asinh()`, `atanh()` |
| Special functions           | `lgamma()`, `gamma()`, `digamma()`, `trigamma()` |
| Cumulative functions        | `cumsum()`, `cumprod()`, `cummax()`, `cummin()` |
| Array transposition         | `t()`, `aperm()` |
| Matrix decomposition        | `chol()` |
| Matrix diagonals            | `diag()` |

## Expectations and summary functions

The `E()` function is an alias of `mean()`, producing means
within each cell of an `rvar`. For example, given `mu`:

```{r mu}
mu
```

We can get the expectation of each cell of `mu`:

```{r E_mu}
E(mu)
```

Expectations of logical expressions are probabilities, and can be computed either
with `E()` / `mean()` or with `Pr()`. `Pr()` is provided as notational sugar, but
also checks that the input is a logical variable before taking the mean:

```{r Pr}
Pr(mu > 0)
```

More generally, the `rvar` data type provides two types of summary functions: 

1. Summary functions that mimic base-R vector summary functions, except applied to `rvar` vectors. These apply their summaries **over** elements of the input vectors **within** each draw, generally returning an `rvar` of length 1. These functions are prefixed with `rvar_` as a reminder that they return `rvar`s. Here is an example of `rvar_mean()`:
  
    ```{r rvar_mean_mu}
    rvar_mean(mu)
    ```

2. Summary functions that summarise **within** elements of input vectors and **over** draws. These summary functions generally return base arrays (`numeric` or `logical`) of the same shape as the input `rvar`, and are especially useful for diagnostic summaries. These summary functions are not prefixed with `rvar_` as they do not return `rvar`s. Here is an example of `mean()`:

    ```{r mean_mu}
    mean(mu)
    ```

   You should expect the same values from these functions (though in a different shape) when you use them with `summarise_draws()`, for example:

    ```{r summarise_draws_mu_mean}
    summarise_draws(mu, mean)
    ```

Here is a table of both types of summary functions:

|                          | 1. Summarise *within* draws,<br>*over* elements | 2. Summarise *over* draws,<br>*within* elements<br> |
|--------------------------|:------------------------------------------------|:----------------------------------------------------|
| **Output format**<br>of `res = f(x)` | `rvar` of length 1 | `array` of same shape as input `rvar` |
| **Help page**            | `help("rvar-summaries-within-draws")` | `help("rvar-summaries-over-draws")` |
| Numeric summaries        | `rvar_median()`<br>`rvar_sum()`, `rvar_prod()`<br>`rvar_min()`, `rvar_max()`| `median()`<br>`sum()`, `prod()`<br>`min()`, `max()`|
| Mean                     | `rvar_mean()`<br>*N/A* | `mean()`, `E()`<br>`Pr()`: enforces that input is `logical` |
| Spread                   | `rvar_sd()` <br>`rvar_var()` <br> `rvar_mad()` | `sd()`<br>`var()`, `variance()`<br> `mad()`|
| Range                    | `rvar_range()`<br>**Note:** `length(res) == 2` | `range()`<br>**Note:** `dim(res) == c(2, dim(x))` |
| Quantiles                | `rvar_quantile()`<br>**Note:** `length(res) == length(probs)` | `quantile()`<br>**Note:** `dim(res) == c(length(probs), dim(x))` |
| Logical summaries        | `rvar_all()`, `rvar_any()` | `all()`, `any()` |
| Special value predicates | `rvar_is_finite()`<br>`rvar_is_infinite()`<br>`rvar_is_nan()`<br>`rvar_is_na()`<br>**Note:** `dim(res) == dim(x)`. These functions act within draws but do not summarise over elements.        | `is.finite()`<br>`is.infinite()`<br>`is.nan()`<br>`is.na()`<br>**Note:** `res[i] == TRUE` if `x[i]` has any draws matching predicate (except for `is.finite()`, where all draws in `x[i]` must match) |
| Diagnostics              | *N/A* | `ess_basic()`, `ess_bulk()`, `ess_quantile()`, `ess_sd()`, `ess_tail()`,<br>`mcse_mean()`, `mcse_quantile()`, `mcse_sd()`<br>`rhat()`, `rhat_basic()`|

## Constants

Constant `rvar`s can be constructed by converting numeric vectors or arrays into `rvar`s
using `as_rvar()`, which will return an `rvar` with one draw and the same dimensions as its
input:

```{r const}
const <- as_rvar(1:3)
const
```

While normally `rvar`s must have the same number of draws to be used in the
same expression, `rvar`s with one draw are treated like constants, and 
can be combined with other `rvar`s:

```{r mu_plus_const}
mu + const
```

## Using existing R functions and expressions with `rvar`s

While `rvar`s attempt to emulate as much of the functionality of base R arrays
as possible, there are situations in which an existing R function may not 
work directly with an `rvar`. There are several approaches to solving this
problem.

For example, say you wish to generate samples from the
following expression for $\mu$, $\sigma$, and $x$:

$$
\begin{align}
\left[\begin{matrix}\mu_1 \\ \vdots \\ \mu_4 \end{matrix}\right] &\sim \textrm{Normal}\left(\left[\begin{matrix}1 \\ \vdots \\ 4 \end{matrix}\right],1\right)\\
\sigma &\sim \textrm{Gamma}(1,1)\\
\left[\begin{matrix}x_1 \\ \vdots \\ x_4 \end{matrix}\right] &\sim \textrm{Normal}\left(\left[\begin{matrix}\mu_1 \\ \vdots \\ \mu_4 \end{matrix}\right], \sigma\right)
\end{align}
$$

There are three different approaches you might take to doing this: converting
existing R functions with `rfun()`, executing expressions of
random variables with `rdo()`, or evaluating random number generator functions
using `rvar_rng()`.

### Converting functions with `rfun()`

The `rfun()` wrapper converts an existing R function into a new function 
that `rvar`s can be passed to it as arguments, and which
will return `rvar`s. 
We can use `rfun()` to convert the base `rnorm()` and `rgamma()` 
random number generating functions into functions that accept and return `rvar`s:

```{r rfun_defs}
rvar_norm <- rfun(rnorm)
rvar_gamma <- rfun(rgamma)
```

Then we can translate the above example into code using those functions:

```{r rfun_ex}
mu <- rvar_norm(4, mean = 1:4, sd = 1)
sigma <- rvar_gamma(1, shape = 1, rate = 1)
x <- rvar_norm(4, mu, sigma)
x
```

While `rfun()`-converted functions work well for prototyping, they will generally
speaking be slower than functions designed specifically for `rvar`s. Thus, you
may find you need to adopt other strategies (like `rvar_rng()`, described below;
or re-writing functions to support `rvar` directly using math operators and/or the
`draws_of()` function).

### Evaluating expressions with `rdo()`

An alternative to `rfun()` is to use `rdo()`, which can be passed nearly-arbitrary
R expressions. The expression will be executed
multiple times to construct an `rvar`. E.g., we can write an expression for
`mu` like in the above example:

```{r mu_rdo}
mu <- rdo(rnorm(4, mean = 1:4, sd = 1))
mu
```

We can also control the number of draws using the `ndraws` argument:

```{r mu_rdo_ndraws}
mu <- rdo(rnorm(4, mean = 1:4, sd = 1), ndraws = 1000)
mu
```

`rdo()` expressions can also contain other `rvar`s, so long as all `rvar`s in
the expression have the same number of draws. Thus, we can re-write the
example above that used `rfun()` as follows:

```{r rdo_ex}
mu <- rdo(rnorm(4, mean = 1:4, sd = 1))
sigma <- rdo(rgamma(1, shape = 1, rate = 1))
x <- rdo(rnorm(4, mu, sigma))
x
```

Like `rfun()`, `rdo()` is not necessarily fast, so you may find it more useful
for prototyping than production code.

### Evaluating random number generators with `rvar_rng()`

`rvar_rng()` is an alternative to `rfun()`/`rdo()` designed specifically to work
with random number generating functions that follow the typical API of such
functions in base R. Such functions, like `rnorm()`, `rgamma()`, `rbinom()`, etc 
all following this interface:

- They have a first argument, `n`, giving the number of draws to take from the distribution.
- Their arguments for distribution parameters (`mean`, `sd`, `shape`, `rate`, etc.) are vectorized.
- They return a single vector of length `n`, representing `n` draws from the distribution.

You can use any function with this interface with `rvar_rng()`, and it will adapt
it to be able to take `rvar` arguments and return an `rvar`, as follows:

```{r rvar_r_ex}
mu <- rvar_rng(rnorm, 4, mean = 1:4, sd = 1)
sigma <- rvar_rng(rgamma, 1, shape = 1, rate = 1)
x <- rvar_rng(rnorm, 4, mu, sigma)
x
```

In contrast to the `rfun()` and `rdo()` examples above, `rvar_rng()` takes 
advantage of the existing vectorization of the underlying random number generating
function to execute quickly.

## Broadcasting

Broadcasting for `rvar`s does not follow R's vector recycling rules. Instead,
when two variables with different dimensions are being used with basic arithmetic
functions, dimensions are added until both variables have the same number of dimensions.
If two variables $x$ and $y$ differ on the length of dimension $d$, they can
be broadcast to the same size so long as one of the variables has dimension $d$
of size 1. Then that variable will be broadcast up to the same size as the
other variable along that dimension. If two variables disagree on the size of 
a dimension and neither has size 1, it is an error.

For example, consider this random matrix:

```{r X_matrix}
X <- rdo(rnorm(12, 1:12), dim = c(4,3))
X
```

And this vector of length 3:

```{r y_vector}
y <- rdo(rnorm(3, 3:1))
y
```

If we attempt to add `X` and `y`, it will produce an error as vectors are 
by default treated as column vectors, and `y` has length 3 while columns of 
`X` have length 4:

```{r X_plus_y, error = TRUE}
X + y
```

By contrast, R arrays of the same shape will simply recycle `y` until it is the
same length as `X` (regardless of the dimensions). Thus will produce a result,
though likely not the intended result:

```{r mean_X_plus_y}
mean(X) + mean(y)
```

On the other hand, if y were a row vector...

```{r row_y}
row_y = t(y)
row_y
```

...it would have the same number of
columns as `X` and contain only one row, so it can be broadcast along rows of `X`:

```{r X_plus_row_y}
X + row_y
```

## Slicing and conditionals

The `[[` and `[` operators implement all of the base array slicing operations, 
including numeric, character, and logical indices, as well as slicing arrays
using a matrix of indices with `[`. The main difference between `rvar` slicing
and base array slicing is that `rvar`s default to `drop = FALSE`; i.e. they
retain all dimensions of the original `rvar` array. For a complete list of 
`rvar` slicing types, see `help("rvar-slice")`.

In addition to the base slicing operations, `rvar` also implements three
slicing/conditioning methods that allow you to use other `rvar`s to define
a slice. 

To demonstrate these operations, consider an `rvar` vector of two components:

```{r}
component = rvar_rng(rnorm, 2, mean = c(1, 5))
component
```

Perhaps we want to create a mixture of these two components, `mixture`, with a
mixing proportion of 0.75. We could create an index, `i`, that is a random
variable indicating which component (1 or 2) determines the value of `mixture`
on each draw:

```{r}
i = rvar_rng(rbinom, 1, size = 1, p = 0.75) + 1L
i
```

We can use several different approaches to create the mixture distribution

### Subsetting `rvar`s by draw: `x[<logical rvar>]`

A slice `x[i]` where `i` is a scalar logical `rvar` returns a new rvar with the same
shape as `x`, but containing only those draws where `i` is `TRUE`. Thus,
we can use `i == 2` to select draws from the second component and overwrite
them in the first component, creating the mixture distribution:

```{r}
mixture = component[[1]]
mixture[i == 2] = component[[2]][i == 2]
mixture
```

The resulting mixture looks like this:

```{r mixture, eval = requireNamespace("ggplot2", quietly = TRUE) && requireNamespace("ggdist", quietly = TRUE)}
library(ggplot2)

ggplot() + ggdist::stat_slab(aes(xdist = mixture))
```

See `vignette("slabinterval", package = "ggdist")` for more examples of 
visualizing distribution-type objects, including `rvar`s.

### Conditionals using `rvar_ifelse()`

You could create the same mixture using `rvar_ifelse(test, yes, no)`, which 
broadcasts `test`, `yes`, and `no` to the same shape, then returns a new
`rvar` containing draws from `yes` where `test == TRUE` and draws from `no`
where `test == FALSE`.

Thus, we can create the mixture as follows:

```{r}
x = rvar_ifelse(i == 1, component[[1]], component[[2]])
x
```

### Selecting different elements in each draw: `x[[<numeric rvar>]]`

The slice `x[[i]]`, where `i` is a scalar numeric rvar, generalizes indexing when 
`i` is a scalar numeric. Within each draw of `x`, it selects the element of `x`
corresponding to the value of `i` within that same draw.

Thus, since `i` in our example is a scalar numeric `rvar` whose values are 
either `1` or `2` within each draw, you can use it as an index directly on 
`component` to create the mixture:

```{r}
x = component[[i]]
x
```

This approach is also nice because it generalizes easily to more than two components.

## Applying functions over `rvar`s

The `rvar` data type supplies an implementation of `as.list()`, which should give 
compatibility with the base R family of functions for applying functions over arrays:
`apply()`, `lapply()`, `vapply()`, `sapply()`, etc. You can also manually use
`as.list()` to convert an `rvar` into a list along its first dimension, which
may be necessary for compatibility with some functions (like `purrr:map()`).

For example, given this multidimensional `rvar`...

```{r multidim_array}
set.seed(3456)
x <- rvar_rng(rnorm, 24, mean = 1:24)
dim(x) <- c(2,3,4)
x
```

... you can apply functions along the margins using `apply()` (here, a silly example):

```{r apply}
apply(x, c(1,2), length)
```

One exception is that while `apply()` will work with an `rvar` input if your function
returns base data types (like numerics), it will not give you simplified `rvar`
arrays if your function returns an `rvar`. Thus, we supply the `rvar_apply()`
function, which takes in either base arrays or `rvar` arrays and returns `rvar`
arrays, and which also uses the `rvar` broadcasting rules to combine the
results of the applied function.

For example, you can use `rvar_apply()` with `rvar_mean()` to compute the distributions of
means along one margin of an array:

```{r rvar_apply_one_dim}
rvar_apply(x, 1, rvar_mean)
```

Or along multiple dimensions:

```{r rvar_apply_multi_dim}
rvar_apply(x, c(2,3), rvar_mean)
```

## Looping over `draws` and `rvar`s

The `rvar` datatype is also used in `for_each_draw()`, which allows you to loop
over draws in a `draws` object or an `rvar`. `for_each_draw(x, expr)` converts `x`
into a `draws_rvars()` object, then loops over each draw of `x`, executing the
provided expression, `expr`. The expression can use the variables in `x` as if
they were regular R arrays.

One application of `for_each_draw()` is in constructing base-R plots of individual
draws (for `ggplot2`-based plotting of `rvar`s, see the next section and the
[ggdist](https://mjskay.github.io/ggdist/) package). For example, it can be used
to construct a parallel coordinates plot:

```{r eight_schools_parcoord, fig.width = 6, fig.height = 4}
eight_schools <- as_draws_rvars(example_draws())

plot(1, type = "n",
  xlim = c(1, length(eight_schools$theta)),
  ylim = range(range(eight_schools$theta)),
  xlab = "school", ylab = "theta"
)

# use for_each_draw() to make a parallel coordinates plot of all draws
# of eight_schools$theta
for_each_draw(eight_schools, {
  lines(seq_along(theta), theta, col = rgb(1, 0, 0, 0.05))
})

# add means and 90% intervals
lines(seq_along(eight_schools$theta), mean(eight_schools$theta))
with(summarise_draws(eight_schools$theta), 
  segments(seq_along(eight_schools$theta), y0 = q5, y1 = q95)
)
```

As `for_each_draw()` will be slower than most other ways of manipulating draws, 
this function should generally not be used unless needed.


## Using `rvar`s in data frames and in ggplot2

`rvar`s can be used as columns in `data.frame()` or `tibble()` objects:

```{r data_frame_with_y}
df <- data.frame(group = c("a","b","c","d"), mu)
df
```

This makes them convenient for adding predictions to a data frame alongside
the data used to generate the predictions. `rvar`s can then be visualized with ggplot2
by passing them to the `xdist` and `ydist` aesthetics of the `stat_...` family of
geometries in the [ggdist](https://mjskay.github.io/ggdist/) package, 
such as `stat_halfeye()`, `stat_lineribbon()`, and `stat_dotsinterval()`.
For example:

```{r data_frame_plot, eval = requireNamespace("ggplot2", quietly = TRUE) && requireNamespace("ggdist", quietly = TRUE)}
library(ggplot2)
library(ggdist)

ggplot(df) +
  stat_halfeye(aes(y = group, xdist = mu))
```

See `vignette("slabinterval", package = "ggdist")` or 
`vignette("tidy-posterior", package = "tidybayes")` for more examples.

---
title: "Pareto-khat diagnostics"
author: "Aki Vehtari"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{Pareto-khat diagnostics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction 

The paper

* Aki Vehtari, Daniel Simpson, Andrew Gelman, Yuling Yao, and
  Jonah Gabry (2024). Pareto smoothed importance sampling.
  *Journal of Machine Learning Research*, 25(72):1-58.

presents Pareto smoothed importance sampling, but also
Pareto-$\hat{k}$ diagnostic that can be used when estimating any
expectation based on finite sample. This vignette illustrates the use of
these diagnostics.
The individual diagnostic functions are `pareto_khat()`, `pareto_min_ss()`, `pareto_convergence_rate()` and `pareto_khat_threshold()`. The function `pareto_diags()` will return all of these.

Additionally, the `pareto_smooth()` function can be used to transform draws by smoothing the tail(s).
## Example

```{r setup}
library(posterior)
library(dplyr)
options(pillar.neg = FALSE, pillar.subtle=FALSE, pillar.sigfig=2)
```

### Simulated data

Generate `xn` a simulated MCMC sample with 4 chains each with 1000
iterations using AR process with marginal normal(0,1)

```{r simulate-data-1}
N <- 1000
phi <- 0.3
set.seed(6534)
dr <- array(data=replicate(4,as.numeric(arima.sim(n = N,
                                                list(ar = c(phi)),
                                                sd = sqrt((1-phi^2))))),
         dim=c(N,4,1)) %>%
  as_draws_df() %>%
  set_variables('xn')
```

Transform `xn` via cdf-inverse-cdf so that we have variables that
have marginally distributions $t_3$, $t_{2.5}$, $t_2$, $t_{1.5}$,
and $t_1$. These all have thick tails. In addition $t_2$,
$t_{1.5}$, and $t_1$ have infinite variance, and $t_1$ (aka Cauchy)
has infinite mean. 

```{r simulate-data-2}
drt <- dr %>%
  mutate_variables(xt3=qt(pnorm(xn), df=3),
                   xt2_5=qt(pnorm(xn), df=2.5),
                   xt2=qt(pnorm(xn), df=2),
                   xt1_5=qt(pnorm(xn), df=1.5),
                   xt1=qt(pnorm(xn), df=1))
```

### MCMC convergence diagnostics

We examine the draws with the default `summarise_draws()`.

```{r summarise_draws}
drt %>%
  summarise_draws()
```

All the usual convergence diagnostics $\widehat{R}$, Bulk-ESS, and
Tail-ESS look good, which is fine as they have been designed to
work also with infinite variance (Vehtari et al., 2020).

If these variables would present variables of interest for which
we would like to estimate means, we would be also interested in
Monte Carlo standard error (MCSE, see case study [How many
iterations to run and how many digits to
report](https://users.aalto.fi/~ave/casestudies/Digits/digits.html)).

```{r summarise_draws-mcse}
drt %>%
  summarise_draws(mean, sd, mcse_mean, ess_bulk, ess_basic)
```

Here MCSE for mean is based on standard deviation and Basic-ESS,
but these assume finite variance. We did sample also from
distributions with infinite variance, but given a finite sample size, the
empirical variance estimates are always finite, and thus we get
overoptimistic MCSE.

## Pareto-$\hat{k}$

To diagnose whether our variables of interest may have infinite
variance and even infinite mean, we can use Pareto-$\hat{k}$
diagnostic.

```{r summarise_draws-pareto_khat}
drt %>%
  summarise_draws(mean, sd, mcse_mean, ess_basic, pareto_khat)
```

$\hat{k} \leq 0$ indicates that all moments exist, and the inverse
of positive $\hat{k}$ tells estimate for the number of finite (fractional)
moments. Thus, $\hat{k}\geq 1/2$ indicates infinite variance,
and $\hat{k}\geq 1$ indicates infinite mean. Sometimes very thick
distribution tails may affect also sampling, but assuming sampling
did go well, and we would be interested only in quantiles, infinite
variance and mean are not a problem. But if we are interested in mean,
then we need to care about the number of (fractional) moments. Here we
see $\hat{k} \geq 1/2$ for $t_2$, $t_{1.5}$, and $t_{1}$, and
we should not trust their `mcse_mean` values. Without trustworthy MCSE
estimate we don't have good estimate of how accurate the mean estimate is.
Furthermore, as $\hat{k} \geq 1$ for $t_{1}$, the mean is not finite and
the mean estimate is not valid.

## Pareto smoothing

If we really do need those mean estimates, we can improve
trustworthiness by Pareto smoothing,  which replaces
extreme tail draws with expected ordered statistics of Pareto
distribution fitted to the tails of the distribution. Pareto
smoothed mean estimate (computed using Pareto smoothed draws) has
finite variance with a cost of some bias which we know when it is
negligible. As a thumb rule when $\hat{k}<0.7$, the bias is
negligible.

We do Pareto smoothing for all the variables.

```{r pareto_smooth}
drts <- drt %>% 
  mutate_variables(xt3_s=pareto_smooth(xt3),
                   xt2_5_s=pareto_smooth(xt2_5),
                   xt2_s=pareto_smooth(xt2),
                   xt1_5_s=pareto_smooth(xt1_5),
                   xt1_s=pareto_smooth(xt1)) %>%
  subset_draws(variable="_s", regex=TRUE)
```

Now the `mcse_mean` values are more trustworthy when $\hat{k} < 0.7$.
When $\hat{k}>0.7$ both bias and variance grow so fast that Pareto smoothing
rarely helps (see more details in the paper).

```{r summarise_draws-pareto_khat-2}
drts %>%
  summarise_draws(mean, mcse_mean, ess_basic, pareto_khat)
```


## Minimum sample size required

The bias and variance depend on the sample size, and we can
use additional diagnostic `min_ss` which tells the minimum sample size needed
so that `mcse_mean` can be trusted.

```{r summarise_draws-min_ss}
drt %>%
  summarise_draws(mean, mcse_mean, ess_basic,
                  pareto_khat, min_ss=pareto_min_ss)
```

Here required `min_ss` is smaller than `ess_basic` for all except $t_1$, for
which there is no hope.

## Convergence rate

Given finite variance, the central limit theorem states that to halve
MCSE we need four times bigger sample size. With Pareto smoothing,
we can go further, but the convergence rate decreases when $\hat{k}$ increases.

```{r summarise_draws-conv_rate}
drt %>%
  summarise_draws(mean, mcse_mean, ess_basic,
                  pareto_khat, min_ss=pareto_min_ss,
                  conv_rate=pareto_convergence_rate)
```

We see that with $t_2$, $t_{1.5}$, and $t_1$  we need $4^{1/0.86}\approx 5$,
$4^{1/0.60}\approx 10$, and $4^{1/0}\approx \infty$ times bigger sample sizes to
halve MCSE for mean.

## Pareto-$\hat{k}$-threshold

The final Pareto diagnostic, $\hat{k}$-threshold, is useful for smaller sample sizes. Here we select only 100 iterations per chain to get total of 400 draws.

```{r summarise_draws-khat_threshold}
drt %>%
  subset_draws(iteration=1:100) %>%
  summarise_draws(mean, mcse_mean, ess_basic,
                  pareto_khat, min_ss=pareto_min_ss,
                  khat_thres=pareto_khat_threshold,
                  conv_rate=pareto_convergence_rate)
```

With only 400 draws, we can trust the Pareto smoothed result only when
$\hat{k}<0.62$. For $t_{1.5}$ $\hat{k}\approx 0.64$, and `min_ss` reveals
we would probably need more than 560 draws to be on the safe side.

## Pareto diagnostics

We can get all these diagnostics with `pareto_diags()`, and it's
easy to use it also for derived quantities.

```{r summarise_draws-pareto_diags}
drt %>%
  mutate_variables(xt2_5_sq=xt2_5^2) %>%
  subset_draws(variable="xt2_5_sq") %>%
  summarise_draws(mean, mcse_mean,
                  pareto_diags)
```

## Discussion

All these diagnostics are presented in Section 3 and summarized in
Table 1 in PSIS paper (Vehtari et al., 2024).

If you don't need to estimate means of thick tailed distributions,
and there are no sampling issues due to thick tails, then you don't
need to check existence of finite variance, and thus there is no
need to check Pareto-$\hat{k}$ for all the parameters and derived
quantities.

It is possible that the distribution has finite variance, but
pre-asymptotically given a finite sample size the behavior can be
similar to infinite variance. Thus the diagnostic is useful even in
cases where theory guarantees finite variance.

## Reference

Vehtari, A., Simpson, D., Gelman, A., Yao, Y., & Gabry, J. (2024).
Pareto smoothed importance sampling.
*Journal of Machine Learning Research*, 25(72):1-58.

Vehtari A., Gelman A., Simpson D., Carpenter B., & Bürkner P. C. (2020).
Rank-normalization, folding, and localization: An improved Rhat for assessing
convergence of MCMC. *Bayesian Analysis*, 16(2):667-718.
	    

---
title: "rvar: The Random Variable Datatype"
author: "Matthew Kay"
date: "`r Sys.Date()`"
output: 
  html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{rvar: The Random Variable Datatype}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

## Introduction

This vignette describes the `rvar()` datatype, a multidimensional, sample-based
representation of random variables designed to act as much like base R arrays
as possible (e.g., by supporting many math operators and functions). This format
is also the basis of the `draws_rvars()` format.

The `rvar()` datatype is inspired by the [rv](https://cran.r-project.org/package=rv)
package and [Kerman and Gelman (2007)](https://doi.org/10.1007%2Fs11222-007-9020-4),
though with a slightly different backing format (multidimensional arrays). It is
also designed to interoperate with vectorized distributions in the
[distributional](https://pkg.mitchelloharawild.com/distributional/) package,
to be able to be used inside `data.frame()`s and `tibble()`s, and to be used
with distribution visualizations in the [ggdist](https://mjskay.github.io/ggdist/) package.

```{r setup, include = FALSE}
library(posterior)

set.seed(1234)
```

## The `rvars` datatype

The `rvar()` datatype is a wrapper around a multidimensional array where the
first dimension is the number of draws in the random variable. The most direct
way to create a random variable is to pass such an array to the `rvar()` function.

For example, to create a "scalar" `rvar`, one would pass a one-dimensional array
or a vector whose length (here `4000`) is the desired number of draws:

```{r x_rvar_rnorm}
x <- rvar(rnorm(4000, mean = 1, sd = 1))
x
```

The default display of an `rvar` shows the mean and standard deviation of
each element of the array.

We can create random vectors by adding an additional dimension beyond just the
draws dimension to the input array:

```{r x_rvar_array}
n <- 4   # length of output vector
x <- rvar(array(rnorm(4000*n, mean = 1, sd = 1), dim = c(4000, n)))
x
```

Or we can create a random matrix:

```{r x_matrix}
rows <- 4
cols <- 3
x <- rvar(array(rnorm(4000 * rows * cols, mean = 1, sd = 1), dim = c(4000, rows, cols)))
x
```

Or any array up to an arbitrary number of dimensions. The array backing an `rvar`
can be accessed (and modified, with caution) via `draws_of()`:

```{r str_draws}
str(draws_of(x))
```

While the above examples assume all draws come from a single chain, `rvar`s
can also contain samples from multiple chains. For example, if your array of
draws has iterations as the first dimension and chains as the second dimension,
you can use `with_chains = TRUE` to create an `rvar` that includes chain
information:

```{r x_matrix_with_chains}
iterations <- 1000
chains <- 4
rows <- 4
cols <- 3
x_array <- array(
  rnorm(iterations * chains * rows * cols, mean = 1, sd = 1),
  dim = c(iterations, chains, rows, cols)
)
x <- rvar(x_array, with_chains = TRUE)
x
```

Manual construction and modification of `rvar`s in this way is not always recommended
unless you need it for performance reasons: several other higher-level interfaces
to constructing and manipulating `rvar`s are described below.

### `rvar_factor` and `rvar_ordered` subtypes {#rvar_factor}

You can also use `rvar`s to represent discrete distributions, using the `rvar_factor()`
and `rvar_ordered()` subtypes. If you attempt to create an `rvar` using character
values or a `factor`, it will automatically be treated as an `rvar_factor`:

```{r rvar_factor}
x <- rvar(sample(c("a","b","c"), 4000, prob = c(0.7, 0.2, 0.1), replace = TRUE))
x
```

Numeric arrays with a `"levels"` attribute can also be passed to `rvar_factor()`.
This (along with conversion of character values) means output from 
`rstanarm::posterior_predict()` and `brms::posterior_predict()` on categorical
models can be passed directly to `rvar_factor()`.

The default display shows the mode (as returned by `modal_category()`) and
normalized entropy (`entropy()`), which is Shannon entropy scaled by the maximum
possible entropy for a distribution with the same number of levels: thus 0
means all probability is concentrated in one category, and 1 means the distribution
is uniform.

You can construct an ordered factor using `rvar_ordered()` (or by passing an
`ordered()` vector to `rvar()`):

```{r rvar_ordered}
x <- rvar_ordered(sample(c("a","b","c"), 4000, prob = c(0.7, 0.2, 0.1), replace = TRUE))
x
```

For `rvar_ordered()`, the default display is mode and dissention (`dissent()`),
which is 0 when all probability is concentrated in one category, and 1 when
the distribution is bimodal at opposite ends of the scale.

`rvar_factor`s attempt to mimic `factor()` and `rvar_ordered`s attempt to mimic
`ordered()`s, by implementing factor-specific functions like `levels()`. Comparison
operations are also implemented where valid. For example, in `x` as defined above, 
approximately 90% of draws should be less than `"b"` (which means the `"a"` and
`"b"` levels):

```{r x_lte_b}
x <= "b"
```

`rvar`s also supply an implementation of `match()` and `%in%`, which can be
especially useful with `rvar_factor`s:

```{r x_in_ac}
x %in% c("a", "c")
```

## The `draws_rvars` datatype

The `draws_rvars()` datatype, like all `draws` datatypes in posterior, contains
multiple variables in a joint sample from some distribution (e.g. a posterior or 
prior distribution).

You can construct `draws_rvars()` objects directly using the `draws_rvars()`
function. The input `rvar`s must have the same number of chains and iterations,
but can otherwise have different shapes:

```{r draws_rvars}
d <- draws_rvars(x = x, y = rvar(rnorm(iterations * chains), nchains = 4))
d
```

Existing objects can also be converted to the `draws_rvars()` format
using `as_draws_rvars()`. Below is the `example_draws("multi_normal")` dataset
converted into the `draws_rvars()` format. This dataset has 100 iterations from 
4 chains from the posterior of a a 3-dimensional multivariate normal model. The 
`mu` variable is a mean vector of length 3 and the `Sigma` variable is a $3 \times 3$
covariance matrix:

```{r post}
post <- as_draws_rvars(example_draws("multi_normal"))
post
```

The `draws_rvars()` datatype works much the same way that other `draws` formats
do; see the main package vignette at `vignette("posterior")` for an introduction 
to `draws` objects. One difference is that `draws_rvars` counts variables
differently, because it allows variables to be multidimensional. For example,
the `post` object above contains two variables, `mu` and `Sigma`:

```{r variables_draws_rvars}
variables(post)
```

But converted to a `draws_list()`, it contains one variable for each
combination of the dimensions of its variables:

```{r variables_draws_list}
variables(as_draws_list(post))
```

## Math with `rvar`s

The `rvar()` datatype implements most math operations, including basic arithmetic,
functions in the *Math* and *Summary* groups, like `log()` and `exp()` (see 
`help("groupGeneric")` for a list), and more. Binary operators can be performed
between multiple `rvar`s or between `rvar`s and `numeric`s. A simple example:

```{r mu_plus_1}
mu <- post$mu
Sigma <- post$Sigma

mu + 1
```

Matrix multiplication is also implemented (using a tensor product under the hood).
Because the normal matrix multiplication operator in R (`%*%`) cannot be properly
implemented for S3 datatypes, `rvar` uses `%**%` instead. A trivial example:

```{r matrix_mult}
Sigma %**% diag(1:3)
```

The set of mathematical functions and operators supported by `rvar`s
includes:

| Group                       | Functions and operators |
|:----------------------------|:------------------------|
| Arithmetic operators        | `+`, `-`, `*`, `/`, `^`, `%%`, `%/%` |
| Logical operators           | `&`, `|`, `!` |
| Comparison operators        | `==`, `!=`, `<`, `<=`, `>=`, `>` |
| Value matching              | `match()`, `%in%` |
| Matrix multiplication       | `%**%` |
| Basic functions             | `abs()`, `sign()`<br>`sqrt()`<br>`floor()`, `ceiling()`, `trunc()`, `round()`, `signif()` |
| Logarithms and exponentials | `exp()`, `expm1()`<br>`log()`, `log10()`, `log2()`, `log1p()` |
| Trigonometric functions     | `cos()`, `sin()`, `tan()`<br>`cospi()`, `sinpi()`, `tanpi()`<br>`acos()`, `asin()`, `atan()`|
| Hyperbolic functions        | `cosh()`, `sinh()`, `tanh()`<br>`acosh()`, `asinh()`, `atanh()` |
| Special functions           | `lgamma()`, `gamma()`, `digamma()`, `trigamma()` |
| Cumulative functions        | `cumsum()`, `cumprod()`, `cummax()`, `cummin()` |
| Array transposition         | `t()`, `aperm()` |
| Matrix decomposition        | `chol()` |
| Matrix diagonals            | `diag()` |

## Expectations and summary functions

The `E()` function is an alias of `mean()`, producing means
within each cell of an `rvar`. For example, given `mu`:

```{r mu}
mu
```

We can get the expectation of each cell of `mu`:

```{r E_mu}
E(mu)
```

Expectations of logical expressions are probabilities, and can be computed either
with `E()` / `mean()` or with `Pr()`. `Pr()` is provided as notational sugar, but
also checks that the input is a logical variable before taking the mean:

```{r Pr}
Pr(mu > 0)
```

More generally, the `rvar` data type provides two types of summary functions: 

1. Summary functions that mimic base-R vector summary functions, except applied to `rvar` vectors. These apply their summaries **over** elements of the input vectors **within** each draw, generally returning an `rvar` of length 1. These functions are prefixed with `rvar_` as a reminder that they return `rvar`s. Here is an example of `rvar_mean()`:
  
    ```{r rvar_mean_mu}
    rvar_mean(mu)
    ```

2. Summary functions that summarise **within** elements of input vectors and **over** draws. These summary functions generally return base arrays (`numeric` or `logical`) of the same shape as the input `rvar`, and are especially useful for diagnostic summaries. These summary functions are not prefixed with `rvar_` as they do not return `rvar`s. Here is an example of `mean()`:

    ```{r mean_mu}
    mean(mu)
    ```

   You should expect the same values from these functions (though in a different shape) when you use them with `summarise_draws()`, for example:

    ```{r summarise_draws_mu_mean}
    summarise_draws(mu, mean)
    ```

Here is a table of both types of summary functions:

|                          | 1. Summarise *within* draws,<br>*over* elements | 2. Summarise *over* draws,<br>*within* elements<br> |
|--------------------------|:------------------------------------------------|:----------------------------------------------------|
| **Output format**<br>of `res = f(x)` | `rvar` of length 1 | `array` of same shape as input `rvar` |
| **Help page**            | `help("rvar-summaries-within-draws")` | `help("rvar-summaries-over-draws")` |
| Numeric summaries        | `rvar_median()`<br>`rvar_sum()`, `rvar_prod()`<br>`rvar_min()`, `rvar_max()`| `median()`<br>`sum()`, `prod()`<br>`min()`, `max()`|
| Mean                     | `rvar_mean()`<br>*N/A* | `mean()`, `E()`<br>`Pr()`: enforces that input is `logical` |
| Spread                   | `rvar_sd()` <br>`rvar_var()` <br> `rvar_mad()` | `sd()`<br>`var()`, `variance()`<br> `mad()`|
| Range                    | `rvar_range()`<br>**Note:** `length(res) == 2` | `range()`<br>**Note:** `dim(res) == c(2, dim(x))` |
| Quantiles                | `rvar_quantile()`<br>**Note:** `length(res) == length(probs)` | `quantile()`<br>**Note:** `dim(res) == c(length(probs), dim(x))` |
| Logical summaries        | `rvar_all()`, `rvar_any()` | `all()`, `any()` |
| Special value predicates | `rvar_is_finite()`<br>`rvar_is_infinite()`<br>`rvar_is_nan()`<br>`rvar_is_na()`<br>**Note:** `dim(res) == dim(x)`. These functions act within draws but do not summarise over elements.        | `is.finite()`<br>`is.infinite()`<br>`is.nan()`<br>`is.na()`<br>**Note:** `res[i] == TRUE` if `x[i]` has any draws matching predicate (except for `is.finite()`, where all draws in `x[i]` must match) |
| Diagnostics              | *N/A* | `ess_basic()`, `ess_bulk()`, `ess_quantile()`, `ess_sd()`, `ess_tail()`,<br>`mcse_mean()`, `mcse_quantile()`, `mcse_sd()`<br>`rhat()`, `rhat_basic()`|

## Constants

Constant `rvar`s can be constructed by converting numeric vectors or arrays into `rvar`s
using `as_rvar()`, which will return an `rvar` with one draw and the same dimensions as its
input:

```{r const}
const <- as_rvar(1:3)
const
```

While normally `rvar`s must have the same number of draws to be used in the
same expression, `rvar`s with one draw are treated like constants, and 
can be combined with other `rvar`s:

```{r mu_plus_const}
mu + const
```

## Using existing R functions and expressions with `rvar`s

While `rvar`s attempt to emulate as much of the functionality of base R arrays
as possible, there are situations in which an existing R function may not 
work directly with an `rvar`. There are several approaches to solving this
problem.

For example, say you wish to generate samples from the
following expression for $\mu$, $\sigma$, and $x$:

$$
\begin{align}
\left[\begin{matrix}\mu_1 \\ \vdots \\ \mu_4 \end{matrix}\right] &\sim \textrm{Normal}\left(\left[\begin{matrix}1 \\ \vdots \\ 4 \end{matrix}\right],1\right)\\
\sigma &\sim \textrm{Gamma}(1,1)\\
\left[\begin{matrix}x_1 \\ \vdots \\ x_4 \end{matrix}\right] &\sim \textrm{Normal}\left(\left[\begin{matrix}\mu_1 \\ \vdots \\ \mu_4 \end{matrix}\right], \sigma\right)
\end{align}
$$

There are three different approaches you might take to doing this: converting
existing R functions with `rfun()`, executing expressions of
random variables with `rdo()`, or evaluating random number generator functions
using `rvar_rng()`.

### Converting functions with `rfun()`

The `rfun()` wrapper converts an existing R function into a new function 
that `rvar`s can be passed to it as arguments, and which
will return `rvar`s. 
We can use `rfun()` to convert the base `rnorm()` and `rgamma()` 
random number generating functions into functions that accept and return `rvar`s:

```{r rfun_defs}
rvar_norm <- rfun(rnorm)
rvar_gamma <- rfun(rgamma)
```

Then we can translate the above example into code using those functions:

```{r rfun_ex}
mu <- rvar_norm(4, mean = 1:4, sd = 1)
sigma <- rvar_gamma(1, shape = 1, rate = 1)
x <- rvar_norm(4, mu, sigma)
x
```

While `rfun()`-converted functions work well for prototyping, they will generally
speaking be slower than functions designed specifically for `rvar`s. Thus, you
may find you need to adopt other strategies (like `rvar_rng()`, described below;
or re-writing functions to support `rvar` directly using math operators and/or the
`draws_of()` function).

### Evaluating expressions with `rdo()`

An alternative to `rfun()` is to use `rdo()`, which can be passed nearly-arbitrary
R expressions. The expression will be executed
multiple times to construct an `rvar`. E.g., we can write an expression for
`mu` like in the above example:

```{r mu_rdo}
mu <- rdo(rnorm(4, mean = 1:4, sd = 1))
mu
```

We can also control the number of draws using the `ndraws` argument:

```{r mu_rdo_ndraws}
mu <- rdo(rnorm(4, mean = 1:4, sd = 1), ndraws = 1000)
mu
```

`rdo()` expressions can also contain other `rvar`s, so long as all `rvar`s in
the expression have the same number of draws. Thus, we can re-write the
example above that used `rfun()` as follows:

```{r rdo_ex}
mu <- rdo(rnorm(4, mean = 1:4, sd = 1))
sigma <- rdo(rgamma(1, shape = 1, rate = 1))
x <- rdo(rnorm(4, mu, sigma))
x
```

Like `rfun()`, `rdo()` is not necessarily fast, so you may find it more useful
for prototyping than production code.

### Evaluating random number generators with `rvar_rng()`

`rvar_rng()` is an alternative to `rfun()`/`rdo()` designed specifically to work
with random number generating functions that follow the typical API of such
functions in base R. Such functions, like `rnorm()`, `rgamma()`, `rbinom()`, etc 
all following this interface:

- They have a first argument, `n`, giving the number of draws to take from the distribution.
- Their arguments for distribution parameters (`mean`, `sd`, `shape`, `rate`, etc.) are vectorized.
- They return a single vector of length `n`, representing `n` draws from the distribution.

You can use any function with this interface with `rvar_rng()`, and it will adapt
it to be able to take `rvar` arguments and return an `rvar`, as follows:

```{r rvar_r_ex}
mu <- rvar_rng(rnorm, 4, mean = 1:4, sd = 1)
sigma <- rvar_rng(rgamma, 1, shape = 1, rate = 1)
x <- rvar_rng(rnorm, 4, mu, sigma)
x
```

In contrast to the `rfun()` and `rdo()` examples above, `rvar_rng()` takes 
advantage of the existing vectorization of the underlying random number generating
function to execute quickly.

## Broadcasting

Broadcasting for `rvar`s does not follow R's vector recycling rules. Instead,
when two variables with different dimensions are being used with basic arithmetic
functions, dimensions are added until both variables have the same number of dimensions.
If two variables $x$ and $y$ differ on the length of dimension $d$, they can
be broadcast to the same size so long as one of the variables has dimension $d$
of size 1. Then that variable will be broadcast up to the same size as the
other variable along that dimension. If two variables disagree on the size of 
a dimension and neither has size 1, it is an error.

For example, consider this random matrix:

```{r X_matrix}
X <- rdo(rnorm(12, 1:12), dim = c(4,3))
X
```

And this vector of length 3:

```{r y_vector}
y <- rdo(rnorm(3, 3:1))
y
```

If we attempt to add `X` and `y`, it will produce an error as vectors are 
by default treated as column vectors, and `y` has length 3 while columns of 
`X` have length 4:

```{r X_plus_y, error = TRUE}
X + y
```

By contrast, R arrays of the same shape will simply recycle `y` until it is the
same length as `X` (regardless of the dimensions). Thus will produce a result,
though likely not the intended result:

```{r mean_X_plus_y}
mean(X) + mean(y)
```

On the other hand, if y were a row vector...

```{r row_y}
row_y = t(y)
row_y
```

...it would have the same number of
columns as `X` and contain only one row, so it can be broadcast along rows of `X`:

```{r X_plus_row_y}
X + row_y
```

## Applying functions over `rvar`s

The `rvar` data type supplies an implementation of `as.list()`, which should give 
compatibility with the base R family of functions for applying functions over arrays:
`apply()`, `lapply()`, `vapply()`, `sapply()`, etc. You can also manually use
`as.list()` to convert an `rvar` into a list along its first dimension, which
may be necessary for compatibility with some functions (like `purrr:map()`).

For example, given this multidimensional `rvar`...

```{r multidim_array}
set.seed(3456)
x <- rvar_rng(rnorm, 24, mean = 1:24)
dim(x) <- c(2,3,4)
x
```

... you can apply functions along the margins using `apply()` (here, a silly example):

```{r apply}
apply(x, c(1,2), length)
```

One exception is that while `apply()` will work with an `rvar` input if your function
returns base data types (like numerics), it will not give you simplified `rvar`
arrays if your function returns an `rvar`. Thus, we supply the `rvar_apply()`
function, which takes in either base arrays or `rvar` arrays and returns `rvar`
arrays, and which also uses the `rvar` broadcasting rules to combine the
results of the applied function.

For example, you can use `rvar_apply()` with `rvar_mean()` to compute the distributions of
means along one margin of an array:

```{r rvar_apply_one_dim}
rvar_apply(x, 1, rvar_mean)
```

Or along multiple dimensions:

```{r rvar_apply_multi_dim}
rvar_apply(x, c(2,3), rvar_mean)
```

## Looping over `draws` and `rvar`s

The `rvar` datatype is also used in `for_each_draw()`, which allows you to loop
over draws in a `draws` object or an `rvar`. `for_each_draw(x, expr)` converts `x`
into a `draws_rvars()` object, then loops over each draw of `x`, executing the
provided expression, `expr`. The expression can use the variables in `x` as if
they were regular R arrays.

One application of `for_each_draw()` is in constructing base-R plots of individual
draws (for `ggplot2`-based plotting of `rvar`s, see the next section and the
[ggdist](https://mjskay.github.io/ggdist/) package). For example, it can be used
to construct a parallel coordinates plot:

```{r eight_schools_parcoord, fig.width = 6, fig.height = 4}
eight_schools <- as_draws_rvars(example_draws())

plot(1, type = "n",
  xlim = c(1, length(eight_schools$theta)),
  ylim = range(range(eight_schools$theta)),
  xlab = "school", ylab = "theta"
)

# use for_each_draw() to make a parallel coordinates plot of all draws
# of eight_schools$theta
for_each_draw(eight_schools, {
  lines(seq_along(theta), theta, col = rgb(1, 0, 0, 0.05))
})

# add means and 90% intervals
lines(seq_along(eight_schools$theta), mean(eight_schools$theta))
with(summarise_draws(eight_schools$theta), 
  segments(seq_along(eight_schools$theta), y0 = q5, y1 = q95)
)
```

As `for_each_draw()` will be slower than most other ways of manipulating draws, 
this function should generally not be used unless needed.


## Using `rvar`s in data frames and in ggplot2

`rvar`s can be used as columns in `data.frame()` or `tibble()` objects:

```{r data_frame_with_y}
data.frame(x = c("a","b","c"), y)
```

This makes them convenient for adding predictions to a data frame alongside
the data used to generate the predictions. `rvar`s can then be visualized with ggplot2
by passing them to the `xdist` and `ydist` aesthetics of the `stat_...` family of
geometries in the [ggdist](https://mjskay.github.io/ggdist/) package, 
such as `stat_halfeye()`, `stat_lineribbon()`, and `stat_dotsinterval()`.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_draws_df.R
\name{draws_df}
\alias{draws_df}
\alias{as_draws_df}
\alias{as_draws_df.default}
\alias{as_draws_df.data.frame}
\alias{as_draws_df.draws_df}
\alias{as_draws_df.draws_matrix}
\alias{as_draws_df.draws_array}
\alias{as_draws_df.draws_list}
\alias{as_draws_df.draws_rvars}
\alias{as_draws_df.mcmc}
\alias{as_draws_df.mcmc.list}
\alias{is_draws_df}
\title{The \code{draws_df} format}
\usage{
as_draws_df(x, ...)

\method{as_draws_df}{default}(x, ...)

\method{as_draws_df}{data.frame}(x, ...)

\method{as_draws_df}{draws_df}(x, ...)

\method{as_draws_df}{draws_matrix}(x, ...)

\method{as_draws_df}{draws_array}(x, ...)

\method{as_draws_df}{draws_list}(x, ...)

\method{as_draws_df}{draws_rvars}(x, ...)

\method{as_draws_df}{mcmc}(x, ...)

\method{as_draws_df}{mcmc.list}(x, ...)

draws_df(..., .nchains = 1)

is_draws_df(x)
}
\arguments{
\item{x}{An object to convert to a \code{draws_df} object.}

\item{...}{For \code{as_draws_df()}:
Arguments passed to individual methods (if applicable).
For \code{draws_df()}: Named
arguments containing numeric vectors each defining a separate variable.}

\item{.nchains}{(positive integer) The number of chains. The default is \code{1}.}
}
\value{
A \code{draws_df} object, which has classes
\code{c("draws_df", "draws", class(tibble::tibble()))}.
}
\description{
The \code{as_draws_df()} methods convert
objects to the \code{draws_df} format.
The \code{draws_df()} function creates an object of the
\code{draws_df} format based on a set of numeric vectors.
See \strong{Details}.
}
\details{
Objects of class \code{"draws_df"} are \link[tibble:tibble]{tibble} data
frames. They have one column per variable as well as additional metadata
columns \code{".iteration"}, \code{".chain"}, and \code{".draw"}. The difference between
the \code{".iteration"} and \code{".draw"} columns is that the former is relative to
the MCMC chain while the latter ignores the chain information and has all
unique values. See \strong{Examples}.

If a \code{data.frame}-like object is supplied to \code{as_draws_df} that contains
columns named \code{".iteration"} or \code{".chain"}, they will be treated as
iteration and chain indices, respectively. See \strong{Examples}.
}
\examples{
x1 <- as_draws_df(example_draws())
class(x1)
print(x1)
str(x1)

x2 <- draws_df(a = rnorm(10), b = rnorm(10), c = 1)
class(x2)
print(x2)
str(x2)

# the difference between iteration and draw is clearer when contrasting
# the head and tail of the data frame
print(head(x1), reserved = TRUE, max_variables = 2)
print(tail(x1), reserved = TRUE, max_variables = 2)

# manually supply chain information
xnew <- data.frame(mu = rnorm(10), .chain = rep(1:2, each = 5))
xnew <- as_draws_df(xnew)
print(xnew)

}
\seealso{
Other formats: 
\code{\link{draws_array}()},
\code{\link{draws_list}()},
\code{\link{draws_matrix}()},
\code{\link{draws_rvars}()},
\code{\link{draws}}
}
\concept{formats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-.R
\name{rvar}
\alias{rvar}
\title{Random variables of arbitrary dimension}
\usage{
rvar(
  x = double(),
  dim = NULL,
  dimnames = NULL,
  nchains = NULL,
  with_chains = FALSE
)
}
\arguments{
\item{x}{(multiple options) The object to convert to an \code{rvar}:
\itemize{
\item A vector of draws from a distribution.
\item An array where the first dimension represents draws from a distribution.
The resulting \code{\link{rvar}} will have dimension \code{dim(x)[-1]}; that is,
everything except the first dimension is used for the shape of the
variable, and the first dimension is used to index draws from the
distribution (see \strong{Examples}). Optionally,
if \code{with_chains == TRUE}, the first dimension indexes the iteration and the
second dimension indexes the chain (see \code{with_chains}).
\item An \code{rvar}.
}}

\item{dim}{(integer vector) One or more integers giving the maximal indices
in each dimension to override the dimensions of the \code{\link{rvar}} to be created
(see \code{\link[=dim]{dim()}}). If \code{NULL} (the default), \code{dim} is determined by the input.
\strong{NOTE:} This argument controls the dimensions of the \code{\link{rvar}}, not the
underlying array, so you cannot change the number of draws using this
argument.}

\item{dimnames}{(list) Character vectors giving the names in each dimension
to override the names of the dimensions of the \code{\link{rvar}} to be created (see
\code{\link[=dimnames]{dimnames()}}). If \code{NULL} (the default), this is determined by the input.
\strong{NOTE:} This argument controls the names of the dimensions of the
\code{\link{rvar}}, not the underlying array.}

\item{nchains}{(positive integer) The number of chains. The if \code{NULL} (the default),
\code{1} is used unless \code{x} is already an \code{\link{rvar}}, in which case the number of
chains it has is used.}

\item{with_chains}{(logical) Does \code{x} include a dimension for chains?
If \code{FALSE} (the default), chains are not included, the first dimension of
the input array should index draws, and the \code{nchains} argument can be
used to determine the number of chains. If \code{TRUE}, the \code{nchains} argument
is ignored and the second dimension of \code{x} is used to index chains.
Internally, the array will be converted to a format without the chain index.
Ignored when \code{x} is already an \code{\link{rvar}}.}
}
\value{
An object of class \code{"rvar"} representing a random variable.
}
\description{
Random variables backed by arrays of arbitrary dimension
}
\details{
The \code{"rvar"} class internally represents random variables as arrays of arbitrary
dimension, where the first dimension is used to index draws from the distribution.
Most mathematical operators and functions are supported, including efficient matrix
multiplication and vector and array-style indexing. The intent is that an \code{rvar}
works as closely as possible to how a base vector/matrix/array does, with a few
differences:
\itemize{
\item The default behavior when subsetting is not to drop extra dimensions (i.e.
the default \code{drop} argument for \code{[} is \code{FALSE}, not \code{TRUE}).
\item Rather than base R-style recycling, \code{rvar}s use a limited form of broadcasting:
if an operation is being performed on two vectors with different size of the same
dimension, the smaller vector will be recycled up to the size of the larger one
along that dimension so long as it has size 1.
}

For functions that expect base numeric arrays and for which \code{rvar}s cannot be
used directly as arguments, you can use \code{\link[=rfun]{rfun()}} or \code{\link[=rdo]{rdo()}} to translate your
code into code that executes across draws from one or more random variables
and returns a random variable as output. Typically \code{\link[=rdo]{rdo()}} offers the most
straightforward translation.

As \code{\link[=rfun]{rfun()}} and \code{\link[=rdo]{rdo()}} incur some performance cost, you can also operate directly
on the underlying array using the \code{\link[=draws_of]{draws_of()}} function. To re-use existing
random number generator functions to efficiently create \code{rvar}s, use \code{\link[=rvar_rng]{rvar_rng()}}.
}
\examples{

set.seed(1234)

# To create a "scalar" `rvar`, pass a one-dimensional array or a vector
# whose length (here `4000`) is the desired number of draws:
x <- rvar(rnorm(4000, mean = 1, sd = 1))
x

# Create random vectors by adding an additional dimension:
n <- 4   # length of output vector
x <- rvar(array(rnorm(4000 * n, mean = rep(1:n, each = 4000), sd = 1), dim = c(4000, n)))
x

# Create a random matrix:
rows <- 4
cols <- 3
x <- rvar(array(rnorm(4000 * rows * cols, mean = 1, sd = 1), dim = c(4000, rows, cols)))
x

# If the input sample comes from multiple chains, we can indicate that using the
# nchains argument (here, 1000 draws each from 4 chains):
x <- rvar(rnorm(4000, mean = 1, sd = 1), nchains = 4)
x

# Or if the input sample has chain information as its second dimension, we can
# use with_chains to create the rvar
x <- rvar(array(rnorm(4000, mean = 1, sd = 1), dim = c(1000, 4)), with_chains = TRUE)
x

}
\seealso{
\code{\link[=as_rvar]{as_rvar()}} to convert objects to \code{rvar}s. See \code{\link[=rdo]{rdo()}}, \code{\link[=rfun]{rfun()}}, and
\code{\link[=rvar_rng]{rvar_rng()}} for higher-level interfaces for creating \code{rvar}s.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/uniformity_test.R
\name{uniformity_test}
\alias{uniformity_test}
\title{Uniformity test for PIT values}
\usage{
uniformity_test(pit, test)
}
\arguments{
\item{pit}{Numeric vector of PIT values in \verb{[0, 1]}.}

\item{test}{Character string. One of \code{"POT"}, \code{"PIET"}, or \code{"PRIT"}.
See details above.}
}
\value{
A list with components:
\itemize{
\item \code{pvalue}: Global p-value from the Cauchy combination test.
\item \code{pointwise}: Shapley values (contributions of each PIT value to the
test statistic). Non-negative values indicate points contributing to
evidence against uniformity.
}
}
\description{
Tests whether PIT (probability integral transform) values deviate from
uniformity while accounting for dependence among them. LOO-PIT values are
not independent because each LOO predictive distribution is conditioned on
nearly the same data (all but one observation), which induces correlation.
Standard uniformity tests that assume independence can have inflated
Type I error or reduced power. This function implements dependence-aware
procedures from Tesso and Vehtari (2026).
}
\details{
Three test variants are available:
\itemize{
\item \strong{POT} (Pointwise Order Tests): Uses beta distributions for order
statistics. Best for continuous PIT values; recommended as default for
LOO-PIT model checking. Good power against diverse departures from
uniformity.
\item \strong{PIET} (Pointwise Inverse-CDF Evaluation Tests): Uses a continuous
reference distribution (exponential) via inverse-CDF transformation.
Best for detecting tail deviations; maintains valid Type I error control
under strong dependence.
\item \strong{PRIT} (Pointwise Rank-based Individual Tests): Uses binomial
distributions on scaled ECDF (ranks). Intended for discrete or
rank-based PIT values.
}

All procedures compute pointwise p-values, aggregate them via the (truncated)
Cauchy combination test (Liu and Xie, 2020; Chen et al., 2025), and derive
Shapley values to quantify each point's contribution to the overall test.
}
\references{
Tesso, H., and Vehtari, A. (2026). LOO-PIT predictive model checking.
arXiv preprint arXiv:2603.02928.

Liu, Y., and Xie, J. (2020). Cauchy combination test: a powerful test
with analytic p-value calculation under arbitrary dependency structures.
\emph{Journal of the American Statistical Association}, 115(529), 393-402.

Chen, B., Xu, W., and Gao, X. (2025). Truncated Cauchy combination test:
a robust and powerful p-value combination method with arbitrary
correlations. arXiv preprint arXiv:2506.12489.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-summaries-over-draws.R
\name{rvar-summaries-over-draws}
\alias{rvar-summaries-over-draws}
\alias{E}
\alias{mean.rvar}
\alias{Pr}
\alias{Pr.default}
\alias{Pr.logical}
\alias{Pr.rvar}
\alias{median.rvar}
\alias{min.rvar}
\alias{max.rvar}
\alias{sum.rvar}
\alias{prod.rvar}
\alias{all.rvar}
\alias{any.rvar}
\alias{Summary.rvar}
\alias{variance.rvar}
\alias{var}
\alias{var.default}
\alias{var.rvar}
\alias{sd}
\alias{sd.default}
\alias{sd.rvar}
\alias{mad}
\alias{mad.default}
\alias{mad.rvar}
\alias{mad.rvar_ordered}
\alias{range.rvar}
\alias{is.finite.rvar}
\alias{is.infinite.rvar}
\alias{is.nan.rvar}
\alias{is.na.rvar}
\title{Summaries of random variables within array elements, over draws}
\usage{
E(x, ...)

\method{mean}{rvar}(x, ...)

Pr(x, ...)

\method{Pr}{default}(x, ...)

\method{Pr}{logical}(x, ...)

\method{Pr}{rvar}(x, ...)

\method{median}{rvar}(x, ...)

\method{min}{rvar}(x, ...)

\method{max}{rvar}(x, ...)

\method{sum}{rvar}(x, ...)

\method{prod}{rvar}(x, ...)

\method{all}{rvar}(x, ...)

\method{any}{rvar}(x, ...)

\method{Summary}{rvar}(...)

\method{variance}{rvar}(x, ...)

var(x, ...)

\method{var}{default}(x, ...)

\method{var}{rvar}(x, ...)

sd(x, ...)

\method{sd}{default}(x, ...)

\method{sd}{rvar}(x, ...)

mad(x, ...)

\method{mad}{default}(x, ...)

\method{mad}{rvar}(x, ...)

\method{mad}{rvar_ordered}(x, ...)

\method{range}{rvar}(x, ...)

\method{is.finite}{rvar}(x)

\method{is.infinite}{rvar}(x)

\method{is.nan}{rvar}(x)

\method{is.na}{rvar}(x)
}
\arguments{
\item{x}{(rvar) An \code{\link{rvar}}.}

\item{...}{Further arguments passed to underlying functions (e.g.,
\code{base::mean()} or \code{base::median()}), such as \code{na.rm}.}
}
\value{
A numeric or logical vector with the same dimensions as the given random variable, where
each entry in the vector is the mean, median, or variance of the corresponding entry in \code{x}.
}
\description{
Compute summaries within elements of an \code{\link{rvar}} and over draws of each element,
producing an array of the same shape as the input random variable (except in
the case of \code{range()}, see \strong{Details}).
}
\details{
Summaries include expectations (\code{E()} or \code{mean()}), probabilities (\code{Pr()}),
medians (\code{median()}), spread (\code{var()}, \code{variance()}, \code{sd()}, \code{mad()}), sums and
products (\code{sum()}, \code{prod()}), extrema and ranges (\code{min()}, \code{max()}, \code{range()}),
logical summaries (\code{all()}, \code{any()}), and special value predicates (\code{is.finite()},
\code{is.infinite()}, \code{is.nan()}, \code{is.na()}).

Unless otherwise stated, these functions return a numeric array with the same shape
(same dimensions) as the input \code{\link{rvar}}, \code{x}.

\code{range(x)} returns an array with dimensions \code{c(2, dim(x))}, where the last
dimension contains the minimum and maximum values.

\code{is.infinite(x)}, \code{is.nan(x)}, and \code{is.na(x)} return logical arrays, where each
element is \code{TRUE} if \strong{any} draws in its corresponding element in \code{x} match
the predicate. Each elements in the result of \code{is.finite(x)} is \code{TRUE} if
\strong{all} draws in the corresponding element in \code{x} are finite.

Both \code{E()}, \code{mean()}, and \code{Pr()} return the means of each element in the input.
\code{Pr()} additionally checks that the provided \code{\link{rvar}}
is a logical variable (hence, taking its expectation results in a probability).

For consistency, \code{E()} and \code{Pr()} are also defined for base arrays so that
they can be used as summary functions in \code{summarise_draws()}.
}
\examples{

set.seed(5678)
x = rvar_rng(rnorm, 4, mean = 1:4, sd = 2)

# These should all be ~= c(1, 2, 3, 4)
E(x)
mean(x)
median(x)

# This ...
Pr(x < 1.5)
# ... should be about the same as this:
pnorm(1.5, mean = 1:4, sd = 2)

}
\seealso{
\link{rvar-summaries-within-draws} for summary functions within draws.
\link{rvar-dist} for density, CDF, and quantile functions of random variables.

Other rvar-summaries: 
\code{\link{rvar-summaries-within-draws}},
\code{\link{rvar_is_finite}()}
}
\concept{rvar-summaries}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc.R
\name{is_constant}
\alias{is_constant}
\title{Check if vector is constant}
\usage{
is_constant(x, tol = .Machine$double.eps)
}
\arguments{
\item{x}{(vector) vector to check if is constant}

\item{tol}{(numeric) tolerance to consider two values equal.
Default is \code{.Machine$double.eps}.}
}
\description{
check if a vector is constant, up to a defined tolerance.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-apply.R
\name{rvar_apply}
\alias{rvar_apply}
\title{Random variable resulting from a function applied over margins of an array or random variable}
\usage{
rvar_apply(.x, .margin, .f, ...)
}
\arguments{
\item{.x}{An array or an \code{\link{rvar}}.}

\item{.margin}{(multiple options) The subscripts which the function will be applied over:
\itemize{
\item An integer vector. E.g., for a matrix \code{1} indicates rows, \code{2} indicates
columns, \code{c(1, 2)} indicates rows and columns.
\item A character vector of dimension names if \code{.x} has named dimensions.
}}

\item{.f}{(function) The function to be applied. The function \code{.f} must
return an \code{\link{rvar}} and the dimensions of the result of \code{.f} applied to each
margin of \code{.x} must be able to be broadcasted to a common shape (otherwise
the resulting \code{\link{rvar}} cannot be simplified). See \strong{Details}.}

\item{...}{Optional arguments passed to \code{.f}.}
}
\value{
An \code{\link{rvar}}.

If the result of each call to \code{.f} returns an \code{\link{rvar}} of dimension \code{d} after
being broadcast to a common shape, then \code{rvar_apply()} returns an \code{\link{rvar}} of
dimension \code{c(d, dim(.x)[.margin])}. If the last dimension of the result would
be \code{1}, it is dropped (other dimensions equal to \code{1} are retained). If \code{d} is
\code{0}, the result has length \code{0} but not necessarily the 'correct' dimension.
}
\description{
Returns an \code{\link{rvar}} obtained by applying a function to margins of an array or \code{\link{rvar}}.
Acts like \code{apply()}, except that the function supplied (\code{.f}) should return an \code{\link{rvar}},
and the final result is always an \code{\link{rvar}}.
}
\details{
This function acts much like \code{apply()}, except that the function passed to it (\code{.f})
must return \code{\link{rvar}}s, and the result is simplified into an \code{\link{rvar}}. Unlike
\code{apply()}, it also keeps the dimensions of the returned values along each margin,
rather than simplifying each margin to a vector, and if the results of \code{.f} do
not all have the same dimensions, it applies the \code{\link{rvar}} broadcasting rules to
bind results together rather than using vector recycling.

If you wish to apply functions over \code{\link{rvar}}s where the result is not intended to
be simplified into an \code{\link{rvar}}, you can use the standard \code{apply()}, \code{lapply()},
\code{sapply()}, or \code{vapply()} functions.
}
\examples{

set.seed(3456)
x <- rvar_rng(rnorm, 24, mean = 1:24)
dim(x) <- c(2,3,4)

# we can find the distributions of marginal means of the above array
# using rvar_mean along with rvar_apply
rvar_apply(x, 1, rvar_mean)
rvar_apply(x, 2:3, rvar_mean)

}
\seealso{
\code{\link[=as_rvar]{as_rvar()}} to convert objects to \code{rvar}s. See \code{\link[=rdo]{rdo()}}, \code{\link[=rfun]{rfun()}}, and
\code{\link[=rvar_rng]{rvar_rng()}} for higher-level interfaces for creating \code{rvar}s.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draws-index.R
\name{set_variables}
\alias{set_variables}
\title{Set variable names in \code{draws} objects}
\usage{
set_variables(x, variables, ...)
}
\arguments{
\item{x}{(draws) A \code{\link{draws}} object.}

\item{variables}{(character) new variable names.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
Returns a \code{\link{draws}} object of the same format as \code{x}, with
variables named as specified.
}
\description{
Set variable names for all variables in a \code{\link{draws}} object. Useful
when using pipe operators.
}
\examples{
x <- as_draws(matrix(rnorm(100), ncol = 2))
variables(x)

x <- set_variables(x, c("theta[1]", "theta[2]"))
variables(x)

# this is equivalent to
variables(x) <- c("theta[1]", "theta[2]")
variables(x)

}
\seealso{
\code{\link{variables}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{ess_tail}
\alias{ess_tail}
\alias{ess_tail.default}
\alias{ess_tail.rvar}
\title{Tail effective sample size (tail-ESS)}
\usage{
ess_tail(x, ...)

\method{ess_tail}{default}(x, ...)

\method{ess_tail}{rvar}(x, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
If the input is an array, returns a single numeric value. If any of the draws
is non-finite, that is, \code{NA}, \code{NaN}, \code{Inf}, or \code{-Inf}, the returned output
will be (numeric) \code{NA}. Also, if all draws within any of the chains of a
variable are the same (constant), the returned output will be (numeric) \code{NA}
as well. The reason for the latter is that, for constant draws, we cannot
distinguish between variables that are supposed to be constant (e.g., a
diagonal element of a correlation matrix is always 1) or variables that just
happened to be constant because of a failure of convergence or other problems
in the sampling process.

If the input is an \code{\link{rvar}}, returns an array of the same dimensions as the
\code{\link{rvar}}, where each element is equal to the value that would be returned by
passing the draws array for that element of the \code{\link{rvar}} to this function.
}
\description{
Compute a tail effective sample size estimate (tail-ESS) for a single
variable. Tail-ESS is useful as a diagnostic for the sampling efficiency in
the tails of the posterior. It is defined as the minimum of the effective
sample sizes for 5\% and 95\% quantiles. For the bulk effective sample
size see \code{\link[=ess_bulk]{ess_bulk()}}. See Vehtari (2021) for an in-depth
comparison of different effective sample size estimators.
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
ess_tail(mu)

d <- as_draws_rvars(example_draws("multi_normal"))
ess_tail(d$Sigma)

}
\references{
Aki Vehtari, Andrew Gelman, Daniel Simpson, Bob Carpenter, and
Paul-Christian Bürkner (2021). Rank-normalization, folding, and
localization: An improved R-hat for assessing convergence of
MCMC (with discussion). \emph{Bayesian Data Analysis}. 16(2), 667-–718.
doi:10.1214/20-BA1221

Aki Vehtari (2021). Comparison of MCMC effective sample size estimators.
Retrieved from https://avehtari.github.io/rhat_ess/ess_comparison.html
}
\seealso{
Other diagnostics: 
\code{\link{ess_basic}()},
\code{\link{ess_bulk}()},
\code{\link{ess_quantile}()},
\code{\link{ess_sd}()},
\code{\link{mcse_mean}()},
\code{\link{mcse_quantile}()},
\code{\link{mcse_sd}()},
\code{\link{rhat_basic}()},
\code{\link{rhat}()},
\code{\link{rstar}()}
}
\concept{diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarise_draws.R
\name{print.draws_summary}
\alias{print.draws_summary}
\title{Print summaries of \code{draws} objects}
\usage{
\method{print}{draws_summary}(x, ..., num_args = NULL)
}
\arguments{
\item{x}{(draws_summary) A \code{"draws_summary"} object as output by \code{\link[=summarise_draws]{summarise_draws()}}.}

\item{...}{Additional arguments passed to \code{\link[tibble:formatting]{tibble::print.tbl_df()}}}

\item{num_args}{(named list) Optional arguments passed to
\link[tibble:num]{num()} for pretty printing of summaries. If \code{NULL}
(the default), uses the arguments stored in the \code{"num_args"} attribute of
\code{x}, as set by the \code{.num_args} argument of \code{\link[=summarise_draws]{summarise_draws()}}, which
itself can be controlled globally via the \code{posterior.num_args}
\link[base:options]{option}.}
}
\value{
An invisible version of the input object.
}
\description{
Print output from \code{\link[=summarise_draws]{summarise_draws()}}.
}
\examples{
x <- example_draws("eight_schools")

# adjust how summaries are printed when calling summarise_draws()...
summarise_draws(x, .num_args = list(sigfig = 2, notation = "dec"))

# ... or when printing
s <- summarise_draws(x)
print(s, num_args = list(sigfig = 2, notation = "dec"))
print(s, num_args = list(digits = 3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_variable_array.R
\name{extract_variable_array}
\alias{extract_variable_array}
\alias{extract_variable_array.default}
\alias{extract_variable_array.draws}
\title{Extract array of a single (possibly indexed) variable}
\usage{
extract_variable_array(x, variable, ...)

\method{extract_variable_array}{default}(x, variable, ...)

\method{extract_variable_array}{draws}(x, variable, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{variable}{(string) The name of the variable to extract. To extract all
dimensions from variables with indices (e.g. \code{"x[1]"}), provide the base
variable name (e.g. \code{"x"}).}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
An \code{array} with dimension \code{niterations(x)} x \code{nchains(x)} x any remaining
dimensions determined by the indices of the variable \code{x}.
}
\description{
Extract an array of draws of a single variable, including any dimensions of
variables with indices.
}
\examples{
x <- example_draws(example = "multi_normal")

mu <- extract_variable_array(x, variable = "mu")
str(mu)

mu1 <- extract_variable_array(x, variable = "mu[1]")
str(mu1)

Sigma <- extract_variable_array(x, variable = "Sigma")
str(Sigma)

}
\seealso{
Other variable extraction methods: 
\code{\link{extract_variable}()},
\code{\link{extract_variable_matrix}()}
}
\concept{variable extraction methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight_draws.R
\name{weights.draws}
\alias{weights.draws}
\title{Extract Weights from Draws Objects}
\usage{
\method{weights}{draws}(object, log = FALSE, normalize = TRUE, ...)
}
\arguments{
\item{object}{(draws) A \code{\link{draws}} object.}

\item{log}{(logical) Should the weights be returned on the log scale?
Defaults to \code{FALSE}.}

\item{normalize}{(logical) Should the weights be normalized to sum to 1 on
the standard scale? Defaults to \code{TRUE}.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A vector of weights, with one weight per draw.
}
\description{
Extract weights from \code{\link{draws}} objects, with one weight per draw.
See \code{\link{weight_draws}} for details how to add weights to \code{\link{draws}} objects.
}
\examples{
x <- example_draws()

# sample some random weights for illustration
wts <- rexp(ndraws(x))
head(wts)

# add weights
x <- weight_draws(x, weights = wts)

# extract weights
head(weights(x)) # defaults to normalized weights
head(weights(x, normalize=FALSE)) # recover original weights
head(weights(x, log=TRUE)) # get normalized log-weights

# add weights which are already on the log scale
log_wts <- log(wts)
head(log_wts)

x <- weight_draws(x, weights = log_wts, log = TRUE)
# extract weights
head(weights(x))
head(weights(x, log=TRUE, normalize = FALSE)) # recover original log_wts

}
\seealso{
\code{\link{weight_draws}}, \code{\link{resample_draws}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-slice.R
\name{rvar-slice}
\alias{rvar-slice}
\alias{[[.rvar}
\alias{[[<-.rvar}
\alias{[.rvar}
\alias{[<-.rvar}
\title{Random variable slicing}
\usage{
\method{[[}{rvar}(x, i, ...)

\method{[[}{rvar}(x, i, ...) <- value

\method{[}{rvar}(x, ..., drop = FALSE)

\method{[}{rvar}(x, i, ...) <- value
}
\arguments{
\item{x}{an \code{\link{rvar}}.}

\item{i, ...}{indices; see \emph{Details}.}

\item{value}{(\code{rvar} or coercable to \code{rvar}) Value to insert into
\code{x} at the location determined by the indices.}

\item{drop}{(logical) Should singular dimensions be dropped when slicing
array \code{\link{rvar}}s? Unlike base array slicing operations, defaults to \code{FALSE}.}
}
\description{
Operations for slicing \code{\link{rvar}}s and replacing parts of \code{\link{rvar}}s.
}
\details{
The \code{\link{rvar}} slicing operators (\code{[} and \code{[[}) attempt to implement the same
semantics as the \link[base:Extract]{base array slicing operators}. There are some
exceptions; most notably, \code{\link{rvar}} slicing defaults to \code{drop = FALSE} instead
of \code{drop = TRUE}.
}
\section{Extracting or replacing single elements with \code{[[}}{


The \code{[[} operator extracts (or replaces) single elements. It always
returns (or replaces) a scalar (length-1) \code{\link{rvar}}.

The \code{x[[i,...]]} operator can be used as follows:
\itemize{
\item \verb{x[[<numeric>]]} for scalar numeric \code{i}: gives the \code{i}th element of \code{x}. If \code{x} is
multidimensional (i.e. \code{length(dim(x)) > 1}), extra dimensions are ignored
when indexing. For example, if \code{x} is a \eqn{6 \times 2} \code{\link{rvar}} array, the
7th element, \code{x[[7]]}, will be the first element of the second column, \code{x[1,2]}.
\item \verb{x[[<numeric rvar>]]} for scalar numeric \code{\link{rvar}} \code{i}: a generalization of indexing when
\code{i} is a scalar numeric. Within each draw of \code{x}, selects the element
corresponding to the value of \code{i} within that same draw.
\item \verb{x[[<character>]]} for scalar character \code{i}: gives the element of \code{x} with name
equal to \code{i}. \strong{Unlike with base arrays}, does not work with
multidimensional \code{\link{rvar}}s.
\item \code{x[[i_1,i_2,...,i_n]]} for scalar numeric or character \code{i_1}, \code{i_2}, etc.
Must provide exactly the same number of indices as dimensions in \code{x}. Selects
the element at the corresponding position in the \code{\link{rvar}} by number and/or
dimname (as a string).
}
}

\section{Extracting or replacing multiple elements with \code{[}}{


The \code{[} operator extracts (or replaces) multiple elements. It always returns
(or replaces) a possibly-multidimensional \code{\link{rvar}}.

The \code{x[i,...]} operator can be used as follows:
\itemize{
\item \verb{x[<logical>]} for vector logical \code{i}: \code{i} is recycled to the same length as \code{x},
ignoring multiple dimensions in \code{x}, then an \code{\link{rvar}} vector is returned
containing the elements in \code{x} where \code{i} is \code{TRUE}.
\item \verb{x[<logical rvar>]} for scalar logical \code{\link{rvar}} \code{i}: returns an \code{\link{rvar}} the same shape
as \code{x} containing only those draws where \code{i} is \code{TRUE}.
\item \verb{x[<numeric>]} for vector numeric \code{i}: an \code{\link{rvar}} vector is returned
containing the \code{i}th elements of \code{x}, ignoring dimensions.
\item \verb{x[<matrix>]} for numeric matrix \code{i}, where \code{ncol(i) == length(dim(x))}: each row
of \code{i} should give the multidimensional index for a single element in \code{x}. The
result is an \code{\link{rvar}} vector of length \code{nrow(i)} containing elements of \code{x}
selected by each row of \code{i}.
\item \code{x[i_1,i_2,...,i_n]} for vector numeric, character, or logical \code{i_1},
\code{i_2}, etc. Returns a slice of \code{x} containing all elements from the dimensions
specified in \code{i_1}, \code{i_2}, etc. If an argument is left empty, all elements
from that dimension are included. Unlike base arrays, trailing dimensions
can be omitted entirely and will still be selected; for example, if \code{x} has
three dimensions, both \code{x[1,,]} and \code{x[1,]} can be used to create a
slice that includes all elements from the last two dimensions. Unlike base
arrays, \code{[} defaults to \code{drop = FALSE}, so results retain the same number of
dimensions as \code{x}.
}
}

\examples{
x <- rvar(array(1:24, dim = c(4,2,3)))
dimnames(x) <- list(c("a","b"), c("d","e","f"))
x

## Slicing single elements
# x[[<numeric>]]
x[[2]]

# x[[<numeric rvar>]]
# notice the draws of x[1:4]...
draws_of(x[1:4])
x[[rvar(c(1,3,4,4))]]
# ... x[[rvar(c(1,3,4,4))]] creates a mixures of those draws
draws_of(x[[rvar(c(1,3,4,4))]])

# x[[i_1,i_2,...]]
x[[2,"e"]]


## Slicing multiple elements
# x[<logical>]
x[c(TRUE,TRUE,FALSE)]

# x[<logical rvar>]
# select every other draw
x[rvar(c(TRUE,FALSE,TRUE,FALSE))]

# x[<numeric>]
x[1:3]

# x[<matrix>]
x[rbind(
  c(1,2),
  c(1,3),
  c(2,2)
)]

# x[i_1,i_2,...,i_n]
x[1,]
x[1,2:3]
x[,2:3]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_draws.R
\name{draws}
\alias{draws}
\alias{as_draws}
\alias{is_draws}
\title{Transform to \code{draws} objects}
\usage{
as_draws(x, ...)

is_draws(x)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
If possible, a \code{draws} object in the closest supported format to \code{x}.
The formats are linked to in the \strong{See Also} section below.
}
\description{
Try to transform an \R object to a format supported by the \pkg{posterior}
package.
}
\details{
The class \code{"draws"} is the parent class of all supported formats,
which also have their own subclasses of the form \code{"draws_{format}"} (e.g.
\code{"draws_array"}).
}
\examples{
# create some random draws
x <- matrix(rnorm(30), nrow = 10)
colnames(x) <- c("a", "b", "c")
str(x)

# transform to a draws object
y <- as_draws(x)
str(y)

# remove the draws classes from the object
class(y) <- class(y)[-(1:2)]
str(y)

}
\seealso{
Other formats: 
\code{\link{draws_array}()},
\code{\link{draws_df}()},
\code{\link{draws_list}()},
\code{\link{draws_matrix}()},
\code{\link{draws_rvars}()}
}
\concept{formats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_draws_matrix.R
\name{[.draws_matrix}
\alias{[.draws_matrix}
\title{Extract parts of a \code{draws_matrix} object}
\usage{
\method{[}{draws_matrix}(x, i, j, ..., drop = FALSE)
}
\arguments{
\item{x, i, j, ..., drop}{Same as in the default extraction method but with
\code{drop} being set to \code{FALSE} by default.}
}
\value{
An object of class \code{"draws_matrix"} unless any of the dimensions
was dropped during the extraction.
}
\description{
Extract parts of a \code{draws_matrix} object. They are strictly defined as
matrices (draws x variable) so dropping any of the
dimensions breaks the expected structure of the object. Accordingly, no
dropping of dimensions is done by default even if the extracted slices are of
length 1. If \code{drop} is manually set to \code{TRUE} and any of the dimensions is
actually dropped, this will lead to dropping the \code{"draws_matrix"} class as
well.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-.R
\name{match}
\alias{match}
\alias{match.default}
\alias{match.rvar}
\alias{\%in\%}
\title{Value Matching}
\usage{
match(x, table, ...)

\method{match}{default}(x, ...)

\method{match}{rvar}(x, ...)

x \%in\% table
}
\arguments{
\item{x}{(multiple options) the values to be matched. Can be:
\itemize{
\item A base vector: see \code{\link[base:match]{base::match()}}
\item An \link{rvar}
}}

\item{table}{(vector) the values to be matched against.}

\item{...}{
  Arguments passed on to \code{\link[base:match]{base::match}}
  \describe{
    \item{\code{nomatch}}{the value to be returned in the case when no match is
    found.  Note that it is coerced to \code{integer}.}
    \item{\code{incomparables}}{a vector of values that cannot be matched.  Any
    value in \code{x} matching a value in this vector is assigned the
    \code{nomatch} value.  For historical reasons, \code{FALSE} is
    equivalent to \code{NULL}.}
  }}
}
\value{
When \code{x} is a base vector, a vector of the same length as \code{x}.

When \code{x} is an \link{rvar}, an \link{rvar} the same shape as \code{x}.
}
\description{
Generic version of \code{\link[base:match]{base::match()}}. For base vectors, returns a vector of the
positions of (first) matches of its first argument in its second. For \link{rvar}s,
returns an \link{rvar} of the matches.
}
\details{
For more information on how match behaves with base vectors, see \code{\link[base:match]{base::match()}}.

When \code{x} is an \link{rvar}, the draws of \code{x} are matched against \code{table} using
\code{\link[base:match]{base::match()}}, and the result is returned as an \link{rvar}.

The implementation of \code{\%in\%} here is identical to \code{base::\%in\%}, except
it uses the generic version of \code{match()} so that non-base vectors (such
as \link{rvar}s) are supported.
}
\examples{
x <- rvar(c("a","b","b","c","d"))
x \%in\% c("b","d")

# for additional examples, see base::match()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/for_each_draw.R
\name{for_each_draw}
\alias{for_each_draw}
\title{Loop over draws}
\usage{
for_each_draw(x, expr)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{expr}{(expression) A bare expression that can contain references to
variables in \code{x} by name. This expression will be executed once per draw
of \code{x}, where references to variables in \code{x} resolve to the value of that
variable in that draw. The expression supports \link{quasiquotation}.}
}
\value{
As \code{for_each_draw()} is used primarily for its side effects (the expression
executed for each draw of \code{x}), it returns the input \code{x} invisibly.
}
\description{
Executes an expression once for every draw in a \code{draws} object. Used
primarily for its side effects and returns the input \code{x} invisibly.
}
\details{
If \code{x} is not in the \code{\link{draws_rvars}} format, it is first converted to that
format. This allows the variables in \code{x} to include their dimensions (i.e,
to act as R vectors and arrays) when being referred to in \code{expr}.

Within \code{expr}, use \code{.draw} to refer to the draw index, which will be a value
between 1 and \code{ndraws(x)}. \code{expr} is executed in the calling environment of
\code{for_each_draw()}, so it can use variables in that environment (however, due
to the use of data masking, to modify variables in that environment, one
must use \verb{<<-}.)
}
\examples{
eight_schools <- as_draws_rvars(example_draws())


# 1. A simple example --- looping over draws and printing each draw
# NOTE: You probably don't want to do this in practice! This example is
# just intended to show what for_each_draw() is doing. If you just want to
# print the draws of an rvar, it is probably better to use draws_of()
for_each_draw(eight_schools, {
  print(mu)
})


# 2. A more complex example --- building a parallel coordinates plot
# First, construct the plot bounds
plot(1, type = "n",
  xlim = c(1, length(eight_schools$theta)),
  ylim = range(range(eight_schools$theta)),
  xlab = "school", ylab = "theta"
)

# Then, use for_each_draw() to make a parallel coordinates plot of all draws
# of eight_schools$theta. Use resample_draws(eight_schools, n = ...)
# in place of eight_schools if a smaller sample is desired for the plot.
for_each_draw(eight_schools, {
  lines(seq_along(theta), theta, col = rgb(1, 0, 0, 0.05))
})

# Finally, add means and 90\% intervals
lines(seq_along(eight_schools$theta), mean(eight_schools$theta))
with(summarise_draws(eight_schools$theta),
  segments(seq_along(eight_schools$theta), y0 = q5, y1 = q95)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{u_scale}
\alias{u_scale}
\title{Rank uniformization}
\usage{
u_scale(x, c = 3/8)
}
\arguments{
\item{x}{(numeric) A scalar, vector, matrix, or array of values.}

\item{c}{(numeric) Fractional offset used in the back-transformation of ranks.
Defaults to \code{3/8}.}
}
\value{
A numeric array of uniformized values with the same size
and dimension as the input.
}
\description{
Compute rank uniformization for a numeric array. First replace each value by
its rank. Average rank for ties are used to conserve the number of unique
values of discrete quantities. Second, uniformize ranks to the scale
\verb{[1/(2S), 1-1/(2S)]}, where \code{S} is the number of values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-dist.R
\name{rvar-dist}
\alias{rvar-dist}
\alias{density.rvar}
\alias{density.rvar_factor}
\alias{cdf.rvar}
\alias{cdf.rvar_factor}
\alias{cdf.rvar_ordered}
\alias{quantile.rvar}
\alias{quantile.rvar_factor}
\alias{quantile.rvar_ordered}
\title{Density, CDF, and quantile functions of random variables}
\usage{
\method{density}{rvar}(x, at, ...)

\method{density}{rvar_factor}(x, at, ...)

\method{cdf}{rvar}(x, q, ...)

\method{cdf}{rvar_factor}(x, q, ...)

\method{cdf}{rvar_ordered}(x, q, ...)

\method{quantile}{rvar}(x, probs, ...)

\method{quantile}{rvar_factor}(x, probs, ...)

\method{quantile}{rvar_ordered}(x, probs, ...)
}
\arguments{
\item{x}{(rvar) An \code{\link{rvar}} object.}

\item{...}{Additional arguments passed onto underlying methods:
\itemize{
\item For \code{density()}, these are passed to \code{\link[stats:density]{stats::density()}}.
\item For \code{cdf()}, these are ignored.
\item For \code{quantile()}, these are passed to \code{\link[stats:quantile]{stats::quantile()}}.
}}

\item{q, at}{(numeric vector) One or more quantiles.}

\item{probs}{(numeric vector) One or more probabilities in \verb{[0,1]}.}
}
\value{
If \code{x} is a scalar \code{\link{rvar}}, returns a vector of the same length as the input
(\code{q}, \code{at}, or \code{probs}) containing values from the corresponding function
of the given \code{\link{rvar}}.

If \code{x} has length greater than 1, returns an array with dimensions
\code{c(length(y), dim(x))} where \code{y} is \code{q}, \code{at}, or \code{probs}, where each
\code{result[i,...]} is the value of the corresponding function,\code{f(y[i])}, for
the corresponding cell in the input array, \code{x[...]}.
}
\description{
The probability density function (\code{density()}), cumulative distribution
function (\code{cdf()}), and quantile function / inverse CDF (\code{quantile()}) of
an \code{\link{rvar}}.
}
\examples{

set.seed(1234)
x = rvar(rnorm(100))

density(x, seq(-2, 2, length.out = 10))
cdf(x, seq(-2, 2, length.out = 10))
quantile(x, ppoints(10))

x2 = c(rvar(rnorm(100, mean = -0.5)), rvar(rnorm(100, mean = 0.5)))
density(x2, seq(-2, 2, length.out = 10))
cdf(x2, seq(-2, 2, length.out = 10))
quantile(x2, ppoints(10))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-math.R
\name{diag,rvar-method}
\alias{diag,rvar-method}
\title{Matrix diagonals (including for random variables)}
\usage{
\S4method{diag}{rvar}(x = 1, nrow, ncol, names = TRUE)
}
\arguments{
\item{x}{(numeric,rvar) a matrix, vector, 1D array, missing, or a 1- or
2-dimensional \code{\link{rvar}}.}

\item{nrow, ncol}{optional dimensions for the result when \code{x} is
    not a matrix.}

\item{names}{(when \code{x} is a matrix) logical indicating if the
    resulting vector, the diagonal of \code{x}, should inherit
    \code{\link[base]{names}} from \code{dimnames(x)} if available.}
}
\value{
For \code{\link{rvar}}s, has two modes:
\enumerate{
\item \code{x} is a matrix-like \code{\link{rvar}}: it returns the diagonal as a vector-like \code{\link{rvar}}
\item \code{x} is a vector-like \code{\link{rvar}}: it returns a matrix-like \code{\link{rvar}} with \code{x} as
the diagonal and zero for off-diagonal entries.
}
}
\description{
Extract the diagonal of a matrix or construct a matrix, including random
matrices (2-dimensional \code{\link{rvar}}s). Makes \code{\link[base:diag]{base::diag()}} generic.
}
\details{
Makes \code{\link[base:diag]{base::diag()}} into a generic function. See that function's documentation
for usage with \code{\link{numeric}}s and for usage of \code{\link{diag<-}}, which is also supported
by \code{\link{rvar}}.
}
\examples{

# Sigma is a 3x3 covariance matrix
Sigma <- as_draws_rvars(example_draws("multi_normal"))$Sigma
Sigma

diag(Sigma)

diag(Sigma) <- 1:3
Sigma

diag(as_rvar(1:3))

}
\seealso{
\code{\link[base:diag]{base::diag()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-summaries-within-draws.R
\name{rvar_is_finite}
\alias{rvar_is_finite}
\alias{rvar_is_infinite}
\alias{rvar_is_nan}
\alias{rvar_is_na}
\title{Special value predicates for random variables}
\usage{
rvar_is_finite(x)

rvar_is_infinite(x)

rvar_is_nan(x)

rvar_is_na(x)
}
\arguments{
\item{x}{(rvar) An \code{\link{rvar}}.}
}
\value{
A logical \code{\link{rvar}} of the same length as the input.
}
\description{
Compute special value predicates (checking for finite / infinite values, \code{NaN}, and \code{NA})
on all draws within a random variable, returning a random variable.
}
\details{
These functions return a new \code{\link{rvar}} that is the result of applying
\code{is.finite()}, \code{is.infinite()}, \code{is.nan()}, or \code{is.na()} to every draw
in the input random variable.
}
\examples{

x <- rvar(c(1, Inf, -Inf, NaN, NA))
x

rvar_is_finite(x)
rvar_is_infinite(x)
rvar_is_nan(x)
rvar_is_na(x)

}
\seealso{
\link{rvar-summaries-over-draws} for summary functions across draws, including
implementations of \code{is.finite()}, \code{is.infinite()}, \code{is.nan()}, and \code{is.na()} for
\code{rvar}s.

Other rvar-summaries: 
\code{\link{rvar-summaries-over-draws}},
\code{\link{rvar-summaries-within-draws}}
}
\concept{rvar-summaries}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nested_rhat.R
\name{rhat_nested}
\alias{rhat_nested}
\alias{rhat_nested.default}
\alias{rhat_nested.rvar}
\title{Nested Rhat convergence diagnostic}
\usage{
rhat_nested(x, ...)

\method{rhat_nested}{default}(x, superchain_ids, ...)

\method{rhat_nested}{rvar}(x, superchain_ids, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{superchain_ids}{(numeric) Vector of length nchains specifying
which superchain each chain belongs to. There should be equal
numbers of chains in each superchain. All chains within the same
superchain are assumed to have been initialized at the same
point.}
}
\value{
If the input is an array, returns a single numeric value. If any of the draws
is non-finite, that is, \code{NA}, \code{NaN}, \code{Inf}, or \code{-Inf}, the returned output
will be (numeric) \code{NA}. Also, if all draws within any of the chains of a
variable are the same (constant), the returned output will be (numeric) \code{NA}
as well. The reason for the latter is that, for constant draws, we cannot
distinguish between variables that are supposed to be constant (e.g., a
diagonal element of a correlation matrix is always 1) or variables that just
happened to be constant because of a failure of convergence or other problems
in the sampling process.

If the input is an \code{\link{rvar}}, returns an array of the same dimensions as the
\code{\link{rvar}}, where each element is equal to the value that would be returned by
passing the draws array for that element of the \code{\link{rvar}} to this function.
}
\description{
Compute the nested Rhat convergence diagnostic for a single
variable as proposed in Margossian et al. (2024).
}
\details{
Nested Rhat is a convergence diagnostic useful when
running many short chains. It is calculated on superchains, which
are groups of chains that have been initialized at the same
point.

Note that there is a slight difference in the calculation of Rhat
and nested Rhat, as nested Rhat is lower bounded by 1. This means
that nested Rhat with one chain per superchain will not be
exactly equal to basic Rhat (see Footnote 3 in Margossian et
al. (2024)).
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
rhat_nested(mu, superchain_ids = c(1, 1, 2, 2))

d <- as_draws_rvars(example_draws("multi_normal"))
rhat_nested(d$Sigma, superchain_ids = c(1, 1, 2, 2))

}
\references{
Charles C. Margossian, Matthew D. Hoffman, Pavel Sountsov, Lionel
Riou-Durand, Aki Vehtari and Andrew Gelman (2024). Nested R-hat:
Assessing the convergence of Markov chain Monte Carlo when running
many short chains. \emph{Bayesian Analysis}. doi:10.1214/24-BA1453
}
\seealso{
Other diagnostics: 
\code{\link{ess_basic}()},
\code{\link{ess_bulk}()},
\code{\link{ess_quantile}()},
\code{\link{ess_sd}()},
\code{\link{ess_tail}()},
\code{\link{mcse_mean}()},
\code{\link{mcse_quantile}()},
\code{\link{mcse_sd}()},
\code{\link{pareto_diags}()},
\code{\link{pareto_khat}()},
\code{\link{rhat}()},
\code{\link{rhat_basic}()},
\code{\link{rstar}()}
}
\concept{diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\encoding{UTF-8}
\name{print.draws_rvars}
\alias{print.draws_rvars}
\title{Print \code{draws_rvars} objects}
\usage{
\method{print}{draws_rvars}(
  x,
  digits = 2,
  max_variables = getOption("posterior.max_variables", 8),
  summary = getOption("posterior.rvar_summary", "mean_sd"),
  reserved = FALSE,
  ...
)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{digits}{(nonnegative integer) The minimum number of significant digits
to print. If \code{NULL}, defaults to \code{getOption("posterior.digits", 2)}.}

\item{max_variables}{(positive integer) The maximum number of variables to
print. Can be controlled globally via the \code{"posterior.max_variables"}
\link[base:options]{option}.}

\item{summary}{(string) The style of summary to display:
\itemize{
\item \code{"mean_sd"} displays \verb{mean ± sd}
\item \code{"median_mad"} displays \verb{median ± mad}
\item \code{"mode_entropy"} displays \verb{mode <entropy>}, and is used automatically for
\code{\link{rvar_factor}}s. It shows normalized entropy, which ranges from
0 (all probability in one category) to 1 (uniform). See \code{\link[=entropy]{entropy()}}.
\item \code{"mode_dissent"} displays \verb{mode <dissent>}, and is used automatically for
\code{\link{rvar_ordered}}s. It shows Tastle and Wierman's (2007) \emph{dissention}
measure, which ranges from 0 (all probability in one category) through
0.5 (uniform) to 1 (bimodal: all probability split equally between the
first and last category). See \code{\link[=dissent]{dissent()}}.
\item \code{NULL} uses \code{getOption("posterior.rvar_summary")} (default \verb{"mean_sd})
}}

\item{reserved}{(logical) Should reserved variables be included in the
output? Defaults to \code{FALSE}. See \code{\link{reserved_variables}} for an overview of
currently reserved variable names.}

\item{...}{Further arguments passed to the underlying \code{\link[=print]{print()}} methods.}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Pretty printing for \code{\link{draws_rvars}} objects.
}
\examples{
x <- as_draws_rvars(example_draws())
print(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{ess_mean}
\alias{ess_mean}
\alias{ess_mean.rvar}
\title{Effective sample size for the mean}
\usage{
ess_mean(x, ...)

\method{ess_mean}{rvar}(x, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
If the input is an array, returns a single numeric value. If any of the draws
is non-finite, that is, \code{NA}, \code{NaN}, \code{Inf}, or \code{-Inf}, the returned output
will be (numeric) \code{NA}. Also, if all draws within any of the chains of a
variable are the same (constant), the returned output will be (numeric) \code{NA}
as well. The reason for the latter is that, for constant draws, we cannot
distinguish between variables that are supposed to be constant (e.g., a
diagonal element of a correlation matrix is always 1) or variables that just
happened to be constant because of a failure of convergence or other problems
in the sampling process.

If the input is an \code{\link{rvar}}, returns an array of the same dimensions as the
\code{\link{rvar}}, where each element is equal to the value that would be returned by
passing the draws array for that element of the \code{\link{rvar}} to this function.
}
\description{
Compute an effective sample size estimate for a mean (expectation)
estimate of a single variable.
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
ess_mean(mu)

d <- as_draws_rvars(example_draws("multi_normal"))
ess_mean(d$Sigma)

}
\references{
Andrew Gelman, John B. Carlin, Hal S. Stern, David B. Dunson, Aki Vehtari and
Donald B. Rubin (2013). \emph{Bayesian Data Analysis, Third Edition}. Chapman and
Hall/CRC.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-math.R
\name{chol.rvar}
\alias{chol.rvar}
\title{Cholesky decomposition of random matrix}
\usage{
\method{chol}{rvar}(x, ...)
}
\arguments{
\item{x}{(rvar) A 2-dimensional \code{\link{rvar}}.}

\item{...}{Additional parameters passed on to \code{chol.tensor()}}
}
\value{
An \code{\link{rvar}} containing the upper triangular factor of the Cholesky
decomposition, i.e., the matrix \eqn{R} such that \eqn{R'R = x}.
}
\description{
Cholesky decomposition of an \code{\link{rvar}} containing a matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-factor.R
\name{is_rvar_factor}
\alias{is_rvar_factor}
\alias{is_rvar_ordered}
\title{Is \code{x} a factor random variable?}
\usage{
is_rvar_factor(x)

is_rvar_ordered(x)
}
\arguments{
\item{x}{(any object) An object to test.}
}
\value{
\code{TRUE} if \code{x} is an \code{\link{rvar_factor}} or \code{\link{rvar_ordered}}, \code{FALSE} otherwise.
}
\description{
Test if \code{x} is an \code{\link{rvar_factor}} or \code{\link{rvar_ordered}}.
}
\seealso{
\code{\link[=as_rvar_factor]{as_rvar_factor()}} and \code{\link[=as_rvar_ordered]{as_rvar_ordered()}} to convert objects to
\code{rvar_factor}s and \code{rvar_ordered}s.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate_variables.R
\name{mutate_variables}
\alias{mutate_variables}
\alias{mutate_variables.draws_matrix}
\alias{mutate_variables.draws_array}
\alias{mutate_variables.draws_df}
\alias{mutate_variables.draws_list}
\alias{mutate_variables.draws_rvars}
\title{Mutate variables in \code{draws} objects}
\usage{
mutate_variables(.x, ...)

\method{mutate_variables}{draws_matrix}(.x, ...)

\method{mutate_variables}{draws_array}(.x, ...)

\method{mutate_variables}{draws_df}(.x, ...)

\method{mutate_variables}{draws_list}(.x, ...)

\method{mutate_variables}{draws_rvars}(.x, ...)
}
\arguments{
\item{.x}{(draws) A \code{\link{draws}} object.}

\item{...}{Name-value pairs of expressions, each with either length 1 or the
same length as in the entire input (i.e., number of iterations or draws).
The name of each argument will be the name of a new variable, and the value
will be its corresponding value. Use a \code{NULL} value in \code{mutate_variables}
to drop a variable. New variables overwrite existing variables of the same
name.}
}
\value{
Returns a \code{\link{draws}} object of the same format as \code{.x}, with variables mutated
according to the expressions provided in \code{...}.
}
\description{
Mutate variables in a \code{\link{draws}} object.
}
\details{
In order to mutate variables in \code{\link{draws_matrix}} and \code{\link{draws_array}} objects,
they are transformed to \code{\link{draws_df}} objects first and then transformed back
after mutation. As those transformations are quite expensive for larger
number of draws, we recommend using \code{mutate_variables} on \code{\link{draws_df}} and
\code{\link{draws_list}} objects if speed is an issue.

In \code{\link{draws_rvars}} objects, the output of each expression in \code{...} is
coerced to an \code{\link{rvar}} object if it is not already one using \code{as_rvar()}.
}
\examples{
x <- as_draws_df(example_draws())
x <- subset(x, variable = c("mu", "tau"))

mutate_variables(x, tau2 = tau^2)
mutate_variables(x, scale = 1.96 * tau, lower = mu - scale)

}
\seealso{
\code{\link{variables}}, \code{\link{rename_variables}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-.R
\name{draws_of}
\alias{draws_of}
\alias{draws_of<-}
\title{Get/set array of draws underlying a random variable}
\usage{
draws_of(x, with_chains = FALSE)

draws_of(x, with_chains = FALSE) <- value
}
\arguments{
\item{x}{(rvar) An \code{\link{rvar}} object.}

\item{with_chains}{(logical) Should the array of draws include a dimension for chains?
If \code{FALSE} (the default), chains are not included and the array has dimension
\code{c(ndraws(x), dim(x))}. If \code{TRUE}, chains are included and the array has
dimension \code{c(niterations(x), nchains(x), dim(x))}.}

\item{value}{(array) An array of values to use as the backing array of \code{x}.}
}
\value{
If \code{with_chains = FALSE}, an array with dimensions \code{c(ndraws(x), dim(x))}.

If \code{with_chains = TRUE}, an array with dimensions
\code{c(niterations(x), nchains(x), dim(x))}.
}
\description{
Gets/sets the array-representation that backs an \code{\link{rvar}}. Should be used rarely.
}
\details{
While \code{\link{rvar}}s implement fast versions of basic math operations (including
\link[=rvar-matmult]{matrix multiplication}), sometimes you may need to bypass
the \code{\link{rvar}} abstraction to do what you need to do more efficiently.
\code{draws_of()} allows you to get / set the underlying array of draws in
order to do that.

\code{\link{rvar}}s represent draws internally using arrays of arbitrary dimension, which
is returned by \code{draws_of(x)} and can be set using \code{draws_of(x) <- value}.
The \strong{first} dimension of these arrays is the index of the draws. If
\code{with_chains = TRUE}, then the dimensions of the returned array are modified
so that the first dimension is the index of the iterations and the second
dimension is the index of the chains.
}
\examples{

x <- rvar(1:10, nchains = 2)
x

# draws_of() without arguments will return the array of draws without
# chain information (first dimension is draw)
draws_of(x)

# draws_of() with with_chains = TRUE will reshape the returned array to
# include chain information in the second dimension
draws_of(x, with_chains = TRUE)

# you can also set draws using draws_of(). When with_chains = FALSE the
# existing chain information will be retained ...
draws_of(x) <- 2:11
x

# when with_chains = TRUE the chain information will be set by the
# second dimension of the assigned array
draws_of(x, with_chains = TRUE) <- array(2:11, dim = c(2,5))
x

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_smooth.R
\name{pareto_khat}
\alias{pareto_khat}
\alias{pareto_khat.default}
\alias{pareto_khat.rvar}
\title{Pareto khat diagnostic}
\usage{
pareto_khat(x, ...)

\method{pareto_khat}{default}(
  x,
  tail = c("both", "right", "left"),
  r_eff = NULL,
  ndraws_tail = NULL,
  verbose = FALSE,
  are_log_weights = FALSE,
  ...
)

\method{pareto_khat}{rvar}(x, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{tail}{(string) The tail to diagnose/smooth:
\itemize{
\item \code{"right"}: diagnose/smooth only the right (upper) tail
\item \code{"left"}: diagnose/smooth only the left (lower) tail
\item \code{"both"}: diagnose/smooth both tails and return the maximum k-hat value
}

The default is \code{"both"}.}

\item{r_eff}{(numeric) relative effective sample size estimate. If
\code{r_eff} is NULL, it will be calculated assuming the draws are
from MCMC. Default is NULL.}

\item{ndraws_tail}{(numeric) number of draws for the tail. If
\code{ndraws_tail} is not specified, it will be calculated as
ceiling(3 * sqrt(length(x) / r_eff)) if length(x) > 225 and
length(x) / 5 otherwise (see Appendix H in Vehtari et
al. (2024)).}

\item{verbose}{(logical) Should diagnostic messages be printed? If
\code{TRUE}, messages related to Pareto diagnostics will be
printed. Default is \code{FALSE}.}

\item{are_log_weights}{(logical) Are the draws log weights? Default is
\code{FALSE}. If \code{TRUE} computation will take into account that the
draws are log weights, and only right tail will be smoothed.}
}
\value{
If the input is an array, returns a single numeric value. If any of the draws
is non-finite, that is, \code{NA}, \code{NaN}, \code{Inf}, or \code{-Inf}, the returned output
will be (numeric) \code{NA}. Also, if all draws within any of the chains of a
variable are the same (constant), the returned output will be (numeric) \code{NA}
as well. The reason for the latter is that, for constant draws, we cannot
distinguish between variables that are supposed to be constant (e.g., a
diagonal element of a correlation matrix is always 1) or variables that just
happened to be constant because of a failure of convergence or other problems
in the sampling process.

If the input is an \code{\link{rvar}}, returns an array of the same dimensions as the
\code{\link{rvar}}, where each element is equal to the value that would be returned by
passing the draws array for that element of the \code{\link{rvar}} to this function.
}
\description{
Estimate Pareto k value by fitting a Generalized Pareto
Distribution to one or two tails of x. This can be used to estimate
the number of fractional moments that is useful for convergence
diagnostics. For further details see Vehtari et al. (2024).
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
pareto_khat(mu)

d <- as_draws_rvars(example_draws("multi_normal"))
pareto_khat(d$Sigma)
}
\references{
Aki Vehtari, Daniel Simpson, Andrew Gelman, Yuling Yao and
Jonah Gabry (2024). Pareto Smoothed Importance Sampling.
\emph{Journal of Machine Learning Research}, 25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\code{\link{pareto_diags}} for additional related diagnostics, and
\code{\link{pareto_smooth}} for Pareto smoothed draws.

Other diagnostics: 
\code{\link{ess_basic}()},
\code{\link{ess_bulk}()},
\code{\link{ess_quantile}()},
\code{\link{ess_sd}()},
\code{\link{ess_tail}()},
\code{\link{mcse_mean}()},
\code{\link{mcse_quantile}()},
\code{\link{mcse_sd}()},
\code{\link{pareto_diags}()},
\code{\link{rhat}()},
\code{\link{rhat_basic}()},
\code{\link{rhat_nested}()},
\code{\link{rstar}()}
}
\concept{diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_smooth.R
\name{pareto_diags}
\alias{pareto_diags}
\alias{pareto_diags.default}
\alias{pareto_diags.rvar}
\alias{pareto_khat_threshold}
\alias{pareto_khat_threshold.default}
\alias{pareto_khat_threshold.rvar}
\alias{pareto_min_ss}
\alias{pareto_min_ss.default}
\alias{pareto_min_ss.rvar}
\alias{pareto_convergence_rate}
\alias{pareto_convergence_rate.default}
\alias{pareto_convergence_rate.rvar}
\title{Pareto smoothing diagnostics}
\usage{
pareto_diags(x, ...)

\method{pareto_diags}{default}(
  x,
  tail = c("both", "right", "left"),
  r_eff = NULL,
  ndraws_tail = NULL,
  verbose = FALSE,
  are_log_weights = FALSE,
  ...
)

\method{pareto_diags}{rvar}(x, ...)

pareto_khat_threshold(x, ...)

\method{pareto_khat_threshold}{default}(x, ...)

\method{pareto_khat_threshold}{rvar}(x, ...)

pareto_min_ss(x, ...)

\method{pareto_min_ss}{default}(x, ...)

\method{pareto_min_ss}{rvar}(x, ...)

pareto_convergence_rate(x, ...)

\method{pareto_convergence_rate}{default}(x, ...)

\method{pareto_convergence_rate}{rvar}(x, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{tail}{(string) The tail to diagnose/smooth:
\itemize{
\item \code{"right"}: diagnose/smooth only the right (upper) tail
\item \code{"left"}: diagnose/smooth only the left (lower) tail
\item \code{"both"}: diagnose/smooth both tails and return the maximum k-hat value
}

The default is \code{"both"}.}

\item{r_eff}{(numeric) relative effective sample size estimate. If
\code{r_eff} is NULL, it will be calculated assuming the draws are
from MCMC. Default is NULL.}

\item{ndraws_tail}{(numeric) number of draws for the tail. If
\code{ndraws_tail} is not specified, it will be calculated as
ceiling(3 * sqrt(length(x) / r_eff)) if length(x) > 225 and
length(x) / 5 otherwise (see Appendix H in Vehtari et
al. (2024)).}

\item{verbose}{(logical) Should diagnostic messages be printed? If
\code{TRUE}, messages related to Pareto diagnostics will be
printed. Default is \code{FALSE}.}

\item{are_log_weights}{(logical) Are the draws log weights? Default is
\code{FALSE}. If \code{TRUE} computation will take into account that the
draws are log weights, and only right tail will be smoothed.}
}
\value{
List of Pareto smoothing diagnostics:
\itemize{
\item \code{khat}: estimated Pareto k shape parameter,
\item \code{min_ss}: minimum sample size for reliable Pareto smoothed estimate,
\item \code{khat_threshold}: khat-threshold for reliable Pareto smoothed estimate,
\item \code{convergence_rate}: Pareto smoothed estimate RMSE convergence rate.
}
}
\description{
Compute diagnostics for Pareto smoothing the tail draws of x by
replacing tail draws by order statistics of a generalized Pareto
distribution fit to the tail(s).
}
\details{
When the fitted Generalized Pareto Distribution is used to
smooth the tail values and these smoothed values are used to
compute expectations, the following diagnostics can give further
information about the reliability of these estimates.
\itemize{
\item \code{min_ss}: Minimum sample size for reliable Pareto smoothed
estimate. If the actual sample size is greater than \code{min_ss}, then
Pareto smoothed estimates can be considered reliable. If the actual
sample size is lower than \code{min_ss}, increasing the sample size
might result in more reliable estimates. For further details, see
Section 3.2.3, Equation 11 in Vehtari et al. (2024).
\item \code{khat_threshold}: Threshold below which k-hat values result in
reliable Pareto smoothed estimates. The threshold is lower for
smaller effective sample sizes. If k-hat is larger than the
threshold, increasing the total sample size may improve reliability
of estimates. For further details, see Section 3.2.4, Equation 13
in Vehtari et al. (2024).
\item \code{convergence_rate}: Relative convergence rate compared to the
central limit theorem. Applicable only if the actual sample size
is sufficiently large (greater than \code{min_ss}). The convergence
rate tells the rate at which the variance of an estimate reduces
when the sample size is increased, compared to the central limit
theorem convergence rate. See Appendix B in Vehtari et al. (2024).
}
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
pareto_diags(mu)

d <- as_draws_rvars(example_draws("multi_normal"))
pareto_diags(d$Sigma)
}
\references{
Aki Vehtari, Daniel Simpson, Andrew Gelman, Yuling Yao and
Jonah Gabry (2024). Pareto Smoothed Importance Sampling.
\emph{Journal of Machine Learning Research}, 25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\code{\link{pareto_khat}}, \code{\link{pareto_min_ss}},
\code{\link{pareto_khat_threshold}}, and \code{\link{pareto_convergence_rate}} for
individual diagnostics; and \code{\link{pareto_smooth}} for Pareto smoothing
draws.

Other diagnostics: 
\code{\link{ess_basic}()},
\code{\link{ess_bulk}()},
\code{\link{ess_quantile}()},
\code{\link{ess_sd}()},
\code{\link{ess_tail}()},
\code{\link{mcse_mean}()},
\code{\link{mcse_quantile}()},
\code{\link{mcse_sd}()},
\code{\link{pareto_khat}()},
\code{\link{rhat}()},
\code{\link{rhat_basic}()},
\code{\link{rhat_nested}()},
\code{\link{rstar}()}
}
\concept{diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gpd.R
\name{qgeneralized_pareto}
\alias{qgeneralized_pareto}
\title{Quantile function for the generalized Pareto distribution}
\usage{
qgeneralized_pareto(
  p,
  mu = 0,
  sigma = 1,
  k = 0,
  lower.tail = TRUE,
  log.p = FALSE
)
}
\arguments{
\item{p}{Numeric vector of probabilities.}

\item{mu}{Location parameter.}

\item{sigma}{Scale parameter (must be positive).}

\item{k}{Shape parameter.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \code{P[X <= x]}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are given as \code{log(p)}.}
}
\value{
A numeric vector of quantiles.
}
\description{
Computes the quantile function for a generalized Pareto distribution
with location \code{mu}, scale \code{sigma}, and shape \code{k}.
}
\examples{
qgeneralized_pareto(p = c(0.1, 0.5, 0.9), mu = 0, sigma = 1, k = 0.2)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/order_draws.R
\name{order_draws}
\alias{order_draws}
\alias{order_draws.draws_matrix}
\alias{order_draws.draws_array}
\alias{order_draws.draws_df}
\alias{order_draws.draws_list}
\alias{order_draws.draws_rvars}
\alias{order_draws.rvar}
\title{Order \code{draws} objects}
\usage{
order_draws(x, ...)

\method{order_draws}{draws_matrix}(x, ...)

\method{order_draws}{draws_array}(x, ...)

\method{order_draws}{draws_df}(x, ...)

\method{order_draws}{draws_list}(x, ...)

\method{order_draws}{draws_rvars}(x, ...)

\method{order_draws}{rvar}(x, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Order \code{\link{draws}} objects according to iteration and chain number. By default,
draws objects are ordered but subsetting or extracting parts of them may
leave them in an unordered state.
}
\examples{
x <- as_draws_array(example_draws())
dimnames(x[10:5, 4:3, ])
dimnames(order_draws(x[10:5, 4:3, ]))

}
\seealso{
\code{\link[=repair_draws]{repair_draws()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-rfun.R
\name{rvar_rng}
\alias{rvar_rng}
\title{Create random variables from existing random number generators}
\usage{
rvar_rng(.f, n, ..., ndraws = NULL)
}
\arguments{
\item{.f}{(function) A function (or string naming a function) representing a
random-number generating function that follows the pattern of base random
number generators (like \code{rnorm()}, \code{rbinom()}, etc). It must:
\itemize{
\item Have a first argument, \code{n}, giving the number of draws to take from the
distribution
\item Have vectorized parameter arguments
\item Return a single vector of length \code{n}
}}

\item{n}{(positive integer) The length of the output \code{\link{rvar}} vector (\strong{not}
the number of draws).}

\item{...}{Arguments passed to \code{.f}. These arguments may include \code{\link{rvar}}s,
so long as they are vectors only (no multidimensional \code{\link{rvar}}s are
allowed).}

\item{ndraws}{(positive integer) The number of draws used to construct the
returned random variable if no \code{\link{rvar}}s are supplied in \code{...}. If \code{NULL},
\code{getOption("posterior.rvar_ndraws")} is used (default 4000). If \code{...}
contains \code{\link{rvar}}s, the number of draws in the provided \code{\link{rvar}}s is used
instead of the value of this argument.}
}
\value{
A single-dimensional \code{\link{rvar}} of length \code{n}.
}
\description{
Specialized alternative to \code{rdo()} or \code{rfun()} for creating \code{\link{rvar}}s from
existing random-number generator functions (such as \code{rnorm()}, \code{rbinom()}, etc).
}
\details{
This function unwraps the arrays underlying the input \code{\link{rvar}}s in
\code{...} and then passes them to \code{.f}, relying on the vectorization of \code{.f}
to evaluate it across draws from the input \code{\link{rvar}}s. This is why the arguments
of \code{.f} \strong{must} be vectorized. It asks for \code{n} times the number of draws
in the input \code{\link{rvar}}s (or \code{ndraws} if none are given) draws from the
random number generator \code{.f}, then reshapes the output from \code{.f} into an
\code{\link{rvar}} with length \code{n}.

\code{rvar_rng()} is a fast alternative to \code{rdo()} or \code{rfun()}, but you \strong{must}
ensure that \code{.f} satisfies the preconditions described above for the result
to be correct. Most base random number generators satisfy these conditions.
It is advisable to test against \code{rdo()} or \code{rfun()} (which should be correct,
but slower) if you are uncertain.
}
\examples{

mu <- rvar_rng(rnorm, 10, mean = 1:10, sd = 1)
sigma <- rvar_rng(rgamma, 1, shape = 1, rate = 1)
x <- rvar_rng(rnorm, 10, mu, sigma)
x

}
\seealso{
Other rfun: 
\code{\link{rdo}()},
\code{\link{rfun}()}
}
\concept{rfun}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-rfun.R
\name{rdo}
\alias{rdo}
\title{Execute expressions of random variables}
\usage{
rdo(expr, dim = NULL, ndraws = NULL)
}
\arguments{
\item{expr}{(expression) A bare expression that can (optionally) contain
\code{\link{rvar}}s. The expression supports \link{quasiquotation}.}

\item{dim}{(integer vector) One or more integers giving the maximal indices
in each dimension to override the dimensions of the \code{\link{rvar}} to be created
(see \code{\link[=dim]{dim()}}). If \code{NULL} (the default), \code{dim} is determined by the input.
\strong{NOTE:} This argument controls the dimensions of the \code{\link{rvar}}, not the
underlying array, so you cannot change the number of draws using this
argument.}

\item{ndraws}{(positive integer) The number of draws used to construct new
random variables if no \code{\link{rvar}}s are supplied in \code{expr}. If \code{NULL},
\code{getOption("posterior.rvar_ndraws")} is used (default 4000). If \code{expr}
contains \code{\link{rvar}}s, the number of draws in the provided \code{\link{rvar}}s is used
instead of the value of this argument.}
}
\value{
An \code{\link{rvar}}.
}
\description{
Execute (nearly) arbitrary \R expressions that may include \code{\link{rvar}}s,
producing a new \code{\link{rvar}}.
}
\details{
This function evaluates \code{expr} possibly multiple times, once for each draw of
the \code{\link{rvar}}s it contains, then returns a new \code{\link{rvar}} representing the output of those
expressions. To identify \code{\link{rvar}}s, \code{rdo()} searches the calling environment for any variables
named in \code{expr} for which \code{\link[=is_rvar]{is_rvar()}} evaluates to \code{TRUE}. If \code{expr} contains no \code{\link{rvar}}s,
then it will be executed \code{ndraws} times and an \code{\link{rvar}} with that many draws returned.

\code{rdo()} is not necessarily \emph{fast} (in fact in some cases it may be very slow), but
it has the advantage of allowing a nearly arbitrary R expression to be executed against \code{\link{rvar}}s
simply by wrapping it with \code{rdo( ... )}. This makes it especially useful as a prototyping
tool. If you create code with \code{rdo()} and it is unacceptably slow for your application,
consider rewriting it using math operations directly on \code{\link{rvar}}s (which should be fast),
using \code{\link[=rvar_rng]{rvar_rng()}}, and/or using operations directly on the arrays that back the \code{\link{rvar}}s
(via \code{\link[=draws_of]{draws_of()}}).
}
\examples{

mu <- rdo(rnorm(10, mean = 1:10, sd = 1))
sigma <- rdo(rgamma(1, shape = 1, rate = 1))
x <- rdo(rnorm(10, mu, sigma))
x

}
\seealso{
Other rfun: 
\code{\link{rfun}()},
\code{\link{rvar_rng}()}
}
\concept{rfun}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/posterior-package.R
\docType{package}
\name{posterior-package}
\alias{posterior-package}
\alias{posterior}
\title{Tools for working with posterior (and prior) distributions}
\description{
\if{html}{
   \figure{stanlogo.png}{options: width="50"}
   \url{https://mc-stan.org/posterior/}
}

The \pkg{posterior} package is intended to provide useful tools
for both users and developers of packages for fitting Bayesian models or
working with output from Bayesian models. The primary goals of the package
are to:
\itemize{
\item Efficiently convert between many different useful formats of
draws (samples) from posterior or prior distributions.
\item Provide consistent methods for operations commonly performed on draws,
for example, subsetting, binding, or mutating draws.
\item Provide various summaries of draws in convenient formats.
\item Provide lightweight implementations of state of the art posterior inference
diagnostics.
}
}
\section{Package options}{


The following options are used to format and print \code{\link{draws}} objects,
as in \code{print.draws_array()}, \code{print.draws_df()}, \code{print.draws_list()},
\code{print.draws_matrix()}, and \code{print.draws_rvars()}:
\itemize{
\item \code{posterior.max_draws}: Maximum number of draws to print.
\item \code{posterior.max_iterations}: Maximum number of iterations to print.
\item \code{posterior.max_chains}: Maximum number of chains to print.
\item \code{posterior.max_variables}: Maximum number of variables to print.
}

The following options are used to format and print \code{\link{rvar}} objects,
as in \code{print.rvar()} and \code{print.draws_rvars()}:
\itemize{
\item \code{posterior.rvar_summary}: What style of summary to display:
\code{"mean_sd"} displays \verb{mean ± sd}, \code{"median_mad"} displays \verb{median ± mad}.
\item \code{posterior.digits}: How many significant digits are displayed. This
defaults to a smaller value (\code{2}) than \code{getOption("digits")} because
\code{\link{rvar}}s print two numbers (point summary and uncertainty) next to
each other.
}

The following option is used to construct new \code{\link{rvar}} objects,
as in \code{rfun()} and \code{rdo()}:
\itemize{
\item \code{posterior.rvar_ndraws}: The number of draws used to construct
new random variables when this number cannot be determined
from existing arguments (e.g., other \code{\link{rvar}}s passed to a function).
}

The following options are used to control warning messages:
\itemize{
\item \code{posterior.warn_on_merge_chains}: (logical) Some operations will
trigger an automatic merging of chains, for example, because chains do not
match between two objects involved in a binary operation. Whether this
causes a warning can be controlled by this option.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-dim.R
\name{drop,rvar-method}
\alias{drop,rvar-method}
\title{Drop redundant dimensions}
\usage{
\S4method{drop}{rvar}(x)
}
\arguments{
\item{x}{(rvar) an \code{\link{rvar}}.}
}
\value{
An \code{\link{rvar}} with the same length as \code{x}, but where any entry equal to \code{1}
in \code{dim(x)} has been removed. The exception is if \code{dim(x) == 1}, in which
case \code{dim(drop(x)) == 1} as well (this is because \code{\link{rvar}}s, unlike \code{\link{numeric}}s,
never have \code{NULL} dimensions).
}
\description{
Delete the dimensions of an \code{\link{rvar}} which are of size one. See \code{\link[base:drop]{base::drop()}}
}
\examples{
# Sigma is a 3x3 covariance matrix
Sigma <- as_draws_rvars(example_draws("multi_normal"))$Sigma
Sigma

Sigma[1, ]

drop(Sigma[1, ])

# equivalently ...
Sigma[1, drop = TRUE]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_variable.R
\name{extract_variable}
\alias{extract_variable}
\alias{extract_variable.default}
\alias{extract_variable.draws}
\alias{extract_variable.draws_df}
\alias{extract_variable.draws_list}
\alias{extract_variable.draws_rvars}
\title{Extract draws of a single variable}
\usage{
extract_variable(x, variable, ...)

\method{extract_variable}{default}(x, variable, ...)

\method{extract_variable}{draws}(x, variable, ...)

\method{extract_variable}{draws_df}(x, variable, ...)

\method{extract_variable}{draws_list}(x, variable, ...)

\method{extract_variable}{draws_rvars}(x, variable, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{variable}{(string) The name of the variable to extract. The name must
correspond to a scalar variable or must include indices for array variables
(e.g. \code{"x[1]"}, \code{"y[1,2]"}) that result in a scalar variable. To extract
all dimensions from variables with indices, use \code{\link[=extract_variable_array]{extract_variable_array()}}.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A vector of length equal to the number of draws.
}
\description{
Extract a vector of draws of a single variable.
}
\examples{
x <- example_draws()
mu <- extract_variable(x, variable = "mu")
str(mu)

}
\seealso{
Other variable extraction methods: 
\code{\link{extract_list_of_variable_arrays}()},
\code{\link{extract_variable_array}()},
\code{\link{extract_variable_matrix}()}
}
\concept{variable extraction methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_draws_matrix.R
\name{draws_matrix}
\alias{draws_matrix}
\alias{as_draws_matrix}
\alias{as_draws_matrix.default}
\alias{as_draws_matrix.draws_matrix}
\alias{as_draws_matrix.draws_array}
\alias{as_draws_matrix.draws_df}
\alias{as_draws_matrix.draws_list}
\alias{as_draws_matrix.draws_rvars}
\alias{as_draws_matrix.mcmc}
\alias{as_draws_matrix.mcmc.list}
\alias{is_draws_matrix}
\title{The \code{draws_matrix} format}
\usage{
as_draws_matrix(x, ...)

\method{as_draws_matrix}{default}(x, ...)

\method{as_draws_matrix}{draws_matrix}(x, ...)

\method{as_draws_matrix}{draws_array}(x, ...)

\method{as_draws_matrix}{draws_df}(x, ...)

\method{as_draws_matrix}{draws_list}(x, ...)

\method{as_draws_matrix}{draws_rvars}(x, ...)

\method{as_draws_matrix}{mcmc}(x, ...)

\method{as_draws_matrix}{mcmc.list}(x, ...)

draws_matrix(..., .nchains = 1)

is_draws_matrix(x)
}
\arguments{
\item{x}{An object to convert to a \code{draws_matrix} object.}

\item{...}{For \code{as_draws_matrix()}:
Arguments passed to individual methods (if applicable).
For \code{draws_matrix()}: Named
arguments containing numeric vectors each defining a separate variable.}

\item{.nchains}{(positive integer) The number of chains. The default is \code{1}.}
}
\value{
A \code{draws_matrix} object, which has classes
\code{c("draws_matrix", "draws", "matrix")}.
}
\description{
The \code{as_draws_matrix()} methods convert
objects to the \code{draws_matrix} format.
The \code{draws_matrix()} function creates an object of the
\code{draws_matrix} format based on a set of numeric vectors.
See \strong{Details}.
}
\details{
Objects of class \code{"draws_matrix"} are matrices (2-D arrays) with
dimensions \code{"draw"} and \code{"variable"}. See \strong{Examples}.
}
\examples{
x1 <- as_draws_matrix(example_draws())
class(x1)
print(x1)
str(x1)

x2 <- draws_matrix(a = rnorm(10), b = rnorm(10), c = 1)
class(x2)
print(x2)
str(x2)
}
\seealso{
Other formats: 
\code{\link{draws_array}()},
\code{\link{draws_df}()},
\code{\link{draws_list}()},
\code{\link{draws_rvars}()},
\code{\link{draws}}
}
\concept{formats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{autocorrelation}
\alias{autocorrelation}
\title{Autocorrelation estimates}
\usage{
autocorrelation(x)
}
\arguments{
\item{x}{(numeric vector) A sequence of values.}
}
\value{
A numeric vector of autocorrelations at every lag (scaled by N-lag).
}
\description{
Compute autocorrelation estimates for every lag for the specified
input sequence using a fast Fourier transform approach. The estimate
for lag t is scaled by N-t where N is the length of the sequence.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_smooth.R
\name{ps_khat_threshold}
\alias{ps_khat_threshold}
\title{Pareto k-hat threshold}
\usage{
ps_khat_threshold(ndraws, ...)
}
\arguments{
\item{ndraws}{number of draws}

\item{...}{unused}
}
\value{
threshold
}
\description{
Given number of draws, computes khat threshold for reliable Pareto
smoothed estimate (to have small probability of large error). See
section 3.2.4, equation (13) of Vehtari et al. (2024). This
function is exported to be usable by other packages. For
user-facing diagnostic functions, see \code{\link{pareto_khat_threshold}} and
\code{\link{pareto_diags}}.
}
\seealso{
Other helper-functions: 
\code{\link{ps_convergence_rate}()},
\code{\link{ps_min_ss}()},
\code{\link{ps_tail_length}()}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{z_scale}
\alias{z_scale}
\title{Rank normalization}
\usage{
z_scale(x, c = 3/8)
}
\arguments{
\item{x}{(numeric) A scalar, vector, matrix, or array of values.}

\item{c}{(numeric) Fractional offset used in the back-transformation of ranks.
Defaults to \code{3/8}.}
}
\value{
A numeric array of rank normalized values with the same size
and dimension as the input.
}
\description{
Compute rank normalization for a numeric array. First replace each
value by its rank. Average rank for ties are used to conserve the
number of unique values of discrete quantities. Second, normalize
ranks via the inverse normal transformation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset_draws.R
\name{subset_draws}
\alias{subset_draws}
\alias{subset_draws.draws_matrix}
\alias{subset_draws.draws_array}
\alias{subset_draws.draws_df}
\alias{subset_draws.draws_list}
\alias{subset_draws.draws_rvars}
\alias{subset_draws.rvar}
\alias{subset.draws}
\title{Subset \code{draws} objects}
\usage{
subset_draws(x, ...)

\method{subset_draws}{draws_matrix}(
  x,
  variable = NULL,
  iteration = NULL,
  chain = NULL,
  draw = NULL,
  regex = FALSE,
  unique = TRUE,
  ...
)

\method{subset_draws}{draws_array}(
  x,
  variable = NULL,
  iteration = NULL,
  chain = NULL,
  draw = NULL,
  regex = FALSE,
  unique = TRUE,
  ...
)

\method{subset_draws}{draws_df}(
  x,
  variable = NULL,
  iteration = NULL,
  chain = NULL,
  draw = NULL,
  regex = FALSE,
  unique = TRUE,
  ...
)

\method{subset_draws}{draws_list}(
  x,
  variable = NULL,
  iteration = NULL,
  chain = NULL,
  draw = NULL,
  regex = FALSE,
  unique = TRUE,
  ...
)

\method{subset_draws}{draws_rvars}(
  x,
  variable = NULL,
  iteration = NULL,
  chain = NULL,
  draw = NULL,
  regex = FALSE,
  unique = TRUE,
  ...
)

\method{subset_draws}{rvar}(x, variable = NULL, ...)

\method{subset}{draws}(x, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{variable}{(character vector) The variables to select. All elements of
non-scalar variables can be selected at once.}

\item{iteration}{(integer vector) The iteration indices to select.}

\item{chain}{(integer vector) The chain indices to select.}

\item{draw}{(integer vector) The draw indices to be select. Subsetting draw
indices will lead to an automatic merging of chains via \code{\link{merge_chains}}.}

\item{regex}{(logical) Should \code{variable} should be treated as a
(vector of) regular expressions? Any variable in \code{x} matching at least one
of the regular expressions will be selected. Defaults to \code{FALSE}.}

\item{unique}{(logical) Should duplicated selection of chains, iterations, or
draws be allowed? If \code{TRUE} (the default) only unique chains, iterations,
and draws are selected regardless of how often they appear in the
respective selecting arguments.}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Subset \code{\link{draws}} objects by variables, iterations, chains, and draws indices.
}
\details{
To ensure that multiple consecutive subsetting operations work correctly,
\code{subset()} \emph{\link[=repair_draws]{repairs}} the \code{draws} object before and after
subsetting.
}
\examples{
x <- example_draws()
subset_draws(x, variable = c("mu", "tau"))
subset_draws(x, chain = 2)
subset_draws(x, iteration = 5:10, chain = 3:4)

# extract the first chain twice
subset_draws(x, chain = c(1, 1), unique = FALSE)

# extract all elements of 'theta'
subset_draws(x, variable = "theta")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-factor.R
\name{rvar_factor}
\alias{rvar_factor}
\alias{rvar_ordered}
\title{Factor random variables of arbitrary dimension}
\usage{
rvar_factor(
  x = factor(),
  dim = NULL,
  dimnames = NULL,
  nchains = NULL,
  with_chains = FALSE,
  ...
)

rvar_ordered(
  x = ordered(NULL),
  dim = NULL,
  dimnames = NULL,
  nchains = NULL,
  with_chains = FALSE,
  ...
)
}
\arguments{
\item{x}{(multiple options) The object to convert to an \code{rvar}:
\itemize{
\item A vector of draws from a distribution.
\item An array where the first dimension represents draws from a distribution.
The resulting \code{\link{rvar}} will have dimension \code{dim(x)[-1]}; that is,
everything except the first dimension is used for the shape of the
variable, and the first dimension is used to index draws from the
distribution (see \strong{Examples}). Optionally,
if \code{with_chains == TRUE}, the first dimension indexes the iteration and the
second dimension indexes the chain (see \code{with_chains}).
\item An \code{rvar}.
}}

\item{dim}{(integer vector) One or more integers giving the maximal indices
in each dimension to override the dimensions of the \code{\link{rvar}} to be created
(see \code{\link[=dim]{dim()}}). If \code{NULL} (the default), \code{dim} is determined by the input.
\strong{NOTE:} This argument controls the dimensions of the \code{\link{rvar}}, not the
underlying array, so you cannot change the number of draws using this
argument.}

\item{dimnames}{(list) Character vectors giving the names in each dimension
to override the names of the dimensions of the \code{\link{rvar}} to be created (see
\code{\link[=dimnames]{dimnames()}}). If \code{NULL} (the default), this is determined by the input.
\strong{NOTE:} This argument controls the names of the dimensions of the
\code{\link{rvar}}, not the underlying array.}

\item{nchains}{(positive integer) The number of chains. The if \code{NULL} (the default),
\code{1} is used unless \code{x} is already an \code{\link{rvar}}, in which case the number of
chains it has is used.}

\item{with_chains}{(logical) Does \code{x} include a dimension for chains?
If \code{FALSE} (the default), chains are not included, the first dimension of
the input array should index draws, and the \code{nchains} argument can be
used to determine the number of chains. If \code{TRUE}, the \code{nchains} argument
is ignored and the second dimension of \code{x} is used to index chains.
Internally, the array will be converted to a format without the chain index.
Ignored when \code{x} is already an \code{\link{rvar}}.}

\item{...}{
  Arguments passed on to \code{\link[base:factor]{base::factor}}
  \describe{
    \item{\code{levels}}{an optional vector of the unique values (as character strings)
    that \code{x} might have taken.  The default is the unique set of
    values taken by \code{\link[base]{as.character}(x)}, sorted into
    increasing order \emph{of \code{x}}.  Note that this set can be
    specified as smaller than \code{sort(unique(x))}.}
    \item{\code{labels}}{\emph{either} an optional character vector of
    labels for the levels (in the same order as \code{levels} after
    removing those in \code{exclude}), \emph{or} a character string of
    length 1.  Duplicated values in \code{labels} can be used to map
    different values of \code{x} to the same factor level.}
    \item{\code{exclude}}{a vector of values to be excluded when forming the
    set of levels.  This may be factor with the same level set as \code{x}
    or should be a \code{character}.}
    \item{\code{ordered}}{logical flag to determine if the levels should be regarded
    as ordered (in the order given).}
    \item{\code{nmax}}{an upper bound on the number of levels; see \sQuote{Details}.}
  }}
}
\value{
An object of class \code{"rvar_factor"} representing a \code{factor}-like random variable.
}
\description{
Random variables backed by \link{factor}-like arrays of arbitrary dimension.
}
\details{
A subtype of \code{\link[=rvar]{rvar()}} that represents a (possibly multidimensional) sample of
a \link{factor} or an \link{ordered} factor. It is otherwise very similar to the basic \code{\link[=rvar]{rvar()}}:
it is backed by a multidimensional array with draws as the first dimension.
The primary difference is that the backing array has class \code{"factor"} (for \code{\link[=rvar_factor]{rvar_factor()}})
or \code{c("ordered", "factor")} (for \code{\link[=rvar_ordered]{rvar_ordered()}}). If you
pass a \link{factor} or \link{ordered} factor to \code{\link[=rvar]{rvar()}} it will automatically return
an object with the classes \code{"rvar_factor"} or \code{c("rvar_ordered", "rvar_factor")}.

See \code{\link[=rvar]{rvar()}} for more details on the internals of the random variable datatype.
}
\examples{
set.seed(1234)

# To create a "scalar" `rvar_factor`, pass a one-dimensional array or a vector
# whose length (here `4000`) is the desired number of draws:
x <- rvar(sample(c("a","a","a","b","c"), 4000, replace = TRUE))
x

# Create random vectors by adding an additional dimension:
x_array <- array(c(
    sample(c("a","a","a","b","c"), 4000, replace = TRUE),
    sample(c("a","a","b","c","c"), 4000, replace = TRUE),
    sample(c("b","b","b","b","c"), 4000, replace = TRUE),
    sample(c("d","d","b","b","c"), 4000, replace = TRUE)
  ), dim = c(4000, 4))
rvar_factor(x_array)

# You can also create ordered factors
rvar_ordered(x_array)

# arguments of factor() and ordered() are passed down by the constructor
# e.g. we can reorder levels of an ordered factor:
rvar_ordered(x_array, levels = c("d","c","b","a"))

# Unlike base factors, rvar factors can be matrices or arrays:
rvar_factor(x_array, dim = c(2, 2))

# If the input to rvar_factor() is an array with a `"levels"` attribute, it
# will use those as the levels of the factor
y_array <- t(array(rbinom(3000, 1, c(0.1, 0.5, 0.9)) + 1, dim = c(3, 1000)))
rvar(y_array)
# with levels
attr(y_array, "levels") = c("a", "b")
rvar_factor(y_array)

}
\seealso{
\code{\link[=as_rvar_factor]{as_rvar_factor()}} to convert objects to \code{rvar_factor}s. See \code{\link[=rdo]{rdo()}}, \code{\link[=rfun]{rfun()}}, and
\code{\link[=rvar_rng]{rvar_rng()}} for higher-level interfaces for creating \code{rvar}s.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_variable_matrix.R
\name{extract_variable_matrix}
\alias{extract_variable_matrix}
\alias{extract_variable_matrix.default}
\alias{extract_variable_matrix.draws}
\alias{extract_variable_matrix.draws_rvars}
\title{Extract matrix of a single variable}
\usage{
extract_variable_matrix(x, variable, ...)

\method{extract_variable_matrix}{default}(x, variable, ...)

\method{extract_variable_matrix}{draws}(x, variable, ...)

\method{extract_variable_matrix}{draws_rvars}(x, variable, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{variable}{(string) The name of the variable to extract.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A \code{matrix} with dimension iterations x chains.
}
\description{
Extract an iterations x chains matrix of draws of a single variable.
This is primarily used for convergence diagnostic functions such as \code{\link[=rhat]{rhat()}}.
}
\examples{
x <- example_draws()
mu <- extract_variable_matrix(x, variable = "mu")
dim(mu)
rhat(mu)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{ess_bulk}
\alias{ess_bulk}
\alias{ess_bulk.default}
\alias{ess_bulk.rvar}
\title{Bulk effective sample size (bulk-ESS)}
\usage{
ess_bulk(x, ...)

\method{ess_bulk}{default}(x, ...)

\method{ess_bulk}{rvar}(x, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
If the input is an array, returns a single numeric value. If any of the draws
is non-finite, that is, \code{NA}, \code{NaN}, \code{Inf}, or \code{-Inf}, the returned output
will be (numeric) \code{NA}. Also, if all draws within any of the chains of a
variable are the same (constant), the returned output will be (numeric) \code{NA}
as well. The reason for the latter is that, for constant draws, we cannot
distinguish between variables that are supposed to be constant (e.g., a
diagonal element of a correlation matrix is always 1) or variables that just
happened to be constant because of a failure of convergence or other problems
in the sampling process.

If the input is an \code{\link{rvar}}, returns an array of the same dimensions as the
\code{\link{rvar}}, where each element is equal to the value that would be returned by
passing the draws array for that element of the \code{\link{rvar}} to this function.
}
\description{
Compute a bulk effective sample size estimate (bulk-ESS) for a single
variable. Bulk-ESS is useful as a diagnostic for the sampling efficiency in
the bulk of the posterior. It is defined as the effective sample size for
rank normalized values using split chains. For the tail effective sample size
see \code{\link[=ess_tail]{ess_tail()}}. See Vehtari (2021) for an in-depth
comparison of different effective sample size estimators.
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
ess_bulk(mu)

d <- as_draws_rvars(example_draws("multi_normal"))
ess_bulk(d$Sigma)

}
\references{
Aki Vehtari, Andrew Gelman, Daniel Simpson, Bob Carpenter, and
Paul-Christian Bürkner (2021). Rank-normalization, folding, and
localization: An improved R-hat for assessing convergence of
MCMC (with discussion). \emph{Bayesian Data Analysis}. 16(2), 667-–718.
doi:10.1214/20-BA1221

Aki Vehtari (2021). Comparison of MCMC effective sample size estimators.
Retrieved from https://avehtari.github.io/rhat_ess/ess_comparison.html
}
\seealso{
Other diagnostics: 
\code{\link{ess_basic}()},
\code{\link{ess_quantile}()},
\code{\link{ess_sd}()},
\code{\link{ess_tail}()},
\code{\link{mcse_mean}()},
\code{\link{mcse_quantile}()},
\code{\link{mcse_sd}()},
\code{\link{rhat_basic}()},
\code{\link{rhat}()},
\code{\link{rstar}()}
}
\concept{diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_draws_array.R
\name{draws_array}
\alias{draws_array}
\alias{as_draws_array}
\alias{as_draws_array.default}
\alias{as_draws_array.draws_array}
\alias{as_draws_array.draws_matrix}
\alias{as_draws_array.draws_df}
\alias{as_draws_array.draws_list}
\alias{as_draws_array.draws_rvars}
\alias{as_draws_array.mcmc}
\alias{as_draws_array.mcmc.list}
\alias{is_draws_array}
\title{The \code{draws_array} format}
\usage{
as_draws_array(x, ...)

\method{as_draws_array}{default}(x, ...)

\method{as_draws_array}{draws_array}(x, ...)

\method{as_draws_array}{draws_matrix}(x, ...)

\method{as_draws_array}{draws_df}(x, ...)

\method{as_draws_array}{draws_list}(x, ...)

\method{as_draws_array}{draws_rvars}(x, ...)

\method{as_draws_array}{mcmc}(x, ...)

\method{as_draws_array}{mcmc.list}(x, ...)

draws_array(..., .nchains = 1)

is_draws_array(x)
}
\arguments{
\item{x}{An object to convert to a \code{draws_array} object.}

\item{...}{For \code{as_draws_array()}:
Arguments passed to individual methods (if applicable).
For \code{draws_array()}: Named
arguments containing numeric vectors each defining a separate variable.}

\item{.nchains}{(positive integer) The number of chains. The default is \code{1}.}
}
\value{
A \code{draws_array} object, which has classes
\code{c("draws_array", "draws", "array")}.
}
\description{
The \code{as_draws_array()} methods convert
objects to the \code{draws_array} format.
The \code{draws_array()} function creates an object of the
\code{draws_array} format based on a set of numeric vectors.
See \strong{Details}.
}
\details{
Objects of class \code{"draws_array"} are 3-D arrays with dimensions
\code{"iteration"}, \code{"chain"}, and \code{"variable"}. See \strong{Examples}.
}
\examples{
x1 <- as_draws_array(example_draws())
class(x1)
print(x1)
str(x1)

x2 <- draws_array(a = rnorm(10), b = rnorm(10), c = 1)
class(x2)
print(x2)
str(x2)
}
\seealso{
Other formats: 
\code{\link{draws}},
\code{\link{draws_df}()},
\code{\link{draws_list}()},
\code{\link{draws_matrix}()},
\code{\link{draws_rvars}()}
}
\concept{formats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{diagnostics}
\alias{diagnostics}
\alias{convergence}
\title{List of available convergence diagnostics}
\value{
See individual functions for a description of return types.
}
\description{
A list of available diagnostics and links to their individual help pages.
}
\details{
\tabular{ll}{
   \strong{Function} \tab \strong{Description} \cr
   \code{\link[=ess_basic]{ess_basic()}} \tab Basic version of effective sample size \cr
   \code{\link[=ess_bulk]{ess_bulk()}} \tab Bulk effective sample size \cr
   \code{\link[=ess_tail]{ess_tail()}} \tab Tail effective sample size \cr
   \code{\link[=ess_quantile]{ess_quantile()}} \tab Effective sample sizes for quantiles \cr
   \code{\link[=ess_sd]{ess_sd()}} \tab Effective sample sizes for standard deviations \cr
   \code{\link[=mcse_mean]{mcse_mean()}} \tab Monte Carlo standard error for the mean \cr
   \code{\link[=mcse_quantile]{mcse_quantile()}} \tab Monte Carlo standard error for quantiles \cr
   \code{\link[=mcse_sd]{mcse_sd()}} \tab Monte Carlo standard error for standard deviations \cr
   \code{\link[=rhat_basic]{rhat_basic()}} \tab Basic version of Rhat \cr
   \code{\link[=rhat]{rhat()}} \tab Improved, rank-based version of Rhat \cr
   \code{\link[=rstar]{rstar()}} \tab R* diagnostic \cr
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reserved_variables.R
\name{reserved_variables}
\alias{reserved_variables}
\alias{reserved_variables.default}
\alias{reserved_variables.draws_matrix}
\alias{reserved_variables.draws_array}
\alias{reserved_variables.draws_df}
\alias{reserved_variables.draws_list}
\alias{reserved_variables.draws_rvars}
\title{Reserved variables}
\usage{
reserved_variables(x, ...)

\method{reserved_variables}{default}(x, ...)

\method{reserved_variables}{draws_matrix}(x, ...)

\method{reserved_variables}{draws_array}(x, ...)

\method{reserved_variables}{draws_df}(x, ...)

\method{reserved_variables}{draws_list}(x, ...)

\method{reserved_variables}{draws_rvars}(x, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A character vector of reserved variables used in \code{x}.
}
\description{
Get names of reserved variables from objects in the \pkg{posterior} package.
}
\details{
\code{reserved_variables()} returns the names of reserved variables in use by
an object.

The following variables names are currently reserved for special use cases
in all \code{\link{draws}} formats:
\itemize{
\item \code{.log_weight}: Log weights per draw (see \code{\link{weight_draws}}).
}

Further, specific for the \code{\link{draws_df}} format, there are three additional
reserved variables:
\itemize{
\item \code{.chain}: Chain index per draw
\item \code{.iteration}: Iteration index within each chain
\item \code{.draw}: Draw index across chains
}

More reserved variables may be added in the future.
}
\examples{

x <- example_draws()
reserved_variables(x)

# if we add weights, the `.log_weight` reserved variable is used
x <- weight_draws(x, rexp(ndraws(x)))
reserved_variables(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/thin_draws.R
\name{thin_draws}
\alias{thin_draws}
\alias{thin}
\alias{thin_draws.draws}
\alias{thin_draws.rvar}
\title{Thin \code{draws} objects}
\usage{
thin_draws(x, thin = NULL, ...)

\method{thin_draws}{draws}(x, thin = NULL, ...)

\method{thin_draws}{rvar}(x, thin = NULL, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{thin}{(positive numeric) The period for selecting draws. Must
be between 1 and the number of iterations. If the value is not an
integer, the draws will be selected such that the number of draws
returned is equal to round(ndraws(x) / thin). Intervals between
selected draws will be either ceiling(thin) or floor(thin), such
that the average interval will be close to the thin value. If
\code{NULL}, it will be automatically calculated based on bulk and
tail effective sample size as suggested by Säilynoja et
al. (2022).}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Thin \code{\link{draws}} objects to reduce their size and autocorrelation in
the chains.
}
\examples{
x <- example_draws()
niterations(x)

x <- thin_draws(x, thin = 5)
niterations(x)

}
\references{
Teemu Säilynoja, Paul-Christian Bürkner, and Aki Vehtari (2022).
Graphical test for discrete uniformity and its applications in
goodness-of-fit evaluation and multiple sample comparison.
\emph{Statistics and Computing}. 32, 32. doi:10.1007/s11222-022-10090-6
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split_chains.R
\name{split_chains}
\alias{split_chains}
\title{Split Chains}
\usage{
split_chains(x, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Split chains by halving the number of iterations per chain and doubling the
number of chains.
}
\examples{
x <- example_draws()
niterations(x)
nchains(x)

x <- split_chains(x)
niterations(x)
nchains(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summarise_draws.R
\name{draws_summary}
\alias{draws_summary}
\alias{summarise_draws}
\alias{summarize_draws}
\alias{summarise_draws.draws}
\alias{summary.draws}
\alias{summarise_draws.rvar}
\alias{summary.rvar}
\alias{default_summary_measures}
\alias{default_convergence_measures}
\alias{default_mcse_measures}
\title{Summaries of \code{draws} objects}
\usage{
summarise_draws(.x, ...)

summarize_draws(.x, ...)

\method{summarise_draws}{draws}(
  .x,
  ...,
  .args = list(),
  .num_args = getOption("posterior.num_args", list()),
  .cores = 1
)

\method{summary}{draws}(object, ...)

\method{summarise_draws}{rvar}(.x, ...)

\method{summary}{rvar}(object, ...)

default_summary_measures()

default_convergence_measures()

default_mcse_measures()
}
\arguments{
\item{.x, object}{(draws) A \code{draws} object or one coercible to a \code{draws} object.}

\item{...}{Name-value pairs of summary or \link[=diagnostics]{diagnostic}
functions. The provided names will be used as the names of the columns in
the result \emph{unless} the function returns a named vector, in which case the
latter names are used. The functions can be specified in any format
supported by \link[rlang:as_function]{as_function()}. See \strong{Examples}.}

\item{.args}{(named list) Optional arguments passed to the summary functions.}

\item{.num_args}{(named list) Optional arguments passed to
\link[tibble:num]{num()} for pretty printing of summaries. Can be controlled
globally via the \code{posterior.num_args} \link[base:options]{option}.}

\item{.cores}{(positive integer) The number of cores to use for computing
summaries for different variables in parallel. Coerced to integer if
possible, otherwise errors. The default is \code{.cores = 1}, in which case no
parallelization is implemented. By default, a socket cluster is used on
Windows and forks otherwise.}
}
\value{
The \code{summarise_draws()} methods return a \link[tibble:tibble]{tibble} data frame.
The first column (\code{"variable"}) contains the variable names and the remaining
columns contain summary statistics and diagnostics.

The functions \code{default_summary_measures()}, \code{default_convergence_measures()},
and \code{default_mcse_measures()} return character vectors of names of the
default measures.
}
\description{
The \code{summarise_draws()} (and \code{summarize_draws()}) methods provide a quick way
to get a table of summary statistics and diagnostics. These methods will
convert an object to a \code{draws} object if it isn't already. For convenience, a
\link[base:summary]{summary()} method for \code{draws} and \code{rvar} objects are also
provided as an alias for \code{summarise_draws()} if the input object is a \code{draws}
or \code{rvar} object.
}
\details{
The default summary functions used are the ones specified by
\code{default_summary_measures()} and \code{default_convergence_measures()}:

\code{default_summary_measures()}
\itemize{
\item \code{\link[=mean]{mean()}}
\item \code{\link[=median]{median()}}
\item \code{\link[=sd]{sd()}}
\item \code{\link[=mad]{mad()}}
\item \code{\link[=quantile2]{quantile2()}}
}

\code{default_convergence_measures()}
\itemize{
\item \code{\link[=rhat]{rhat()}}
\item \code{\link[=ess_bulk]{ess_bulk()}}
\item \code{\link[=ess_tail]{ess_tail()}}
}

The \code{var()} function should not be used to compute variances due
to its inconsistent behavior with matrices. Instead, please use
\code{distributional::variance()}.
}
\examples{
x <- example_draws("eight_schools")
class(x)
str(x)

summarise_draws(x)
summarise_draws(x, "mean", "median")
summarise_draws(x, mean, mcse = mcse_mean)
summarise_draws(x, ~quantile(.x, probs = c(0.4, 0.6)))

# using default_*_meaures()
summarise_draws(x, default_summary_measures())
summarise_draws(x, default_convergence_measures())
summarise_draws(x, default_mcse_measures())

# compute variance of variables
summarise_draws(x, var = distributional::variance)

# illustrate use of '.args'
ws <- rexp(ndraws(x))
summarise_draws(x, weighted.mean, .args = list(w = ws))

# adjust how numerical summaries are printed
summarise_draws(x, .num_args = list(sigfig = 2, notation = "dec"))

}
\seealso{
\code{\link{diagnostics}} for a list of available diagnostics and links to
their individual help pages.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{rhat_basic}
\alias{rhat_basic}
\alias{rhat_basic.default}
\alias{rhat_basic.rvar}
\title{Basic version of the Rhat convergence diagnostic}
\usage{
rhat_basic(x, ...)

\method{rhat_basic}{default}(x, split = TRUE, ...)

\method{rhat_basic}{rvar}(x, split = TRUE, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{split}{(logical) Should the estimate be computed on split chains? The
default is \code{TRUE}.}
}
\value{
If the input is an array, returns a single numeric value. If any of the draws
is non-finite, that is, \code{NA}, \code{NaN}, \code{Inf}, or \code{-Inf}, the returned output
will be (numeric) \code{NA}. Also, if all draws within any of the chains of a
variable are the same (constant), the returned output will be (numeric) \code{NA}
as well. The reason for the latter is that, for constant draws, we cannot
distinguish between variables that are supposed to be constant (e.g., a
diagonal element of a correlation matrix is always 1) or variables that just
happened to be constant because of a failure of convergence or other problems
in the sampling process.

If the input is an \code{\link{rvar}}, returns an array of the same dimensions as the
\code{\link{rvar}}, where each element is equal to the value that would be returned by
passing the draws array for that element of the \code{\link{rvar}} to this function.
}
\description{
Compute the basic Rhat convergence diagnostic for a single variable as
described in Gelman et al. (2013) with some changes according to Vehtari et
al. (2021). For practical applications, we strongly recommend the improved
Rhat convergence diagnostic implemented in \code{\link[=rhat]{rhat()}}.
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
rhat_basic(mu)

d <- as_draws_rvars(example_draws("multi_normal"))
rhat_basic(d$Sigma)

}
\references{
Andrew Gelman, John B. Carlin, Hal S. Stern, David B. Dunson, Aki Vehtari and
Donald B. Rubin (2013). \emph{Bayesian Data Analysis, Third Edition}. Chapman and
Hall/CRC.

Aki Vehtari, Andrew Gelman, Daniel Simpson, Bob Carpenter, and
Paul-Christian Bürkner (2021). Rank-normalization, folding, and
localization: An improved R-hat for assessing convergence of
MCMC (with discussion). \emph{Bayesian Analysis}. 16(2), 667-–718.
doi:10.1214/20-BA1221
}
\seealso{
Other diagnostics: 
\code{\link{ess_basic}()},
\code{\link{ess_bulk}()},
\code{\link{ess_quantile}()},
\code{\link{ess_sd}()},
\code{\link{ess_tail}()},
\code{\link{mcse_mean}()},
\code{\link{mcse_quantile}()},
\code{\link{mcse_sd}()},
\code{\link{pareto_diags}()},
\code{\link{pareto_khat}()},
\code{\link{rhat}()},
\code{\link{rhat_nested}()},
\code{\link{rstar}()}
}
\concept{diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/example_draws.R
\name{example_draws}
\alias{example_draws}
\title{Example \code{draws} objects}
\usage{
example_draws(example = "eight_schools")
}
\arguments{
\item{example}{(string) The name of the example \code{draws} object. See
\strong{Details} for available options.}
}
\value{
A \code{draws} object.
}
\description{
Objects for use in examples, vignettes, and tests.
}
\details{
The following example \code{draws} objects are available.

\strong{eight_schools}: A \code{\link{draws_array}} object with 100 iterations
from each of 4 Markov chains obtained by fitting the eight schools model
described in Gelman et al. (2013) with \href{https://mc-stan.org}{Stan}. The
variables are:
\itemize{
\item \code{mu}: Overall mean of the eight schools
\item \code{tau}: Standard deviation between schools
\item \code{theta}: Individual means of each of the eight schools
}

\strong{multi_normal}: A \code{\link{draws_array}} object with 100 iterations from each of
the 4 Markov chains obtained by fitting a 3-dimensional multivariate normal
model to 100 simulated observations. The variables are:
\itemize{
\item \code{mu}: Mean parameter vector of length 3
\item \code{Sigma}: Covariance matrix of dimension 3 x 3
}
}
\note{
These objects are only intended to be used in demonstrations and tests.
They contain fewer iterations and chains than recommended for performing
actual inference.
}
\examples{
draws_eight_schools <- example_draws("eight_schools")
summarise_draws(draws_eight_schools)

draws_multi_normal <- example_draws("multi_normal")
summarise_draws(draws_multi_normal)

}
\references{
Andrew Gelman, John B. Carlin, Hal S. Stern, David B. Dunson, Aki Vehtari and
Donald B. Rubin (2013). Bayesian Data Analysis, Third Edition. Chapman and
Hall/CRC.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resample_draws.R
\name{resample_draws}
\alias{resample_draws}
\alias{resample_draws.draws}
\alias{resample_draws.rvar}
\title{Resample \code{draws} objects}
\usage{
resample_draws(x, ...)

\method{resample_draws}{draws}(x, weights = NULL, method = "stratified", ndraws = NULL, ...)

\method{resample_draws}{rvar}(x, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{weights}{(numeric vector) A vector of positive weights of length
\code{ndraws(x)}. The weights will be internally normalized. If \code{weights} is not
specified, an attempt will be made to extract any weights already stored in
the draws object (via \code{\link[=weight_draws]{weight_draws()}}). If no weights are stored in the
draws object, equal weight is supplied to each draw. How exactly the
weights influence the resampling depends on the \code{method} argument.}

\item{method}{(string) The resampling method to use:
\itemize{
\item \code{"simple"}: simple random resampling with replacement
\item \code{"simple_no_replace"}: simple random resampling without replacement
\item \code{"stratified"}: stratified resampling with replacement
\item \code{"deterministic"}: deterministic resampling with replacement
}

Currently, \code{"stratified"} is the default as it has comparably low variance
and bias with respect to ideal resampling. The latter would sample perfectly
proportional to the weights, but this is not possible in practice due to the
finite number of draws available. For more details about resampling methods,
see Kitagawa (1996).}

\item{ndraws}{(positive integer) The number of draws to be returned. By
default \code{ndraws} is set internally to the total number of draws in \code{x} if
sensible.}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Resample \code{\link{draws}} objects according to provided weights, for example weights
obtained through importance sampling.
}
\details{
Upon usage of \code{resample_draws()}, chains will automatically be merged
due to subsetting of individual draws (see \code{\link{subset_draws}} for details).
Also, weights stored in the \code{draws} object will be removed in the process,
as resampling invalidates existing weights.
}
\examples{
x <- as_draws_df(example_draws())

# random weights for justr for demonstration
w <- runif(ndraws(x), 0, 10)

# use default stratified sampling
x_rs <- resample_draws(x, weights = w)
summarise_draws(x_rs, default_summary_measures())

# use simple random sampling
x_rs <- resample_draws(x, weights = w, method = "simple")
summarise_draws(x_rs, default_summary_measures())

}
\references{
Kitagawa, G., Monte Carlo Filter and Smoother for Non-Gaussian Nonlinear '
State Space Models, \emph{Journal of Computational and Graphical Statistics},
5(1):1-25, 1996.
}
\seealso{
\code{\link[=resample_draws]{resample_draws()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename_variables.R
\name{rename_variables}
\alias{rename_variables}
\alias{rename_variables.draws}
\title{Rename variables in \code{draws} objects}
\usage{
rename_variables(.x, ...)

\method{rename_variables}{draws}(.x, ...)
}
\arguments{
\item{.x}{(draws) A \code{\link{draws}} object.}

\item{...}{One or more expressions, separated by commas, indicating the
variables to rename. The variable names can be unquoted
(\code{new_name = old_name}) or quoted (\code{"new_name" = "old_name"}). For non-scalar
variables, all elements can be renamed together (\code{"new_name" = "old_name"})
or they can be renamed individually (\code{"new_name[1]" = "old_name[1]"}).}
}
\value{
Returns a \code{\link{draws}} object of the same format as \code{.x}, with variables renamed
according to the expressions provided in \code{...}.
}
\description{
Rename variables in a \code{\link{draws}} object.
}
\examples{
x <- as_draws_df(example_draws())
variables(x)

x <- rename_variables(x, mean = mu, sigma = tau)
variables(x)

x <- rename_variables(x, b = `theta[1]`) # or b  = "theta[1]"
variables(x)

# rename all elements of 'theta' at once
x <- rename_variables(x, alpha = theta)
variables(x)

}
\seealso{
\code{\link{variables}}, \code{\link{mutate_variables}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_smooth.R
\name{pareto_smooth}
\alias{pareto_smooth}
\alias{pareto_smooth.rvar}
\alias{pareto_smooth.default}
\title{Pareto smoothing}
\usage{
pareto_smooth(x, ...)

\method{pareto_smooth}{rvar}(x, return_k = FALSE, extra_diags = FALSE, ...)

\method{pareto_smooth}{default}(
  x,
  tail = c("both", "right", "left"),
  r_eff = NULL,
  ndraws_tail = NULL,
  return_k = FALSE,
  extra_diags = FALSE,
  verbose = TRUE,
  are_log_weights = FALSE,
  ...
)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{return_k}{(logical) Should the Pareto khat be included in
output? If \code{TRUE}, output will be a list containing smoothed
draws and diagnostics, otherwise it will be a numeric of the
smoothed draws. Default is \code{FALSE}.}

\item{extra_diags}{(logical) Should extra Pareto khat diagnostics
be included in output? If \code{TRUE}, \code{min_ss}, \code{khat_threshold} and
\code{convergence_rate} for the estimated k value will be
returned. Default is \code{FALSE}.}

\item{tail}{(string) The tail to diagnose/smooth:
\itemize{
\item \code{"right"}: diagnose/smooth only the right (upper) tail
\item \code{"left"}: diagnose/smooth only the left (lower) tail
\item \code{"both"}: diagnose/smooth both tails and return the maximum k-hat value
}

The default is \code{"both"}.}

\item{r_eff}{(numeric) relative effective sample size estimate. If
\code{r_eff} is NULL, it will be calculated assuming the draws are
from MCMC. Default is NULL.}

\item{ndraws_tail}{(numeric) number of draws for the tail. If
\code{ndraws_tail} is not specified, it will be calculated as
ceiling(3 * sqrt(length(x) / r_eff)) if length(x) > 225 and
length(x) / 5 otherwise (see Appendix H in Vehtari et
al. (2024)).}

\item{verbose}{(logical) Should diagnostic messages be printed? If
\code{TRUE}, messages related to Pareto diagnostics will be
printed. Default is \code{FALSE}.}

\item{are_log_weights}{(logical) Are the draws log weights? Default is
\code{FALSE}. If \code{TRUE} computation will take into account that the
draws are log weights, and only right tail will be smoothed.}
}
\value{
Either a vector \code{x} of smoothed values or a named list
containing the vector \code{x} and a named list \code{diagnostics}
containing numeric values:
\itemize{
\item \code{khat}: estimated Pareto k shape parameter, and optionally
\item \code{min_ss}: minimum sample size for reliable Pareto smoothed
estimate
\item \code{khat_threshold}: sample size specific khat threshold for
reliable Pareto smoothed estimates
\item \code{convergence_rate}: Relative convergence rate for
Pareto smoothed estimates
}

If any of the draws is non-finite, that is, \code{NA}, \code{NaN}, \code{Inf}, or
\code{-Inf}, Pareto smoothing will not be performed, and the original
draws will be returned and and diagnostics will be \code{NA} (numeric).
}
\description{
Smooth the tail draws of x by replacing tail draws by order
statistics of a generalized Pareto distribution fit to the
tail(s). For further details see Vehtari et al. (2024).
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
pareto_smooth(mu)

d <- as_draws_rvars(example_draws("multi_normal"))
pareto_smooth(d$Sigma)
}
\references{
Aki Vehtari, Daniel Simpson, Andrew Gelman, Yuling Yao and
Jonah Gabry (2024). Pareto Smoothed Importance Sampling.
\emph{Journal of Machine Learning Research}, 25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\code{\link{pareto_khat}} for only calculating khat, and
\code{\link{pareto_diags}} for additional diagnostics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rstar.R
\name{rstar}
\alias{rstar}
\title{Calculate R* convergence diagnostic}
\usage{
rstar(
  x,
  split = TRUE,
  uncertainty = FALSE,
  method = "rf",
  hyperparameters = NULL,
  training_proportion = 0.7,
  nsimulations = 1000,
  ...
)
}
\arguments{
\item{x}{(draws) A \code{\link{draws_df}} object or one coercible to a \code{draws_df} object.}

\item{split}{(logical) Should the estimate be computed on split chains? The
default is \code{TRUE}.}

\item{uncertainty}{(logical). Indicates whether to provide a vector of R*
values representing uncertainty in the calculated value (if \code{TRUE}) or a
single value (if \code{FALSE}). The default is \code{TRUE.}}

\item{method}{(string) The machine learning classifier to use (must be
available in the \pkg{caret} package). The default is \code{"rf"}, which calls
the random forest classifier.}

\item{hyperparameters}{(named list) Hyperparameter settings passed to the classifier.
The default for the random forest classifier (\code{method = "rf"}) is
\code{list(mtry = floor(sqt(nvariables(x))))}.
The default for the gradient-based model (\code{method = "gbm"}) is
\code{list(interaction.depth = 3, n.trees = 50, shrinkage = 0.1, n.minobsinnode = 10)}.}

\item{training_proportion}{(positive real) The proportion (in \verb{(0,1)}) of
iterations in used to train the classifier. The default is \code{0.7}.}

\item{nsimulations}{(positive integer) The number of R* values in the
returned vector if \code{uncertainty} is \code{TRUE}. The default is \code{1000.}}

\item{...}{Other arguments passed to \code{caret::train()}.}
}
\value{
A numeric vector of length 1 (by default) or length \code{nsimulations}
(if \code{uncertainty = TRUE}).
}
\description{
The \code{rstar()} function generates a measure of convergence for MCMC draws
based on whether it is possible to determine the Markov chain that generated
a draw with probability greater than chance. To do so, it fits a machine
learning classifier to a training set of MCMC draws and evaluates its
predictive accuracy on a testing set: giving the ratio of accuracy to
predicting a chain uniformly at random.
}
\details{
The \code{rstar()} function provides a measure of MCMC convergence based
on whether it is possible to determine the chain that generated a
particular draw with a probability greater than chance. To do so, it fits a
machine learning classifier to a subset of the original MCMC draws (the
training set) and evaluates its predictive accuracy on the remaining draws
(the testing set). If predictive accuracy exceeds chance (i.e. predicting
the chain that generated a draw uniformly at random), the diagnostic
measure R* will be above 1, indicating that convergence has yet to occur.
This statistic is recently developed, and it is currently unclear what is a
reasonable threshold for diagnosing convergence.

The statistic, R*, is stochastic, meaning that each time the test is run,
unless the random seed is fixed, it will generally produce a different
result. To minimize the implications of this stochasticity, it is
recommended to repeatedly run this function to calculate a distribution of
R*; alternatively, an approximation to this distribution can be obtained by
setting \code{uncertainty = TRUE}, although this approximation of uncertainty
will generally have a lower mean.

By default, a random forest classifier is used (\code{method = "rf"}), which tends
to perform best for target distributions of around 4 dimensions and above.
For lower dimensional targets, gradient boosted models (called via
\code{method = "gbm"}) tend to have a higher classification accuracy. On a given
MCMC sample, it is recommended to try both of these classifiers.
}
\examples{
\donttest{
if (require("caret", quietly = TRUE) && require("randomForest", quietly = TRUE)) {
  x <- example_draws("eight_schools")
  print(rstar(x))
  print(rstar(x, split = FALSE))
  print(rstar(x, method = "gbm"))
  # can pass additional arguments to methods
  print(rstar(x, method = "gbm", verbose = FALSE))

  # with uncertainty, returns a vector of R* values
  hist(rstar(x, uncertainty = TRUE))
  hist(rstar(x, uncertainty = TRUE, nsimulations = 100))

  # can use other classification methods in caret library
  print(rstar(x, method = "knn"))
}
}
}
\references{
Ben Lambert, Aki Vehtari (2020) R*: A robust MCMC convergence
diagnostic with uncertainty using gradient-boosted machines.
\emph{arXiv preprint} \code{arXiv:2003.07900}.
}
\seealso{
Other diagnostics: 
\code{\link{ess_basic}()},
\code{\link{ess_bulk}()},
\code{\link{ess_quantile}()},
\code{\link{ess_sd}()},
\code{\link{ess_tail}()},
\code{\link{mcse_mean}()},
\code{\link{mcse_quantile}()},
\code{\link{mcse_sd}()},
\code{\link{rhat_basic}()},
\code{\link{rhat}()}
}
\concept{diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-factor.R
\name{as_rvar_factor}
\alias{as_rvar_factor}
\alias{as_rvar_ordered}
\title{Coerce to a factor random variable}
\usage{
as_rvar_factor(x, dim = NULL, dimnames = NULL, nchains = NULL, ...)

as_rvar_ordered(x, dim = NULL, dimnames = NULL, nchains = NULL, ...)
}
\arguments{
\item{x}{(multiple options) An object that can be converted to an \code{\link{rvar}},
such as a vector, array, or an \code{\link{rvar}} itself.}

\item{dim}{(integer vector) One or more integers giving the maximal indices
in each dimension to override the dimensions of the \code{\link{rvar}} to be created
(see \code{\link[=dim]{dim()}}). If \code{NULL} (the default), \code{dim} is determined by the input.
\strong{NOTE:} This argument controls the dimensions of the \code{\link{rvar}}, not the
underlying array, so you cannot change the number of draws using this
argument.}

\item{dimnames}{(list) Character vectors giving the names in each dimension
to override the names of the dimensions of the \code{\link{rvar}} to be created (see
\code{\link[=dimnames]{dimnames()}}). If \code{NULL} (the default), this is determined by the input.
\strong{NOTE:} This argument controls the names of the dimensions of the
\code{\link{rvar}}, not the underlying array.}

\item{nchains}{(positive integer) The number of chains. The default is \code{1}.}

\item{...}{
  Arguments passed on to \code{\link[base:factor]{base::factor}}
  \describe{
    \item{\code{levels}}{an optional vector of the unique values (as character strings)
    that \code{x} might have taken.  The default is the unique set of
    values taken by \code{\link[base]{as.character}(x)}, sorted into
    increasing order \emph{of \code{x}}.  Note that this set can be
    specified as smaller than \code{sort(unique(x))}.}
    \item{\code{labels}}{\emph{either} an optional character vector of
    labels for the levels (in the same order as \code{levels} after
    removing those in \code{exclude}), \emph{or} a character string of
    length 1.  Duplicated values in \code{labels} can be used to map
    different values of \code{x} to the same factor level.}
    \item{\code{exclude}}{a vector of values to be excluded when forming the
    set of levels.  This may be factor with the same level set as \code{x}
    or should be a \code{character}.}
    \item{\code{ordered}}{logical flag to determine if the levels should be regarded
    as ordered (in the order given).}
    \item{\code{nmax}}{an upper bound on the number of levels; see \sQuote{Details}.}
  }}
}
\value{
An object of class \code{"rvar_factor"} or \code{"rvar_ordered"} representing a random variable.
}
\description{
Convert \code{x} to an \code{\link{rvar_factor}} or \code{\link{rvar_ordered}} object.
}
\details{
For objects that are already \code{\link{rvar}}s, returns them (with modified dimensions
if \code{dim} is not \code{NULL}), possibly adding levels using the unique values of the draws of
the \code{rvar} (if the object is not already factor-like).

For numeric, logical, factor, or character vectors or arrays, returns an \code{\link{rvar_factor}}
or \code{\link{rvar_ordered}} with a single draw and the same dimensions as \code{x}. This is in contrast
to the \code{\link[=rvar_factor]{rvar_factor()}} and \code{\link[=rvar_ordered]{rvar_ordered()}} constructors, which treats the first dimension
of \code{x} as the draws dimension. As a result, \code{as_rvar_factor()} and \code{as_rvar_ordered()}
are useful for creating constants.
}
\examples{

# You can use as_rvar_factor() to create "constant" rvars (having only one draw):
x <- as_rvar_factor("a")
x

# Such constants can be of arbitrary shape:
as_rvar_factor(letters[1:4])
as_rvar_ordered(matrix(letters[1:10], nrow = 5))
as_rvar_factor(array(letters[1:12], dim = c(2, 3, 2)))

}
\seealso{
\code{\link[=rvar]{rvar()}}, \code{\link[=rvar_factor]{rvar_factor()}}, and \code{\link[=rvar_ordered]{rvar_ordered()}} to construct \code{\link{rvar}}s directly.
See \code{\link[=rdo]{rdo()}}, \code{\link[=rfun]{rfun()}}, and \code{\link[=rvar_rng]{rvar_rng()}} for higher-level interfaces for creating \code{rvar}s.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variables.R
\name{variables<-}
\alias{variables<-}
\alias{variables<-.draws_matrix}
\alias{variables<-.draws_array}
\alias{variables<-.draws_df}
\alias{variables<-.draws_list}
\alias{variables<-.draws_rvars}
\alias{set_variables}
\title{Set variable names in \code{draws} objects}
\usage{
variables(x, ...) <- value

\method{variables}{draws_matrix}(x, with_indices = TRUE, ...) <- value

\method{variables}{draws_array}(x, with_indices = TRUE, ...) <- value

\method{variables}{draws_df}(x, with_indices = TRUE, ...) <- value

\method{variables}{draws_list}(x, with_indices = TRUE, ...) <- value

\method{variables}{draws_rvars}(x, with_indices = FALSE, ...) <- value

set_variables(x, variables, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{value, variables}{(character vector) new variable names.}

\item{with_indices}{(logical) Should indices be included in variable
names? For example, if the object includes variables named \code{"x[1]"} and
\code{"x[2]"}, if \code{TRUE}, \code{c("x[1]", "x[2]")} is returned; if \code{FALSE}, only \code{"x"}
is returned. Defaults to \code{TRUE} for all formats except \code{\link[=draws_rvars]{draws_rvars()}}.}
}
\value{
Returns a \code{\link{draws}} object of the same format as \code{x}, with
variables named as specified.
}
\description{
Set variable names for all variables in a \code{\link{draws}} object. The
\code{set_variables()} form is useful when using pipe operators.
}
\details{
\code{variables(x) <- value} allows you to modify the vector of variable names,
similar to how \code{names(x) <- value} works for vectors and lists. For renaming
specific variables, \code{set_variables(x, value)} works equivalently, but is more intuitive
when using the pipe operator.

For renaming specific variables, \code{\link[=rename_variables]{rename_variables()}} may offer a more
convenient approach.
}
\examples{
x <- example_draws()

variables(x)
nvariables(x)
variables(x) <- letters[1:nvariables(x)]

# or equivalently...
x <- set_variables(x, letters[1:nvariables(x)])

}
\seealso{
\code{\link{variables}}, \code{\link{rename_variables}}, \code{\link{draws-index}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pit.R
\name{pareto_pit}
\alias{pareto_pit}
\alias{pareto_pit.default}
\alias{pareto_pit.draws_matrix}
\alias{pareto_pit.rvar}
\title{Pareto-smoothed probability integral transform}
\usage{
pareto_pit(x, y, ...)

\method{pareto_pit}{default}(x, y, weights = NULL, log = FALSE, ndraws_tail = NULL, ...)

\method{pareto_pit}{draws_matrix}(x, y, weights = NULL, log = FALSE, ndraws_tail = NULL, ...)

\method{pareto_pit}{rvar}(x, y, weights = NULL, log = FALSE, ndraws_tail = NULL, ...)
}
\arguments{
\item{x}{(draws) A \code{\link{draws_matrix}} object or one coercible to a
\code{draws_matrix} object, or an \code{\link{rvar}} object.}

\item{y}{(observations) A 1D vector, or an array of dim(x), if x is \code{rvar}.
Each element of \code{y} corresponds to a variable in \code{x}.}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{weights}{A matrix of weights for each draw and variable. \code{weights}
should have one column per variable in \code{x}, and \code{ndraws(x)} rows.}

\item{log}{(logical) Are the weights passed already on the log scale? The
default is \code{FALSE}, that is, expecting \code{weights} to be on the standard
(non-log) scale.}

\item{ndraws_tail}{(integer) Number of tail draws to use for GPD
fitting. If \code{NULL} (the default), computed using \code{\link[=ps_tail_length]{ps_tail_length()}}.}
}
\value{
A numeric vector of length \code{length(y)} containing the PIT values, or
an array of shape \code{dim(y)}, if \code{x} is an \code{rvar}.
}
\description{
Compute PIT values using the empirical CDF, then refine values in
the tails by fitting a generalized Pareto distribution (GPD) to
the tail draws. This gives smoother, more accurate PIT values in
the tails where the ECDF is coarse, and avoids PIT values of 0 and 1.
Due to use of generalized Pareto distribution CDF in tails, the
PIT values are not anymore rank based and continuous uniformity
test is appropriate.
}
\details{
The function first computes raw PIT values identically to
\code{\link[=pit]{pit()}} (including support for weighted draws). It then fits a
GPD to both tails of the draws (using the same approach as
\code{\link[=pareto_smooth]{pareto_smooth()}}) and replaces PIT values for observations falling in
the tail regions:

For a right-tail observation \eqn{y_i > c_R} (where \eqn{c_R} is
the right-tail cutoff):

\deqn{PIT(y_i) = 1 - p_{tail}(1 - F_{GPD}(y_i; c_R, \sigma_R, k_R))}

For a left-tail observation \eqn{y_i < c_L}:

\deqn{PIT(y_i) = p_{tail}(1 - F_{GPD}(-y_i; -c_L, \sigma_L, k_L))}

where \eqn{p_{tail}} is the proportion of (weighted) mass in the
tail.

When (log-)weights in \code{weights} are provided, they are used for
the raw PIT computation (as in \code{\link[=pit]{pit()}}) and for GPD fit.
}
\examples{
x <- example_draws()
y <- rnorm(nvariables(x), 5, 5)
pareto_pit(x, y)

}
\seealso{
\code{\link[=pit]{pit()}} for the unsmoothed version, \code{\link[=pareto_smooth]{pareto_smooth()}} for
Pareto smoothing of draws.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repair_draws.R
\name{repair_draws}
\alias{repair_draws}
\alias{repair_draws.draws_matrix}
\alias{repair_draws.draws_array}
\alias{repair_draws.draws_df}
\alias{repair_draws.draws_list}
\alias{repair_draws.draws_rvars}
\alias{repair_draws.rvar}
\title{Repair indices of \code{draws} objects}
\usage{
repair_draws(x, order = TRUE, ...)

\method{repair_draws}{draws_matrix}(x, order = TRUE, ...)

\method{repair_draws}{draws_array}(x, order = TRUE, ...)

\method{repair_draws}{draws_df}(x, order = TRUE, ...)

\method{repair_draws}{draws_list}(x, order = TRUE, ...)

\method{repair_draws}{draws_rvars}(x, order = TRUE, ...)

\method{repair_draws}{rvar}(x, order = TRUE, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{order}{(logical) Should draws be ordered (via \code{\link[=order_draws]{order_draws()}}) before
repairing indices? Defaults to \code{TRUE}.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Repair indices of \code{draws} objects so that iterations, chains, and draws
are continuously and consistently numbered.
}
\examples{
x <- as_draws_array(example_draws())
(x <- x[10:5, 3:4, ])
repair_draws(x)

}
\seealso{
\code{\link[=order_draws]{order_draws()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-rfun.R
\name{rfun}
\alias{rfun}
\title{Create functions of random variables}
\usage{
rfun(.f, rvar_args = NULL, ndraws = NULL)
}
\arguments{
\item{.f}{(multiple options) A function to turn into a function that accepts
and/or produces random variables:
\itemize{
\item A function
\item A one-sided formula that can be parsed by \code{\link[rlang:as_function]{rlang::as_function()}}
}}

\item{rvar_args}{(character vector) The names of the arguments of \code{.f} that
should be allowed to accept \code{\link{rvar}}s as arguments. If \code{NULL} (the
default), all arguments to \code{.f} are turned into arguments that accept
\code{\link{rvar}}s.}

\item{ndraws}{(positive integer). The number of draws used to construct new
random variables if no \code{\link{rvar}}s are supplied as arguments to the
returned function. If \code{NULL}, \code{getOption("posterior.rvar_ndraws")} is used
(default \code{4000}). If any arguments to the returned function contain
\code{\link{rvar}}s, the number of draws in the provided \code{\link{rvar}}s is used instead of
the value of this argument.}
}
\value{
A function with the same argument specification as \code{.f}, but which can accept and return
\code{\link{rvar}}s.
}
\description{
Function that create functions that can accept and/or produce \code{\link{rvar}}s.
}
\details{
This function wraps an existing function (\code{.f}) such that it returns \code{\link{rvar}}s containing
whatever type of data \code{.f} would normally return.

The returned function, when called, executes \code{.f} possibly multiple times, once for each draw of
the \code{\link{rvar}}s passed to it, then returns a new
\code{\link{rvar}} representing the output of those function evaluations. If the arguments contain no \code{\link{rvar}}s,
then \code{.f} will be executed \code{ndraws} times and an \code{\link{rvar}} with that many draws returned.

Functions created by \code{rfun()} are not necessarily \emph{fast} (in fact in some cases they may be very slow), but
they have the advantage of allowing a nearly arbitrary R functions to be executed against \code{\link{rvar}}s
simply by wrapping them with \code{rfun()}. This makes it especially useful as a prototyping
tool. If you create code with \code{rfun()} and it is unacceptably slow for your application,
consider rewriting it using math operations directly on \code{\link{rvar}}s (which should be fast),
using \code{\link[=rvar_rng]{rvar_rng()}}, and/or using operations directly on the arrays that back the \code{\link{rvar}}s
(via \code{\link[=draws_of]{draws_of()}}).
}
\examples{

rvar_norm <- rfun(rnorm)
rvar_gamma <- rfun(rgamma)

mu <- rvar_norm(10, mean = 1:10, sd = 1)
sigma <- rvar_gamma(1, shape = 1, rate = 1)
x <- rvar_norm(10, mu, sigma)
x

}
\seealso{
Other rfun: 
\code{\link{rdo}()},
\code{\link{rvar_rng}()}
}
\concept{rfun}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{mcse_quantile}
\alias{mcse_quantile}
\alias{mcse_quantile.default}
\alias{mcse_quantile.rvar}
\alias{mcse_median}
\title{Monte Carlo standard error for quantiles}
\usage{
mcse_quantile(x, probs = c(0.05, 0.95), ...)

\method{mcse_quantile}{default}(x, probs = c(0.05, 0.95), names = TRUE, ...)

\method{mcse_quantile}{rvar}(x, probs = c(0.05, 0.95), names = TRUE, ...)

mcse_median(x, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{probs}{(numeric vector) Probabilities in \verb{[0, 1]}.}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{names}{(logical) Should the result have a \code{names} attribute? The
default is \code{TRUE}, but use \code{FALSE} for improved speed if there are many
values in \code{probs}.}
}
\value{
If the input is an array,
returns a numeric vector with one element per quantile. If any of the draws is
non-finite, that is, \code{NA}, \code{NaN}, \code{Inf}, or \code{-Inf}, the returned output will
be a vector of (numeric) \code{NA} values. Also, if all draws of a variable are
the same (constant), the returned output will be a vector of (numeric) \code{NA}
values as well. The reason for the latter is that, for constant draws, we
cannot distinguish between variables that are supposed to be constant (e.g.,
a diagonal element of a correlation matrix is always 1) or variables that
just happened to be constant because of a failure of convergence or other
problems in the sampling process.

If the input is an \code{\link{rvar}} and \code{length(probs) == 1}, returns an array of the
same dimensions as the \code{\link{rvar}}, where each element is equal to the value
that would be returned by passing the draws array for that element of the
\code{\link{rvar}} to this function. If \code{length(probs) > 1}, the first dimension of the
result indexes the input probabilities; i.e. the result has dimension
\code{c(length(probs), dim(x))}.
}
\description{
Compute Monte Carlo standard errors for quantile estimates of a
single variable.
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
mcse_quantile(mu, probs = c(0.1, 0.9))

d <- as_draws_rvars(example_draws("multi_normal"))
mcse_quantile(d$mu)

}
\references{
Aki Vehtari, Andrew Gelman, Daniel Simpson, Bob Carpenter, and
Paul-Christian Bürkner (2021). Rank-normalization, folding, and
localization: An improved R-hat for assessing convergence of
MCMC (with discussion). \emph{Bayesian Data Analysis}. 16(2), 667-–718.
doi:10.1214/20-BA1221
}
\seealso{
Other diagnostics: 
\code{\link{ess_basic}()},
\code{\link{ess_bulk}()},
\code{\link{ess_quantile}()},
\code{\link{ess_sd}()},
\code{\link{ess_tail}()},
\code{\link{mcse_mean}()},
\code{\link{mcse_sd}()},
\code{\link{rhat_basic}()},
\code{\link{rhat}()},
\code{\link{rstar}()}
}
\concept{diagnostics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weight_draws.R
\name{weight_draws}
\alias{weight_draws}
\alias{weight_draws.draws_matrix}
\alias{weight_draws.draws_array}
\alias{weight_draws.draws_df}
\alias{weight_draws.draws_list}
\alias{weight_draws.draws_rvars}
\title{Weight \code{draws} objects}
\usage{
weight_draws(x, weights, ...)

\method{weight_draws}{draws_matrix}(x, weights, log = FALSE, ...)

\method{weight_draws}{draws_array}(x, weights, log = FALSE, ...)

\method{weight_draws}{draws_df}(x, weights, log = FALSE, ...)

\method{weight_draws}{draws_list}(x, weights, log = FALSE, ...)

\method{weight_draws}{draws_rvars}(x, weights, log = FALSE, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{weights}{(numeric vector) A vector of weights of length \code{ndraws(x)}.
Weights will be internally stored on the log scale (in a variable called
\code{.log_weight}) and will not be normalized, but normalized (non-log) weights
can be returned via the \code{\link[=weights.draws]{weights.draws()}} method later.}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{log}{(logicla) Are the weights passed already on the log scale? The
default is \code{FALSE}, that is, expecting \code{weights} to be on the standard
(non-log) scale.}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Add weights to \code{\link{draws}} objects, with one weight per draw, for use in
subsequent weighting operations. For reasons of numerical accuracy, weights
are stored in the form of unnormalized log-weights (in a variable called
\code{.log_weight}). See \code{\link[=weights.draws]{weights.draws()}} for details how to extract weights from
\code{draws} objects.
}
\examples{
x <- example_draws()

# sample some random weights for illustration
wts <- rexp(ndraws(x))
head(wts)

# add weights
x <- weight_draws(x, weights = wts)

# extract weights
head(weights(x)) # defaults to normalized weights
head(weights(x, normalize=FALSE)) # recover original weights
head(weights(x, log=TRUE)) # get normalized log-weights

# add weights which are already on the log scale
log_wts <- log(wts)
head(log_wts)

x <- weight_draws(x, weights = log_wts, log = TRUE)
# extract weights
head(weights(x))
head(weights(x, log=TRUE, normalize = FALSE)) # recover original log_wts

}
\seealso{
\code{\link[=weights.draws]{weights.draws()}}, \code{\link[=resample_draws]{resample_draws()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{r_scale}
\alias{r_scale}
\title{Rank values}
\usage{
r_scale(x)
}
\arguments{
\item{x}{(numeric) A scalar, vector, matrix, or array of values.}
}
\value{
A numeric array of ranked values with the same size
and dimension as the input.
}
\description{
Compute ranks for a numeric array, that is, replace each
value by its rank. Average rank for ties are used to conserve the
number of unique values of discrete quantities.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-cast.R
\name{as_rvar}
\alias{as_rvar}
\alias{as_rvar_numeric}
\title{Coerce to a random variable}
\usage{
as_rvar(x, dim = NULL, dimnames = NULL, nchains = NULL)

as_rvar_numeric(x, dim = NULL, dimnames = NULL, nchains = NULL)
}
\arguments{
\item{x}{(multiple options) An object that can be converted to an \code{\link{rvar}},
such as a vector, array, or an \code{\link{rvar}} itself.}

\item{dim}{(integer vector) One or more integers giving the maximal indices
in each dimension to override the dimensions of the \code{\link{rvar}} to be created
(see \code{\link[=dim]{dim()}}). If \code{NULL} (the default), \code{dim} is determined by the input.
\strong{NOTE:} This argument controls the dimensions of the \code{\link{rvar}}, not the
underlying array, so you cannot change the number of draws using this
argument.}

\item{dimnames}{(list) Character vectors giving the names in each dimension
to override the names of the dimensions of the \code{\link{rvar}} to be created (see
\code{\link[=dimnames]{dimnames()}}). If \code{NULL} (the default), this is determined by the input.
\strong{NOTE:} This argument controls the names of the dimensions of the
\code{\link{rvar}}, not the underlying array.}

\item{nchains}{(positive integer) The number of chains. The default is \code{1}.}
}
\value{
An object of class \code{"rvar"} (or one of its subtypes) representing a random variable.
}
\description{
Convert \code{x} to an \code{\link{rvar}} object.
}
\details{
For objects that are already \code{\link{rvar}}s, returns them (with modified dimensions
if \code{dim} is not \code{NULL}).

For numeric or logical vectors or arrays, returns an \code{\link{rvar}} with a single draw and
the same dimensions as \code{x}. This is in contrast to the \code{\link[=rvar]{rvar()}} constructor, which
treats the first dimension of \code{x} as the draws dimension. As a result, \code{as_rvar()}
is useful for creating constants.

While \code{as_rvar()} attempts to pick the most suitable subtype of \code{\link{rvar}} based on the
type of \code{x} (possibly returning a an \code{\link{rvar_factor}} or \code{\link{rvar_ordered}}),
\code{as_rvar_numeric()} always coerces the draws of the output \code{\link{rvar}} to be \code{\link{numeric}},
and always returns a base \code{\link{rvar}}, never a subtype.
}
\examples{

# You can use as_rvar() to create "constant" rvars (having only one draw):
x <- as_rvar(1)
x

# Such constants can be of arbitrary shape:
as_rvar(1:4)
as_rvar(matrix(1:10, nrow = 5))
as_rvar(array(1:12, dim = c(2, 3, 2)))

# as_rvar_numeric() coerces subtypes of rvar to the base rvar type
y <- as_rvar_factor(c("a", "b", "c"))
y
as_rvar_numeric(y)

}
\seealso{
\code{\link[=rvar]{rvar()}} to construct \code{\link{rvar}}s directly.  See \code{\link[=rdo]{rdo()}}, \code{\link[=rfun]{rfun()}}, and
\code{\link[=rvar_rng]{rvar_rng()}} for higher-level interfaces for creating \code{rvar}s.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/discrete-summaries.R
\name{modal_category}
\alias{modal_category}
\alias{modal_category.default}
\alias{modal_category.rvar}
\title{Modal category}
\usage{
modal_category(x)

\method{modal_category}{default}(x)

\method{modal_category}{rvar}(x)
}
\arguments{
\item{x}{(multiple options) A vector to be interpreted as draws from
a categorical distribution, such as:
\itemize{
\item A \link{factor}
\item A \link{numeric} (should be \link{integer} or integer-like)
\item An \link{rvar}, \link{rvar_factor}, or \link{rvar_ordered}
}}
}
\value{
If \code{x} is a \link{factor} or \link{numeric}, returns a length-1 vector containing
the modal value.

If \code{x} is an \link{rvar}, returns an array of the same shape as \code{x}, where each
cell is the modal value of the draws in the corresponding cell of \code{x}.
}
\description{
Modal category of a vector.
}
\details{
Finds the modal category (i.e., most frequent value) in \code{x}. In the case of
ties, returns the first tie.
}
\examples{
x <- factor(c("a","b","b","c","d"))
modal_category(x)

# in the case of ties, the first tie is returned
y <- factor(c("a","c","c","d","d"))
modal_category(y)

# both together, as an rvar
xy <- c(rvar(x), rvar(y))
xy
modal_category(xy)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/discrete-summaries.R
\name{entropy}
\alias{entropy}
\alias{entropy.default}
\alias{entropy.rvar}
\title{Normalized entropy}
\usage{
entropy(x)

\method{entropy}{default}(x)

\method{entropy}{rvar}(x)
}
\arguments{
\item{x}{(multiple options) A vector to be interpreted as draws from
a categorical distribution, such as:
\itemize{
\item A \link{factor}
\item A \link{numeric} (should be \link{integer} or integer-like)
\item An \link{rvar}, \link{rvar_factor}, or \link{rvar_ordered}
}}
}
\value{
If \code{x} is a \link{factor} or \link{numeric}, returns a length-1 numeric vector with a value
between 0 and 1 (inclusive) giving the normalized Shannon entropy of \code{x}.

If \code{x} is an \link{rvar}, returns an array of the same shape as \code{x}, where each
cell is the normalized Shannon entropy of the draws in the corresponding cell of \code{x}.
}
\description{
Normalized entropy, for measuring dispersion in draws from categorical distributions.
}
\details{
Calculates the normalized Shannon entropy of the draws in \code{x}. This value is
the entropy of \code{x} divided by the maximum entropy of a distribution with \code{n}
categories, where \code{n} is \code{length(unique(x))} for numeric vectors and
\code{length(levels(x))} for factors:

\deqn{-\frac{\sum_{i = 1}^{n} p_i \log(p_i)}{\log(n)}}

This scales the output to be between 0 (all probability in one category)
and 1 (uniform). This form of normalized entropy is referred to as
\eqn{H_\mathrm{REL}} in Wilcox (1967).
}
\examples{
set.seed(1234)

levels <- c("a", "b", "c", "d", "e")

# a uniform distribution: high normalized entropy
x <- factor(
  sample(levels, 4000, replace = TRUE, prob = c(0.2, 0.2, 0.2, 0.2, 0.2)),
  levels = levels
)
entropy(x)

# a unimodal distribution: low normalized entropy
y <- factor(
  sample(levels, 4000, replace = TRUE, prob = c(0.95, 0.02, 0.015, 0.01, 0.005)),
  levels = levels
)
entropy(y)

# both together, as an rvar
xy <- c(rvar(x), rvar(y))
xy
entropy(xy)
}
\references{
Allen R. Wilcox (1967). \emph{Indices of Qualitative Variation}
(No. ORNL-TM-1919). Oak Ridge National Lab., Tenn.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convergence.R
\name{quantile2}
\alias{quantile2}
\alias{quantile2.default}
\alias{quantile2.rvar}
\title{Compute Quantiles}
\usage{
quantile2(x, probs = c(0.05, 0.95), na.rm = FALSE, ...)

\method{quantile2}{default}(x, probs = c(0.05, 0.95), na.rm = FALSE, names = TRUE, ...)

\method{quantile2}{rvar}(x, probs = c(0.05, 0.95), na.rm = FALSE, names = TRUE, ...)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{probs}{(numeric vector) Probabilities in \verb{[0, 1]}.}

\item{na.rm}{(logical) Should \code{NA} and \code{NaN} values be removed from \code{x} prior
to computing quantiles? The default is \code{FALSE}.}

\item{...}{Arguments passed to individual methods (if applicable) and then
on to \code{\link[stats:quantile]{stats::quantile()}}.}

\item{names}{(logical) Should the result have a \code{names} attribute? The
default is \code{TRUE}, but use \code{FALSE} for improved speed if there are many
values in \code{probs}.}
}
\value{
A numeric vector of length \code{length(probs)}. If \code{names = TRUE}, it has a
\link{names} attribute with names like \code{"q5"}, \code{"q95"}, etc, based on the values
of \code{probs}.
}
\description{
Compute quantiles of a sample and return them in a format consistent
with other summary functions in the \pkg{posterior} package.
}
\examples{
mu <- extract_variable_matrix(example_draws(), "mu")
quantile2(mu)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draws-index.R
\name{draws-index}
\alias{draws-index}
\alias{iteration_ids}
\alias{chain_ids}
\alias{draw_ids}
\alias{niterations}
\alias{nchains}
\alias{ndraws}
\title{Index \code{draws} objects}
\usage{
iteration_ids(x)

chain_ids(x)

draw_ids(x)

niterations(x)

nchains(x)

ndraws(x)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}
}
\value{
For \code{iteration_ids()}, \code{chain_ids()}, and \code{draw_ids()}, an integer vector.

For \code{niterations()}, \code{nchains()}, and \code{ndraws()}, a scalar integer.
}
\description{
Index iterations, chains, and draws of \code{\link{draws}} objects.
}
\details{
The methods \code{iteration_ids()}, \code{chain_ids()}, and \code{draw_ids()} return
vectors of all iterations, chains, and draws, respectively. In
contrast, the methods \code{niterations()}, \code{nchains()}, and
\code{ndraws()} return the number of variables, iterations, chains, and draws,
respectively.
}
\examples{
x <- example_draws()

iteration_ids(x)
niterations(x)

chain_ids(x)
nchains(x)

draw_ids(x)
ndraws(x)

}
\seealso{
\code{\link{variables}}, \code{\link{rename_variables}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-print.R
\encoding{UTF-8}
\name{print.rvar}
\alias{print.rvar}
\alias{format.rvar}
\alias{str.rvar}
\title{Print or format a random variable}
\usage{
\method{print}{rvar}(
  x,
  ...,
  summary = NULL,
  digits = NULL,
  color = TRUE,
  width = getOption("width")
)

\method{format}{rvar}(x, ..., summary = NULL, digits = NULL, color = FALSE)

\method{str}{rvar}(
  object,
  ...,
  summary = NULL,
  vec.len = NULL,
  indent.str = paste(rep.int(" ", max(0, nest.lev + 1)), collapse = ".."),
  nest.lev = 0,
  give.attr = TRUE
)
}
\arguments{
\item{x, object}{(rvar) The \code{\link{rvar}} to print.}

\item{...}{Further arguments passed to the underlying \code{\link[=print]{print()}} methods.}

\item{summary}{(string) The style of summary to display:
\itemize{
\item \code{"mean_sd"} displays \verb{mean ± sd}
\item \code{"median_mad"} displays \verb{median ± mad}
\item \code{"mode_entropy"} displays \verb{mode <entropy>}, and is used automatically for
\code{\link{rvar_factor}}s. It shows normalized entropy, which ranges from
0 (all probability in one category) to 1 (uniform). See \code{\link[=entropy]{entropy()}}.
\item \code{"mode_dissent"} displays \verb{mode <dissent>}, and is used automatically for
\code{\link{rvar_ordered}}s. It shows Tastle and Wierman's (2007) \emph{dissention}
measure, which ranges from 0 (all probability in one category) through
0.5 (uniform) to 1 (bimodal: all probability split equally between the
first and last category). See \code{\link[=dissent]{dissent()}}.
\item \code{NULL} uses \code{getOption("posterior.rvar_summary")} (default \verb{"mean_sd})
}}

\item{digits}{(nonnegative integer) The minimum number of significant digits
to print. If \code{NULL}, defaults to \code{getOption("posterior.digits", 2)}.}

\item{color}{(logical) Whether or not to use color when formatting the
output. If \code{TRUE}, the \code{\link[pillar:style_subtle]{pillar::style_num()}} functions may be used to
produce strings containing control sequences to produce colored output on
the terminal.}

\item{width}{The maxmimum width used to print out lists of factor levels
for \code{\link{rvar_factor}}s. See \code{\link[=format]{format()}}.}

\item{vec.len}{(nonnegative integer) How many 'first few' elements are
displayed of each vector. If \code{NULL}, defaults to
\code{getOption("str")$vec.len}, which defaults to 4.}

\item{indent.str}{(string) The indentation string to use.}

\item{nest.lev}{(nonnegative integer) Current nesting level in the recursive
calls to \code{str()}.}

\item{give.attr}{(logical) If \code{TRUE} (default), show attributes as sub
structures.}
}
\value{
For \code{print()}, an invisible version of the input object.

For \code{str()}, nothing; i.e. \code{invisible(NULL)}.

For \code{format()}, a character vector of the same dimensions as \code{x} where each
entry is of the form \code{"mean±sd"} or \code{"median±mad"}, depending on the value
of \code{summary}.
}
\description{
Printing and formatting methods for \code{\link{rvar}}s.
}
\details{
\code{print()} and \code{str()} print out \code{\link{rvar}} objects by summarizing each element
in the random variable with either its mean±sd or median±mad, depending on
the value of \code{summary}. Both functions use the \code{format()} implementation for
\code{\link{rvar}} objects under the hood, which returns a character vector in the
mean±sd or median±mad form.
}
\examples{

set.seed(5678)
x = rbind(
  cbind(rvar(rnorm(1000, 1)), rvar(rnorm(1000, 2))),
  cbind(rvar(rnorm(1000, 3)), rvar(rnorm(1000, 4)))
)

print(x)
print(x, summary = "median_mad")

str(x)

format(x)

}
\references{
William J. Tastle, Mark J. Wierman (2007). Consensus and dissention: A measure
of ordinal dispersion. \emph{International Journal of Approximate Reasoning}.
45(3), 531--545.
\doi{10.1016/j.ijar.2006.06.024}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_chains.R
\name{merge_chains}
\alias{merge_chains}
\alias{merge_chains.draws_matrix}
\alias{merge_chains.draws_array}
\alias{merge_chains.draws_df}
\alias{merge_chains.draws_list}
\alias{merge_chains.rvar}
\alias{merge_chains.draws_rvars}
\title{Merge chains of \code{draws} objects}
\usage{
merge_chains(x, ...)

\method{merge_chains}{draws_matrix}(x, ...)

\method{merge_chains}{draws_array}(x, ...)

\method{merge_chains}{draws_df}(x, ...)

\method{merge_chains}{draws_list}(x, ...)

\method{merge_chains}{rvar}(x, ...)

\method{merge_chains}{draws_rvars}(x, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Merge chains of \code{\link{draws}} objects into a single chain. Some operations will
trigger an automatic merging of chains, for example, because chains do not
match between two objects involved in a binary operation. By default, no
warning will be issued when this happens but you can activate one via
\code{options(posterior.warn_on_merge_chains = TRUE)}.
}
\examples{
x <- example_draws()

# draws_array with 4 chains, 100 iters each
str(x)

# draws_array with 1 chain of 400 iterations
str(merge_chains(x))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/discrete-summaries.R
\name{dissent}
\alias{dissent}
\alias{dissent.default}
\alias{dissent.rvar}
\title{Dissention}
\usage{
dissent(x)

\method{dissent}{default}(x)

\method{dissent}{rvar}(x)
}
\arguments{
\item{x}{(multiple options) A vector to be interpreted as draws from
an ordinal distribution, such as:
\itemize{
\item A \link{factor}
\item A \link{numeric} (should be \link{integer} or integer-like)
\item An \link{rvar}, \link{rvar_factor}, or \link{rvar_ordered}
}}
}
\value{
If \code{x} is a \link{factor} or \link{numeric}, returns a length-1 numeric vector with a value
between 0 and 1 (inclusive) giving the dissention of \code{x}.

If \code{x} is an \link{rvar}, returns an array of the same shape as \code{x}, where each
cell is the dissention of the draws in the corresponding cell of \code{x}.
}
\description{
Dissention, for measuring dispersion in draws from ordinal distributions.
}
\details{
Calculates Tastle and Wierman's (2007) \emph{dissention} measure:

\deqn{-\sum_{i = 1}^{n} p_i \log_2 \left(1 - \frac{|x_i - \mathrm{E}(x)| }{\max(x) - \min(x)} \right)}

This ranges from 0 (all probability in one category) through 0.5 (uniform) to
1 (bimodal: all probability split equally between the first and last category).
}
\examples{
set.seed(1234)

levels <- c("lowest", "low", "neutral", "high", "highest")

# a bimodal distribution: high dissention
x <- ordered(
  sample(levels, 4000, replace = TRUE, prob = c(0.45, 0.04, 0.02, 0.04, 0.45)),
  levels = levels
)
dissent(x)

# a unimodal distribution: low dissention
y <- ordered(
  sample(levels, 4000, replace = TRUE, prob = c(0.95, 0.02, 0.015, 0.01, 0.005)),
  levels = levels
)
dissent(y)

# both together, as an rvar
xy <- c(rvar(x), rvar(y))
xy
dissent(xy)
}
\references{
William J. Tastle, Mark J. Wierman (2007). Consensus and dissention: A measure
of ordinal dispersion. \emph{International Journal of Approximate Reasoning}.
45(3), 531--545.
\doi{10.1016/j.ijar.2006.06.024}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-dist.R, R/rvar-summaries-over-draws.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{cdf}
\alias{variance}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{distributional}{\code{\link[distributional]{cdf}}, \code{\link[distributional]{variance}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-summaries-within-draws.R
\name{rvar-summaries-within-draws}
\alias{rvar-summaries-within-draws}
\alias{rvar_mean}
\alias{rvar_median}
\alias{rvar_sum}
\alias{rvar_prod}
\alias{rvar_min}
\alias{rvar_max}
\alias{rvar_sd}
\alias{rvar_var}
\alias{rvar_mad}
\alias{rvar_range}
\alias{rvar_quantile}
\alias{rvar_all}
\alias{rvar_any}
\title{Summaries of random variables over array elements, within draws}
\usage{
rvar_mean(..., na.rm = FALSE)

rvar_median(..., na.rm = FALSE)

rvar_sum(..., na.rm = FALSE)

rvar_prod(..., na.rm = FALSE)

rvar_min(..., na.rm = FALSE)

rvar_max(..., na.rm = FALSE)

rvar_sd(..., na.rm = FALSE)

rvar_var(..., na.rm = FALSE)

rvar_mad(..., constant = 1.4826, na.rm = FALSE)

rvar_range(..., na.rm = FALSE)

rvar_quantile(..., probs, names = FALSE, na.rm = FALSE)

rvar_all(..., na.rm = FALSE)

rvar_any(..., na.rm = FALSE)
}
\arguments{
\item{...}{(rvar) One or more \code{\link{rvar}}s.}

\item{na.rm}{(logical) Should \code{NA}s be removed from the input before
summaries are computed? The default is \code{FALSE}.}

\item{constant}{(scalar real) For \code{rvar_mad()}, a scale factor for computing
the median absolute deviation. See the details of \code{stats::mad()} for the
justification for the default value.}

\item{probs}{(numeric vector) For \code{rvar_quantile()}, probabilities in \verb{[0, 1]}.}

\item{names}{(logical) For \code{rvar_quantile()}, if \code{TRUE}, the result has a
\code{names} attribute.}
}
\value{
An \code{\link{rvar}} of length 1 (for \code{range()}, length 2; for \code{quantile()}, length
equal to \code{length(probs)}) with the same number
of draws as the input rvar(s) containing the summary statistic computed within
each draw of the input rvar(s).
}
\description{
Compute summaries of random variables over array elements and within draws,
producing a new random variable of length 1 (except in the case of
\code{rvar_range()}, see \strong{Details}).
}
\details{
These functions compute statistics within each draw of the random variable.
For summaries over draws (such as expectations), see \link{rvar-summaries-over-draws}.

Each function defined here corresponds to the base function of the same name
without the \code{rvar_} prefix (e.g., \code{rvar_mean()} calls \code{mean()} under the hood, etc).
}
\examples{

set.seed(5678)
x = rvar_rng(rnorm, 4, mean = 1:4, sd = 2)

# These will give similar results to mean(1:4),
# median(1:4), sum(1:4), prod(1:4), etc
rvar_mean(x)
rvar_median(x)
rvar_sum(x)
rvar_prod(x)
rvar_range(x)
rvar_quantile(x, probs = c(0.25, 0.5, 0.75), names = TRUE)

}
\seealso{
\link{rvar-summaries-over-draws} for summary functions across draws (e.g. expectations).
\link{rvar-dist} for density, CDF, and quantile functions of random variables.

Other rvar-summaries: 
\code{\link{rvar-summaries-over-draws}},
\code{\link{rvar_is_finite}()}
}
\concept{rvar-summaries}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-.R
\name{rvar_ifelse}
\alias{rvar_ifelse}
\title{Random variable ifelse}
\usage{
rvar_ifelse(test, yes, no)
}
\arguments{
\item{test}{(logical \code{\link{rvar}}, or castable to one) logical test determining
whether the value in \code{yes} or \code{no} is assigned in the corresponding position
of the result.}

\item{yes}{(\code{\link{rvar}}, or castable to one) corresponding values assigned for
entries in \code{test} that are \code{TRUE}.}

\item{no}{(\code{\link{rvar}}, or castable to one) corresponding values assigned for
entries in \code{test} that are \code{FALSE}.}
}
\value{
An \code{\link{rvar}} with the common type of \code{yes} and \code{no} (as determined by
\code{vctrs::vec_cast_common()}) and a shape determined by broadcasting \code{test},
\code{yes}, and \code{no} to a common shape (see the section on broadcasting rules in
\code{vignette("rvar")}). For every element of \code{draws_of(test)}, the corresponding
element of \code{draws_of(yes)} or \code{draws_of(no)} is placed into the result,
depending on whether the element of \code{test} is \code{TRUE} or \code{FALSE}.
}
\description{
A version of \code{ifelse()} that returns an \code{\link{rvar}}.
}
\examples{
x <- rvar(1:4)
y <- rvar(5:8)

i <- rvar(c(TRUE,FALSE,TRUE,FALSE))
z <- rvar_ifelse(i, x, y)
z
draws_of(z)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bind_draws.R
\name{bind_draws}
\alias{bind_draws}
\alias{bind_draws.draws_matrix}
\alias{bind_draws.draws_array}
\alias{bind_draws.draws_df}
\alias{bind_draws.draws_list}
\alias{bind_draws.draws_rvars}
\title{Bind \code{draws} objects together}
\usage{
bind_draws(x, ...)

\method{bind_draws}{draws_matrix}(x, ..., along = "variable")

\method{bind_draws}{draws_array}(x, ..., along = "variable")

\method{bind_draws}{draws_df}(x, ..., along = "variable")

\method{bind_draws}{draws_list}(x, ..., along = "variable")

\method{bind_draws}{draws_rvars}(x, ..., along = "variable")
}
\arguments{
\item{x}{(draws) A \code{\link{draws}} object. The draws format of \code{x} will define the
format of the returned draws object.}

\item{...}{(draws) Additional \code{\link{draws}} objects to bind to \code{x}.}

\item{along}{(string) The dimension along which draws objects should be bound
together. Possible values are \code{"variable"} (the default), \code{"chain"},
\code{"iteration"}, and \code{"draw"}. Not all options are supported for all input
formats.}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Bind multiple \code{\link{draws}} objects together to form a single \code{draws} object.
}
\examples{
x1 <- draws_matrix(alpha = rnorm(5), beta = rnorm(5))
x2 <- draws_matrix(alpha = rnorm(5), beta = rnorm(5))
ndraws(x1)
ndraws(x2)
x3 <- bind_draws(x1, x2, along = "draw")
ndraws(x3)

x4 <- draws_matrix(theta = rexp(5))
x5 <- bind_draws(x1, x4, along = "variable")
variables(x5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.draws_list}
\alias{print.draws_list}
\title{Print \code{draws_list} objects}
\usage{
\method{print}{draws_list}(
  x,
  digits = 2,
  max_iterations = getOption("posterior.max_iterations", 10),
  max_chains = getOption("posterior.max_chains", 2),
  max_variables = getOption("posterior.max_variables", 4),
  reserved = FALSE,
  ...
)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{digits}{(nonnegative integer) The minimum number of significant digits
to print. If \code{NULL}, defaults to \code{getOption("posterior.digits", 2)}.}

\item{max_iterations}{(positive integer) The maximum number of iterations to
print. Can be controlled globally via the \code{"posterior.max_iterations"}
\link[base:options]{option}.}

\item{max_chains}{(positive integer) The maximum number of chains to print.
Can be controlled globally via the \code{"posterior.max_chains"}
\link[base:options]{option}.}

\item{max_variables}{(positive integer) The maximum number of variables to
print. Can be controlled globally via the \code{"posterior.max_variables"}
\link[base:options]{option}.}

\item{reserved}{(logical) Should reserved variables be included in the
output? Defaults to \code{FALSE}. See \code{\link{reserved_variables}} for an overview of
currently reserved variable names.}

\item{...}{Further arguments passed to the underlying \code{\link[=print]{print()}} methods.}
}
\value{
A \code{draws} object of the same class as \code{x}.
}
\description{
Pretty printing for \code{\link{draws_list}} objects.
}
\examples{
x <- as_draws_list(example_draws())
print(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_list_of_variable_arrays.R
\name{extract_list_of_variable_arrays}
\alias{extract_list_of_variable_arrays}
\alias{extract_list_of_variable_arrays.default}
\alias{extract_list_of_variable_arrays.draws}
\title{Extract arrays of multiple variables}
\usage{
extract_list_of_variable_arrays(x, variables = NULL, ...)

\method{extract_list_of_variable_arrays}{default}(x, variables = NULL, ...)

\method{extract_list_of_variable_arrays}{draws}(x, variables = NULL, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{variables}{A character vector of variable names to extract, or NULL to
extract all variables. To extract all dimensions from variables with
indices (e.g. \code{"x[1]"}), provide the base variable names
(e.g. \code{"x"}).}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\value{
A named list of arrays, where each array has dimension
\code{niterations(x)} x \code{nchains(x)} x any remaining
dimensions determined by the indices of the variable \code{x}
(if \code{with_chains = TRUE}) or dimension
\code{niterations(x) * nchains(x)} x any remaining dimensions
(if \code{with_chains = FALSE}).
}
\description{
Extract arrays of draws for multiple variables, returning them as a named list
of arrays. Each array has the same structure as returned by
\code{\link{extract_variable_array}}.
}
\examples{
x <- example_draws(example = "multi_normal")

# Extract multiple variables at once
vars <- extract_list_of_variable_arrays(x, c("mu", "Sigma"))
str(vars)

# Extract all variables (uses base variable names)
all_vars <- extract_list_of_variable_arrays(x)
str(all_vars)

# Extract specific indexed variables
vars2 <- extract_list_of_variable_arrays(x, c("mu[1]", "mu[2]"))
str(vars2)

}
\seealso{
Other variable extraction methods: 
\code{\link{extract_variable}()},
\code{\link{extract_variable_array}()},
\code{\link{extract_variable_matrix}()}
}
\concept{variable extraction methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gpd.R
\name{pgeneralized_pareto}
\alias{pgeneralized_pareto}
\title{Distribution function for the generalized Pareto distribution}
\usage{
pgeneralized_pareto(
  q,
  mu = 0,
  sigma = 1,
  k = 0,
  lower.tail = TRUE,
  log.p = FALSE
)
}
\arguments{
\item{q}{Numeric vector of quantiles.}

\item{mu}{Location parameter.}

\item{sigma}{Scale parameter (must be positive).}

\item{k}{Shape parameter.}

\item{lower.tail}{Logical; if \code{TRUE} (default), probabilities are \code{P[X <= x]}.}

\item{log.p}{Logical; if \code{TRUE}, probabilities are returned as \code{log(p)}.}
}
\value{
A numeric vector of probabilities.
}
\description{
Computes the cumulative distribution function (CDF) for a generalized
Pareto distribution with location \code{mu}, scale \code{sigma}, and shape \code{k}.
}
\examples{
pgeneralized_pareto(q = c(1, 2, 5), mu = 0, sigma = 1, k = 0.2)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-math.R
\name{rvar-matmult}
\alias{rvar-matmult}
\alias{\%**\%}
\title{Matrix multiplication of random variables}
\usage{
x \%**\% y
}
\arguments{
\item{x}{(multiple options) The object to be postmultiplied by \code{y}:
\itemize{
\item An \code{\link{rvar}}
\item A \code{\link{numeric}} vector or matrix
\item A \code{\link{logical}} vector or matrix
}

If a vector is used, it is treated as a \emph{row} vector.}

\item{y}{(multiple options) The object to be premultiplied by \code{x}:
\itemize{
\item An \code{\link{rvar}}
\item A \code{\link{numeric}} vector or matrix
\item A \code{\link{logical}} vector or matrix
}

If a vector is used, it is treated as a \emph{column} vector.}
}
\value{
An \code{\link{rvar}} representing the matrix product of \code{x} and \code{y}.
}
\description{
Matrix multiplication of random variables.
}
\details{
If \code{x} or \code{y} are vectors, they are converted into matrices prior to multiplication, with \code{x}
converted to a row vector and \code{y} to a column vector. Numerics and logicals can be multiplied
by \code{\link{rvar}}s and are broadcasted across all draws of the \code{\link{rvar}} argument. Tensor multiplication
is used to efficiently multiply matrices across draws, so if either \code{x} or \code{y} is an \code{\link{rvar}},
\code{x \%**\% y} will be much faster than \code{rdo(x \%*\% y)}.

Because \code{\link{rvar}} is an S3 class and S3 classes cannot properly override \code{\%*\%}, \code{\link{rvar}}s use
\verb{\%**\%} for matrix multiplication.
}
\examples{

# d has mu (mean vector of length 3) and Sigma (3x3 covariance matrix)
d <- as_draws_rvars(example_draws("multi_normal"))
d$Sigma

# trivial example: multiplication by a non-random matrix
d$Sigma \%**\% diag(1:3)

# Decompose Sigma into R s.t. R'R = Sigma ...
R <- chol(d$Sigma)
# ... and recreate Sigma using matrix multiplication
t(R) \%**\% R

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rvar-cast.R
\name{is_rvar}
\alias{is_rvar}
\title{Is \code{x} a random variable?}
\usage{
is_rvar(x)
}
\arguments{
\item{x}{(any object) An object to test.}
}
\value{
\code{TRUE} if \code{x} is an \code{\link{rvar}}, \code{FALSE} otherwise.
}
\description{
Test if \code{x} is an \code{\link{rvar}}.
}
\seealso{
\code{\link[=as_rvar]{as_rvar()}} to convert objects to \code{rvar}s.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_draws_list.R
\name{draws_list}
\alias{draws_list}
\alias{as_draws_list}
\alias{as_draws_list.default}
\alias{as_draws_list.draws_list}
\alias{as_draws_list.draws_matrix}
\alias{as_draws_list.draws_array}
\alias{as_draws_list.draws_df}
\alias{as_draws_list.draws_rvars}
\alias{as_draws_list.mcmc}
\alias{as_draws_list.mcmc.list}
\alias{is_draws_list}
\title{The \code{draws_list} format}
\usage{
as_draws_list(x, ...)

\method{as_draws_list}{default}(x, ...)

\method{as_draws_list}{draws_list}(x, ...)

\method{as_draws_list}{draws_matrix}(x, ...)

\method{as_draws_list}{draws_array}(x, ...)

\method{as_draws_list}{draws_df}(x, ...)

\method{as_draws_list}{draws_rvars}(x, ...)

\method{as_draws_list}{mcmc}(x, ...)

\method{as_draws_list}{mcmc.list}(x, ...)

draws_list(..., .nchains = 1)

is_draws_list(x)
}
\arguments{
\item{x}{An object to convert to a \code{draws_list} object.}

\item{...}{For \code{as_draws_list()}:
Arguments passed to individual methods (if applicable).
For \code{draws_list()}: Named
arguments containing numeric vectors each defining a separate variable.}

\item{.nchains}{(positive integer) The number of chains. The default is \code{1}.}
}
\value{
A \code{draws_list} object, which has classes
\code{c("draws_list", "draws", "list")}.
}
\description{
The \code{as_draws_list()} methods convert
objects to the \code{draws_list} format.
The \code{draws_list()} function creates an object of the
\code{draws_list} format based on a set of numeric vectors.
See \strong{Details}.
}
\details{
Objects of class \code{"draws_list"} are lists with one element per MCMC
chain. Each of these elements is itself a named list of numeric vectors
with one vector per variable. The length of each vector is equal to the
number of saved iterations per chain. See \strong{Examples}.
}
\examples{
x1 <- as_draws_list(example_draws())
class(x1)
print(x1)
str(x1)

x2 <- draws_list(a = rnorm(10), b = rnorm(10), c = 1)
class(x2)
print(x2)
str(x2)
}
\seealso{
Other formats: 
\code{\link{draws_array}()},
\code{\link{draws_df}()},
\code{\link{draws_matrix}()},
\code{\link{draws_rvars}()},
\code{\link{draws}}
}
\concept{formats}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variables.R
\name{variables}
\alias{variables}
\alias{variables.draws_matrix}
\alias{variables.draws_array}
\alias{variables.draws_df}
\alias{variables.draws_list}
\alias{variables.draws_rvars}
\alias{nvariables}
\title{Get variable names from \code{draws} objects}
\usage{
variables(x, ...)

\method{variables}{draws_matrix}(x, reserved = FALSE, with_indices = TRUE, ...)

\method{variables}{draws_array}(x, reserved = FALSE, with_indices = TRUE, ...)

\method{variables}{draws_df}(x, reserved = FALSE, with_indices = TRUE, ...)

\method{variables}{draws_list}(x, reserved = FALSE, with_indices = TRUE, ...)

\method{variables}{draws_rvars}(x, reserved = FALSE, with_indices = FALSE, ...)

nvariables(x, ...)
}
\arguments{
\item{x}{(draws) A \code{draws} object or another \R object for which the method
is defined.}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{reserved}{(logical) Should reserved variables be included in the
output? Defaults to \code{FALSE}. See \code{\link{reserved_variables}} for an overview of
currently reserved variable names.}

\item{with_indices}{(logical) Should indices be included in variable
names? For example, if the object includes variables named \code{"x[1]"} and
\code{"x[2]"}, if \code{TRUE}, \code{c("x[1]", "x[2]")} is returned; if \code{FALSE}, only \code{"x"}
is returned. Defaults to \code{TRUE} for all formats except \code{\link[=draws_rvars]{draws_rvars()}}.}
}
\value{
For \code{variables()}, a character vector.

For \code{nvariables()}, a scalar integer.
}
\description{
Get variable names from \code{\link{draws}} objects.
}
\details{
\code{variables()} returns a vector of all variable names, and \code{nvariables()}
returns the number of variables.
}
\examples{
x <- example_draws()

variables(x)
nvariables(x)
variables(x) <- letters[1:nvariables(x)]
}
\seealso{
\code{\link{variables<-}}, \code{\link{rename_variables}}, \code{\link{draws-index}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gpd.R
\name{gpdfit}
\alias{gpdfit}
\title{Estimate parameters of the Generalized Pareto distribution}
\usage{
gpdfit(x, wip = TRUE, min_grid_pts = 30, sort_x = TRUE, weights = NULL)
}
\arguments{
\item{x}{A numeric vector. The sample from which to estimate the
parameters.}

\item{wip}{Logical indicating whether to adjust \eqn{k} based on a
weakly informative Gaussian prior centered on 0.5. Defaults to
\code{TRUE}.}

\item{min_grid_pts}{The minimum number of grid points used in the
fitting algorithm. The actual number used is \code{min_grid_pts + floor(sqrt(length(x)))}.}

\item{sort_x}{If \code{TRUE} (the default), the first step in the
fitting algorithm is to sort the elements of \code{x}. If \code{x} is
already sorted in ascending order then \code{sort_x} can be set to
\code{FALSE} to skip the initial sorting step.}

\item{weights}{An optional numeric vector of positive weights the same
length as \code{x}. If \code{NULL} (the default), all observations are
weighted equally and the result is identical to the unweighted fit.
Weights are normalized internally to sum to \code{length(x)}.}
}
\value{
A named list with components \code{k} and \code{sigma}.
}
\description{
Given a sample \eqn{x}, Estimate the parameters \eqn{k} and
\eqn{\sigma} of the generalized Pareto distribution (GPD), assuming
the location parameter is 0. By default the fit uses a prior for
\eqn{k} (this is in addition to the prior described by Zhang and
Stephens, 2009), which will stabilize estimates for very small
sample sizes (and low effective sample sizes in the case of MCMC
samples). The weakly informative prior is a Gaussian prior centered
at 0.5 (see details in Vehtari et al., 2024). This is used
internally but is exported for use by other packages.
}
\details{
Here the parameter \eqn{k} is the negative of \eqn{k} in Zhang &
Stephens (2009).
}
\references{
Zhang, J., and Stephens, M. A. (2009). A new and efficient estimation method
for the generalized Pareto distribution. \emph{Technometrics} \strong{51}, 316-325.
}
\seealso{
Other helper-functions: 
\code{\link{ps_convergence_rate}()},
\code{\link{ps_khat_threshold}()},
\code{\link{ps_min_ss}()},
\code{\link{ps_tail_length}()}
}
\concept{helper-functions}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_smooth.R
\name{ps_tail_length}
\alias{ps_tail_length}
\title{Pareto tail length}
\usage{
ps_tail_length(ndraws, r_eff, ...)
}
\arguments{
\item{ndraws}{number of draws}

\item{r_eff}{relative efficiency}

\item{...}{unused}
}
\value{
tail length
}
\description{
Calculate the tail length from number of draws and relative efficiency
r_eff. See Appendix H in Vehtari et al. (2024). This function is
used internally and is exported to be available for other packages.
}
\seealso{
Other helper-functions: 
\code{\link{ps_convergence_rate}()},
\code{\link{ps_khat_threshold}()},
\code{\link{ps_min_ss}()}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_smooth.R
\name{ps_convergence_rate}
\alias{ps_convergence_rate}
\title{Pareto convergence rate}
\usage{
ps_convergence_rate(k, ndraws, ...)
}
\arguments{
\item{k}{pareto-k values}

\item{ndraws}{number of draws}

\item{...}{unused}
}
\value{
convergence rate
}
\description{
Given number of draws and scalar or array of k's, compute the
relative convergence rate of PSIS estimate RMSE. See Appendix B of
Vehtari et al. (2024). This function is exported to be usable by
other packages. For user-facing diagnostic functions, see
\code{\link{pareto_convergence_rate}} and \code{\link{pareto_diags}}.
}
\seealso{
Other helper-functions: 
\code{\link{ps_khat_threshold}()},
\code{\link{ps_min_ss}()},
\code{\link{ps_tail_length}()}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_smooth.R
\name{ps_min_ss}
\alias{ps_min_ss}
\title{Pareto-smoothing minimum sample-size}
\usage{
ps_min_ss(k, ...)
}
\arguments{
\item{k}{pareto k value}

\item{...}{unused}
}
\value{
minimum sample size
}
\description{
Given Pareto-k computes the minimum sample size for reliable Pareto
smoothed estimate (to have small probability of large error). See
section 3.2.3 in Vehtari et al. (2024). This function is exported
to be usable by other packages. For user-facing diagnostic functions, see
\code{\link{pareto_min_ss}} and \code{\link{pareto_diags}}.
}
\seealso{
Other helper-functions: 
\code{\link{ps_convergence_rate}()},
\code{\link{ps_khat_threshold}()},
\code{\link{ps_tail_length}()}
}
\concept{helper-functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pareto_smooth.R
\name{ps_tail}
\alias{ps_tail}
\title{Pareto smooth tail
function to Pareto smooth the tail of a vector. Exported
for usage in other packages, not by users.}
\usage{
ps_tail(
  x,
  ndraws_tail,
  smooth_draws = TRUE,
  tail = c("right", "left"),
  are_log_weights = FALSE,
  ...
)
}
\arguments{
\item{x}{(multiple options) One of:
\itemize{
\item A matrix of draws for a single variable (iterations x chains). See
\code{\link[=extract_variable_matrix]{extract_variable_matrix()}}.
\item An \code{\link{rvar}}.
}}

\item{ndraws_tail}{(numeric) number of draws for the tail. If
\code{ndraws_tail} is not specified, it will be set to \code{length(x)}.}

\item{smooth_draws}{(logical) Should the tails be smoothed? Default is
\code{TRUE}. If \code{FALSE}, \code{k} will be calculated but \code{x} will remain untouched.}

\item{tail}{(string) The tail to diagnose/smooth:
\itemize{
\item \code{"right"}: diagnose/smooth only the right (upper) tail
\item \code{"left"}: diagnose/smooth only the left (lower) tail
}}

\item{are_log_weights}{(logical) Are the draws log weights? Default is
\code{FALSE}. If \code{TRUE} computation will take into account that the
draws are log weights, and only right tail will be smoothed.}

\item{...}{Arguments passed to individual methods (if applicable).}
}
\description{
Pareto smooth tail
function to Pareto smooth the tail of a vector. Exported
for usage in other packages, not by users.
}
\references{
Aki Vehtari, Daniel Simpson, Andrew Gelman, Yuling Yao and
Jonah Gabry (2024). Pareto Smoothed Importance Sampling.
\emph{Journal of Machine Learning Research}, 25(72):1-58.
\href{https://jmlr.org/papers/v25/19-556.html}{PDF}
}
\seealso{
\code{\link{pareto_smooth}} for the user-facing function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pit.R
\name{pit}
\alias{pit}
\alias{pit.default}
\alias{pit.draws_matrix}
\alias{pit.rvar}
\title{Probability integral transform}
\usage{
pit(x, y, ...)

\method{pit}{default}(x, y, weights = NULL, log = FALSE, ...)

\method{pit}{draws_matrix}(x, y, weights = NULL, log = FALSE, ...)

\method{pit}{rvar}(x, y, weights = NULL, log = FALSE, ...)
}
\arguments{
\item{x}{(draws) A \code{\link{draws_matrix}} object or one coercible to a
\code{draws_matrix} object, or an \code{\link{rvar}} object.}

\item{y}{(observations) A 1D vector, or an array of dim(x), if x is \code{rvar}.
Each element of \code{y} corresponds to a variable in \code{x}.}

\item{...}{Arguments passed to individual methods (if applicable).}

\item{weights}{A matrix of weights for each draw and variable. \code{weights}
should have one column per variable in \code{x}, and \code{ndraws(x)} rows.}

\item{log}{(logical) Are the weights passed already on the log scale? The
default is \code{FALSE}, that is, expecting \code{weights} to be on the standard
(non-log) scale.}
}
\value{
A numeric vector of length \code{length(y)} containing the PIT values, or
an array of shape \code{dim(y)}, if \code{x} is an \code{rvar}.
}
\description{
Probability integral transform (PIT). LOO-PIT is given by a weighted sample.
}
\details{
The \code{pit()} function computes the probability integral transform of
\code{y} using the empirical cumulative distribution computed from the draws
in \code{x}. For continuous valued \code{y} and \code{x}, the PIT for the elements of \code{y}
is computed as the empirical cumulative distribution value:

\if{html}{\out{<div class="sourceCode">}}\preformatted{PIT(y_i) = Pr(x_i < y_i),
}\if{html}{\out{</div>}}

where x_i, is the corresponding set of draws in \code{x}. For \code{draws} objects,
this corresponds to the draws of the \emph{i}th variable, and for \code{rvar}
the elements of \code{y} and \code{x} are matched.

The draws in \code{x} can further be provided (log-)weights in

If \code{y} and \code{x} are discrete, randomization is used to obtain continuous PIT
values. (see, e.g., Czado, C., Gneiting, T., Held, L.: Predictive model
assessment for count data.  Biometrics 65(4), 1254–1261 (2009).)
}
\examples{
# PIT for a draws object
x <- example_draws()
# Create a vector of observations
y <- rnorm(nvariables(x), 5, 5)
pit(x, y)

# Compute weighted PIT (for example LOO-PIT)
weights <- matrix(runif(length(x)), ncol = nvariables(x))

pit(x, y, weights)

# PIT for an rvar
x <- rvar(example_draws())
# Create an array of observations with the same dimensions as x.
y_arr <- array(rnorm(length(x), 5, 5), dim = dim(x))
pit(x, y_arr)

}
